#include <assert.h>

#include <aterm2.h>
#include <deprecated.h>
#include "ParsedError.h"

/**
 * Converts a string to an ATermList of integers (ASCII values). 
 * \param[in] str An ASCII string
 * \return An ATermList containing the ASCII values of #arg as ATermInts
 */
ATerm PERR_stringToChars(const char *str) {
  int len = strlen(str);
  int i;
  ATermList result = ATempty;

  for (i = len - 1; i >= 0; i--) {
    result = ATinsert(result, (ATerm) ATmakeInt(str[i]));
  }

  return (ATerm) result;
}

/**
 * Converts an ASCII char to an ATermInt. 
 * \param[in] ch an ASCII character
 * \return An ATerm representing the ASCII value of #arg
 */
ATerm PERR_byteToChar(char ch) {
    return (ATerm) ATmakeInt(ch);
}

/**
 * Converts a list of integers (ASCII values) to a C string. 
 * \param[in] arg An ATermList with ATermInts, such as [32,32,10]
 * \return String containing the characters from #arg as characters
 */
char *PERR_charsToString(ATerm arg) {
  ATermList list = (ATermList) arg;
  int len = ATgetLength(list);
  int i;
  char *str;

  str = (char *) malloc(len+1);
  if (str == NULL) {
      return NULL;
  }

  for (i = 0; !ATisEmpty(list); list = ATgetNext(list), i++) {
    str[i] = (char) ATgetInt((ATermInt) ATgetFirst(list));
  }
  str[i] = '\0';

  return str;
}

char PERR_charToByte(ATerm arg) {
    return (char) ATgetInt((ATermInt) arg);
}


typedef struct ATerm _PERR_OptLayout;
typedef struct ATerm _PERR_Layout;
typedef struct ATerm _PERR_LexLayoutList;
typedef struct ATerm _PERR_LexStrChar;
typedef struct ATerm _PERR_StrChar;
typedef struct ATerm _PERR_LexStrCon;
typedef struct ATerm _PERR_StrCon;
typedef struct ATerm _PERR_LexStrCharChars;
typedef struct ATerm _PERR_LexNatCon;
typedef struct ATerm _PERR_NatCon;
typedef struct ATerm _PERR_LexLayout;
typedef struct ATerm _PERR_Start;
typedef struct ATerm _PERR_Subject;
typedef struct ATerm _PERR_Error;
typedef struct ATerm _PERR_Summary;
typedef struct ATerm _PERR_SubjectList;
typedef struct ATerm _PERR_ErrorList;
typedef struct ATerm _PERR_Location;
typedef struct ATerm _PERR_Area;
typedef struct ATerm _PERR_Slice;
typedef struct ATerm _PERR_AreaAreas;


/**
 * Initializes the full API. Forgetting to call this function before using the API will lead to strange behaviour. ATinit() needs to be called before this function.
 */
void _PERR_initParsedErrorApi(void) {
  init_ParsedError_dict();

}

/**
 * Protect a PERR_OptLayout from the ATerm garbage collector. Every PERR_OptLayout that is not rooted somewhere on the C call stack must be protected. Examples are global variables
 * \param[in] arg pointer to a PERR_OptLayout
 */
void _PERR_protectOptLayout(PERR_OptLayout *arg) {
  ATprotect((ATerm*)((void*) arg));
}

/**
 * Unprotect a PERR_OptLayout from the ATerm garbage collector. This improves the efficiency of the garbage collector, as well as provide opportunity for reclaiming space
 * \param[in] arg pointer to a PERR_OptLayout
 */
void _PERR_unprotectOptLayout(PERR_OptLayout *arg) {
  ATunprotect((ATerm*)((void*) arg));
}

/**
 * Protect a PERR_Layout from the ATerm garbage collector. Every PERR_Layout that is not rooted somewhere on the C call stack must be protected. Examples are global variables
 * \param[in] arg pointer to a PERR_Layout
 */
void _PERR_protectLayout(PERR_Layout *arg) {
  ATprotect((ATerm*)((void*) arg));
}

/**
 * Unprotect a PERR_Layout from the ATerm garbage collector. This improves the efficiency of the garbage collector, as well as provide opportunity for reclaiming space
 * \param[in] arg pointer to a PERR_Layout
 */
void _PERR_unprotectLayout(PERR_Layout *arg) {
  ATunprotect((ATerm*)((void*) arg));
}

/**
 * Protect a PERR_LexLayoutList from the ATerm garbage collector. Every PERR_LexLayoutList that is not rooted somewhere on the C call stack must be protected. Examples are global variables
 * \param[in] arg pointer to a PERR_LexLayoutList
 */
void _PERR_protectLexLayoutList(PERR_LexLayoutList *arg) {
  ATprotect((ATerm*)((void*) arg));
}

/**
 * Unprotect a PERR_LexLayoutList from the ATerm garbage collector. This improves the efficiency of the garbage collector, as well as provide opportunity for reclaiming space
 * \param[in] arg pointer to a PERR_LexLayoutList
 */
void _PERR_unprotectLexLayoutList(PERR_LexLayoutList *arg) {
  ATunprotect((ATerm*)((void*) arg));
}

/**
 * Protect a PERR_LexStrChar from the ATerm garbage collector. Every PERR_LexStrChar that is not rooted somewhere on the C call stack must be protected. Examples are global variables
 * \param[in] arg pointer to a PERR_LexStrChar
 */
void _PERR_protectLexStrChar(PERR_LexStrChar *arg) {
  ATprotect((ATerm*)((void*) arg));
}

/**
 * Unprotect a PERR_LexStrChar from the ATerm garbage collector. This improves the efficiency of the garbage collector, as well as provide opportunity for reclaiming space
 * \param[in] arg pointer to a PERR_LexStrChar
 */
void _PERR_unprotectLexStrChar(PERR_LexStrChar *arg) {
  ATunprotect((ATerm*)((void*) arg));
}

/**
 * Protect a PERR_StrChar from the ATerm garbage collector. Every PERR_StrChar that is not rooted somewhere on the C call stack must be protected. Examples are global variables
 * \param[in] arg pointer to a PERR_StrChar
 */
void _PERR_protectStrChar(PERR_StrChar *arg) {
  ATprotect((ATerm*)((void*) arg));
}

/**
 * Unprotect a PERR_StrChar from the ATerm garbage collector. This improves the efficiency of the garbage collector, as well as provide opportunity for reclaiming space
 * \param[in] arg pointer to a PERR_StrChar
 */
void _PERR_unprotectStrChar(PERR_StrChar *arg) {
  ATunprotect((ATerm*)((void*) arg));
}

/**
 * Protect a PERR_LexStrCon from the ATerm garbage collector. Every PERR_LexStrCon that is not rooted somewhere on the C call stack must be protected. Examples are global variables
 * \param[in] arg pointer to a PERR_LexStrCon
 */
void _PERR_protectLexStrCon(PERR_LexStrCon *arg) {
  ATprotect((ATerm*)((void*) arg));
}

/**
 * Unprotect a PERR_LexStrCon from the ATerm garbage collector. This improves the efficiency of the garbage collector, as well as provide opportunity for reclaiming space
 * \param[in] arg pointer to a PERR_LexStrCon
 */
void _PERR_unprotectLexStrCon(PERR_LexStrCon *arg) {
  ATunprotect((ATerm*)((void*) arg));
}

/**
 * Protect a PERR_StrCon from the ATerm garbage collector. Every PERR_StrCon that is not rooted somewhere on the C call stack must be protected. Examples are global variables
 * \param[in] arg pointer to a PERR_StrCon
 */
void _PERR_protectStrCon(PERR_StrCon *arg) {
  ATprotect((ATerm*)((void*) arg));
}

/**
 * Unprotect a PERR_StrCon from the ATerm garbage collector. This improves the efficiency of the garbage collector, as well as provide opportunity for reclaiming space
 * \param[in] arg pointer to a PERR_StrCon
 */
void _PERR_unprotectStrCon(PERR_StrCon *arg) {
  ATunprotect((ATerm*)((void*) arg));
}

/**
 * Protect a PERR_LexStrCharChars from the ATerm garbage collector. Every PERR_LexStrCharChars that is not rooted somewhere on the C call stack must be protected. Examples are global variables
 * \param[in] arg pointer to a PERR_LexStrCharChars
 */
void _PERR_protectLexStrCharChars(PERR_LexStrCharChars *arg) {
  ATprotect((ATerm*)((void*) arg));
}

/**
 * Unprotect a PERR_LexStrCharChars from the ATerm garbage collector. This improves the efficiency of the garbage collector, as well as provide opportunity for reclaiming space
 * \param[in] arg pointer to a PERR_LexStrCharChars
 */
void _PERR_unprotectLexStrCharChars(PERR_LexStrCharChars *arg) {
  ATunprotect((ATerm*)((void*) arg));
}

/**
 * Protect a PERR_LexNatCon from the ATerm garbage collector. Every PERR_LexNatCon that is not rooted somewhere on the C call stack must be protected. Examples are global variables
 * \param[in] arg pointer to a PERR_LexNatCon
 */
void _PERR_protectLexNatCon(PERR_LexNatCon *arg) {
  ATprotect((ATerm*)((void*) arg));
}

/**
 * Unprotect a PERR_LexNatCon from the ATerm garbage collector. This improves the efficiency of the garbage collector, as well as provide opportunity for reclaiming space
 * \param[in] arg pointer to a PERR_LexNatCon
 */
void _PERR_unprotectLexNatCon(PERR_LexNatCon *arg) {
  ATunprotect((ATerm*)((void*) arg));
}

/**
 * Protect a PERR_NatCon from the ATerm garbage collector. Every PERR_NatCon that is not rooted somewhere on the C call stack must be protected. Examples are global variables
 * \param[in] arg pointer to a PERR_NatCon
 */
void _PERR_protectNatCon(PERR_NatCon *arg) {
  ATprotect((ATerm*)((void*) arg));
}

/**
 * Unprotect a PERR_NatCon from the ATerm garbage collector. This improves the efficiency of the garbage collector, as well as provide opportunity for reclaiming space
 * \param[in] arg pointer to a PERR_NatCon
 */
void _PERR_unprotectNatCon(PERR_NatCon *arg) {
  ATunprotect((ATerm*)((void*) arg));
}

/**
 * Protect a PERR_LexLayout from the ATerm garbage collector. Every PERR_LexLayout that is not rooted somewhere on the C call stack must be protected. Examples are global variables
 * \param[in] arg pointer to a PERR_LexLayout
 */
void _PERR_protectLexLayout(PERR_LexLayout *arg) {
  ATprotect((ATerm*)((void*) arg));
}

/**
 * Unprotect a PERR_LexLayout from the ATerm garbage collector. This improves the efficiency of the garbage collector, as well as provide opportunity for reclaiming space
 * \param[in] arg pointer to a PERR_LexLayout
 */
void _PERR_unprotectLexLayout(PERR_LexLayout *arg) {
  ATunprotect((ATerm*)((void*) arg));
}

/**
 * Protect a PERR_Start from the ATerm garbage collector. Every PERR_Start that is not rooted somewhere on the C call stack must be protected. Examples are global variables
 * \param[in] arg pointer to a PERR_Start
 */
void _PERR_protectStart(PERR_Start *arg) {
  ATprotect((ATerm*)((void*) arg));
}

/**
 * Unprotect a PERR_Start from the ATerm garbage collector. This improves the efficiency of the garbage collector, as well as provide opportunity for reclaiming space
 * \param[in] arg pointer to a PERR_Start
 */
void _PERR_unprotectStart(PERR_Start *arg) {
  ATunprotect((ATerm*)((void*) arg));
}

/**
 * Protect a PERR_Subject from the ATerm garbage collector. Every PERR_Subject that is not rooted somewhere on the C call stack must be protected. Examples are global variables
 * \param[in] arg pointer to a PERR_Subject
 */
void _PERR_protectSubject(PERR_Subject *arg) {
  ATprotect((ATerm*)((void*) arg));
}

/**
 * Unprotect a PERR_Subject from the ATerm garbage collector. This improves the efficiency of the garbage collector, as well as provide opportunity for reclaiming space
 * \param[in] arg pointer to a PERR_Subject
 */
void _PERR_unprotectSubject(PERR_Subject *arg) {
  ATunprotect((ATerm*)((void*) arg));
}

/**
 * Protect a PERR_Error from the ATerm garbage collector. Every PERR_Error that is not rooted somewhere on the C call stack must be protected. Examples are global variables
 * \param[in] arg pointer to a PERR_Error
 */
void _PERR_protectError(PERR_Error *arg) {
  ATprotect((ATerm*)((void*) arg));
}

/**
 * Unprotect a PERR_Error from the ATerm garbage collector. This improves the efficiency of the garbage collector, as well as provide opportunity for reclaiming space
 * \param[in] arg pointer to a PERR_Error
 */
void _PERR_unprotectError(PERR_Error *arg) {
  ATunprotect((ATerm*)((void*) arg));
}

/**
 * Protect a PERR_Summary from the ATerm garbage collector. Every PERR_Summary that is not rooted somewhere on the C call stack must be protected. Examples are global variables
 * \param[in] arg pointer to a PERR_Summary
 */
void _PERR_protectSummary(PERR_Summary *arg) {
  ATprotect((ATerm*)((void*) arg));
}

/**
 * Unprotect a PERR_Summary from the ATerm garbage collector. This improves the efficiency of the garbage collector, as well as provide opportunity for reclaiming space
 * \param[in] arg pointer to a PERR_Summary
 */
void _PERR_unprotectSummary(PERR_Summary *arg) {
  ATunprotect((ATerm*)((void*) arg));
}

/**
 * Protect a PERR_SubjectList from the ATerm garbage collector. Every PERR_SubjectList that is not rooted somewhere on the C call stack must be protected. Examples are global variables
 * \param[in] arg pointer to a PERR_SubjectList
 */
void _PERR_protectSubjectList(PERR_SubjectList *arg) {
  ATprotect((ATerm*)((void*) arg));
}

/**
 * Unprotect a PERR_SubjectList from the ATerm garbage collector. This improves the efficiency of the garbage collector, as well as provide opportunity for reclaiming space
 * \param[in] arg pointer to a PERR_SubjectList
 */
void _PERR_unprotectSubjectList(PERR_SubjectList *arg) {
  ATunprotect((ATerm*)((void*) arg));
}

/**
 * Protect a PERR_ErrorList from the ATerm garbage collector. Every PERR_ErrorList that is not rooted somewhere on the C call stack must be protected. Examples are global variables
 * \param[in] arg pointer to a PERR_ErrorList
 */
void _PERR_protectErrorList(PERR_ErrorList *arg) {
  ATprotect((ATerm*)((void*) arg));
}

/**
 * Unprotect a PERR_ErrorList from the ATerm garbage collector. This improves the efficiency of the garbage collector, as well as provide opportunity for reclaiming space
 * \param[in] arg pointer to a PERR_ErrorList
 */
void _PERR_unprotectErrorList(PERR_ErrorList *arg) {
  ATunprotect((ATerm*)((void*) arg));
}

/**
 * Protect a PERR_Location from the ATerm garbage collector. Every PERR_Location that is not rooted somewhere on the C call stack must be protected. Examples are global variables
 * \param[in] arg pointer to a PERR_Location
 */
void _PERR_protectLocation(PERR_Location *arg) {
  ATprotect((ATerm*)((void*) arg));
}

/**
 * Unprotect a PERR_Location from the ATerm garbage collector. This improves the efficiency of the garbage collector, as well as provide opportunity for reclaiming space
 * \param[in] arg pointer to a PERR_Location
 */
void _PERR_unprotectLocation(PERR_Location *arg) {
  ATunprotect((ATerm*)((void*) arg));
}

/**
 * Protect a PERR_Area from the ATerm garbage collector. Every PERR_Area that is not rooted somewhere on the C call stack must be protected. Examples are global variables
 * \param[in] arg pointer to a PERR_Area
 */
void _PERR_protectArea(PERR_Area *arg) {
  ATprotect((ATerm*)((void*) arg));
}

/**
 * Unprotect a PERR_Area from the ATerm garbage collector. This improves the efficiency of the garbage collector, as well as provide opportunity for reclaiming space
 * \param[in] arg pointer to a PERR_Area
 */
void _PERR_unprotectArea(PERR_Area *arg) {
  ATunprotect((ATerm*)((void*) arg));
}

/**
 * Protect a PERR_Slice from the ATerm garbage collector. Every PERR_Slice that is not rooted somewhere on the C call stack must be protected. Examples are global variables
 * \param[in] arg pointer to a PERR_Slice
 */
void _PERR_protectSlice(PERR_Slice *arg) {
  ATprotect((ATerm*)((void*) arg));
}

/**
 * Unprotect a PERR_Slice from the ATerm garbage collector. This improves the efficiency of the garbage collector, as well as provide opportunity for reclaiming space
 * \param[in] arg pointer to a PERR_Slice
 */
void _PERR_unprotectSlice(PERR_Slice *arg) {
  ATunprotect((ATerm*)((void*) arg));
}

/**
 * Protect a PERR_AreaAreas from the ATerm garbage collector. Every PERR_AreaAreas that is not rooted somewhere on the C call stack must be protected. Examples are global variables
 * \param[in] arg pointer to a PERR_AreaAreas
 */
void _PERR_protectAreaAreas(PERR_AreaAreas *arg) {
  ATprotect((ATerm*)((void*) arg));
}

/**
 * Unprotect a PERR_AreaAreas from the ATerm garbage collector. This improves the efficiency of the garbage collector, as well as provide opportunity for reclaiming space
 * \param[in] arg pointer to a PERR_AreaAreas
 */
void _PERR_unprotectAreaAreas(PERR_AreaAreas *arg) {
  ATunprotect((ATerm*)((void*) arg));
}

/**
 * Transforms an ATerm to a PERR_OptLayout. This is just a wrapper for a cast, so no structural validation is done!
 * \param[in] t ATerm to be converted
 * \return PERR_OptLayout that was encoded by \arg
 */
PERR_OptLayout _PERR_OptLayoutFromTerm(ATerm t) {
  return (((union {PERR_OptLayout target; ATerm source; })(t)).target);
}

/**
 * Transforms a PERR_OptLayoutto an ATerm. This is just a wrapper for a cast.
 * \param[in] arg PERR_OptLayout to be converted
 * \return ATerm that represents the PERR_OptLayout
 */
ATerm _PERR_OptLayoutToTerm(PERR_OptLayout arg) {
  return (((union {PERR_OptLayout source; ATerm target; })(arg)).target);
}

/**
 * Transforms an ATerm to a PERR_Layout. This is just a wrapper for a cast, so no structural validation is done!
 * \param[in] t ATerm to be converted
 * \return PERR_Layout that was encoded by \arg
 */
PERR_Layout _PERR_LayoutFromTerm(ATerm t) {
  return (((union {PERR_Layout target; ATerm source; })(t)).target);
}

/**
 * Transforms a PERR_Layoutto an ATerm. This is just a wrapper for a cast.
 * \param[in] arg PERR_Layout to be converted
 * \return ATerm that represents the PERR_Layout
 */
ATerm _PERR_LayoutToTerm(PERR_Layout arg) {
  return (((union {PERR_Layout source; ATerm target; })(arg)).target);
}

/**
 * Transforms an ATerm to a PERR_LexLayoutList. This is just a wrapper for a cast, so no structural validation is done!
 * \param[in] t ATerm to be converted
 * \return PERR_LexLayoutList that was encoded by \arg
 */
PERR_LexLayoutList _PERR_LexLayoutListFromTerm(ATerm t) {
  return (((union {PERR_LexLayoutList target; ATerm source; })(t)).target);
}

/**
 * Transforms a PERR_LexLayoutListto an ATerm. This is just a wrapper for a cast.
 * \param[in] arg PERR_LexLayoutList to be converted
 * \return ATerm that represents the PERR_LexLayoutList
 */
ATerm _PERR_LexLayoutListToTerm(PERR_LexLayoutList arg) {
  return (((union {PERR_LexLayoutList source; ATerm target; })(arg)).target);
}

/**
 * Transforms an ATerm to a PERR_LexStrChar. This is just a wrapper for a cast, so no structural validation is done!
 * \param[in] t ATerm to be converted
 * \return PERR_LexStrChar that was encoded by \arg
 */
PERR_LexStrChar _PERR_LexStrCharFromTerm(ATerm t) {
  return (((union {PERR_LexStrChar target; ATerm source; })(t)).target);
}

/**
 * Transforms a PERR_LexStrCharto an ATerm. This is just a wrapper for a cast.
 * \param[in] arg PERR_LexStrChar to be converted
 * \return ATerm that represents the PERR_LexStrChar
 */
ATerm _PERR_LexStrCharToTerm(PERR_LexStrChar arg) {
  return (((union {PERR_LexStrChar source; ATerm target; })(arg)).target);
}

/**
 * Transforms an ATerm to a PERR_StrChar. This is just a wrapper for a cast, so no structural validation is done!
 * \param[in] t ATerm to be converted
 * \return PERR_StrChar that was encoded by \arg
 */
PERR_StrChar _PERR_StrCharFromTerm(ATerm t) {
  return (((union {PERR_StrChar target; ATerm source; })(t)).target);
}

/**
 * Transforms a PERR_StrCharto an ATerm. This is just a wrapper for a cast.
 * \param[in] arg PERR_StrChar to be converted
 * \return ATerm that represents the PERR_StrChar
 */
ATerm _PERR_StrCharToTerm(PERR_StrChar arg) {
  return (((union {PERR_StrChar source; ATerm target; })(arg)).target);
}

/**
 * Transforms an ATerm to a PERR_LexStrCon. This is just a wrapper for a cast, so no structural validation is done!
 * \param[in] t ATerm to be converted
 * \return PERR_LexStrCon that was encoded by \arg
 */
PERR_LexStrCon _PERR_LexStrConFromTerm(ATerm t) {
  return (((union {PERR_LexStrCon target; ATerm source; })(t)).target);
}

/**
 * Transforms a PERR_LexStrConto an ATerm. This is just a wrapper for a cast.
 * \param[in] arg PERR_LexStrCon to be converted
 * \return ATerm that represents the PERR_LexStrCon
 */
ATerm _PERR_LexStrConToTerm(PERR_LexStrCon arg) {
  return (((union {PERR_LexStrCon source; ATerm target; })(arg)).target);
}

/**
 * Transforms an ATerm to a PERR_StrCon. This is just a wrapper for a cast, so no structural validation is done!
 * \param[in] t ATerm to be converted
 * \return PERR_StrCon that was encoded by \arg
 */
PERR_StrCon _PERR_StrConFromTerm(ATerm t) {
  return (((union {PERR_StrCon target; ATerm source; })(t)).target);
}

/**
 * Transforms a PERR_StrConto an ATerm. This is just a wrapper for a cast.
 * \param[in] arg PERR_StrCon to be converted
 * \return ATerm that represents the PERR_StrCon
 */
ATerm _PERR_StrConToTerm(PERR_StrCon arg) {
  return (((union {PERR_StrCon source; ATerm target; })(arg)).target);
}

/**
 * Transforms an ATerm to a PERR_LexStrCharChars. This is just a wrapper for a cast, so no structural validation is done!
 * \param[in] t ATerm to be converted
 * \return PERR_LexStrCharChars that was encoded by \arg
 */
PERR_LexStrCharChars _PERR_LexStrCharCharsFromTerm(ATerm t) {
  return (((union {PERR_LexStrCharChars target; ATerm source; })(t)).target);
}

/**
 * Transforms a PERR_LexStrCharCharsto an ATerm. This is just a wrapper for a cast.
 * \param[in] arg PERR_LexStrCharChars to be converted
 * \return ATerm that represents the PERR_LexStrCharChars
 */
ATerm _PERR_LexStrCharCharsToTerm(PERR_LexStrCharChars arg) {
  return (((union {PERR_LexStrCharChars source; ATerm target; })(arg)).target);
}

/**
 * Transforms an ATerm to a PERR_LexNatCon. This is just a wrapper for a cast, so no structural validation is done!
 * \param[in] t ATerm to be converted
 * \return PERR_LexNatCon that was encoded by \arg
 */
PERR_LexNatCon _PERR_LexNatConFromTerm(ATerm t) {
  return (((union {PERR_LexNatCon target; ATerm source; })(t)).target);
}

/**
 * Transforms a PERR_LexNatConto an ATerm. This is just a wrapper for a cast.
 * \param[in] arg PERR_LexNatCon to be converted
 * \return ATerm that represents the PERR_LexNatCon
 */
ATerm _PERR_LexNatConToTerm(PERR_LexNatCon arg) {
  return (((union {PERR_LexNatCon source; ATerm target; })(arg)).target);
}

/**
 * Transforms an ATerm to a PERR_NatCon. This is just a wrapper for a cast, so no structural validation is done!
 * \param[in] t ATerm to be converted
 * \return PERR_NatCon that was encoded by \arg
 */
PERR_NatCon _PERR_NatConFromTerm(ATerm t) {
  return (((union {PERR_NatCon target; ATerm source; })(t)).target);
}

/**
 * Transforms a PERR_NatConto an ATerm. This is just a wrapper for a cast.
 * \param[in] arg PERR_NatCon to be converted
 * \return ATerm that represents the PERR_NatCon
 */
ATerm _PERR_NatConToTerm(PERR_NatCon arg) {
  return (((union {PERR_NatCon source; ATerm target; })(arg)).target);
}

/**
 * Transforms an ATerm to a PERR_LexLayout. This is just a wrapper for a cast, so no structural validation is done!
 * \param[in] t ATerm to be converted
 * \return PERR_LexLayout that was encoded by \arg
 */
PERR_LexLayout _PERR_LexLayoutFromTerm(ATerm t) {
  return (((union {PERR_LexLayout target; ATerm source; })(t)).target);
}

/**
 * Transforms a PERR_LexLayoutto an ATerm. This is just a wrapper for a cast.
 * \param[in] arg PERR_LexLayout to be converted
 * \return ATerm that represents the PERR_LexLayout
 */
ATerm _PERR_LexLayoutToTerm(PERR_LexLayout arg) {
  return (((union {PERR_LexLayout source; ATerm target; })(arg)).target);
}

/**
 * Transforms an ATerm to a PERR_Start. This is just a wrapper for a cast, so no structural validation is done!
 * \param[in] t ATerm to be converted
 * \return PERR_Start that was encoded by \arg
 */
PERR_Start _PERR_StartFromTerm(ATerm t) {
  return (((union {PERR_Start target; ATerm source; })(t)).target);
}

/**
 * Transforms a PERR_Startto an ATerm. This is just a wrapper for a cast.
 * \param[in] arg PERR_Start to be converted
 * \return ATerm that represents the PERR_Start
 */
ATerm _PERR_StartToTerm(PERR_Start arg) {
  return (((union {PERR_Start source; ATerm target; })(arg)).target);
}

/**
 * Transforms an ATerm to a PERR_Subject. This is just a wrapper for a cast, so no structural validation is done!
 * \param[in] t ATerm to be converted
 * \return PERR_Subject that was encoded by \arg
 */
PERR_Subject _PERR_SubjectFromTerm(ATerm t) {
  return (((union {PERR_Subject target; ATerm source; })(t)).target);
}

/**
 * Transforms a PERR_Subjectto an ATerm. This is just a wrapper for a cast.
 * \param[in] arg PERR_Subject to be converted
 * \return ATerm that represents the PERR_Subject
 */
ATerm _PERR_SubjectToTerm(PERR_Subject arg) {
  return (((union {PERR_Subject source; ATerm target; })(arg)).target);
}

/**
 * Transforms an ATerm to a PERR_Error. This is just a wrapper for a cast, so no structural validation is done!
 * \param[in] t ATerm to be converted
 * \return PERR_Error that was encoded by \arg
 */
PERR_Error _PERR_ErrorFromTerm(ATerm t) {
  return (((union {PERR_Error target; ATerm source; })(t)).target);
}

/**
 * Transforms a PERR_Errorto an ATerm. This is just a wrapper for a cast.
 * \param[in] arg PERR_Error to be converted
 * \return ATerm that represents the PERR_Error
 */
ATerm _PERR_ErrorToTerm(PERR_Error arg) {
  return (((union {PERR_Error source; ATerm target; })(arg)).target);
}

/**
 * Transforms an ATerm to a PERR_Summary. This is just a wrapper for a cast, so no structural validation is done!
 * \param[in] t ATerm to be converted
 * \return PERR_Summary that was encoded by \arg
 */
PERR_Summary _PERR_SummaryFromTerm(ATerm t) {
  return (((union {PERR_Summary target; ATerm source; })(t)).target);
}

/**
 * Transforms a PERR_Summaryto an ATerm. This is just a wrapper for a cast.
 * \param[in] arg PERR_Summary to be converted
 * \return ATerm that represents the PERR_Summary
 */
ATerm _PERR_SummaryToTerm(PERR_Summary arg) {
  return (((union {PERR_Summary source; ATerm target; })(arg)).target);
}

/**
 * Transforms an ATerm to a PERR_SubjectList. This is just a wrapper for a cast, so no structural validation is done!
 * \param[in] t ATerm to be converted
 * \return PERR_SubjectList that was encoded by \arg
 */
PERR_SubjectList _PERR_SubjectListFromTerm(ATerm t) {
  return (((union {PERR_SubjectList target; ATerm source; })(t)).target);
}

/**
 * Transforms a PERR_SubjectListto an ATerm. This is just a wrapper for a cast.
 * \param[in] arg PERR_SubjectList to be converted
 * \return ATerm that represents the PERR_SubjectList
 */
ATerm _PERR_SubjectListToTerm(PERR_SubjectList arg) {
  return (((union {PERR_SubjectList source; ATerm target; })(arg)).target);
}

/**
 * Transforms an ATerm to a PERR_ErrorList. This is just a wrapper for a cast, so no structural validation is done!
 * \param[in] t ATerm to be converted
 * \return PERR_ErrorList that was encoded by \arg
 */
PERR_ErrorList _PERR_ErrorListFromTerm(ATerm t) {
  return (((union {PERR_ErrorList target; ATerm source; })(t)).target);
}

/**
 * Transforms a PERR_ErrorListto an ATerm. This is just a wrapper for a cast.
 * \param[in] arg PERR_ErrorList to be converted
 * \return ATerm that represents the PERR_ErrorList
 */
ATerm _PERR_ErrorListToTerm(PERR_ErrorList arg) {
  return (((union {PERR_ErrorList source; ATerm target; })(arg)).target);
}

/**
 * Transforms an ATerm to a PERR_Location. This is just a wrapper for a cast, so no structural validation is done!
 * \param[in] t ATerm to be converted
 * \return PERR_Location that was encoded by \arg
 */
PERR_Location _PERR_LocationFromTerm(ATerm t) {
  return (((union {PERR_Location target; ATerm source; })(t)).target);
}

/**
 * Transforms a PERR_Locationto an ATerm. This is just a wrapper for a cast.
 * \param[in] arg PERR_Location to be converted
 * \return ATerm that represents the PERR_Location
 */
ATerm _PERR_LocationToTerm(PERR_Location arg) {
  return (((union {PERR_Location source; ATerm target; })(arg)).target);
}

/**
 * Transforms an ATerm to a PERR_Area. This is just a wrapper for a cast, so no structural validation is done!
 * \param[in] t ATerm to be converted
 * \return PERR_Area that was encoded by \arg
 */
PERR_Area _PERR_AreaFromTerm(ATerm t) {
  return (((union {PERR_Area target; ATerm source; })(t)).target);
}

/**
 * Transforms a PERR_Areato an ATerm. This is just a wrapper for a cast.
 * \param[in] arg PERR_Area to be converted
 * \return ATerm that represents the PERR_Area
 */
ATerm _PERR_AreaToTerm(PERR_Area arg) {
  return (((union {PERR_Area source; ATerm target; })(arg)).target);
}

/**
 * Transforms an ATerm to a PERR_Slice. This is just a wrapper for a cast, so no structural validation is done!
 * \param[in] t ATerm to be converted
 * \return PERR_Slice that was encoded by \arg
 */
PERR_Slice _PERR_SliceFromTerm(ATerm t) {
  return (((union {PERR_Slice target; ATerm source; })(t)).target);
}

/**
 * Transforms a PERR_Sliceto an ATerm. This is just a wrapper for a cast.
 * \param[in] arg PERR_Slice to be converted
 * \return ATerm that represents the PERR_Slice
 */
ATerm _PERR_SliceToTerm(PERR_Slice arg) {
  return (((union {PERR_Slice source; ATerm target; })(arg)).target);
}

/**
 * Transforms an ATerm to a PERR_AreaAreas. This is just a wrapper for a cast, so no structural validation is done!
 * \param[in] t ATerm to be converted
 * \return PERR_AreaAreas that was encoded by \arg
 */
PERR_AreaAreas _PERR_AreaAreasFromTerm(ATerm t) {
  return (((union {PERR_AreaAreas target; ATerm source; })(t)).target);
}

/**
 * Transforms a PERR_AreaAreasto an ATerm. This is just a wrapper for a cast.
 * \param[in] arg PERR_AreaAreas to be converted
 * \return ATerm that represents the PERR_AreaAreas
 */
ATerm _PERR_AreaAreasToTerm(PERR_AreaAreas arg) {
  return (((union {PERR_AreaAreas source; ATerm target; })(arg)).target);
}

/**
 * Retrieve the length of a PERR_LexLayoutList. 
 * \param[in] arg input PERR_LexLayoutList
 * \return The number of elements in the PERR_LexLayoutList
 */
int _PERR_getLexLayoutListLength(PERR_LexLayoutList arg) {
  return ATgetLength((ATermList) arg);
}

/**
 * Reverse a PERR_LexLayoutList. 
 * \param[in] arg PERR_LexLayoutList to be reversed
 * \return a reversed #arg
 */
PERR_LexLayoutList _PERR_reverseLexLayoutList(PERR_LexLayoutList arg) {
  return (PERR_LexLayoutList) ATreverse((ATermList) arg);
}

/**
 * Append a PERR_LexLayout to the end of a PERR_LexLayoutList. 
 * \param[in] arg PERR_LexLayoutList to append the PERR_LexLayout to
 * \param[in] elem PERR_LexLayout to be appended
 * \return new PERR_LexLayoutList with #elem appended
 */
PERR_LexLayoutList _PERR_appendLexLayoutList(PERR_LexLayoutList arg, PERR_LexLayout elem) {
  return (PERR_LexLayoutList) ATappend((ATermList) arg, (ATerm) ((ATerm) elem));
}

/**
 * Concatenate two PERR_LexLayoutLists. 
 * \param[in] arg0 first PERR_LexLayoutList
 * \param[in] arg1 second PERR_LexLayoutList
 * \return PERR_LexLayoutList with the elements of #arg0 before the elements of #arg1
 */
PERR_LexLayoutList _PERR_concatLexLayoutList(PERR_LexLayoutList arg0, PERR_LexLayoutList arg1) {
  return (PERR_LexLayoutList) ATconcat((ATermList) arg0, (ATermList) arg1);
}

/**
 * Extract a sublist from a PERR_LexLayoutList. 
 * \param[in] arg PERR_LexLayoutList to extract a slice from
 * \param[in] start inclusive start index of the sublist
 * \param[in] end exclusive end index of the sublist
 * \return new PERR_LexLayoutList with a first element the element at index #start from #arg, and as last element the element at index (#end - 1).
 */
PERR_LexLayoutList _PERR_sliceLexLayoutList(PERR_LexLayoutList arg, int start, int end) {
  return (PERR_LexLayoutList) ATgetSlice((ATermList) arg, start, end);
}

/**
 * Retrieve the PERR_LexLayout at #index from a PERR_LexLayoutList. 
 * \param[in] arg PERR_LexLayoutList to retrieve the PERR_LexLayout from
 * \param[in] index index to use to point in the PERR_LexLayoutList
 * \return PERR_LexLayout at position #index in #arg
 */
PERR_LexLayout _PERR_getLexLayoutListLexLayoutAt(PERR_LexLayoutList arg, int index) {
 return (PERR_LexLayout)ATelementAt((ATermList) arg,index);
}

/**
 * Replace the PERR_LexLayout at #index from a PERR_LexLayoutList by a new one. 
 * \param[in] arg PERR_LexLayoutList to retrieve the PERR_LexLayout from
 * \param[in] elem new PERR_LexLayout to replace another
 * \param[in] index index to use to point in the PERR_LexLayoutList
 * \return A new PERR_LexLayoutListwith #elem replaced in #arg at position #index
 */
PERR_LexLayoutList _PERR_replaceLexLayoutListLexLayoutAt(PERR_LexLayoutList arg, PERR_LexLayout elem, int index) {
 return (PERR_LexLayoutList) ATreplace((ATermList) arg, (ATerm) ((ATerm) elem), index);
}

/**
 * Builds a PERR_LexLayoutList of 2 consecutive elements. 
 * \param[in] elem1 One PERR_LexLayout element of the new PERR_LexLayoutList
 * \param[in] elem2 One PERR_LexLayout element of the new PERR_LexLayoutList
 * \return A new PERR_LexLayoutList consisting of 2 PERR_LexLayouts
 */
PERR_LexLayoutList _PERR_makeLexLayoutList2(PERR_LexLayout elem1, PERR_LexLayout elem2) {
  return (PERR_LexLayoutList) ATmakeList2((ATerm) ((ATerm) elem1), (ATerm) ((ATerm) elem2));
}

/**
 * Builds a PERR_LexLayoutList of 3 consecutive elements. 
 * \param[in] elem1 One PERR_LexLayout element of the new PERR_LexLayoutList
 * \param[in] elem2 One PERR_LexLayout element of the new PERR_LexLayoutList
 * \param[in] elem3 One PERR_LexLayout element of the new PERR_LexLayoutList
 * \return A new PERR_LexLayoutList consisting of 3 PERR_LexLayouts
 */
PERR_LexLayoutList _PERR_makeLexLayoutList3(PERR_LexLayout elem1, PERR_LexLayout elem2, PERR_LexLayout elem3) {
  return (PERR_LexLayoutList) ATmakeList3((ATerm) ((ATerm) elem1), (ATerm) ((ATerm) elem2), (ATerm) ((ATerm) elem3));
}

/**
 * Builds a PERR_LexLayoutList of 4 consecutive elements. 
 * \param[in] elem1 One PERR_LexLayout element of the new PERR_LexLayoutList
 * \param[in] elem2 One PERR_LexLayout element of the new PERR_LexLayoutList
 * \param[in] elem3 One PERR_LexLayout element of the new PERR_LexLayoutList
 * \param[in] elem4 One PERR_LexLayout element of the new PERR_LexLayoutList
 * \return A new PERR_LexLayoutList consisting of 4 PERR_LexLayouts
 */
PERR_LexLayoutList _PERR_makeLexLayoutList4(PERR_LexLayout elem1, PERR_LexLayout elem2, PERR_LexLayout elem3, PERR_LexLayout elem4) {
  return (PERR_LexLayoutList) ATmakeList4((ATerm) ((ATerm) elem1), (ATerm) ((ATerm) elem2), (ATerm) ((ATerm) elem3), (ATerm) ((ATerm) elem4));
}

/**
 * Builds a PERR_LexLayoutList of 5 consecutive elements. 
 * \param[in] elem1 One PERR_LexLayout element of the new PERR_LexLayoutList
 * \param[in] elem2 One PERR_LexLayout element of the new PERR_LexLayoutList
 * \param[in] elem3 One PERR_LexLayout element of the new PERR_LexLayoutList
 * \param[in] elem4 One PERR_LexLayout element of the new PERR_LexLayoutList
 * \param[in] elem5 One PERR_LexLayout element of the new PERR_LexLayoutList
 * \return A new PERR_LexLayoutList consisting of 5 PERR_LexLayouts
 */
PERR_LexLayoutList _PERR_makeLexLayoutList5(PERR_LexLayout elem1, PERR_LexLayout elem2, PERR_LexLayout elem3, PERR_LexLayout elem4, PERR_LexLayout elem5) {
  return (PERR_LexLayoutList) ATmakeList5((ATerm) ((ATerm) elem1), (ATerm) ((ATerm) elem2), (ATerm) ((ATerm) elem3), (ATerm) ((ATerm) elem4), (ATerm) ((ATerm) elem5));
}

/**
 * Builds a PERR_LexLayoutList of 6 consecutive elements. 
 * \param[in] elem1 One PERR_LexLayout element of the new PERR_LexLayoutList
 * \param[in] elem2 One PERR_LexLayout element of the new PERR_LexLayoutList
 * \param[in] elem3 One PERR_LexLayout element of the new PERR_LexLayoutList
 * \param[in] elem4 One PERR_LexLayout element of the new PERR_LexLayoutList
 * \param[in] elem5 One PERR_LexLayout element of the new PERR_LexLayoutList
 * \param[in] elem6 One PERR_LexLayout element of the new PERR_LexLayoutList
 * \return A new PERR_LexLayoutList consisting of 6 PERR_LexLayouts
 */
PERR_LexLayoutList _PERR_makeLexLayoutList6(PERR_LexLayout elem1, PERR_LexLayout elem2, PERR_LexLayout elem3, PERR_LexLayout elem4, PERR_LexLayout elem5, PERR_LexLayout elem6) {
  return (PERR_LexLayoutList) ATmakeList6((ATerm) ((ATerm) elem1), (ATerm) ((ATerm) elem2), (ATerm) ((ATerm) elem3), (ATerm) ((ATerm) elem4), (ATerm) ((ATerm) elem5), (ATerm) ((ATerm) elem6));
}

/**
 * Retrieve the length of a PERR_LexStrCharChars. 
 * \param[in] arg input PERR_LexStrCharChars
 * \return The number of elements in the PERR_LexStrCharChars
 */
int _PERR_getLexStrCharCharsLength(PERR_LexStrCharChars arg) {
  return ATgetLength((ATermList) arg);
}

/**
 * Reverse a PERR_LexStrCharChars. 
 * \param[in] arg PERR_LexStrCharChars to be reversed
 * \return a reversed #arg
 */
PERR_LexStrCharChars _PERR_reverseLexStrCharChars(PERR_LexStrCharChars arg) {
  return (PERR_LexStrCharChars) ATreverse((ATermList) arg);
}

/**
 * Append a PERR_LexStrChar to the end of a PERR_LexStrCharChars. 
 * \param[in] arg PERR_LexStrCharChars to append the PERR_LexStrChar to
 * \param[in] elem PERR_LexStrChar to be appended
 * \return new PERR_LexStrCharChars with #elem appended
 */
PERR_LexStrCharChars _PERR_appendLexStrCharChars(PERR_LexStrCharChars arg, PERR_LexStrChar elem) {
  return (PERR_LexStrCharChars) ATappend((ATermList) arg, (ATerm) ((ATerm) elem));
}

/**
 * Concatenate two PERR_LexStrCharCharss. 
 * \param[in] arg0 first PERR_LexStrCharChars
 * \param[in] arg1 second PERR_LexStrCharChars
 * \return PERR_LexStrCharChars with the elements of #arg0 before the elements of #arg1
 */
PERR_LexStrCharChars _PERR_concatLexStrCharChars(PERR_LexStrCharChars arg0, PERR_LexStrCharChars arg1) {
  return (PERR_LexStrCharChars) ATconcat((ATermList) arg0, (ATermList) arg1);
}

/**
 * Extract a sublist from a PERR_LexStrCharChars. 
 * \param[in] arg PERR_LexStrCharChars to extract a slice from
 * \param[in] start inclusive start index of the sublist
 * \param[in] end exclusive end index of the sublist
 * \return new PERR_LexStrCharChars with a first element the element at index #start from #arg, and as last element the element at index (#end - 1).
 */
PERR_LexStrCharChars _PERR_sliceLexStrCharChars(PERR_LexStrCharChars arg, int start, int end) {
  return (PERR_LexStrCharChars) ATgetSlice((ATermList) arg, start, end);
}

/**
 * Retrieve the PERR_LexStrChar at #index from a PERR_LexStrCharChars. 
 * \param[in] arg PERR_LexStrCharChars to retrieve the PERR_LexStrChar from
 * \param[in] index index to use to point in the PERR_LexStrCharChars
 * \return PERR_LexStrChar at position #index in #arg
 */
PERR_LexStrChar _PERR_getLexStrCharCharsLexStrCharAt(PERR_LexStrCharChars arg, int index) {
 return (PERR_LexStrChar)ATelementAt((ATermList) arg,index);
}

/**
 * Replace the PERR_LexStrChar at #index from a PERR_LexStrCharChars by a new one. 
 * \param[in] arg PERR_LexStrCharChars to retrieve the PERR_LexStrChar from
 * \param[in] elem new PERR_LexStrChar to replace another
 * \param[in] index index to use to point in the PERR_LexStrCharChars
 * \return A new PERR_LexStrCharCharswith #elem replaced in #arg at position #index
 */
PERR_LexStrCharChars _PERR_replaceLexStrCharCharsLexStrCharAt(PERR_LexStrCharChars arg, PERR_LexStrChar elem, int index) {
 return (PERR_LexStrCharChars) ATreplace((ATermList) arg, (ATerm) ((ATerm) elem), index);
}

/**
 * Builds a PERR_LexStrCharChars of 2 consecutive elements. 
 * \param[in] elem1 One PERR_LexStrChar element of the new PERR_LexStrCharChars
 * \param[in] elem2 One PERR_LexStrChar element of the new PERR_LexStrCharChars
 * \return A new PERR_LexStrCharChars consisting of 2 PERR_LexStrChars
 */
PERR_LexStrCharChars _PERR_makeLexStrCharChars2(PERR_LexStrChar elem1, PERR_LexStrChar elem2) {
  return (PERR_LexStrCharChars) ATmakeList2((ATerm) ((ATerm) elem1), (ATerm) ((ATerm) elem2));
}

/**
 * Builds a PERR_LexStrCharChars of 3 consecutive elements. 
 * \param[in] elem1 One PERR_LexStrChar element of the new PERR_LexStrCharChars
 * \param[in] elem2 One PERR_LexStrChar element of the new PERR_LexStrCharChars
 * \param[in] elem3 One PERR_LexStrChar element of the new PERR_LexStrCharChars
 * \return A new PERR_LexStrCharChars consisting of 3 PERR_LexStrChars
 */
PERR_LexStrCharChars _PERR_makeLexStrCharChars3(PERR_LexStrChar elem1, PERR_LexStrChar elem2, PERR_LexStrChar elem3) {
  return (PERR_LexStrCharChars) ATmakeList3((ATerm) ((ATerm) elem1), (ATerm) ((ATerm) elem2), (ATerm) ((ATerm) elem3));
}

/**
 * Builds a PERR_LexStrCharChars of 4 consecutive elements. 
 * \param[in] elem1 One PERR_LexStrChar element of the new PERR_LexStrCharChars
 * \param[in] elem2 One PERR_LexStrChar element of the new PERR_LexStrCharChars
 * \param[in] elem3 One PERR_LexStrChar element of the new PERR_LexStrCharChars
 * \param[in] elem4 One PERR_LexStrChar element of the new PERR_LexStrCharChars
 * \return A new PERR_LexStrCharChars consisting of 4 PERR_LexStrChars
 */
PERR_LexStrCharChars _PERR_makeLexStrCharChars4(PERR_LexStrChar elem1, PERR_LexStrChar elem2, PERR_LexStrChar elem3, PERR_LexStrChar elem4) {
  return (PERR_LexStrCharChars) ATmakeList4((ATerm) ((ATerm) elem1), (ATerm) ((ATerm) elem2), (ATerm) ((ATerm) elem3), (ATerm) ((ATerm) elem4));
}

/**
 * Builds a PERR_LexStrCharChars of 5 consecutive elements. 
 * \param[in] elem1 One PERR_LexStrChar element of the new PERR_LexStrCharChars
 * \param[in] elem2 One PERR_LexStrChar element of the new PERR_LexStrCharChars
 * \param[in] elem3 One PERR_LexStrChar element of the new PERR_LexStrCharChars
 * \param[in] elem4 One PERR_LexStrChar element of the new PERR_LexStrCharChars
 * \param[in] elem5 One PERR_LexStrChar element of the new PERR_LexStrCharChars
 * \return A new PERR_LexStrCharChars consisting of 5 PERR_LexStrChars
 */
PERR_LexStrCharChars _PERR_makeLexStrCharChars5(PERR_LexStrChar elem1, PERR_LexStrChar elem2, PERR_LexStrChar elem3, PERR_LexStrChar elem4, PERR_LexStrChar elem5) {
  return (PERR_LexStrCharChars) ATmakeList5((ATerm) ((ATerm) elem1), (ATerm) ((ATerm) elem2), (ATerm) ((ATerm) elem3), (ATerm) ((ATerm) elem4), (ATerm) ((ATerm) elem5));
}

/**
 * Builds a PERR_LexStrCharChars of 6 consecutive elements. 
 * \param[in] elem1 One PERR_LexStrChar element of the new PERR_LexStrCharChars
 * \param[in] elem2 One PERR_LexStrChar element of the new PERR_LexStrCharChars
 * \param[in] elem3 One PERR_LexStrChar element of the new PERR_LexStrCharChars
 * \param[in] elem4 One PERR_LexStrChar element of the new PERR_LexStrCharChars
 * \param[in] elem5 One PERR_LexStrChar element of the new PERR_LexStrCharChars
 * \param[in] elem6 One PERR_LexStrChar element of the new PERR_LexStrCharChars
 * \return A new PERR_LexStrCharChars consisting of 6 PERR_LexStrChars
 */
PERR_LexStrCharChars _PERR_makeLexStrCharChars6(PERR_LexStrChar elem1, PERR_LexStrChar elem2, PERR_LexStrChar elem3, PERR_LexStrChar elem4, PERR_LexStrChar elem5, PERR_LexStrChar elem6) {
  return (PERR_LexStrCharChars) ATmakeList6((ATerm) ((ATerm) elem1), (ATerm) ((ATerm) elem2), (ATerm) ((ATerm) elem3), (ATerm) ((ATerm) elem4), (ATerm) ((ATerm) elem5), (ATerm) ((ATerm) elem6));
}

/**
 * Retrieve the number of elements in a PERR_SubjectList. 
 * \param[in] arg input PERR_SubjectList
 * \return The number of elements in #arg, excluding any separators
 */
int _PERR_getSubjectListLength(PERR_SubjectList arg) {
  if (ATisEmpty((ATermList) arg)) {
    return 0;
  }
  return (ATgetLength((ATermList) arg) / 4) + 1;
}

/**
 * Reverses the elements of a PERR_SubjectList. Note that separators are reversed with the list, but the order in which each set of separators inbetween two elements occurs does not change
 * \param[in] arg PERR_SubjectList to be reversed
 * \return #arg reversed
 */
PERR_SubjectList PERR_reverseSubjectList(PERR_SubjectList arg) {
  ATermList list = (ATermList) arg;
  ATerm head;
  ATerm sep0;
  ATerm sep1;
  ATerm sep2;
  ATermList result;

 if (ATisEmpty(list) || ATgetLength(list) == 1) {
    return arg;
  }

  result = ATmakeList1(ATgetFirst(list));
      list = ATgetNext(list);
  sep0 = ATgetFirst(list);
  list = ATgetNext(list);
  sep1 = ATgetFirst(list);
  list = ATgetNext(list);
  sep2 = ATgetFirst(list);
  list = ATgetNext(list);

  while (!ATisEmpty(list)) {
    result = ATinsert(result, sep2);
    result = ATinsert(result, sep1);
    result = ATinsert(result, sep0);

   head = ATgetFirst(list);
   result = ATinsert(result, head);
    list = ATgetNext(list);

   if (!ATisEmpty(list)) {
  sep0 = ATgetFirst(list);
  list = ATgetNext(list);
  sep1 = ATgetFirst(list);
  list = ATgetNext(list);
  sep2 = ATgetFirst(list);
  list = ATgetNext(list);
   }
  }

  return (PERR_SubjectList) result;
}

/**
 * Append a PERR_Subject to the end of a PERR_SubjectList. 
 * \param[in] arg PERR_SubjectList to append the PERR_Subject to
 * \param[in] elem PERR_Subject to be appended
 * \return new PERR_SubjectList with #elem appended after the separators
 */
PERR_SubjectList PERR_appendSubjectList(PERR_SubjectList arg0, PERR_OptLayout wsAfterHead, PERR_OptLayout wsAfterSep, PERR_Subject arg1) {
  return PERR_concatSubjectList(arg0, wsAfterHead, wsAfterSep, PERR_makeSubjectListSingle(arg1));
}

/**
 * Concatenate two PERR_SubjectLists. 
 * \param[in] arg0 first PERR_SubjectList
 * \param[in] arg1 second PERR_SubjectList
 * \return PERR_SubjectList with the elements of #arg0 before the elements of #arg1, with the separators in between.
 */
PERR_SubjectList PERR_concatSubjectList(PERR_SubjectList arg0, PERR_OptLayout wsAfterHead, PERR_OptLayout wsAfterSep, PERR_SubjectList arg1) {
  if (ATisEmpty((ATermList) arg0)) {
    return arg1;
  }
  arg1 = PERR_makeSubjectListMany((PERR_Subject)ATgetFirst((ATermList) arg0), wsAfterHead, wsAfterSep,  arg1);
  arg1 = (PERR_SubjectList) ATgetNext((ATermList) arg1);
  return (PERR_SubjectList) ATconcat((ATermList) arg0, (ATermList) arg1);
}

/**
 * Extract a sublist from a PERR_SubjectList. 
 * \param[in] arg PERR_SubjectList to extract a slice from
 * \param[in] start inclusive start index of the sublist
 * \param[in] end exclusive end index of the sublist
 * \return new PERR_SubjectList with a first element the element at index #start from #arg, and as last element the element at index (#end - 1).
 */
PERR_SubjectList _PERR_sliceSubjectList(PERR_SubjectList arg, int start, int end) {
  return (PERR_SubjectList) ATgetSlice((ATermList) arg, start * 4, end * 4);
}

/**
 * Retrieve the PERR_Subject at #index from a PERR_SubjectList. 
 * \param[in] arg PERR_SubjectList to retrieve the PERR_Subject from
 * \param[in] index index to use to point in the PERR_SubjectList
 * \return PERR_Subject at position #index in #arg
 */
PERR_Subject _PERR_getSubjectListSubjectAt(PERR_SubjectList arg, int index) {
 return (PERR_Subject)ATelementAt((ATermList) arg,index * 4);
}

/**
 * Replace the PERR_Subject at #index from a PERR_SubjectList by a new one. 
 * \param[in] arg PERR_SubjectList to retrieve the PERR_Subject from
 * \param[in] elem new PERR_Subject to replace another
 * \param[in] index index to use to point in the PERR_SubjectList
 * \return A new PERR_SubjectListwith #elem replaced in #arg at position #index
 */
PERR_SubjectList _PERR_replaceSubjectListSubjectAt(PERR_SubjectList arg, PERR_Subject elem, int index) {
 return (PERR_SubjectList) ATreplace((ATermList) arg, (ATerm) ((ATerm) elem), index * 4);
}

/**
 * Builds a PERR_SubjectList of 2 consecutive elements. The elements are separated.
 * \param[in] elem1 One PERR_Subject element of the new PERR_SubjectList
 * \param[in] elem2 One PERR_Subject element of the new PERR_SubjectList
 * \return A new PERR_SubjectList consisting of 2 PERR_Subjects
 */
PERR_SubjectList PERR_makeSubjectList2(PERR_OptLayout wsAfterHead, PERR_OptLayout wsAfterSep, PERR_Subject elem1, PERR_Subject elem2) {
  return PERR_makeSubjectListMany(elem1, wsAfterHead, wsAfterSep, PERR_makeSubjectListSingle(elem2));
}

/**
 * Builds a PERR_SubjectList of 3 consecutive elements. The elements are separated.
 * \param[in] elem1 One PERR_Subject element of the new PERR_SubjectList
 * \param[in] elem2 One PERR_Subject element of the new PERR_SubjectList
 * \param[in] elem3 One PERR_Subject element of the new PERR_SubjectList
 * \return A new PERR_SubjectList consisting of 3 PERR_Subjects
 */
PERR_SubjectList PERR_makeSubjectList3(PERR_OptLayout wsAfterHead, PERR_OptLayout wsAfterSep, PERR_Subject elem1, PERR_Subject elem2, PERR_Subject elem3) {
  return PERR_makeSubjectListMany(elem1, wsAfterHead, wsAfterSep, PERR_makeSubjectList2(wsAfterHead, wsAfterSep, elem2, elem3));
}

/**
 * Builds a PERR_SubjectList of 4 consecutive elements. The elements are separated.
 * \param[in] elem1 One PERR_Subject element of the new PERR_SubjectList
 * \param[in] elem2 One PERR_Subject element of the new PERR_SubjectList
 * \param[in] elem3 One PERR_Subject element of the new PERR_SubjectList
 * \param[in] elem4 One PERR_Subject element of the new PERR_SubjectList
 * \return A new PERR_SubjectList consisting of 4 PERR_Subjects
 */
PERR_SubjectList PERR_makeSubjectList4(PERR_OptLayout wsAfterHead, PERR_OptLayout wsAfterSep, PERR_Subject elem1, PERR_Subject elem2, PERR_Subject elem3, PERR_Subject elem4) {
  return PERR_makeSubjectListMany(elem1, wsAfterHead, wsAfterSep, PERR_makeSubjectList3(wsAfterHead, wsAfterSep, elem2, elem3, elem4));
}

/**
 * Builds a PERR_SubjectList of 5 consecutive elements. The elements are separated.
 * \param[in] elem1 One PERR_Subject element of the new PERR_SubjectList
 * \param[in] elem2 One PERR_Subject element of the new PERR_SubjectList
 * \param[in] elem3 One PERR_Subject element of the new PERR_SubjectList
 * \param[in] elem4 One PERR_Subject element of the new PERR_SubjectList
 * \param[in] elem5 One PERR_Subject element of the new PERR_SubjectList
 * \return A new PERR_SubjectList consisting of 5 PERR_Subjects
 */
PERR_SubjectList PERR_makeSubjectList5(PERR_OptLayout wsAfterHead, PERR_OptLayout wsAfterSep, PERR_Subject elem1, PERR_Subject elem2, PERR_Subject elem3, PERR_Subject elem4, PERR_Subject elem5) {
  return PERR_makeSubjectListMany(elem1, wsAfterHead, wsAfterSep, PERR_makeSubjectList4(wsAfterHead, wsAfterSep, elem2, elem3, elem4, elem5));
}

/**
 * Builds a PERR_SubjectList of 6 consecutive elements. The elements are separated.
 * \param[in] elem1 One PERR_Subject element of the new PERR_SubjectList
 * \param[in] elem2 One PERR_Subject element of the new PERR_SubjectList
 * \param[in] elem3 One PERR_Subject element of the new PERR_SubjectList
 * \param[in] elem4 One PERR_Subject element of the new PERR_SubjectList
 * \param[in] elem5 One PERR_Subject element of the new PERR_SubjectList
 * \param[in] elem6 One PERR_Subject element of the new PERR_SubjectList
 * \return A new PERR_SubjectList consisting of 6 PERR_Subjects
 */
PERR_SubjectList PERR_makeSubjectList6(PERR_OptLayout wsAfterHead, PERR_OptLayout wsAfterSep, PERR_Subject elem1, PERR_Subject elem2, PERR_Subject elem3, PERR_Subject elem4, PERR_Subject elem5, PERR_Subject elem6) {
  return PERR_makeSubjectListMany(elem1, wsAfterHead, wsAfterSep, PERR_makeSubjectList5(wsAfterHead, wsAfterSep, elem2, elem3, elem4, elem5, elem6));
}

/**
 * Retrieve the number of elements in a PERR_ErrorList. 
 * \param[in] arg input PERR_ErrorList
 * \return The number of elements in #arg, excluding any separators
 */
int _PERR_getErrorListLength(PERR_ErrorList arg) {
  if (ATisEmpty((ATermList) arg)) {
    return 0;
  }
  return (ATgetLength((ATermList) arg) / 4) + 1;
}

/**
 * Reverses the elements of a PERR_ErrorList. Note that separators are reversed with the list, but the order in which each set of separators inbetween two elements occurs does not change
 * \param[in] arg PERR_ErrorList to be reversed
 * \return #arg reversed
 */
PERR_ErrorList PERR_reverseErrorList(PERR_ErrorList arg) {
  ATermList list = (ATermList) arg;
  ATerm head;
  ATerm sep0;
  ATerm sep1;
  ATerm sep2;
  ATermList result;

 if (ATisEmpty(list) || ATgetLength(list) == 1) {
    return arg;
  }

  result = ATmakeList1(ATgetFirst(list));
      list = ATgetNext(list);
  sep0 = ATgetFirst(list);
  list = ATgetNext(list);
  sep1 = ATgetFirst(list);
  list = ATgetNext(list);
  sep2 = ATgetFirst(list);
  list = ATgetNext(list);

  while (!ATisEmpty(list)) {
    result = ATinsert(result, sep2);
    result = ATinsert(result, sep1);
    result = ATinsert(result, sep0);

   head = ATgetFirst(list);
   result = ATinsert(result, head);
    list = ATgetNext(list);

   if (!ATisEmpty(list)) {
  sep0 = ATgetFirst(list);
  list = ATgetNext(list);
  sep1 = ATgetFirst(list);
  list = ATgetNext(list);
  sep2 = ATgetFirst(list);
  list = ATgetNext(list);
   }
  }

  return (PERR_ErrorList) result;
}

/**
 * Append a PERR_Error to the end of a PERR_ErrorList. 
 * \param[in] arg PERR_ErrorList to append the PERR_Error to
 * \param[in] elem PERR_Error to be appended
 * \return new PERR_ErrorList with #elem appended after the separators
 */
PERR_ErrorList PERR_appendErrorList(PERR_ErrorList arg0, PERR_OptLayout wsAfterHead, PERR_OptLayout wsAfterSep, PERR_Error arg1) {
  return PERR_concatErrorList(arg0, wsAfterHead, wsAfterSep, PERR_makeErrorListSingle(arg1));
}

/**
 * Concatenate two PERR_ErrorLists. 
 * \param[in] arg0 first PERR_ErrorList
 * \param[in] arg1 second PERR_ErrorList
 * \return PERR_ErrorList with the elements of #arg0 before the elements of #arg1, with the separators in between.
 */
PERR_ErrorList PERR_concatErrorList(PERR_ErrorList arg0, PERR_OptLayout wsAfterHead, PERR_OptLayout wsAfterSep, PERR_ErrorList arg1) {
  if (ATisEmpty((ATermList) arg0)) {
    return arg1;
  }
  arg1 = PERR_makeErrorListMany((PERR_Error)ATgetFirst((ATermList) arg0), wsAfterHead, wsAfterSep,  arg1);
  arg1 = (PERR_ErrorList) ATgetNext((ATermList) arg1);
  return (PERR_ErrorList) ATconcat((ATermList) arg0, (ATermList) arg1);
}

/**
 * Extract a sublist from a PERR_ErrorList. 
 * \param[in] arg PERR_ErrorList to extract a slice from
 * \param[in] start inclusive start index of the sublist
 * \param[in] end exclusive end index of the sublist
 * \return new PERR_ErrorList with a first element the element at index #start from #arg, and as last element the element at index (#end - 1).
 */
PERR_ErrorList _PERR_sliceErrorList(PERR_ErrorList arg, int start, int end) {
  return (PERR_ErrorList) ATgetSlice((ATermList) arg, start * 4, end * 4);
}

/**
 * Retrieve the PERR_Error at #index from a PERR_ErrorList. 
 * \param[in] arg PERR_ErrorList to retrieve the PERR_Error from
 * \param[in] index index to use to point in the PERR_ErrorList
 * \return PERR_Error at position #index in #arg
 */
PERR_Error _PERR_getErrorListErrorAt(PERR_ErrorList arg, int index) {
 return (PERR_Error)ATelementAt((ATermList) arg,index * 4);
}

/**
 * Replace the PERR_Error at #index from a PERR_ErrorList by a new one. 
 * \param[in] arg PERR_ErrorList to retrieve the PERR_Error from
 * \param[in] elem new PERR_Error to replace another
 * \param[in] index index to use to point in the PERR_ErrorList
 * \return A new PERR_ErrorListwith #elem replaced in #arg at position #index
 */
PERR_ErrorList _PERR_replaceErrorListErrorAt(PERR_ErrorList arg, PERR_Error elem, int index) {
 return (PERR_ErrorList) ATreplace((ATermList) arg, (ATerm) ((ATerm) elem), index * 4);
}

/**
 * Builds a PERR_ErrorList of 2 consecutive elements. The elements are separated.
 * \param[in] elem1 One PERR_Error element of the new PERR_ErrorList
 * \param[in] elem2 One PERR_Error element of the new PERR_ErrorList
 * \return A new PERR_ErrorList consisting of 2 PERR_Errors
 */
PERR_ErrorList PERR_makeErrorList2(PERR_OptLayout wsAfterHead, PERR_OptLayout wsAfterSep, PERR_Error elem1, PERR_Error elem2) {
  return PERR_makeErrorListMany(elem1, wsAfterHead, wsAfterSep, PERR_makeErrorListSingle(elem2));
}

/**
 * Builds a PERR_ErrorList of 3 consecutive elements. The elements are separated.
 * \param[in] elem1 One PERR_Error element of the new PERR_ErrorList
 * \param[in] elem2 One PERR_Error element of the new PERR_ErrorList
 * \param[in] elem3 One PERR_Error element of the new PERR_ErrorList
 * \return A new PERR_ErrorList consisting of 3 PERR_Errors
 */
PERR_ErrorList PERR_makeErrorList3(PERR_OptLayout wsAfterHead, PERR_OptLayout wsAfterSep, PERR_Error elem1, PERR_Error elem2, PERR_Error elem3) {
  return PERR_makeErrorListMany(elem1, wsAfterHead, wsAfterSep, PERR_makeErrorList2(wsAfterHead, wsAfterSep, elem2, elem3));
}

/**
 * Builds a PERR_ErrorList of 4 consecutive elements. The elements are separated.
 * \param[in] elem1 One PERR_Error element of the new PERR_ErrorList
 * \param[in] elem2 One PERR_Error element of the new PERR_ErrorList
 * \param[in] elem3 One PERR_Error element of the new PERR_ErrorList
 * \param[in] elem4 One PERR_Error element of the new PERR_ErrorList
 * \return A new PERR_ErrorList consisting of 4 PERR_Errors
 */
PERR_ErrorList PERR_makeErrorList4(PERR_OptLayout wsAfterHead, PERR_OptLayout wsAfterSep, PERR_Error elem1, PERR_Error elem2, PERR_Error elem3, PERR_Error elem4) {
  return PERR_makeErrorListMany(elem1, wsAfterHead, wsAfterSep, PERR_makeErrorList3(wsAfterHead, wsAfterSep, elem2, elem3, elem4));
}

/**
 * Builds a PERR_ErrorList of 5 consecutive elements. The elements are separated.
 * \param[in] elem1 One PERR_Error element of the new PERR_ErrorList
 * \param[in] elem2 One PERR_Error element of the new PERR_ErrorList
 * \param[in] elem3 One PERR_Error element of the new PERR_ErrorList
 * \param[in] elem4 One PERR_Error element of the new PERR_ErrorList
 * \param[in] elem5 One PERR_Error element of the new PERR_ErrorList
 * \return A new PERR_ErrorList consisting of 5 PERR_Errors
 */
PERR_ErrorList PERR_makeErrorList5(PERR_OptLayout wsAfterHead, PERR_OptLayout wsAfterSep, PERR_Error elem1, PERR_Error elem2, PERR_Error elem3, PERR_Error elem4, PERR_Error elem5) {
  return PERR_makeErrorListMany(elem1, wsAfterHead, wsAfterSep, PERR_makeErrorList4(wsAfterHead, wsAfterSep, elem2, elem3, elem4, elem5));
}

/**
 * Builds a PERR_ErrorList of 6 consecutive elements. The elements are separated.
 * \param[in] elem1 One PERR_Error element of the new PERR_ErrorList
 * \param[in] elem2 One PERR_Error element of the new PERR_ErrorList
 * \param[in] elem3 One PERR_Error element of the new PERR_ErrorList
 * \param[in] elem4 One PERR_Error element of the new PERR_ErrorList
 * \param[in] elem5 One PERR_Error element of the new PERR_ErrorList
 * \param[in] elem6 One PERR_Error element of the new PERR_ErrorList
 * \return A new PERR_ErrorList consisting of 6 PERR_Errors
 */
PERR_ErrorList PERR_makeErrorList6(PERR_OptLayout wsAfterHead, PERR_OptLayout wsAfterSep, PERR_Error elem1, PERR_Error elem2, PERR_Error elem3, PERR_Error elem4, PERR_Error elem5, PERR_Error elem6) {
  return PERR_makeErrorListMany(elem1, wsAfterHead, wsAfterSep, PERR_makeErrorList5(wsAfterHead, wsAfterSep, elem2, elem3, elem4, elem5, elem6));
}

/**
 * Retrieve the number of elements in a PERR_AreaAreas. 
 * \param[in] arg input PERR_AreaAreas
 * \return The number of elements in #arg, excluding any separators
 */
int _PERR_getAreaAreasLength(PERR_AreaAreas arg) {
  if (ATisEmpty((ATermList) arg)) {
    return 0;
  }
  return (ATgetLength((ATermList) arg) / 4) + 1;
}

/**
 * Reverses the elements of a PERR_AreaAreas. Note that separators are reversed with the list, but the order in which each set of separators inbetween two elements occurs does not change
 * \param[in] arg PERR_AreaAreas to be reversed
 * \return #arg reversed
 */
PERR_AreaAreas PERR_reverseAreaAreas(PERR_AreaAreas arg) {
  ATermList list = (ATermList) arg;
  ATerm head;
  ATerm sep0;
  ATerm sep1;
  ATerm sep2;
  ATermList result;

 if (ATisEmpty(list) || ATgetLength(list) == 1) {
    return arg;
  }

  result = ATmakeList1(ATgetFirst(list));
      list = ATgetNext(list);
  sep0 = ATgetFirst(list);
  list = ATgetNext(list);
  sep1 = ATgetFirst(list);
  list = ATgetNext(list);
  sep2 = ATgetFirst(list);
  list = ATgetNext(list);

  while (!ATisEmpty(list)) {
    result = ATinsert(result, sep2);
    result = ATinsert(result, sep1);
    result = ATinsert(result, sep0);

   head = ATgetFirst(list);
   result = ATinsert(result, head);
    list = ATgetNext(list);

   if (!ATisEmpty(list)) {
  sep0 = ATgetFirst(list);
  list = ATgetNext(list);
  sep1 = ATgetFirst(list);
  list = ATgetNext(list);
  sep2 = ATgetFirst(list);
  list = ATgetNext(list);
   }
  }

  return (PERR_AreaAreas) result;
}

/**
 * Append a PERR_Area to the end of a PERR_AreaAreas. 
 * \param[in] arg PERR_AreaAreas to append the PERR_Area to
 * \param[in] elem PERR_Area to be appended
 * \return new PERR_AreaAreas with #elem appended after the separators
 */
PERR_AreaAreas PERR_appendAreaAreas(PERR_AreaAreas arg0, PERR_OptLayout wsAfterHead, PERR_OptLayout wsAfterSep, PERR_Area arg1) {
  return PERR_concatAreaAreas(arg0, wsAfterHead, wsAfterSep, PERR_makeAreaAreasSingle(arg1));
}

/**
 * Concatenate two PERR_AreaAreass. 
 * \param[in] arg0 first PERR_AreaAreas
 * \param[in] arg1 second PERR_AreaAreas
 * \return PERR_AreaAreas with the elements of #arg0 before the elements of #arg1, with the separators in between.
 */
PERR_AreaAreas PERR_concatAreaAreas(PERR_AreaAreas arg0, PERR_OptLayout wsAfterHead, PERR_OptLayout wsAfterSep, PERR_AreaAreas arg1) {
  if (ATisEmpty((ATermList) arg0)) {
    return arg1;
  }
  arg1 = PERR_makeAreaAreasMany((PERR_Area)ATgetFirst((ATermList) arg0), wsAfterHead, wsAfterSep,  arg1);
  arg1 = (PERR_AreaAreas) ATgetNext((ATermList) arg1);
  return (PERR_AreaAreas) ATconcat((ATermList) arg0, (ATermList) arg1);
}

/**
 * Extract a sublist from a PERR_AreaAreas. 
 * \param[in] arg PERR_AreaAreas to extract a slice from
 * \param[in] start inclusive start index of the sublist
 * \param[in] end exclusive end index of the sublist
 * \return new PERR_AreaAreas with a first element the element at index #start from #arg, and as last element the element at index (#end - 1).
 */
PERR_AreaAreas _PERR_sliceAreaAreas(PERR_AreaAreas arg, int start, int end) {
  return (PERR_AreaAreas) ATgetSlice((ATermList) arg, start * 4, end * 4);
}

/**
 * Retrieve the PERR_Area at #index from a PERR_AreaAreas. 
 * \param[in] arg PERR_AreaAreas to retrieve the PERR_Area from
 * \param[in] index index to use to point in the PERR_AreaAreas
 * \return PERR_Area at position #index in #arg
 */
PERR_Area _PERR_getAreaAreasAreaAt(PERR_AreaAreas arg, int index) {
 return (PERR_Area)ATelementAt((ATermList) arg,index * 4);
}

/**
 * Replace the PERR_Area at #index from a PERR_AreaAreas by a new one. 
 * \param[in] arg PERR_AreaAreas to retrieve the PERR_Area from
 * \param[in] elem new PERR_Area to replace another
 * \param[in] index index to use to point in the PERR_AreaAreas
 * \return A new PERR_AreaAreaswith #elem replaced in #arg at position #index
 */
PERR_AreaAreas _PERR_replaceAreaAreasAreaAt(PERR_AreaAreas arg, PERR_Area elem, int index) {
 return (PERR_AreaAreas) ATreplace((ATermList) arg, (ATerm) ((ATerm) elem), index * 4);
}

/**
 * Builds a PERR_AreaAreas of 2 consecutive elements. The elements are separated.
 * \param[in] elem1 One PERR_Area element of the new PERR_AreaAreas
 * \param[in] elem2 One PERR_Area element of the new PERR_AreaAreas
 * \return A new PERR_AreaAreas consisting of 2 PERR_Areas
 */
PERR_AreaAreas PERR_makeAreaAreas2(PERR_OptLayout wsAfterHead, PERR_OptLayout wsAfterSep, PERR_Area elem1, PERR_Area elem2) {
  return PERR_makeAreaAreasMany(elem1, wsAfterHead, wsAfterSep, PERR_makeAreaAreasSingle(elem2));
}

/**
 * Builds a PERR_AreaAreas of 3 consecutive elements. The elements are separated.
 * \param[in] elem1 One PERR_Area element of the new PERR_AreaAreas
 * \param[in] elem2 One PERR_Area element of the new PERR_AreaAreas
 * \param[in] elem3 One PERR_Area element of the new PERR_AreaAreas
 * \return A new PERR_AreaAreas consisting of 3 PERR_Areas
 */
PERR_AreaAreas PERR_makeAreaAreas3(PERR_OptLayout wsAfterHead, PERR_OptLayout wsAfterSep, PERR_Area elem1, PERR_Area elem2, PERR_Area elem3) {
  return PERR_makeAreaAreasMany(elem1, wsAfterHead, wsAfterSep, PERR_makeAreaAreas2(wsAfterHead, wsAfterSep, elem2, elem3));
}

/**
 * Builds a PERR_AreaAreas of 4 consecutive elements. The elements are separated.
 * \param[in] elem1 One PERR_Area element of the new PERR_AreaAreas
 * \param[in] elem2 One PERR_Area element of the new PERR_AreaAreas
 * \param[in] elem3 One PERR_Area element of the new PERR_AreaAreas
 * \param[in] elem4 One PERR_Area element of the new PERR_AreaAreas
 * \return A new PERR_AreaAreas consisting of 4 PERR_Areas
 */
PERR_AreaAreas PERR_makeAreaAreas4(PERR_OptLayout wsAfterHead, PERR_OptLayout wsAfterSep, PERR_Area elem1, PERR_Area elem2, PERR_Area elem3, PERR_Area elem4) {
  return PERR_makeAreaAreasMany(elem1, wsAfterHead, wsAfterSep, PERR_makeAreaAreas3(wsAfterHead, wsAfterSep, elem2, elem3, elem4));
}

/**
 * Builds a PERR_AreaAreas of 5 consecutive elements. The elements are separated.
 * \param[in] elem1 One PERR_Area element of the new PERR_AreaAreas
 * \param[in] elem2 One PERR_Area element of the new PERR_AreaAreas
 * \param[in] elem3 One PERR_Area element of the new PERR_AreaAreas
 * \param[in] elem4 One PERR_Area element of the new PERR_AreaAreas
 * \param[in] elem5 One PERR_Area element of the new PERR_AreaAreas
 * \return A new PERR_AreaAreas consisting of 5 PERR_Areas
 */
PERR_AreaAreas PERR_makeAreaAreas5(PERR_OptLayout wsAfterHead, PERR_OptLayout wsAfterSep, PERR_Area elem1, PERR_Area elem2, PERR_Area elem3, PERR_Area elem4, PERR_Area elem5) {
  return PERR_makeAreaAreasMany(elem1, wsAfterHead, wsAfterSep, PERR_makeAreaAreas4(wsAfterHead, wsAfterSep, elem2, elem3, elem4, elem5));
}

/**
 * Builds a PERR_AreaAreas of 6 consecutive elements. The elements are separated.
 * \param[in] elem1 One PERR_Area element of the new PERR_AreaAreas
 * \param[in] elem2 One PERR_Area element of the new PERR_AreaAreas
 * \param[in] elem3 One PERR_Area element of the new PERR_AreaAreas
 * \param[in] elem4 One PERR_Area element of the new PERR_AreaAreas
 * \param[in] elem5 One PERR_Area element of the new PERR_AreaAreas
 * \param[in] elem6 One PERR_Area element of the new PERR_AreaAreas
 * \return A new PERR_AreaAreas consisting of 6 PERR_Areas
 */
PERR_AreaAreas PERR_makeAreaAreas6(PERR_OptLayout wsAfterHead, PERR_OptLayout wsAfterSep, PERR_Area elem1, PERR_Area elem2, PERR_Area elem3, PERR_Area elem4, PERR_Area elem5, PERR_Area elem6) {
  return PERR_makeAreaAreasMany(elem1, wsAfterHead, wsAfterSep, PERR_makeAreaAreas5(wsAfterHead, wsAfterSep, elem2, elem3, elem4, elem5, elem6));
}

/**
 * Constructs a Absent of type PERR_OptLayout. Like all ATerm types, PERR_OptLayouts are maximally shared.
 * \return A pointer to a Absent, either newly constructed or shared
 */
PERR_OptLayout PERR_makeOptLayoutAbsent(void) {
  return (PERR_OptLayout)(ATerm)ATmakeAppl2(PERR_afun0, (ATerm)ATmakeAppl3(PERR_afun1, (ATerm)ATempty, (ATerm)ATmakeAppl1(PERR_afun2, (ATerm)ATmakeAppl1(PERR_afun3, (ATerm)ATmakeAppl0(PERR_afun4))), (ATerm)ATmakeAppl0(PERR_afun5)), (ATerm)ATempty);
}
/**
 * Constructs a Present of type PERR_OptLayout. Like all ATerm types, PERR_OptLayouts are maximally shared.
 * \param[in] layout a child of the new Present
 * \return A pointer to a Present, either newly constructed or shared
 */
PERR_OptLayout PERR_makeOptLayoutPresent(PERR_Layout layout) {
  return (PERR_OptLayout)(ATerm)ATmakeAppl2(PERR_afun0, (ATerm)ATmakeAppl3(PERR_afun1, (ATerm)ATmakeList1((ATerm)ATmakeAppl1(PERR_afun2, (ATerm)ATmakeAppl0(PERR_afun4))), (ATerm)ATmakeAppl1(PERR_afun2, (ATerm)ATmakeAppl1(PERR_afun3, (ATerm)ATmakeAppl0(PERR_afun4))), (ATerm)ATmakeAppl0(PERR_afun5)), (ATerm)ATmakeList1((ATerm) layout));
}
/**
 * Constructs a Lex-to-cf of type PERR_Layout. Like all ATerm types, PERR_Layouts are maximally shared.
 * \param[in] list a child of the new Lex-to-cf
 * \return A pointer to a Lex-to-cf, either newly constructed or shared
 */
PERR_Layout PERR_makeLayoutLexToCf(PERR_LexLayoutList list) {
  return (PERR_Layout)(ATerm)ATmakeAppl2(PERR_afun0, (ATerm)ATmakeAppl3(PERR_afun1, (ATerm)ATmakeList1((ATerm)ATmakeAppl1(PERR_afun6, (ATerm)ATmakeAppl1(PERR_afun7, (ATerm)ATmakeAppl0(PERR_afun4)))), (ATerm)ATmakeAppl1(PERR_afun2, (ATerm)ATmakeAppl0(PERR_afun4)), (ATerm)ATmakeAppl0(PERR_afun5)), (ATerm)ATmakeList1((ATerm)ATmakeAppl2(PERR_afun0, (ATerm)ATmakeAppl1(PERR_afun8, (ATerm)ATmakeAppl1(PERR_afun6, (ATerm)ATmakeAppl1(PERR_afun7, (ATerm)ATmakeAppl0(PERR_afun4)))), (ATerm) list)));
}
/**
 * Constructs a empty of type PERR_LexLayoutList. Like all ATerm types, PERR_LexLayoutLists are maximally shared.
 * \return A pointer to a empty, either newly constructed or shared
 */
PERR_LexLayoutList PERR_makeLexLayoutListEmpty(void) {
  return (PERR_LexLayoutList)(ATerm)ATempty;
}
/**
 * Constructs a single of type PERR_LexLayoutList. Like all ATerm types, PERR_LexLayoutLists are maximally shared.
 * \param[in] head a child of the new single
 * \return A pointer to a single, either newly constructed or shared
 */
PERR_LexLayoutList PERR_makeLexLayoutListSingle(PERR_LexLayout head) {
  return (PERR_LexLayoutList)(ATerm)ATmakeList1((ATerm) head);
}
/**
 * Constructs a many of type PERR_LexLayoutList. Like all ATerm types, PERR_LexLayoutLists are maximally shared.
 * \param[in] head a child of the new many
 * \param[in] tail a child of the new many
 * \return A pointer to a many, either newly constructed or shared
 */
PERR_LexLayoutList PERR_makeLexLayoutListMany(PERR_LexLayout head, PERR_LexLayoutList tail) {
  return (PERR_LexLayoutList)(ATerm)ATinsert((ATermList)tail, (ATerm) head);
}
/**
 * Constructs a newline of type PERR_LexStrChar. Like all ATerm types, PERR_LexStrChars are maximally shared.
 * \return A pointer to a newline, either newly constructed or shared
 */
PERR_LexStrChar PERR_makeLexStrCharNewline(void) {
  return (PERR_LexStrChar)(ATerm)ATmakeAppl2(PERR_afun0, (ATerm)ATmakeAppl3(PERR_afun1, (ATerm)ATmakeList1((ATerm)ATmakeAppl1(PERR_afun9, (ATerm)ATmakeAppl0(PERR_afun10))), (ATerm)ATmakeAppl1(PERR_afun6, (ATerm)ATmakeAppl1(PERR_afun11, (ATerm)ATmakeAppl0(PERR_afun12))), (ATerm)ATmakeAppl1(PERR_afun13, (ATerm)ATmakeList1((ATerm)ATmakeAppl1(PERR_afun14, (ATerm)ATmakeAppl1(PERR_afun15, (ATerm)ATmakeAppl0(PERR_afun16)))))), (ATerm)ATmakeList1((ATerm)ATmakeAppl2(PERR_afun0, (ATerm)ATmakeAppl3(PERR_afun1, (ATerm)ATinsert(ATmakeList1((ATerm)ATmakeAppl1(PERR_afun17, (ATerm)ATmakeList1((ATerm)ATmakeInt(110)))), (ATerm)ATmakeAppl1(PERR_afun17, (ATerm)ATmakeList1((ATerm)ATmakeInt(92)))), (ATerm)ATmakeAppl1(PERR_afun9, (ATerm)ATmakeAppl0(PERR_afun10)), (ATerm)ATmakeAppl0(PERR_afun5)), (ATerm)ATinsert(ATmakeList1((ATerm)ATmakeInt(110)), (ATerm)ATmakeInt(92)))));
}
/**
 * Constructs a tab of type PERR_LexStrChar. Like all ATerm types, PERR_LexStrChars are maximally shared.
 * \return A pointer to a tab, either newly constructed or shared
 */
PERR_LexStrChar PERR_makeLexStrCharTab(void) {
  return (PERR_LexStrChar)(ATerm)ATmakeAppl2(PERR_afun0, (ATerm)ATmakeAppl3(PERR_afun1, (ATerm)ATmakeList1((ATerm)ATmakeAppl1(PERR_afun9, (ATerm)ATmakeAppl0(PERR_afun18))), (ATerm)ATmakeAppl1(PERR_afun6, (ATerm)ATmakeAppl1(PERR_afun11, (ATerm)ATmakeAppl0(PERR_afun12))), (ATerm)ATmakeAppl1(PERR_afun13, (ATerm)ATmakeList1((ATerm)ATmakeAppl1(PERR_afun14, (ATerm)ATmakeAppl1(PERR_afun15, (ATerm)ATmakeAppl0(PERR_afun19)))))), (ATerm)ATmakeList1((ATerm)ATmakeAppl2(PERR_afun0, (ATerm)ATmakeAppl3(PERR_afun1, (ATerm)ATinsert(ATmakeList1((ATerm)ATmakeAppl1(PERR_afun17, (ATerm)ATmakeList1((ATerm)ATmakeInt(116)))), (ATerm)ATmakeAppl1(PERR_afun17, (ATerm)ATmakeList1((ATerm)ATmakeInt(92)))), (ATerm)ATmakeAppl1(PERR_afun9, (ATerm)ATmakeAppl0(PERR_afun18)), (ATerm)ATmakeAppl0(PERR_afun5)), (ATerm)ATinsert(ATmakeList1((ATerm)ATmakeInt(116)), (ATerm)ATmakeInt(92)))));
}
/**
 * Constructs a quote of type PERR_LexStrChar. Like all ATerm types, PERR_LexStrChars are maximally shared.
 * \return A pointer to a quote, either newly constructed or shared
 */
PERR_LexStrChar PERR_makeLexStrCharQuote(void) {
  return (PERR_LexStrChar)(ATerm)ATmakeAppl2(PERR_afun0, (ATerm)ATmakeAppl3(PERR_afun1, (ATerm)ATmakeList1((ATerm)ATmakeAppl1(PERR_afun9, (ATerm)ATmakeAppl0(PERR_afun20))), (ATerm)ATmakeAppl1(PERR_afun6, (ATerm)ATmakeAppl1(PERR_afun11, (ATerm)ATmakeAppl0(PERR_afun12))), (ATerm)ATmakeAppl1(PERR_afun13, (ATerm)ATmakeList1((ATerm)ATmakeAppl1(PERR_afun14, (ATerm)ATmakeAppl1(PERR_afun15, (ATerm)ATmakeAppl0(PERR_afun21)))))), (ATerm)ATmakeList1((ATerm)ATmakeAppl2(PERR_afun0, (ATerm)ATmakeAppl3(PERR_afun1, (ATerm)ATinsert(ATmakeList1((ATerm)ATmakeAppl1(PERR_afun17, (ATerm)ATmakeList1((ATerm)ATmakeInt(34)))), (ATerm)ATmakeAppl1(PERR_afun17, (ATerm)ATmakeList1((ATerm)ATmakeInt(92)))), (ATerm)ATmakeAppl1(PERR_afun9, (ATerm)ATmakeAppl0(PERR_afun20)), (ATerm)ATmakeAppl0(PERR_afun5)), (ATerm)ATinsert(ATmakeList1((ATerm)ATmakeInt(34)), (ATerm)ATmakeInt(92)))));
}
/**
 * Constructs a backslash of type PERR_LexStrChar. Like all ATerm types, PERR_LexStrChars are maximally shared.
 * \return A pointer to a backslash, either newly constructed or shared
 */
PERR_LexStrChar PERR_makeLexStrCharBackslash(void) {
  return (PERR_LexStrChar)(ATerm)ATmakeAppl2(PERR_afun0, (ATerm)ATmakeAppl3(PERR_afun1, (ATerm)ATmakeList1((ATerm)ATmakeAppl1(PERR_afun9, (ATerm)ATmakeAppl0(PERR_afun22))), (ATerm)ATmakeAppl1(PERR_afun6, (ATerm)ATmakeAppl1(PERR_afun11, (ATerm)ATmakeAppl0(PERR_afun12))), (ATerm)ATmakeAppl1(PERR_afun13, (ATerm)ATmakeList1((ATerm)ATmakeAppl1(PERR_afun14, (ATerm)ATmakeAppl1(PERR_afun15, (ATerm)ATmakeAppl0(PERR_afun23)))))), (ATerm)ATmakeList1((ATerm)ATmakeAppl2(PERR_afun0, (ATerm)ATmakeAppl3(PERR_afun1, (ATerm)ATinsert(ATmakeList1((ATerm)ATmakeAppl1(PERR_afun17, (ATerm)ATmakeList1((ATerm)ATmakeInt(92)))), (ATerm)ATmakeAppl1(PERR_afun17, (ATerm)ATmakeList1((ATerm)ATmakeInt(92)))), (ATerm)ATmakeAppl1(PERR_afun9, (ATerm)ATmakeAppl0(PERR_afun22)), (ATerm)ATmakeAppl0(PERR_afun5)), (ATerm)ATinsert(ATmakeList1((ATerm)ATmakeInt(92)), (ATerm)ATmakeInt(92)))));
}
/**
 * Constructs a decimal of type PERR_LexStrChar. Like all ATerm types, PERR_LexStrChars are maximally shared.
 * \param[in] a a child of the new decimal
 * \param[in] b a child of the new decimal
 * \param[in] c a child of the new decimal
 * \return A pointer to a decimal, either newly constructed or shared
 */
PERR_LexStrChar PERR_makeLexStrCharDecimal(char a, char b, char c) {
  return (PERR_LexStrChar)(ATerm)ATmakeAppl2(PERR_afun0, (ATerm)ATmakeAppl3(PERR_afun1, (ATerm)ATinsert(ATinsert(ATinsert(ATmakeList1((ATerm)ATmakeAppl1(PERR_afun17, (ATerm)ATmakeList1((ATerm)ATmakeAppl2(PERR_afun24, (ATerm)ATmakeInt(48), (ATerm)ATmakeInt(57))))), (ATerm)ATmakeAppl1(PERR_afun17, (ATerm)ATmakeList1((ATerm)ATmakeAppl2(PERR_afun24, (ATerm)ATmakeInt(48), (ATerm)ATmakeInt(57))))), (ATerm)ATmakeAppl1(PERR_afun17, (ATerm)ATmakeList1((ATerm)ATmakeAppl2(PERR_afun24, (ATerm)ATmakeInt(48), (ATerm)ATmakeInt(57))))), (ATerm)ATmakeAppl1(PERR_afun9, (ATerm)ATmakeAppl0(PERR_afun25))), (ATerm)ATmakeAppl1(PERR_afun6, (ATerm)ATmakeAppl1(PERR_afun11, (ATerm)ATmakeAppl0(PERR_afun12))), (ATerm)ATmakeAppl1(PERR_afun13, (ATerm)ATmakeList1((ATerm)ATmakeAppl1(PERR_afun14, (ATerm)ATmakeAppl1(PERR_afun15, (ATerm)ATmakeAppl0(PERR_afun26)))))), (ATerm)ATinsert(ATinsert(ATinsert(ATmakeList1((ATerm) ((ATerm) PERR_byteToChar(c))), (ATerm) ((ATerm) PERR_byteToChar(b))), (ATerm) ((ATerm) PERR_byteToChar(a))), (ATerm)ATmakeAppl2(PERR_afun0, (ATerm)ATmakeAppl3(PERR_afun1, (ATerm)ATmakeList1((ATerm)ATmakeAppl1(PERR_afun17, (ATerm)ATmakeList1((ATerm)ATmakeInt(92)))), (ATerm)ATmakeAppl1(PERR_afun9, (ATerm)ATmakeAppl0(PERR_afun25)), (ATerm)ATmakeAppl0(PERR_afun5)), (ATerm)ATmakeList1((ATerm)ATmakeInt(92)))));
}
/**
 * Constructs a normal of type PERR_LexStrChar. Like all ATerm types, PERR_LexStrChars are maximally shared.
 * \param[in] ch a child of the new normal
 * \return A pointer to a normal, either newly constructed or shared
 */
PERR_LexStrChar PERR_makeLexStrCharNormal(char ch) {
  return (PERR_LexStrChar)(ATerm)ATmakeAppl2(PERR_afun0, (ATerm)ATmakeAppl3(PERR_afun1, (ATerm)ATmakeList1((ATerm)ATmakeAppl1(PERR_afun17, (ATerm)ATinsert(ATinsert(ATmakeList1((ATerm)ATmakeAppl2(PERR_afun24, (ATerm)ATmakeInt(93), (ATerm)ATmakeInt(255))), (ATerm)ATmakeAppl2(PERR_afun24, (ATerm)ATmakeInt(35), (ATerm)ATmakeInt(91))), (ATerm)ATmakeAppl2(PERR_afun24, (ATerm)ATmakeInt(32), (ATerm)ATmakeInt(33))))), (ATerm)ATmakeAppl1(PERR_afun6, (ATerm)ATmakeAppl1(PERR_afun11, (ATerm)ATmakeAppl0(PERR_afun12))), (ATerm)ATmakeAppl1(PERR_afun13, (ATerm)ATmakeList1((ATerm)ATmakeAppl1(PERR_afun14, (ATerm)ATmakeAppl1(PERR_afun15, (ATerm)ATmakeAppl0(PERR_afun27)))))), (ATerm)ATmakeList1((ATerm) ((ATerm) PERR_byteToChar(ch))));
}
/**
 * Constructs a Lex-to-cf of type PERR_StrChar. Like all ATerm types, PERR_StrChars are maximally shared.
 * \param[in] StrChar a child of the new Lex-to-cf
 * \return A pointer to a Lex-to-cf, either newly constructed or shared
 */
PERR_StrChar PERR_makeStrCharLexToCf(PERR_LexStrChar StrChar) {
  return (PERR_StrChar)(ATerm)ATmakeAppl2(PERR_afun0, (ATerm)ATmakeAppl3(PERR_afun1, (ATerm)ATmakeList1((ATerm)ATmakeAppl1(PERR_afun6, (ATerm)ATmakeAppl1(PERR_afun11, (ATerm)ATmakeAppl0(PERR_afun12)))), (ATerm)ATmakeAppl1(PERR_afun2, (ATerm)ATmakeAppl1(PERR_afun11, (ATerm)ATmakeAppl0(PERR_afun12))), (ATerm)ATmakeAppl0(PERR_afun5)), (ATerm)ATmakeList1((ATerm) StrChar));
}
/**
 * Constructs a default of type PERR_LexStrCon. Like all ATerm types, PERR_LexStrCons are maximally shared.
 * \param[in] chars a child of the new default
 * \return A pointer to a default, either newly constructed or shared
 */
PERR_LexStrCon PERR_makeLexStrConDefault(PERR_LexStrCharChars chars) {
  return (PERR_LexStrCon)(ATerm)ATmakeAppl2(PERR_afun0, (ATerm)ATmakeAppl3(PERR_afun1, (ATerm)ATinsert(ATinsert(ATmakeList1((ATerm)ATmakeAppl1(PERR_afun17, (ATerm)ATmakeList1((ATerm)ATmakeInt(34)))), (ATerm)ATmakeAppl1(PERR_afun6, (ATerm)ATmakeAppl1(PERR_afun28, (ATerm)ATmakeAppl1(PERR_afun11, (ATerm)ATmakeAppl0(PERR_afun12))))), (ATerm)ATmakeAppl1(PERR_afun17, (ATerm)ATmakeList1((ATerm)ATmakeInt(34)))), (ATerm)ATmakeAppl1(PERR_afun6, (ATerm)ATmakeAppl1(PERR_afun11, (ATerm)ATmakeAppl0(PERR_afun29))), (ATerm)ATmakeAppl1(PERR_afun13, (ATerm)ATmakeList1((ATerm)ATmakeAppl1(PERR_afun14, (ATerm)ATmakeAppl1(PERR_afun15, (ATerm)ATmakeAppl0(PERR_afun30)))))), (ATerm)ATinsert(ATinsert(ATmakeList1((ATerm)ATmakeInt(34)), (ATerm)ATmakeAppl2(PERR_afun0, (ATerm)ATmakeAppl1(PERR_afun8, (ATerm)ATmakeAppl1(PERR_afun6, (ATerm)ATmakeAppl1(PERR_afun28, (ATerm)ATmakeAppl1(PERR_afun11, (ATerm)ATmakeAppl0(PERR_afun12))))), (ATerm) chars)), (ATerm)ATmakeInt(34)));
}
/**
 * Constructs a Lex-to-cf of type PERR_StrCon. Like all ATerm types, PERR_StrCons are maximally shared.
 * \param[in] StrCon a child of the new Lex-to-cf
 * \return A pointer to a Lex-to-cf, either newly constructed or shared
 */
PERR_StrCon PERR_makeStrConLexToCf(PERR_LexStrCon StrCon) {
  return (PERR_StrCon)(ATerm)ATmakeAppl2(PERR_afun0, (ATerm)ATmakeAppl3(PERR_afun1, (ATerm)ATmakeList1((ATerm)ATmakeAppl1(PERR_afun6, (ATerm)ATmakeAppl1(PERR_afun11, (ATerm)ATmakeAppl0(PERR_afun29)))), (ATerm)ATmakeAppl1(PERR_afun2, (ATerm)ATmakeAppl1(PERR_afun11, (ATerm)ATmakeAppl0(PERR_afun29))), (ATerm)ATmakeAppl0(PERR_afun5)), (ATerm)ATmakeList1((ATerm) StrCon));
}
/**
 * Constructs a empty of type PERR_LexStrCharChars. Like all ATerm types, PERR_LexStrCharCharss are maximally shared.
 * \return A pointer to a empty, either newly constructed or shared
 */
PERR_LexStrCharChars PERR_makeLexStrCharCharsEmpty(void) {
  return (PERR_LexStrCharChars)(ATerm)ATempty;
}
/**
 * Constructs a single of type PERR_LexStrCharChars. Like all ATerm types, PERR_LexStrCharCharss are maximally shared.
 * \param[in] head a child of the new single
 * \return A pointer to a single, either newly constructed or shared
 */
PERR_LexStrCharChars PERR_makeLexStrCharCharsSingle(PERR_LexStrChar head) {
  return (PERR_LexStrCharChars)(ATerm)ATmakeList1((ATerm) head);
}
/**
 * Constructs a many of type PERR_LexStrCharChars. Like all ATerm types, PERR_LexStrCharCharss are maximally shared.
 * \param[in] head a child of the new many
 * \param[in] tail a child of the new many
 * \return A pointer to a many, either newly constructed or shared
 */
PERR_LexStrCharChars PERR_makeLexStrCharCharsMany(PERR_LexStrChar head, PERR_LexStrCharChars tail) {
  return (PERR_LexStrCharChars)(ATerm)ATinsert((ATermList)tail, (ATerm) head);
}
/**
 * Constructs a digits of type PERR_LexNatCon. Like all ATerm types, PERR_LexNatCons are maximally shared.
 * \param[in] list a child of the new digits
 * \return A pointer to a digits, either newly constructed or shared
 */
PERR_LexNatCon PERR_makeLexNatConDigits(const char* list) {
  return (PERR_LexNatCon)(ATerm)ATmakeAppl2(PERR_afun0, (ATerm)ATmakeAppl3(PERR_afun1, (ATerm)ATmakeList1((ATerm)ATmakeAppl1(PERR_afun6, (ATerm)ATmakeAppl1(PERR_afun7, (ATerm)ATmakeAppl1(PERR_afun17, (ATerm)ATmakeList1((ATerm)ATmakeAppl2(PERR_afun24, (ATerm)ATmakeInt(48), (ATerm)ATmakeInt(57))))))), (ATerm)ATmakeAppl1(PERR_afun6, (ATerm)ATmakeAppl1(PERR_afun11, (ATerm)ATmakeAppl0(PERR_afun31))), (ATerm)ATmakeAppl1(PERR_afun13, (ATerm)ATmakeList1((ATerm)ATmakeAppl1(PERR_afun14, (ATerm)ATmakeAppl1(PERR_afun15, (ATerm)ATmakeAppl0(PERR_afun32)))))), (ATerm)ATmakeList1((ATerm)ATmakeAppl2(PERR_afun0, (ATerm)ATmakeAppl1(PERR_afun8, (ATerm)ATmakeAppl1(PERR_afun6, (ATerm)ATmakeAppl1(PERR_afun7, (ATerm)ATmakeAppl1(PERR_afun17, (ATerm)ATmakeList1((ATerm)ATmakeAppl2(PERR_afun24, (ATerm)ATmakeInt(48), (ATerm)ATmakeInt(57))))))), (ATerm) ((ATerm) PERR_stringToChars(list)))));
}
/**
 * Constructs a Lex-to-cf of type PERR_NatCon. Like all ATerm types, PERR_NatCons are maximally shared.
 * \param[in] NatCon a child of the new Lex-to-cf
 * \return A pointer to a Lex-to-cf, either newly constructed or shared
 */
PERR_NatCon PERR_makeNatConLexToCf(PERR_LexNatCon NatCon) {
  return (PERR_NatCon)(ATerm)ATmakeAppl2(PERR_afun0, (ATerm)ATmakeAppl3(PERR_afun1, (ATerm)ATmakeList1((ATerm)ATmakeAppl1(PERR_afun6, (ATerm)ATmakeAppl1(PERR_afun11, (ATerm)ATmakeAppl0(PERR_afun31)))), (ATerm)ATmakeAppl1(PERR_afun2, (ATerm)ATmakeAppl1(PERR_afun11, (ATerm)ATmakeAppl0(PERR_afun31))), (ATerm)ATmakeAppl0(PERR_afun5)), (ATerm)ATmakeList1((ATerm) NatCon));
}
/**
 * Constructs a whitespace of type PERR_LexLayout. Like all ATerm types, PERR_LexLayouts are maximally shared.
 * \param[in] ch a child of the new whitespace
 * \return A pointer to a whitespace, either newly constructed or shared
 */
PERR_LexLayout PERR_makeLexLayoutWhitespace(char ch) {
  return (PERR_LexLayout)(ATerm)ATmakeAppl2(PERR_afun0, (ATerm)ATmakeAppl3(PERR_afun1, (ATerm)ATmakeList1((ATerm)ATmakeAppl1(PERR_afun17, (ATerm)ATinsert(ATinsert(ATmakeList1((ATerm)ATmakeInt(32)), (ATerm)ATmakeInt(13)), (ATerm)ATmakeAppl2(PERR_afun24, (ATerm)ATmakeInt(9), (ATerm)ATmakeInt(10))))), (ATerm)ATmakeAppl1(PERR_afun6, (ATerm)ATmakeAppl0(PERR_afun4)), (ATerm)ATmakeAppl1(PERR_afun13, (ATerm)ATmakeList1((ATerm)ATmakeAppl1(PERR_afun14, (ATerm)ATmakeAppl1(PERR_afun15, (ATerm)ATmakeAppl0(PERR_afun33)))))), (ATerm)ATmakeList1((ATerm) ((ATerm) PERR_byteToChar(ch))));
}
/**
 * Constructs a Summary of type PERR_Start. Like all ATerm types, PERR_Starts are maximally shared.
 * \param[in] wsBefore a child of the new Summary
 * \param[in] topSummary a child of the new Summary
 * \param[in] wsAfter a child of the new Summary
 * \param[in] ambCnt a child of the new Summary
 * \return A pointer to a Summary, either newly constructed or shared
 */
PERR_Start PERR_makeStartSummary(PERR_OptLayout wsBefore, PERR_Summary topSummary, PERR_OptLayout wsAfter, int ambCnt) {
  return (PERR_Start)(ATerm)ATmakeAppl2(PERR_afun34, (ATerm)ATmakeAppl2(PERR_afun0, (ATerm)ATmakeAppl3(PERR_afun1, (ATerm)ATinsert(ATinsert(ATmakeList1((ATerm)ATmakeAppl1(PERR_afun2, (ATerm)ATmakeAppl1(PERR_afun3, (ATerm)ATmakeAppl0(PERR_afun4)))), (ATerm)ATmakeAppl1(PERR_afun2, (ATerm)ATmakeAppl1(PERR_afun11, (ATerm)ATmakeAppl0(PERR_afun35)))), (ATerm)ATmakeAppl1(PERR_afun2, (ATerm)ATmakeAppl1(PERR_afun3, (ATerm)ATmakeAppl0(PERR_afun4)))), (ATerm)ATmakeAppl1(PERR_afun11, (ATerm)ATmakeAppl0(PERR_afun36)), (ATerm)ATmakeAppl0(PERR_afun5)), (ATerm)ATinsert(ATinsert(ATmakeList1((ATerm) wsAfter), (ATerm) topSummary), (ATerm) wsBefore)), (ATerm) (ATerm) ATmakeInt(ambCnt));
}
/**
 * Constructs a Error of type PERR_Start. Like all ATerm types, PERR_Starts are maximally shared.
 * \param[in] wsBefore a child of the new Error
 * \param[in] topError a child of the new Error
 * \param[in] wsAfter a child of the new Error
 * \param[in] ambCnt a child of the new Error
 * \return A pointer to a Error, either newly constructed or shared
 */
PERR_Start PERR_makeStartError(PERR_OptLayout wsBefore, PERR_Error topError, PERR_OptLayout wsAfter, int ambCnt) {
  return (PERR_Start)(ATerm)ATmakeAppl2(PERR_afun34, (ATerm)ATmakeAppl2(PERR_afun0, (ATerm)ATmakeAppl3(PERR_afun1, (ATerm)ATinsert(ATinsert(ATmakeList1((ATerm)ATmakeAppl1(PERR_afun2, (ATerm)ATmakeAppl1(PERR_afun3, (ATerm)ATmakeAppl0(PERR_afun4)))), (ATerm)ATmakeAppl1(PERR_afun2, (ATerm)ATmakeAppl1(PERR_afun11, (ATerm)ATmakeAppl0(PERR_afun37)))), (ATerm)ATmakeAppl1(PERR_afun2, (ATerm)ATmakeAppl1(PERR_afun3, (ATerm)ATmakeAppl0(PERR_afun4)))), (ATerm)ATmakeAppl1(PERR_afun11, (ATerm)ATmakeAppl0(PERR_afun36)), (ATerm)ATmakeAppl0(PERR_afun5)), (ATerm)ATinsert(ATinsert(ATmakeList1((ATerm) wsAfter), (ATerm) topError), (ATerm) wsBefore)), (ATerm) (ATerm) ATmakeInt(ambCnt));
}
/**
 * Constructs a subject of type PERR_Subject. Like all ATerm types, PERR_Subjects are maximally shared.
 * \param[in] wsAfterSubject a child of the new subject
 * \param[in] wsAfterParenOpen a child of the new subject
 * \param[in] description a child of the new subject
 * \param[in] wsAfterDescription a child of the new subject
 * \return A pointer to a subject, either newly constructed or shared
 */
PERR_Subject PERR_makeSubjectSubject(PERR_OptLayout wsAfterSubject, PERR_OptLayout wsAfterParenOpen, PERR_StrCon description, PERR_OptLayout wsAfterDescription) {
  return (PERR_Subject)(ATerm)ATmakeAppl2(PERR_afun0, (ATerm)ATmakeAppl3(PERR_afun1, (ATerm)ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATmakeList1((ATerm)ATmakeAppl1(PERR_afun9, (ATerm)ATmakeAppl0(PERR_afun38))), (ATerm)ATmakeAppl1(PERR_afun2, (ATerm)ATmakeAppl1(PERR_afun3, (ATerm)ATmakeAppl0(PERR_afun4)))), (ATerm)ATmakeAppl1(PERR_afun2, (ATerm)ATmakeAppl1(PERR_afun11, (ATerm)ATmakeAppl0(PERR_afun29)))), (ATerm)ATmakeAppl1(PERR_afun2, (ATerm)ATmakeAppl1(PERR_afun3, (ATerm)ATmakeAppl0(PERR_afun4)))), (ATerm)ATmakeAppl1(PERR_afun9, (ATerm)ATmakeAppl0(PERR_afun39))), (ATerm)ATmakeAppl1(PERR_afun2, (ATerm)ATmakeAppl1(PERR_afun3, (ATerm)ATmakeAppl0(PERR_afun4)))), (ATerm)ATmakeAppl1(PERR_afun9, (ATerm)ATmakeAppl0(PERR_afun40))), (ATerm)ATmakeAppl1(PERR_afun2, (ATerm)ATmakeAppl1(PERR_afun11, (ATerm)ATmakeAppl0(PERR_afun41))), (ATerm)ATmakeAppl1(PERR_afun13, (ATerm)ATmakeList1((ATerm)ATmakeAppl1(PERR_afun14, (ATerm)ATmakeAppl1(PERR_afun15, (ATerm)ATmakeAppl0(PERR_afun40)))))), (ATerm)ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATmakeList1((ATerm)ATmakeAppl2(PERR_afun0, (ATerm)ATmakeAppl3(PERR_afun1, (ATerm)ATmakeList1((ATerm)ATmakeAppl1(PERR_afun17, (ATerm)ATmakeList1((ATerm)ATmakeInt(41)))), (ATerm)ATmakeAppl1(PERR_afun9, (ATerm)ATmakeAppl0(PERR_afun38)), (ATerm)ATmakeAppl0(PERR_afun5)), (ATerm)ATmakeList1((ATerm)ATmakeInt(41)))), (ATerm) wsAfterDescription), (ATerm) description), (ATerm) wsAfterParenOpen), (ATerm)ATmakeAppl2(PERR_afun0, (ATerm)ATmakeAppl3(PERR_afun1, (ATerm)ATmakeList1((ATerm)ATmakeAppl1(PERR_afun17, (ATerm)ATmakeList1((ATerm)ATmakeInt(40)))), (ATerm)ATmakeAppl1(PERR_afun9, (ATerm)ATmakeAppl0(PERR_afun39)), (ATerm)ATmakeAppl0(PERR_afun5)), (ATerm)ATmakeList1((ATerm)ATmakeInt(40)))), (ATerm) wsAfterSubject), (ATerm)ATmakeAppl2(PERR_afun0, (ATerm)ATmakeAppl3(PERR_afun1, (ATerm)ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATmakeList1((ATerm)ATmakeAppl1(PERR_afun17, (ATerm)ATmakeList1((ATerm)ATmakeInt(116)))), (ATerm)ATmakeAppl1(PERR_afun17, (ATerm)ATmakeList1((ATerm)ATmakeInt(99)))), (ATerm)ATmakeAppl1(PERR_afun17, (ATerm)ATmakeList1((ATerm)ATmakeInt(101)))), (ATerm)ATmakeAppl1(PERR_afun17, (ATerm)ATmakeList1((ATerm)ATmakeInt(106)))), (ATerm)ATmakeAppl1(PERR_afun17, (ATerm)ATmakeList1((ATerm)ATmakeInt(98)))), (ATerm)ATmakeAppl1(PERR_afun17, (ATerm)ATmakeList1((ATerm)ATmakeInt(117)))), (ATerm)ATmakeAppl1(PERR_afun17, (ATerm)ATmakeList1((ATerm)ATmakeInt(115)))), (ATerm)ATmakeAppl1(PERR_afun9, (ATerm)ATmakeAppl0(PERR_afun40)), (ATerm)ATmakeAppl0(PERR_afun5)), (ATerm)ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATmakeList1((ATerm)ATmakeInt(116)), (ATerm)ATmakeInt(99)), (ATerm)ATmakeInt(101)), (ATerm)ATmakeInt(106)), (ATerm)ATmakeInt(98)), (ATerm)ATmakeInt(117)), (ATerm)ATmakeInt(115)))));
}
/**
 * Constructs a localized of type PERR_Subject. Like all ATerm types, PERR_Subjects are maximally shared.
 * \param[in] wsAfterLocalized a child of the new localized
 * \param[in] wsAfterParenOpen a child of the new localized
 * \param[in] description a child of the new localized
 * \param[in] wsAfterDescription a child of the new localized
 * \param[in] wsAfterComma a child of the new localized
 * \param[in] Location a child of the new localized
 * \param[in] wsAfterLocation a child of the new localized
 * \return A pointer to a localized, either newly constructed or shared
 */
PERR_Subject PERR_makeSubjectLocalized(PERR_OptLayout wsAfterLocalized, PERR_OptLayout wsAfterParenOpen, PERR_StrCon description, PERR_OptLayout wsAfterDescription, PERR_OptLayout wsAfterComma, PERR_Location Location, PERR_OptLayout wsAfterLocation) {
  return (PERR_Subject)(ATerm)ATmakeAppl2(PERR_afun0, (ATerm)ATmakeAppl3(PERR_afun1, (ATerm)ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATmakeList1((ATerm)ATmakeAppl1(PERR_afun9, (ATerm)ATmakeAppl0(PERR_afun38))), (ATerm)ATmakeAppl1(PERR_afun2, (ATerm)ATmakeAppl1(PERR_afun3, (ATerm)ATmakeAppl0(PERR_afun4)))), (ATerm)ATmakeAppl1(PERR_afun2, (ATerm)ATmakeAppl1(PERR_afun11, (ATerm)ATmakeAppl0(PERR_afun42)))), (ATerm)ATmakeAppl1(PERR_afun2, (ATerm)ATmakeAppl1(PERR_afun3, (ATerm)ATmakeAppl0(PERR_afun4)))), (ATerm)ATmakeAppl1(PERR_afun9, (ATerm)ATmakeAppl0(PERR_afun43))), (ATerm)ATmakeAppl1(PERR_afun2, (ATerm)ATmakeAppl1(PERR_afun3, (ATerm)ATmakeAppl0(PERR_afun4)))), (ATerm)ATmakeAppl1(PERR_afun2, (ATerm)ATmakeAppl1(PERR_afun11, (ATerm)ATmakeAppl0(PERR_afun29)))), (ATerm)ATmakeAppl1(PERR_afun2, (ATerm)ATmakeAppl1(PERR_afun3, (ATerm)ATmakeAppl0(PERR_afun4)))), (ATerm)ATmakeAppl1(PERR_afun9, (ATerm)ATmakeAppl0(PERR_afun39))), (ATerm)ATmakeAppl1(PERR_afun2, (ATerm)ATmakeAppl1(PERR_afun3, (ATerm)ATmakeAppl0(PERR_afun4)))), (ATerm)ATmakeAppl1(PERR_afun9, (ATerm)ATmakeAppl0(PERR_afun44))), (ATerm)ATmakeAppl1(PERR_afun2, (ATerm)ATmakeAppl1(PERR_afun11, (ATerm)ATmakeAppl0(PERR_afun41))), (ATerm)ATmakeAppl1(PERR_afun13, (ATerm)ATmakeList1((ATerm)ATmakeAppl1(PERR_afun14, (ATerm)ATmakeAppl1(PERR_afun15, (ATerm)ATmakeAppl0(PERR_afun44)))))), (ATerm)ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATmakeList1((ATerm)ATmakeAppl2(PERR_afun0, (ATerm)ATmakeAppl3(PERR_afun1, (ATerm)ATmakeList1((ATerm)ATmakeAppl1(PERR_afun17, (ATerm)ATmakeList1((ATerm)ATmakeInt(41)))), (ATerm)ATmakeAppl1(PERR_afun9, (ATerm)ATmakeAppl0(PERR_afun38)), (ATerm)ATmakeAppl0(PERR_afun5)), (ATerm)ATmakeList1((ATerm)ATmakeInt(41)))), (ATerm) wsAfterLocation), (ATerm) Location), (ATerm) wsAfterComma), (ATerm)ATmakeAppl2(PERR_afun0, (ATerm)ATmakeAppl3(PERR_afun1, (ATerm)ATmakeList1((ATerm)ATmakeAppl1(PERR_afun17, (ATerm)ATmakeList1((ATerm)ATmakeInt(44)))), (ATerm)ATmakeAppl1(PERR_afun9, (ATerm)ATmakeAppl0(PERR_afun43)), (ATerm)ATmakeAppl0(PERR_afun5)), (ATerm)ATmakeList1((ATerm)ATmakeInt(44)))), (ATerm) wsAfterDescription), (ATerm) description), (ATerm) wsAfterParenOpen), (ATerm)ATmakeAppl2(PERR_afun0, (ATerm)ATmakeAppl3(PERR_afun1, (ATerm)ATmakeList1((ATerm)ATmakeAppl1(PERR_afun17, (ATerm)ATmakeList1((ATerm)ATmakeInt(40)))), (ATerm)ATmakeAppl1(PERR_afun9, (ATerm)ATmakeAppl0(PERR_afun39)), (ATerm)ATmakeAppl0(PERR_afun5)), (ATerm)ATmakeList1((ATerm)ATmakeInt(40)))), (ATerm) wsAfterLocalized), (ATerm)ATmakeAppl2(PERR_afun0, (ATerm)ATmakeAppl3(PERR_afun1, (ATerm)ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATmakeList1((ATerm)ATmakeAppl1(PERR_afun17, (ATerm)ATmakeList1((ATerm)ATmakeInt(100)))), (ATerm)ATmakeAppl1(PERR_afun17, (ATerm)ATmakeList1((ATerm)ATmakeInt(101)))), (ATerm)ATmakeAppl1(PERR_afun17, (ATerm)ATmakeList1((ATerm)ATmakeInt(122)))), (ATerm)ATmakeAppl1(PERR_afun17, (ATerm)ATmakeList1((ATerm)ATmakeInt(105)))), (ATerm)ATmakeAppl1(PERR_afun17, (ATerm)ATmakeList1((ATerm)ATmakeInt(108)))), (ATerm)ATmakeAppl1(PERR_afun17, (ATerm)ATmakeList1((ATerm)ATmakeInt(97)))), (ATerm)ATmakeAppl1(PERR_afun17, (ATerm)ATmakeList1((ATerm)ATmakeInt(99)))), (ATerm)ATmakeAppl1(PERR_afun17, (ATerm)ATmakeList1((ATerm)ATmakeInt(111)))), (ATerm)ATmakeAppl1(PERR_afun17, (ATerm)ATmakeList1((ATerm)ATmakeInt(108)))), (ATerm)ATmakeAppl1(PERR_afun9, (ATerm)ATmakeAppl0(PERR_afun44)), (ATerm)ATmakeAppl0(PERR_afun5)), (ATerm)ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATmakeList1((ATerm)ATmakeInt(100)), (ATerm)ATmakeInt(101)), (ATerm)ATmakeInt(122)), (ATerm)ATmakeInt(105)), (ATerm)ATmakeInt(108)), (ATerm)ATmakeInt(97)), (ATerm)ATmakeInt(99)), (ATerm)ATmakeInt(111)), (ATerm)ATmakeInt(108)))));
}
/**
 * Constructs a info of type PERR_Error. Like all ATerm types, PERR_Errors are maximally shared.
 * \param[in] wsAfterInfo a child of the new info
 * \param[in] wsAfterParenOpen a child of the new info
 * \param[in] description a child of the new info
 * \param[in] wsAfterDescription a child of the new info
 * \param[in] wsAfterComma a child of the new info
 * \param[in] wsAfterBracketOpen a child of the new info
 * \param[in] list a child of the new info
 * \param[in] wsAfterList a child of the new info
 * \param[in] wsAfterBracketClose a child of the new info
 * \return A pointer to a info, either newly constructed or shared
 */
PERR_Error PERR_makeErrorInfo(PERR_OptLayout wsAfterInfo, PERR_OptLayout wsAfterParenOpen, PERR_StrCon description, PERR_OptLayout wsAfterDescription, PERR_OptLayout wsAfterComma, PERR_OptLayout wsAfterBracketOpen, PERR_SubjectList list, PERR_OptLayout wsAfterList, PERR_OptLayout wsAfterBracketClose) {
  return (PERR_Error)(ATerm)ATmakeAppl2(PERR_afun0, (ATerm)ATmakeAppl3(PERR_afun1, (ATerm)ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATmakeList1((ATerm)ATmakeAppl1(PERR_afun9, (ATerm)ATmakeAppl0(PERR_afun38))), (ATerm)ATmakeAppl1(PERR_afun2, (ATerm)ATmakeAppl1(PERR_afun3, (ATerm)ATmakeAppl0(PERR_afun4)))), (ATerm)ATmakeAppl1(PERR_afun9, (ATerm)ATmakeAppl0(PERR_afun45))), (ATerm)ATmakeAppl1(PERR_afun2, (ATerm)ATmakeAppl1(PERR_afun3, (ATerm)ATmakeAppl0(PERR_afun4)))), (ATerm)ATmakeAppl1(PERR_afun2, (ATerm)ATmakeAppl2(PERR_afun46, (ATerm)ATmakeAppl1(PERR_afun11, (ATerm)ATmakeAppl0(PERR_afun41)), (ATerm)ATmakeAppl1(PERR_afun9, (ATerm)ATmakeAppl0(PERR_afun43))))), (ATerm)ATmakeAppl1(PERR_afun2, (ATerm)ATmakeAppl1(PERR_afun3, (ATerm)ATmakeAppl0(PERR_afun4)))), (ATerm)ATmakeAppl1(PERR_afun9, (ATerm)ATmakeAppl0(PERR_afun47))), (ATerm)ATmakeAppl1(PERR_afun2, (ATerm)ATmakeAppl1(PERR_afun3, (ATerm)ATmakeAppl0(PERR_afun4)))), (ATerm)ATmakeAppl1(PERR_afun9, (ATerm)ATmakeAppl0(PERR_afun43))), (ATerm)ATmakeAppl1(PERR_afun2, (ATerm)ATmakeAppl1(PERR_afun3, (ATerm)ATmakeAppl0(PERR_afun4)))), (ATerm)ATmakeAppl1(PERR_afun2, (ATerm)ATmakeAppl1(PERR_afun11, (ATerm)ATmakeAppl0(PERR_afun29)))), (ATerm)ATmakeAppl1(PERR_afun2, (ATerm)ATmakeAppl1(PERR_afun3, (ATerm)ATmakeAppl0(PERR_afun4)))), (ATerm)ATmakeAppl1(PERR_afun9, (ATerm)ATmakeAppl0(PERR_afun39))), (ATerm)ATmakeAppl1(PERR_afun2, (ATerm)ATmakeAppl1(PERR_afun3, (ATerm)ATmakeAppl0(PERR_afun4)))), (ATerm)ATmakeAppl1(PERR_afun9, (ATerm)ATmakeAppl0(PERR_afun48))), (ATerm)ATmakeAppl1(PERR_afun2, (ATerm)ATmakeAppl1(PERR_afun11, (ATerm)ATmakeAppl0(PERR_afun37))), (ATerm)ATmakeAppl1(PERR_afun13, (ATerm)ATmakeList1((ATerm)ATmakeAppl1(PERR_afun14, (ATerm)ATmakeAppl1(PERR_afun15, (ATerm)ATmakeAppl0(PERR_afun48)))))), (ATerm)ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATmakeList1((ATerm)ATmakeAppl2(PERR_afun0, (ATerm)ATmakeAppl3(PERR_afun1, (ATerm)ATmakeList1((ATerm)ATmakeAppl1(PERR_afun17, (ATerm)ATmakeList1((ATerm)ATmakeInt(41)))), (ATerm)ATmakeAppl1(PERR_afun9, (ATerm)ATmakeAppl0(PERR_afun38)), (ATerm)ATmakeAppl0(PERR_afun5)), (ATerm)ATmakeList1((ATerm)ATmakeInt(41)))), (ATerm) wsAfterBracketClose), (ATerm)ATmakeAppl2(PERR_afun0, (ATerm)ATmakeAppl3(PERR_afun1, (ATerm)ATmakeList1((ATerm)ATmakeAppl1(PERR_afun17, (ATerm)ATmakeList1((ATerm)ATmakeInt(93)))), (ATerm)ATmakeAppl1(PERR_afun9, (ATerm)ATmakeAppl0(PERR_afun45)), (ATerm)ATmakeAppl0(PERR_afun5)), (ATerm)ATmakeList1((ATerm)ATmakeInt(93)))), (ATerm) wsAfterList), (ATerm)ATmakeAppl2(PERR_afun0, (ATerm)ATmakeAppl1(PERR_afun8, (ATerm)ATmakeAppl1(PERR_afun2, (ATerm)ATmakeAppl2(PERR_afun46, (ATerm)ATmakeAppl1(PERR_afun11, (ATerm)ATmakeAppl0(PERR_afun41)), (ATerm)ATmakeAppl1(PERR_afun9, (ATerm)ATmakeAppl0(PERR_afun43))))), (ATerm) list)), (ATerm) wsAfterBracketOpen), (ATerm)ATmakeAppl2(PERR_afun0, (ATerm)ATmakeAppl3(PERR_afun1, (ATerm)ATmakeList1((ATerm)ATmakeAppl1(PERR_afun17, (ATerm)ATmakeList1((ATerm)ATmakeInt(91)))), (ATerm)ATmakeAppl1(PERR_afun9, (ATerm)ATmakeAppl0(PERR_afun47)), (ATerm)ATmakeAppl0(PERR_afun5)), (ATerm)ATmakeList1((ATerm)ATmakeInt(91)))), (ATerm) wsAfterComma), (ATerm)ATmakeAppl2(PERR_afun0, (ATerm)ATmakeAppl3(PERR_afun1, (ATerm)ATmakeList1((ATerm)ATmakeAppl1(PERR_afun17, (ATerm)ATmakeList1((ATerm)ATmakeInt(44)))), (ATerm)ATmakeAppl1(PERR_afun9, (ATerm)ATmakeAppl0(PERR_afun43)), (ATerm)ATmakeAppl0(PERR_afun5)), (ATerm)ATmakeList1((ATerm)ATmakeInt(44)))), (ATerm) wsAfterDescription), (ATerm) description), (ATerm) wsAfterParenOpen), (ATerm)ATmakeAppl2(PERR_afun0, (ATerm)ATmakeAppl3(PERR_afun1, (ATerm)ATmakeList1((ATerm)ATmakeAppl1(PERR_afun17, (ATerm)ATmakeList1((ATerm)ATmakeInt(40)))), (ATerm)ATmakeAppl1(PERR_afun9, (ATerm)ATmakeAppl0(PERR_afun39)), (ATerm)ATmakeAppl0(PERR_afun5)), (ATerm)ATmakeList1((ATerm)ATmakeInt(40)))), (ATerm) wsAfterInfo), (ATerm)ATmakeAppl2(PERR_afun0, (ATerm)ATmakeAppl3(PERR_afun1, (ATerm)ATinsert(ATinsert(ATinsert(ATmakeList1((ATerm)ATmakeAppl1(PERR_afun17, (ATerm)ATmakeList1((ATerm)ATmakeInt(111)))), (ATerm)ATmakeAppl1(PERR_afun17, (ATerm)ATmakeList1((ATerm)ATmakeInt(102)))), (ATerm)ATmakeAppl1(PERR_afun17, (ATerm)ATmakeList1((ATerm)ATmakeInt(110)))), (ATerm)ATmakeAppl1(PERR_afun17, (ATerm)ATmakeList1((ATerm)ATmakeInt(105)))), (ATerm)ATmakeAppl1(PERR_afun9, (ATerm)ATmakeAppl0(PERR_afun48)), (ATerm)ATmakeAppl0(PERR_afun5)), (ATerm)ATinsert(ATinsert(ATinsert(ATmakeList1((ATerm)ATmakeInt(111)), (ATerm)ATmakeInt(102)), (ATerm)ATmakeInt(110)), (ATerm)ATmakeInt(105)))));
}
/**
 * Constructs a warning of type PERR_Error. Like all ATerm types, PERR_Errors are maximally shared.
 * \param[in] wsAfterWarning a child of the new warning
 * \param[in] wsAfterParenOpen a child of the new warning
 * \param[in] description a child of the new warning
 * \param[in] wsAfterDescription a child of the new warning
 * \param[in] wsAfterComma a child of the new warning
 * \param[in] wsAfterBracketOpen a child of the new warning
 * \param[in] list a child of the new warning
 * \param[in] wsAfterList a child of the new warning
 * \param[in] wsAfterBracketClose a child of the new warning
 * \return A pointer to a warning, either newly constructed or shared
 */
PERR_Error PERR_makeErrorWarning(PERR_OptLayout wsAfterWarning, PERR_OptLayout wsAfterParenOpen, PERR_StrCon description, PERR_OptLayout wsAfterDescription, PERR_OptLayout wsAfterComma, PERR_OptLayout wsAfterBracketOpen, PERR_SubjectList list, PERR_OptLayout wsAfterList, PERR_OptLayout wsAfterBracketClose) {
  return (PERR_Error)(ATerm)ATmakeAppl2(PERR_afun0, (ATerm)ATmakeAppl3(PERR_afun1, (ATerm)ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATmakeList1((ATerm)ATmakeAppl1(PERR_afun9, (ATerm)ATmakeAppl0(PERR_afun38))), (ATerm)ATmakeAppl1(PERR_afun2, (ATerm)ATmakeAppl1(PERR_afun3, (ATerm)ATmakeAppl0(PERR_afun4)))), (ATerm)ATmakeAppl1(PERR_afun9, (ATerm)ATmakeAppl0(PERR_afun45))), (ATerm)ATmakeAppl1(PERR_afun2, (ATerm)ATmakeAppl1(PERR_afun3, (ATerm)ATmakeAppl0(PERR_afun4)))), (ATerm)ATmakeAppl1(PERR_afun2, (ATerm)ATmakeAppl2(PERR_afun46, (ATerm)ATmakeAppl1(PERR_afun11, (ATerm)ATmakeAppl0(PERR_afun41)), (ATerm)ATmakeAppl1(PERR_afun9, (ATerm)ATmakeAppl0(PERR_afun43))))), (ATerm)ATmakeAppl1(PERR_afun2, (ATerm)ATmakeAppl1(PERR_afun3, (ATerm)ATmakeAppl0(PERR_afun4)))), (ATerm)ATmakeAppl1(PERR_afun9, (ATerm)ATmakeAppl0(PERR_afun47))), (ATerm)ATmakeAppl1(PERR_afun2, (ATerm)ATmakeAppl1(PERR_afun3, (ATerm)ATmakeAppl0(PERR_afun4)))), (ATerm)ATmakeAppl1(PERR_afun9, (ATerm)ATmakeAppl0(PERR_afun43))), (ATerm)ATmakeAppl1(PERR_afun2, (ATerm)ATmakeAppl1(PERR_afun3, (ATerm)ATmakeAppl0(PERR_afun4)))), (ATerm)ATmakeAppl1(PERR_afun2, (ATerm)ATmakeAppl1(PERR_afun11, (ATerm)ATmakeAppl0(PERR_afun29)))), (ATerm)ATmakeAppl1(PERR_afun2, (ATerm)ATmakeAppl1(PERR_afun3, (ATerm)ATmakeAppl0(PERR_afun4)))), (ATerm)ATmakeAppl1(PERR_afun9, (ATerm)ATmakeAppl0(PERR_afun39))), (ATerm)ATmakeAppl1(PERR_afun2, (ATerm)ATmakeAppl1(PERR_afun3, (ATerm)ATmakeAppl0(PERR_afun4)))), (ATerm)ATmakeAppl1(PERR_afun9, (ATerm)ATmakeAppl0(PERR_afun49))), (ATerm)ATmakeAppl1(PERR_afun2, (ATerm)ATmakeAppl1(PERR_afun11, (ATerm)ATmakeAppl0(PERR_afun37))), (ATerm)ATmakeAppl1(PERR_afun13, (ATerm)ATmakeList1((ATerm)ATmakeAppl1(PERR_afun14, (ATerm)ATmakeAppl1(PERR_afun15, (ATerm)ATmakeAppl0(PERR_afun49)))))), (ATerm)ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATmakeList1((ATerm)ATmakeAppl2(PERR_afun0, (ATerm)ATmakeAppl3(PERR_afun1, (ATerm)ATmakeList1((ATerm)ATmakeAppl1(PERR_afun17, (ATerm)ATmakeList1((ATerm)ATmakeInt(41)))), (ATerm)ATmakeAppl1(PERR_afun9, (ATerm)ATmakeAppl0(PERR_afun38)), (ATerm)ATmakeAppl0(PERR_afun5)), (ATerm)ATmakeList1((ATerm)ATmakeInt(41)))), (ATerm) wsAfterBracketClose), (ATerm)ATmakeAppl2(PERR_afun0, (ATerm)ATmakeAppl3(PERR_afun1, (ATerm)ATmakeList1((ATerm)ATmakeAppl1(PERR_afun17, (ATerm)ATmakeList1((ATerm)ATmakeInt(93)))), (ATerm)ATmakeAppl1(PERR_afun9, (ATerm)ATmakeAppl0(PERR_afun45)), (ATerm)ATmakeAppl0(PERR_afun5)), (ATerm)ATmakeList1((ATerm)ATmakeInt(93)))), (ATerm) wsAfterList), (ATerm)ATmakeAppl2(PERR_afun0, (ATerm)ATmakeAppl1(PERR_afun8, (ATerm)ATmakeAppl1(PERR_afun2, (ATerm)ATmakeAppl2(PERR_afun46, (ATerm)ATmakeAppl1(PERR_afun11, (ATerm)ATmakeAppl0(PERR_afun41)), (ATerm)ATmakeAppl1(PERR_afun9, (ATerm)ATmakeAppl0(PERR_afun43))))), (ATerm) list)), (ATerm) wsAfterBracketOpen), (ATerm)ATmakeAppl2(PERR_afun0, (ATerm)ATmakeAppl3(PERR_afun1, (ATerm)ATmakeList1((ATerm)ATmakeAppl1(PERR_afun17, (ATerm)ATmakeList1((ATerm)ATmakeInt(91)))), (ATerm)ATmakeAppl1(PERR_afun9, (ATerm)ATmakeAppl0(PERR_afun47)), (ATerm)ATmakeAppl0(PERR_afun5)), (ATerm)ATmakeList1((ATerm)ATmakeInt(91)))), (ATerm) wsAfterComma), (ATerm)ATmakeAppl2(PERR_afun0, (ATerm)ATmakeAppl3(PERR_afun1, (ATerm)ATmakeList1((ATerm)ATmakeAppl1(PERR_afun17, (ATerm)ATmakeList1((ATerm)ATmakeInt(44)))), (ATerm)ATmakeAppl1(PERR_afun9, (ATerm)ATmakeAppl0(PERR_afun43)), (ATerm)ATmakeAppl0(PERR_afun5)), (ATerm)ATmakeList1((ATerm)ATmakeInt(44)))), (ATerm) wsAfterDescription), (ATerm) description), (ATerm) wsAfterParenOpen), (ATerm)ATmakeAppl2(PERR_afun0, (ATerm)ATmakeAppl3(PERR_afun1, (ATerm)ATmakeList1((ATerm)ATmakeAppl1(PERR_afun17, (ATerm)ATmakeList1((ATerm)ATmakeInt(40)))), (ATerm)ATmakeAppl1(PERR_afun9, (ATerm)ATmakeAppl0(PERR_afun39)), (ATerm)ATmakeAppl0(PERR_afun5)), (ATerm)ATmakeList1((ATerm)ATmakeInt(40)))), (ATerm) wsAfterWarning), (ATerm)ATmakeAppl2(PERR_afun0, (ATerm)ATmakeAppl3(PERR_afun1, (ATerm)ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATmakeList1((ATerm)ATmakeAppl1(PERR_afun17, (ATerm)ATmakeList1((ATerm)ATmakeInt(103)))), (ATerm)ATmakeAppl1(PERR_afun17, (ATerm)ATmakeList1((ATerm)ATmakeInt(110)))), (ATerm)ATmakeAppl1(PERR_afun17, (ATerm)ATmakeList1((ATerm)ATmakeInt(105)))), (ATerm)ATmakeAppl1(PERR_afun17, (ATerm)ATmakeList1((ATerm)ATmakeInt(110)))), (ATerm)ATmakeAppl1(PERR_afun17, (ATerm)ATmakeList1((ATerm)ATmakeInt(114)))), (ATerm)ATmakeAppl1(PERR_afun17, (ATerm)ATmakeList1((ATerm)ATmakeInt(97)))), (ATerm)ATmakeAppl1(PERR_afun17, (ATerm)ATmakeList1((ATerm)ATmakeInt(119)))), (ATerm)ATmakeAppl1(PERR_afun9, (ATerm)ATmakeAppl0(PERR_afun49)), (ATerm)ATmakeAppl0(PERR_afun5)), (ATerm)ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATmakeList1((ATerm)ATmakeInt(103)), (ATerm)ATmakeInt(110)), (ATerm)ATmakeInt(105)), (ATerm)ATmakeInt(110)), (ATerm)ATmakeInt(114)), (ATerm)ATmakeInt(97)), (ATerm)ATmakeInt(119)))));
}
/**
 * Constructs a error of type PERR_Error. Like all ATerm types, PERR_Errors are maximally shared.
 * \param[in] wsAfterError a child of the new error
 * \param[in] wsAfterParenOpen a child of the new error
 * \param[in] description a child of the new error
 * \param[in] wsAfterDescription a child of the new error
 * \param[in] wsAfterComma a child of the new error
 * \param[in] wsAfterBracketOpen a child of the new error
 * \param[in] list a child of the new error
 * \param[in] wsAfterList a child of the new error
 * \param[in] wsAfterBracketClose a child of the new error
 * \return A pointer to a error, either newly constructed or shared
 */
PERR_Error PERR_makeErrorError(PERR_OptLayout wsAfterError, PERR_OptLayout wsAfterParenOpen, PERR_StrCon description, PERR_OptLayout wsAfterDescription, PERR_OptLayout wsAfterComma, PERR_OptLayout wsAfterBracketOpen, PERR_SubjectList list, PERR_OptLayout wsAfterList, PERR_OptLayout wsAfterBracketClose) {
  return (PERR_Error)(ATerm)ATmakeAppl2(PERR_afun0, (ATerm)ATmakeAppl3(PERR_afun1, (ATerm)ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATmakeList1((ATerm)ATmakeAppl1(PERR_afun9, (ATerm)ATmakeAppl0(PERR_afun38))), (ATerm)ATmakeAppl1(PERR_afun2, (ATerm)ATmakeAppl1(PERR_afun3, (ATerm)ATmakeAppl0(PERR_afun4)))), (ATerm)ATmakeAppl1(PERR_afun9, (ATerm)ATmakeAppl0(PERR_afun45))), (ATerm)ATmakeAppl1(PERR_afun2, (ATerm)ATmakeAppl1(PERR_afun3, (ATerm)ATmakeAppl0(PERR_afun4)))), (ATerm)ATmakeAppl1(PERR_afun2, (ATerm)ATmakeAppl2(PERR_afun46, (ATerm)ATmakeAppl1(PERR_afun11, (ATerm)ATmakeAppl0(PERR_afun41)), (ATerm)ATmakeAppl1(PERR_afun9, (ATerm)ATmakeAppl0(PERR_afun43))))), (ATerm)ATmakeAppl1(PERR_afun2, (ATerm)ATmakeAppl1(PERR_afun3, (ATerm)ATmakeAppl0(PERR_afun4)))), (ATerm)ATmakeAppl1(PERR_afun9, (ATerm)ATmakeAppl0(PERR_afun47))), (ATerm)ATmakeAppl1(PERR_afun2, (ATerm)ATmakeAppl1(PERR_afun3, (ATerm)ATmakeAppl0(PERR_afun4)))), (ATerm)ATmakeAppl1(PERR_afun9, (ATerm)ATmakeAppl0(PERR_afun43))), (ATerm)ATmakeAppl1(PERR_afun2, (ATerm)ATmakeAppl1(PERR_afun3, (ATerm)ATmakeAppl0(PERR_afun4)))), (ATerm)ATmakeAppl1(PERR_afun2, (ATerm)ATmakeAppl1(PERR_afun11, (ATerm)ATmakeAppl0(PERR_afun29)))), (ATerm)ATmakeAppl1(PERR_afun2, (ATerm)ATmakeAppl1(PERR_afun3, (ATerm)ATmakeAppl0(PERR_afun4)))), (ATerm)ATmakeAppl1(PERR_afun9, (ATerm)ATmakeAppl0(PERR_afun39))), (ATerm)ATmakeAppl1(PERR_afun2, (ATerm)ATmakeAppl1(PERR_afun3, (ATerm)ATmakeAppl0(PERR_afun4)))), (ATerm)ATmakeAppl1(PERR_afun9, (ATerm)ATmakeAppl0(PERR_afun50))), (ATerm)ATmakeAppl1(PERR_afun2, (ATerm)ATmakeAppl1(PERR_afun11, (ATerm)ATmakeAppl0(PERR_afun37))), (ATerm)ATmakeAppl1(PERR_afun13, (ATerm)ATmakeList1((ATerm)ATmakeAppl1(PERR_afun14, (ATerm)ATmakeAppl1(PERR_afun15, (ATerm)ATmakeAppl0(PERR_afun50)))))), (ATerm)ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATmakeList1((ATerm)ATmakeAppl2(PERR_afun0, (ATerm)ATmakeAppl3(PERR_afun1, (ATerm)ATmakeList1((ATerm)ATmakeAppl1(PERR_afun17, (ATerm)ATmakeList1((ATerm)ATmakeInt(41)))), (ATerm)ATmakeAppl1(PERR_afun9, (ATerm)ATmakeAppl0(PERR_afun38)), (ATerm)ATmakeAppl0(PERR_afun5)), (ATerm)ATmakeList1((ATerm)ATmakeInt(41)))), (ATerm) wsAfterBracketClose), (ATerm)ATmakeAppl2(PERR_afun0, (ATerm)ATmakeAppl3(PERR_afun1, (ATerm)ATmakeList1((ATerm)ATmakeAppl1(PERR_afun17, (ATerm)ATmakeList1((ATerm)ATmakeInt(93)))), (ATerm)ATmakeAppl1(PERR_afun9, (ATerm)ATmakeAppl0(PERR_afun45)), (ATerm)ATmakeAppl0(PERR_afun5)), (ATerm)ATmakeList1((ATerm)ATmakeInt(93)))), (ATerm) wsAfterList), (ATerm)ATmakeAppl2(PERR_afun0, (ATerm)ATmakeAppl1(PERR_afun8, (ATerm)ATmakeAppl1(PERR_afun2, (ATerm)ATmakeAppl2(PERR_afun46, (ATerm)ATmakeAppl1(PERR_afun11, (ATerm)ATmakeAppl0(PERR_afun41)), (ATerm)ATmakeAppl1(PERR_afun9, (ATerm)ATmakeAppl0(PERR_afun43))))), (ATerm) list)), (ATerm) wsAfterBracketOpen), (ATerm)ATmakeAppl2(PERR_afun0, (ATerm)ATmakeAppl3(PERR_afun1, (ATerm)ATmakeList1((ATerm)ATmakeAppl1(PERR_afun17, (ATerm)ATmakeList1((ATerm)ATmakeInt(91)))), (ATerm)ATmakeAppl1(PERR_afun9, (ATerm)ATmakeAppl0(PERR_afun47)), (ATerm)ATmakeAppl0(PERR_afun5)), (ATerm)ATmakeList1((ATerm)ATmakeInt(91)))), (ATerm) wsAfterComma), (ATerm)ATmakeAppl2(PERR_afun0, (ATerm)ATmakeAppl3(PERR_afun1, (ATerm)ATmakeList1((ATerm)ATmakeAppl1(PERR_afun17, (ATerm)ATmakeList1((ATerm)ATmakeInt(44)))), (ATerm)ATmakeAppl1(PERR_afun9, (ATerm)ATmakeAppl0(PERR_afun43)), (ATerm)ATmakeAppl0(PERR_afun5)), (ATerm)ATmakeList1((ATerm)ATmakeInt(44)))), (ATerm) wsAfterDescription), (ATerm) description), (ATerm) wsAfterParenOpen), (ATerm)ATmakeAppl2(PERR_afun0, (ATerm)ATmakeAppl3(PERR_afun1, (ATerm)ATmakeList1((ATerm)ATmakeAppl1(PERR_afun17, (ATerm)ATmakeList1((ATerm)ATmakeInt(40)))), (ATerm)ATmakeAppl1(PERR_afun9, (ATerm)ATmakeAppl0(PERR_afun39)), (ATerm)ATmakeAppl0(PERR_afun5)), (ATerm)ATmakeList1((ATerm)ATmakeInt(40)))), (ATerm) wsAfterError), (ATerm)ATmakeAppl2(PERR_afun0, (ATerm)ATmakeAppl3(PERR_afun1, (ATerm)ATinsert(ATinsert(ATinsert(ATinsert(ATmakeList1((ATerm)ATmakeAppl1(PERR_afun17, (ATerm)ATmakeList1((ATerm)ATmakeInt(114)))), (ATerm)ATmakeAppl1(PERR_afun17, (ATerm)ATmakeList1((ATerm)ATmakeInt(111)))), (ATerm)ATmakeAppl1(PERR_afun17, (ATerm)ATmakeList1((ATerm)ATmakeInt(114)))), (ATerm)ATmakeAppl1(PERR_afun17, (ATerm)ATmakeList1((ATerm)ATmakeInt(114)))), (ATerm)ATmakeAppl1(PERR_afun17, (ATerm)ATmakeList1((ATerm)ATmakeInt(101)))), (ATerm)ATmakeAppl1(PERR_afun9, (ATerm)ATmakeAppl0(PERR_afun50)), (ATerm)ATmakeAppl0(PERR_afun5)), (ATerm)ATinsert(ATinsert(ATinsert(ATinsert(ATmakeList1((ATerm)ATmakeInt(114)), (ATerm)ATmakeInt(111)), (ATerm)ATmakeInt(114)), (ATerm)ATmakeInt(114)), (ATerm)ATmakeInt(101)))));
}
/**
 * Constructs a fatal of type PERR_Error. Like all ATerm types, PERR_Errors are maximally shared.
 * \param[in] wsAfterFatal a child of the new fatal
 * \param[in] wsAfterParenOpen a child of the new fatal
 * \param[in] description a child of the new fatal
 * \param[in] wsAfterDescription a child of the new fatal
 * \param[in] wsAfterComma a child of the new fatal
 * \param[in] wsAfterBracketOpen a child of the new fatal
 * \param[in] list a child of the new fatal
 * \param[in] wsAfterList a child of the new fatal
 * \param[in] wsAfterBracketClose a child of the new fatal
 * \return A pointer to a fatal, either newly constructed or shared
 */
PERR_Error PERR_makeErrorFatal(PERR_OptLayout wsAfterFatal, PERR_OptLayout wsAfterParenOpen, PERR_StrCon description, PERR_OptLayout wsAfterDescription, PERR_OptLayout wsAfterComma, PERR_OptLayout wsAfterBracketOpen, PERR_SubjectList list, PERR_OptLayout wsAfterList, PERR_OptLayout wsAfterBracketClose) {
  return (PERR_Error)(ATerm)ATmakeAppl2(PERR_afun0, (ATerm)ATmakeAppl3(PERR_afun1, (ATerm)ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATmakeList1((ATerm)ATmakeAppl1(PERR_afun9, (ATerm)ATmakeAppl0(PERR_afun38))), (ATerm)ATmakeAppl1(PERR_afun2, (ATerm)ATmakeAppl1(PERR_afun3, (ATerm)ATmakeAppl0(PERR_afun4)))), (ATerm)ATmakeAppl1(PERR_afun9, (ATerm)ATmakeAppl0(PERR_afun45))), (ATerm)ATmakeAppl1(PERR_afun2, (ATerm)ATmakeAppl1(PERR_afun3, (ATerm)ATmakeAppl0(PERR_afun4)))), (ATerm)ATmakeAppl1(PERR_afun2, (ATerm)ATmakeAppl2(PERR_afun46, (ATerm)ATmakeAppl1(PERR_afun11, (ATerm)ATmakeAppl0(PERR_afun41)), (ATerm)ATmakeAppl1(PERR_afun9, (ATerm)ATmakeAppl0(PERR_afun43))))), (ATerm)ATmakeAppl1(PERR_afun2, (ATerm)ATmakeAppl1(PERR_afun3, (ATerm)ATmakeAppl0(PERR_afun4)))), (ATerm)ATmakeAppl1(PERR_afun9, (ATerm)ATmakeAppl0(PERR_afun47))), (ATerm)ATmakeAppl1(PERR_afun2, (ATerm)ATmakeAppl1(PERR_afun3, (ATerm)ATmakeAppl0(PERR_afun4)))), (ATerm)ATmakeAppl1(PERR_afun9, (ATerm)ATmakeAppl0(PERR_afun43))), (ATerm)ATmakeAppl1(PERR_afun2, (ATerm)ATmakeAppl1(PERR_afun3, (ATerm)ATmakeAppl0(PERR_afun4)))), (ATerm)ATmakeAppl1(PERR_afun2, (ATerm)ATmakeAppl1(PERR_afun11, (ATerm)ATmakeAppl0(PERR_afun29)))), (ATerm)ATmakeAppl1(PERR_afun2, (ATerm)ATmakeAppl1(PERR_afun3, (ATerm)ATmakeAppl0(PERR_afun4)))), (ATerm)ATmakeAppl1(PERR_afun9, (ATerm)ATmakeAppl0(PERR_afun39))), (ATerm)ATmakeAppl1(PERR_afun2, (ATerm)ATmakeAppl1(PERR_afun3, (ATerm)ATmakeAppl0(PERR_afun4)))), (ATerm)ATmakeAppl1(PERR_afun9, (ATerm)ATmakeAppl0(PERR_afun51))), (ATerm)ATmakeAppl1(PERR_afun2, (ATerm)ATmakeAppl1(PERR_afun11, (ATerm)ATmakeAppl0(PERR_afun37))), (ATerm)ATmakeAppl1(PERR_afun13, (ATerm)ATmakeList1((ATerm)ATmakeAppl1(PERR_afun14, (ATerm)ATmakeAppl1(PERR_afun15, (ATerm)ATmakeAppl0(PERR_afun51)))))), (ATerm)ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATmakeList1((ATerm)ATmakeAppl2(PERR_afun0, (ATerm)ATmakeAppl3(PERR_afun1, (ATerm)ATmakeList1((ATerm)ATmakeAppl1(PERR_afun17, (ATerm)ATmakeList1((ATerm)ATmakeInt(41)))), (ATerm)ATmakeAppl1(PERR_afun9, (ATerm)ATmakeAppl0(PERR_afun38)), (ATerm)ATmakeAppl0(PERR_afun5)), (ATerm)ATmakeList1((ATerm)ATmakeInt(41)))), (ATerm) wsAfterBracketClose), (ATerm)ATmakeAppl2(PERR_afun0, (ATerm)ATmakeAppl3(PERR_afun1, (ATerm)ATmakeList1((ATerm)ATmakeAppl1(PERR_afun17, (ATerm)ATmakeList1((ATerm)ATmakeInt(93)))), (ATerm)ATmakeAppl1(PERR_afun9, (ATerm)ATmakeAppl0(PERR_afun45)), (ATerm)ATmakeAppl0(PERR_afun5)), (ATerm)ATmakeList1((ATerm)ATmakeInt(93)))), (ATerm) wsAfterList), (ATerm)ATmakeAppl2(PERR_afun0, (ATerm)ATmakeAppl1(PERR_afun8, (ATerm)ATmakeAppl1(PERR_afun2, (ATerm)ATmakeAppl2(PERR_afun46, (ATerm)ATmakeAppl1(PERR_afun11, (ATerm)ATmakeAppl0(PERR_afun41)), (ATerm)ATmakeAppl1(PERR_afun9, (ATerm)ATmakeAppl0(PERR_afun43))))), (ATerm) list)), (ATerm) wsAfterBracketOpen), (ATerm)ATmakeAppl2(PERR_afun0, (ATerm)ATmakeAppl3(PERR_afun1, (ATerm)ATmakeList1((ATerm)ATmakeAppl1(PERR_afun17, (ATerm)ATmakeList1((ATerm)ATmakeInt(91)))), (ATerm)ATmakeAppl1(PERR_afun9, (ATerm)ATmakeAppl0(PERR_afun47)), (ATerm)ATmakeAppl0(PERR_afun5)), (ATerm)ATmakeList1((ATerm)ATmakeInt(91)))), (ATerm) wsAfterComma), (ATerm)ATmakeAppl2(PERR_afun0, (ATerm)ATmakeAppl3(PERR_afun1, (ATerm)ATmakeList1((ATerm)ATmakeAppl1(PERR_afun17, (ATerm)ATmakeList1((ATerm)ATmakeInt(44)))), (ATerm)ATmakeAppl1(PERR_afun9, (ATerm)ATmakeAppl0(PERR_afun43)), (ATerm)ATmakeAppl0(PERR_afun5)), (ATerm)ATmakeList1((ATerm)ATmakeInt(44)))), (ATerm) wsAfterDescription), (ATerm) description), (ATerm) wsAfterParenOpen), (ATerm)ATmakeAppl2(PERR_afun0, (ATerm)ATmakeAppl3(PERR_afun1, (ATerm)ATmakeList1((ATerm)ATmakeAppl1(PERR_afun17, (ATerm)ATmakeList1((ATerm)ATmakeInt(40)))), (ATerm)ATmakeAppl1(PERR_afun9, (ATerm)ATmakeAppl0(PERR_afun39)), (ATerm)ATmakeAppl0(PERR_afun5)), (ATerm)ATmakeList1((ATerm)ATmakeInt(40)))), (ATerm) wsAfterFatal), (ATerm)ATmakeAppl2(PERR_afun0, (ATerm)ATmakeAppl3(PERR_afun1, (ATerm)ATinsert(ATinsert(ATinsert(ATinsert(ATmakeList1((ATerm)ATmakeAppl1(PERR_afun17, (ATerm)ATmakeList1((ATerm)ATmakeInt(108)))), (ATerm)ATmakeAppl1(PERR_afun17, (ATerm)ATmakeList1((ATerm)ATmakeInt(97)))), (ATerm)ATmakeAppl1(PERR_afun17, (ATerm)ATmakeList1((ATerm)ATmakeInt(116)))), (ATerm)ATmakeAppl1(PERR_afun17, (ATerm)ATmakeList1((ATerm)ATmakeInt(97)))), (ATerm)ATmakeAppl1(PERR_afun17, (ATerm)ATmakeList1((ATerm)ATmakeInt(102)))), (ATerm)ATmakeAppl1(PERR_afun9, (ATerm)ATmakeAppl0(PERR_afun51)), (ATerm)ATmakeAppl0(PERR_afun5)), (ATerm)ATinsert(ATinsert(ATinsert(ATinsert(ATmakeList1((ATerm)ATmakeInt(108)), (ATerm)ATmakeInt(97)), (ATerm)ATmakeInt(116)), (ATerm)ATmakeInt(97)), (ATerm)ATmakeInt(102)))));
}
/**
 * Constructs a summary of type PERR_Summary. Like all ATerm types, PERR_Summarys are maximally shared.
 * \param[in] wsAfterSummary a child of the new summary
 * \param[in] wsAfterParenOpen a child of the new summary
 * \param[in] producer a child of the new summary
 * \param[in] wsAfterProducer a child of the new summary
 * \param[in] wsAfterComma a child of the new summary
 * \param[in] id a child of the new summary
 * \param[in] wsAfterId a child of the new summary
 * \param[in] wsAfterComma1 a child of the new summary
 * \param[in] wsAfterBracketOpen a child of the new summary
 * \param[in] list a child of the new summary
 * \param[in] wsAfterList a child of the new summary
 * \param[in] wsAfterBracketClose a child of the new summary
 * \return A pointer to a summary, either newly constructed or shared
 */
PERR_Summary PERR_makeSummarySummary(PERR_OptLayout wsAfterSummary, PERR_OptLayout wsAfterParenOpen, PERR_StrCon producer, PERR_OptLayout wsAfterProducer, PERR_OptLayout wsAfterComma, PERR_StrCon id, PERR_OptLayout wsAfterId, PERR_OptLayout wsAfterComma1, PERR_OptLayout wsAfterBracketOpen, PERR_ErrorList list, PERR_OptLayout wsAfterList, PERR_OptLayout wsAfterBracketClose) {
  return (PERR_Summary)(ATerm)ATmakeAppl2(PERR_afun0, (ATerm)ATmakeAppl3(PERR_afun1, (ATerm)ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATmakeList1((ATerm)ATmakeAppl1(PERR_afun9, (ATerm)ATmakeAppl0(PERR_afun38))), (ATerm)ATmakeAppl1(PERR_afun2, (ATerm)ATmakeAppl1(PERR_afun3, (ATerm)ATmakeAppl0(PERR_afun4)))), (ATerm)ATmakeAppl1(PERR_afun9, (ATerm)ATmakeAppl0(PERR_afun45))), (ATerm)ATmakeAppl1(PERR_afun2, (ATerm)ATmakeAppl1(PERR_afun3, (ATerm)ATmakeAppl0(PERR_afun4)))), (ATerm)ATmakeAppl1(PERR_afun2, (ATerm)ATmakeAppl2(PERR_afun46, (ATerm)ATmakeAppl1(PERR_afun11, (ATerm)ATmakeAppl0(PERR_afun37)), (ATerm)ATmakeAppl1(PERR_afun9, (ATerm)ATmakeAppl0(PERR_afun43))))), (ATerm)ATmakeAppl1(PERR_afun2, (ATerm)ATmakeAppl1(PERR_afun3, (ATerm)ATmakeAppl0(PERR_afun4)))), (ATerm)ATmakeAppl1(PERR_afun9, (ATerm)ATmakeAppl0(PERR_afun47))), (ATerm)ATmakeAppl1(PERR_afun2, (ATerm)ATmakeAppl1(PERR_afun3, (ATerm)ATmakeAppl0(PERR_afun4)))), (ATerm)ATmakeAppl1(PERR_afun9, (ATerm)ATmakeAppl0(PERR_afun43))), (ATerm)ATmakeAppl1(PERR_afun2, (ATerm)ATmakeAppl1(PERR_afun3, (ATerm)ATmakeAppl0(PERR_afun4)))), (ATerm)ATmakeAppl1(PERR_afun2, (ATerm)ATmakeAppl1(PERR_afun11, (ATerm)ATmakeAppl0(PERR_afun29)))), (ATerm)ATmakeAppl1(PERR_afun2, (ATerm)ATmakeAppl1(PERR_afun3, (ATerm)ATmakeAppl0(PERR_afun4)))), (ATerm)ATmakeAppl1(PERR_afun9, (ATerm)ATmakeAppl0(PERR_afun43))), (ATerm)ATmakeAppl1(PERR_afun2, (ATerm)ATmakeAppl1(PERR_afun3, (ATerm)ATmakeAppl0(PERR_afun4)))), (ATerm)ATmakeAppl1(PERR_afun2, (ATerm)ATmakeAppl1(PERR_afun11, (ATerm)ATmakeAppl0(PERR_afun29)))), (ATerm)ATmakeAppl1(PERR_afun2, (ATerm)ATmakeAppl1(PERR_afun3, (ATerm)ATmakeAppl0(PERR_afun4)))), (ATerm)ATmakeAppl1(PERR_afun9, (ATerm)ATmakeAppl0(PERR_afun39))), (ATerm)ATmakeAppl1(PERR_afun2, (ATerm)ATmakeAppl1(PERR_afun3, (ATerm)ATmakeAppl0(PERR_afun4)))), (ATerm)ATmakeAppl1(PERR_afun9, (ATerm)ATmakeAppl0(PERR_afun52))), (ATerm)ATmakeAppl1(PERR_afun2, (ATerm)ATmakeAppl1(PERR_afun11, (ATerm)ATmakeAppl0(PERR_afun35))), (ATerm)ATmakeAppl1(PERR_afun13, (ATerm)ATmakeList1((ATerm)ATmakeAppl1(PERR_afun14, (ATerm)ATmakeAppl1(PERR_afun15, (ATerm)ATmakeAppl0(PERR_afun52)))))), (ATerm)ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATmakeList1((ATerm)ATmakeAppl2(PERR_afun0, (ATerm)ATmakeAppl3(PERR_afun1, (ATerm)ATmakeList1((ATerm)ATmakeAppl1(PERR_afun17, (ATerm)ATmakeList1((ATerm)ATmakeInt(41)))), (ATerm)ATmakeAppl1(PERR_afun9, (ATerm)ATmakeAppl0(PERR_afun38)), (ATerm)ATmakeAppl0(PERR_afun5)), (ATerm)ATmakeList1((ATerm)ATmakeInt(41)))), (ATerm) wsAfterBracketClose), (ATerm)ATmakeAppl2(PERR_afun0, (ATerm)ATmakeAppl3(PERR_afun1, (ATerm)ATmakeList1((ATerm)ATmakeAppl1(PERR_afun17, (ATerm)ATmakeList1((ATerm)ATmakeInt(93)))), (ATerm)ATmakeAppl1(PERR_afun9, (ATerm)ATmakeAppl0(PERR_afun45)), (ATerm)ATmakeAppl0(PERR_afun5)), (ATerm)ATmakeList1((ATerm)ATmakeInt(93)))), (ATerm) wsAfterList), (ATerm)ATmakeAppl2(PERR_afun0, (ATerm)ATmakeAppl1(PERR_afun8, (ATerm)ATmakeAppl1(PERR_afun2, (ATerm)ATmakeAppl2(PERR_afun46, (ATerm)ATmakeAppl1(PERR_afun11, (ATerm)ATmakeAppl0(PERR_afun37)), (ATerm)ATmakeAppl1(PERR_afun9, (ATerm)ATmakeAppl0(PERR_afun43))))), (ATerm) list)), (ATerm) wsAfterBracketOpen), (ATerm)ATmakeAppl2(PERR_afun0, (ATerm)ATmakeAppl3(PERR_afun1, (ATerm)ATmakeList1((ATerm)ATmakeAppl1(PERR_afun17, (ATerm)ATmakeList1((ATerm)ATmakeInt(91)))), (ATerm)ATmakeAppl1(PERR_afun9, (ATerm)ATmakeAppl0(PERR_afun47)), (ATerm)ATmakeAppl0(PERR_afun5)), (ATerm)ATmakeList1((ATerm)ATmakeInt(91)))), (ATerm) wsAfterComma1), (ATerm)ATmakeAppl2(PERR_afun0, (ATerm)ATmakeAppl3(PERR_afun1, (ATerm)ATmakeList1((ATerm)ATmakeAppl1(PERR_afun17, (ATerm)ATmakeList1((ATerm)ATmakeInt(44)))), (ATerm)ATmakeAppl1(PERR_afun9, (ATerm)ATmakeAppl0(PERR_afun43)), (ATerm)ATmakeAppl0(PERR_afun5)), (ATerm)ATmakeList1((ATerm)ATmakeInt(44)))), (ATerm) wsAfterId), (ATerm) id), (ATerm) wsAfterComma), (ATerm)ATmakeAppl2(PERR_afun0, (ATerm)ATmakeAppl3(PERR_afun1, (ATerm)ATmakeList1((ATerm)ATmakeAppl1(PERR_afun17, (ATerm)ATmakeList1((ATerm)ATmakeInt(44)))), (ATerm)ATmakeAppl1(PERR_afun9, (ATerm)ATmakeAppl0(PERR_afun43)), (ATerm)ATmakeAppl0(PERR_afun5)), (ATerm)ATmakeList1((ATerm)ATmakeInt(44)))), (ATerm) wsAfterProducer), (ATerm) producer), (ATerm) wsAfterParenOpen), (ATerm)ATmakeAppl2(PERR_afun0, (ATerm)ATmakeAppl3(PERR_afun1, (ATerm)ATmakeList1((ATerm)ATmakeAppl1(PERR_afun17, (ATerm)ATmakeList1((ATerm)ATmakeInt(40)))), (ATerm)ATmakeAppl1(PERR_afun9, (ATerm)ATmakeAppl0(PERR_afun39)), (ATerm)ATmakeAppl0(PERR_afun5)), (ATerm)ATmakeList1((ATerm)ATmakeInt(40)))), (ATerm) wsAfterSummary), (ATerm)ATmakeAppl2(PERR_afun0, (ATerm)ATmakeAppl3(PERR_afun1, (ATerm)ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATmakeList1((ATerm)ATmakeAppl1(PERR_afun17, (ATerm)ATmakeList1((ATerm)ATmakeInt(121)))), (ATerm)ATmakeAppl1(PERR_afun17, (ATerm)ATmakeList1((ATerm)ATmakeInt(114)))), (ATerm)ATmakeAppl1(PERR_afun17, (ATerm)ATmakeList1((ATerm)ATmakeInt(97)))), (ATerm)ATmakeAppl1(PERR_afun17, (ATerm)ATmakeList1((ATerm)ATmakeInt(109)))), (ATerm)ATmakeAppl1(PERR_afun17, (ATerm)ATmakeList1((ATerm)ATmakeInt(109)))), (ATerm)ATmakeAppl1(PERR_afun17, (ATerm)ATmakeList1((ATerm)ATmakeInt(117)))), (ATerm)ATmakeAppl1(PERR_afun17, (ATerm)ATmakeList1((ATerm)ATmakeInt(115)))), (ATerm)ATmakeAppl1(PERR_afun9, (ATerm)ATmakeAppl0(PERR_afun52)), (ATerm)ATmakeAppl0(PERR_afun5)), (ATerm)ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATmakeList1((ATerm)ATmakeInt(121)), (ATerm)ATmakeInt(114)), (ATerm)ATmakeInt(97)), (ATerm)ATmakeInt(109)), (ATerm)ATmakeInt(109)), (ATerm)ATmakeInt(117)), (ATerm)ATmakeInt(115)))));
}
/**
 * Constructs a empty of type PERR_SubjectList. Like all ATerm types, PERR_SubjectLists are maximally shared.
 * \return A pointer to a empty, either newly constructed or shared
 */
PERR_SubjectList PERR_makeSubjectListEmpty(void) {
  return (PERR_SubjectList)(ATerm)ATempty;
}
/**
 * Constructs a single of type PERR_SubjectList. Like all ATerm types, PERR_SubjectLists are maximally shared.
 * \param[in] head a child of the new single
 * \return A pointer to a single, either newly constructed or shared
 */
PERR_SubjectList PERR_makeSubjectListSingle(PERR_Subject head) {
  return (PERR_SubjectList)(ATerm)ATmakeList1((ATerm) head);
}
/**
 * Constructs a many of type PERR_SubjectList. Like all ATerm types, PERR_SubjectLists are maximally shared.
 * \param[in] head a child of the new many
 * \param[in] wsAfterHead a child of the new many
 * \param[in] wsAfterSep a child of the new many
 * \param[in] tail a child of the new many
 * \return A pointer to a many, either newly constructed or shared
 */
PERR_SubjectList PERR_makeSubjectListMany(PERR_Subject head, PERR_OptLayout wsAfterHead, PERR_OptLayout wsAfterSep, PERR_SubjectList tail) {
  if (PERR_isSubjectListEmpty(tail)) {
    return PERR_makeSubjectListSingle(head);
  }
  return (PERR_SubjectList)(ATerm)ATinsert(ATinsert(ATinsert(ATinsert((ATermList)tail, (ATerm) wsAfterSep), (ATerm)ATmakeAppl2(PERR_afun0, (ATerm)ATmakeAppl3(PERR_afun1, (ATerm)ATmakeList1((ATerm)ATmakeAppl1(PERR_afun17, (ATerm)ATmakeList1((ATerm)ATmakeInt(44)))), (ATerm)ATmakeAppl1(PERR_afun9, (ATerm)ATmakeAppl0(PERR_afun43)), (ATerm)ATmakeAppl0(PERR_afun5)), (ATerm)ATmakeList1((ATerm)ATmakeInt(44)))), (ATerm) wsAfterHead), (ATerm) head);
}
/**
 * Constructs a empty of type PERR_ErrorList. Like all ATerm types, PERR_ErrorLists are maximally shared.
 * \return A pointer to a empty, either newly constructed or shared
 */
PERR_ErrorList PERR_makeErrorListEmpty(void) {
  return (PERR_ErrorList)(ATerm)ATempty;
}
/**
 * Constructs a single of type PERR_ErrorList. Like all ATerm types, PERR_ErrorLists are maximally shared.
 * \param[in] head a child of the new single
 * \return A pointer to a single, either newly constructed or shared
 */
PERR_ErrorList PERR_makeErrorListSingle(PERR_Error head) {
  return (PERR_ErrorList)(ATerm)ATmakeList1((ATerm) head);
}
/**
 * Constructs a many of type PERR_ErrorList. Like all ATerm types, PERR_ErrorLists are maximally shared.
 * \param[in] head a child of the new many
 * \param[in] wsAfterHead a child of the new many
 * \param[in] wsAfterSep a child of the new many
 * \param[in] tail a child of the new many
 * \return A pointer to a many, either newly constructed or shared
 */
PERR_ErrorList PERR_makeErrorListMany(PERR_Error head, PERR_OptLayout wsAfterHead, PERR_OptLayout wsAfterSep, PERR_ErrorList tail) {
  if (PERR_isErrorListEmpty(tail)) {
    return PERR_makeErrorListSingle(head);
  }
  return (PERR_ErrorList)(ATerm)ATinsert(ATinsert(ATinsert(ATinsert((ATermList)tail, (ATerm) wsAfterSep), (ATerm)ATmakeAppl2(PERR_afun0, (ATerm)ATmakeAppl3(PERR_afun1, (ATerm)ATmakeList1((ATerm)ATmakeAppl1(PERR_afun17, (ATerm)ATmakeList1((ATerm)ATmakeInt(44)))), (ATerm)ATmakeAppl1(PERR_afun9, (ATerm)ATmakeAppl0(PERR_afun43)), (ATerm)ATmakeAppl0(PERR_afun5)), (ATerm)ATmakeList1((ATerm)ATmakeInt(44)))), (ATerm) wsAfterHead), (ATerm) head);
}
/**
 * Constructs a file of type PERR_Location. Like all ATerm types, PERR_Locations are maximally shared.
 * \param[in] wsAfterFile a child of the new file
 * \param[in] wsAfterParenOpen a child of the new file
 * \param[in] filename a child of the new file
 * \param[in] wsAfterFilename a child of the new file
 * \return A pointer to a file, either newly constructed or shared
 */
PERR_Location PERR_makeLocationFile(PERR_OptLayout wsAfterFile, PERR_OptLayout wsAfterParenOpen, PERR_StrCon filename, PERR_OptLayout wsAfterFilename) {
  return (PERR_Location)(ATerm)ATmakeAppl2(PERR_afun0, (ATerm)ATmakeAppl3(PERR_afun1, (ATerm)ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATmakeList1((ATerm)ATmakeAppl1(PERR_afun9, (ATerm)ATmakeAppl0(PERR_afun38))), (ATerm)ATmakeAppl1(PERR_afun2, (ATerm)ATmakeAppl1(PERR_afun3, (ATerm)ATmakeAppl0(PERR_afun4)))), (ATerm)ATmakeAppl1(PERR_afun2, (ATerm)ATmakeAppl1(PERR_afun11, (ATerm)ATmakeAppl0(PERR_afun29)))), (ATerm)ATmakeAppl1(PERR_afun2, (ATerm)ATmakeAppl1(PERR_afun3, (ATerm)ATmakeAppl0(PERR_afun4)))), (ATerm)ATmakeAppl1(PERR_afun9, (ATerm)ATmakeAppl0(PERR_afun39))), (ATerm)ATmakeAppl1(PERR_afun2, (ATerm)ATmakeAppl1(PERR_afun3, (ATerm)ATmakeAppl0(PERR_afun4)))), (ATerm)ATmakeAppl1(PERR_afun9, (ATerm)ATmakeAppl0(PERR_afun53))), (ATerm)ATmakeAppl1(PERR_afun2, (ATerm)ATmakeAppl1(PERR_afun11, (ATerm)ATmakeAppl0(PERR_afun42))), (ATerm)ATmakeAppl1(PERR_afun13, (ATerm)ATmakeList1((ATerm)ATmakeAppl1(PERR_afun14, (ATerm)ATmakeAppl1(PERR_afun15, (ATerm)ATmakeAppl0(PERR_afun53)))))), (ATerm)ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATmakeList1((ATerm)ATmakeAppl2(PERR_afun0, (ATerm)ATmakeAppl3(PERR_afun1, (ATerm)ATmakeList1((ATerm)ATmakeAppl1(PERR_afun17, (ATerm)ATmakeList1((ATerm)ATmakeInt(41)))), (ATerm)ATmakeAppl1(PERR_afun9, (ATerm)ATmakeAppl0(PERR_afun38)), (ATerm)ATmakeAppl0(PERR_afun5)), (ATerm)ATmakeList1((ATerm)ATmakeInt(41)))), (ATerm) wsAfterFilename), (ATerm) filename), (ATerm) wsAfterParenOpen), (ATerm)ATmakeAppl2(PERR_afun0, (ATerm)ATmakeAppl3(PERR_afun1, (ATerm)ATmakeList1((ATerm)ATmakeAppl1(PERR_afun17, (ATerm)ATmakeList1((ATerm)ATmakeInt(40)))), (ATerm)ATmakeAppl1(PERR_afun9, (ATerm)ATmakeAppl0(PERR_afun39)), (ATerm)ATmakeAppl0(PERR_afun5)), (ATerm)ATmakeList1((ATerm)ATmakeInt(40)))), (ATerm) wsAfterFile), (ATerm)ATmakeAppl2(PERR_afun0, (ATerm)ATmakeAppl3(PERR_afun1, (ATerm)ATinsert(ATinsert(ATinsert(ATmakeList1((ATerm)ATmakeAppl1(PERR_afun17, (ATerm)ATmakeList1((ATerm)ATmakeInt(101)))), (ATerm)ATmakeAppl1(PERR_afun17, (ATerm)ATmakeList1((ATerm)ATmakeInt(108)))), (ATerm)ATmakeAppl1(PERR_afun17, (ATerm)ATmakeList1((ATerm)ATmakeInt(105)))), (ATerm)ATmakeAppl1(PERR_afun17, (ATerm)ATmakeList1((ATerm)ATmakeInt(102)))), (ATerm)ATmakeAppl1(PERR_afun9, (ATerm)ATmakeAppl0(PERR_afun53)), (ATerm)ATmakeAppl0(PERR_afun5)), (ATerm)ATinsert(ATinsert(ATinsert(ATmakeList1((ATerm)ATmakeInt(101)), (ATerm)ATmakeInt(108)), (ATerm)ATmakeInt(105)), (ATerm)ATmakeInt(102)))));
}
/**
 * Constructs a area of type PERR_Location. Like all ATerm types, PERR_Locations are maximally shared.
 * \param[in] wsAfterA a child of the new area
 * \param[in] wsAfterParenOpen a child of the new area
 * \param[in] Area a child of the new area
 * \param[in] wsAfterArea a child of the new area
 * \return A pointer to a area, either newly constructed or shared
 */
PERR_Location PERR_makeLocationArea(PERR_OptLayout wsAfterA, PERR_OptLayout wsAfterParenOpen, PERR_Area Area, PERR_OptLayout wsAfterArea) {
  return (PERR_Location)(ATerm)ATmakeAppl2(PERR_afun0, (ATerm)ATmakeAppl3(PERR_afun1, (ATerm)ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATmakeList1((ATerm)ATmakeAppl1(PERR_afun9, (ATerm)ATmakeAppl0(PERR_afun38))), (ATerm)ATmakeAppl1(PERR_afun2, (ATerm)ATmakeAppl1(PERR_afun3, (ATerm)ATmakeAppl0(PERR_afun4)))), (ATerm)ATmakeAppl1(PERR_afun2, (ATerm)ATmakeAppl1(PERR_afun11, (ATerm)ATmakeAppl0(PERR_afun54)))), (ATerm)ATmakeAppl1(PERR_afun2, (ATerm)ATmakeAppl1(PERR_afun3, (ATerm)ATmakeAppl0(PERR_afun4)))), (ATerm)ATmakeAppl1(PERR_afun9, (ATerm)ATmakeAppl0(PERR_afun39))), (ATerm)ATmakeAppl1(PERR_afun2, (ATerm)ATmakeAppl1(PERR_afun3, (ATerm)ATmakeAppl0(PERR_afun4)))), (ATerm)ATmakeAppl1(PERR_afun9, (ATerm)ATmakeAppl0(PERR_afun55))), (ATerm)ATmakeAppl1(PERR_afun2, (ATerm)ATmakeAppl1(PERR_afun11, (ATerm)ATmakeAppl0(PERR_afun42))), (ATerm)ATmakeAppl1(PERR_afun13, (ATerm)ATmakeList1((ATerm)ATmakeAppl1(PERR_afun14, (ATerm)ATmakeAppl1(PERR_afun15, (ATerm)ATmakeAppl0(PERR_afun55)))))), (ATerm)ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATmakeList1((ATerm)ATmakeAppl2(PERR_afun0, (ATerm)ATmakeAppl3(PERR_afun1, (ATerm)ATmakeList1((ATerm)ATmakeAppl1(PERR_afun17, (ATerm)ATmakeList1((ATerm)ATmakeInt(41)))), (ATerm)ATmakeAppl1(PERR_afun9, (ATerm)ATmakeAppl0(PERR_afun38)), (ATerm)ATmakeAppl0(PERR_afun5)), (ATerm)ATmakeList1((ATerm)ATmakeInt(41)))), (ATerm) wsAfterArea), (ATerm) Area), (ATerm) wsAfterParenOpen), (ATerm)ATmakeAppl2(PERR_afun0, (ATerm)ATmakeAppl3(PERR_afun1, (ATerm)ATmakeList1((ATerm)ATmakeAppl1(PERR_afun17, (ATerm)ATmakeList1((ATerm)ATmakeInt(40)))), (ATerm)ATmakeAppl1(PERR_afun9, (ATerm)ATmakeAppl0(PERR_afun39)), (ATerm)ATmakeAppl0(PERR_afun5)), (ATerm)ATmakeList1((ATerm)ATmakeInt(40)))), (ATerm) wsAfterA), (ATerm)ATmakeAppl2(PERR_afun0, (ATerm)ATmakeAppl3(PERR_afun1, (ATerm)ATinsert(ATinsert(ATinsert(ATmakeList1((ATerm)ATmakeAppl1(PERR_afun17, (ATerm)ATmakeList1((ATerm)ATmakeInt(97)))), (ATerm)ATmakeAppl1(PERR_afun17, (ATerm)ATmakeList1((ATerm)ATmakeInt(101)))), (ATerm)ATmakeAppl1(PERR_afun17, (ATerm)ATmakeList1((ATerm)ATmakeInt(114)))), (ATerm)ATmakeAppl1(PERR_afun17, (ATerm)ATmakeList1((ATerm)ATmakeInt(97)))), (ATerm)ATmakeAppl1(PERR_afun9, (ATerm)ATmakeAppl0(PERR_afun55)), (ATerm)ATmakeAppl0(PERR_afun5)), (ATerm)ATinsert(ATinsert(ATinsert(ATmakeList1((ATerm)ATmakeInt(97)), (ATerm)ATmakeInt(101)), (ATerm)ATmakeInt(114)), (ATerm)ATmakeInt(97)))));
}
/**
 * Constructs a area-in-file of type PERR_Location. Like all ATerm types, PERR_Locations are maximally shared.
 * \param[in] wsAfterAreaInFile a child of the new area-in-file
 * \param[in] wsAfterParenOpen a child of the new area-in-file
 * \param[in] filename a child of the new area-in-file
 * \param[in] wsAfterFilename a child of the new area-in-file
 * \param[in] wsAfterComma a child of the new area-in-file
 * \param[in] Area a child of the new area-in-file
 * \param[in] wsAfterArea a child of the new area-in-file
 * \return A pointer to a area-in-file, either newly constructed or shared
 */
PERR_Location PERR_makeLocationAreaInFile(PERR_OptLayout wsAfterAreaInFile, PERR_OptLayout wsAfterParenOpen, PERR_StrCon filename, PERR_OptLayout wsAfterFilename, PERR_OptLayout wsAfterComma, PERR_Area Area, PERR_OptLayout wsAfterArea) {
  return (PERR_Location)(ATerm)ATmakeAppl2(PERR_afun0, (ATerm)ATmakeAppl3(PERR_afun1, (ATerm)ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATmakeList1((ATerm)ATmakeAppl1(PERR_afun9, (ATerm)ATmakeAppl0(PERR_afun38))), (ATerm)ATmakeAppl1(PERR_afun2, (ATerm)ATmakeAppl1(PERR_afun3, (ATerm)ATmakeAppl0(PERR_afun4)))), (ATerm)ATmakeAppl1(PERR_afun2, (ATerm)ATmakeAppl1(PERR_afun11, (ATerm)ATmakeAppl0(PERR_afun54)))), (ATerm)ATmakeAppl1(PERR_afun2, (ATerm)ATmakeAppl1(PERR_afun3, (ATerm)ATmakeAppl0(PERR_afun4)))), (ATerm)ATmakeAppl1(PERR_afun9, (ATerm)ATmakeAppl0(PERR_afun43))), (ATerm)ATmakeAppl1(PERR_afun2, (ATerm)ATmakeAppl1(PERR_afun3, (ATerm)ATmakeAppl0(PERR_afun4)))), (ATerm)ATmakeAppl1(PERR_afun2, (ATerm)ATmakeAppl1(PERR_afun11, (ATerm)ATmakeAppl0(PERR_afun29)))), (ATerm)ATmakeAppl1(PERR_afun2, (ATerm)ATmakeAppl1(PERR_afun3, (ATerm)ATmakeAppl0(PERR_afun4)))), (ATerm)ATmakeAppl1(PERR_afun9, (ATerm)ATmakeAppl0(PERR_afun39))), (ATerm)ATmakeAppl1(PERR_afun2, (ATerm)ATmakeAppl1(PERR_afun3, (ATerm)ATmakeAppl0(PERR_afun4)))), (ATerm)ATmakeAppl1(PERR_afun9, (ATerm)ATmakeAppl0(PERR_afun56))), (ATerm)ATmakeAppl1(PERR_afun2, (ATerm)ATmakeAppl1(PERR_afun11, (ATerm)ATmakeAppl0(PERR_afun42))), (ATerm)ATmakeAppl1(PERR_afun13, (ATerm)ATmakeList1((ATerm)ATmakeAppl1(PERR_afun14, (ATerm)ATmakeAppl1(PERR_afun15, (ATerm)ATmakeAppl0(PERR_afun56)))))), (ATerm)ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATmakeList1((ATerm)ATmakeAppl2(PERR_afun0, (ATerm)ATmakeAppl3(PERR_afun1, (ATerm)ATmakeList1((ATerm)ATmakeAppl1(PERR_afun17, (ATerm)ATmakeList1((ATerm)ATmakeInt(41)))), (ATerm)ATmakeAppl1(PERR_afun9, (ATerm)ATmakeAppl0(PERR_afun38)), (ATerm)ATmakeAppl0(PERR_afun5)), (ATerm)ATmakeList1((ATerm)ATmakeInt(41)))), (ATerm) wsAfterArea), (ATerm) Area), (ATerm) wsAfterComma), (ATerm)ATmakeAppl2(PERR_afun0, (ATerm)ATmakeAppl3(PERR_afun1, (ATerm)ATmakeList1((ATerm)ATmakeAppl1(PERR_afun17, (ATerm)ATmakeList1((ATerm)ATmakeInt(44)))), (ATerm)ATmakeAppl1(PERR_afun9, (ATerm)ATmakeAppl0(PERR_afun43)), (ATerm)ATmakeAppl0(PERR_afun5)), (ATerm)ATmakeList1((ATerm)ATmakeInt(44)))), (ATerm) wsAfterFilename), (ATerm) filename), (ATerm) wsAfterParenOpen), (ATerm)ATmakeAppl2(PERR_afun0, (ATerm)ATmakeAppl3(PERR_afun1, (ATerm)ATmakeList1((ATerm)ATmakeAppl1(PERR_afun17, (ATerm)ATmakeList1((ATerm)ATmakeInt(40)))), (ATerm)ATmakeAppl1(PERR_afun9, (ATerm)ATmakeAppl0(PERR_afun39)), (ATerm)ATmakeAppl0(PERR_afun5)), (ATerm)ATmakeList1((ATerm)ATmakeInt(40)))), (ATerm) wsAfterAreaInFile), (ATerm)ATmakeAppl2(PERR_afun0, (ATerm)ATmakeAppl3(PERR_afun1, (ATerm)ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATmakeList1((ATerm)ATmakeAppl1(PERR_afun17, (ATerm)ATmakeList1((ATerm)ATmakeInt(101)))), (ATerm)ATmakeAppl1(PERR_afun17, (ATerm)ATmakeList1((ATerm)ATmakeInt(108)))), (ATerm)ATmakeAppl1(PERR_afun17, (ATerm)ATmakeList1((ATerm)ATmakeInt(105)))), (ATerm)ATmakeAppl1(PERR_afun17, (ATerm)ATmakeList1((ATerm)ATmakeInt(102)))), (ATerm)ATmakeAppl1(PERR_afun17, (ATerm)ATmakeList1((ATerm)ATmakeInt(45)))), (ATerm)ATmakeAppl1(PERR_afun17, (ATerm)ATmakeList1((ATerm)ATmakeInt(110)))), (ATerm)ATmakeAppl1(PERR_afun17, (ATerm)ATmakeList1((ATerm)ATmakeInt(105)))), (ATerm)ATmakeAppl1(PERR_afun17, (ATerm)ATmakeList1((ATerm)ATmakeInt(45)))), (ATerm)ATmakeAppl1(PERR_afun17, (ATerm)ATmakeList1((ATerm)ATmakeInt(97)))), (ATerm)ATmakeAppl1(PERR_afun17, (ATerm)ATmakeList1((ATerm)ATmakeInt(101)))), (ATerm)ATmakeAppl1(PERR_afun17, (ATerm)ATmakeList1((ATerm)ATmakeInt(114)))), (ATerm)ATmakeAppl1(PERR_afun17, (ATerm)ATmakeList1((ATerm)ATmakeInt(97)))), (ATerm)ATmakeAppl1(PERR_afun9, (ATerm)ATmakeAppl0(PERR_afun56)), (ATerm)ATmakeAppl0(PERR_afun5)), (ATerm)ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATmakeList1((ATerm)ATmakeInt(101)), (ATerm)ATmakeInt(108)), (ATerm)ATmakeInt(105)), (ATerm)ATmakeInt(102)), (ATerm)ATmakeInt(45)), (ATerm)ATmakeInt(110)), (ATerm)ATmakeInt(105)), (ATerm)ATmakeInt(45)), (ATerm)ATmakeInt(97)), (ATerm)ATmakeInt(101)), (ATerm)ATmakeInt(114)), (ATerm)ATmakeInt(97)))));
}
/**
 * Constructs a area of type PERR_Area. Like all ATerm types, PERR_Areas are maximally shared.
 * \param[in] wsAfterArea a child of the new area
 * \param[in] wsAfterParenOpen a child of the new area
 * \param[in] beginLine a child of the new area
 * \param[in] wsAfterBeginLine a child of the new area
 * \param[in] wsAfterComma a child of the new area
 * \param[in] beginColumn a child of the new area
 * \param[in] wsAfterBeginColumn a child of the new area
 * \param[in] wsAfterComma1 a child of the new area
 * \param[in] endLine a child of the new area
 * \param[in] wsAfterEndLine a child of the new area
 * \param[in] wsAfterComma2 a child of the new area
 * \param[in] endColumn a child of the new area
 * \param[in] wsAfterEndColumn a child of the new area
 * \param[in] wsAfterComma3 a child of the new area
 * \param[in] offset a child of the new area
 * \param[in] wsAfterOffset a child of the new area
 * \param[in] wsAfterComma4 a child of the new area
 * \param[in] length a child of the new area
 * \param[in] wsAfterLength a child of the new area
 * \return A pointer to a area, either newly constructed or shared
 */
PERR_Area PERR_makeAreaArea(PERR_OptLayout wsAfterArea, PERR_OptLayout wsAfterParenOpen, PERR_NatCon beginLine, PERR_OptLayout wsAfterBeginLine, PERR_OptLayout wsAfterComma, PERR_NatCon beginColumn, PERR_OptLayout wsAfterBeginColumn, PERR_OptLayout wsAfterComma1, PERR_NatCon endLine, PERR_OptLayout wsAfterEndLine, PERR_OptLayout wsAfterComma2, PERR_NatCon endColumn, PERR_OptLayout wsAfterEndColumn, PERR_OptLayout wsAfterComma3, PERR_NatCon offset, PERR_OptLayout wsAfterOffset, PERR_OptLayout wsAfterComma4, PERR_NatCon length, PERR_OptLayout wsAfterLength) {
  return (PERR_Area)(ATerm)ATmakeAppl2(PERR_afun0, (ATerm)ATmakeAppl3(PERR_afun1, (ATerm)ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATmakeList1((ATerm)ATmakeAppl1(PERR_afun9, (ATerm)ATmakeAppl0(PERR_afun38))), (ATerm)ATmakeAppl1(PERR_afun2, (ATerm)ATmakeAppl1(PERR_afun3, (ATerm)ATmakeAppl0(PERR_afun4)))), (ATerm)ATmakeAppl1(PERR_afun2, (ATerm)ATmakeAppl1(PERR_afun11, (ATerm)ATmakeAppl0(PERR_afun31)))), (ATerm)ATmakeAppl1(PERR_afun2, (ATerm)ATmakeAppl1(PERR_afun3, (ATerm)ATmakeAppl0(PERR_afun4)))), (ATerm)ATmakeAppl1(PERR_afun9, (ATerm)ATmakeAppl0(PERR_afun43))), (ATerm)ATmakeAppl1(PERR_afun2, (ATerm)ATmakeAppl1(PERR_afun3, (ATerm)ATmakeAppl0(PERR_afun4)))), (ATerm)ATmakeAppl1(PERR_afun2, (ATerm)ATmakeAppl1(PERR_afun11, (ATerm)ATmakeAppl0(PERR_afun31)))), (ATerm)ATmakeAppl1(PERR_afun2, (ATerm)ATmakeAppl1(PERR_afun3, (ATerm)ATmakeAppl0(PERR_afun4)))), (ATerm)ATmakeAppl1(PERR_afun9, (ATerm)ATmakeAppl0(PERR_afun43))), (ATerm)ATmakeAppl1(PERR_afun2, (ATerm)ATmakeAppl1(PERR_afun3, (ATerm)ATmakeAppl0(PERR_afun4)))), (ATerm)ATmakeAppl1(PERR_afun2, (ATerm)ATmakeAppl1(PERR_afun11, (ATerm)ATmakeAppl0(PERR_afun31)))), (ATerm)ATmakeAppl1(PERR_afun2, (ATerm)ATmakeAppl1(PERR_afun3, (ATerm)ATmakeAppl0(PERR_afun4)))), (ATerm)ATmakeAppl1(PERR_afun9, (ATerm)ATmakeAppl0(PERR_afun43))), (ATerm)ATmakeAppl1(PERR_afun2, (ATerm)ATmakeAppl1(PERR_afun3, (ATerm)ATmakeAppl0(PERR_afun4)))), (ATerm)ATmakeAppl1(PERR_afun2, (ATerm)ATmakeAppl1(PERR_afun11, (ATerm)ATmakeAppl0(PERR_afun31)))), (ATerm)ATmakeAppl1(PERR_afun2, (ATerm)ATmakeAppl1(PERR_afun3, (ATerm)ATmakeAppl0(PERR_afun4)))), (ATerm)ATmakeAppl1(PERR_afun9, (ATerm)ATmakeAppl0(PERR_afun43))), (ATerm)ATmakeAppl1(PERR_afun2, (ATerm)ATmakeAppl1(PERR_afun3, (ATerm)ATmakeAppl0(PERR_afun4)))), (ATerm)ATmakeAppl1(PERR_afun2, (ATerm)ATmakeAppl1(PERR_afun11, (ATerm)ATmakeAppl0(PERR_afun31)))), (ATerm)ATmakeAppl1(PERR_afun2, (ATerm)ATmakeAppl1(PERR_afun3, (ATerm)ATmakeAppl0(PERR_afun4)))), (ATerm)ATmakeAppl1(PERR_afun9, (ATerm)ATmakeAppl0(PERR_afun43))), (ATerm)ATmakeAppl1(PERR_afun2, (ATerm)ATmakeAppl1(PERR_afun3, (ATerm)ATmakeAppl0(PERR_afun4)))), (ATerm)ATmakeAppl1(PERR_afun2, (ATerm)ATmakeAppl1(PERR_afun11, (ATerm)ATmakeAppl0(PERR_afun31)))), (ATerm)ATmakeAppl1(PERR_afun2, (ATerm)ATmakeAppl1(PERR_afun3, (ATerm)ATmakeAppl0(PERR_afun4)))), (ATerm)ATmakeAppl1(PERR_afun9, (ATerm)ATmakeAppl0(PERR_afun39))), (ATerm)ATmakeAppl1(PERR_afun2, (ATerm)ATmakeAppl1(PERR_afun3, (ATerm)ATmakeAppl0(PERR_afun4)))), (ATerm)ATmakeAppl1(PERR_afun9, (ATerm)ATmakeAppl0(PERR_afun55))), (ATerm)ATmakeAppl1(PERR_afun2, (ATerm)ATmakeAppl1(PERR_afun11, (ATerm)ATmakeAppl0(PERR_afun54))), (ATerm)ATmakeAppl1(PERR_afun13, (ATerm)ATmakeList1((ATerm)ATmakeAppl1(PERR_afun14, (ATerm)ATmakeAppl1(PERR_afun15, (ATerm)ATmakeAppl0(PERR_afun55)))))), (ATerm)ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATmakeList1((ATerm)ATmakeAppl2(PERR_afun0, (ATerm)ATmakeAppl3(PERR_afun1, (ATerm)ATmakeList1((ATerm)ATmakeAppl1(PERR_afun17, (ATerm)ATmakeList1((ATerm)ATmakeInt(41)))), (ATerm)ATmakeAppl1(PERR_afun9, (ATerm)ATmakeAppl0(PERR_afun38)), (ATerm)ATmakeAppl0(PERR_afun5)), (ATerm)ATmakeList1((ATerm)ATmakeInt(41)))), (ATerm) wsAfterLength), (ATerm) length), (ATerm) wsAfterComma4), (ATerm)ATmakeAppl2(PERR_afun0, (ATerm)ATmakeAppl3(PERR_afun1, (ATerm)ATmakeList1((ATerm)ATmakeAppl1(PERR_afun17, (ATerm)ATmakeList1((ATerm)ATmakeInt(44)))), (ATerm)ATmakeAppl1(PERR_afun9, (ATerm)ATmakeAppl0(PERR_afun43)), (ATerm)ATmakeAppl0(PERR_afun5)), (ATerm)ATmakeList1((ATerm)ATmakeInt(44)))), (ATerm) wsAfterOffset), (ATerm) offset), (ATerm) wsAfterComma3), (ATerm)ATmakeAppl2(PERR_afun0, (ATerm)ATmakeAppl3(PERR_afun1, (ATerm)ATmakeList1((ATerm)ATmakeAppl1(PERR_afun17, (ATerm)ATmakeList1((ATerm)ATmakeInt(44)))), (ATerm)ATmakeAppl1(PERR_afun9, (ATerm)ATmakeAppl0(PERR_afun43)), (ATerm)ATmakeAppl0(PERR_afun5)), (ATerm)ATmakeList1((ATerm)ATmakeInt(44)))), (ATerm) wsAfterEndColumn), (ATerm) endColumn), (ATerm) wsAfterComma2), (ATerm)ATmakeAppl2(PERR_afun0, (ATerm)ATmakeAppl3(PERR_afun1, (ATerm)ATmakeList1((ATerm)ATmakeAppl1(PERR_afun17, (ATerm)ATmakeList1((ATerm)ATmakeInt(44)))), (ATerm)ATmakeAppl1(PERR_afun9, (ATerm)ATmakeAppl0(PERR_afun43)), (ATerm)ATmakeAppl0(PERR_afun5)), (ATerm)ATmakeList1((ATerm)ATmakeInt(44)))), (ATerm) wsAfterEndLine), (ATerm) endLine), (ATerm) wsAfterComma1), (ATerm)ATmakeAppl2(PERR_afun0, (ATerm)ATmakeAppl3(PERR_afun1, (ATerm)ATmakeList1((ATerm)ATmakeAppl1(PERR_afun17, (ATerm)ATmakeList1((ATerm)ATmakeInt(44)))), (ATerm)ATmakeAppl1(PERR_afun9, (ATerm)ATmakeAppl0(PERR_afun43)), (ATerm)ATmakeAppl0(PERR_afun5)), (ATerm)ATmakeList1((ATerm)ATmakeInt(44)))), (ATerm) wsAfterBeginColumn), (ATerm) beginColumn), (ATerm) wsAfterComma), (ATerm)ATmakeAppl2(PERR_afun0, (ATerm)ATmakeAppl3(PERR_afun1, (ATerm)ATmakeList1((ATerm)ATmakeAppl1(PERR_afun17, (ATerm)ATmakeList1((ATerm)ATmakeInt(44)))), (ATerm)ATmakeAppl1(PERR_afun9, (ATerm)ATmakeAppl0(PERR_afun43)), (ATerm)ATmakeAppl0(PERR_afun5)), (ATerm)ATmakeList1((ATerm)ATmakeInt(44)))), (ATerm) wsAfterBeginLine), (ATerm) beginLine), (ATerm) wsAfterParenOpen), (ATerm)ATmakeAppl2(PERR_afun0, (ATerm)ATmakeAppl3(PERR_afun1, (ATerm)ATmakeList1((ATerm)ATmakeAppl1(PERR_afun17, (ATerm)ATmakeList1((ATerm)ATmakeInt(40)))), (ATerm)ATmakeAppl1(PERR_afun9, (ATerm)ATmakeAppl0(PERR_afun39)), (ATerm)ATmakeAppl0(PERR_afun5)), (ATerm)ATmakeList1((ATerm)ATmakeInt(40)))), (ATerm) wsAfterArea), (ATerm)ATmakeAppl2(PERR_afun0, (ATerm)ATmakeAppl3(PERR_afun1, (ATerm)ATinsert(ATinsert(ATinsert(ATmakeList1((ATerm)ATmakeAppl1(PERR_afun17, (ATerm)ATmakeList1((ATerm)ATmakeInt(97)))), (ATerm)ATmakeAppl1(PERR_afun17, (ATerm)ATmakeList1((ATerm)ATmakeInt(101)))), (ATerm)ATmakeAppl1(PERR_afun17, (ATerm)ATmakeList1((ATerm)ATmakeInt(114)))), (ATerm)ATmakeAppl1(PERR_afun17, (ATerm)ATmakeList1((ATerm)ATmakeInt(97)))), (ATerm)ATmakeAppl1(PERR_afun9, (ATerm)ATmakeAppl0(PERR_afun55)), (ATerm)ATmakeAppl0(PERR_afun5)), (ATerm)ATinsert(ATinsert(ATinsert(ATmakeList1((ATerm)ATmakeInt(97)), (ATerm)ATmakeInt(101)), (ATerm)ATmakeInt(114)), (ATerm)ATmakeInt(97)))));
}
/**
 * Constructs a slice of type PERR_Slice. Like all ATerm types, PERR_Slices are maximally shared.
 * \param[in] wsAfterSlice a child of the new slice
 * \param[in] wsAfterParenOpen a child of the new slice
 * \param[in] id a child of the new slice
 * \param[in] wsAfterId a child of the new slice
 * \param[in] wsAfterComma a child of the new slice
 * \param[in] wsAfterBracketOpen a child of the new slice
 * \param[in] areas a child of the new slice
 * \param[in] wsAfterAreas a child of the new slice
 * \param[in] wsAfterBracketClose a child of the new slice
 * \return A pointer to a slice, either newly constructed or shared
 */
PERR_Slice PERR_makeSliceSlice(PERR_OptLayout wsAfterSlice, PERR_OptLayout wsAfterParenOpen, PERR_StrCon id, PERR_OptLayout wsAfterId, PERR_OptLayout wsAfterComma, PERR_OptLayout wsAfterBracketOpen, PERR_AreaAreas areas, PERR_OptLayout wsAfterAreas, PERR_OptLayout wsAfterBracketClose) {
  return (PERR_Slice)(ATerm)ATmakeAppl2(PERR_afun0, (ATerm)ATmakeAppl3(PERR_afun1, (ATerm)ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATmakeList1((ATerm)ATmakeAppl1(PERR_afun9, (ATerm)ATmakeAppl0(PERR_afun38))), (ATerm)ATmakeAppl1(PERR_afun2, (ATerm)ATmakeAppl1(PERR_afun3, (ATerm)ATmakeAppl0(PERR_afun4)))), (ATerm)ATmakeAppl1(PERR_afun9, (ATerm)ATmakeAppl0(PERR_afun45))), (ATerm)ATmakeAppl1(PERR_afun2, (ATerm)ATmakeAppl1(PERR_afun3, (ATerm)ATmakeAppl0(PERR_afun4)))), (ATerm)ATmakeAppl1(PERR_afun2, (ATerm)ATmakeAppl2(PERR_afun46, (ATerm)ATmakeAppl1(PERR_afun11, (ATerm)ATmakeAppl0(PERR_afun54)), (ATerm)ATmakeAppl1(PERR_afun9, (ATerm)ATmakeAppl0(PERR_afun43))))), (ATerm)ATmakeAppl1(PERR_afun2, (ATerm)ATmakeAppl1(PERR_afun3, (ATerm)ATmakeAppl0(PERR_afun4)))), (ATerm)ATmakeAppl1(PERR_afun9, (ATerm)ATmakeAppl0(PERR_afun47))), (ATerm)ATmakeAppl1(PERR_afun2, (ATerm)ATmakeAppl1(PERR_afun3, (ATerm)ATmakeAppl0(PERR_afun4)))), (ATerm)ATmakeAppl1(PERR_afun9, (ATerm)ATmakeAppl0(PERR_afun43))), (ATerm)ATmakeAppl1(PERR_afun2, (ATerm)ATmakeAppl1(PERR_afun3, (ATerm)ATmakeAppl0(PERR_afun4)))), (ATerm)ATmakeAppl1(PERR_afun2, (ATerm)ATmakeAppl1(PERR_afun11, (ATerm)ATmakeAppl0(PERR_afun29)))), (ATerm)ATmakeAppl1(PERR_afun2, (ATerm)ATmakeAppl1(PERR_afun3, (ATerm)ATmakeAppl0(PERR_afun4)))), (ATerm)ATmakeAppl1(PERR_afun9, (ATerm)ATmakeAppl0(PERR_afun39))), (ATerm)ATmakeAppl1(PERR_afun2, (ATerm)ATmakeAppl1(PERR_afun3, (ATerm)ATmakeAppl0(PERR_afun4)))), (ATerm)ATmakeAppl1(PERR_afun9, (ATerm)ATmakeAppl0(PERR_afun57))), (ATerm)ATmakeAppl1(PERR_afun2, (ATerm)ATmakeAppl1(PERR_afun11, (ATerm)ATmakeAppl0(PERR_afun58))), (ATerm)ATmakeAppl1(PERR_afun13, (ATerm)ATmakeList1((ATerm)ATmakeAppl1(PERR_afun14, (ATerm)ATmakeAppl1(PERR_afun15, (ATerm)ATmakeAppl0(PERR_afun57)))))), (ATerm)ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATinsert(ATmakeList1((ATerm)ATmakeAppl2(PERR_afun0, (ATerm)ATmakeAppl3(PERR_afun1, (ATerm)ATmakeList1((ATerm)ATmakeAppl1(PERR_afun17, (ATerm)ATmakeList1((ATerm)ATmakeInt(41)))), (ATerm)ATmakeAppl1(PERR_afun9, (ATerm)ATmakeAppl0(PERR_afun38)), (ATerm)ATmakeAppl0(PERR_afun5)), (ATerm)ATmakeList1((ATerm)ATmakeInt(41)))), (ATerm) wsAfterBracketClose), (ATerm)ATmakeAppl2(PERR_afun0, (ATerm)ATmakeAppl3(PERR_afun1, (ATerm)ATmakeList1((ATerm)ATmakeAppl1(PERR_afun17, (ATerm)ATmakeList1((ATerm)ATmakeInt(93)))), (ATerm)ATmakeAppl1(PERR_afun9, (ATerm)ATmakeAppl0(PERR_afun45)), (ATerm)ATmakeAppl0(PERR_afun5)), (ATerm)ATmakeList1((ATerm)ATmakeInt(93)))), (ATerm) wsAfterAreas), (ATerm)ATmakeAppl2(PERR_afun0, (ATerm)ATmakeAppl1(PERR_afun8, (ATerm)ATmakeAppl1(PERR_afun2, (ATerm)ATmakeAppl2(PERR_afun46, (ATerm)ATmakeAppl1(PERR_afun11, (ATerm)ATmakeAppl0(PERR_afun54)), (ATerm)ATmakeAppl1(PERR_afun9, (ATerm)ATmakeAppl0(PERR_afun43))))), (ATerm) areas)), (ATerm) wsAfterBracketOpen), (ATerm)ATmakeAppl2(PERR_afun0, (ATerm)ATmakeAppl3(PERR_afun1, (ATerm)ATmakeList1((ATerm)ATmakeAppl1(PERR_afun17, (ATerm)ATmakeList1((ATerm)ATmakeInt(91)))), (ATerm)ATmakeAppl1(PERR_afun9, (ATerm)ATmakeAppl0(PERR_afun47)), (ATerm)ATmakeAppl0(PERR_afun5)), (ATerm)ATmakeList1((ATerm)ATmakeInt(91)))), (ATerm) wsAfterComma), (ATerm)ATmakeAppl2(PERR_afun0, (ATerm)ATmakeAppl3(PERR_afun1, (ATerm)ATmakeList1((ATerm)ATmakeAppl1(PERR_afun17, (ATerm)ATmakeList1((ATerm)ATmakeInt(44)))), (ATerm)ATmakeAppl1(PERR_afun9, (ATerm)ATmakeAppl0(PERR_afun43)), (ATerm)ATmakeAppl0(PERR_afun5)), (ATerm)ATmakeList1((ATerm)ATmakeInt(44)))), (ATerm) wsAfterId), (ATerm) id), (ATerm) wsAfterParenOpen), (ATerm)ATmakeAppl2(PERR_afun0, (ATerm)ATmakeAppl3(PERR_afun1, (ATerm)ATmakeList1((ATerm)ATmakeAppl1(PERR_afun17, (ATerm)ATmakeList1((ATerm)ATmakeInt(40)))), (ATerm)ATmakeAppl1(PERR_afun9, (ATerm)ATmakeAppl0(PERR_afun39)), (ATerm)ATmakeAppl0(PERR_afun5)), (ATerm)ATmakeList1((ATerm)ATmakeInt(40)))), (ATerm) wsAfterSlice), (ATerm)ATmakeAppl2(PERR_afun0, (ATerm)ATmakeAppl3(PERR_afun1, (ATerm)ATinsert(ATinsert(ATinsert(ATinsert(ATmakeList1((ATerm)ATmakeAppl1(PERR_afun17, (ATerm)ATmakeList1((ATerm)ATmakeInt(101)))), (ATerm)ATmakeAppl1(PERR_afun17, (ATerm)ATmakeList1((ATerm)ATmakeInt(99)))), (ATerm)ATmakeAppl1(PERR_afun17, (ATerm)ATmakeList1((ATerm)ATmakeInt(105)))), (ATerm)ATmakeAppl1(PERR_afun17, (ATerm)ATmakeList1((ATerm)ATmakeInt(108)))), (ATerm)ATmakeAppl1(PERR_afun17, (ATerm)ATmakeList1((ATerm)ATmakeInt(115)))), (ATerm)ATmakeAppl1(PERR_afun9, (ATerm)ATmakeAppl0(PERR_afun57)), (ATerm)ATmakeAppl0(PERR_afun5)), (ATerm)ATinsert(ATinsert(ATinsert(ATinsert(ATmakeList1((ATerm)ATmakeInt(101)), (ATerm)ATmakeInt(99)), (ATerm)ATmakeInt(105)), (ATerm)ATmakeInt(108)), (ATerm)ATmakeInt(115)))));
}
/**
 * Constructs a empty of type PERR_AreaAreas. Like all ATerm types, PERR_AreaAreass are maximally shared.
 * \return A pointer to a empty, either newly constructed or shared
 */
PERR_AreaAreas PERR_makeAreaAreasEmpty(void) {
  return (PERR_AreaAreas)(ATerm)ATempty;
}
/**
 * Constructs a single of type PERR_AreaAreas. Like all ATerm types, PERR_AreaAreass are maximally shared.
 * \param[in] head a child of the new single
 * \return A pointer to a single, either newly constructed or shared
 */
PERR_AreaAreas PERR_makeAreaAreasSingle(PERR_Area head) {
  return (PERR_AreaAreas)(ATerm)ATmakeList1((ATerm) head);
}
/**
 * Constructs a many of type PERR_AreaAreas. Like all ATerm types, PERR_AreaAreass are maximally shared.
 * \param[in] head a child of the new many
 * \param[in] wsAfterHead a child of the new many
 * \param[in] wsAfterSep a child of the new many
 * \param[in] tail a child of the new many
 * \return A pointer to a many, either newly constructed or shared
 */
PERR_AreaAreas PERR_makeAreaAreasMany(PERR_Area head, PERR_OptLayout wsAfterHead, PERR_OptLayout wsAfterSep, PERR_AreaAreas tail) {
  if (PERR_isAreaAreasEmpty(tail)) {
    return PERR_makeAreaAreasSingle(head);
  }
  return (PERR_AreaAreas)(ATerm)ATinsert(ATinsert(ATinsert(ATinsert((ATermList)tail, (ATerm) wsAfterSep), (ATerm)ATmakeAppl2(PERR_afun0, (ATerm)ATmakeAppl3(PERR_afun1, (ATerm)ATmakeList1((ATerm)ATmakeAppl1(PERR_afun17, (ATerm)ATmakeList1((ATerm)ATmakeInt(44)))), (ATerm)ATmakeAppl1(PERR_afun9, (ATerm)ATmakeAppl0(PERR_afun43)), (ATerm)ATmakeAppl0(PERR_afun5)), (ATerm)ATmakeList1((ATerm)ATmakeInt(44)))), (ATerm) wsAfterHead), (ATerm) head);
}

/**
 * Tests equality of two PERR_OptLayouts. A constant time operation.
 * \param[in] arg0 first PERR_OptLayout to be compared
 * \param[in] arg1 second PERR_OptLayout to be compared
 * \return ATtrue if #arg0 was equal to #arg1, ATfalse otherwise
 */
ATbool _PERR_isEqualOptLayout(PERR_OptLayout arg0, PERR_OptLayout arg1) {
  return ATisEqual((ATerm)arg0, (ATerm)arg1);
}

/**
 * Tests equality of two PERR_Layouts. A constant time operation.
 * \param[in] arg0 first PERR_Layout to be compared
 * \param[in] arg1 second PERR_Layout to be compared
 * \return ATtrue if #arg0 was equal to #arg1, ATfalse otherwise
 */
ATbool _PERR_isEqualLayout(PERR_Layout arg0, PERR_Layout arg1) {
  return ATisEqual((ATerm)arg0, (ATerm)arg1);
}

/**
 * Tests equality of two PERR_LexLayoutLists. A constant time operation.
 * \param[in] arg0 first PERR_LexLayoutList to be compared
 * \param[in] arg1 second PERR_LexLayoutList to be compared
 * \return ATtrue if #arg0 was equal to #arg1, ATfalse otherwise
 */
ATbool _PERR_isEqualLexLayoutList(PERR_LexLayoutList arg0, PERR_LexLayoutList arg1) {
  return ATisEqual((ATerm)arg0, (ATerm)arg1);
}

/**
 * Tests equality of two PERR_LexStrChars. A constant time operation.
 * \param[in] arg0 first PERR_LexStrChar to be compared
 * \param[in] arg1 second PERR_LexStrChar to be compared
 * \return ATtrue if #arg0 was equal to #arg1, ATfalse otherwise
 */
ATbool _PERR_isEqualLexStrChar(PERR_LexStrChar arg0, PERR_LexStrChar arg1) {
  return ATisEqual((ATerm)arg0, (ATerm)arg1);
}

/**
 * Tests equality of two PERR_StrChars. A constant time operation.
 * \param[in] arg0 first PERR_StrChar to be compared
 * \param[in] arg1 second PERR_StrChar to be compared
 * \return ATtrue if #arg0 was equal to #arg1, ATfalse otherwise
 */
ATbool _PERR_isEqualStrChar(PERR_StrChar arg0, PERR_StrChar arg1) {
  return ATisEqual((ATerm)arg0, (ATerm)arg1);
}

/**
 * Tests equality of two PERR_LexStrCons. A constant time operation.
 * \param[in] arg0 first PERR_LexStrCon to be compared
 * \param[in] arg1 second PERR_LexStrCon to be compared
 * \return ATtrue if #arg0 was equal to #arg1, ATfalse otherwise
 */
ATbool _PERR_isEqualLexStrCon(PERR_LexStrCon arg0, PERR_LexStrCon arg1) {
  return ATisEqual((ATerm)arg0, (ATerm)arg1);
}

/**
 * Tests equality of two PERR_StrCons. A constant time operation.
 * \param[in] arg0 first PERR_StrCon to be compared
 * \param[in] arg1 second PERR_StrCon to be compared
 * \return ATtrue if #arg0 was equal to #arg1, ATfalse otherwise
 */
ATbool _PERR_isEqualStrCon(PERR_StrCon arg0, PERR_StrCon arg1) {
  return ATisEqual((ATerm)arg0, (ATerm)arg1);
}

/**
 * Tests equality of two PERR_LexStrCharCharss. A constant time operation.
 * \param[in] arg0 first PERR_LexStrCharChars to be compared
 * \param[in] arg1 second PERR_LexStrCharChars to be compared
 * \return ATtrue if #arg0 was equal to #arg1, ATfalse otherwise
 */
ATbool _PERR_isEqualLexStrCharChars(PERR_LexStrCharChars arg0, PERR_LexStrCharChars arg1) {
  return ATisEqual((ATerm)arg0, (ATerm)arg1);
}

/**
 * Tests equality of two PERR_LexNatCons. A constant time operation.
 * \param[in] arg0 first PERR_LexNatCon to be compared
 * \param[in] arg1 second PERR_LexNatCon to be compared
 * \return ATtrue if #arg0 was equal to #arg1, ATfalse otherwise
 */
ATbool _PERR_isEqualLexNatCon(PERR_LexNatCon arg0, PERR_LexNatCon arg1) {
  return ATisEqual((ATerm)arg0, (ATerm)arg1);
}

/**
 * Tests equality of two PERR_NatCons. A constant time operation.
 * \param[in] arg0 first PERR_NatCon to be compared
 * \param[in] arg1 second PERR_NatCon to be compared
 * \return ATtrue if #arg0 was equal to #arg1, ATfalse otherwise
 */
ATbool _PERR_isEqualNatCon(PERR_NatCon arg0, PERR_NatCon arg1) {
  return ATisEqual((ATerm)arg0, (ATerm)arg1);
}

/**
 * Tests equality of two PERR_LexLayouts. A constant time operation.
 * \param[in] arg0 first PERR_LexLayout to be compared
 * \param[in] arg1 second PERR_LexLayout to be compared
 * \return ATtrue if #arg0 was equal to #arg1, ATfalse otherwise
 */
ATbool _PERR_isEqualLexLayout(PERR_LexLayout arg0, PERR_LexLayout arg1) {
  return ATisEqual((ATerm)arg0, (ATerm)arg1);
}

/**
 * Tests equality of two PERR_Starts. A constant time operation.
 * \param[in] arg0 first PERR_Start to be compared
 * \param[in] arg1 second PERR_Start to be compared
 * \return ATtrue if #arg0 was equal to #arg1, ATfalse otherwise
 */
ATbool _PERR_isEqualStart(PERR_Start arg0, PERR_Start arg1) {
  return ATisEqual((ATerm)arg0, (ATerm)arg1);
}

/**
 * Tests equality of two PERR_Subjects. A constant time operation.
 * \param[in] arg0 first PERR_Subject to be compared
 * \param[in] arg1 second PERR_Subject to be compared
 * \return ATtrue if #arg0 was equal to #arg1, ATfalse otherwise
 */
ATbool _PERR_isEqualSubject(PERR_Subject arg0, PERR_Subject arg1) {
  return ATisEqual((ATerm)arg0, (ATerm)arg1);
}

/**
 * Tests equality of two PERR_Errors. A constant time operation.
 * \param[in] arg0 first PERR_Error to be compared
 * \param[in] arg1 second PERR_Error to be compared
 * \return ATtrue if #arg0 was equal to #arg1, ATfalse otherwise
 */
ATbool _PERR_isEqualError(PERR_Error arg0, PERR_Error arg1) {
  return ATisEqual((ATerm)arg0, (ATerm)arg1);
}

/**
 * Tests equality of two PERR_Summarys. A constant time operation.
 * \param[in] arg0 first PERR_Summary to be compared
 * \param[in] arg1 second PERR_Summary to be compared
 * \return ATtrue if #arg0 was equal to #arg1, ATfalse otherwise
 */
ATbool _PERR_isEqualSummary(PERR_Summary arg0, PERR_Summary arg1) {
  return ATisEqual((ATerm)arg0, (ATerm)arg1);
}

/**
 * Tests equality of two PERR_SubjectLists. A constant time operation.
 * \param[in] arg0 first PERR_SubjectList to be compared
 * \param[in] arg1 second PERR_SubjectList to be compared
 * \return ATtrue if #arg0 was equal to #arg1, ATfalse otherwise
 */
ATbool _PERR_isEqualSubjectList(PERR_SubjectList arg0, PERR_SubjectList arg1) {
  return ATisEqual((ATerm)arg0, (ATerm)arg1);
}

/**
 * Tests equality of two PERR_ErrorLists. A constant time operation.
 * \param[in] arg0 first PERR_ErrorList to be compared
 * \param[in] arg1 second PERR_ErrorList to be compared
 * \return ATtrue if #arg0 was equal to #arg1, ATfalse otherwise
 */
ATbool _PERR_isEqualErrorList(PERR_ErrorList arg0, PERR_ErrorList arg1) {
  return ATisEqual((ATerm)arg0, (ATerm)arg1);
}

/**
 * Tests equality of two PERR_Locations. A constant time operation.
 * \param[in] arg0 first PERR_Location to be compared
 * \param[in] arg1 second PERR_Location to be compared
 * \return ATtrue if #arg0 was equal to #arg1, ATfalse otherwise
 */
ATbool _PERR_isEqualLocation(PERR_Location arg0, PERR_Location arg1) {
  return ATisEqual((ATerm)arg0, (ATerm)arg1);
}

/**
 * Tests equality of two PERR_Areas. A constant time operation.
 * \param[in] arg0 first PERR_Area to be compared
 * \param[in] arg1 second PERR_Area to be compared
 * \return ATtrue if #arg0 was equal to #arg1, ATfalse otherwise
 */
ATbool _PERR_isEqualArea(PERR_Area arg0, PERR_Area arg1) {
  return ATisEqual((ATerm)arg0, (ATerm)arg1);
}

/**
 * Tests equality of two PERR_Slices. A constant time operation.
 * \param[in] arg0 first PERR_Slice to be compared
 * \param[in] arg1 second PERR_Slice to be compared
 * \return ATtrue if #arg0 was equal to #arg1, ATfalse otherwise
 */
ATbool _PERR_isEqualSlice(PERR_Slice arg0, PERR_Slice arg1) {
  return ATisEqual((ATerm)arg0, (ATerm)arg1);
}

/**
 * Tests equality of two PERR_AreaAreass. A constant time operation.
 * \param[in] arg0 first PERR_AreaAreas to be compared
 * \param[in] arg1 second PERR_AreaAreas to be compared
 * \return ATtrue if #arg0 was equal to #arg1, ATfalse otherwise
 */
ATbool _PERR_isEqualAreaAreas(PERR_AreaAreas arg0, PERR_AreaAreas arg1) {
  return ATisEqual((ATerm)arg0, (ATerm)arg1);
}

/**
 * Assert whether a PERR_OptLayout is any of the valid alternatives, or not. This analysis does not go any deeper than the top level
 * \param[in] arg input PERR_OptLayout
 * \return ATtrue if #arg corresponds to the expected signature, or ATfalse otherwise
 */
ATbool PERR_isValidOptLayout(PERR_OptLayout arg) {
  if (PERR_isOptLayoutAbsent(arg)) {
    return ATtrue;
  }
  else if (PERR_isOptLayoutPresent(arg)) {
    return ATtrue;
  }
  return ATfalse;
}

/**
 * Assert whether a PERR_OptLayout is a Absent by checking against the following ATerm pattern: appl(prod([],cf(opt(layout)),no-attrs),[]). May not be used to assert correctness of the PERR_OptLayout
 * \param[in] arg input PERR_OptLayout
 * \return ATtrue if #arg corresponds to the signature of a Absent, or ATfalse otherwise
 */
inline ATbool PERR_isOptLayoutAbsent(PERR_OptLayout arg){
  /* checking for: appl */
  if (ATgetType((ATerm)arg) == AT_APPL && ATgetAFun((ATermAppl)arg) == PERR_afun0) {
    ATerm arg_arg0 = ATgetArgument(arg, 0);
    /* checking for: prod */
    if (ATgetType((ATerm)arg_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0) == PERR_afun1) {
      ATerm arg_arg0_arg0 = ATgetArgument(arg_arg0, 0);
      if (ATisEmpty((ATermList)arg_arg0_arg0)) {
        ATerm arg_arg0_arg1 = ATgetArgument(arg_arg0, 1);
        /* checking for: cf */
        if (ATgetType((ATerm)arg_arg0_arg1) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg1) == PERR_afun2) {
          ATerm arg_arg0_arg1_arg0 = ATgetArgument(arg_arg0_arg1, 0);
          /* checking for: opt */
          if (ATgetType((ATerm)arg_arg0_arg1_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg1_arg0) == PERR_afun3) {
            ATerm arg_arg0_arg1_arg0_arg0 = ATgetArgument(arg_arg0_arg1_arg0, 0);
            /* checking for: layout */
            if (ATgetType((ATerm)arg_arg0_arg1_arg0_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg1_arg0_arg0) == PERR_afun4) {
              ATerm arg_arg0_arg2 = ATgetArgument(arg_arg0, 2);
              /* checking for: no-attrs */
              if (ATgetType((ATerm)arg_arg0_arg2) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg2) == PERR_afun5) {
                ATerm arg_arg1 = ATgetArgument(arg, 1);
                if (ATisEmpty((ATermList)arg_arg1)) {
                  return ATtrue;
                }
              }
            }
          }
        }
      }
    }
  }
  return ATfalse;
}

/**
 * Assert whether a PERR_OptLayout is a Present by checking against the following ATerm pattern: appl(prod([cf(layout)],cf(opt(layout)),no-attrs),[<"layout"("Layout")>]). May not be used to assert correctness of the PERR_OptLayout
 * \param[in] arg input PERR_OptLayout
 * \return ATtrue if #arg corresponds to the signature of a Present, or ATfalse otherwise
 */
inline ATbool PERR_isOptLayoutPresent(PERR_OptLayout arg){
  /* checking for: appl */
  if (ATgetType((ATerm)arg) == AT_APPL && ATgetAFun((ATermAppl)arg) == PERR_afun0) {
    ATerm arg_arg0 = ATgetArgument(arg, 0);
    /* checking for: prod */
    if (ATgetType((ATerm)arg_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0) == PERR_afun1) {
      ATerm arg_arg0_arg0 = ATgetArgument(arg_arg0, 0);
      if (ATgetType((ATerm)arg_arg0_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0) == ATfalse) {
        ATerm arg_arg0_arg0_head;
        ATermList arg_arg0_arg0_list = (ATermList)arg_arg0_arg0;
        arg_arg0_arg0_head = ATgetFirst(arg_arg0_arg0_list);
        arg_arg0_arg0_list = ATgetNext(arg_arg0_arg0_list);
        /* checking for: cf */
        if (ATgetType((ATerm)arg_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head) == PERR_afun2) {
          ATerm arg_arg0_arg0_head_arg0 = ATgetArgument(arg_arg0_arg0_head, 0);
          /* checking for: layout */
          if (ATgetType((ATerm)arg_arg0_arg0_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0) == PERR_afun4) {
            if (ATgetType((ATerm)arg_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0_list) == ATtrue) {
              ATerm arg_arg0_arg1 = ATgetArgument(arg_arg0, 1);
              /* checking for: cf */
              if (ATgetType((ATerm)arg_arg0_arg1) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg1) == PERR_afun2) {
                ATerm arg_arg0_arg1_arg0 = ATgetArgument(arg_arg0_arg1, 0);
                /* checking for: opt */
                if (ATgetType((ATerm)arg_arg0_arg1_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg1_arg0) == PERR_afun3) {
                  ATerm arg_arg0_arg1_arg0_arg0 = ATgetArgument(arg_arg0_arg1_arg0, 0);
                  /* checking for: layout */
                  if (ATgetType((ATerm)arg_arg0_arg1_arg0_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg1_arg0_arg0) == PERR_afun4) {
                    ATerm arg_arg0_arg2 = ATgetArgument(arg_arg0, 2);
                    /* checking for: no-attrs */
                    if (ATgetType((ATerm)arg_arg0_arg2) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg2) == PERR_afun5) {
                      ATerm arg_arg1 = ATgetArgument(arg, 1);
                      if (ATgetType((ATerm)arg_arg1) == AT_LIST && ATisEmpty((ATermList)arg_arg1) == ATfalse) {
                        ATermList arg_arg1_list = (ATermList)arg_arg1;
                        arg_arg1_list = ATgetNext(arg_arg1_list);
                        if (ATgetType((ATerm)arg_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_list) == ATtrue) {
                          return ATtrue;
                        }
                      }
                    }
                  }
                }
              }
            }
          }
        }
      }
    }
  }
  return ATfalse;
}

/**
 * Assert whether a PERR_OptLayout has a layout. 
 * \param[in] arg input PERR_OptLayout
 * \return ATtrue if the PERR_OptLayout had a layout, or ATfalse otherwise
 */
ATbool PERR_hasOptLayoutLayout(PERR_OptLayout arg) {
  if (PERR_isOptLayoutPresent(arg)) {
    return ATtrue;
  }
  return ATfalse;
}

/**
 * Get the layout PERR_Layout of a PERR_OptLayout. Note that the precondition is that this PERR_OptLayout actually has a layout
 * \param[in] arg input PERR_OptLayout
 * \return the layout of #arg, if it exist or an undefined value if it does not
 */
PERR_Layout PERR_getOptLayoutLayout(PERR_OptLayout arg) {
  
    return (PERR_Layout)ATgetFirst((ATermList)ATgetArgument((ATermAppl)arg, 1));
}

/**
 * Set the layout of a PERR_OptLayout. The precondition being that this PERR_OptLayout actually has a layout
 * \param[in] arg input PERR_OptLayout
 * \param[in] layout new PERR_Layout to set in #arg
 * \return A new PERR_OptLayout with layout at the right place, or a core dump if #arg did not have a layout
 */
PERR_OptLayout PERR_setOptLayoutLayout(PERR_OptLayout arg, PERR_Layout layout) {
  if (PERR_isOptLayoutPresent(arg)) {
    return (PERR_OptLayout)ATsetArgument((ATermAppl)arg, (ATerm)ATreplace((ATermList)ATgetArgument((ATermAppl)arg, 1), (ATerm)((ATerm) layout), 0), 1);
  }

  ATabort("OptLayout has no Layout: %t\n", arg);
  return (PERR_OptLayout)NULL;
}

/**
 * Assert whether a PERR_Layout is any of the valid alternatives, or not. This analysis does not go any deeper than the top level
 * \param[in] arg input PERR_Layout
 * \return ATtrue if #arg corresponds to the expected signature, or ATfalse otherwise
 */
ATbool PERR_isValidLayout(PERR_Layout arg) {
  if (PERR_isLayoutLexToCf(arg)) {
    return ATtrue;
  }
  return ATfalse;
}

/**
 * Assert whether a PERR_Layout is a Lex-to-cf by checking against the following ATerm pattern: appl(prod([lex(iter(layout))],cf(layout),no-attrs),[appl(list(lex(iter(layout))),<"list"("Lex-Layout-list")>)]). Always returns ATtrue
 * \param[in] arg input PERR_Layout
 * \return ATtrue if #arg corresponds to the signature of a Lex-to-cf, or ATfalse otherwise
 */
inline ATbool PERR_isLayoutLexToCf(PERR_Layout arg){
  /* checking for: appl */
  if (ATgetType((ATerm)arg) == AT_APPL && ATgetAFun((ATermAppl)arg) == PERR_afun0) {
    ATerm arg_arg0 = ATgetArgument(arg, 0);
    /* checking for: prod */
    if (ATgetType((ATerm)arg_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0) == PERR_afun1) {
      ATerm arg_arg0_arg0 = ATgetArgument(arg_arg0, 0);
      if (ATgetType((ATerm)arg_arg0_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0) == ATfalse) {
        ATerm arg_arg0_arg0_head;
        ATermList arg_arg0_arg0_list = (ATermList)arg_arg0_arg0;
        arg_arg0_arg0_head = ATgetFirst(arg_arg0_arg0_list);
        arg_arg0_arg0_list = ATgetNext(arg_arg0_arg0_list);
        /* checking for: lex */
        if (ATgetType((ATerm)arg_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head) == PERR_afun6) {
          ATerm arg_arg0_arg0_head_arg0 = ATgetArgument(arg_arg0_arg0_head, 0);
          /* checking for: iter */
          if (ATgetType((ATerm)arg_arg0_arg0_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0) == PERR_afun7) {
            ATerm arg_arg0_arg0_head_arg0_arg0 = ATgetArgument(arg_arg0_arg0_head_arg0, 0);
            /* checking for: layout */
            if (ATgetType((ATerm)arg_arg0_arg0_head_arg0_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0_arg0) == PERR_afun4) {
              if (ATgetType((ATerm)arg_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0_list) == ATtrue) {
                ATerm arg_arg0_arg1 = ATgetArgument(arg_arg0, 1);
                /* checking for: cf */
                if (ATgetType((ATerm)arg_arg0_arg1) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg1) == PERR_afun2) {
                  ATerm arg_arg0_arg1_arg0 = ATgetArgument(arg_arg0_arg1, 0);
                  /* checking for: layout */
                  if (ATgetType((ATerm)arg_arg0_arg1_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg1_arg0) == PERR_afun4) {
                    ATerm arg_arg0_arg2 = ATgetArgument(arg_arg0, 2);
                    /* checking for: no-attrs */
                    if (ATgetType((ATerm)arg_arg0_arg2) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg2) == PERR_afun5) {
                      ATerm arg_arg1 = ATgetArgument(arg, 1);
                      if (ATgetType((ATerm)arg_arg1) == AT_LIST && ATisEmpty((ATermList)arg_arg1) == ATfalse) {
                        ATerm arg_arg1_head;
                        ATermList arg_arg1_list = (ATermList)arg_arg1;
                        arg_arg1_head = ATgetFirst(arg_arg1_list);
                        arg_arg1_list = ATgetNext(arg_arg1_list);
                        /* checking for: appl */
                        if (ATgetType((ATerm)arg_arg1_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head) == PERR_afun0) {
                          ATerm arg_arg1_head_arg0 = ATgetArgument(arg_arg1_head, 0);
                          /* checking for: list */
                          if (ATgetType((ATerm)arg_arg1_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0) == PERR_afun8) {
                            ATerm arg_arg1_head_arg0_arg0 = ATgetArgument(arg_arg1_head_arg0, 0);
                            /* checking for: lex */
                            if (ATgetType((ATerm)arg_arg1_head_arg0_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg0) == PERR_afun6) {
                              ATerm arg_arg1_head_arg0_arg0_arg0 = ATgetArgument(arg_arg1_head_arg0_arg0, 0);
                              /* checking for: iter */
                              if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg0_arg0) == PERR_afun7) {
                                ATerm arg_arg1_head_arg0_arg0_arg0_arg0 = ATgetArgument(arg_arg1_head_arg0_arg0_arg0, 0);
                                /* checking for: layout */
                                if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_arg0_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg0_arg0_arg0) == PERR_afun4) {
                                  if (ATgetType((ATerm)arg_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_list) == ATtrue) {
                                    return ATtrue;
                                  }
                                }
                              }
                            }
                          }
                        }
                      }
                    }
                  }
                }
              }
            }
          }
        }
      }
    }
  }
  return ATfalse;
}

/**
 * Assert whether a PERR_Layout has a list. 
 * \param[in] arg input PERR_Layout
 * \return ATtrue if the PERR_Layout had a list, or ATfalse otherwise
 */
ATbool PERR_hasLayoutList(PERR_Layout arg) {
  if (PERR_isLayoutLexToCf(arg)) {
    return ATtrue;
  }
  return ATfalse;
}

/**
 * Get the list PERR_LexLayoutList of a PERR_Layout. Note that the precondition is that this PERR_Layout actually has a list
 * \param[in] arg input PERR_Layout
 * \return the list of #arg, if it exist or an undefined value if it does not
 */
PERR_LexLayoutList PERR_getLayoutList(PERR_Layout arg) {
  
    return (PERR_LexLayoutList)ATgetArgument((ATermAppl)ATgetFirst((ATermList)ATgetArgument((ATermAppl)arg, 1)), 1);
}

/**
 * Set the list of a PERR_Layout. The precondition being that this PERR_Layout actually has a list
 * \param[in] arg input PERR_Layout
 * \param[in] list new PERR_LexLayoutList to set in #arg
 * \return A new PERR_Layout with list at the right place, or a core dump if #arg did not have a list
 */
PERR_Layout PERR_setLayoutList(PERR_Layout arg, PERR_LexLayoutList list) {
  if (PERR_isLayoutLexToCf(arg)) {
    return (PERR_Layout)ATsetArgument((ATermAppl)arg, (ATerm)ATreplace((ATermList)ATgetArgument((ATermAppl)arg, 1), (ATerm)ATsetArgument((ATermAppl)ATgetFirst((ATermList)ATgetArgument((ATermAppl)arg, 1)), (ATerm)((ATerm) list), 1), 0), 1);
  }

  ATabort("Layout has no List: %t\n", arg);
  return (PERR_Layout)NULL;
}

/**
 * Assert whether a PERR_LexLayoutList is any of the valid alternatives, or not. This analysis does not go any deeper than the top level
 * \param[in] arg input PERR_LexLayoutList
 * \return ATtrue if #arg corresponds to the expected signature, or ATfalse otherwise
 */
ATbool PERR_isValidLexLayoutList(PERR_LexLayoutList arg) {
  if (PERR_isLexLayoutListEmpty(arg)) {
    return ATtrue;
  }
  else if (PERR_isLexLayoutListSingle(arg)) {
    return ATtrue;
  }
  else if (PERR_isLexLayoutListMany(arg)) {
    return ATtrue;
  }
  return ATfalse;
}

/**
 * Assert whether a PERR_LexLayoutList is a empty by checking against the following ATerm pattern: []. May not be used to assert correctness of the PERR_LexLayoutList
 * \param[in] arg input PERR_LexLayoutList
 * \return ATtrue if #arg corresponds to the signature of a empty, or ATfalse otherwise
 */
inline ATbool PERR_isLexLayoutListEmpty(PERR_LexLayoutList arg){
  if (ATisEmpty((ATermList)arg)) {
    return ATtrue;
  }
  return ATfalse;
}

/**
 * Assert whether a PERR_LexLayoutList is a single by checking against the following ATerm pattern: [<head(Lex-Layout)>]. May not be used to assert correctness of the PERR_LexLayoutList
 * \param[in] arg input PERR_LexLayoutList
 * \return ATtrue if #arg corresponds to the signature of a single, or ATfalse otherwise
 */
inline ATbool PERR_isLexLayoutListSingle(PERR_LexLayoutList arg){
  if (ATgetType((ATerm)arg) == AT_LIST && ATisEmpty((ATermList)arg) == ATfalse) {
    ATermList arg_list = (ATermList)arg;
    arg_list = ATgetNext(arg_list);
    if (ATgetType((ATerm)arg_list) == AT_LIST && ATisEmpty((ATermList)arg_list) == ATtrue) {
      return ATtrue;
    }
  }
  return ATfalse;
}

/**
 * Assert whether a PERR_LexLayoutList is a many by checking against the following ATerm pattern: [<head(Lex-Layout)>,<[tail(Lex-Layout-list)]>]. May not be used to assert correctness of the PERR_LexLayoutList
 * \param[in] arg input PERR_LexLayoutList
 * \return ATtrue if #arg corresponds to the signature of a many, or ATfalse otherwise
 */
inline ATbool PERR_isLexLayoutListMany(PERR_LexLayoutList arg){
  if (ATgetType((ATerm)arg) == AT_LIST && ATisEmpty((ATermList)arg) == ATfalse) {
    return ATtrue;
  }
  return ATfalse;
}

/**
 * Assert whether a PERR_LexLayoutList has a head. 
 * \param[in] arg input PERR_LexLayoutList
 * \return ATtrue if the PERR_LexLayoutList had a head, or ATfalse otherwise
 */
ATbool PERR_hasLexLayoutListHead(PERR_LexLayoutList arg) {
  if (PERR_isLexLayoutListSingle(arg)) {
    return ATtrue;
  }
  else if (PERR_isLexLayoutListMany(arg)) {
    return ATtrue;
  }
  return ATfalse;
}

/**
 * Assert whether a PERR_LexLayoutList has a tail. 
 * \param[in] arg input PERR_LexLayoutList
 * \return ATtrue if the PERR_LexLayoutList had a tail, or ATfalse otherwise
 */
ATbool PERR_hasLexLayoutListTail(PERR_LexLayoutList arg) {
  if (PERR_isLexLayoutListMany(arg)) {
    return ATtrue;
  }
  return ATfalse;
}

/**
 * Get the head PERR_LexLayout of a PERR_LexLayoutList. Note that the precondition is that this PERR_LexLayoutList actually has a head
 * \param[in] arg input PERR_LexLayoutList
 * \return the head of #arg, if it exist or an undefined value if it does not
 */
PERR_LexLayout PERR_getLexLayoutListHead(PERR_LexLayoutList arg) {
  if (PERR_isLexLayoutListSingle(arg)) {
    return (PERR_LexLayout)ATgetFirst((ATermList)arg);
  }
  else 
    return (PERR_LexLayout)ATgetFirst((ATermList)arg);
}

/**
 * Get the tail PERR_LexLayoutList of a PERR_LexLayoutList. Note that the precondition is that this PERR_LexLayoutList actually has a tail
 * \param[in] arg input PERR_LexLayoutList
 * \return the tail of #arg, if it exist or an undefined value if it does not
 */
PERR_LexLayoutList PERR_getLexLayoutListTail(PERR_LexLayoutList arg) {
  
    return (PERR_LexLayoutList)ATgetNext((ATermList)arg);
}

/**
 * Set the head of a PERR_LexLayoutList. The precondition being that this PERR_LexLayoutList actually has a head
 * \param[in] arg input PERR_LexLayoutList
 * \param[in] head new PERR_LexLayout to set in #arg
 * \return A new PERR_LexLayoutList with head at the right place, or a core dump if #arg did not have a head
 */
PERR_LexLayoutList PERR_setLexLayoutListHead(PERR_LexLayoutList arg, PERR_LexLayout head) {
  if (PERR_isLexLayoutListSingle(arg)) {
    return (PERR_LexLayoutList)ATreplace((ATermList)arg, (ATerm)((ATerm) head), 0);
  }
  else if (PERR_isLexLayoutListMany(arg)) {
    return (PERR_LexLayoutList)ATreplace((ATermList)arg, (ATerm)((ATerm) head), 0);
  }

  ATabort("LexLayoutList has no Head: %t\n", arg);
  return (PERR_LexLayoutList)NULL;
}

/**
 * Set the tail of a PERR_LexLayoutList. The precondition being that this PERR_LexLayoutList actually has a tail
 * \param[in] arg input PERR_LexLayoutList
 * \param[in] tail new PERR_LexLayoutList to set in #arg
 * \return A new PERR_LexLayoutList with tail at the right place, or a core dump if #arg did not have a tail
 */
PERR_LexLayoutList PERR_setLexLayoutListTail(PERR_LexLayoutList arg, PERR_LexLayoutList tail) {
  if (PERR_isLexLayoutListMany(arg)) {
    return (PERR_LexLayoutList)ATreplaceTail((ATermList)arg, (ATermList)((ATerm) tail), 1);
  }

  ATabort("LexLayoutList has no Tail: %t\n", arg);
  return (PERR_LexLayoutList)NULL;
}

/**
 * Assert whether a PERR_LexStrChar is any of the valid alternatives, or not. This analysis does not go any deeper than the top level
 * \param[in] arg input PERR_LexStrChar
 * \return ATtrue if #arg corresponds to the expected signature, or ATfalse otherwise
 */
ATbool PERR_isValidLexStrChar(PERR_LexStrChar arg) {
  if (PERR_isLexStrCharNewline(arg)) {
    return ATtrue;
  }
  else if (PERR_isLexStrCharTab(arg)) {
    return ATtrue;
  }
  else if (PERR_isLexStrCharQuote(arg)) {
    return ATtrue;
  }
  else if (PERR_isLexStrCharBackslash(arg)) {
    return ATtrue;
  }
  else if (PERR_isLexStrCharDecimal(arg)) {
    return ATtrue;
  }
  else if (PERR_isLexStrCharNormal(arg)) {
    return ATtrue;
  }
  return ATfalse;
}

/**
 * Assert whether a PERR_LexStrChar is a newline by checking against the following ATerm pattern: appl(prod([lit("\\n")],lex(sort("StrChar")),attrs([term(cons("newline"))])),[appl(prod([char-class([92]),char-class([110])],lit("\\n"),no-attrs),[92,110])]). May not be used to assert correctness of the PERR_LexStrChar
 * \param[in] arg input PERR_LexStrChar
 * \return ATtrue if #arg corresponds to the signature of a newline, or ATfalse otherwise
 */
inline ATbool PERR_isLexStrCharNewline(PERR_LexStrChar arg){
  /* checking for: appl */
  if (ATgetType((ATerm)arg) == AT_APPL && ATgetAFun((ATermAppl)arg) == PERR_afun0) {
    ATerm arg_arg0 = ATgetArgument(arg, 0);
    /* checking for: prod */
    if (ATgetType((ATerm)arg_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0) == PERR_afun1) {
      ATerm arg_arg0_arg0 = ATgetArgument(arg_arg0, 0);
      if (ATgetType((ATerm)arg_arg0_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0) == ATfalse) {
        ATerm arg_arg0_arg0_head;
        ATermList arg_arg0_arg0_list = (ATermList)arg_arg0_arg0;
        arg_arg0_arg0_head = ATgetFirst(arg_arg0_arg0_list);
        arg_arg0_arg0_list = ATgetNext(arg_arg0_arg0_list);
        /* checking for: lit */
        if (ATgetType((ATerm)arg_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head) == PERR_afun9) {
          ATerm arg_arg0_arg0_head_arg0 = ATgetArgument(arg_arg0_arg0_head, 0);
          /* checking for: \n */
          if (ATgetType((ATerm)arg_arg0_arg0_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0) == PERR_afun10) {
            if (ATgetType((ATerm)arg_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0_list) == ATtrue) {
              ATerm arg_arg0_arg1 = ATgetArgument(arg_arg0, 1);
              /* checking for: lex */
              if (ATgetType((ATerm)arg_arg0_arg1) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg1) == PERR_afun6) {
                ATerm arg_arg0_arg1_arg0 = ATgetArgument(arg_arg0_arg1, 0);
                /* checking for: sort */
                if (ATgetType((ATerm)arg_arg0_arg1_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg1_arg0) == PERR_afun11) {
                  ATerm arg_arg0_arg1_arg0_arg0 = ATgetArgument(arg_arg0_arg1_arg0, 0);
                  /* checking for: StrChar */
                  if (ATgetType((ATerm)arg_arg0_arg1_arg0_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg1_arg0_arg0) == PERR_afun12) {
                    ATerm arg_arg0_arg2 = ATgetArgument(arg_arg0, 2);
                    /* checking for: attrs */
                    if (ATgetType((ATerm)arg_arg0_arg2) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg2) == PERR_afun13) {
                      ATerm arg_arg0_arg2_arg0 = ATgetArgument(arg_arg0_arg2, 0);
                      if (ATgetType((ATerm)arg_arg0_arg2_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg2_arg0) == ATfalse) {
                        ATerm arg_arg0_arg2_arg0_head;
                        ATermList arg_arg0_arg2_arg0_list = (ATermList)arg_arg0_arg2_arg0;
                        arg_arg0_arg2_arg0_head = ATgetFirst(arg_arg0_arg2_arg0_list);
                        arg_arg0_arg2_arg0_list = ATgetNext(arg_arg0_arg2_arg0_list);
                        /* checking for: term */
                        if (ATgetType((ATerm)arg_arg0_arg2_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg2_arg0_head) == PERR_afun14) {
                          ATerm arg_arg0_arg2_arg0_head_arg0 = ATgetArgument(arg_arg0_arg2_arg0_head, 0);
                          /* checking for: cons */
                          if (ATgetType((ATerm)arg_arg0_arg2_arg0_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg2_arg0_head_arg0) == PERR_afun15) {
                            ATerm arg_arg0_arg2_arg0_head_arg0_arg0 = ATgetArgument(arg_arg0_arg2_arg0_head_arg0, 0);
                            /* checking for: newline */
                            if (ATgetType((ATerm)arg_arg0_arg2_arg0_head_arg0_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg2_arg0_head_arg0_arg0) == PERR_afun16) {
                              if (ATgetType((ATerm)arg_arg0_arg2_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg2_arg0_list) == ATtrue) {
                                ATerm arg_arg1 = ATgetArgument(arg, 1);
                                if (ATgetType((ATerm)arg_arg1) == AT_LIST && ATisEmpty((ATermList)arg_arg1) == ATfalse) {
                                  ATerm arg_arg1_head;
                                  ATermList arg_arg1_list = (ATermList)arg_arg1;
                                  arg_arg1_head = ATgetFirst(arg_arg1_list);
                                  arg_arg1_list = ATgetNext(arg_arg1_list);
                                  /* checking for: appl */
                                  if (ATgetType((ATerm)arg_arg1_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head) == PERR_afun0) {
                                    ATerm arg_arg1_head_arg0 = ATgetArgument(arg_arg1_head, 0);
                                    /* checking for: prod */
                                    if (ATgetType((ATerm)arg_arg1_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0) == PERR_afun1) {
                                      ATerm arg_arg1_head_arg0_arg0 = ATgetArgument(arg_arg1_head_arg0, 0);
                                      if (ATgetType((ATerm)arg_arg1_head_arg0_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0) == ATfalse) {
                                        ATerm arg_arg1_head_arg0_arg0_head;
                                        ATermList arg_arg1_head_arg0_arg0_list = (ATermList)arg_arg1_head_arg0_arg0;
                                        arg_arg1_head_arg0_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_list);
                                        arg_arg1_head_arg0_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_list);
                                        /* checking for: char-class */
                                        if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg0_head) == PERR_afun17) {
                                          ATerm arg_arg1_head_arg0_arg0_head_arg0 = ATgetArgument(arg_arg1_head_arg0_arg0_head, 0);
                                          if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0) == ATfalse) {
                                            ATerm arg_arg1_head_arg0_arg0_head_arg0_head;
                                            ATermList arg_arg1_head_arg0_arg0_head_arg0_list = (ATermList)arg_arg1_head_arg0_arg0_head_arg0;
                                            arg_arg1_head_arg0_arg0_head_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_head_arg0_list);
                                            arg_arg1_head_arg0_arg0_head_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_head_arg0_list);
                                            if (ATgetInt((ATermInt)arg_arg1_head_arg0_arg0_head_arg0_head) == 92) {

                                              if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0_list) == ATtrue) {
                                                if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_list) == ATfalse) {
                                                  arg_arg1_head_arg0_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_list);
                                                  arg_arg1_head_arg0_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_list);
                                                  /* checking for: char-class */
                                                  if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg0_head) == PERR_afun17) {
                                                    ATerm arg_arg1_head_arg0_arg0_head_arg0 = ATgetArgument(arg_arg1_head_arg0_arg0_head, 0);
                                                    if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0) == ATfalse) {
                                                      ATerm arg_arg1_head_arg0_arg0_head_arg0_head;
                                                      ATermList arg_arg1_head_arg0_arg0_head_arg0_list = (ATermList)arg_arg1_head_arg0_arg0_head_arg0;
                                                      arg_arg1_head_arg0_arg0_head_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                      arg_arg1_head_arg0_arg0_head_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                      if (ATgetInt((ATermInt)arg_arg1_head_arg0_arg0_head_arg0_head) == 110) {

                                                        if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0_list) == ATtrue) {
                                                          if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_list) == ATtrue) {
                                                            ATerm arg_arg1_head_arg0_arg1 = ATgetArgument(arg_arg1_head_arg0, 1);
                                                            /* checking for: lit */
                                                            if (ATgetType((ATerm)arg_arg1_head_arg0_arg1) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg1) == PERR_afun9) {
                                                              ATerm arg_arg1_head_arg0_arg1_arg0 = ATgetArgument(arg_arg1_head_arg0_arg1, 0);
                                                              /* checking for: \n */
                                                              if (ATgetType((ATerm)arg_arg1_head_arg0_arg1_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg1_arg0) == PERR_afun10) {
                                                                ATerm arg_arg1_head_arg0_arg2 = ATgetArgument(arg_arg1_head_arg0, 2);
                                                                /* checking for: no-attrs */
                                                                if (ATgetType((ATerm)arg_arg1_head_arg0_arg2) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg2) == PERR_afun5) {
                                                                  ATerm arg_arg1_head_arg1 = ATgetArgument(arg_arg1_head, 1);
                                                                  if (ATgetType((ATerm)arg_arg1_head_arg1) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg1) == ATfalse) {
                                                                    ATerm arg_arg1_head_arg1_head;
                                                                    ATermList arg_arg1_head_arg1_list = (ATermList)arg_arg1_head_arg1;
                                                                    arg_arg1_head_arg1_head = ATgetFirst(arg_arg1_head_arg1_list);
                                                                    arg_arg1_head_arg1_list = ATgetNext(arg_arg1_head_arg1_list);
                                                                    if (ATgetInt((ATermInt)arg_arg1_head_arg1_head) == 92) {

                                                                      if (ATgetType((ATerm)arg_arg1_head_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg1_list) == ATfalse) {
                                                                        arg_arg1_head_arg1_head = ATgetFirst(arg_arg1_head_arg1_list);
                                                                        arg_arg1_head_arg1_list = ATgetNext(arg_arg1_head_arg1_list);
                                                                        if (ATgetInt((ATermInt)arg_arg1_head_arg1_head) == 110) {

                                                                          if (ATgetType((ATerm)arg_arg1_head_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg1_list) == ATtrue) {
                                                                            if (ATgetType((ATerm)arg_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_list) == ATtrue) {
                                                                              return ATtrue;
                                                                            }
                                                                          }
                                                                        }
                                                                      }
                                                                    }
                                                                  }
                                                                }
                                                              }
                                                            }
                                                          }
                                                        }
                                                      }
                                                    }
                                                  }
                                                }
                                              }
                                            }
                                          }
                                        }
                                      }
                                    }
                                  }
                                }
                              }
                            }
                          }
                        }
                      }
                    }
                  }
                }
              }
            }
          }
        }
      }
    }
  }
  return ATfalse;
}

/**
 * Assert whether a PERR_LexStrChar is a tab by checking against the following ATerm pattern: appl(prod([lit("\\t")],lex(sort("StrChar")),attrs([term(cons("tab"))])),[appl(prod([char-class([92]),char-class([116])],lit("\\t"),no-attrs),[92,116])]). May not be used to assert correctness of the PERR_LexStrChar
 * \param[in] arg input PERR_LexStrChar
 * \return ATtrue if #arg corresponds to the signature of a tab, or ATfalse otherwise
 */
inline ATbool PERR_isLexStrCharTab(PERR_LexStrChar arg){
  /* checking for: appl */
  if (ATgetType((ATerm)arg) == AT_APPL && ATgetAFun((ATermAppl)arg) == PERR_afun0) {
    ATerm arg_arg0 = ATgetArgument(arg, 0);
    /* checking for: prod */
    if (ATgetType((ATerm)arg_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0) == PERR_afun1) {
      ATerm arg_arg0_arg0 = ATgetArgument(arg_arg0, 0);
      if (ATgetType((ATerm)arg_arg0_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0) == ATfalse) {
        ATerm arg_arg0_arg0_head;
        ATermList arg_arg0_arg0_list = (ATermList)arg_arg0_arg0;
        arg_arg0_arg0_head = ATgetFirst(arg_arg0_arg0_list);
        arg_arg0_arg0_list = ATgetNext(arg_arg0_arg0_list);
        /* checking for: lit */
        if (ATgetType((ATerm)arg_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head) == PERR_afun9) {
          ATerm arg_arg0_arg0_head_arg0 = ATgetArgument(arg_arg0_arg0_head, 0);
          /* checking for: \t */
          if (ATgetType((ATerm)arg_arg0_arg0_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0) == PERR_afun18) {
            if (ATgetType((ATerm)arg_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0_list) == ATtrue) {
              ATerm arg_arg0_arg1 = ATgetArgument(arg_arg0, 1);
              /* checking for: lex */
              if (ATgetType((ATerm)arg_arg0_arg1) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg1) == PERR_afun6) {
                ATerm arg_arg0_arg1_arg0 = ATgetArgument(arg_arg0_arg1, 0);
                /* checking for: sort */
                if (ATgetType((ATerm)arg_arg0_arg1_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg1_arg0) == PERR_afun11) {
                  ATerm arg_arg0_arg1_arg0_arg0 = ATgetArgument(arg_arg0_arg1_arg0, 0);
                  /* checking for: StrChar */
                  if (ATgetType((ATerm)arg_arg0_arg1_arg0_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg1_arg0_arg0) == PERR_afun12) {
                    ATerm arg_arg0_arg2 = ATgetArgument(arg_arg0, 2);
                    /* checking for: attrs */
                    if (ATgetType((ATerm)arg_arg0_arg2) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg2) == PERR_afun13) {
                      ATerm arg_arg0_arg2_arg0 = ATgetArgument(arg_arg0_arg2, 0);
                      if (ATgetType((ATerm)arg_arg0_arg2_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg2_arg0) == ATfalse) {
                        ATerm arg_arg0_arg2_arg0_head;
                        ATermList arg_arg0_arg2_arg0_list = (ATermList)arg_arg0_arg2_arg0;
                        arg_arg0_arg2_arg0_head = ATgetFirst(arg_arg0_arg2_arg0_list);
                        arg_arg0_arg2_arg0_list = ATgetNext(arg_arg0_arg2_arg0_list);
                        /* checking for: term */
                        if (ATgetType((ATerm)arg_arg0_arg2_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg2_arg0_head) == PERR_afun14) {
                          ATerm arg_arg0_arg2_arg0_head_arg0 = ATgetArgument(arg_arg0_arg2_arg0_head, 0);
                          /* checking for: cons */
                          if (ATgetType((ATerm)arg_arg0_arg2_arg0_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg2_arg0_head_arg0) == PERR_afun15) {
                            ATerm arg_arg0_arg2_arg0_head_arg0_arg0 = ATgetArgument(arg_arg0_arg2_arg0_head_arg0, 0);
                            /* checking for: tab */
                            if (ATgetType((ATerm)arg_arg0_arg2_arg0_head_arg0_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg2_arg0_head_arg0_arg0) == PERR_afun19) {
                              if (ATgetType((ATerm)arg_arg0_arg2_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg2_arg0_list) == ATtrue) {
                                ATerm arg_arg1 = ATgetArgument(arg, 1);
                                if (ATgetType((ATerm)arg_arg1) == AT_LIST && ATisEmpty((ATermList)arg_arg1) == ATfalse) {
                                  ATerm arg_arg1_head;
                                  ATermList arg_arg1_list = (ATermList)arg_arg1;
                                  arg_arg1_head = ATgetFirst(arg_arg1_list);
                                  arg_arg1_list = ATgetNext(arg_arg1_list);
                                  /* checking for: appl */
                                  if (ATgetType((ATerm)arg_arg1_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head) == PERR_afun0) {
                                    ATerm arg_arg1_head_arg0 = ATgetArgument(arg_arg1_head, 0);
                                    /* checking for: prod */
                                    if (ATgetType((ATerm)arg_arg1_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0) == PERR_afun1) {
                                      ATerm arg_arg1_head_arg0_arg0 = ATgetArgument(arg_arg1_head_arg0, 0);
                                      if (ATgetType((ATerm)arg_arg1_head_arg0_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0) == ATfalse) {
                                        ATerm arg_arg1_head_arg0_arg0_head;
                                        ATermList arg_arg1_head_arg0_arg0_list = (ATermList)arg_arg1_head_arg0_arg0;
                                        arg_arg1_head_arg0_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_list);
                                        arg_arg1_head_arg0_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_list);
                                        /* checking for: char-class */
                                        if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg0_head) == PERR_afun17) {
                                          ATerm arg_arg1_head_arg0_arg0_head_arg0 = ATgetArgument(arg_arg1_head_arg0_arg0_head, 0);
                                          if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0) == ATfalse) {
                                            ATerm arg_arg1_head_arg0_arg0_head_arg0_head;
                                            ATermList arg_arg1_head_arg0_arg0_head_arg0_list = (ATermList)arg_arg1_head_arg0_arg0_head_arg0;
                                            arg_arg1_head_arg0_arg0_head_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_head_arg0_list);
                                            arg_arg1_head_arg0_arg0_head_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_head_arg0_list);
                                            if (ATgetInt((ATermInt)arg_arg1_head_arg0_arg0_head_arg0_head) == 92) {

                                              if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0_list) == ATtrue) {
                                                if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_list) == ATfalse) {
                                                  arg_arg1_head_arg0_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_list);
                                                  arg_arg1_head_arg0_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_list);
                                                  /* checking for: char-class */
                                                  if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg0_head) == PERR_afun17) {
                                                    ATerm arg_arg1_head_arg0_arg0_head_arg0 = ATgetArgument(arg_arg1_head_arg0_arg0_head, 0);
                                                    if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0) == ATfalse) {
                                                      ATerm arg_arg1_head_arg0_arg0_head_arg0_head;
                                                      ATermList arg_arg1_head_arg0_arg0_head_arg0_list = (ATermList)arg_arg1_head_arg0_arg0_head_arg0;
                                                      arg_arg1_head_arg0_arg0_head_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                      arg_arg1_head_arg0_arg0_head_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                      if (ATgetInt((ATermInt)arg_arg1_head_arg0_arg0_head_arg0_head) == 116) {

                                                        if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0_list) == ATtrue) {
                                                          if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_list) == ATtrue) {
                                                            ATerm arg_arg1_head_arg0_arg1 = ATgetArgument(arg_arg1_head_arg0, 1);
                                                            /* checking for: lit */
                                                            if (ATgetType((ATerm)arg_arg1_head_arg0_arg1) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg1) == PERR_afun9) {
                                                              ATerm arg_arg1_head_arg0_arg1_arg0 = ATgetArgument(arg_arg1_head_arg0_arg1, 0);
                                                              /* checking for: \t */
                                                              if (ATgetType((ATerm)arg_arg1_head_arg0_arg1_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg1_arg0) == PERR_afun18) {
                                                                ATerm arg_arg1_head_arg0_arg2 = ATgetArgument(arg_arg1_head_arg0, 2);
                                                                /* checking for: no-attrs */
                                                                if (ATgetType((ATerm)arg_arg1_head_arg0_arg2) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg2) == PERR_afun5) {
                                                                  ATerm arg_arg1_head_arg1 = ATgetArgument(arg_arg1_head, 1);
                                                                  if (ATgetType((ATerm)arg_arg1_head_arg1) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg1) == ATfalse) {
                                                                    ATerm arg_arg1_head_arg1_head;
                                                                    ATermList arg_arg1_head_arg1_list = (ATermList)arg_arg1_head_arg1;
                                                                    arg_arg1_head_arg1_head = ATgetFirst(arg_arg1_head_arg1_list);
                                                                    arg_arg1_head_arg1_list = ATgetNext(arg_arg1_head_arg1_list);
                                                                    if (ATgetInt((ATermInt)arg_arg1_head_arg1_head) == 92) {

                                                                      if (ATgetType((ATerm)arg_arg1_head_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg1_list) == ATfalse) {
                                                                        arg_arg1_head_arg1_head = ATgetFirst(arg_arg1_head_arg1_list);
                                                                        arg_arg1_head_arg1_list = ATgetNext(arg_arg1_head_arg1_list);
                                                                        if (ATgetInt((ATermInt)arg_arg1_head_arg1_head) == 116) {

                                                                          if (ATgetType((ATerm)arg_arg1_head_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg1_list) == ATtrue) {
                                                                            if (ATgetType((ATerm)arg_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_list) == ATtrue) {
                                                                              return ATtrue;
                                                                            }
                                                                          }
                                                                        }
                                                                      }
                                                                    }
                                                                  }
                                                                }
                                                              }
                                                            }
                                                          }
                                                        }
                                                      }
                                                    }
                                                  }
                                                }
                                              }
                                            }
                                          }
                                        }
                                      }
                                    }
                                  }
                                }
                              }
                            }
                          }
                        }
                      }
                    }
                  }
                }
              }
            }
          }
        }
      }
    }
  }
  return ATfalse;
}

/**
 * Assert whether a PERR_LexStrChar is a quote by checking against the following ATerm pattern: appl(prod([lit("\\\"")],lex(sort("StrChar")),attrs([term(cons("quote"))])),[appl(prod([char-class([92]),char-class([34])],lit("\\\""),no-attrs),[92,34])]). May not be used to assert correctness of the PERR_LexStrChar
 * \param[in] arg input PERR_LexStrChar
 * \return ATtrue if #arg corresponds to the signature of a quote, or ATfalse otherwise
 */
inline ATbool PERR_isLexStrCharQuote(PERR_LexStrChar arg){
  /* checking for: appl */
  if (ATgetType((ATerm)arg) == AT_APPL && ATgetAFun((ATermAppl)arg) == PERR_afun0) {
    ATerm arg_arg0 = ATgetArgument(arg, 0);
    /* checking for: prod */
    if (ATgetType((ATerm)arg_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0) == PERR_afun1) {
      ATerm arg_arg0_arg0 = ATgetArgument(arg_arg0, 0);
      if (ATgetType((ATerm)arg_arg0_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0) == ATfalse) {
        ATerm arg_arg0_arg0_head;
        ATermList arg_arg0_arg0_list = (ATermList)arg_arg0_arg0;
        arg_arg0_arg0_head = ATgetFirst(arg_arg0_arg0_list);
        arg_arg0_arg0_list = ATgetNext(arg_arg0_arg0_list);
        /* checking for: lit */
        if (ATgetType((ATerm)arg_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head) == PERR_afun9) {
          ATerm arg_arg0_arg0_head_arg0 = ATgetArgument(arg_arg0_arg0_head, 0);
          /* checking for: \" */
          if (ATgetType((ATerm)arg_arg0_arg0_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0) == PERR_afun20) {
            if (ATgetType((ATerm)arg_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0_list) == ATtrue) {
              ATerm arg_arg0_arg1 = ATgetArgument(arg_arg0, 1);
              /* checking for: lex */
              if (ATgetType((ATerm)arg_arg0_arg1) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg1) == PERR_afun6) {
                ATerm arg_arg0_arg1_arg0 = ATgetArgument(arg_arg0_arg1, 0);
                /* checking for: sort */
                if (ATgetType((ATerm)arg_arg0_arg1_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg1_arg0) == PERR_afun11) {
                  ATerm arg_arg0_arg1_arg0_arg0 = ATgetArgument(arg_arg0_arg1_arg0, 0);
                  /* checking for: StrChar */
                  if (ATgetType((ATerm)arg_arg0_arg1_arg0_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg1_arg0_arg0) == PERR_afun12) {
                    ATerm arg_arg0_arg2 = ATgetArgument(arg_arg0, 2);
                    /* checking for: attrs */
                    if (ATgetType((ATerm)arg_arg0_arg2) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg2) == PERR_afun13) {
                      ATerm arg_arg0_arg2_arg0 = ATgetArgument(arg_arg0_arg2, 0);
                      if (ATgetType((ATerm)arg_arg0_arg2_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg2_arg0) == ATfalse) {
                        ATerm arg_arg0_arg2_arg0_head;
                        ATermList arg_arg0_arg2_arg0_list = (ATermList)arg_arg0_arg2_arg0;
                        arg_arg0_arg2_arg0_head = ATgetFirst(arg_arg0_arg2_arg0_list);
                        arg_arg0_arg2_arg0_list = ATgetNext(arg_arg0_arg2_arg0_list);
                        /* checking for: term */
                        if (ATgetType((ATerm)arg_arg0_arg2_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg2_arg0_head) == PERR_afun14) {
                          ATerm arg_arg0_arg2_arg0_head_arg0 = ATgetArgument(arg_arg0_arg2_arg0_head, 0);
                          /* checking for: cons */
                          if (ATgetType((ATerm)arg_arg0_arg2_arg0_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg2_arg0_head_arg0) == PERR_afun15) {
                            ATerm arg_arg0_arg2_arg0_head_arg0_arg0 = ATgetArgument(arg_arg0_arg2_arg0_head_arg0, 0);
                            /* checking for: quote */
                            if (ATgetType((ATerm)arg_arg0_arg2_arg0_head_arg0_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg2_arg0_head_arg0_arg0) == PERR_afun21) {
                              if (ATgetType((ATerm)arg_arg0_arg2_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg2_arg0_list) == ATtrue) {
                                ATerm arg_arg1 = ATgetArgument(arg, 1);
                                if (ATgetType((ATerm)arg_arg1) == AT_LIST && ATisEmpty((ATermList)arg_arg1) == ATfalse) {
                                  ATerm arg_arg1_head;
                                  ATermList arg_arg1_list = (ATermList)arg_arg1;
                                  arg_arg1_head = ATgetFirst(arg_arg1_list);
                                  arg_arg1_list = ATgetNext(arg_arg1_list);
                                  /* checking for: appl */
                                  if (ATgetType((ATerm)arg_arg1_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head) == PERR_afun0) {
                                    ATerm arg_arg1_head_arg0 = ATgetArgument(arg_arg1_head, 0);
                                    /* checking for: prod */
                                    if (ATgetType((ATerm)arg_arg1_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0) == PERR_afun1) {
                                      ATerm arg_arg1_head_arg0_arg0 = ATgetArgument(arg_arg1_head_arg0, 0);
                                      if (ATgetType((ATerm)arg_arg1_head_arg0_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0) == ATfalse) {
                                        ATerm arg_arg1_head_arg0_arg0_head;
                                        ATermList arg_arg1_head_arg0_arg0_list = (ATermList)arg_arg1_head_arg0_arg0;
                                        arg_arg1_head_arg0_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_list);
                                        arg_arg1_head_arg0_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_list);
                                        /* checking for: char-class */
                                        if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg0_head) == PERR_afun17) {
                                          ATerm arg_arg1_head_arg0_arg0_head_arg0 = ATgetArgument(arg_arg1_head_arg0_arg0_head, 0);
                                          if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0) == ATfalse) {
                                            ATerm arg_arg1_head_arg0_arg0_head_arg0_head;
                                            ATermList arg_arg1_head_arg0_arg0_head_arg0_list = (ATermList)arg_arg1_head_arg0_arg0_head_arg0;
                                            arg_arg1_head_arg0_arg0_head_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_head_arg0_list);
                                            arg_arg1_head_arg0_arg0_head_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_head_arg0_list);
                                            if (ATgetInt((ATermInt)arg_arg1_head_arg0_arg0_head_arg0_head) == 92) {

                                              if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0_list) == ATtrue) {
                                                if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_list) == ATfalse) {
                                                  arg_arg1_head_arg0_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_list);
                                                  arg_arg1_head_arg0_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_list);
                                                  /* checking for: char-class */
                                                  if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg0_head) == PERR_afun17) {
                                                    ATerm arg_arg1_head_arg0_arg0_head_arg0 = ATgetArgument(arg_arg1_head_arg0_arg0_head, 0);
                                                    if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0) == ATfalse) {
                                                      ATerm arg_arg1_head_arg0_arg0_head_arg0_head;
                                                      ATermList arg_arg1_head_arg0_arg0_head_arg0_list = (ATermList)arg_arg1_head_arg0_arg0_head_arg0;
                                                      arg_arg1_head_arg0_arg0_head_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                      arg_arg1_head_arg0_arg0_head_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                      if (ATgetInt((ATermInt)arg_arg1_head_arg0_arg0_head_arg0_head) == 34) {

                                                        if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0_list) == ATtrue) {
                                                          if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_list) == ATtrue) {
                                                            ATerm arg_arg1_head_arg0_arg1 = ATgetArgument(arg_arg1_head_arg0, 1);
                                                            /* checking for: lit */
                                                            if (ATgetType((ATerm)arg_arg1_head_arg0_arg1) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg1) == PERR_afun9) {
                                                              ATerm arg_arg1_head_arg0_arg1_arg0 = ATgetArgument(arg_arg1_head_arg0_arg1, 0);
                                                              /* checking for: \" */
                                                              if (ATgetType((ATerm)arg_arg1_head_arg0_arg1_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg1_arg0) == PERR_afun20) {
                                                                ATerm arg_arg1_head_arg0_arg2 = ATgetArgument(arg_arg1_head_arg0, 2);
                                                                /* checking for: no-attrs */
                                                                if (ATgetType((ATerm)arg_arg1_head_arg0_arg2) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg2) == PERR_afun5) {
                                                                  ATerm arg_arg1_head_arg1 = ATgetArgument(arg_arg1_head, 1);
                                                                  if (ATgetType((ATerm)arg_arg1_head_arg1) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg1) == ATfalse) {
                                                                    ATerm arg_arg1_head_arg1_head;
                                                                    ATermList arg_arg1_head_arg1_list = (ATermList)arg_arg1_head_arg1;
                                                                    arg_arg1_head_arg1_head = ATgetFirst(arg_arg1_head_arg1_list);
                                                                    arg_arg1_head_arg1_list = ATgetNext(arg_arg1_head_arg1_list);
                                                                    if (ATgetInt((ATermInt)arg_arg1_head_arg1_head) == 92) {

                                                                      if (ATgetType((ATerm)arg_arg1_head_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg1_list) == ATfalse) {
                                                                        arg_arg1_head_arg1_head = ATgetFirst(arg_arg1_head_arg1_list);
                                                                        arg_arg1_head_arg1_list = ATgetNext(arg_arg1_head_arg1_list);
                                                                        if (ATgetInt((ATermInt)arg_arg1_head_arg1_head) == 34) {

                                                                          if (ATgetType((ATerm)arg_arg1_head_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg1_list) == ATtrue) {
                                                                            if (ATgetType((ATerm)arg_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_list) == ATtrue) {
                                                                              return ATtrue;
                                                                            }
                                                                          }
                                                                        }
                                                                      }
                                                                    }
                                                                  }
                                                                }
                                                              }
                                                            }
                                                          }
                                                        }
                                                      }
                                                    }
                                                  }
                                                }
                                              }
                                            }
                                          }
                                        }
                                      }
                                    }
                                  }
                                }
                              }
                            }
                          }
                        }
                      }
                    }
                  }
                }
              }
            }
          }
        }
      }
    }
  }
  return ATfalse;
}

/**
 * Assert whether a PERR_LexStrChar is a backslash by checking against the following ATerm pattern: appl(prod([lit("\\\\")],lex(sort("StrChar")),attrs([term(cons("backslash"))])),[appl(prod([char-class([92]),char-class([92])],lit("\\\\"),no-attrs),[92,92])]). May not be used to assert correctness of the PERR_LexStrChar
 * \param[in] arg input PERR_LexStrChar
 * \return ATtrue if #arg corresponds to the signature of a backslash, or ATfalse otherwise
 */
inline ATbool PERR_isLexStrCharBackslash(PERR_LexStrChar arg){
  /* checking for: appl */
  if (ATgetType((ATerm)arg) == AT_APPL && ATgetAFun((ATermAppl)arg) == PERR_afun0) {
    ATerm arg_arg0 = ATgetArgument(arg, 0);
    /* checking for: prod */
    if (ATgetType((ATerm)arg_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0) == PERR_afun1) {
      ATerm arg_arg0_arg0 = ATgetArgument(arg_arg0, 0);
      if (ATgetType((ATerm)arg_arg0_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0) == ATfalse) {
        ATerm arg_arg0_arg0_head;
        ATermList arg_arg0_arg0_list = (ATermList)arg_arg0_arg0;
        arg_arg0_arg0_head = ATgetFirst(arg_arg0_arg0_list);
        arg_arg0_arg0_list = ATgetNext(arg_arg0_arg0_list);
        /* checking for: lit */
        if (ATgetType((ATerm)arg_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head) == PERR_afun9) {
          ATerm arg_arg0_arg0_head_arg0 = ATgetArgument(arg_arg0_arg0_head, 0);
          /* checking for: \\ */
          if (ATgetType((ATerm)arg_arg0_arg0_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0) == PERR_afun22) {
            if (ATgetType((ATerm)arg_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0_list) == ATtrue) {
              ATerm arg_arg0_arg1 = ATgetArgument(arg_arg0, 1);
              /* checking for: lex */
              if (ATgetType((ATerm)arg_arg0_arg1) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg1) == PERR_afun6) {
                ATerm arg_arg0_arg1_arg0 = ATgetArgument(arg_arg0_arg1, 0);
                /* checking for: sort */
                if (ATgetType((ATerm)arg_arg0_arg1_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg1_arg0) == PERR_afun11) {
                  ATerm arg_arg0_arg1_arg0_arg0 = ATgetArgument(arg_arg0_arg1_arg0, 0);
                  /* checking for: StrChar */
                  if (ATgetType((ATerm)arg_arg0_arg1_arg0_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg1_arg0_arg0) == PERR_afun12) {
                    ATerm arg_arg0_arg2 = ATgetArgument(arg_arg0, 2);
                    /* checking for: attrs */
                    if (ATgetType((ATerm)arg_arg0_arg2) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg2) == PERR_afun13) {
                      ATerm arg_arg0_arg2_arg0 = ATgetArgument(arg_arg0_arg2, 0);
                      if (ATgetType((ATerm)arg_arg0_arg2_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg2_arg0) == ATfalse) {
                        ATerm arg_arg0_arg2_arg0_head;
                        ATermList arg_arg0_arg2_arg0_list = (ATermList)arg_arg0_arg2_arg0;
                        arg_arg0_arg2_arg0_head = ATgetFirst(arg_arg0_arg2_arg0_list);
                        arg_arg0_arg2_arg0_list = ATgetNext(arg_arg0_arg2_arg0_list);
                        /* checking for: term */
                        if (ATgetType((ATerm)arg_arg0_arg2_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg2_arg0_head) == PERR_afun14) {
                          ATerm arg_arg0_arg2_arg0_head_arg0 = ATgetArgument(arg_arg0_arg2_arg0_head, 0);
                          /* checking for: cons */
                          if (ATgetType((ATerm)arg_arg0_arg2_arg0_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg2_arg0_head_arg0) == PERR_afun15) {
                            ATerm arg_arg0_arg2_arg0_head_arg0_arg0 = ATgetArgument(arg_arg0_arg2_arg0_head_arg0, 0);
                            /* checking for: backslash */
                            if (ATgetType((ATerm)arg_arg0_arg2_arg0_head_arg0_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg2_arg0_head_arg0_arg0) == PERR_afun23) {
                              if (ATgetType((ATerm)arg_arg0_arg2_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg2_arg0_list) == ATtrue) {
                                ATerm arg_arg1 = ATgetArgument(arg, 1);
                                if (ATgetType((ATerm)arg_arg1) == AT_LIST && ATisEmpty((ATermList)arg_arg1) == ATfalse) {
                                  ATerm arg_arg1_head;
                                  ATermList arg_arg1_list = (ATermList)arg_arg1;
                                  arg_arg1_head = ATgetFirst(arg_arg1_list);
                                  arg_arg1_list = ATgetNext(arg_arg1_list);
                                  /* checking for: appl */
                                  if (ATgetType((ATerm)arg_arg1_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head) == PERR_afun0) {
                                    ATerm arg_arg1_head_arg0 = ATgetArgument(arg_arg1_head, 0);
                                    /* checking for: prod */
                                    if (ATgetType((ATerm)arg_arg1_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0) == PERR_afun1) {
                                      ATerm arg_arg1_head_arg0_arg0 = ATgetArgument(arg_arg1_head_arg0, 0);
                                      if (ATgetType((ATerm)arg_arg1_head_arg0_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0) == ATfalse) {
                                        ATerm arg_arg1_head_arg0_arg0_head;
                                        ATermList arg_arg1_head_arg0_arg0_list = (ATermList)arg_arg1_head_arg0_arg0;
                                        arg_arg1_head_arg0_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_list);
                                        arg_arg1_head_arg0_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_list);
                                        /* checking for: char-class */
                                        if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg0_head) == PERR_afun17) {
                                          ATerm arg_arg1_head_arg0_arg0_head_arg0 = ATgetArgument(arg_arg1_head_arg0_arg0_head, 0);
                                          if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0) == ATfalse) {
                                            ATerm arg_arg1_head_arg0_arg0_head_arg0_head;
                                            ATermList arg_arg1_head_arg0_arg0_head_arg0_list = (ATermList)arg_arg1_head_arg0_arg0_head_arg0;
                                            arg_arg1_head_arg0_arg0_head_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_head_arg0_list);
                                            arg_arg1_head_arg0_arg0_head_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_head_arg0_list);
                                            if (ATgetInt((ATermInt)arg_arg1_head_arg0_arg0_head_arg0_head) == 92) {

                                              if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0_list) == ATtrue) {
                                                if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_list) == ATfalse) {
                                                  arg_arg1_head_arg0_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_list);
                                                  arg_arg1_head_arg0_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_list);
                                                  /* checking for: char-class */
                                                  if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg0_head) == PERR_afun17) {
                                                    ATerm arg_arg1_head_arg0_arg0_head_arg0 = ATgetArgument(arg_arg1_head_arg0_arg0_head, 0);
                                                    if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0) == ATfalse) {
                                                      ATerm arg_arg1_head_arg0_arg0_head_arg0_head;
                                                      ATermList arg_arg1_head_arg0_arg0_head_arg0_list = (ATermList)arg_arg1_head_arg0_arg0_head_arg0;
                                                      arg_arg1_head_arg0_arg0_head_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                      arg_arg1_head_arg0_arg0_head_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                      if (ATgetInt((ATermInt)arg_arg1_head_arg0_arg0_head_arg0_head) == 92) {

                                                        if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0_list) == ATtrue) {
                                                          if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_list) == ATtrue) {
                                                            ATerm arg_arg1_head_arg0_arg1 = ATgetArgument(arg_arg1_head_arg0, 1);
                                                            /* checking for: lit */
                                                            if (ATgetType((ATerm)arg_arg1_head_arg0_arg1) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg1) == PERR_afun9) {
                                                              ATerm arg_arg1_head_arg0_arg1_arg0 = ATgetArgument(arg_arg1_head_arg0_arg1, 0);
                                                              /* checking for: \\ */
                                                              if (ATgetType((ATerm)arg_arg1_head_arg0_arg1_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg1_arg0) == PERR_afun22) {
                                                                ATerm arg_arg1_head_arg0_arg2 = ATgetArgument(arg_arg1_head_arg0, 2);
                                                                /* checking for: no-attrs */
                                                                if (ATgetType((ATerm)arg_arg1_head_arg0_arg2) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg2) == PERR_afun5) {
                                                                  ATerm arg_arg1_head_arg1 = ATgetArgument(arg_arg1_head, 1);
                                                                  if (ATgetType((ATerm)arg_arg1_head_arg1) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg1) == ATfalse) {
                                                                    ATerm arg_arg1_head_arg1_head;
                                                                    ATermList arg_arg1_head_arg1_list = (ATermList)arg_arg1_head_arg1;
                                                                    arg_arg1_head_arg1_head = ATgetFirst(arg_arg1_head_arg1_list);
                                                                    arg_arg1_head_arg1_list = ATgetNext(arg_arg1_head_arg1_list);
                                                                    if (ATgetInt((ATermInt)arg_arg1_head_arg1_head) == 92) {

                                                                      if (ATgetType((ATerm)arg_arg1_head_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg1_list) == ATfalse) {
                                                                        arg_arg1_head_arg1_head = ATgetFirst(arg_arg1_head_arg1_list);
                                                                        arg_arg1_head_arg1_list = ATgetNext(arg_arg1_head_arg1_list);
                                                                        if (ATgetInt((ATermInt)arg_arg1_head_arg1_head) == 92) {

                                                                          if (ATgetType((ATerm)arg_arg1_head_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg1_list) == ATtrue) {
                                                                            if (ATgetType((ATerm)arg_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_list) == ATtrue) {
                                                                              return ATtrue;
                                                                            }
                                                                          }
                                                                        }
                                                                      }
                                                                    }
                                                                  }
                                                                }
                                                              }
                                                            }
                                                          }
                                                        }
                                                      }
                                                    }
                                                  }
                                                }
                                              }
                                            }
                                          }
                                        }
                                      }
                                    }
                                  }
                                }
                              }
                            }
                          }
                        }
                      }
                    }
                  }
                }
              }
            }
          }
        }
      }
    }
  }
  return ATfalse;
}

/**
 * Assert whether a PERR_LexStrChar is a decimal by checking against the following ATerm pattern: appl(prod([lit("\\"),char-class([range(48,57)]),char-class([range(48,57)]),char-class([range(48,57)])],lex(sort("StrChar")),attrs([term(cons("decimal"))])),[appl(prod([char-class([92])],lit("\\"),no-attrs),[92]),<"a"(char)>,<"b"(char)>,<"c"(char)>]). May not be used to assert correctness of the PERR_LexStrChar
 * \param[in] arg input PERR_LexStrChar
 * \return ATtrue if #arg corresponds to the signature of a decimal, or ATfalse otherwise
 */
inline ATbool PERR_isLexStrCharDecimal(PERR_LexStrChar arg){
  /* checking for: appl */
  if (ATgetType((ATerm)arg) == AT_APPL && ATgetAFun((ATermAppl)arg) == PERR_afun0) {
    ATerm arg_arg0 = ATgetArgument(arg, 0);
    /* checking for: prod */
    if (ATgetType((ATerm)arg_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0) == PERR_afun1) {
      ATerm arg_arg0_arg0 = ATgetArgument(arg_arg0, 0);
      if (ATgetType((ATerm)arg_arg0_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0) == ATfalse) {
        ATerm arg_arg0_arg0_head;
        ATermList arg_arg0_arg0_list = (ATermList)arg_arg0_arg0;
        arg_arg0_arg0_head = ATgetFirst(arg_arg0_arg0_list);
        arg_arg0_arg0_list = ATgetNext(arg_arg0_arg0_list);
        /* checking for: lit */
        if (ATgetType((ATerm)arg_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head) == PERR_afun9) {
          ATerm arg_arg0_arg0_head_arg0 = ATgetArgument(arg_arg0_arg0_head, 0);
          /* checking for: \ */
          if (ATgetType((ATerm)arg_arg0_arg0_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0) == PERR_afun25) {
            if (ATgetType((ATerm)arg_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0_list) == ATfalse) {
              arg_arg0_arg0_head = ATgetFirst(arg_arg0_arg0_list);
              arg_arg0_arg0_list = ATgetNext(arg_arg0_arg0_list);
              /* checking for: char-class */
              if (ATgetType((ATerm)arg_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head) == PERR_afun17) {
                ATerm arg_arg0_arg0_head_arg0 = ATgetArgument(arg_arg0_arg0_head, 0);
                if (ATgetType((ATerm)arg_arg0_arg0_head_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0_head_arg0) == ATfalse) {
                  ATerm arg_arg0_arg0_head_arg0_head;
                  ATermList arg_arg0_arg0_head_arg0_list = (ATermList)arg_arg0_arg0_head_arg0;
                  arg_arg0_arg0_head_arg0_head = ATgetFirst(arg_arg0_arg0_head_arg0_list);
                  arg_arg0_arg0_head_arg0_list = ATgetNext(arg_arg0_arg0_head_arg0_list);
                  /* checking for: range */
                  if (ATgetType((ATerm)arg_arg0_arg0_head_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0_head) == PERR_afun24) {
                    ATerm arg_arg0_arg0_head_arg0_head_arg0 = ATgetArgument(arg_arg0_arg0_head_arg0_head, 0);
                    if (ATgetInt((ATermInt)arg_arg0_arg0_head_arg0_head_arg0) == 48) {

                      ATerm arg_arg0_arg0_head_arg0_head_arg1 = ATgetArgument(arg_arg0_arg0_head_arg0_head, 1);
                      if (ATgetInt((ATermInt)arg_arg0_arg0_head_arg0_head_arg1) == 57) {

                        if (ATgetType((ATerm)arg_arg0_arg0_head_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0_head_arg0_list) == ATtrue) {
                          if (ATgetType((ATerm)arg_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0_list) == ATfalse) {
                            arg_arg0_arg0_head = ATgetFirst(arg_arg0_arg0_list);
                            arg_arg0_arg0_list = ATgetNext(arg_arg0_arg0_list);
                            /* checking for: char-class */
                            if (ATgetType((ATerm)arg_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head) == PERR_afun17) {
                              ATerm arg_arg0_arg0_head_arg0 = ATgetArgument(arg_arg0_arg0_head, 0);
                              if (ATgetType((ATerm)arg_arg0_arg0_head_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0_head_arg0) == ATfalse) {
                                ATerm arg_arg0_arg0_head_arg0_head;
                                ATermList arg_arg0_arg0_head_arg0_list = (ATermList)arg_arg0_arg0_head_arg0;
                                arg_arg0_arg0_head_arg0_head = ATgetFirst(arg_arg0_arg0_head_arg0_list);
                                arg_arg0_arg0_head_arg0_list = ATgetNext(arg_arg0_arg0_head_arg0_list);
                                /* checking for: range */
                                if (ATgetType((ATerm)arg_arg0_arg0_head_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0_head) == PERR_afun24) {
                                  ATerm arg_arg0_arg0_head_arg0_head_arg0 = ATgetArgument(arg_arg0_arg0_head_arg0_head, 0);
                                  if (ATgetInt((ATermInt)arg_arg0_arg0_head_arg0_head_arg0) == 48) {

                                    ATerm arg_arg0_arg0_head_arg0_head_arg1 = ATgetArgument(arg_arg0_arg0_head_arg0_head, 1);
                                    if (ATgetInt((ATermInt)arg_arg0_arg0_head_arg0_head_arg1) == 57) {

                                      if (ATgetType((ATerm)arg_arg0_arg0_head_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0_head_arg0_list) == ATtrue) {
                                        if (ATgetType((ATerm)arg_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0_list) == ATfalse) {
                                          arg_arg0_arg0_head = ATgetFirst(arg_arg0_arg0_list);
                                          arg_arg0_arg0_list = ATgetNext(arg_arg0_arg0_list);
                                          /* checking for: char-class */
                                          if (ATgetType((ATerm)arg_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head) == PERR_afun17) {
                                            ATerm arg_arg0_arg0_head_arg0 = ATgetArgument(arg_arg0_arg0_head, 0);
                                            if (ATgetType((ATerm)arg_arg0_arg0_head_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0_head_arg0) == ATfalse) {
                                              ATerm arg_arg0_arg0_head_arg0_head;
                                              ATermList arg_arg0_arg0_head_arg0_list = (ATermList)arg_arg0_arg0_head_arg0;
                                              arg_arg0_arg0_head_arg0_head = ATgetFirst(arg_arg0_arg0_head_arg0_list);
                                              arg_arg0_arg0_head_arg0_list = ATgetNext(arg_arg0_arg0_head_arg0_list);
                                              /* checking for: range */
                                              if (ATgetType((ATerm)arg_arg0_arg0_head_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0_head) == PERR_afun24) {
                                                ATerm arg_arg0_arg0_head_arg0_head_arg0 = ATgetArgument(arg_arg0_arg0_head_arg0_head, 0);
                                                if (ATgetInt((ATermInt)arg_arg0_arg0_head_arg0_head_arg0) == 48) {

                                                  ATerm arg_arg0_arg0_head_arg0_head_arg1 = ATgetArgument(arg_arg0_arg0_head_arg0_head, 1);
                                                  if (ATgetInt((ATermInt)arg_arg0_arg0_head_arg0_head_arg1) == 57) {

                                                    if (ATgetType((ATerm)arg_arg0_arg0_head_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0_head_arg0_list) == ATtrue) {
                                                      if (ATgetType((ATerm)arg_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0_list) == ATtrue) {
                                                        ATerm arg_arg0_arg1 = ATgetArgument(arg_arg0, 1);
                                                        /* checking for: lex */
                                                        if (ATgetType((ATerm)arg_arg0_arg1) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg1) == PERR_afun6) {
                                                          ATerm arg_arg0_arg1_arg0 = ATgetArgument(arg_arg0_arg1, 0);
                                                          /* checking for: sort */
                                                          if (ATgetType((ATerm)arg_arg0_arg1_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg1_arg0) == PERR_afun11) {
                                                            ATerm arg_arg0_arg1_arg0_arg0 = ATgetArgument(arg_arg0_arg1_arg0, 0);
                                                            /* checking for: StrChar */
                                                            if (ATgetType((ATerm)arg_arg0_arg1_arg0_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg1_arg0_arg0) == PERR_afun12) {
                                                              ATerm arg_arg0_arg2 = ATgetArgument(arg_arg0, 2);
                                                              /* checking for: attrs */
                                                              if (ATgetType((ATerm)arg_arg0_arg2) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg2) == PERR_afun13) {
                                                                ATerm arg_arg0_arg2_arg0 = ATgetArgument(arg_arg0_arg2, 0);
                                                                if (ATgetType((ATerm)arg_arg0_arg2_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg2_arg0) == ATfalse) {
                                                                  ATerm arg_arg0_arg2_arg0_head;
                                                                  ATermList arg_arg0_arg2_arg0_list = (ATermList)arg_arg0_arg2_arg0;
                                                                  arg_arg0_arg2_arg0_head = ATgetFirst(arg_arg0_arg2_arg0_list);
                                                                  arg_arg0_arg2_arg0_list = ATgetNext(arg_arg0_arg2_arg0_list);
                                                                  /* checking for: term */
                                                                  if (ATgetType((ATerm)arg_arg0_arg2_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg2_arg0_head) == PERR_afun14) {
                                                                    ATerm arg_arg0_arg2_arg0_head_arg0 = ATgetArgument(arg_arg0_arg2_arg0_head, 0);
                                                                    /* checking for: cons */
                                                                    if (ATgetType((ATerm)arg_arg0_arg2_arg0_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg2_arg0_head_arg0) == PERR_afun15) {
                                                                      ATerm arg_arg0_arg2_arg0_head_arg0_arg0 = ATgetArgument(arg_arg0_arg2_arg0_head_arg0, 0);
                                                                      /* checking for: decimal */
                                                                      if (ATgetType((ATerm)arg_arg0_arg2_arg0_head_arg0_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg2_arg0_head_arg0_arg0) == PERR_afun26) {
                                                                        if (ATgetType((ATerm)arg_arg0_arg2_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg2_arg0_list) == ATtrue) {
                                                                          ATerm arg_arg1 = ATgetArgument(arg, 1);
                                                                          if (ATgetType((ATerm)arg_arg1) == AT_LIST && ATisEmpty((ATermList)arg_arg1) == ATfalse) {
                                                                            ATerm arg_arg1_head;
                                                                            ATermList arg_arg1_list = (ATermList)arg_arg1;
                                                                            arg_arg1_head = ATgetFirst(arg_arg1_list);
                                                                            arg_arg1_list = ATgetNext(arg_arg1_list);
                                                                            /* checking for: appl */
                                                                            if (ATgetType((ATerm)arg_arg1_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head) == PERR_afun0) {
                                                                              ATerm arg_arg1_head_arg0 = ATgetArgument(arg_arg1_head, 0);
                                                                              /* checking for: prod */
                                                                              if (ATgetType((ATerm)arg_arg1_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0) == PERR_afun1) {
                                                                                ATerm arg_arg1_head_arg0_arg0 = ATgetArgument(arg_arg1_head_arg0, 0);
                                                                                if (ATgetType((ATerm)arg_arg1_head_arg0_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0) == ATfalse) {
                                                                                  ATerm arg_arg1_head_arg0_arg0_head;
                                                                                  ATermList arg_arg1_head_arg0_arg0_list = (ATermList)arg_arg1_head_arg0_arg0;
                                                                                  arg_arg1_head_arg0_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_list);
                                                                                  arg_arg1_head_arg0_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_list);
                                                                                  /* checking for: char-class */
                                                                                  if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg0_head) == PERR_afun17) {
                                                                                    ATerm arg_arg1_head_arg0_arg0_head_arg0 = ATgetArgument(arg_arg1_head_arg0_arg0_head, 0);
                                                                                    if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0) == ATfalse) {
                                                                                      ATerm arg_arg1_head_arg0_arg0_head_arg0_head;
                                                                                      ATermList arg_arg1_head_arg0_arg0_head_arg0_list = (ATermList)arg_arg1_head_arg0_arg0_head_arg0;
                                                                                      arg_arg1_head_arg0_arg0_head_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                      arg_arg1_head_arg0_arg0_head_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                      if (ATgetInt((ATermInt)arg_arg1_head_arg0_arg0_head_arg0_head) == 92) {

                                                                                        if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0_list) == ATtrue) {
                                                                                          if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_list) == ATtrue) {
                                                                                            ATerm arg_arg1_head_arg0_arg1 = ATgetArgument(arg_arg1_head_arg0, 1);
                                                                                            /* checking for: lit */
                                                                                            if (ATgetType((ATerm)arg_arg1_head_arg0_arg1) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg1) == PERR_afun9) {
                                                                                              ATerm arg_arg1_head_arg0_arg1_arg0 = ATgetArgument(arg_arg1_head_arg0_arg1, 0);
                                                                                              /* checking for: \ */
                                                                                              if (ATgetType((ATerm)arg_arg1_head_arg0_arg1_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg1_arg0) == PERR_afun25) {
                                                                                                ATerm arg_arg1_head_arg0_arg2 = ATgetArgument(arg_arg1_head_arg0, 2);
                                                                                                /* checking for: no-attrs */
                                                                                                if (ATgetType((ATerm)arg_arg1_head_arg0_arg2) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg2) == PERR_afun5) {
                                                                                                  ATerm arg_arg1_head_arg1 = ATgetArgument(arg_arg1_head, 1);
                                                                                                  if (ATgetType((ATerm)arg_arg1_head_arg1) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg1) == ATfalse) {
                                                                                                    ATerm arg_arg1_head_arg1_head;
                                                                                                    ATermList arg_arg1_head_arg1_list = (ATermList)arg_arg1_head_arg1;
                                                                                                    arg_arg1_head_arg1_head = ATgetFirst(arg_arg1_head_arg1_list);
                                                                                                    arg_arg1_head_arg1_list = ATgetNext(arg_arg1_head_arg1_list);
                                                                                                    if (ATgetInt((ATermInt)arg_arg1_head_arg1_head) == 92) {

                                                                                                      if (ATgetType((ATerm)arg_arg1_head_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg1_list) == ATtrue) {
                                                                                                        if (ATgetType((ATerm)arg_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_list) == ATfalse) {
                                                                                                          arg_arg1_head = ATgetFirst(arg_arg1_list);
                                                                                                          arg_arg1_list = ATgetNext(arg_arg1_list);
                                                                                                          if (ATgetType((ATerm)arg_arg1_head) == AT_INT) {
                                                                                                            if (ATgetType((ATerm)arg_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_list) == ATfalse) {
                                                                                                              arg_arg1_head = ATgetFirst(arg_arg1_list);
                                                                                                              arg_arg1_list = ATgetNext(arg_arg1_list);
                                                                                                              if (ATgetType((ATerm)arg_arg1_head) == AT_INT) {
                                                                                                                if (ATgetType((ATerm)arg_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_list) == ATfalse) {
                                                                                                                  arg_arg1_head = ATgetFirst(arg_arg1_list);
                                                                                                                  arg_arg1_list = ATgetNext(arg_arg1_list);
                                                                                                                  if (ATgetType((ATerm)arg_arg1_head) == AT_INT) {
                                                                                                                    if (ATgetType((ATerm)arg_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_list) == ATtrue) {
                                                                                                                      return ATtrue;
                                                                                                                    }
                                                                                                                  }
                                                                                                                }
                                                                                                              }
                                                                                                            }
                                                                                                          }
                                                                                                        }
                                                                                                      }
                                                                                                    }
                                                                                                  }
                                                                                                }
                                                                                              }
                                                                                            }
                                                                                          }
                                                                                        }
                                                                                      }
                                                                                    }
                                                                                  }
                                                                                }
                                                                              }
                                                                            }
                                                                          }
                                                                        }
                                                                      }
                                                                    }
                                                                  }
                                                                }
                                                              }
                                                            }
                                                          }
                                                        }
                                                      }
                                                    }
                                                  }
                                                }
                                              }
                                            }
                                          }
                                        }
                                      }
                                    }
                                  }
                                }
                              }
                            }
                          }
                        }
                      }
                    }
                  }
                }
              }
            }
          }
        }
      }
    }
  }
  return ATfalse;
}

/**
 * Assert whether a PERR_LexStrChar is a normal by checking against the following ATerm pattern: appl(prod([char-class([range(32,33),range(35,91),range(93,255)])],lex(sort("StrChar")),attrs([term(cons("normal"))])),[<"ch"(char)>]). May not be used to assert correctness of the PERR_LexStrChar
 * \param[in] arg input PERR_LexStrChar
 * \return ATtrue if #arg corresponds to the signature of a normal, or ATfalse otherwise
 */
inline ATbool PERR_isLexStrCharNormal(PERR_LexStrChar arg){
  /* checking for: appl */
  if (ATgetType((ATerm)arg) == AT_APPL && ATgetAFun((ATermAppl)arg) == PERR_afun0) {
    ATerm arg_arg0 = ATgetArgument(arg, 0);
    /* checking for: prod */
    if (ATgetType((ATerm)arg_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0) == PERR_afun1) {
      ATerm arg_arg0_arg0 = ATgetArgument(arg_arg0, 0);
      if (ATgetType((ATerm)arg_arg0_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0) == ATfalse) {
        ATerm arg_arg0_arg0_head;
        ATermList arg_arg0_arg0_list = (ATermList)arg_arg0_arg0;
        arg_arg0_arg0_head = ATgetFirst(arg_arg0_arg0_list);
        arg_arg0_arg0_list = ATgetNext(arg_arg0_arg0_list);
        /* checking for: char-class */
        if (ATgetType((ATerm)arg_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head) == PERR_afun17) {
          ATerm arg_arg0_arg0_head_arg0 = ATgetArgument(arg_arg0_arg0_head, 0);
          if (ATgetType((ATerm)arg_arg0_arg0_head_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0_head_arg0) == ATfalse) {
            ATerm arg_arg0_arg0_head_arg0_head;
            ATermList arg_arg0_arg0_head_arg0_list = (ATermList)arg_arg0_arg0_head_arg0;
            arg_arg0_arg0_head_arg0_head = ATgetFirst(arg_arg0_arg0_head_arg0_list);
            arg_arg0_arg0_head_arg0_list = ATgetNext(arg_arg0_arg0_head_arg0_list);
            /* checking for: range */
            if (ATgetType((ATerm)arg_arg0_arg0_head_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0_head) == PERR_afun24) {
              ATerm arg_arg0_arg0_head_arg0_head_arg0 = ATgetArgument(arg_arg0_arg0_head_arg0_head, 0);
              if (ATgetInt((ATermInt)arg_arg0_arg0_head_arg0_head_arg0) == 32) {

                ATerm arg_arg0_arg0_head_arg0_head_arg1 = ATgetArgument(arg_arg0_arg0_head_arg0_head, 1);
                if (ATgetInt((ATermInt)arg_arg0_arg0_head_arg0_head_arg1) == 33) {

                  if (ATgetType((ATerm)arg_arg0_arg0_head_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0_head_arg0_list) == ATfalse) {
                    arg_arg0_arg0_head_arg0_head = ATgetFirst(arg_arg0_arg0_head_arg0_list);
                    arg_arg0_arg0_head_arg0_list = ATgetNext(arg_arg0_arg0_head_arg0_list);
                    /* checking for: range */
                    if (ATgetType((ATerm)arg_arg0_arg0_head_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0_head) == PERR_afun24) {
                      ATerm arg_arg0_arg0_head_arg0_head_arg0 = ATgetArgument(arg_arg0_arg0_head_arg0_head, 0);
                      if (ATgetInt((ATermInt)arg_arg0_arg0_head_arg0_head_arg0) == 35) {

                        ATerm arg_arg0_arg0_head_arg0_head_arg1 = ATgetArgument(arg_arg0_arg0_head_arg0_head, 1);
                        if (ATgetInt((ATermInt)arg_arg0_arg0_head_arg0_head_arg1) == 91) {

                          if (ATgetType((ATerm)arg_arg0_arg0_head_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0_head_arg0_list) == ATfalse) {
                            arg_arg0_arg0_head_arg0_head = ATgetFirst(arg_arg0_arg0_head_arg0_list);
                            arg_arg0_arg0_head_arg0_list = ATgetNext(arg_arg0_arg0_head_arg0_list);
                            /* checking for: range */
                            if (ATgetType((ATerm)arg_arg0_arg0_head_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0_head) == PERR_afun24) {
                              ATerm arg_arg0_arg0_head_arg0_head_arg0 = ATgetArgument(arg_arg0_arg0_head_arg0_head, 0);
                              if (ATgetInt((ATermInt)arg_arg0_arg0_head_arg0_head_arg0) == 93) {

                                ATerm arg_arg0_arg0_head_arg0_head_arg1 = ATgetArgument(arg_arg0_arg0_head_arg0_head, 1);
                                if (ATgetInt((ATermInt)arg_arg0_arg0_head_arg0_head_arg1) == 255) {

                                  if (ATgetType((ATerm)arg_arg0_arg0_head_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0_head_arg0_list) == ATtrue) {
                                    if (ATgetType((ATerm)arg_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0_list) == ATtrue) {
                                      ATerm arg_arg0_arg1 = ATgetArgument(arg_arg0, 1);
                                      /* checking for: lex */
                                      if (ATgetType((ATerm)arg_arg0_arg1) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg1) == PERR_afun6) {
                                        ATerm arg_arg0_arg1_arg0 = ATgetArgument(arg_arg0_arg1, 0);
                                        /* checking for: sort */
                                        if (ATgetType((ATerm)arg_arg0_arg1_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg1_arg0) == PERR_afun11) {
                                          ATerm arg_arg0_arg1_arg0_arg0 = ATgetArgument(arg_arg0_arg1_arg0, 0);
                                          /* checking for: StrChar */
                                          if (ATgetType((ATerm)arg_arg0_arg1_arg0_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg1_arg0_arg0) == PERR_afun12) {
                                            ATerm arg_arg0_arg2 = ATgetArgument(arg_arg0, 2);
                                            /* checking for: attrs */
                                            if (ATgetType((ATerm)arg_arg0_arg2) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg2) == PERR_afun13) {
                                              ATerm arg_arg0_arg2_arg0 = ATgetArgument(arg_arg0_arg2, 0);
                                              if (ATgetType((ATerm)arg_arg0_arg2_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg2_arg0) == ATfalse) {
                                                ATerm arg_arg0_arg2_arg0_head;
                                                ATermList arg_arg0_arg2_arg0_list = (ATermList)arg_arg0_arg2_arg0;
                                                arg_arg0_arg2_arg0_head = ATgetFirst(arg_arg0_arg2_arg0_list);
                                                arg_arg0_arg2_arg0_list = ATgetNext(arg_arg0_arg2_arg0_list);
                                                /* checking for: term */
                                                if (ATgetType((ATerm)arg_arg0_arg2_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg2_arg0_head) == PERR_afun14) {
                                                  ATerm arg_arg0_arg2_arg0_head_arg0 = ATgetArgument(arg_arg0_arg2_arg0_head, 0);
                                                  /* checking for: cons */
                                                  if (ATgetType((ATerm)arg_arg0_arg2_arg0_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg2_arg0_head_arg0) == PERR_afun15) {
                                                    ATerm arg_arg0_arg2_arg0_head_arg0_arg0 = ATgetArgument(arg_arg0_arg2_arg0_head_arg0, 0);
                                                    /* checking for: normal */
                                                    if (ATgetType((ATerm)arg_arg0_arg2_arg0_head_arg0_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg2_arg0_head_arg0_arg0) == PERR_afun27) {
                                                      if (ATgetType((ATerm)arg_arg0_arg2_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg2_arg0_list) == ATtrue) {
                                                        ATerm arg_arg1 = ATgetArgument(arg, 1);
                                                        if (ATgetType((ATerm)arg_arg1) == AT_LIST && ATisEmpty((ATermList)arg_arg1) == ATfalse) {
                                                          ATerm arg_arg1_head;
                                                          ATermList arg_arg1_list = (ATermList)arg_arg1;
                                                          arg_arg1_head = ATgetFirst(arg_arg1_list);
                                                          arg_arg1_list = ATgetNext(arg_arg1_list);
                                                          if (ATgetType((ATerm)arg_arg1_head) == AT_INT) {
                                                            if (ATgetType((ATerm)arg_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_list) == ATtrue) {
                                                              return ATtrue;
                                                            }
                                                          }
                                                        }
                                                      }
                                                    }
                                                  }
                                                }
                                              }
                                            }
                                          }
                                        }
                                      }
                                    }
                                  }
                                }
                              }
                            }
                          }
                        }
                      }
                    }
                  }
                }
              }
            }
          }
        }
      }
    }
  }
  return ATfalse;
}

/**
 * Assert whether a PERR_LexStrChar has a a. 
 * \param[in] arg input PERR_LexStrChar
 * \return ATtrue if the PERR_LexStrChar had a a, or ATfalse otherwise
 */
ATbool PERR_hasLexStrCharA(PERR_LexStrChar arg) {
  if (PERR_isLexStrCharDecimal(arg)) {
    return ATtrue;
  }
  return ATfalse;
}

/**
 * Assert whether a PERR_LexStrChar has a b. 
 * \param[in] arg input PERR_LexStrChar
 * \return ATtrue if the PERR_LexStrChar had a b, or ATfalse otherwise
 */
ATbool PERR_hasLexStrCharB(PERR_LexStrChar arg) {
  if (PERR_isLexStrCharDecimal(arg)) {
    return ATtrue;
  }
  return ATfalse;
}

/**
 * Assert whether a PERR_LexStrChar has a c. 
 * \param[in] arg input PERR_LexStrChar
 * \return ATtrue if the PERR_LexStrChar had a c, or ATfalse otherwise
 */
ATbool PERR_hasLexStrCharC(PERR_LexStrChar arg) {
  if (PERR_isLexStrCharDecimal(arg)) {
    return ATtrue;
  }
  return ATfalse;
}

/**
 * Assert whether a PERR_LexStrChar has a ch. 
 * \param[in] arg input PERR_LexStrChar
 * \return ATtrue if the PERR_LexStrChar had a ch, or ATfalse otherwise
 */
ATbool PERR_hasLexStrCharCh(PERR_LexStrChar arg) {
  if (PERR_isLexStrCharNormal(arg)) {
    return ATtrue;
  }
  return ATfalse;
}

/**
 * Get the a char of a PERR_LexStrChar. Note that the precondition is that this PERR_LexStrChar actually has a a
 * \param[in] arg input PERR_LexStrChar
 * \return the a of #arg, if it exist or an undefined value if it does not
 */
char PERR_getLexStrCharA(PERR_LexStrChar arg) {
  
    return (char)PERR_charToByte((ATerm)ATelementAt((ATermList)ATgetArgument((ATermAppl)arg, 1), 1));
}

/**
 * Get the b char of a PERR_LexStrChar. Note that the precondition is that this PERR_LexStrChar actually has a b
 * \param[in] arg input PERR_LexStrChar
 * \return the b of #arg, if it exist or an undefined value if it does not
 */
char PERR_getLexStrCharB(PERR_LexStrChar arg) {
  
    return (char)PERR_charToByte((ATerm)ATelementAt((ATermList)ATgetArgument((ATermAppl)arg, 1), 2));
}

/**
 * Get the c char of a PERR_LexStrChar. Note that the precondition is that this PERR_LexStrChar actually has a c
 * \param[in] arg input PERR_LexStrChar
 * \return the c of #arg, if it exist or an undefined value if it does not
 */
char PERR_getLexStrCharC(PERR_LexStrChar arg) {
  
    return (char)PERR_charToByte((ATerm)ATelementAt((ATermList)ATgetArgument((ATermAppl)arg, 1), 3));
}

/**
 * Get the ch char of a PERR_LexStrChar. Note that the precondition is that this PERR_LexStrChar actually has a ch
 * \param[in] arg input PERR_LexStrChar
 * \return the ch of #arg, if it exist or an undefined value if it does not
 */
char PERR_getLexStrCharCh(PERR_LexStrChar arg) {
  
    return (char)PERR_charToByte((ATerm)ATgetFirst((ATermList)ATgetArgument((ATermAppl)arg, 1)));
}

/**
 * Set the a of a PERR_LexStrChar. The precondition being that this PERR_LexStrChar actually has a a
 * \param[in] arg input PERR_LexStrChar
 * \param[in] a new char to set in #arg
 * \return A new PERR_LexStrChar with a at the right place, or a core dump if #arg did not have a a
 */
PERR_LexStrChar PERR_setLexStrCharA(PERR_LexStrChar arg, char a) {
  if (PERR_isLexStrCharDecimal(arg)) {
    return (PERR_LexStrChar)ATsetArgument((ATermAppl)arg, (ATerm)ATreplace((ATermList)ATgetArgument((ATermAppl)arg, 1), (ATerm)((ATerm) ((ATerm) PERR_byteToChar(a))), 1), 1);
  }

  ATabort("LexStrChar has no A: %t\n", arg);
  return (PERR_LexStrChar)NULL;
}

/**
 * Set the b of a PERR_LexStrChar. The precondition being that this PERR_LexStrChar actually has a b
 * \param[in] arg input PERR_LexStrChar
 * \param[in] b new char to set in #arg
 * \return A new PERR_LexStrChar with b at the right place, or a core dump if #arg did not have a b
 */
PERR_LexStrChar PERR_setLexStrCharB(PERR_LexStrChar arg, char b) {
  if (PERR_isLexStrCharDecimal(arg)) {
    return (PERR_LexStrChar)ATsetArgument((ATermAppl)arg, (ATerm)ATreplace((ATermList)ATgetArgument((ATermAppl)arg, 1), (ATerm)((ATerm) ((ATerm) PERR_byteToChar(b))), 2), 1);
  }

  ATabort("LexStrChar has no B: %t\n", arg);
  return (PERR_LexStrChar)NULL;
}

/**
 * Set the c of a PERR_LexStrChar. The precondition being that this PERR_LexStrChar actually has a c
 * \param[in] arg input PERR_LexStrChar
 * \param[in] c new char to set in #arg
 * \return A new PERR_LexStrChar with c at the right place, or a core dump if #arg did not have a c
 */
PERR_LexStrChar PERR_setLexStrCharC(PERR_LexStrChar arg, char c) {
  if (PERR_isLexStrCharDecimal(arg)) {
    return (PERR_LexStrChar)ATsetArgument((ATermAppl)arg, (ATerm)ATreplace((ATermList)ATgetArgument((ATermAppl)arg, 1), (ATerm)((ATerm) ((ATerm) PERR_byteToChar(c))), 3), 1);
  }

  ATabort("LexStrChar has no C: %t\n", arg);
  return (PERR_LexStrChar)NULL;
}

/**
 * Set the ch of a PERR_LexStrChar. The precondition being that this PERR_LexStrChar actually has a ch
 * \param[in] arg input PERR_LexStrChar
 * \param[in] ch new char to set in #arg
 * \return A new PERR_LexStrChar with ch at the right place, or a core dump if #arg did not have a ch
 */
PERR_LexStrChar PERR_setLexStrCharCh(PERR_LexStrChar arg, char ch) {
  if (PERR_isLexStrCharNormal(arg)) {
    return (PERR_LexStrChar)ATsetArgument((ATermAppl)arg, (ATerm)ATreplace((ATermList)ATgetArgument((ATermAppl)arg, 1), (ATerm)((ATerm) ((ATerm) PERR_byteToChar(ch))), 0), 1);
  }

  ATabort("LexStrChar has no Ch: %t\n", arg);
  return (PERR_LexStrChar)NULL;
}

/**
 * Assert whether a PERR_StrChar is any of the valid alternatives, or not. This analysis does not go any deeper than the top level
 * \param[in] arg input PERR_StrChar
 * \return ATtrue if #arg corresponds to the expected signature, or ATfalse otherwise
 */
ATbool PERR_isValidStrChar(PERR_StrChar arg) {
  if (PERR_isStrCharLexToCf(arg)) {
    return ATtrue;
  }
  return ATfalse;
}

/**
 * Assert whether a PERR_StrChar is a Lex-to-cf by checking against the following ATerm pattern: appl(prod([lex(sort("StrChar"))],cf(sort("StrChar")),no-attrs),[<"StrChar"("Lex-StrChar")>]). Always returns ATtrue
 * \param[in] arg input PERR_StrChar
 * \return ATtrue if #arg corresponds to the signature of a Lex-to-cf, or ATfalse otherwise
 */
inline ATbool PERR_isStrCharLexToCf(PERR_StrChar arg){
  /* checking for: appl */
  if (ATgetType((ATerm)arg) == AT_APPL && ATgetAFun((ATermAppl)arg) == PERR_afun0) {
    ATerm arg_arg0 = ATgetArgument(arg, 0);
    /* checking for: prod */
    if (ATgetType((ATerm)arg_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0) == PERR_afun1) {
      ATerm arg_arg0_arg0 = ATgetArgument(arg_arg0, 0);
      if (ATgetType((ATerm)arg_arg0_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0) == ATfalse) {
        ATerm arg_arg0_arg0_head;
        ATermList arg_arg0_arg0_list = (ATermList)arg_arg0_arg0;
        arg_arg0_arg0_head = ATgetFirst(arg_arg0_arg0_list);
        arg_arg0_arg0_list = ATgetNext(arg_arg0_arg0_list);
        /* checking for: lex */
        if (ATgetType((ATerm)arg_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head) == PERR_afun6) {
          ATerm arg_arg0_arg0_head_arg0 = ATgetArgument(arg_arg0_arg0_head, 0);
          /* checking for: sort */
          if (ATgetType((ATerm)arg_arg0_arg0_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0) == PERR_afun11) {
            ATerm arg_arg0_arg0_head_arg0_arg0 = ATgetArgument(arg_arg0_arg0_head_arg0, 0);
            /* checking for: StrChar */
            if (ATgetType((ATerm)arg_arg0_arg0_head_arg0_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0_arg0) == PERR_afun12) {
              if (ATgetType((ATerm)arg_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0_list) == ATtrue) {
                ATerm arg_arg0_arg1 = ATgetArgument(arg_arg0, 1);
                /* checking for: cf */
                if (ATgetType((ATerm)arg_arg0_arg1) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg1) == PERR_afun2) {
                  ATerm arg_arg0_arg1_arg0 = ATgetArgument(arg_arg0_arg1, 0);
                  /* checking for: sort */
                  if (ATgetType((ATerm)arg_arg0_arg1_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg1_arg0) == PERR_afun11) {
                    ATerm arg_arg0_arg1_arg0_arg0 = ATgetArgument(arg_arg0_arg1_arg0, 0);
                    /* checking for: StrChar */
                    if (ATgetType((ATerm)arg_arg0_arg1_arg0_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg1_arg0_arg0) == PERR_afun12) {
                      ATerm arg_arg0_arg2 = ATgetArgument(arg_arg0, 2);
                      /* checking for: no-attrs */
                      if (ATgetType((ATerm)arg_arg0_arg2) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg2) == PERR_afun5) {
                        ATerm arg_arg1 = ATgetArgument(arg, 1);
                        if (ATgetType((ATerm)arg_arg1) == AT_LIST && ATisEmpty((ATermList)arg_arg1) == ATfalse) {
                          ATermList arg_arg1_list = (ATermList)arg_arg1;
                          arg_arg1_list = ATgetNext(arg_arg1_list);
                          if (ATgetType((ATerm)arg_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_list) == ATtrue) {
                            return ATtrue;
                          }
                        }
                      }
                    }
                  }
                }
              }
            }
          }
        }
      }
    }
  }
  return ATfalse;
}

/**
 * Assert whether a PERR_StrChar has a StrChar. 
 * \param[in] arg input PERR_StrChar
 * \return ATtrue if the PERR_StrChar had a StrChar, or ATfalse otherwise
 */
ATbool PERR_hasStrCharStrChar(PERR_StrChar arg) {
  if (PERR_isStrCharLexToCf(arg)) {
    return ATtrue;
  }
  return ATfalse;
}

/**
 * Get the StrChar PERR_LexStrChar of a PERR_StrChar. Note that the precondition is that this PERR_StrChar actually has a StrChar
 * \param[in] arg input PERR_StrChar
 * \return the StrChar of #arg, if it exist or an undefined value if it does not
 */
PERR_LexStrChar PERR_getStrCharStrChar(PERR_StrChar arg) {
  
    return (PERR_LexStrChar)ATgetFirst((ATermList)ATgetArgument((ATermAppl)arg, 1));
}

/**
 * Set the StrChar of a PERR_StrChar. The precondition being that this PERR_StrChar actually has a StrChar
 * \param[in] arg input PERR_StrChar
 * \param[in] StrChar new PERR_LexStrChar to set in #arg
 * \return A new PERR_StrChar with StrChar at the right place, or a core dump if #arg did not have a StrChar
 */
PERR_StrChar PERR_setStrCharStrChar(PERR_StrChar arg, PERR_LexStrChar StrChar) {
  if (PERR_isStrCharLexToCf(arg)) {
    return (PERR_StrChar)ATsetArgument((ATermAppl)arg, (ATerm)ATreplace((ATermList)ATgetArgument((ATermAppl)arg, 1), (ATerm)((ATerm) StrChar), 0), 1);
  }

  ATabort("StrChar has no StrChar: %t\n", arg);
  return (PERR_StrChar)NULL;
}

/**
 * Assert whether a PERR_LexStrCon is any of the valid alternatives, or not. This analysis does not go any deeper than the top level
 * \param[in] arg input PERR_LexStrCon
 * \return ATtrue if #arg corresponds to the expected signature, or ATfalse otherwise
 */
ATbool PERR_isValidLexStrCon(PERR_LexStrCon arg) {
  if (PERR_isLexStrConDefault(arg)) {
    return ATtrue;
  }
  return ATfalse;
}

/**
 * Assert whether a PERR_LexStrCon is a default by checking against the following ATerm pattern: appl(prod([char-class([34]),lex(iter-star(sort("StrChar"))),char-class([34])],lex(sort("StrCon")),attrs([term(cons("default"))])),[34,appl(list(lex(iter-star(sort("StrChar")))),<"chars"("Lex-StrChar-chars")>),34]). Always returns ATtrue
 * \param[in] arg input PERR_LexStrCon
 * \return ATtrue if #arg corresponds to the signature of a default, or ATfalse otherwise
 */
inline ATbool PERR_isLexStrConDefault(PERR_LexStrCon arg){
  /* checking for: appl */
  if (ATgetType((ATerm)arg) == AT_APPL && ATgetAFun((ATermAppl)arg) == PERR_afun0) {
    ATerm arg_arg0 = ATgetArgument(arg, 0);
    /* checking for: prod */
    if (ATgetType((ATerm)arg_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0) == PERR_afun1) {
      ATerm arg_arg0_arg0 = ATgetArgument(arg_arg0, 0);
      if (ATgetType((ATerm)arg_arg0_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0) == ATfalse) {
        ATerm arg_arg0_arg0_head;
        ATermList arg_arg0_arg0_list = (ATermList)arg_arg0_arg0;
        arg_arg0_arg0_head = ATgetFirst(arg_arg0_arg0_list);
        arg_arg0_arg0_list = ATgetNext(arg_arg0_arg0_list);
        /* checking for: char-class */
        if (ATgetType((ATerm)arg_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head) == PERR_afun17) {
          ATerm arg_arg0_arg0_head_arg0 = ATgetArgument(arg_arg0_arg0_head, 0);
          if (ATgetType((ATerm)arg_arg0_arg0_head_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0_head_arg0) == ATfalse) {
            ATerm arg_arg0_arg0_head_arg0_head;
            ATermList arg_arg0_arg0_head_arg0_list = (ATermList)arg_arg0_arg0_head_arg0;
            arg_arg0_arg0_head_arg0_head = ATgetFirst(arg_arg0_arg0_head_arg0_list);
            arg_arg0_arg0_head_arg0_list = ATgetNext(arg_arg0_arg0_head_arg0_list);
            if (ATgetInt((ATermInt)arg_arg0_arg0_head_arg0_head) == 34) {

              if (ATgetType((ATerm)arg_arg0_arg0_head_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0_head_arg0_list) == ATtrue) {
                if (ATgetType((ATerm)arg_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0_list) == ATfalse) {
                  arg_arg0_arg0_head = ATgetFirst(arg_arg0_arg0_list);
                  arg_arg0_arg0_list = ATgetNext(arg_arg0_arg0_list);
                  /* checking for: lex */
                  if (ATgetType((ATerm)arg_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head) == PERR_afun6) {
                    ATerm arg_arg0_arg0_head_arg0 = ATgetArgument(arg_arg0_arg0_head, 0);
                    /* checking for: iter-star */
                    if (ATgetType((ATerm)arg_arg0_arg0_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0) == PERR_afun28) {
                      ATerm arg_arg0_arg0_head_arg0_arg0 = ATgetArgument(arg_arg0_arg0_head_arg0, 0);
                      /* checking for: sort */
                      if (ATgetType((ATerm)arg_arg0_arg0_head_arg0_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0_arg0) == PERR_afun11) {
                        ATerm arg_arg0_arg0_head_arg0_arg0_arg0 = ATgetArgument(arg_arg0_arg0_head_arg0_arg0, 0);
                        /* checking for: StrChar */
                        if (ATgetType((ATerm)arg_arg0_arg0_head_arg0_arg0_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0_arg0_arg0) == PERR_afun12) {
                          if (ATgetType((ATerm)arg_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0_list) == ATfalse) {
                            arg_arg0_arg0_head = ATgetFirst(arg_arg0_arg0_list);
                            arg_arg0_arg0_list = ATgetNext(arg_arg0_arg0_list);
                            /* checking for: char-class */
                            if (ATgetType((ATerm)arg_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head) == PERR_afun17) {
                              ATerm arg_arg0_arg0_head_arg0 = ATgetArgument(arg_arg0_arg0_head, 0);
                              if (ATgetType((ATerm)arg_arg0_arg0_head_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0_head_arg0) == ATfalse) {
                                ATerm arg_arg0_arg0_head_arg0_head;
                                ATermList arg_arg0_arg0_head_arg0_list = (ATermList)arg_arg0_arg0_head_arg0;
                                arg_arg0_arg0_head_arg0_head = ATgetFirst(arg_arg0_arg0_head_arg0_list);
                                arg_arg0_arg0_head_arg0_list = ATgetNext(arg_arg0_arg0_head_arg0_list);
                                if (ATgetInt((ATermInt)arg_arg0_arg0_head_arg0_head) == 34) {

                                  if (ATgetType((ATerm)arg_arg0_arg0_head_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0_head_arg0_list) == ATtrue) {
                                    if (ATgetType((ATerm)arg_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0_list) == ATtrue) {
                                      ATerm arg_arg0_arg1 = ATgetArgument(arg_arg0, 1);
                                      /* checking for: lex */
                                      if (ATgetType((ATerm)arg_arg0_arg1) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg1) == PERR_afun6) {
                                        ATerm arg_arg0_arg1_arg0 = ATgetArgument(arg_arg0_arg1, 0);
                                        /* checking for: sort */
                                        if (ATgetType((ATerm)arg_arg0_arg1_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg1_arg0) == PERR_afun11) {
                                          ATerm arg_arg0_arg1_arg0_arg0 = ATgetArgument(arg_arg0_arg1_arg0, 0);
                                          /* checking for: StrCon */
                                          if (ATgetType((ATerm)arg_arg0_arg1_arg0_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg1_arg0_arg0) == PERR_afun29) {
                                            ATerm arg_arg0_arg2 = ATgetArgument(arg_arg0, 2);
                                            /* checking for: attrs */
                                            if (ATgetType((ATerm)arg_arg0_arg2) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg2) == PERR_afun13) {
                                              ATerm arg_arg0_arg2_arg0 = ATgetArgument(arg_arg0_arg2, 0);
                                              if (ATgetType((ATerm)arg_arg0_arg2_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg2_arg0) == ATfalse) {
                                                ATerm arg_arg0_arg2_arg0_head;
                                                ATermList arg_arg0_arg2_arg0_list = (ATermList)arg_arg0_arg2_arg0;
                                                arg_arg0_arg2_arg0_head = ATgetFirst(arg_arg0_arg2_arg0_list);
                                                arg_arg0_arg2_arg0_list = ATgetNext(arg_arg0_arg2_arg0_list);
                                                /* checking for: term */
                                                if (ATgetType((ATerm)arg_arg0_arg2_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg2_arg0_head) == PERR_afun14) {
                                                  ATerm arg_arg0_arg2_arg0_head_arg0 = ATgetArgument(arg_arg0_arg2_arg0_head, 0);
                                                  /* checking for: cons */
                                                  if (ATgetType((ATerm)arg_arg0_arg2_arg0_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg2_arg0_head_arg0) == PERR_afun15) {
                                                    ATerm arg_arg0_arg2_arg0_head_arg0_arg0 = ATgetArgument(arg_arg0_arg2_arg0_head_arg0, 0);
                                                    /* checking for: default */
                                                    if (ATgetType((ATerm)arg_arg0_arg2_arg0_head_arg0_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg2_arg0_head_arg0_arg0) == PERR_afun30) {
                                                      if (ATgetType((ATerm)arg_arg0_arg2_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg2_arg0_list) == ATtrue) {
                                                        ATerm arg_arg1 = ATgetArgument(arg, 1);
                                                        if (ATgetType((ATerm)arg_arg1) == AT_LIST && ATisEmpty((ATermList)arg_arg1) == ATfalse) {
                                                          ATerm arg_arg1_head;
                                                          ATermList arg_arg1_list = (ATermList)arg_arg1;
                                                          arg_arg1_head = ATgetFirst(arg_arg1_list);
                                                          arg_arg1_list = ATgetNext(arg_arg1_list);
                                                          if (ATgetInt((ATermInt)arg_arg1_head) == 34) {

                                                            if (ATgetType((ATerm)arg_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_list) == ATfalse) {
                                                              arg_arg1_head = ATgetFirst(arg_arg1_list);
                                                              arg_arg1_list = ATgetNext(arg_arg1_list);
                                                              /* checking for: appl */
                                                              if (ATgetType((ATerm)arg_arg1_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head) == PERR_afun0) {
                                                                ATerm arg_arg1_head_arg0 = ATgetArgument(arg_arg1_head, 0);
                                                                /* checking for: list */
                                                                if (ATgetType((ATerm)arg_arg1_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0) == PERR_afun8) {
                                                                  ATerm arg_arg1_head_arg0_arg0 = ATgetArgument(arg_arg1_head_arg0, 0);
                                                                  /* checking for: lex */
                                                                  if (ATgetType((ATerm)arg_arg1_head_arg0_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg0) == PERR_afun6) {
                                                                    ATerm arg_arg1_head_arg0_arg0_arg0 = ATgetArgument(arg_arg1_head_arg0_arg0, 0);
                                                                    /* checking for: iter-star */
                                                                    if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg0_arg0) == PERR_afun28) {
                                                                      ATerm arg_arg1_head_arg0_arg0_arg0_arg0 = ATgetArgument(arg_arg1_head_arg0_arg0_arg0, 0);
                                                                      /* checking for: sort */
                                                                      if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_arg0_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg0_arg0_arg0) == PERR_afun11) {
                                                                        ATerm arg_arg1_head_arg0_arg0_arg0_arg0_arg0 = ATgetArgument(arg_arg1_head_arg0_arg0_arg0_arg0, 0);
                                                                        /* checking for: StrChar */
                                                                        if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_arg0_arg0_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg0_arg0_arg0_arg0) == PERR_afun12) {
                                                                          if (ATgetType((ATerm)arg_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_list) == ATfalse) {
                                                                            arg_arg1_head = ATgetFirst(arg_arg1_list);
                                                                            arg_arg1_list = ATgetNext(arg_arg1_list);
                                                                            if (ATgetInt((ATermInt)arg_arg1_head) == 34) {

                                                                              if (ATgetType((ATerm)arg_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_list) == ATtrue) {
                                                                                return ATtrue;
                                                                              }
                                                                            }
                                                                          }
                                                                        }
                                                                      }
                                                                    }
                                                                  }
                                                                }
                                                              }
                                                            }
                                                          }
                                                        }
                                                      }
                                                    }
                                                  }
                                                }
                                              }
                                            }
                                          }
                                        }
                                      }
                                    }
                                  }
                                }
                              }
                            }
                          }
                        }
                      }
                    }
                  }
                }
              }
            }
          }
        }
      }
    }
  }
  return ATfalse;
}

/**
 * Assert whether a PERR_LexStrCon has a chars. 
 * \param[in] arg input PERR_LexStrCon
 * \return ATtrue if the PERR_LexStrCon had a chars, or ATfalse otherwise
 */
ATbool PERR_hasLexStrConChars(PERR_LexStrCon arg) {
  if (PERR_isLexStrConDefault(arg)) {
    return ATtrue;
  }
  return ATfalse;
}

/**
 * Get the chars PERR_LexStrCharChars of a PERR_LexStrCon. Note that the precondition is that this PERR_LexStrCon actually has a chars
 * \param[in] arg input PERR_LexStrCon
 * \return the chars of #arg, if it exist or an undefined value if it does not
 */
PERR_LexStrCharChars PERR_getLexStrConChars(PERR_LexStrCon arg) {
  
    return (PERR_LexStrCharChars)ATgetArgument((ATermAppl)ATelementAt((ATermList)ATgetArgument((ATermAppl)arg, 1), 1), 1);
}

/**
 * Set the chars of a PERR_LexStrCon. The precondition being that this PERR_LexStrCon actually has a chars
 * \param[in] arg input PERR_LexStrCon
 * \param[in] chars new PERR_LexStrCharChars to set in #arg
 * \return A new PERR_LexStrCon with chars at the right place, or a core dump if #arg did not have a chars
 */
PERR_LexStrCon PERR_setLexStrConChars(PERR_LexStrCon arg, PERR_LexStrCharChars chars) {
  if (PERR_isLexStrConDefault(arg)) {
    return (PERR_LexStrCon)ATsetArgument((ATermAppl)arg, (ATerm)ATreplace((ATermList)ATgetArgument((ATermAppl)arg, 1), (ATerm)ATsetArgument((ATermAppl)ATelementAt((ATermList)ATgetArgument((ATermAppl)arg, 1), 1), (ATerm)((ATerm) chars), 1), 1), 1);
  }

  ATabort("LexStrCon has no Chars: %t\n", arg);
  return (PERR_LexStrCon)NULL;
}

/**
 * Assert whether a PERR_StrCon is any of the valid alternatives, or not. This analysis does not go any deeper than the top level
 * \param[in] arg input PERR_StrCon
 * \return ATtrue if #arg corresponds to the expected signature, or ATfalse otherwise
 */
ATbool PERR_isValidStrCon(PERR_StrCon arg) {
  if (PERR_isStrConLexToCf(arg)) {
    return ATtrue;
  }
  return ATfalse;
}

/**
 * Assert whether a PERR_StrCon is a Lex-to-cf by checking against the following ATerm pattern: appl(prod([lex(sort("StrCon"))],cf(sort("StrCon")),no-attrs),[<"StrCon"("Lex-StrCon")>]). Always returns ATtrue
 * \param[in] arg input PERR_StrCon
 * \return ATtrue if #arg corresponds to the signature of a Lex-to-cf, or ATfalse otherwise
 */
inline ATbool PERR_isStrConLexToCf(PERR_StrCon arg){
  /* checking for: appl */
  if (ATgetType((ATerm)arg) == AT_APPL && ATgetAFun((ATermAppl)arg) == PERR_afun0) {
    ATerm arg_arg0 = ATgetArgument(arg, 0);
    /* checking for: prod */
    if (ATgetType((ATerm)arg_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0) == PERR_afun1) {
      ATerm arg_arg0_arg0 = ATgetArgument(arg_arg0, 0);
      if (ATgetType((ATerm)arg_arg0_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0) == ATfalse) {
        ATerm arg_arg0_arg0_head;
        ATermList arg_arg0_arg0_list = (ATermList)arg_arg0_arg0;
        arg_arg0_arg0_head = ATgetFirst(arg_arg0_arg0_list);
        arg_arg0_arg0_list = ATgetNext(arg_arg0_arg0_list);
        /* checking for: lex */
        if (ATgetType((ATerm)arg_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head) == PERR_afun6) {
          ATerm arg_arg0_arg0_head_arg0 = ATgetArgument(arg_arg0_arg0_head, 0);
          /* checking for: sort */
          if (ATgetType((ATerm)arg_arg0_arg0_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0) == PERR_afun11) {
            ATerm arg_arg0_arg0_head_arg0_arg0 = ATgetArgument(arg_arg0_arg0_head_arg0, 0);
            /* checking for: StrCon */
            if (ATgetType((ATerm)arg_arg0_arg0_head_arg0_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0_arg0) == PERR_afun29) {
              if (ATgetType((ATerm)arg_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0_list) == ATtrue) {
                ATerm arg_arg0_arg1 = ATgetArgument(arg_arg0, 1);
                /* checking for: cf */
                if (ATgetType((ATerm)arg_arg0_arg1) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg1) == PERR_afun2) {
                  ATerm arg_arg0_arg1_arg0 = ATgetArgument(arg_arg0_arg1, 0);
                  /* checking for: sort */
                  if (ATgetType((ATerm)arg_arg0_arg1_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg1_arg0) == PERR_afun11) {
                    ATerm arg_arg0_arg1_arg0_arg0 = ATgetArgument(arg_arg0_arg1_arg0, 0);
                    /* checking for: StrCon */
                    if (ATgetType((ATerm)arg_arg0_arg1_arg0_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg1_arg0_arg0) == PERR_afun29) {
                      ATerm arg_arg0_arg2 = ATgetArgument(arg_arg0, 2);
                      /* checking for: no-attrs */
                      if (ATgetType((ATerm)arg_arg0_arg2) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg2) == PERR_afun5) {
                        ATerm arg_arg1 = ATgetArgument(arg, 1);
                        if (ATgetType((ATerm)arg_arg1) == AT_LIST && ATisEmpty((ATermList)arg_arg1) == ATfalse) {
                          ATermList arg_arg1_list = (ATermList)arg_arg1;
                          arg_arg1_list = ATgetNext(arg_arg1_list);
                          if (ATgetType((ATerm)arg_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_list) == ATtrue) {
                            return ATtrue;
                          }
                        }
                      }
                    }
                  }
                }
              }
            }
          }
        }
      }
    }
  }
  return ATfalse;
}

/**
 * Assert whether a PERR_StrCon has a StrCon. 
 * \param[in] arg input PERR_StrCon
 * \return ATtrue if the PERR_StrCon had a StrCon, or ATfalse otherwise
 */
ATbool PERR_hasStrConStrCon(PERR_StrCon arg) {
  if (PERR_isStrConLexToCf(arg)) {
    return ATtrue;
  }
  return ATfalse;
}

/**
 * Get the StrCon PERR_LexStrCon of a PERR_StrCon. Note that the precondition is that this PERR_StrCon actually has a StrCon
 * \param[in] arg input PERR_StrCon
 * \return the StrCon of #arg, if it exist or an undefined value if it does not
 */
PERR_LexStrCon PERR_getStrConStrCon(PERR_StrCon arg) {
  
    return (PERR_LexStrCon)ATgetFirst((ATermList)ATgetArgument((ATermAppl)arg, 1));
}

/**
 * Set the StrCon of a PERR_StrCon. The precondition being that this PERR_StrCon actually has a StrCon
 * \param[in] arg input PERR_StrCon
 * \param[in] StrCon new PERR_LexStrCon to set in #arg
 * \return A new PERR_StrCon with StrCon at the right place, or a core dump if #arg did not have a StrCon
 */
PERR_StrCon PERR_setStrConStrCon(PERR_StrCon arg, PERR_LexStrCon StrCon) {
  if (PERR_isStrConLexToCf(arg)) {
    return (PERR_StrCon)ATsetArgument((ATermAppl)arg, (ATerm)ATreplace((ATermList)ATgetArgument((ATermAppl)arg, 1), (ATerm)((ATerm) StrCon), 0), 1);
  }

  ATabort("StrCon has no StrCon: %t\n", arg);
  return (PERR_StrCon)NULL;
}

/**
 * Assert whether a PERR_LexStrCharChars is any of the valid alternatives, or not. This analysis does not go any deeper than the top level
 * \param[in] arg input PERR_LexStrCharChars
 * \return ATtrue if #arg corresponds to the expected signature, or ATfalse otherwise
 */
ATbool PERR_isValidLexStrCharChars(PERR_LexStrCharChars arg) {
  if (PERR_isLexStrCharCharsEmpty(arg)) {
    return ATtrue;
  }
  else if (PERR_isLexStrCharCharsSingle(arg)) {
    return ATtrue;
  }
  else if (PERR_isLexStrCharCharsMany(arg)) {
    return ATtrue;
  }
  return ATfalse;
}

/**
 * Assert whether a PERR_LexStrCharChars is a empty by checking against the following ATerm pattern: []. May not be used to assert correctness of the PERR_LexStrCharChars
 * \param[in] arg input PERR_LexStrCharChars
 * \return ATtrue if #arg corresponds to the signature of a empty, or ATfalse otherwise
 */
inline ATbool PERR_isLexStrCharCharsEmpty(PERR_LexStrCharChars arg){
  if (ATisEmpty((ATermList)arg)) {
    return ATtrue;
  }
  return ATfalse;
}

/**
 * Assert whether a PERR_LexStrCharChars is a single by checking against the following ATerm pattern: [<head(Lex-StrChar)>]. May not be used to assert correctness of the PERR_LexStrCharChars
 * \param[in] arg input PERR_LexStrCharChars
 * \return ATtrue if #arg corresponds to the signature of a single, or ATfalse otherwise
 */
inline ATbool PERR_isLexStrCharCharsSingle(PERR_LexStrCharChars arg){
  if (ATgetType((ATerm)arg) == AT_LIST && ATisEmpty((ATermList)arg) == ATfalse) {
    ATermList arg_list = (ATermList)arg;
    arg_list = ATgetNext(arg_list);
    if (ATgetType((ATerm)arg_list) == AT_LIST && ATisEmpty((ATermList)arg_list) == ATtrue) {
      return ATtrue;
    }
  }
  return ATfalse;
}

/**
 * Assert whether a PERR_LexStrCharChars is a many by checking against the following ATerm pattern: [<head(Lex-StrChar)>,<[tail(Lex-StrChar-chars)]>]. May not be used to assert correctness of the PERR_LexStrCharChars
 * \param[in] arg input PERR_LexStrCharChars
 * \return ATtrue if #arg corresponds to the signature of a many, or ATfalse otherwise
 */
inline ATbool PERR_isLexStrCharCharsMany(PERR_LexStrCharChars arg){
  if (ATgetType((ATerm)arg) == AT_LIST && ATisEmpty((ATermList)arg) == ATfalse) {
    return ATtrue;
  }
  return ATfalse;
}

/**
 * Assert whether a PERR_LexStrCharChars has a head. 
 * \param[in] arg input PERR_LexStrCharChars
 * \return ATtrue if the PERR_LexStrCharChars had a head, or ATfalse otherwise
 */
ATbool PERR_hasLexStrCharCharsHead(PERR_LexStrCharChars arg) {
  if (PERR_isLexStrCharCharsSingle(arg)) {
    return ATtrue;
  }
  else if (PERR_isLexStrCharCharsMany(arg)) {
    return ATtrue;
  }
  return ATfalse;
}

/**
 * Assert whether a PERR_LexStrCharChars has a tail. 
 * \param[in] arg input PERR_LexStrCharChars
 * \return ATtrue if the PERR_LexStrCharChars had a tail, or ATfalse otherwise
 */
ATbool PERR_hasLexStrCharCharsTail(PERR_LexStrCharChars arg) {
  if (PERR_isLexStrCharCharsMany(arg)) {
    return ATtrue;
  }
  return ATfalse;
}

/**
 * Get the head PERR_LexStrChar of a PERR_LexStrCharChars. Note that the precondition is that this PERR_LexStrCharChars actually has a head
 * \param[in] arg input PERR_LexStrCharChars
 * \return the head of #arg, if it exist or an undefined value if it does not
 */
PERR_LexStrChar PERR_getLexStrCharCharsHead(PERR_LexStrCharChars arg) {
  if (PERR_isLexStrCharCharsSingle(arg)) {
    return (PERR_LexStrChar)ATgetFirst((ATermList)arg);
  }
  else 
    return (PERR_LexStrChar)ATgetFirst((ATermList)arg);
}

/**
 * Get the tail PERR_LexStrCharChars of a PERR_LexStrCharChars. Note that the precondition is that this PERR_LexStrCharChars actually has a tail
 * \param[in] arg input PERR_LexStrCharChars
 * \return the tail of #arg, if it exist or an undefined value if it does not
 */
PERR_LexStrCharChars PERR_getLexStrCharCharsTail(PERR_LexStrCharChars arg) {
  
    return (PERR_LexStrCharChars)ATgetNext((ATermList)arg);
}

/**
 * Set the head of a PERR_LexStrCharChars. The precondition being that this PERR_LexStrCharChars actually has a head
 * \param[in] arg input PERR_LexStrCharChars
 * \param[in] head new PERR_LexStrChar to set in #arg
 * \return A new PERR_LexStrCharChars with head at the right place, or a core dump if #arg did not have a head
 */
PERR_LexStrCharChars PERR_setLexStrCharCharsHead(PERR_LexStrCharChars arg, PERR_LexStrChar head) {
  if (PERR_isLexStrCharCharsSingle(arg)) {
    return (PERR_LexStrCharChars)ATreplace((ATermList)arg, (ATerm)((ATerm) head), 0);
  }
  else if (PERR_isLexStrCharCharsMany(arg)) {
    return (PERR_LexStrCharChars)ATreplace((ATermList)arg, (ATerm)((ATerm) head), 0);
  }

  ATabort("LexStrCharChars has no Head: %t\n", arg);
  return (PERR_LexStrCharChars)NULL;
}

/**
 * Set the tail of a PERR_LexStrCharChars. The precondition being that this PERR_LexStrCharChars actually has a tail
 * \param[in] arg input PERR_LexStrCharChars
 * \param[in] tail new PERR_LexStrCharChars to set in #arg
 * \return A new PERR_LexStrCharChars with tail at the right place, or a core dump if #arg did not have a tail
 */
PERR_LexStrCharChars PERR_setLexStrCharCharsTail(PERR_LexStrCharChars arg, PERR_LexStrCharChars tail) {
  if (PERR_isLexStrCharCharsMany(arg)) {
    return (PERR_LexStrCharChars)ATreplaceTail((ATermList)arg, (ATermList)((ATerm) tail), 1);
  }

  ATabort("LexStrCharChars has no Tail: %t\n", arg);
  return (PERR_LexStrCharChars)NULL;
}

/**
 * Assert whether a PERR_LexNatCon is any of the valid alternatives, or not. This analysis does not go any deeper than the top level
 * \param[in] arg input PERR_LexNatCon
 * \return ATtrue if #arg corresponds to the expected signature, or ATfalse otherwise
 */
ATbool PERR_isValidLexNatCon(PERR_LexNatCon arg) {
  if (PERR_isLexNatConDigits(arg)) {
    return ATtrue;
  }
  return ATfalse;
}

/**
 * Assert whether a PERR_LexNatCon is a digits by checking against the following ATerm pattern: appl(prod([lex(iter(char-class([range(48,57)])))],lex(sort("NatCon")),attrs([term(cons("digits"))])),[appl(list(lex(iter(char-class([range(48,57)])))),<"list"(chars)>)]). Always returns ATtrue
 * \param[in] arg input PERR_LexNatCon
 * \return ATtrue if #arg corresponds to the signature of a digits, or ATfalse otherwise
 */
inline ATbool PERR_isLexNatConDigits(PERR_LexNatCon arg){
  /* checking for: appl */
  if (ATgetType((ATerm)arg) == AT_APPL && ATgetAFun((ATermAppl)arg) == PERR_afun0) {
    ATerm arg_arg0 = ATgetArgument(arg, 0);
    /* checking for: prod */
    if (ATgetType((ATerm)arg_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0) == PERR_afun1) {
      ATerm arg_arg0_arg0 = ATgetArgument(arg_arg0, 0);
      if (ATgetType((ATerm)arg_arg0_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0) == ATfalse) {
        ATerm arg_arg0_arg0_head;
        ATermList arg_arg0_arg0_list = (ATermList)arg_arg0_arg0;
        arg_arg0_arg0_head = ATgetFirst(arg_arg0_arg0_list);
        arg_arg0_arg0_list = ATgetNext(arg_arg0_arg0_list);
        /* checking for: lex */
        if (ATgetType((ATerm)arg_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head) == PERR_afun6) {
          ATerm arg_arg0_arg0_head_arg0 = ATgetArgument(arg_arg0_arg0_head, 0);
          /* checking for: iter */
          if (ATgetType((ATerm)arg_arg0_arg0_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0) == PERR_afun7) {
            ATerm arg_arg0_arg0_head_arg0_arg0 = ATgetArgument(arg_arg0_arg0_head_arg0, 0);
            /* checking for: char-class */
            if (ATgetType((ATerm)arg_arg0_arg0_head_arg0_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0_arg0) == PERR_afun17) {
              ATerm arg_arg0_arg0_head_arg0_arg0_arg0 = ATgetArgument(arg_arg0_arg0_head_arg0_arg0, 0);
              if (ATgetType((ATerm)arg_arg0_arg0_head_arg0_arg0_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0_head_arg0_arg0_arg0) == ATfalse) {
                ATerm arg_arg0_arg0_head_arg0_arg0_arg0_head;
                ATermList arg_arg0_arg0_head_arg0_arg0_arg0_list = (ATermList)arg_arg0_arg0_head_arg0_arg0_arg0;
                arg_arg0_arg0_head_arg0_arg0_arg0_head = ATgetFirst(arg_arg0_arg0_head_arg0_arg0_arg0_list);
                arg_arg0_arg0_head_arg0_arg0_arg0_list = ATgetNext(arg_arg0_arg0_head_arg0_arg0_arg0_list);
                /* checking for: range */
                if (ATgetType((ATerm)arg_arg0_arg0_head_arg0_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0_arg0_arg0_head) == PERR_afun24) {
                  ATerm arg_arg0_arg0_head_arg0_arg0_arg0_head_arg0 = ATgetArgument(arg_arg0_arg0_head_arg0_arg0_arg0_head, 0);
                  if (ATgetInt((ATermInt)arg_arg0_arg0_head_arg0_arg0_arg0_head_arg0) == 48) {

                    ATerm arg_arg0_arg0_head_arg0_arg0_arg0_head_arg1 = ATgetArgument(arg_arg0_arg0_head_arg0_arg0_arg0_head, 1);
                    if (ATgetInt((ATermInt)arg_arg0_arg0_head_arg0_arg0_arg0_head_arg1) == 57) {

                      if (ATgetType((ATerm)arg_arg0_arg0_head_arg0_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0_head_arg0_arg0_arg0_list) == ATtrue) {
                        if (ATgetType((ATerm)arg_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0_list) == ATtrue) {
                          ATerm arg_arg0_arg1 = ATgetArgument(arg_arg0, 1);
                          /* checking for: lex */
                          if (ATgetType((ATerm)arg_arg0_arg1) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg1) == PERR_afun6) {
                            ATerm arg_arg0_arg1_arg0 = ATgetArgument(arg_arg0_arg1, 0);
                            /* checking for: sort */
                            if (ATgetType((ATerm)arg_arg0_arg1_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg1_arg0) == PERR_afun11) {
                              ATerm arg_arg0_arg1_arg0_arg0 = ATgetArgument(arg_arg0_arg1_arg0, 0);
                              /* checking for: NatCon */
                              if (ATgetType((ATerm)arg_arg0_arg1_arg0_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg1_arg0_arg0) == PERR_afun31) {
                                ATerm arg_arg0_arg2 = ATgetArgument(arg_arg0, 2);
                                /* checking for: attrs */
                                if (ATgetType((ATerm)arg_arg0_arg2) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg2) == PERR_afun13) {
                                  ATerm arg_arg0_arg2_arg0 = ATgetArgument(arg_arg0_arg2, 0);
                                  if (ATgetType((ATerm)arg_arg0_arg2_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg2_arg0) == ATfalse) {
                                    ATerm arg_arg0_arg2_arg0_head;
                                    ATermList arg_arg0_arg2_arg0_list = (ATermList)arg_arg0_arg2_arg0;
                                    arg_arg0_arg2_arg0_head = ATgetFirst(arg_arg0_arg2_arg0_list);
                                    arg_arg0_arg2_arg0_list = ATgetNext(arg_arg0_arg2_arg0_list);
                                    /* checking for: term */
                                    if (ATgetType((ATerm)arg_arg0_arg2_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg2_arg0_head) == PERR_afun14) {
                                      ATerm arg_arg0_arg2_arg0_head_arg0 = ATgetArgument(arg_arg0_arg2_arg0_head, 0);
                                      /* checking for: cons */
                                      if (ATgetType((ATerm)arg_arg0_arg2_arg0_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg2_arg0_head_arg0) == PERR_afun15) {
                                        ATerm arg_arg0_arg2_arg0_head_arg0_arg0 = ATgetArgument(arg_arg0_arg2_arg0_head_arg0, 0);
                                        /* checking for: digits */
                                        if (ATgetType((ATerm)arg_arg0_arg2_arg0_head_arg0_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg2_arg0_head_arg0_arg0) == PERR_afun32) {
                                          if (ATgetType((ATerm)arg_arg0_arg2_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg2_arg0_list) == ATtrue) {
                                            ATerm arg_arg1 = ATgetArgument(arg, 1);
                                            if (ATgetType((ATerm)arg_arg1) == AT_LIST && ATisEmpty((ATermList)arg_arg1) == ATfalse) {
                                              ATerm arg_arg1_head;
                                              ATermList arg_arg1_list = (ATermList)arg_arg1;
                                              arg_arg1_head = ATgetFirst(arg_arg1_list);
                                              arg_arg1_list = ATgetNext(arg_arg1_list);
                                              /* checking for: appl */
                                              if (ATgetType((ATerm)arg_arg1_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head) == PERR_afun0) {
                                                ATerm arg_arg1_head_arg0 = ATgetArgument(arg_arg1_head, 0);
                                                /* checking for: list */
                                                if (ATgetType((ATerm)arg_arg1_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0) == PERR_afun8) {
                                                  ATerm arg_arg1_head_arg0_arg0 = ATgetArgument(arg_arg1_head_arg0, 0);
                                                  /* checking for: lex */
                                                  if (ATgetType((ATerm)arg_arg1_head_arg0_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg0) == PERR_afun6) {
                                                    ATerm arg_arg1_head_arg0_arg0_arg0 = ATgetArgument(arg_arg1_head_arg0_arg0, 0);
                                                    /* checking for: iter */
                                                    if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg0_arg0) == PERR_afun7) {
                                                      ATerm arg_arg1_head_arg0_arg0_arg0_arg0 = ATgetArgument(arg_arg1_head_arg0_arg0_arg0, 0);
                                                      /* checking for: char-class */
                                                      if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_arg0_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg0_arg0_arg0) == PERR_afun17) {
                                                        ATerm arg_arg1_head_arg0_arg0_arg0_arg0_arg0 = ATgetArgument(arg_arg1_head_arg0_arg0_arg0_arg0, 0);
                                                        if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_arg0_arg0_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_arg0_arg0_arg0) == ATfalse) {
                                                          ATerm arg_arg1_head_arg0_arg0_arg0_arg0_arg0_head;
                                                          ATermList arg_arg1_head_arg0_arg0_arg0_arg0_arg0_list = (ATermList)arg_arg1_head_arg0_arg0_arg0_arg0_arg0;
                                                          arg_arg1_head_arg0_arg0_arg0_arg0_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_arg0_arg0_arg0_list);
                                                          arg_arg1_head_arg0_arg0_arg0_arg0_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_arg0_arg0_arg0_list);
                                                          /* checking for: range */
                                                          if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_arg0_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg0_arg0_arg0_arg0_head) == PERR_afun24) {
                                                            ATerm arg_arg1_head_arg0_arg0_arg0_arg0_arg0_head_arg0 = ATgetArgument(arg_arg1_head_arg0_arg0_arg0_arg0_arg0_head, 0);
                                                            if (ATgetInt((ATermInt)arg_arg1_head_arg0_arg0_arg0_arg0_arg0_head_arg0) == 48) {

                                                              ATerm arg_arg1_head_arg0_arg0_arg0_arg0_arg0_head_arg1 = ATgetArgument(arg_arg1_head_arg0_arg0_arg0_arg0_arg0_head, 1);
                                                              if (ATgetInt((ATermInt)arg_arg1_head_arg0_arg0_arg0_arg0_arg0_head_arg1) == 57) {

                                                                if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_arg0_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_arg0_arg0_arg0_list) == ATtrue) {
                                                                  ATerm arg_arg1_head_arg1 = ATgetArgument(arg_arg1_head, 1);
                                                                  if (ATgetType((ATerm)arg_arg1_head_arg1) == AT_LIST) {
                                                                    if (ATgetType((ATerm)arg_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_list) == ATtrue) {
                                                                      return ATtrue;
                                                                    }
                                                                  }
                                                                }
                                                              }
                                                            }
                                                          }
                                                        }
                                                      }
                                                    }
                                                  }
                                                }
                                              }
                                            }
                                          }
                                        }
                                      }
                                    }
                                  }
                                }
                              }
                            }
                          }
                        }
                      }
                    }
                  }
                }
              }
            }
          }
        }
      }
    }
  }
  return ATfalse;
}

/**
 * Assert whether a PERR_LexNatCon has a list. 
 * \param[in] arg input PERR_LexNatCon
 * \return ATtrue if the PERR_LexNatCon had a list, or ATfalse otherwise
 */
ATbool PERR_hasLexNatConList(PERR_LexNatCon arg) {
  if (PERR_isLexNatConDigits(arg)) {
    return ATtrue;
  }
  return ATfalse;
}

/**
 * Get the list char* of a PERR_LexNatCon. Note that the precondition is that this PERR_LexNatCon actually has a list
 * \param[in] arg input PERR_LexNatCon
 * \return the list of #arg, if it exist or an undefined value if it does not
 */
char* PERR_getLexNatConList(PERR_LexNatCon arg) {
  
    return (char*)PERR_charsToString((ATerm)ATgetArgument((ATermAppl)ATgetFirst((ATermList)ATgetArgument((ATermAppl)arg, 1)), 1));
}

/**
 * Set the list of a PERR_LexNatCon. The precondition being that this PERR_LexNatCon actually has a list
 * \param[in] arg input PERR_LexNatCon
 * \param[in] list new const char* to set in #arg
 * \return A new PERR_LexNatCon with list at the right place, or a core dump if #arg did not have a list
 */
PERR_LexNatCon PERR_setLexNatConList(PERR_LexNatCon arg, const char* list) {
  if (PERR_isLexNatConDigits(arg)) {
    return (PERR_LexNatCon)ATsetArgument((ATermAppl)arg, (ATerm)ATreplace((ATermList)ATgetArgument((ATermAppl)arg, 1), (ATerm)ATsetArgument((ATermAppl)ATgetFirst((ATermList)ATgetArgument((ATermAppl)arg, 1)), (ATerm)((ATerm) ((ATerm) PERR_stringToChars(list))), 1), 0), 1);
  }

  ATabort("LexNatCon has no List: %t\n", arg);
  return (PERR_LexNatCon)NULL;
}

/**
 * Assert whether a PERR_NatCon is any of the valid alternatives, or not. This analysis does not go any deeper than the top level
 * \param[in] arg input PERR_NatCon
 * \return ATtrue if #arg corresponds to the expected signature, or ATfalse otherwise
 */
ATbool PERR_isValidNatCon(PERR_NatCon arg) {
  if (PERR_isNatConLexToCf(arg)) {
    return ATtrue;
  }
  return ATfalse;
}

/**
 * Assert whether a PERR_NatCon is a Lex-to-cf by checking against the following ATerm pattern: appl(prod([lex(sort("NatCon"))],cf(sort("NatCon")),no-attrs),[<"NatCon"("Lex-NatCon")>]). Always returns ATtrue
 * \param[in] arg input PERR_NatCon
 * \return ATtrue if #arg corresponds to the signature of a Lex-to-cf, or ATfalse otherwise
 */
inline ATbool PERR_isNatConLexToCf(PERR_NatCon arg){
  /* checking for: appl */
  if (ATgetType((ATerm)arg) == AT_APPL && ATgetAFun((ATermAppl)arg) == PERR_afun0) {
    ATerm arg_arg0 = ATgetArgument(arg, 0);
    /* checking for: prod */
    if (ATgetType((ATerm)arg_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0) == PERR_afun1) {
      ATerm arg_arg0_arg0 = ATgetArgument(arg_arg0, 0);
      if (ATgetType((ATerm)arg_arg0_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0) == ATfalse) {
        ATerm arg_arg0_arg0_head;
        ATermList arg_arg0_arg0_list = (ATermList)arg_arg0_arg0;
        arg_arg0_arg0_head = ATgetFirst(arg_arg0_arg0_list);
        arg_arg0_arg0_list = ATgetNext(arg_arg0_arg0_list);
        /* checking for: lex */
        if (ATgetType((ATerm)arg_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head) == PERR_afun6) {
          ATerm arg_arg0_arg0_head_arg0 = ATgetArgument(arg_arg0_arg0_head, 0);
          /* checking for: sort */
          if (ATgetType((ATerm)arg_arg0_arg0_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0) == PERR_afun11) {
            ATerm arg_arg0_arg0_head_arg0_arg0 = ATgetArgument(arg_arg0_arg0_head_arg0, 0);
            /* checking for: NatCon */
            if (ATgetType((ATerm)arg_arg0_arg0_head_arg0_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0_arg0) == PERR_afun31) {
              if (ATgetType((ATerm)arg_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0_list) == ATtrue) {
                ATerm arg_arg0_arg1 = ATgetArgument(arg_arg0, 1);
                /* checking for: cf */
                if (ATgetType((ATerm)arg_arg0_arg1) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg1) == PERR_afun2) {
                  ATerm arg_arg0_arg1_arg0 = ATgetArgument(arg_arg0_arg1, 0);
                  /* checking for: sort */
                  if (ATgetType((ATerm)arg_arg0_arg1_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg1_arg0) == PERR_afun11) {
                    ATerm arg_arg0_arg1_arg0_arg0 = ATgetArgument(arg_arg0_arg1_arg0, 0);
                    /* checking for: NatCon */
                    if (ATgetType((ATerm)arg_arg0_arg1_arg0_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg1_arg0_arg0) == PERR_afun31) {
                      ATerm arg_arg0_arg2 = ATgetArgument(arg_arg0, 2);
                      /* checking for: no-attrs */
                      if (ATgetType((ATerm)arg_arg0_arg2) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg2) == PERR_afun5) {
                        ATerm arg_arg1 = ATgetArgument(arg, 1);
                        if (ATgetType((ATerm)arg_arg1) == AT_LIST && ATisEmpty((ATermList)arg_arg1) == ATfalse) {
                          ATermList arg_arg1_list = (ATermList)arg_arg1;
                          arg_arg1_list = ATgetNext(arg_arg1_list);
                          if (ATgetType((ATerm)arg_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_list) == ATtrue) {
                            return ATtrue;
                          }
                        }
                      }
                    }
                  }
                }
              }
            }
          }
        }
      }
    }
  }
  return ATfalse;
}

/**
 * Assert whether a PERR_NatCon has a NatCon. 
 * \param[in] arg input PERR_NatCon
 * \return ATtrue if the PERR_NatCon had a NatCon, or ATfalse otherwise
 */
ATbool PERR_hasNatConNatCon(PERR_NatCon arg) {
  if (PERR_isNatConLexToCf(arg)) {
    return ATtrue;
  }
  return ATfalse;
}

/**
 * Get the NatCon PERR_LexNatCon of a PERR_NatCon. Note that the precondition is that this PERR_NatCon actually has a NatCon
 * \param[in] arg input PERR_NatCon
 * \return the NatCon of #arg, if it exist or an undefined value if it does not
 */
PERR_LexNatCon PERR_getNatConNatCon(PERR_NatCon arg) {
  
    return (PERR_LexNatCon)ATgetFirst((ATermList)ATgetArgument((ATermAppl)arg, 1));
}

/**
 * Set the NatCon of a PERR_NatCon. The precondition being that this PERR_NatCon actually has a NatCon
 * \param[in] arg input PERR_NatCon
 * \param[in] NatCon new PERR_LexNatCon to set in #arg
 * \return A new PERR_NatCon with NatCon at the right place, or a core dump if #arg did not have a NatCon
 */
PERR_NatCon PERR_setNatConNatCon(PERR_NatCon arg, PERR_LexNatCon NatCon) {
  if (PERR_isNatConLexToCf(arg)) {
    return (PERR_NatCon)ATsetArgument((ATermAppl)arg, (ATerm)ATreplace((ATermList)ATgetArgument((ATermAppl)arg, 1), (ATerm)((ATerm) NatCon), 0), 1);
  }

  ATabort("NatCon has no NatCon: %t\n", arg);
  return (PERR_NatCon)NULL;
}

/**
 * Assert whether a PERR_LexLayout is any of the valid alternatives, or not. This analysis does not go any deeper than the top level
 * \param[in] arg input PERR_LexLayout
 * \return ATtrue if #arg corresponds to the expected signature, or ATfalse otherwise
 */
ATbool PERR_isValidLexLayout(PERR_LexLayout arg) {
  if (PERR_isLexLayoutWhitespace(arg)) {
    return ATtrue;
  }
  return ATfalse;
}

/**
 * Assert whether a PERR_LexLayout is a whitespace by checking against the following ATerm pattern: appl(prod([char-class([range(9,10),13,32])],lex(layout),attrs([term(cons("whitespace"))])),[<"ch"(char)>]). Always returns ATtrue
 * \param[in] arg input PERR_LexLayout
 * \return ATtrue if #arg corresponds to the signature of a whitespace, or ATfalse otherwise
 */
inline ATbool PERR_isLexLayoutWhitespace(PERR_LexLayout arg){
  /* checking for: appl */
  if (ATgetType((ATerm)arg) == AT_APPL && ATgetAFun((ATermAppl)arg) == PERR_afun0) {
    ATerm arg_arg0 = ATgetArgument(arg, 0);
    /* checking for: prod */
    if (ATgetType((ATerm)arg_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0) == PERR_afun1) {
      ATerm arg_arg0_arg0 = ATgetArgument(arg_arg0, 0);
      if (ATgetType((ATerm)arg_arg0_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0) == ATfalse) {
        ATerm arg_arg0_arg0_head;
        ATermList arg_arg0_arg0_list = (ATermList)arg_arg0_arg0;
        arg_arg0_arg0_head = ATgetFirst(arg_arg0_arg0_list);
        arg_arg0_arg0_list = ATgetNext(arg_arg0_arg0_list);
        /* checking for: char-class */
        if (ATgetType((ATerm)arg_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head) == PERR_afun17) {
          ATerm arg_arg0_arg0_head_arg0 = ATgetArgument(arg_arg0_arg0_head, 0);
          if (ATgetType((ATerm)arg_arg0_arg0_head_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0_head_arg0) == ATfalse) {
            ATerm arg_arg0_arg0_head_arg0_head;
            ATermList arg_arg0_arg0_head_arg0_list = (ATermList)arg_arg0_arg0_head_arg0;
            arg_arg0_arg0_head_arg0_head = ATgetFirst(arg_arg0_arg0_head_arg0_list);
            arg_arg0_arg0_head_arg0_list = ATgetNext(arg_arg0_arg0_head_arg0_list);
            /* checking for: range */
            if (ATgetType((ATerm)arg_arg0_arg0_head_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0_head) == PERR_afun24) {
              ATerm arg_arg0_arg0_head_arg0_head_arg0 = ATgetArgument(arg_arg0_arg0_head_arg0_head, 0);
              if (ATgetInt((ATermInt)arg_arg0_arg0_head_arg0_head_arg0) == 9) {

                ATerm arg_arg0_arg0_head_arg0_head_arg1 = ATgetArgument(arg_arg0_arg0_head_arg0_head, 1);
                if (ATgetInt((ATermInt)arg_arg0_arg0_head_arg0_head_arg1) == 10) {

                  if (ATgetType((ATerm)arg_arg0_arg0_head_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0_head_arg0_list) == ATfalse) {
                    arg_arg0_arg0_head_arg0_head = ATgetFirst(arg_arg0_arg0_head_arg0_list);
                    arg_arg0_arg0_head_arg0_list = ATgetNext(arg_arg0_arg0_head_arg0_list);
                    if (ATgetInt((ATermInt)arg_arg0_arg0_head_arg0_head) == 13) {

                      if (ATgetType((ATerm)arg_arg0_arg0_head_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0_head_arg0_list) == ATfalse) {
                        arg_arg0_arg0_head_arg0_head = ATgetFirst(arg_arg0_arg0_head_arg0_list);
                        arg_arg0_arg0_head_arg0_list = ATgetNext(arg_arg0_arg0_head_arg0_list);
                        if (ATgetInt((ATermInt)arg_arg0_arg0_head_arg0_head) == 32) {

                          if (ATgetType((ATerm)arg_arg0_arg0_head_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0_head_arg0_list) == ATtrue) {
                            if (ATgetType((ATerm)arg_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0_list) == ATtrue) {
                              ATerm arg_arg0_arg1 = ATgetArgument(arg_arg0, 1);
                              /* checking for: lex */
                              if (ATgetType((ATerm)arg_arg0_arg1) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg1) == PERR_afun6) {
                                ATerm arg_arg0_arg1_arg0 = ATgetArgument(arg_arg0_arg1, 0);
                                /* checking for: layout */
                                if (ATgetType((ATerm)arg_arg0_arg1_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg1_arg0) == PERR_afun4) {
                                  ATerm arg_arg0_arg2 = ATgetArgument(arg_arg0, 2);
                                  /* checking for: attrs */
                                  if (ATgetType((ATerm)arg_arg0_arg2) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg2) == PERR_afun13) {
                                    ATerm arg_arg0_arg2_arg0 = ATgetArgument(arg_arg0_arg2, 0);
                                    if (ATgetType((ATerm)arg_arg0_arg2_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg2_arg0) == ATfalse) {
                                      ATerm arg_arg0_arg2_arg0_head;
                                      ATermList arg_arg0_arg2_arg0_list = (ATermList)arg_arg0_arg2_arg0;
                                      arg_arg0_arg2_arg0_head = ATgetFirst(arg_arg0_arg2_arg0_list);
                                      arg_arg0_arg2_arg0_list = ATgetNext(arg_arg0_arg2_arg0_list);
                                      /* checking for: term */
                                      if (ATgetType((ATerm)arg_arg0_arg2_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg2_arg0_head) == PERR_afun14) {
                                        ATerm arg_arg0_arg2_arg0_head_arg0 = ATgetArgument(arg_arg0_arg2_arg0_head, 0);
                                        /* checking for: cons */
                                        if (ATgetType((ATerm)arg_arg0_arg2_arg0_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg2_arg0_head_arg0) == PERR_afun15) {
                                          ATerm arg_arg0_arg2_arg0_head_arg0_arg0 = ATgetArgument(arg_arg0_arg2_arg0_head_arg0, 0);
                                          /* checking for: whitespace */
                                          if (ATgetType((ATerm)arg_arg0_arg2_arg0_head_arg0_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg2_arg0_head_arg0_arg0) == PERR_afun33) {
                                            if (ATgetType((ATerm)arg_arg0_arg2_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg2_arg0_list) == ATtrue) {
                                              ATerm arg_arg1 = ATgetArgument(arg, 1);
                                              if (ATgetType((ATerm)arg_arg1) == AT_LIST && ATisEmpty((ATermList)arg_arg1) == ATfalse) {
                                                ATerm arg_arg1_head;
                                                ATermList arg_arg1_list = (ATermList)arg_arg1;
                                                arg_arg1_head = ATgetFirst(arg_arg1_list);
                                                arg_arg1_list = ATgetNext(arg_arg1_list);
                                                if (ATgetType((ATerm)arg_arg1_head) == AT_INT) {
                                                  if (ATgetType((ATerm)arg_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_list) == ATtrue) {
                                                    return ATtrue;
                                                  }
                                                }
                                              }
                                            }
                                          }
                                        }
                                      }
                                    }
                                  }
                                }
                              }
                            }
                          }
                        }
                      }
                    }
                  }
                }
              }
            }
          }
        }
      }
    }
  }
  return ATfalse;
}

/**
 * Assert whether a PERR_LexLayout has a ch. 
 * \param[in] arg input PERR_LexLayout
 * \return ATtrue if the PERR_LexLayout had a ch, or ATfalse otherwise
 */
ATbool PERR_hasLexLayoutCh(PERR_LexLayout arg) {
  if (PERR_isLexLayoutWhitespace(arg)) {
    return ATtrue;
  }
  return ATfalse;
}

/**
 * Get the ch char of a PERR_LexLayout. Note that the precondition is that this PERR_LexLayout actually has a ch
 * \param[in] arg input PERR_LexLayout
 * \return the ch of #arg, if it exist or an undefined value if it does not
 */
char PERR_getLexLayoutCh(PERR_LexLayout arg) {
  
    return (char)PERR_charToByte((ATerm)ATgetFirst((ATermList)ATgetArgument((ATermAppl)arg, 1)));
}

/**
 * Set the ch of a PERR_LexLayout. The precondition being that this PERR_LexLayout actually has a ch
 * \param[in] arg input PERR_LexLayout
 * \param[in] ch new char to set in #arg
 * \return A new PERR_LexLayout with ch at the right place, or a core dump if #arg did not have a ch
 */
PERR_LexLayout PERR_setLexLayoutCh(PERR_LexLayout arg, char ch) {
  if (PERR_isLexLayoutWhitespace(arg)) {
    return (PERR_LexLayout)ATsetArgument((ATermAppl)arg, (ATerm)ATreplace((ATermList)ATgetArgument((ATermAppl)arg, 1), (ATerm)((ATerm) ((ATerm) PERR_byteToChar(ch))), 0), 1);
  }

  ATabort("LexLayout has no Ch: %t\n", arg);
  return (PERR_LexLayout)NULL;
}

/**
 * Assert whether a PERR_Start is any of the valid alternatives, or not. This analysis does not go any deeper than the top level
 * \param[in] arg input PERR_Start
 * \return ATtrue if #arg corresponds to the expected signature, or ATfalse otherwise
 */
ATbool PERR_isValidStart(PERR_Start arg) {
  if (PERR_isStartSummary(arg)) {
    return ATtrue;
  }
  else if (PERR_isStartError(arg)) {
    return ATtrue;
  }
  return ATfalse;
}

/**
 * Assert whether a PERR_Start is a Summary by checking against the following ATerm pattern: parsetree(appl(prod([cf(opt(layout)),cf(sort("Summary")),cf(opt(layout))],sort("<START>"),no-attrs),[<"ws-before"("OptLayout")>,<"top-Summary"("Summary")>,<"ws-after"("OptLayout")>]),<amb-cnt(int)>). May not be used to assert correctness of the PERR_Start
 * \param[in] arg input PERR_Start
 * \return ATtrue if #arg corresponds to the signature of a Summary, or ATfalse otherwise
 */
inline ATbool PERR_isStartSummary(PERR_Start arg){
  /* checking for: parsetree */
  if (ATgetType((ATerm)arg) == AT_APPL && ATgetAFun((ATermAppl)arg) == PERR_afun34) {
    ATerm arg_arg0 = ATgetArgument(arg, 0);
    /* checking for: appl */
    if (ATgetType((ATerm)arg_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0) == PERR_afun0) {
      ATerm arg_arg0_arg0 = ATgetArgument(arg_arg0, 0);
      /* checking for: prod */
      if (ATgetType((ATerm)arg_arg0_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0) == PERR_afun1) {
        ATerm arg_arg0_arg0_arg0 = ATgetArgument(arg_arg0_arg0, 0);
        if (ATgetType((ATerm)arg_arg0_arg0_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0_arg0) == ATfalse) {
          ATerm arg_arg0_arg0_arg0_head;
          ATermList arg_arg0_arg0_arg0_list = (ATermList)arg_arg0_arg0_arg0;
          arg_arg0_arg0_arg0_head = ATgetFirst(arg_arg0_arg0_arg0_list);
          arg_arg0_arg0_arg0_list = ATgetNext(arg_arg0_arg0_arg0_list);
          /* checking for: cf */
          if (ATgetType((ATerm)arg_arg0_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_arg0_head) == PERR_afun2) {
            ATerm arg_arg0_arg0_arg0_head_arg0 = ATgetArgument(arg_arg0_arg0_arg0_head, 0);
            /* checking for: opt */
            if (ATgetType((ATerm)arg_arg0_arg0_arg0_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_arg0_head_arg0) == PERR_afun3) {
              ATerm arg_arg0_arg0_arg0_head_arg0_arg0 = ATgetArgument(arg_arg0_arg0_arg0_head_arg0, 0);
              /* checking for: layout */
              if (ATgetType((ATerm)arg_arg0_arg0_arg0_head_arg0_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_arg0_head_arg0_arg0) == PERR_afun4) {
                if (ATgetType((ATerm)arg_arg0_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0_arg0_list) == ATfalse) {
                  arg_arg0_arg0_arg0_head = ATgetFirst(arg_arg0_arg0_arg0_list);
                  arg_arg0_arg0_arg0_list = ATgetNext(arg_arg0_arg0_arg0_list);
                  /* checking for: cf */
                  if (ATgetType((ATerm)arg_arg0_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_arg0_head) == PERR_afun2) {
                    ATerm arg_arg0_arg0_arg0_head_arg0 = ATgetArgument(arg_arg0_arg0_arg0_head, 0);
                    /* checking for: sort */
                    if (ATgetType((ATerm)arg_arg0_arg0_arg0_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_arg0_head_arg0) == PERR_afun11) {
                      ATerm arg_arg0_arg0_arg0_head_arg0_arg0 = ATgetArgument(arg_arg0_arg0_arg0_head_arg0, 0);
                      /* checking for: Summary */
                      if (ATgetType((ATerm)arg_arg0_arg0_arg0_head_arg0_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_arg0_head_arg0_arg0) == PERR_afun35) {
                        if (ATgetType((ATerm)arg_arg0_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0_arg0_list) == ATfalse) {
                          arg_arg0_arg0_arg0_head = ATgetFirst(arg_arg0_arg0_arg0_list);
                          arg_arg0_arg0_arg0_list = ATgetNext(arg_arg0_arg0_arg0_list);
                          /* checking for: cf */
                          if (ATgetType((ATerm)arg_arg0_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_arg0_head) == PERR_afun2) {
                            ATerm arg_arg0_arg0_arg0_head_arg0 = ATgetArgument(arg_arg0_arg0_arg0_head, 0);
                            /* checking for: opt */
                            if (ATgetType((ATerm)arg_arg0_arg0_arg0_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_arg0_head_arg0) == PERR_afun3) {
                              ATerm arg_arg0_arg0_arg0_head_arg0_arg0 = ATgetArgument(arg_arg0_arg0_arg0_head_arg0, 0);
                              /* checking for: layout */
                              if (ATgetType((ATerm)arg_arg0_arg0_arg0_head_arg0_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_arg0_head_arg0_arg0) == PERR_afun4) {
                                if (ATgetType((ATerm)arg_arg0_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0_arg0_list) == ATtrue) {
                                  ATerm arg_arg0_arg0_arg1 = ATgetArgument(arg_arg0_arg0, 1);
                                  /* checking for: sort */
                                  if (ATgetType((ATerm)arg_arg0_arg0_arg1) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_arg1) == PERR_afun11) {
                                    ATerm arg_arg0_arg0_arg1_arg0 = ATgetArgument(arg_arg0_arg0_arg1, 0);
                                    /* checking for: <START> */
                                    if (ATgetType((ATerm)arg_arg0_arg0_arg1_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_arg1_arg0) == PERR_afun36) {
                                      ATerm arg_arg0_arg0_arg2 = ATgetArgument(arg_arg0_arg0, 2);
                                      /* checking for: no-attrs */
                                      if (ATgetType((ATerm)arg_arg0_arg0_arg2) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_arg2) == PERR_afun5) {
                                        ATerm arg_arg0_arg1 = ATgetArgument(arg_arg0, 1);
                                        if (ATgetType((ATerm)arg_arg0_arg1) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg1) == ATfalse) {
                                          ATermList arg_arg0_arg1_list = (ATermList)arg_arg0_arg1;
                                          arg_arg0_arg1_list = ATgetNext(arg_arg0_arg1_list);
                                          if (ATgetType((ATerm)arg_arg0_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg1_list) == ATfalse) {
                                            arg_arg0_arg1_list = ATgetNext(arg_arg0_arg1_list);
                                            if (ATgetType((ATerm)arg_arg0_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg1_list) == ATfalse) {
                                              arg_arg0_arg1_list = ATgetNext(arg_arg0_arg1_list);
                                              if (ATgetType((ATerm)arg_arg0_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg1_list) == ATtrue) {
                                                ATerm arg_arg1 = ATgetArgument(arg, 1);
                                                if (ATgetType((ATerm)arg_arg1) == AT_INT) {
                                                  return ATtrue;
                                                }
                                              }
                                            }
                                          }
                                        }
                                      }
                                    }
                                  }
                                }
                              }
                            }
                          }
                        }
                      }
                    }
                  }
                }
              }
            }
          }
        }
      }
    }
  }
  return ATfalse;
}

/**
 * Assert whether a PERR_Start is a Error by checking against the following ATerm pattern: parsetree(appl(prod([cf(opt(layout)),cf(sort("Error")),cf(opt(layout))],sort("<START>"),no-attrs),[<"ws-before"("OptLayout")>,<"top-Error"("Error")>,<"ws-after"("OptLayout")>]),<amb-cnt(int)>). May not be used to assert correctness of the PERR_Start
 * \param[in] arg input PERR_Start
 * \return ATtrue if #arg corresponds to the signature of a Error, or ATfalse otherwise
 */
inline ATbool PERR_isStartError(PERR_Start arg){
  /* checking for: parsetree */
  if (ATgetType((ATerm)arg) == AT_APPL && ATgetAFun((ATermAppl)arg) == PERR_afun34) {
    ATerm arg_arg0 = ATgetArgument(arg, 0);
    /* checking for: appl */
    if (ATgetType((ATerm)arg_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0) == PERR_afun0) {
      ATerm arg_arg0_arg0 = ATgetArgument(arg_arg0, 0);
      /* checking for: prod */
      if (ATgetType((ATerm)arg_arg0_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0) == PERR_afun1) {
        ATerm arg_arg0_arg0_arg0 = ATgetArgument(arg_arg0_arg0, 0);
        if (ATgetType((ATerm)arg_arg0_arg0_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0_arg0) == ATfalse) {
          ATerm arg_arg0_arg0_arg0_head;
          ATermList arg_arg0_arg0_arg0_list = (ATermList)arg_arg0_arg0_arg0;
          arg_arg0_arg0_arg0_head = ATgetFirst(arg_arg0_arg0_arg0_list);
          arg_arg0_arg0_arg0_list = ATgetNext(arg_arg0_arg0_arg0_list);
          /* checking for: cf */
          if (ATgetType((ATerm)arg_arg0_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_arg0_head) == PERR_afun2) {
            ATerm arg_arg0_arg0_arg0_head_arg0 = ATgetArgument(arg_arg0_arg0_arg0_head, 0);
            /* checking for: opt */
            if (ATgetType((ATerm)arg_arg0_arg0_arg0_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_arg0_head_arg0) == PERR_afun3) {
              ATerm arg_arg0_arg0_arg0_head_arg0_arg0 = ATgetArgument(arg_arg0_arg0_arg0_head_arg0, 0);
              /* checking for: layout */
              if (ATgetType((ATerm)arg_arg0_arg0_arg0_head_arg0_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_arg0_head_arg0_arg0) == PERR_afun4) {
                if (ATgetType((ATerm)arg_arg0_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0_arg0_list) == ATfalse) {
                  arg_arg0_arg0_arg0_head = ATgetFirst(arg_arg0_arg0_arg0_list);
                  arg_arg0_arg0_arg0_list = ATgetNext(arg_arg0_arg0_arg0_list);
                  /* checking for: cf */
                  if (ATgetType((ATerm)arg_arg0_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_arg0_head) == PERR_afun2) {
                    ATerm arg_arg0_arg0_arg0_head_arg0 = ATgetArgument(arg_arg0_arg0_arg0_head, 0);
                    /* checking for: sort */
                    if (ATgetType((ATerm)arg_arg0_arg0_arg0_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_arg0_head_arg0) == PERR_afun11) {
                      ATerm arg_arg0_arg0_arg0_head_arg0_arg0 = ATgetArgument(arg_arg0_arg0_arg0_head_arg0, 0);
                      /* checking for: Error */
                      if (ATgetType((ATerm)arg_arg0_arg0_arg0_head_arg0_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_arg0_head_arg0_arg0) == PERR_afun37) {
                        if (ATgetType((ATerm)arg_arg0_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0_arg0_list) == ATfalse) {
                          arg_arg0_arg0_arg0_head = ATgetFirst(arg_arg0_arg0_arg0_list);
                          arg_arg0_arg0_arg0_list = ATgetNext(arg_arg0_arg0_arg0_list);
                          /* checking for: cf */
                          if (ATgetType((ATerm)arg_arg0_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_arg0_head) == PERR_afun2) {
                            ATerm arg_arg0_arg0_arg0_head_arg0 = ATgetArgument(arg_arg0_arg0_arg0_head, 0);
                            /* checking for: opt */
                            if (ATgetType((ATerm)arg_arg0_arg0_arg0_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_arg0_head_arg0) == PERR_afun3) {
                              ATerm arg_arg0_arg0_arg0_head_arg0_arg0 = ATgetArgument(arg_arg0_arg0_arg0_head_arg0, 0);
                              /* checking for: layout */
                              if (ATgetType((ATerm)arg_arg0_arg0_arg0_head_arg0_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_arg0_head_arg0_arg0) == PERR_afun4) {
                                if (ATgetType((ATerm)arg_arg0_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0_arg0_list) == ATtrue) {
                                  ATerm arg_arg0_arg0_arg1 = ATgetArgument(arg_arg0_arg0, 1);
                                  /* checking for: sort */
                                  if (ATgetType((ATerm)arg_arg0_arg0_arg1) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_arg1) == PERR_afun11) {
                                    ATerm arg_arg0_arg0_arg1_arg0 = ATgetArgument(arg_arg0_arg0_arg1, 0);
                                    /* checking for: <START> */
                                    if (ATgetType((ATerm)arg_arg0_arg0_arg1_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_arg1_arg0) == PERR_afun36) {
                                      ATerm arg_arg0_arg0_arg2 = ATgetArgument(arg_arg0_arg0, 2);
                                      /* checking for: no-attrs */
                                      if (ATgetType((ATerm)arg_arg0_arg0_arg2) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_arg2) == PERR_afun5) {
                                        ATerm arg_arg0_arg1 = ATgetArgument(arg_arg0, 1);
                                        if (ATgetType((ATerm)arg_arg0_arg1) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg1) == ATfalse) {
                                          ATermList arg_arg0_arg1_list = (ATermList)arg_arg0_arg1;
                                          arg_arg0_arg1_list = ATgetNext(arg_arg0_arg1_list);
                                          if (ATgetType((ATerm)arg_arg0_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg1_list) == ATfalse) {
                                            arg_arg0_arg1_list = ATgetNext(arg_arg0_arg1_list);
                                            if (ATgetType((ATerm)arg_arg0_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg1_list) == ATfalse) {
                                              arg_arg0_arg1_list = ATgetNext(arg_arg0_arg1_list);
                                              if (ATgetType((ATerm)arg_arg0_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg1_list) == ATtrue) {
                                                ATerm arg_arg1 = ATgetArgument(arg, 1);
                                                if (ATgetType((ATerm)arg_arg1) == AT_INT) {
                                                  return ATtrue;
                                                }
                                              }
                                            }
                                          }
                                        }
                                      }
                                    }
                                  }
                                }
                              }
                            }
                          }
                        }
                      }
                    }
                  }
                }
              }
            }
          }
        }
      }
    }
  }
  return ATfalse;
}

/**
 * Assert whether a PERR_Start has a ws-before. 
 * \param[in] arg input PERR_Start
 * \return ATtrue if the PERR_Start had a ws-before, or ATfalse otherwise
 */
ATbool PERR_hasStartWsBefore(PERR_Start arg) {
  if (PERR_isStartSummary(arg)) {
    return ATtrue;
  }
  else if (PERR_isStartError(arg)) {
    return ATtrue;
  }
  return ATfalse;
}

/**
 * Assert whether a PERR_Start has a top-Summary. 
 * \param[in] arg input PERR_Start
 * \return ATtrue if the PERR_Start had a top-Summary, or ATfalse otherwise
 */
ATbool PERR_hasStartTopSummary(PERR_Start arg) {
  if (PERR_isStartSummary(arg)) {
    return ATtrue;
  }
  return ATfalse;
}

/**
 * Assert whether a PERR_Start has a ws-after. 
 * \param[in] arg input PERR_Start
 * \return ATtrue if the PERR_Start had a ws-after, or ATfalse otherwise
 */
ATbool PERR_hasStartWsAfter(PERR_Start arg) {
  if (PERR_isStartSummary(arg)) {
    return ATtrue;
  }
  else if (PERR_isStartError(arg)) {
    return ATtrue;
  }
  return ATfalse;
}

/**
 * Assert whether a PERR_Start has a amb-cnt. 
 * \param[in] arg input PERR_Start
 * \return ATtrue if the PERR_Start had a amb-cnt, or ATfalse otherwise
 */
ATbool PERR_hasStartAmbCnt(PERR_Start arg) {
  if (PERR_isStartSummary(arg)) {
    return ATtrue;
  }
  else if (PERR_isStartError(arg)) {
    return ATtrue;
  }
  return ATfalse;
}

/**
 * Assert whether a PERR_Start has a top-Error. 
 * \param[in] arg input PERR_Start
 * \return ATtrue if the PERR_Start had a top-Error, or ATfalse otherwise
 */
ATbool PERR_hasStartTopError(PERR_Start arg) {
  if (PERR_isStartError(arg)) {
    return ATtrue;
  }
  return ATfalse;
}

/**
 * Get the ws-before PERR_OptLayout of a PERR_Start. Note that the precondition is that this PERR_Start actually has a ws-before
 * \param[in] arg input PERR_Start
 * \return the ws-before of #arg, if it exist or an undefined value if it does not
 */
PERR_OptLayout PERR_getStartWsBefore(PERR_Start arg) {
  if (PERR_isStartSummary(arg)) {
    return (PERR_OptLayout)ATgetFirst((ATermList)ATgetArgument((ATermAppl)ATgetArgument((ATermAppl)arg, 0), 1));
  }
  else 
    return (PERR_OptLayout)ATgetFirst((ATermList)ATgetArgument((ATermAppl)ATgetArgument((ATermAppl)arg, 0), 1));
}

/**
 * Get the top-Summary PERR_Summary of a PERR_Start. Note that the precondition is that this PERR_Start actually has a top-Summary
 * \param[in] arg input PERR_Start
 * \return the top-Summary of #arg, if it exist or an undefined value if it does not
 */
PERR_Summary PERR_getStartTopSummary(PERR_Start arg) {
  
    return (PERR_Summary)ATelementAt((ATermList)ATgetArgument((ATermAppl)ATgetArgument((ATermAppl)arg, 0), 1), 1);
}

/**
 * Get the ws-after PERR_OptLayout of a PERR_Start. Note that the precondition is that this PERR_Start actually has a ws-after
 * \param[in] arg input PERR_Start
 * \return the ws-after of #arg, if it exist or an undefined value if it does not
 */
PERR_OptLayout PERR_getStartWsAfter(PERR_Start arg) {
  if (PERR_isStartSummary(arg)) {
    return (PERR_OptLayout)ATelementAt((ATermList)ATgetArgument((ATermAppl)ATgetArgument((ATermAppl)arg, 0), 1), 2);
  }
  else 
    return (PERR_OptLayout)ATelementAt((ATermList)ATgetArgument((ATermAppl)ATgetArgument((ATermAppl)arg, 0), 1), 2);
}

/**
 * Get the amb-cnt int of a PERR_Start. Note that the precondition is that this PERR_Start actually has a amb-cnt
 * \param[in] arg input PERR_Start
 * \return the amb-cnt of #arg, if it exist or an undefined value if it does not
 */
int PERR_getStartAmbCnt(PERR_Start arg) {
  if (PERR_isStartSummary(arg)) {
    return (int)ATgetInt((ATermInt) ATgetArgument((ATermAppl)arg, 1));
  }
  else 
    return (int)ATgetInt((ATermInt) ATgetArgument((ATermAppl)arg, 1));
}

/**
 * Get the top-Error PERR_Error of a PERR_Start. Note that the precondition is that this PERR_Start actually has a top-Error
 * \param[in] arg input PERR_Start
 * \return the top-Error of #arg, if it exist or an undefined value if it does not
 */
PERR_Error PERR_getStartTopError(PERR_Start arg) {
  
    return (PERR_Error)ATelementAt((ATermList)ATgetArgument((ATermAppl)ATgetArgument((ATermAppl)arg, 0), 1), 1);
}

/**
 * Set the ws-before of a PERR_Start. The precondition being that this PERR_Start actually has a ws-before
 * \param[in] arg input PERR_Start
 * \param[in] wsBefore new PERR_OptLayout to set in #arg
 * \return A new PERR_Start with wsBefore at the right place, or a core dump if #arg did not have a wsBefore
 */
PERR_Start PERR_setStartWsBefore(PERR_Start arg, PERR_OptLayout wsBefore) {
  if (PERR_isStartSummary(arg)) {
    return (PERR_Start)ATsetArgument((ATermAppl)arg, (ATerm)ATsetArgument((ATermAppl)ATgetArgument((ATermAppl)arg, 0), (ATerm)ATreplace((ATermList)ATgetArgument((ATermAppl)ATgetArgument((ATermAppl)arg, 0), 1), (ATerm)((ATerm) wsBefore), 0), 1), 0);
  }
  else if (PERR_isStartError(arg)) {
    return (PERR_Start)ATsetArgument((ATermAppl)arg, (ATerm)ATsetArgument((ATermAppl)ATgetArgument((ATermAppl)arg, 0), (ATerm)ATreplace((ATermList)ATgetArgument((ATermAppl)ATgetArgument((ATermAppl)arg, 0), 1), (ATerm)((ATerm) wsBefore), 0), 1), 0);
  }

  ATabort("Start has no WsBefore: %t\n", arg);
  return (PERR_Start)NULL;
}

/**
 * Set the top-Summary of a PERR_Start. The precondition being that this PERR_Start actually has a top-Summary
 * \param[in] arg input PERR_Start
 * \param[in] topSummary new PERR_Summary to set in #arg
 * \return A new PERR_Start with topSummary at the right place, or a core dump if #arg did not have a topSummary
 */
PERR_Start PERR_setStartTopSummary(PERR_Start arg, PERR_Summary topSummary) {
  if (PERR_isStartSummary(arg)) {
    return (PERR_Start)ATsetArgument((ATermAppl)arg, (ATerm)ATsetArgument((ATermAppl)ATgetArgument((ATermAppl)arg, 0), (ATerm)ATreplace((ATermList)ATgetArgument((ATermAppl)ATgetArgument((ATermAppl)arg, 0), 1), (ATerm)((ATerm) topSummary), 1), 1), 0);
  }

  ATabort("Start has no TopSummary: %t\n", arg);
  return (PERR_Start)NULL;
}

/**
 * Set the ws-after of a PERR_Start. The precondition being that this PERR_Start actually has a ws-after
 * \param[in] arg input PERR_Start
 * \param[in] wsAfter new PERR_OptLayout to set in #arg
 * \return A new PERR_Start with wsAfter at the right place, or a core dump if #arg did not have a wsAfter
 */
PERR_Start PERR_setStartWsAfter(PERR_Start arg, PERR_OptLayout wsAfter) {
  if (PERR_isStartSummary(arg)) {
    return (PERR_Start)ATsetArgument((ATermAppl)arg, (ATerm)ATsetArgument((ATermAppl)ATgetArgument((ATermAppl)arg, 0), (ATerm)ATreplace((ATermList)ATgetArgument((ATermAppl)ATgetArgument((ATermAppl)arg, 0), 1), (ATerm)((ATerm) wsAfter), 2), 1), 0);
  }
  else if (PERR_isStartError(arg)) {
    return (PERR_Start)ATsetArgument((ATermAppl)arg, (ATerm)ATsetArgument((ATermAppl)ATgetArgument((ATermAppl)arg, 0), (ATerm)ATreplace((ATermList)ATgetArgument((ATermAppl)ATgetArgument((ATermAppl)arg, 0), 1), (ATerm)((ATerm) wsAfter), 2), 1), 0);
  }

  ATabort("Start has no WsAfter: %t\n", arg);
  return (PERR_Start)NULL;
}

/**
 * Set the amb-cnt of a PERR_Start. The precondition being that this PERR_Start actually has a amb-cnt
 * \param[in] arg input PERR_Start
 * \param[in] ambCnt new int to set in #arg
 * \return A new PERR_Start with ambCnt at the right place, or a core dump if #arg did not have a ambCnt
 */
PERR_Start PERR_setStartAmbCnt(PERR_Start arg, int ambCnt) {
  if (PERR_isStartSummary(arg)) {
    return (PERR_Start)ATsetArgument((ATermAppl)arg, (ATerm)((ATerm) (ATerm) ATmakeInt(ambCnt)), 1);
  }
  else if (PERR_isStartError(arg)) {
    return (PERR_Start)ATsetArgument((ATermAppl)arg, (ATerm)((ATerm) (ATerm) ATmakeInt(ambCnt)), 1);
  }

  ATabort("Start has no AmbCnt: %t\n", arg);
  return (PERR_Start)NULL;
}

/**
 * Set the top-Error of a PERR_Start. The precondition being that this PERR_Start actually has a top-Error
 * \param[in] arg input PERR_Start
 * \param[in] topError new PERR_Error to set in #arg
 * \return A new PERR_Start with topError at the right place, or a core dump if #arg did not have a topError
 */
PERR_Start PERR_setStartTopError(PERR_Start arg, PERR_Error topError) {
  if (PERR_isStartError(arg)) {
    return (PERR_Start)ATsetArgument((ATermAppl)arg, (ATerm)ATsetArgument((ATermAppl)ATgetArgument((ATermAppl)arg, 0), (ATerm)ATreplace((ATermList)ATgetArgument((ATermAppl)ATgetArgument((ATermAppl)arg, 0), 1), (ATerm)((ATerm) topError), 1), 1), 0);
  }

  ATabort("Start has no TopError: %t\n", arg);
  return (PERR_Start)NULL;
}

/**
 * Assert whether a PERR_Subject is any of the valid alternatives, or not. This analysis does not go any deeper than the top level
 * \param[in] arg input PERR_Subject
 * \return ATtrue if #arg corresponds to the expected signature, or ATfalse otherwise
 */
ATbool PERR_isValidSubject(PERR_Subject arg) {
  if (PERR_isSubjectSubject(arg)) {
    return ATtrue;
  }
  else if (PERR_isSubjectLocalized(arg)) {
    return ATtrue;
  }
  return ATfalse;
}

/**
 * Assert whether a PERR_Subject is a subject by checking against the following ATerm pattern: appl(prod([lit("subject"),cf(opt(layout)),lit("("),cf(opt(layout)),cf(sort("StrCon")),cf(opt(layout)),lit(")")],cf(sort("Subject")),attrs([term(cons("subject"))])),[appl(prod([char-class([115]),char-class([117]),char-class([98]),char-class([106]),char-class([101]),char-class([99]),char-class([116])],lit("subject"),no-attrs),[115,117,98,106,101,99,116]),<"ws-after-subject"("OptLayout")>,appl(prod([char-class([40])],lit("("),no-attrs),[40]),<"ws-after-("("OptLayout")>,<"description"("StrCon")>,<"ws-after-description"("OptLayout")>,appl(prod([char-class([41])],lit(")"),no-attrs),[41])]). May not be used to assert correctness of the PERR_Subject
 * \param[in] arg input PERR_Subject
 * \return ATtrue if #arg corresponds to the signature of a subject, or ATfalse otherwise
 */
inline ATbool PERR_isSubjectSubject(PERR_Subject arg){
  /* checking for: appl */
  if (ATgetType((ATerm)arg) == AT_APPL && ATgetAFun((ATermAppl)arg) == PERR_afun0) {
    ATerm arg_arg0 = ATgetArgument(arg, 0);
    /* checking for: prod */
    if (ATgetType((ATerm)arg_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0) == PERR_afun1) {
      ATerm arg_arg0_arg0 = ATgetArgument(arg_arg0, 0);
      if (ATgetType((ATerm)arg_arg0_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0) == ATfalse) {
        ATerm arg_arg0_arg0_head;
        ATermList arg_arg0_arg0_list = (ATermList)arg_arg0_arg0;
        arg_arg0_arg0_head = ATgetFirst(arg_arg0_arg0_list);
        arg_arg0_arg0_list = ATgetNext(arg_arg0_arg0_list);
        /* checking for: lit */
        if (ATgetType((ATerm)arg_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head) == PERR_afun9) {
          ATerm arg_arg0_arg0_head_arg0 = ATgetArgument(arg_arg0_arg0_head, 0);
          /* checking for: subject */
          if (ATgetType((ATerm)arg_arg0_arg0_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0) == PERR_afun40) {
            if (ATgetType((ATerm)arg_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0_list) == ATfalse) {
              arg_arg0_arg0_head = ATgetFirst(arg_arg0_arg0_list);
              arg_arg0_arg0_list = ATgetNext(arg_arg0_arg0_list);
              /* checking for: cf */
              if (ATgetType((ATerm)arg_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head) == PERR_afun2) {
                ATerm arg_arg0_arg0_head_arg0 = ATgetArgument(arg_arg0_arg0_head, 0);
                /* checking for: opt */
                if (ATgetType((ATerm)arg_arg0_arg0_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0) == PERR_afun3) {
                  ATerm arg_arg0_arg0_head_arg0_arg0 = ATgetArgument(arg_arg0_arg0_head_arg0, 0);
                  /* checking for: layout */
                  if (ATgetType((ATerm)arg_arg0_arg0_head_arg0_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0_arg0) == PERR_afun4) {
                    if (ATgetType((ATerm)arg_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0_list) == ATfalse) {
                      arg_arg0_arg0_head = ATgetFirst(arg_arg0_arg0_list);
                      arg_arg0_arg0_list = ATgetNext(arg_arg0_arg0_list);
                      /* checking for: lit */
                      if (ATgetType((ATerm)arg_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head) == PERR_afun9) {
                        ATerm arg_arg0_arg0_head_arg0 = ATgetArgument(arg_arg0_arg0_head, 0);
                        /* checking for: ( */
                        if (ATgetType((ATerm)arg_arg0_arg0_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0) == PERR_afun39) {
                          if (ATgetType((ATerm)arg_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0_list) == ATfalse) {
                            arg_arg0_arg0_head = ATgetFirst(arg_arg0_arg0_list);
                            arg_arg0_arg0_list = ATgetNext(arg_arg0_arg0_list);
                            /* checking for: cf */
                            if (ATgetType((ATerm)arg_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head) == PERR_afun2) {
                              ATerm arg_arg0_arg0_head_arg0 = ATgetArgument(arg_arg0_arg0_head, 0);
                              /* checking for: opt */
                              if (ATgetType((ATerm)arg_arg0_arg0_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0) == PERR_afun3) {
                                ATerm arg_arg0_arg0_head_arg0_arg0 = ATgetArgument(arg_arg0_arg0_head_arg0, 0);
                                /* checking for: layout */
                                if (ATgetType((ATerm)arg_arg0_arg0_head_arg0_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0_arg0) == PERR_afun4) {
                                  if (ATgetType((ATerm)arg_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0_list) == ATfalse) {
                                    arg_arg0_arg0_head = ATgetFirst(arg_arg0_arg0_list);
                                    arg_arg0_arg0_list = ATgetNext(arg_arg0_arg0_list);
                                    /* checking for: cf */
                                    if (ATgetType((ATerm)arg_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head) == PERR_afun2) {
                                      ATerm arg_arg0_arg0_head_arg0 = ATgetArgument(arg_arg0_arg0_head, 0);
                                      /* checking for: sort */
                                      if (ATgetType((ATerm)arg_arg0_arg0_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0) == PERR_afun11) {
                                        ATerm arg_arg0_arg0_head_arg0_arg0 = ATgetArgument(arg_arg0_arg0_head_arg0, 0);
                                        /* checking for: StrCon */
                                        if (ATgetType((ATerm)arg_arg0_arg0_head_arg0_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0_arg0) == PERR_afun29) {
                                          if (ATgetType((ATerm)arg_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0_list) == ATfalse) {
                                            arg_arg0_arg0_head = ATgetFirst(arg_arg0_arg0_list);
                                            arg_arg0_arg0_list = ATgetNext(arg_arg0_arg0_list);
                                            /* checking for: cf */
                                            if (ATgetType((ATerm)arg_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head) == PERR_afun2) {
                                              ATerm arg_arg0_arg0_head_arg0 = ATgetArgument(arg_arg0_arg0_head, 0);
                                              /* checking for: opt */
                                              if (ATgetType((ATerm)arg_arg0_arg0_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0) == PERR_afun3) {
                                                ATerm arg_arg0_arg0_head_arg0_arg0 = ATgetArgument(arg_arg0_arg0_head_arg0, 0);
                                                /* checking for: layout */
                                                if (ATgetType((ATerm)arg_arg0_arg0_head_arg0_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0_arg0) == PERR_afun4) {
                                                  if (ATgetType((ATerm)arg_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0_list) == ATfalse) {
                                                    arg_arg0_arg0_head = ATgetFirst(arg_arg0_arg0_list);
                                                    arg_arg0_arg0_list = ATgetNext(arg_arg0_arg0_list);
                                                    /* checking for: lit */
                                                    if (ATgetType((ATerm)arg_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head) == PERR_afun9) {
                                                      ATerm arg_arg0_arg0_head_arg0 = ATgetArgument(arg_arg0_arg0_head, 0);
                                                      /* checking for: ) */
                                                      if (ATgetType((ATerm)arg_arg0_arg0_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0) == PERR_afun38) {
                                                        if (ATgetType((ATerm)arg_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0_list) == ATtrue) {
                                                          ATerm arg_arg0_arg1 = ATgetArgument(arg_arg0, 1);
                                                          /* checking for: cf */
                                                          if (ATgetType((ATerm)arg_arg0_arg1) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg1) == PERR_afun2) {
                                                            ATerm arg_arg0_arg1_arg0 = ATgetArgument(arg_arg0_arg1, 0);
                                                            /* checking for: sort */
                                                            if (ATgetType((ATerm)arg_arg0_arg1_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg1_arg0) == PERR_afun11) {
                                                              ATerm arg_arg0_arg1_arg0_arg0 = ATgetArgument(arg_arg0_arg1_arg0, 0);
                                                              /* checking for: Subject */
                                                              if (ATgetType((ATerm)arg_arg0_arg1_arg0_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg1_arg0_arg0) == PERR_afun41) {
                                                                ATerm arg_arg0_arg2 = ATgetArgument(arg_arg0, 2);
                                                                /* checking for: attrs */
                                                                if (ATgetType((ATerm)arg_arg0_arg2) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg2) == PERR_afun13) {
                                                                  ATerm arg_arg0_arg2_arg0 = ATgetArgument(arg_arg0_arg2, 0);
                                                                  if (ATgetType((ATerm)arg_arg0_arg2_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg2_arg0) == ATfalse) {
                                                                    ATerm arg_arg0_arg2_arg0_head;
                                                                    ATermList arg_arg0_arg2_arg0_list = (ATermList)arg_arg0_arg2_arg0;
                                                                    arg_arg0_arg2_arg0_head = ATgetFirst(arg_arg0_arg2_arg0_list);
                                                                    arg_arg0_arg2_arg0_list = ATgetNext(arg_arg0_arg2_arg0_list);
                                                                    /* checking for: term */
                                                                    if (ATgetType((ATerm)arg_arg0_arg2_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg2_arg0_head) == PERR_afun14) {
                                                                      ATerm arg_arg0_arg2_arg0_head_arg0 = ATgetArgument(arg_arg0_arg2_arg0_head, 0);
                                                                      /* checking for: cons */
                                                                      if (ATgetType((ATerm)arg_arg0_arg2_arg0_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg2_arg0_head_arg0) == PERR_afun15) {
                                                                        ATerm arg_arg0_arg2_arg0_head_arg0_arg0 = ATgetArgument(arg_arg0_arg2_arg0_head_arg0, 0);
                                                                        /* checking for: subject */
                                                                        if (ATgetType((ATerm)arg_arg0_arg2_arg0_head_arg0_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg2_arg0_head_arg0_arg0) == PERR_afun40) {
                                                                          if (ATgetType((ATerm)arg_arg0_arg2_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg2_arg0_list) == ATtrue) {
                                                                            ATerm arg_arg1 = ATgetArgument(arg, 1);
                                                                            if (ATgetType((ATerm)arg_arg1) == AT_LIST && ATisEmpty((ATermList)arg_arg1) == ATfalse) {
                                                                              ATerm arg_arg1_head;
                                                                              ATermList arg_arg1_list = (ATermList)arg_arg1;
                                                                              arg_arg1_head = ATgetFirst(arg_arg1_list);
                                                                              arg_arg1_list = ATgetNext(arg_arg1_list);
                                                                              /* checking for: appl */
                                                                              if (ATgetType((ATerm)arg_arg1_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head) == PERR_afun0) {
                                                                                ATerm arg_arg1_head_arg0 = ATgetArgument(arg_arg1_head, 0);
                                                                                /* checking for: prod */
                                                                                if (ATgetType((ATerm)arg_arg1_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0) == PERR_afun1) {
                                                                                  ATerm arg_arg1_head_arg0_arg0 = ATgetArgument(arg_arg1_head_arg0, 0);
                                                                                  if (ATgetType((ATerm)arg_arg1_head_arg0_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0) == ATfalse) {
                                                                                    ATerm arg_arg1_head_arg0_arg0_head;
                                                                                    ATermList arg_arg1_head_arg0_arg0_list = (ATermList)arg_arg1_head_arg0_arg0;
                                                                                    arg_arg1_head_arg0_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_list);
                                                                                    arg_arg1_head_arg0_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_list);
                                                                                    /* checking for: char-class */
                                                                                    if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg0_head) == PERR_afun17) {
                                                                                      ATerm arg_arg1_head_arg0_arg0_head_arg0 = ATgetArgument(arg_arg1_head_arg0_arg0_head, 0);
                                                                                      if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0) == ATfalse) {
                                                                                        ATerm arg_arg1_head_arg0_arg0_head_arg0_head;
                                                                                        ATermList arg_arg1_head_arg0_arg0_head_arg0_list = (ATermList)arg_arg1_head_arg0_arg0_head_arg0;
                                                                                        arg_arg1_head_arg0_arg0_head_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                        arg_arg1_head_arg0_arg0_head_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                        if (ATgetInt((ATermInt)arg_arg1_head_arg0_arg0_head_arg0_head) == 115) {

                                                                                          if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0_list) == ATtrue) {
                                                                                            if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_list) == ATfalse) {
                                                                                              arg_arg1_head_arg0_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_list);
                                                                                              arg_arg1_head_arg0_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_list);
                                                                                              /* checking for: char-class */
                                                                                              if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg0_head) == PERR_afun17) {
                                                                                                ATerm arg_arg1_head_arg0_arg0_head_arg0 = ATgetArgument(arg_arg1_head_arg0_arg0_head, 0);
                                                                                                if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0) == ATfalse) {
                                                                                                  ATerm arg_arg1_head_arg0_arg0_head_arg0_head;
                                                                                                  ATermList arg_arg1_head_arg0_arg0_head_arg0_list = (ATermList)arg_arg1_head_arg0_arg0_head_arg0;
                                                                                                  arg_arg1_head_arg0_arg0_head_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                  arg_arg1_head_arg0_arg0_head_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                  if (ATgetInt((ATermInt)arg_arg1_head_arg0_arg0_head_arg0_head) == 117) {

                                                                                                    if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0_list) == ATtrue) {
                                                                                                      if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_list) == ATfalse) {
                                                                                                        arg_arg1_head_arg0_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_list);
                                                                                                        arg_arg1_head_arg0_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_list);
                                                                                                        /* checking for: char-class */
                                                                                                        if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg0_head) == PERR_afun17) {
                                                                                                          ATerm arg_arg1_head_arg0_arg0_head_arg0 = ATgetArgument(arg_arg1_head_arg0_arg0_head, 0);
                                                                                                          if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0) == ATfalse) {
                                                                                                            ATerm arg_arg1_head_arg0_arg0_head_arg0_head;
                                                                                                            ATermList arg_arg1_head_arg0_arg0_head_arg0_list = (ATermList)arg_arg1_head_arg0_arg0_head_arg0;
                                                                                                            arg_arg1_head_arg0_arg0_head_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                            arg_arg1_head_arg0_arg0_head_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                            if (ATgetInt((ATermInt)arg_arg1_head_arg0_arg0_head_arg0_head) == 98) {

                                                                                                              if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0_list) == ATtrue) {
                                                                                                                if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_list) == ATfalse) {
                                                                                                                  arg_arg1_head_arg0_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_list);
                                                                                                                  arg_arg1_head_arg0_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_list);
                                                                                                                  /* checking for: char-class */
                                                                                                                  if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg0_head) == PERR_afun17) {
                                                                                                                    ATerm arg_arg1_head_arg0_arg0_head_arg0 = ATgetArgument(arg_arg1_head_arg0_arg0_head, 0);
                                                                                                                    if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0) == ATfalse) {
                                                                                                                      ATerm arg_arg1_head_arg0_arg0_head_arg0_head;
                                                                                                                      ATermList arg_arg1_head_arg0_arg0_head_arg0_list = (ATermList)arg_arg1_head_arg0_arg0_head_arg0;
                                                                                                                      arg_arg1_head_arg0_arg0_head_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                                      arg_arg1_head_arg0_arg0_head_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                                      if (ATgetInt((ATermInt)arg_arg1_head_arg0_arg0_head_arg0_head) == 106) {

                                                                                                                        if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0_list) == ATtrue) {
                                                                                                                          if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_list) == ATfalse) {
                                                                                                                            arg_arg1_head_arg0_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_list);
                                                                                                                            arg_arg1_head_arg0_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_list);
                                                                                                                            /* checking for: char-class */
                                                                                                                            if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg0_head) == PERR_afun17) {
                                                                                                                              ATerm arg_arg1_head_arg0_arg0_head_arg0 = ATgetArgument(arg_arg1_head_arg0_arg0_head, 0);
                                                                                                                              if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0) == ATfalse) {
                                                                                                                                ATerm arg_arg1_head_arg0_arg0_head_arg0_head;
                                                                                                                                ATermList arg_arg1_head_arg0_arg0_head_arg0_list = (ATermList)arg_arg1_head_arg0_arg0_head_arg0;
                                                                                                                                arg_arg1_head_arg0_arg0_head_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                                                arg_arg1_head_arg0_arg0_head_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                                                if (ATgetInt((ATermInt)arg_arg1_head_arg0_arg0_head_arg0_head) == 101) {

                                                                                                                                  if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0_list) == ATtrue) {
                                                                                                                                    if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_list) == ATfalse) {
                                                                                                                                      arg_arg1_head_arg0_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_list);
                                                                                                                                      arg_arg1_head_arg0_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_list);
                                                                                                                                      /* checking for: char-class */
                                                                                                                                      if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg0_head) == PERR_afun17) {
                                                                                                                                        ATerm arg_arg1_head_arg0_arg0_head_arg0 = ATgetArgument(arg_arg1_head_arg0_arg0_head, 0);
                                                                                                                                        if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0) == ATfalse) {
                                                                                                                                          ATerm arg_arg1_head_arg0_arg0_head_arg0_head;
                                                                                                                                          ATermList arg_arg1_head_arg0_arg0_head_arg0_list = (ATermList)arg_arg1_head_arg0_arg0_head_arg0;
                                                                                                                                          arg_arg1_head_arg0_arg0_head_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                                                          arg_arg1_head_arg0_arg0_head_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                                                          if (ATgetInt((ATermInt)arg_arg1_head_arg0_arg0_head_arg0_head) == 99) {

                                                                                                                                            if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0_list) == ATtrue) {
                                                                                                                                              if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_list) == ATfalse) {
                                                                                                                                                arg_arg1_head_arg0_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_list);
                                                                                                                                                arg_arg1_head_arg0_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_list);
                                                                                                                                                /* checking for: char-class */
                                                                                                                                                if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg0_head) == PERR_afun17) {
                                                                                                                                                  ATerm arg_arg1_head_arg0_arg0_head_arg0 = ATgetArgument(arg_arg1_head_arg0_arg0_head, 0);
                                                                                                                                                  if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0) == ATfalse) {
                                                                                                                                                    ATerm arg_arg1_head_arg0_arg0_head_arg0_head;
                                                                                                                                                    ATermList arg_arg1_head_arg0_arg0_head_arg0_list = (ATermList)arg_arg1_head_arg0_arg0_head_arg0;
                                                                                                                                                    arg_arg1_head_arg0_arg0_head_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                                                                    arg_arg1_head_arg0_arg0_head_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                                                                    if (ATgetInt((ATermInt)arg_arg1_head_arg0_arg0_head_arg0_head) == 116) {

                                                                                                                                                      if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0_list) == ATtrue) {
                                                                                                                                                        if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_list) == ATtrue) {
                                                                                                                                                          ATerm arg_arg1_head_arg0_arg1 = ATgetArgument(arg_arg1_head_arg0, 1);
                                                                                                                                                          /* checking for: lit */
                                                                                                                                                          if (ATgetType((ATerm)arg_arg1_head_arg0_arg1) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg1) == PERR_afun9) {
                                                                                                                                                            ATerm arg_arg1_head_arg0_arg1_arg0 = ATgetArgument(arg_arg1_head_arg0_arg1, 0);
                                                                                                                                                            /* checking for: subject */
                                                                                                                                                            if (ATgetType((ATerm)arg_arg1_head_arg0_arg1_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg1_arg0) == PERR_afun40) {
                                                                                                                                                              ATerm arg_arg1_head_arg0_arg2 = ATgetArgument(arg_arg1_head_arg0, 2);
                                                                                                                                                              /* checking for: no-attrs */
                                                                                                                                                              if (ATgetType((ATerm)arg_arg1_head_arg0_arg2) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg2) == PERR_afun5) {
                                                                                                                                                                ATerm arg_arg1_head_arg1 = ATgetArgument(arg_arg1_head, 1);
                                                                                                                                                                if (ATgetType((ATerm)arg_arg1_head_arg1) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg1) == ATfalse) {
                                                                                                                                                                  ATerm arg_arg1_head_arg1_head;
                                                                                                                                                                  ATermList arg_arg1_head_arg1_list = (ATermList)arg_arg1_head_arg1;
                                                                                                                                                                  arg_arg1_head_arg1_head = ATgetFirst(arg_arg1_head_arg1_list);
                                                                                                                                                                  arg_arg1_head_arg1_list = ATgetNext(arg_arg1_head_arg1_list);
                                                                                                                                                                  if (ATgetInt((ATermInt)arg_arg1_head_arg1_head) == 115) {

                                                                                                                                                                    if (ATgetType((ATerm)arg_arg1_head_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg1_list) == ATfalse) {
                                                                                                                                                                      arg_arg1_head_arg1_head = ATgetFirst(arg_arg1_head_arg1_list);
                                                                                                                                                                      arg_arg1_head_arg1_list = ATgetNext(arg_arg1_head_arg1_list);
                                                                                                                                                                      if (ATgetInt((ATermInt)arg_arg1_head_arg1_head) == 117) {

                                                                                                                                                                        if (ATgetType((ATerm)arg_arg1_head_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg1_list) == ATfalse) {
                                                                                                                                                                          arg_arg1_head_arg1_head = ATgetFirst(arg_arg1_head_arg1_list);
                                                                                                                                                                          arg_arg1_head_arg1_list = ATgetNext(arg_arg1_head_arg1_list);
                                                                                                                                                                          if (ATgetInt((ATermInt)arg_arg1_head_arg1_head) == 98) {

                                                                                                                                                                            if (ATgetType((ATerm)arg_arg1_head_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg1_list) == ATfalse) {
                                                                                                                                                                              arg_arg1_head_arg1_head = ATgetFirst(arg_arg1_head_arg1_list);
                                                                                                                                                                              arg_arg1_head_arg1_list = ATgetNext(arg_arg1_head_arg1_list);
                                                                                                                                                                              if (ATgetInt((ATermInt)arg_arg1_head_arg1_head) == 106) {

                                                                                                                                                                                if (ATgetType((ATerm)arg_arg1_head_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg1_list) == ATfalse) {
                                                                                                                                                                                  arg_arg1_head_arg1_head = ATgetFirst(arg_arg1_head_arg1_list);
                                                                                                                                                                                  arg_arg1_head_arg1_list = ATgetNext(arg_arg1_head_arg1_list);
                                                                                                                                                                                  if (ATgetInt((ATermInt)arg_arg1_head_arg1_head) == 101) {

                                                                                                                                                                                    if (ATgetType((ATerm)arg_arg1_head_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg1_list) == ATfalse) {
                                                                                                                                                                                      arg_arg1_head_arg1_head = ATgetFirst(arg_arg1_head_arg1_list);
                                                                                                                                                                                      arg_arg1_head_arg1_list = ATgetNext(arg_arg1_head_arg1_list);
                                                                                                                                                                                      if (ATgetInt((ATermInt)arg_arg1_head_arg1_head) == 99) {

                                                                                                                                                                                        if (ATgetType((ATerm)arg_arg1_head_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg1_list) == ATfalse) {
                                                                                                                                                                                          arg_arg1_head_arg1_head = ATgetFirst(arg_arg1_head_arg1_list);
                                                                                                                                                                                          arg_arg1_head_arg1_list = ATgetNext(arg_arg1_head_arg1_list);
                                                                                                                                                                                          if (ATgetInt((ATermInt)arg_arg1_head_arg1_head) == 116) {

                                                                                                                                                                                            if (ATgetType((ATerm)arg_arg1_head_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg1_list) == ATtrue) {
                                                                                                                                                                                              if (ATgetType((ATerm)arg_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_list) == ATfalse) {
                                                                                                                                                                                                arg_arg1_list = ATgetNext(arg_arg1_list);
                                                                                                                                                                                                if (ATgetType((ATerm)arg_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_list) == ATfalse) {
                                                                                                                                                                                                  arg_arg1_head = ATgetFirst(arg_arg1_list);
                                                                                                                                                                                                  arg_arg1_list = ATgetNext(arg_arg1_list);
                                                                                                                                                                                                  /* checking for: appl */
                                                                                                                                                                                                  if (ATgetType((ATerm)arg_arg1_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head) == PERR_afun0) {
                                                                                                                                                                                                    ATerm arg_arg1_head_arg0 = ATgetArgument(arg_arg1_head, 0);
                                                                                                                                                                                                    /* checking for: prod */
                                                                                                                                                                                                    if (ATgetType((ATerm)arg_arg1_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0) == PERR_afun1) {
                                                                                                                                                                                                      ATerm arg_arg1_head_arg0_arg0 = ATgetArgument(arg_arg1_head_arg0, 0);
                                                                                                                                                                                                      if (ATgetType((ATerm)arg_arg1_head_arg0_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0) == ATfalse) {
                                                                                                                                                                                                        ATerm arg_arg1_head_arg0_arg0_head;
                                                                                                                                                                                                        ATermList arg_arg1_head_arg0_arg0_list = (ATermList)arg_arg1_head_arg0_arg0;
                                                                                                                                                                                                        arg_arg1_head_arg0_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_list);
                                                                                                                                                                                                        arg_arg1_head_arg0_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_list);
                                                                                                                                                                                                        /* checking for: char-class */
                                                                                                                                                                                                        if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg0_head) == PERR_afun17) {
                                                                                                                                                                                                          ATerm arg_arg1_head_arg0_arg0_head_arg0 = ATgetArgument(arg_arg1_head_arg0_arg0_head, 0);
                                                                                                                                                                                                          if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0) == ATfalse) {
                                                                                                                                                                                                            ATerm arg_arg1_head_arg0_arg0_head_arg0_head;
                                                                                                                                                                                                            ATermList arg_arg1_head_arg0_arg0_head_arg0_list = (ATermList)arg_arg1_head_arg0_arg0_head_arg0;
                                                                                                                                                                                                            arg_arg1_head_arg0_arg0_head_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                                                                                                                            arg_arg1_head_arg0_arg0_head_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                                                                                                                            if (ATgetInt((ATermInt)arg_arg1_head_arg0_arg0_head_arg0_head) == 40) {

                                                                                                                                                                                                              if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0_list) == ATtrue) {
                                                                                                                                                                                                                if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_list) == ATtrue) {
                                                                                                                                                                                                                  ATerm arg_arg1_head_arg0_arg1 = ATgetArgument(arg_arg1_head_arg0, 1);
                                                                                                                                                                                                                  /* checking for: lit */
                                                                                                                                                                                                                  if (ATgetType((ATerm)arg_arg1_head_arg0_arg1) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg1) == PERR_afun9) {
                                                                                                                                                                                                                    ATerm arg_arg1_head_arg0_arg1_arg0 = ATgetArgument(arg_arg1_head_arg0_arg1, 0);
                                                                                                                                                                                                                    /* checking for: ( */
                                                                                                                                                                                                                    if (ATgetType((ATerm)arg_arg1_head_arg0_arg1_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg1_arg0) == PERR_afun39) {
                                                                                                                                                                                                                      ATerm arg_arg1_head_arg0_arg2 = ATgetArgument(arg_arg1_head_arg0, 2);
                                                                                                                                                                                                                      /* checking for: no-attrs */
                                                                                                                                                                                                                      if (ATgetType((ATerm)arg_arg1_head_arg0_arg2) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg2) == PERR_afun5) {
                                                                                                                                                                                                                        ATerm arg_arg1_head_arg1 = ATgetArgument(arg_arg1_head, 1);
                                                                                                                                                                                                                        if (ATgetType((ATerm)arg_arg1_head_arg1) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg1) == ATfalse) {
                                                                                                                                                                                                                          ATerm arg_arg1_head_arg1_head;
                                                                                                                                                                                                                          ATermList arg_arg1_head_arg1_list = (ATermList)arg_arg1_head_arg1;
                                                                                                                                                                                                                          arg_arg1_head_arg1_head = ATgetFirst(arg_arg1_head_arg1_list);
                                                                                                                                                                                                                          arg_arg1_head_arg1_list = ATgetNext(arg_arg1_head_arg1_list);
                                                                                                                                                                                                                          if (ATgetInt((ATermInt)arg_arg1_head_arg1_head) == 40) {

                                                                                                                                                                                                                            if (ATgetType((ATerm)arg_arg1_head_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg1_list) == ATtrue) {
                                                                                                                                                                                                                              if (ATgetType((ATerm)arg_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_list) == ATfalse) {
                                                                                                                                                                                                                                arg_arg1_list = ATgetNext(arg_arg1_list);
                                                                                                                                                                                                                                if (ATgetType((ATerm)arg_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_list) == ATfalse) {
                                                                                                                                                                                                                                  arg_arg1_list = ATgetNext(arg_arg1_list);
                                                                                                                                                                                                                                  if (ATgetType((ATerm)arg_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_list) == ATfalse) {
                                                                                                                                                                                                                                    arg_arg1_list = ATgetNext(arg_arg1_list);
                                                                                                                                                                                                                                    if (ATgetType((ATerm)arg_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_list) == ATfalse) {
                                                                                                                                                                                                                                      arg_arg1_head = ATgetFirst(arg_arg1_list);
                                                                                                                                                                                                                                      arg_arg1_list = ATgetNext(arg_arg1_list);
                                                                                                                                                                                                                                      /* checking for: appl */
                                                                                                                                                                                                                                      if (ATgetType((ATerm)arg_arg1_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head) == PERR_afun0) {
                                                                                                                                                                                                                                        ATerm arg_arg1_head_arg0 = ATgetArgument(arg_arg1_head, 0);
                                                                                                                                                                                                                                        /* checking for: prod */
                                                                                                                                                                                                                                        if (ATgetType((ATerm)arg_arg1_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0) == PERR_afun1) {
                                                                                                                                                                                                                                          ATerm arg_arg1_head_arg0_arg0 = ATgetArgument(arg_arg1_head_arg0, 0);
                                                                                                                                                                                                                                          if (ATgetType((ATerm)arg_arg1_head_arg0_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0) == ATfalse) {
                                                                                                                                                                                                                                            ATerm arg_arg1_head_arg0_arg0_head;
                                                                                                                                                                                                                                            ATermList arg_arg1_head_arg0_arg0_list = (ATermList)arg_arg1_head_arg0_arg0;
                                                                                                                                                                                                                                            arg_arg1_head_arg0_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_list);
                                                                                                                                                                                                                                            arg_arg1_head_arg0_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_list);
                                                                                                                                                                                                                                            /* checking for: char-class */
                                                                                                                                                                                                                                            if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg0_head) == PERR_afun17) {
                                                                                                                                                                                                                                              ATerm arg_arg1_head_arg0_arg0_head_arg0 = ATgetArgument(arg_arg1_head_arg0_arg0_head, 0);
                                                                                                                                                                                                                                              if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0) == ATfalse) {
                                                                                                                                                                                                                                                ATerm arg_arg1_head_arg0_arg0_head_arg0_head;
                                                                                                                                                                                                                                                ATermList arg_arg1_head_arg0_arg0_head_arg0_list = (ATermList)arg_arg1_head_arg0_arg0_head_arg0;
                                                                                                                                                                                                                                                arg_arg1_head_arg0_arg0_head_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                                                                                                                                                                arg_arg1_head_arg0_arg0_head_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                                                                                                                                                                if (ATgetInt((ATermInt)arg_arg1_head_arg0_arg0_head_arg0_head) == 41) {

                                                                                                                                                                                                                                                  if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0_list) == ATtrue) {
                                                                                                                                                                                                                                                    if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_list) == ATtrue) {
                                                                                                                                                                                                                                                      ATerm arg_arg1_head_arg0_arg1 = ATgetArgument(arg_arg1_head_arg0, 1);
                                                                                                                                                                                                                                                      /* checking for: lit */
                                                                                                                                                                                                                                                      if (ATgetType((ATerm)arg_arg1_head_arg0_arg1) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg1) == PERR_afun9) {
                                                                                                                                                                                                                                                        ATerm arg_arg1_head_arg0_arg1_arg0 = ATgetArgument(arg_arg1_head_arg0_arg1, 0);
                                                                                                                                                                                                                                                        /* checking for: ) */
                                                                                                                                                                                                                                                        if (ATgetType((ATerm)arg_arg1_head_arg0_arg1_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg1_arg0) == PERR_afun38) {
                                                                                                                                                                                                                                                          ATerm arg_arg1_head_arg0_arg2 = ATgetArgument(arg_arg1_head_arg0, 2);
                                                                                                                                                                                                                                                          /* checking for: no-attrs */
                                                                                                                                                                                                                                                          if (ATgetType((ATerm)arg_arg1_head_arg0_arg2) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg2) == PERR_afun5) {
                                                                                                                                                                                                                                                            ATerm arg_arg1_head_arg1 = ATgetArgument(arg_arg1_head, 1);
                                                                                                                                                                                                                                                            if (ATgetType((ATerm)arg_arg1_head_arg1) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg1) == ATfalse) {
                                                                                                                                                                                                                                                              ATerm arg_arg1_head_arg1_head;
                                                                                                                                                                                                                                                              ATermList arg_arg1_head_arg1_list = (ATermList)arg_arg1_head_arg1;
                                                                                                                                                                                                                                                              arg_arg1_head_arg1_head = ATgetFirst(arg_arg1_head_arg1_list);
                                                                                                                                                                                                                                                              arg_arg1_head_arg1_list = ATgetNext(arg_arg1_head_arg1_list);
                                                                                                                                                                                                                                                              if (ATgetInt((ATermInt)arg_arg1_head_arg1_head) == 41) {

                                                                                                                                                                                                                                                                if (ATgetType((ATerm)arg_arg1_head_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg1_list) == ATtrue) {
                                                                                                                                                                                                                                                                  if (ATgetType((ATerm)arg_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_list) == ATtrue) {
                                                                                                                                                                                                                                                                    return ATtrue;
                                                                                                                                                                                                                                                                  }
                                                                                                                                                                                                                                                                }
                                                                                                                                                                                                                                                              }
                                                                                                                                                                                                                                                            }
                                                                                                                                                                                                                                                          }
                                                                                                                                                                                                                                                        }
                                                                                                                                                                                                                                                      }
                                                                                                                                                                                                                                                    }
                                                                                                                                                                                                                                                  }
                                                                                                                                                                                                                                                }
                                                                                                                                                                                                                                              }
                                                                                                                                                                                                                                            }
                                                                                                                                                                                                                                          }
                                                                                                                                                                                                                                        }
                                                                                                                                                                                                                                      }
                                                                                                                                                                                                                                    }
                                                                                                                                                                                                                                  }
                                                                                                                                                                                                                                }
                                                                                                                                                                                                                              }
                                                                                                                                                                                                                            }
                                                                                                                                                                                                                          }
                                                                                                                                                                                                                        }
                                                                                                                                                                                                                      }
                                                                                                                                                                                                                    }
                                                                                                                                                                                                                  }
                                                                                                                                                                                                                }
                                                                                                                                                                                                              }
                                                                                                                                                                                                            }
                                                                                                                                                                                                          }
                                                                                                                                                                                                        }
                                                                                                                                                                                                      }
                                                                                                                                                                                                    }
                                                                                                                                                                                                  }
                                                                                                                                                                                                }
                                                                                                                                                                                              }
                                                                                                                                                                                            }
                                                                                                                                                                                          }
                                                                                                                                                                                        }
                                                                                                                                                                                      }
                                                                                                                                                                                    }
                                                                                                                                                                                  }
                                                                                                                                                                                }
                                                                                                                                                                              }
                                                                                                                                                                            }
                                                                                                                                                                          }
                                                                                                                                                                        }
                                                                                                                                                                      }
                                                                                                                                                                    }
                                                                                                                                                                  }
                                                                                                                                                                }
                                                                                                                                                              }
                                                                                                                                                            }
                                                                                                                                                          }
                                                                                                                                                        }
                                                                                                                                                      }
                                                                                                                                                    }
                                                                                                                                                  }
                                                                                                                                                }
                                                                                                                                              }
                                                                                                                                            }
                                                                                                                                          }
                                                                                                                                        }
                                                                                                                                      }
                                                                                                                                    }
                                                                                                                                  }
                                                                                                                                }
                                                                                                                              }
                                                                                                                            }
                                                                                                                          }
                                                                                                                        }
                                                                                                                      }
                                                                                                                    }
                                                                                                                  }
                                                                                                                }
                                                                                                              }
                                                                                                            }
                                                                                                          }
                                                                                                        }
                                                                                                      }
                                                                                                    }
                                                                                                  }
                                                                                                }
                                                                                              }
                                                                                            }
                                                                                          }
                                                                                        }
                                                                                      }
                                                                                    }
                                                                                  }
                                                                                }
                                                                              }
                                                                            }
                                                                          }
                                                                        }
                                                                      }
                                                                    }
                                                                  }
                                                                }
                                                              }
                                                            }
                                                          }
                                                        }
                                                      }
                                                    }
                                                  }
                                                }
                                              }
                                            }
                                          }
                                        }
                                      }
                                    }
                                  }
                                }
                              }
                            }
                          }
                        }
                      }
                    }
                  }
                }
              }
            }
          }
        }
      }
    }
  }
  return ATfalse;
}

/**
 * Assert whether a PERR_Subject is a localized by checking against the following ATerm pattern: appl(prod([lit("localized"),cf(opt(layout)),lit("("),cf(opt(layout)),cf(sort("StrCon")),cf(opt(layout)),lit(","),cf(opt(layout)),cf(sort("Location")),cf(opt(layout)),lit(")")],cf(sort("Subject")),attrs([term(cons("localized"))])),[appl(prod([char-class([108]),char-class([111]),char-class([99]),char-class([97]),char-class([108]),char-class([105]),char-class([122]),char-class([101]),char-class([100])],lit("localized"),no-attrs),[108,111,99,97,108,105,122,101,100]),<"ws-after-localized"("OptLayout")>,appl(prod([char-class([40])],lit("("),no-attrs),[40]),<"ws-after-("("OptLayout")>,<"description"("StrCon")>,<"ws-after-description"("OptLayout")>,appl(prod([char-class([44])],lit(","),no-attrs),[44]),<"ws-after-,"("OptLayout")>,<"Location"("Location")>,<"ws-after-Location"("OptLayout")>,appl(prod([char-class([41])],lit(")"),no-attrs),[41])]). May not be used to assert correctness of the PERR_Subject
 * \param[in] arg input PERR_Subject
 * \return ATtrue if #arg corresponds to the signature of a localized, or ATfalse otherwise
 */
inline ATbool PERR_isSubjectLocalized(PERR_Subject arg){
  /* checking for: appl */
  if (ATgetType((ATerm)arg) == AT_APPL && ATgetAFun((ATermAppl)arg) == PERR_afun0) {
    ATerm arg_arg0 = ATgetArgument(arg, 0);
    /* checking for: prod */
    if (ATgetType((ATerm)arg_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0) == PERR_afun1) {
      ATerm arg_arg0_arg0 = ATgetArgument(arg_arg0, 0);
      if (ATgetType((ATerm)arg_arg0_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0) == ATfalse) {
        ATerm arg_arg0_arg0_head;
        ATermList arg_arg0_arg0_list = (ATermList)arg_arg0_arg0;
        arg_arg0_arg0_head = ATgetFirst(arg_arg0_arg0_list);
        arg_arg0_arg0_list = ATgetNext(arg_arg0_arg0_list);
        /* checking for: lit */
        if (ATgetType((ATerm)arg_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head) == PERR_afun9) {
          ATerm arg_arg0_arg0_head_arg0 = ATgetArgument(arg_arg0_arg0_head, 0);
          /* checking for: localized */
          if (ATgetType((ATerm)arg_arg0_arg0_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0) == PERR_afun44) {
            if (ATgetType((ATerm)arg_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0_list) == ATfalse) {
              arg_arg0_arg0_head = ATgetFirst(arg_arg0_arg0_list);
              arg_arg0_arg0_list = ATgetNext(arg_arg0_arg0_list);
              /* checking for: cf */
              if (ATgetType((ATerm)arg_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head) == PERR_afun2) {
                ATerm arg_arg0_arg0_head_arg0 = ATgetArgument(arg_arg0_arg0_head, 0);
                /* checking for: opt */
                if (ATgetType((ATerm)arg_arg0_arg0_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0) == PERR_afun3) {
                  ATerm arg_arg0_arg0_head_arg0_arg0 = ATgetArgument(arg_arg0_arg0_head_arg0, 0);
                  /* checking for: layout */
                  if (ATgetType((ATerm)arg_arg0_arg0_head_arg0_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0_arg0) == PERR_afun4) {
                    if (ATgetType((ATerm)arg_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0_list) == ATfalse) {
                      arg_arg0_arg0_head = ATgetFirst(arg_arg0_arg0_list);
                      arg_arg0_arg0_list = ATgetNext(arg_arg0_arg0_list);
                      /* checking for: lit */
                      if (ATgetType((ATerm)arg_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head) == PERR_afun9) {
                        ATerm arg_arg0_arg0_head_arg0 = ATgetArgument(arg_arg0_arg0_head, 0);
                        /* checking for: ( */
                        if (ATgetType((ATerm)arg_arg0_arg0_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0) == PERR_afun39) {
                          if (ATgetType((ATerm)arg_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0_list) == ATfalse) {
                            arg_arg0_arg0_head = ATgetFirst(arg_arg0_arg0_list);
                            arg_arg0_arg0_list = ATgetNext(arg_arg0_arg0_list);
                            /* checking for: cf */
                            if (ATgetType((ATerm)arg_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head) == PERR_afun2) {
                              ATerm arg_arg0_arg0_head_arg0 = ATgetArgument(arg_arg0_arg0_head, 0);
                              /* checking for: opt */
                              if (ATgetType((ATerm)arg_arg0_arg0_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0) == PERR_afun3) {
                                ATerm arg_arg0_arg0_head_arg0_arg0 = ATgetArgument(arg_arg0_arg0_head_arg0, 0);
                                /* checking for: layout */
                                if (ATgetType((ATerm)arg_arg0_arg0_head_arg0_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0_arg0) == PERR_afun4) {
                                  if (ATgetType((ATerm)arg_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0_list) == ATfalse) {
                                    arg_arg0_arg0_head = ATgetFirst(arg_arg0_arg0_list);
                                    arg_arg0_arg0_list = ATgetNext(arg_arg0_arg0_list);
                                    /* checking for: cf */
                                    if (ATgetType((ATerm)arg_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head) == PERR_afun2) {
                                      ATerm arg_arg0_arg0_head_arg0 = ATgetArgument(arg_arg0_arg0_head, 0);
                                      /* checking for: sort */
                                      if (ATgetType((ATerm)arg_arg0_arg0_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0) == PERR_afun11) {
                                        ATerm arg_arg0_arg0_head_arg0_arg0 = ATgetArgument(arg_arg0_arg0_head_arg0, 0);
                                        /* checking for: StrCon */
                                        if (ATgetType((ATerm)arg_arg0_arg0_head_arg0_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0_arg0) == PERR_afun29) {
                                          if (ATgetType((ATerm)arg_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0_list) == ATfalse) {
                                            arg_arg0_arg0_head = ATgetFirst(arg_arg0_arg0_list);
                                            arg_arg0_arg0_list = ATgetNext(arg_arg0_arg0_list);
                                            /* checking for: cf */
                                            if (ATgetType((ATerm)arg_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head) == PERR_afun2) {
                                              ATerm arg_arg0_arg0_head_arg0 = ATgetArgument(arg_arg0_arg0_head, 0);
                                              /* checking for: opt */
                                              if (ATgetType((ATerm)arg_arg0_arg0_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0) == PERR_afun3) {
                                                ATerm arg_arg0_arg0_head_arg0_arg0 = ATgetArgument(arg_arg0_arg0_head_arg0, 0);
                                                /* checking for: layout */
                                                if (ATgetType((ATerm)arg_arg0_arg0_head_arg0_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0_arg0) == PERR_afun4) {
                                                  if (ATgetType((ATerm)arg_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0_list) == ATfalse) {
                                                    arg_arg0_arg0_head = ATgetFirst(arg_arg0_arg0_list);
                                                    arg_arg0_arg0_list = ATgetNext(arg_arg0_arg0_list);
                                                    /* checking for: lit */
                                                    if (ATgetType((ATerm)arg_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head) == PERR_afun9) {
                                                      ATerm arg_arg0_arg0_head_arg0 = ATgetArgument(arg_arg0_arg0_head, 0);
                                                      /* checking for: , */
                                                      if (ATgetType((ATerm)arg_arg0_arg0_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0) == PERR_afun43) {
                                                        if (ATgetType((ATerm)arg_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0_list) == ATfalse) {
                                                          arg_arg0_arg0_head = ATgetFirst(arg_arg0_arg0_list);
                                                          arg_arg0_arg0_list = ATgetNext(arg_arg0_arg0_list);
                                                          /* checking for: cf */
                                                          if (ATgetType((ATerm)arg_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head) == PERR_afun2) {
                                                            ATerm arg_arg0_arg0_head_arg0 = ATgetArgument(arg_arg0_arg0_head, 0);
                                                            /* checking for: opt */
                                                            if (ATgetType((ATerm)arg_arg0_arg0_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0) == PERR_afun3) {
                                                              ATerm arg_arg0_arg0_head_arg0_arg0 = ATgetArgument(arg_arg0_arg0_head_arg0, 0);
                                                              /* checking for: layout */
                                                              if (ATgetType((ATerm)arg_arg0_arg0_head_arg0_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0_arg0) == PERR_afun4) {
                                                                if (ATgetType((ATerm)arg_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0_list) == ATfalse) {
                                                                  arg_arg0_arg0_head = ATgetFirst(arg_arg0_arg0_list);
                                                                  arg_arg0_arg0_list = ATgetNext(arg_arg0_arg0_list);
                                                                  /* checking for: cf */
                                                                  if (ATgetType((ATerm)arg_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head) == PERR_afun2) {
                                                                    ATerm arg_arg0_arg0_head_arg0 = ATgetArgument(arg_arg0_arg0_head, 0);
                                                                    /* checking for: sort */
                                                                    if (ATgetType((ATerm)arg_arg0_arg0_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0) == PERR_afun11) {
                                                                      ATerm arg_arg0_arg0_head_arg0_arg0 = ATgetArgument(arg_arg0_arg0_head_arg0, 0);
                                                                      /* checking for: Location */
                                                                      if (ATgetType((ATerm)arg_arg0_arg0_head_arg0_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0_arg0) == PERR_afun42) {
                                                                        if (ATgetType((ATerm)arg_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0_list) == ATfalse) {
                                                                          arg_arg0_arg0_head = ATgetFirst(arg_arg0_arg0_list);
                                                                          arg_arg0_arg0_list = ATgetNext(arg_arg0_arg0_list);
                                                                          /* checking for: cf */
                                                                          if (ATgetType((ATerm)arg_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head) == PERR_afun2) {
                                                                            ATerm arg_arg0_arg0_head_arg0 = ATgetArgument(arg_arg0_arg0_head, 0);
                                                                            /* checking for: opt */
                                                                            if (ATgetType((ATerm)arg_arg0_arg0_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0) == PERR_afun3) {
                                                                              ATerm arg_arg0_arg0_head_arg0_arg0 = ATgetArgument(arg_arg0_arg0_head_arg0, 0);
                                                                              /* checking for: layout */
                                                                              if (ATgetType((ATerm)arg_arg0_arg0_head_arg0_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0_arg0) == PERR_afun4) {
                                                                                if (ATgetType((ATerm)arg_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0_list) == ATfalse) {
                                                                                  arg_arg0_arg0_head = ATgetFirst(arg_arg0_arg0_list);
                                                                                  arg_arg0_arg0_list = ATgetNext(arg_arg0_arg0_list);
                                                                                  /* checking for: lit */
                                                                                  if (ATgetType((ATerm)arg_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head) == PERR_afun9) {
                                                                                    ATerm arg_arg0_arg0_head_arg0 = ATgetArgument(arg_arg0_arg0_head, 0);
                                                                                    /* checking for: ) */
                                                                                    if (ATgetType((ATerm)arg_arg0_arg0_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0) == PERR_afun38) {
                                                                                      if (ATgetType((ATerm)arg_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0_list) == ATtrue) {
                                                                                        ATerm arg_arg0_arg1 = ATgetArgument(arg_arg0, 1);
                                                                                        /* checking for: cf */
                                                                                        if (ATgetType((ATerm)arg_arg0_arg1) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg1) == PERR_afun2) {
                                                                                          ATerm arg_arg0_arg1_arg0 = ATgetArgument(arg_arg0_arg1, 0);
                                                                                          /* checking for: sort */
                                                                                          if (ATgetType((ATerm)arg_arg0_arg1_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg1_arg0) == PERR_afun11) {
                                                                                            ATerm arg_arg0_arg1_arg0_arg0 = ATgetArgument(arg_arg0_arg1_arg0, 0);
                                                                                            /* checking for: Subject */
                                                                                            if (ATgetType((ATerm)arg_arg0_arg1_arg0_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg1_arg0_arg0) == PERR_afun41) {
                                                                                              ATerm arg_arg0_arg2 = ATgetArgument(arg_arg0, 2);
                                                                                              /* checking for: attrs */
                                                                                              if (ATgetType((ATerm)arg_arg0_arg2) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg2) == PERR_afun13) {
                                                                                                ATerm arg_arg0_arg2_arg0 = ATgetArgument(arg_arg0_arg2, 0);
                                                                                                if (ATgetType((ATerm)arg_arg0_arg2_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg2_arg0) == ATfalse) {
                                                                                                  ATerm arg_arg0_arg2_arg0_head;
                                                                                                  ATermList arg_arg0_arg2_arg0_list = (ATermList)arg_arg0_arg2_arg0;
                                                                                                  arg_arg0_arg2_arg0_head = ATgetFirst(arg_arg0_arg2_arg0_list);
                                                                                                  arg_arg0_arg2_arg0_list = ATgetNext(arg_arg0_arg2_arg0_list);
                                                                                                  /* checking for: term */
                                                                                                  if (ATgetType((ATerm)arg_arg0_arg2_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg2_arg0_head) == PERR_afun14) {
                                                                                                    ATerm arg_arg0_arg2_arg0_head_arg0 = ATgetArgument(arg_arg0_arg2_arg0_head, 0);
                                                                                                    /* checking for: cons */
                                                                                                    if (ATgetType((ATerm)arg_arg0_arg2_arg0_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg2_arg0_head_arg0) == PERR_afun15) {
                                                                                                      ATerm arg_arg0_arg2_arg0_head_arg0_arg0 = ATgetArgument(arg_arg0_arg2_arg0_head_arg0, 0);
                                                                                                      /* checking for: localized */
                                                                                                      if (ATgetType((ATerm)arg_arg0_arg2_arg0_head_arg0_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg2_arg0_head_arg0_arg0) == PERR_afun44) {
                                                                                                        if (ATgetType((ATerm)arg_arg0_arg2_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg2_arg0_list) == ATtrue) {
                                                                                                          ATerm arg_arg1 = ATgetArgument(arg, 1);
                                                                                                          if (ATgetType((ATerm)arg_arg1) == AT_LIST && ATisEmpty((ATermList)arg_arg1) == ATfalse) {
                                                                                                            ATerm arg_arg1_head;
                                                                                                            ATermList arg_arg1_list = (ATermList)arg_arg1;
                                                                                                            arg_arg1_head = ATgetFirst(arg_arg1_list);
                                                                                                            arg_arg1_list = ATgetNext(arg_arg1_list);
                                                                                                            /* checking for: appl */
                                                                                                            if (ATgetType((ATerm)arg_arg1_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head) == PERR_afun0) {
                                                                                                              ATerm arg_arg1_head_arg0 = ATgetArgument(arg_arg1_head, 0);
                                                                                                              /* checking for: prod */
                                                                                                              if (ATgetType((ATerm)arg_arg1_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0) == PERR_afun1) {
                                                                                                                ATerm arg_arg1_head_arg0_arg0 = ATgetArgument(arg_arg1_head_arg0, 0);
                                                                                                                if (ATgetType((ATerm)arg_arg1_head_arg0_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0) == ATfalse) {
                                                                                                                  ATerm arg_arg1_head_arg0_arg0_head;
                                                                                                                  ATermList arg_arg1_head_arg0_arg0_list = (ATermList)arg_arg1_head_arg0_arg0;
                                                                                                                  arg_arg1_head_arg0_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_list);
                                                                                                                  arg_arg1_head_arg0_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_list);
                                                                                                                  /* checking for: char-class */
                                                                                                                  if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg0_head) == PERR_afun17) {
                                                                                                                    ATerm arg_arg1_head_arg0_arg0_head_arg0 = ATgetArgument(arg_arg1_head_arg0_arg0_head, 0);
                                                                                                                    if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0) == ATfalse) {
                                                                                                                      ATerm arg_arg1_head_arg0_arg0_head_arg0_head;
                                                                                                                      ATermList arg_arg1_head_arg0_arg0_head_arg0_list = (ATermList)arg_arg1_head_arg0_arg0_head_arg0;
                                                                                                                      arg_arg1_head_arg0_arg0_head_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                                      arg_arg1_head_arg0_arg0_head_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                                      if (ATgetInt((ATermInt)arg_arg1_head_arg0_arg0_head_arg0_head) == 108) {

                                                                                                                        if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0_list) == ATtrue) {
                                                                                                                          if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_list) == ATfalse) {
                                                                                                                            arg_arg1_head_arg0_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_list);
                                                                                                                            arg_arg1_head_arg0_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_list);
                                                                                                                            /* checking for: char-class */
                                                                                                                            if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg0_head) == PERR_afun17) {
                                                                                                                              ATerm arg_arg1_head_arg0_arg0_head_arg0 = ATgetArgument(arg_arg1_head_arg0_arg0_head, 0);
                                                                                                                              if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0) == ATfalse) {
                                                                                                                                ATerm arg_arg1_head_arg0_arg0_head_arg0_head;
                                                                                                                                ATermList arg_arg1_head_arg0_arg0_head_arg0_list = (ATermList)arg_arg1_head_arg0_arg0_head_arg0;
                                                                                                                                arg_arg1_head_arg0_arg0_head_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                                                arg_arg1_head_arg0_arg0_head_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                                                if (ATgetInt((ATermInt)arg_arg1_head_arg0_arg0_head_arg0_head) == 111) {

                                                                                                                                  if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0_list) == ATtrue) {
                                                                                                                                    if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_list) == ATfalse) {
                                                                                                                                      arg_arg1_head_arg0_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_list);
                                                                                                                                      arg_arg1_head_arg0_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_list);
                                                                                                                                      /* checking for: char-class */
                                                                                                                                      if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg0_head) == PERR_afun17) {
                                                                                                                                        ATerm arg_arg1_head_arg0_arg0_head_arg0 = ATgetArgument(arg_arg1_head_arg0_arg0_head, 0);
                                                                                                                                        if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0) == ATfalse) {
                                                                                                                                          ATerm arg_arg1_head_arg0_arg0_head_arg0_head;
                                                                                                                                          ATermList arg_arg1_head_arg0_arg0_head_arg0_list = (ATermList)arg_arg1_head_arg0_arg0_head_arg0;
                                                                                                                                          arg_arg1_head_arg0_arg0_head_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                                                          arg_arg1_head_arg0_arg0_head_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                                                          if (ATgetInt((ATermInt)arg_arg1_head_arg0_arg0_head_arg0_head) == 99) {

                                                                                                                                            if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0_list) == ATtrue) {
                                                                                                                                              if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_list) == ATfalse) {
                                                                                                                                                arg_arg1_head_arg0_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_list);
                                                                                                                                                arg_arg1_head_arg0_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_list);
                                                                                                                                                /* checking for: char-class */
                                                                                                                                                if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg0_head) == PERR_afun17) {
                                                                                                                                                  ATerm arg_arg1_head_arg0_arg0_head_arg0 = ATgetArgument(arg_arg1_head_arg0_arg0_head, 0);
                                                                                                                                                  if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0) == ATfalse) {
                                                                                                                                                    ATerm arg_arg1_head_arg0_arg0_head_arg0_head;
                                                                                                                                                    ATermList arg_arg1_head_arg0_arg0_head_arg0_list = (ATermList)arg_arg1_head_arg0_arg0_head_arg0;
                                                                                                                                                    arg_arg1_head_arg0_arg0_head_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                                                                    arg_arg1_head_arg0_arg0_head_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                                                                    if (ATgetInt((ATermInt)arg_arg1_head_arg0_arg0_head_arg0_head) == 97) {

                                                                                                                                                      if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0_list) == ATtrue) {
                                                                                                                                                        if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_list) == ATfalse) {
                                                                                                                                                          arg_arg1_head_arg0_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_list);
                                                                                                                                                          arg_arg1_head_arg0_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_list);
                                                                                                                                                          /* checking for: char-class */
                                                                                                                                                          if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg0_head) == PERR_afun17) {
                                                                                                                                                            ATerm arg_arg1_head_arg0_arg0_head_arg0 = ATgetArgument(arg_arg1_head_arg0_arg0_head, 0);
                                                                                                                                                            if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0) == ATfalse) {
                                                                                                                                                              ATerm arg_arg1_head_arg0_arg0_head_arg0_head;
                                                                                                                                                              ATermList arg_arg1_head_arg0_arg0_head_arg0_list = (ATermList)arg_arg1_head_arg0_arg0_head_arg0;
                                                                                                                                                              arg_arg1_head_arg0_arg0_head_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                                                                              arg_arg1_head_arg0_arg0_head_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                                                                              if (ATgetInt((ATermInt)arg_arg1_head_arg0_arg0_head_arg0_head) == 108) {

                                                                                                                                                                if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0_list) == ATtrue) {
                                                                                                                                                                  if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_list) == ATfalse) {
                                                                                                                                                                    arg_arg1_head_arg0_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_list);
                                                                                                                                                                    arg_arg1_head_arg0_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_list);
                                                                                                                                                                    /* checking for: char-class */
                                                                                                                                                                    if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg0_head) == PERR_afun17) {
                                                                                                                                                                      ATerm arg_arg1_head_arg0_arg0_head_arg0 = ATgetArgument(arg_arg1_head_arg0_arg0_head, 0);
                                                                                                                                                                      if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0) == ATfalse) {
                                                                                                                                                                        ATerm arg_arg1_head_arg0_arg0_head_arg0_head;
                                                                                                                                                                        ATermList arg_arg1_head_arg0_arg0_head_arg0_list = (ATermList)arg_arg1_head_arg0_arg0_head_arg0;
                                                                                                                                                                        arg_arg1_head_arg0_arg0_head_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                                                                                        arg_arg1_head_arg0_arg0_head_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                                                                                        if (ATgetInt((ATermInt)arg_arg1_head_arg0_arg0_head_arg0_head) == 105) {

                                                                                                                                                                          if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0_list) == ATtrue) {
                                                                                                                                                                            if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_list) == ATfalse) {
                                                                                                                                                                              arg_arg1_head_arg0_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_list);
                                                                                                                                                                              arg_arg1_head_arg0_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_list);
                                                                                                                                                                              /* checking for: char-class */
                                                                                                                                                                              if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg0_head) == PERR_afun17) {
                                                                                                                                                                                ATerm arg_arg1_head_arg0_arg0_head_arg0 = ATgetArgument(arg_arg1_head_arg0_arg0_head, 0);
                                                                                                                                                                                if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0) == ATfalse) {
                                                                                                                                                                                  ATerm arg_arg1_head_arg0_arg0_head_arg0_head;
                                                                                                                                                                                  ATermList arg_arg1_head_arg0_arg0_head_arg0_list = (ATermList)arg_arg1_head_arg0_arg0_head_arg0;
                                                                                                                                                                                  arg_arg1_head_arg0_arg0_head_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                                                                                                  arg_arg1_head_arg0_arg0_head_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                                                                                                  if (ATgetInt((ATermInt)arg_arg1_head_arg0_arg0_head_arg0_head) == 122) {

                                                                                                                                                                                    if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0_list) == ATtrue) {
                                                                                                                                                                                      if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_list) == ATfalse) {
                                                                                                                                                                                        arg_arg1_head_arg0_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_list);
                                                                                                                                                                                        arg_arg1_head_arg0_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_list);
                                                                                                                                                                                        /* checking for: char-class */
                                                                                                                                                                                        if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg0_head) == PERR_afun17) {
                                                                                                                                                                                          ATerm arg_arg1_head_arg0_arg0_head_arg0 = ATgetArgument(arg_arg1_head_arg0_arg0_head, 0);
                                                                                                                                                                                          if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0) == ATfalse) {
                                                                                                                                                                                            ATerm arg_arg1_head_arg0_arg0_head_arg0_head;
                                                                                                                                                                                            ATermList arg_arg1_head_arg0_arg0_head_arg0_list = (ATermList)arg_arg1_head_arg0_arg0_head_arg0;
                                                                                                                                                                                            arg_arg1_head_arg0_arg0_head_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                                                                                                            arg_arg1_head_arg0_arg0_head_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                                                                                                            if (ATgetInt((ATermInt)arg_arg1_head_arg0_arg0_head_arg0_head) == 101) {

                                                                                                                                                                                              if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0_list) == ATtrue) {
                                                                                                                                                                                                if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_list) == ATfalse) {
                                                                                                                                                                                                  arg_arg1_head_arg0_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_list);
                                                                                                                                                                                                  arg_arg1_head_arg0_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_list);
                                                                                                                                                                                                  /* checking for: char-class */
                                                                                                                                                                                                  if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg0_head) == PERR_afun17) {
                                                                                                                                                                                                    ATerm arg_arg1_head_arg0_arg0_head_arg0 = ATgetArgument(arg_arg1_head_arg0_arg0_head, 0);
                                                                                                                                                                                                    if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0) == ATfalse) {
                                                                                                                                                                                                      ATerm arg_arg1_head_arg0_arg0_head_arg0_head;
                                                                                                                                                                                                      ATermList arg_arg1_head_arg0_arg0_head_arg0_list = (ATermList)arg_arg1_head_arg0_arg0_head_arg0;
                                                                                                                                                                                                      arg_arg1_head_arg0_arg0_head_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                                                                                                                      arg_arg1_head_arg0_arg0_head_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                                                                                                                      if (ATgetInt((ATermInt)arg_arg1_head_arg0_arg0_head_arg0_head) == 100) {

                                                                                                                                                                                                        if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0_list) == ATtrue) {
                                                                                                                                                                                                          if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_list) == ATtrue) {
                                                                                                                                                                                                            ATerm arg_arg1_head_arg0_arg1 = ATgetArgument(arg_arg1_head_arg0, 1);
                                                                                                                                                                                                            /* checking for: lit */
                                                                                                                                                                                                            if (ATgetType((ATerm)arg_arg1_head_arg0_arg1) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg1) == PERR_afun9) {
                                                                                                                                                                                                              ATerm arg_arg1_head_arg0_arg1_arg0 = ATgetArgument(arg_arg1_head_arg0_arg1, 0);
                                                                                                                                                                                                              /* checking for: localized */
                                                                                                                                                                                                              if (ATgetType((ATerm)arg_arg1_head_arg0_arg1_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg1_arg0) == PERR_afun44) {
                                                                                                                                                                                                                ATerm arg_arg1_head_arg0_arg2 = ATgetArgument(arg_arg1_head_arg0, 2);
                                                                                                                                                                                                                /* checking for: no-attrs */
                                                                                                                                                                                                                if (ATgetType((ATerm)arg_arg1_head_arg0_arg2) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg2) == PERR_afun5) {
                                                                                                                                                                                                                  ATerm arg_arg1_head_arg1 = ATgetArgument(arg_arg1_head, 1);
                                                                                                                                                                                                                  if (ATgetType((ATerm)arg_arg1_head_arg1) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg1) == ATfalse) {
                                                                                                                                                                                                                    ATerm arg_arg1_head_arg1_head;
                                                                                                                                                                                                                    ATermList arg_arg1_head_arg1_list = (ATermList)arg_arg1_head_arg1;
                                                                                                                                                                                                                    arg_arg1_head_arg1_head = ATgetFirst(arg_arg1_head_arg1_list);
                                                                                                                                                                                                                    arg_arg1_head_arg1_list = ATgetNext(arg_arg1_head_arg1_list);
                                                                                                                                                                                                                    if (ATgetInt((ATermInt)arg_arg1_head_arg1_head) == 108) {

                                                                                                                                                                                                                      if (ATgetType((ATerm)arg_arg1_head_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg1_list) == ATfalse) {
                                                                                                                                                                                                                        arg_arg1_head_arg1_head = ATgetFirst(arg_arg1_head_arg1_list);
                                                                                                                                                                                                                        arg_arg1_head_arg1_list = ATgetNext(arg_arg1_head_arg1_list);
                                                                                                                                                                                                                        if (ATgetInt((ATermInt)arg_arg1_head_arg1_head) == 111) {

                                                                                                                                                                                                                          if (ATgetType((ATerm)arg_arg1_head_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg1_list) == ATfalse) {
                                                                                                                                                                                                                            arg_arg1_head_arg1_head = ATgetFirst(arg_arg1_head_arg1_list);
                                                                                                                                                                                                                            arg_arg1_head_arg1_list = ATgetNext(arg_arg1_head_arg1_list);
                                                                                                                                                                                                                            if (ATgetInt((ATermInt)arg_arg1_head_arg1_head) == 99) {

                                                                                                                                                                                                                              if (ATgetType((ATerm)arg_arg1_head_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg1_list) == ATfalse) {
                                                                                                                                                                                                                                arg_arg1_head_arg1_head = ATgetFirst(arg_arg1_head_arg1_list);
                                                                                                                                                                                                                                arg_arg1_head_arg1_list = ATgetNext(arg_arg1_head_arg1_list);
                                                                                                                                                                                                                                if (ATgetInt((ATermInt)arg_arg1_head_arg1_head) == 97) {

                                                                                                                                                                                                                                  if (ATgetType((ATerm)arg_arg1_head_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg1_list) == ATfalse) {
                                                                                                                                                                                                                                    arg_arg1_head_arg1_head = ATgetFirst(arg_arg1_head_arg1_list);
                                                                                                                                                                                                                                    arg_arg1_head_arg1_list = ATgetNext(arg_arg1_head_arg1_list);
                                                                                                                                                                                                                                    if (ATgetInt((ATermInt)arg_arg1_head_arg1_head) == 108) {

                                                                                                                                                                                                                                      if (ATgetType((ATerm)arg_arg1_head_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg1_list) == ATfalse) {
                                                                                                                                                                                                                                        arg_arg1_head_arg1_head = ATgetFirst(arg_arg1_head_arg1_list);
                                                                                                                                                                                                                                        arg_arg1_head_arg1_list = ATgetNext(arg_arg1_head_arg1_list);
                                                                                                                                                                                                                                        if (ATgetInt((ATermInt)arg_arg1_head_arg1_head) == 105) {

                                                                                                                                                                                                                                          if (ATgetType((ATerm)arg_arg1_head_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg1_list) == ATfalse) {
                                                                                                                                                                                                                                            arg_arg1_head_arg1_head = ATgetFirst(arg_arg1_head_arg1_list);
                                                                                                                                                                                                                                            arg_arg1_head_arg1_list = ATgetNext(arg_arg1_head_arg1_list);
                                                                                                                                                                                                                                            if (ATgetInt((ATermInt)arg_arg1_head_arg1_head) == 122) {

                                                                                                                                                                                                                                              if (ATgetType((ATerm)arg_arg1_head_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg1_list) == ATfalse) {
                                                                                                                                                                                                                                                arg_arg1_head_arg1_head = ATgetFirst(arg_arg1_head_arg1_list);
                                                                                                                                                                                                                                                arg_arg1_head_arg1_list = ATgetNext(arg_arg1_head_arg1_list);
                                                                                                                                                                                                                                                if (ATgetInt((ATermInt)arg_arg1_head_arg1_head) == 101) {

                                                                                                                                                                                                                                                  if (ATgetType((ATerm)arg_arg1_head_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg1_list) == ATfalse) {
                                                                                                                                                                                                                                                    arg_arg1_head_arg1_head = ATgetFirst(arg_arg1_head_arg1_list);
                                                                                                                                                                                                                                                    arg_arg1_head_arg1_list = ATgetNext(arg_arg1_head_arg1_list);
                                                                                                                                                                                                                                                    if (ATgetInt((ATermInt)arg_arg1_head_arg1_head) == 100) {

                                                                                                                                                                                                                                                      if (ATgetType((ATerm)arg_arg1_head_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg1_list) == ATtrue) {
                                                                                                                                                                                                                                                        if (ATgetType((ATerm)arg_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_list) == ATfalse) {
                                                                                                                                                                                                                                                          arg_arg1_list = ATgetNext(arg_arg1_list);
                                                                                                                                                                                                                                                          if (ATgetType((ATerm)arg_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_list) == ATfalse) {
                                                                                                                                                                                                                                                            arg_arg1_head = ATgetFirst(arg_arg1_list);
                                                                                                                                                                                                                                                            arg_arg1_list = ATgetNext(arg_arg1_list);
                                                                                                                                                                                                                                                            /* checking for: appl */
                                                                                                                                                                                                                                                            if (ATgetType((ATerm)arg_arg1_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head) == PERR_afun0) {
                                                                                                                                                                                                                                                              ATerm arg_arg1_head_arg0 = ATgetArgument(arg_arg1_head, 0);
                                                                                                                                                                                                                                                              /* checking for: prod */
                                                                                                                                                                                                                                                              if (ATgetType((ATerm)arg_arg1_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0) == PERR_afun1) {
                                                                                                                                                                                                                                                                ATerm arg_arg1_head_arg0_arg0 = ATgetArgument(arg_arg1_head_arg0, 0);
                                                                                                                                                                                                                                                                if (ATgetType((ATerm)arg_arg1_head_arg0_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0) == ATfalse) {
                                                                                                                                                                                                                                                                  ATerm arg_arg1_head_arg0_arg0_head;
                                                                                                                                                                                                                                                                  ATermList arg_arg1_head_arg0_arg0_list = (ATermList)arg_arg1_head_arg0_arg0;
                                                                                                                                                                                                                                                                  arg_arg1_head_arg0_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_list);
                                                                                                                                                                                                                                                                  arg_arg1_head_arg0_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_list);
                                                                                                                                                                                                                                                                  /* checking for: char-class */
                                                                                                                                                                                                                                                                  if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg0_head) == PERR_afun17) {
                                                                                                                                                                                                                                                                    ATerm arg_arg1_head_arg0_arg0_head_arg0 = ATgetArgument(arg_arg1_head_arg0_arg0_head, 0);
                                                                                                                                                                                                                                                                    if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0) == ATfalse) {
                                                                                                                                                                                                                                                                      ATerm arg_arg1_head_arg0_arg0_head_arg0_head;
                                                                                                                                                                                                                                                                      ATermList arg_arg1_head_arg0_arg0_head_arg0_list = (ATermList)arg_arg1_head_arg0_arg0_head_arg0;
                                                                                                                                                                                                                                                                      arg_arg1_head_arg0_arg0_head_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                                                                                                                                                                                      arg_arg1_head_arg0_arg0_head_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                                                                                                                                                                                      if (ATgetInt((ATermInt)arg_arg1_head_arg0_arg0_head_arg0_head) == 40) {

                                                                                                                                                                                                                                                                        if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0_list) == ATtrue) {
                                                                                                                                                                                                                                                                          if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_list) == ATtrue) {
                                                                                                                                                                                                                                                                            ATerm arg_arg1_head_arg0_arg1 = ATgetArgument(arg_arg1_head_arg0, 1);
                                                                                                                                                                                                                                                                            /* checking for: lit */
                                                                                                                                                                                                                                                                            if (ATgetType((ATerm)arg_arg1_head_arg0_arg1) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg1) == PERR_afun9) {
                                                                                                                                                                                                                                                                              ATerm arg_arg1_head_arg0_arg1_arg0 = ATgetArgument(arg_arg1_head_arg0_arg1, 0);
                                                                                                                                                                                                                                                                              /* checking for: ( */
                                                                                                                                                                                                                                                                              if (ATgetType((ATerm)arg_arg1_head_arg0_arg1_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg1_arg0) == PERR_afun39) {
                                                                                                                                                                                                                                                                                ATerm arg_arg1_head_arg0_arg2 = ATgetArgument(arg_arg1_head_arg0, 2);
                                                                                                                                                                                                                                                                                /* checking for: no-attrs */
                                                                                                                                                                                                                                                                                if (ATgetType((ATerm)arg_arg1_head_arg0_arg2) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg2) == PERR_afun5) {
                                                                                                                                                                                                                                                                                  ATerm arg_arg1_head_arg1 = ATgetArgument(arg_arg1_head, 1);
                                                                                                                                                                                                                                                                                  if (ATgetType((ATerm)arg_arg1_head_arg1) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg1) == ATfalse) {
                                                                                                                                                                                                                                                                                    ATerm arg_arg1_head_arg1_head;
                                                                                                                                                                                                                                                                                    ATermList arg_arg1_head_arg1_list = (ATermList)arg_arg1_head_arg1;
                                                                                                                                                                                                                                                                                    arg_arg1_head_arg1_head = ATgetFirst(arg_arg1_head_arg1_list);
                                                                                                                                                                                                                                                                                    arg_arg1_head_arg1_list = ATgetNext(arg_arg1_head_arg1_list);
                                                                                                                                                                                                                                                                                    if (ATgetInt((ATermInt)arg_arg1_head_arg1_head) == 40) {

                                                                                                                                                                                                                                                                                      if (ATgetType((ATerm)arg_arg1_head_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg1_list) == ATtrue) {
                                                                                                                                                                                                                                                                                        if (ATgetType((ATerm)arg_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_list) == ATfalse) {
                                                                                                                                                                                                                                                                                          arg_arg1_list = ATgetNext(arg_arg1_list);
                                                                                                                                                                                                                                                                                          if (ATgetType((ATerm)arg_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_list) == ATfalse) {
                                                                                                                                                                                                                                                                                            arg_arg1_list = ATgetNext(arg_arg1_list);
                                                                                                                                                                                                                                                                                            if (ATgetType((ATerm)arg_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_list) == ATfalse) {
                                                                                                                                                                                                                                                                                              arg_arg1_list = ATgetNext(arg_arg1_list);
                                                                                                                                                                                                                                                                                              if (ATgetType((ATerm)arg_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_list) == ATfalse) {
                                                                                                                                                                                                                                                                                                arg_arg1_head = ATgetFirst(arg_arg1_list);
                                                                                                                                                                                                                                                                                                arg_arg1_list = ATgetNext(arg_arg1_list);
                                                                                                                                                                                                                                                                                                /* checking for: appl */
                                                                                                                                                                                                                                                                                                if (ATgetType((ATerm)arg_arg1_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head) == PERR_afun0) {
                                                                                                                                                                                                                                                                                                  ATerm arg_arg1_head_arg0 = ATgetArgument(arg_arg1_head, 0);
                                                                                                                                                                                                                                                                                                  /* checking for: prod */
                                                                                                                                                                                                                                                                                                  if (ATgetType((ATerm)arg_arg1_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0) == PERR_afun1) {
                                                                                                                                                                                                                                                                                                    ATerm arg_arg1_head_arg0_arg0 = ATgetArgument(arg_arg1_head_arg0, 0);
                                                                                                                                                                                                                                                                                                    if (ATgetType((ATerm)arg_arg1_head_arg0_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0) == ATfalse) {
                                                                                                                                                                                                                                                                                                      ATerm arg_arg1_head_arg0_arg0_head;
                                                                                                                                                                                                                                                                                                      ATermList arg_arg1_head_arg0_arg0_list = (ATermList)arg_arg1_head_arg0_arg0;
                                                                                                                                                                                                                                                                                                      arg_arg1_head_arg0_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_list);
                                                                                                                                                                                                                                                                                                      arg_arg1_head_arg0_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_list);
                                                                                                                                                                                                                                                                                                      /* checking for: char-class */
                                                                                                                                                                                                                                                                                                      if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg0_head) == PERR_afun17) {
                                                                                                                                                                                                                                                                                                        ATerm arg_arg1_head_arg0_arg0_head_arg0 = ATgetArgument(arg_arg1_head_arg0_arg0_head, 0);
                                                                                                                                                                                                                                                                                                        if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0) == ATfalse) {
                                                                                                                                                                                                                                                                                                          ATerm arg_arg1_head_arg0_arg0_head_arg0_head;
                                                                                                                                                                                                                                                                                                          ATermList arg_arg1_head_arg0_arg0_head_arg0_list = (ATermList)arg_arg1_head_arg0_arg0_head_arg0;
                                                                                                                                                                                                                                                                                                          arg_arg1_head_arg0_arg0_head_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                                                                                                                                                                                                                          arg_arg1_head_arg0_arg0_head_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                                                                                                                                                                                                                          if (ATgetInt((ATermInt)arg_arg1_head_arg0_arg0_head_arg0_head) == 44) {

                                                                                                                                                                                                                                                                                                            if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0_list) == ATtrue) {
                                                                                                                                                                                                                                                                                                              if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_list) == ATtrue) {
                                                                                                                                                                                                                                                                                                                ATerm arg_arg1_head_arg0_arg1 = ATgetArgument(arg_arg1_head_arg0, 1);
                                                                                                                                                                                                                                                                                                                /* checking for: lit */
                                                                                                                                                                                                                                                                                                                if (ATgetType((ATerm)arg_arg1_head_arg0_arg1) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg1) == PERR_afun9) {
                                                                                                                                                                                                                                                                                                                  ATerm arg_arg1_head_arg0_arg1_arg0 = ATgetArgument(arg_arg1_head_arg0_arg1, 0);
                                                                                                                                                                                                                                                                                                                  /* checking for: , */
                                                                                                                                                                                                                                                                                                                  if (ATgetType((ATerm)arg_arg1_head_arg0_arg1_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg1_arg0) == PERR_afun43) {
                                                                                                                                                                                                                                                                                                                    ATerm arg_arg1_head_arg0_arg2 = ATgetArgument(arg_arg1_head_arg0, 2);
                                                                                                                                                                                                                                                                                                                    /* checking for: no-attrs */
                                                                                                                                                                                                                                                                                                                    if (ATgetType((ATerm)arg_arg1_head_arg0_arg2) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg2) == PERR_afun5) {
                                                                                                                                                                                                                                                                                                                      ATerm arg_arg1_head_arg1 = ATgetArgument(arg_arg1_head, 1);
                                                                                                                                                                                                                                                                                                                      if (ATgetType((ATerm)arg_arg1_head_arg1) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg1) == ATfalse) {
                                                                                                                                                                                                                                                                                                                        ATerm arg_arg1_head_arg1_head;
                                                                                                                                                                                                                                                                                                                        ATermList arg_arg1_head_arg1_list = (ATermList)arg_arg1_head_arg1;
                                                                                                                                                                                                                                                                                                                        arg_arg1_head_arg1_head = ATgetFirst(arg_arg1_head_arg1_list);
                                                                                                                                                                                                                                                                                                                        arg_arg1_head_arg1_list = ATgetNext(arg_arg1_head_arg1_list);
                                                                                                                                                                                                                                                                                                                        if (ATgetInt((ATermInt)arg_arg1_head_arg1_head) == 44) {

                                                                                                                                                                                                                                                                                                                          if (ATgetType((ATerm)arg_arg1_head_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg1_list) == ATtrue) {
                                                                                                                                                                                                                                                                                                                            if (ATgetType((ATerm)arg_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_list) == ATfalse) {
                                                                                                                                                                                                                                                                                                                              arg_arg1_list = ATgetNext(arg_arg1_list);
                                                                                                                                                                                                                                                                                                                              if (ATgetType((ATerm)arg_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_list) == ATfalse) {
                                                                                                                                                                                                                                                                                                                                arg_arg1_list = ATgetNext(arg_arg1_list);
                                                                                                                                                                                                                                                                                                                                if (ATgetType((ATerm)arg_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_list) == ATfalse) {
                                                                                                                                                                                                                                                                                                                                  arg_arg1_list = ATgetNext(arg_arg1_list);
                                                                                                                                                                                                                                                                                                                                  if (ATgetType((ATerm)arg_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_list) == ATfalse) {
                                                                                                                                                                                                                                                                                                                                    arg_arg1_head = ATgetFirst(arg_arg1_list);
                                                                                                                                                                                                                                                                                                                                    arg_arg1_list = ATgetNext(arg_arg1_list);
                                                                                                                                                                                                                                                                                                                                    /* checking for: appl */
                                                                                                                                                                                                                                                                                                                                    if (ATgetType((ATerm)arg_arg1_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head) == PERR_afun0) {
                                                                                                                                                                                                                                                                                                                                      ATerm arg_arg1_head_arg0 = ATgetArgument(arg_arg1_head, 0);
                                                                                                                                                                                                                                                                                                                                      /* checking for: prod */
                                                                                                                                                                                                                                                                                                                                      if (ATgetType((ATerm)arg_arg1_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0) == PERR_afun1) {
                                                                                                                                                                                                                                                                                                                                        ATerm arg_arg1_head_arg0_arg0 = ATgetArgument(arg_arg1_head_arg0, 0);
                                                                                                                                                                                                                                                                                                                                        if (ATgetType((ATerm)arg_arg1_head_arg0_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0) == ATfalse) {
                                                                                                                                                                                                                                                                                                                                          ATerm arg_arg1_head_arg0_arg0_head;
                                                                                                                                                                                                                                                                                                                                          ATermList arg_arg1_head_arg0_arg0_list = (ATermList)arg_arg1_head_arg0_arg0;
                                                                                                                                                                                                                                                                                                                                          arg_arg1_head_arg0_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_list);
                                                                                                                                                                                                                                                                                                                                          arg_arg1_head_arg0_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_list);
                                                                                                                                                                                                                                                                                                                                          /* checking for: char-class */
                                                                                                                                                                                                                                                                                                                                          if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg0_head) == PERR_afun17) {
                                                                                                                                                                                                                                                                                                                                            ATerm arg_arg1_head_arg0_arg0_head_arg0 = ATgetArgument(arg_arg1_head_arg0_arg0_head, 0);
                                                                                                                                                                                                                                                                                                                                            if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0) == ATfalse) {
                                                                                                                                                                                                                                                                                                                                              ATerm arg_arg1_head_arg0_arg0_head_arg0_head;
                                                                                                                                                                                                                                                                                                                                              ATermList arg_arg1_head_arg0_arg0_head_arg0_list = (ATermList)arg_arg1_head_arg0_arg0_head_arg0;
                                                                                                                                                                                                                                                                                                                                              arg_arg1_head_arg0_arg0_head_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                                                                                                                                                                                                                                                              arg_arg1_head_arg0_arg0_head_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                                                                                                                                                                                                                                                              if (ATgetInt((ATermInt)arg_arg1_head_arg0_arg0_head_arg0_head) == 41) {

                                                                                                                                                                                                                                                                                                                                                if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0_list) == ATtrue) {
                                                                                                                                                                                                                                                                                                                                                  if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_list) == ATtrue) {
                                                                                                                                                                                                                                                                                                                                                    ATerm arg_arg1_head_arg0_arg1 = ATgetArgument(arg_arg1_head_arg0, 1);
                                                                                                                                                                                                                                                                                                                                                    /* checking for: lit */
                                                                                                                                                                                                                                                                                                                                                    if (ATgetType((ATerm)arg_arg1_head_arg0_arg1) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg1) == PERR_afun9) {
                                                                                                                                                                                                                                                                                                                                                      ATerm arg_arg1_head_arg0_arg1_arg0 = ATgetArgument(arg_arg1_head_arg0_arg1, 0);
                                                                                                                                                                                                                                                                                                                                                      /* checking for: ) */
                                                                                                                                                                                                                                                                                                                                                      if (ATgetType((ATerm)arg_arg1_head_arg0_arg1_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg1_arg0) == PERR_afun38) {
                                                                                                                                                                                                                                                                                                                                                        ATerm arg_arg1_head_arg0_arg2 = ATgetArgument(arg_arg1_head_arg0, 2);
                                                                                                                                                                                                                                                                                                                                                        /* checking for: no-attrs */
                                                                                                                                                                                                                                                                                                                                                        if (ATgetType((ATerm)arg_arg1_head_arg0_arg2) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg2) == PERR_afun5) {
                                                                                                                                                                                                                                                                                                                                                          ATerm arg_arg1_head_arg1 = ATgetArgument(arg_arg1_head, 1);
                                                                                                                                                                                                                                                                                                                                                          if (ATgetType((ATerm)arg_arg1_head_arg1) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg1) == ATfalse) {
                                                                                                                                                                                                                                                                                                                                                            ATerm arg_arg1_head_arg1_head;
                                                                                                                                                                                                                                                                                                                                                            ATermList arg_arg1_head_arg1_list = (ATermList)arg_arg1_head_arg1;
                                                                                                                                                                                                                                                                                                                                                            arg_arg1_head_arg1_head = ATgetFirst(arg_arg1_head_arg1_list);
                                                                                                                                                                                                                                                                                                                                                            arg_arg1_head_arg1_list = ATgetNext(arg_arg1_head_arg1_list);
                                                                                                                                                                                                                                                                                                                                                            if (ATgetInt((ATermInt)arg_arg1_head_arg1_head) == 41) {

                                                                                                                                                                                                                                                                                                                                                              if (ATgetType((ATerm)arg_arg1_head_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg1_list) == ATtrue) {
                                                                                                                                                                                                                                                                                                                                                                if (ATgetType((ATerm)arg_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_list) == ATtrue) {
                                                                                                                                                                                                                                                                                                                                                                  return ATtrue;
                                                                                                                                                                                                                                                                                                                                                                }
                                                                                                                                                                                                                                                                                                                                                              }
                                                                                                                                                                                                                                                                                                                                                            }
                                                                                                                                                                                                                                                                                                                                                          }
                                                                                                                                                                                                                                                                                                                                                        }
                                                                                                                                                                                                                                                                                                                                                      }
                                                                                                                                                                                                                                                                                                                                                    }
                                                                                                                                                                                                                                                                                                                                                  }
                                                                                                                                                                                                                                                                                                                                                }
                                                                                                                                                                                                                                                                                                                                              }
                                                                                                                                                                                                                                                                                                                                            }
                                                                                                                                                                                                                                                                                                                                          }
                                                                                                                                                                                                                                                                                                                                        }
                                                                                                                                                                                                                                                                                                                                      }
                                                                                                                                                                                                                                                                                                                                    }
                                                                                                                                                                                                                                                                                                                                  }
                                                                                                                                                                                                                                                                                                                                }
                                                                                                                                                                                                                                                                                                                              }
                                                                                                                                                                                                                                                                                                                            }
                                                                                                                                                                                                                                                                                                                          }
                                                                                                                                                                                                                                                                                                                        }
                                                                                                                                                                                                                                                                                                                      }
                                                                                                                                                                                                                                                                                                                    }
                                                                                                                                                                                                                                                                                                                  }
                                                                                                                                                                                                                                                                                                                }
                                                                                                                                                                                                                                                                                                              }
                                                                                                                                                                                                                                                                                                            }
                                                                                                                                                                                                                                                                                                          }
                                                                                                                                                                                                                                                                                                        }
                                                                                                                                                                                                                                                                                                      }
                                                                                                                                                                                                                                                                                                    }
                                                                                                                                                                                                                                                                                                  }
                                                                                                                                                                                                                                                                                                }
                                                                                                                                                                                                                                                                                              }
                                                                                                                                                                                                                                                                                            }
                                                                                                                                                                                                                                                                                          }
                                                                                                                                                                                                                                                                                        }
                                                                                                                                                                                                                                                                                      }
                                                                                                                                                                                                                                                                                    }
                                                                                                                                                                                                                                                                                  }
                                                                                                                                                                                                                                                                                }
                                                                                                                                                                                                                                                                              }
                                                                                                                                                                                                                                                                            }
                                                                                                                                                                                                                                                                          }
                                                                                                                                                                                                                                                                        }
                                                                                                                                                                                                                                                                      }
                                                                                                                                                                                                                                                                    }
                                                                                                                                                                                                                                                                  }
                                                                                                                                                                                                                                                                }
                                                                                                                                                                                                                                                              }
                                                                                                                                                                                                                                                            }
                                                                                                                                                                                                                                                          }
                                                                                                                                                                                                                                                        }
                                                                                                                                                                                                                                                      }
                                                                                                                                                                                                                                                    }
                                                                                                                                                                                                                                                  }
                                                                                                                                                                                                                                                }
                                                                                                                                                                                                                                              }
                                                                                                                                                                                                                                            }
                                                                                                                                                                                                                                          }
                                                                                                                                                                                                                                        }
                                                                                                                                                                                                                                      }
                                                                                                                                                                                                                                    }
                                                                                                                                                                                                                                  }
                                                                                                                                                                                                                                }
                                                                                                                                                                                                                              }
                                                                                                                                                                                                                            }
                                                                                                                                                                                                                          }
                                                                                                                                                                                                                        }
                                                                                                                                                                                                                      }
                                                                                                                                                                                                                    }
                                                                                                                                                                                                                  }
                                                                                                                                                                                                                }
                                                                                                                                                                                                              }
                                                                                                                                                                                                            }
                                                                                                                                                                                                          }
                                                                                                                                                                                                        }
                                                                                                                                                                                                      }
                                                                                                                                                                                                    }
                                                                                                                                                                                                  }
                                                                                                                                                                                                }
                                                                                                                                                                                              }
                                                                                                                                                                                            }
                                                                                                                                                                                          }
                                                                                                                                                                                        }
                                                                                                                                                                                      }
                                                                                                                                                                                    }
                                                                                                                                                                                  }
                                                                                                                                                                                }
                                                                                                                                                                              }
                                                                                                                                                                            }
                                                                                                                                                                          }
                                                                                                                                                                        }
                                                                                                                                                                      }
                                                                                                                                                                    }
                                                                                                                                                                  }
                                                                                                                                                                }
                                                                                                                                                              }
                                                                                                                                                            }
                                                                                                                                                          }
                                                                                                                                                        }
                                                                                                                                                      }
                                                                                                                                                    }
                                                                                                                                                  }
                                                                                                                                                }
                                                                                                                                              }
                                                                                                                                            }
                                                                                                                                          }
                                                                                                                                        }
                                                                                                                                      }
                                                                                                                                    }
                                                                                                                                  }
                                                                                                                                }
                                                                                                                              }
                                                                                                                            }
                                                                                                                          }
                                                                                                                        }
                                                                                                                      }
                                                                                                                    }
                                                                                                                  }
                                                                                                                }
                                                                                                              }
                                                                                                            }
                                                                                                          }
                                                                                                        }
                                                                                                      }
                                                                                                    }
                                                                                                  }
                                                                                                }
                                                                                              }
                                                                                            }
                                                                                          }
                                                                                        }
                                                                                      }
                                                                                    }
                                                                                  }
                                                                                }
                                                                              }
                                                                            }
                                                                          }
                                                                        }
                                                                      }
                                                                    }
                                                                  }
                                                                }
                                                              }
                                                            }
                                                          }
                                                        }
                                                      }
                                                    }
                                                  }
                                                }
                                              }
                                            }
                                          }
                                        }
                                      }
                                    }
                                  }
                                }
                              }
                            }
                          }
                        }
                      }
                    }
                  }
                }
              }
            }
          }
        }
      }
    }
  }
  return ATfalse;
}

/**
 * Assert whether a PERR_Subject has a ws-after-subject. 
 * \param[in] arg input PERR_Subject
 * \return ATtrue if the PERR_Subject had a ws-after-subject, or ATfalse otherwise
 */
ATbool PERR_hasSubjectWsAfterSubject(PERR_Subject arg) {
  if (PERR_isSubjectSubject(arg)) {
    return ATtrue;
  }
  return ATfalse;
}

/**
 * Assert whether a PERR_Subject has a ws-after-(. 
 * \param[in] arg input PERR_Subject
 * \return ATtrue if the PERR_Subject had a ws-after-(, or ATfalse otherwise
 */
ATbool PERR_hasSubjectWsAfterParenOpen(PERR_Subject arg) {
  if (PERR_isSubjectSubject(arg)) {
    return ATtrue;
  }
  else if (PERR_isSubjectLocalized(arg)) {
    return ATtrue;
  }
  return ATfalse;
}

/**
 * Assert whether a PERR_Subject has a description. 
 * \param[in] arg input PERR_Subject
 * \return ATtrue if the PERR_Subject had a description, or ATfalse otherwise
 */
ATbool PERR_hasSubjectDescription(PERR_Subject arg) {
  if (PERR_isSubjectSubject(arg)) {
    return ATtrue;
  }
  else if (PERR_isSubjectLocalized(arg)) {
    return ATtrue;
  }
  return ATfalse;
}

/**
 * Assert whether a PERR_Subject has a ws-after-description. 
 * \param[in] arg input PERR_Subject
 * \return ATtrue if the PERR_Subject had a ws-after-description, or ATfalse otherwise
 */
ATbool PERR_hasSubjectWsAfterDescription(PERR_Subject arg) {
  if (PERR_isSubjectSubject(arg)) {
    return ATtrue;
  }
  else if (PERR_isSubjectLocalized(arg)) {
    return ATtrue;
  }
  return ATfalse;
}

/**
 * Assert whether a PERR_Subject has a ws-after-localized. 
 * \param[in] arg input PERR_Subject
 * \return ATtrue if the PERR_Subject had a ws-after-localized, or ATfalse otherwise
 */
ATbool PERR_hasSubjectWsAfterLocalized(PERR_Subject arg) {
  if (PERR_isSubjectLocalized(arg)) {
    return ATtrue;
  }
  return ATfalse;
}

/**
 * Assert whether a PERR_Subject has a ws-after-,. 
 * \param[in] arg input PERR_Subject
 * \return ATtrue if the PERR_Subject had a ws-after-,, or ATfalse otherwise
 */
ATbool PERR_hasSubjectWsAfterComma(PERR_Subject arg) {
  if (PERR_isSubjectLocalized(arg)) {
    return ATtrue;
  }
  return ATfalse;
}

/**
 * Assert whether a PERR_Subject has a Location. 
 * \param[in] arg input PERR_Subject
 * \return ATtrue if the PERR_Subject had a Location, or ATfalse otherwise
 */
ATbool PERR_hasSubjectLocation(PERR_Subject arg) {
  if (PERR_isSubjectLocalized(arg)) {
    return ATtrue;
  }
  return ATfalse;
}

/**
 * Assert whether a PERR_Subject has a ws-after-Location. 
 * \param[in] arg input PERR_Subject
 * \return ATtrue if the PERR_Subject had a ws-after-Location, or ATfalse otherwise
 */
ATbool PERR_hasSubjectWsAfterLocation(PERR_Subject arg) {
  if (PERR_isSubjectLocalized(arg)) {
    return ATtrue;
  }
  return ATfalse;
}

/**
 * Get the ws-after-subject PERR_OptLayout of a PERR_Subject. Note that the precondition is that this PERR_Subject actually has a ws-after-subject
 * \param[in] arg input PERR_Subject
 * \return the ws-after-subject of #arg, if it exist or an undefined value if it does not
 */
PERR_OptLayout PERR_getSubjectWsAfterSubject(PERR_Subject arg) {
  
    return (PERR_OptLayout)ATelementAt((ATermList)ATgetArgument((ATermAppl)arg, 1), 1);
}

/**
 * Get the ws-after-( PERR_OptLayout of a PERR_Subject. Note that the precondition is that this PERR_Subject actually has a ws-after-(
 * \param[in] arg input PERR_Subject
 * \return the ws-after-( of #arg, if it exist or an undefined value if it does not
 */
PERR_OptLayout PERR_getSubjectWsAfterParenOpen(PERR_Subject arg) {
  if (PERR_isSubjectSubject(arg)) {
    return (PERR_OptLayout)ATelementAt((ATermList)ATgetArgument((ATermAppl)arg, 1), 3);
  }
  else 
    return (PERR_OptLayout)ATelementAt((ATermList)ATgetArgument((ATermAppl)arg, 1), 3);
}

/**
 * Get the description PERR_StrCon of a PERR_Subject. Note that the precondition is that this PERR_Subject actually has a description
 * \param[in] arg input PERR_Subject
 * \return the description of #arg, if it exist or an undefined value if it does not
 */
PERR_StrCon PERR_getSubjectDescription(PERR_Subject arg) {
  if (PERR_isSubjectSubject(arg)) {
    return (PERR_StrCon)ATelementAt((ATermList)ATgetArgument((ATermAppl)arg, 1), 4);
  }
  else 
    return (PERR_StrCon)ATelementAt((ATermList)ATgetArgument((ATermAppl)arg, 1), 4);
}

/**
 * Get the ws-after-description PERR_OptLayout of a PERR_Subject. Note that the precondition is that this PERR_Subject actually has a ws-after-description
 * \param[in] arg input PERR_Subject
 * \return the ws-after-description of #arg, if it exist or an undefined value if it does not
 */
PERR_OptLayout PERR_getSubjectWsAfterDescription(PERR_Subject arg) {
  if (PERR_isSubjectSubject(arg)) {
    return (PERR_OptLayout)ATelementAt((ATermList)ATgetArgument((ATermAppl)arg, 1), 5);
  }
  else 
    return (PERR_OptLayout)ATelementAt((ATermList)ATgetArgument((ATermAppl)arg, 1), 5);
}

/**
 * Get the ws-after-localized PERR_OptLayout of a PERR_Subject. Note that the precondition is that this PERR_Subject actually has a ws-after-localized
 * \param[in] arg input PERR_Subject
 * \return the ws-after-localized of #arg, if it exist or an undefined value if it does not
 */
PERR_OptLayout PERR_getSubjectWsAfterLocalized(PERR_Subject arg) {
  
    return (PERR_OptLayout)ATelementAt((ATermList)ATgetArgument((ATermAppl)arg, 1), 1);
}

/**
 * Get the ws-after-, PERR_OptLayout of a PERR_Subject. Note that the precondition is that this PERR_Subject actually has a ws-after-,
 * \param[in] arg input PERR_Subject
 * \return the ws-after-, of #arg, if it exist or an undefined value if it does not
 */
PERR_OptLayout PERR_getSubjectWsAfterComma(PERR_Subject arg) {
  
    return (PERR_OptLayout)ATelementAt((ATermList)ATgetArgument((ATermAppl)arg, 1), 7);
}

/**
 * Get the Location PERR_Location of a PERR_Subject. Note that the precondition is that this PERR_Subject actually has a Location
 * \param[in] arg input PERR_Subject
 * \return the Location of #arg, if it exist or an undefined value if it does not
 */
PERR_Location PERR_getSubjectLocation(PERR_Subject arg) {
  
    return (PERR_Location)ATelementAt((ATermList)ATgetArgument((ATermAppl)arg, 1), 8);
}

/**
 * Get the ws-after-Location PERR_OptLayout of a PERR_Subject. Note that the precondition is that this PERR_Subject actually has a ws-after-Location
 * \param[in] arg input PERR_Subject
 * \return the ws-after-Location of #arg, if it exist or an undefined value if it does not
 */
PERR_OptLayout PERR_getSubjectWsAfterLocation(PERR_Subject arg) {
  
    return (PERR_OptLayout)ATelementAt((ATermList)ATgetArgument((ATermAppl)arg, 1), 9);
}

/**
 * Set the ws-after-subject of a PERR_Subject. The precondition being that this PERR_Subject actually has a ws-after-subject
 * \param[in] arg input PERR_Subject
 * \param[in] wsAfterSubject new PERR_OptLayout to set in #arg
 * \return A new PERR_Subject with wsAfterSubject at the right place, or a core dump if #arg did not have a wsAfterSubject
 */
PERR_Subject PERR_setSubjectWsAfterSubject(PERR_Subject arg, PERR_OptLayout wsAfterSubject) {
  if (PERR_isSubjectSubject(arg)) {
    return (PERR_Subject)ATsetArgument((ATermAppl)arg, (ATerm)ATreplace((ATermList)ATgetArgument((ATermAppl)arg, 1), (ATerm)((ATerm) wsAfterSubject), 1), 1);
  }

  ATabort("Subject has no WsAfterSubject: %t\n", arg);
  return (PERR_Subject)NULL;
}

/**
 * Set the ws-after-( of a PERR_Subject. The precondition being that this PERR_Subject actually has a ws-after-(
 * \param[in] arg input PERR_Subject
 * \param[in] wsAfterParenOpen new PERR_OptLayout to set in #arg
 * \return A new PERR_Subject with wsAfterParenOpen at the right place, or a core dump if #arg did not have a wsAfterParenOpen
 */
PERR_Subject PERR_setSubjectWsAfterParenOpen(PERR_Subject arg, PERR_OptLayout wsAfterParenOpen) {
  if (PERR_isSubjectSubject(arg)) {
    return (PERR_Subject)ATsetArgument((ATermAppl)arg, (ATerm)ATreplace((ATermList)ATgetArgument((ATermAppl)arg, 1), (ATerm)((ATerm) wsAfterParenOpen), 3), 1);
  }
  else if (PERR_isSubjectLocalized(arg)) {
    return (PERR_Subject)ATsetArgument((ATermAppl)arg, (ATerm)ATreplace((ATermList)ATgetArgument((ATermAppl)arg, 1), (ATerm)((ATerm) wsAfterParenOpen), 3), 1);
  }

  ATabort("Subject has no WsAfterParenOpen: %t\n", arg);
  return (PERR_Subject)NULL;
}

/**
 * Set the description of a PERR_Subject. The precondition being that this PERR_Subject actually has a description
 * \param[in] arg input PERR_Subject
 * \param[in] description new PERR_StrCon to set in #arg
 * \return A new PERR_Subject with description at the right place, or a core dump if #arg did not have a description
 */
PERR_Subject PERR_setSubjectDescription(PERR_Subject arg, PERR_StrCon description) {
  if (PERR_isSubjectSubject(arg)) {
    return (PERR_Subject)ATsetArgument((ATermAppl)arg, (ATerm)ATreplace((ATermList)ATgetArgument((ATermAppl)arg, 1), (ATerm)((ATerm) description), 4), 1);
  }
  else if (PERR_isSubjectLocalized(arg)) {
    return (PERR_Subject)ATsetArgument((ATermAppl)arg, (ATerm)ATreplace((ATermList)ATgetArgument((ATermAppl)arg, 1), (ATerm)((ATerm) description), 4), 1);
  }

  ATabort("Subject has no Description: %t\n", arg);
  return (PERR_Subject)NULL;
}

/**
 * Set the ws-after-description of a PERR_Subject. The precondition being that this PERR_Subject actually has a ws-after-description
 * \param[in] arg input PERR_Subject
 * \param[in] wsAfterDescription new PERR_OptLayout to set in #arg
 * \return A new PERR_Subject with wsAfterDescription at the right place, or a core dump if #arg did not have a wsAfterDescription
 */
PERR_Subject PERR_setSubjectWsAfterDescription(PERR_Subject arg, PERR_OptLayout wsAfterDescription) {
  if (PERR_isSubjectSubject(arg)) {
    return (PERR_Subject)ATsetArgument((ATermAppl)arg, (ATerm)ATreplace((ATermList)ATgetArgument((ATermAppl)arg, 1), (ATerm)((ATerm) wsAfterDescription), 5), 1);
  }
  else if (PERR_isSubjectLocalized(arg)) {
    return (PERR_Subject)ATsetArgument((ATermAppl)arg, (ATerm)ATreplace((ATermList)ATgetArgument((ATermAppl)arg, 1), (ATerm)((ATerm) wsAfterDescription), 5), 1);
  }

  ATabort("Subject has no WsAfterDescription: %t\n", arg);
  return (PERR_Subject)NULL;
}

/**
 * Set the ws-after-localized of a PERR_Subject. The precondition being that this PERR_Subject actually has a ws-after-localized
 * \param[in] arg input PERR_Subject
 * \param[in] wsAfterLocalized new PERR_OptLayout to set in #arg
 * \return A new PERR_Subject with wsAfterLocalized at the right place, or a core dump if #arg did not have a wsAfterLocalized
 */
PERR_Subject PERR_setSubjectWsAfterLocalized(PERR_Subject arg, PERR_OptLayout wsAfterLocalized) {
  if (PERR_isSubjectLocalized(arg)) {
    return (PERR_Subject)ATsetArgument((ATermAppl)arg, (ATerm)ATreplace((ATermList)ATgetArgument((ATermAppl)arg, 1), (ATerm)((ATerm) wsAfterLocalized), 1), 1);
  }

  ATabort("Subject has no WsAfterLocalized: %t\n", arg);
  return (PERR_Subject)NULL;
}

/**
 * Set the ws-after-, of a PERR_Subject. The precondition being that this PERR_Subject actually has a ws-after-,
 * \param[in] arg input PERR_Subject
 * \param[in] wsAfterComma new PERR_OptLayout to set in #arg
 * \return A new PERR_Subject with wsAfterComma at the right place, or a core dump if #arg did not have a wsAfterComma
 */
PERR_Subject PERR_setSubjectWsAfterComma(PERR_Subject arg, PERR_OptLayout wsAfterComma) {
  if (PERR_isSubjectLocalized(arg)) {
    return (PERR_Subject)ATsetArgument((ATermAppl)arg, (ATerm)ATreplace((ATermList)ATgetArgument((ATermAppl)arg, 1), (ATerm)((ATerm) wsAfterComma), 7), 1);
  }

  ATabort("Subject has no WsAfterComma: %t\n", arg);
  return (PERR_Subject)NULL;
}

/**
 * Set the Location of a PERR_Subject. The precondition being that this PERR_Subject actually has a Location
 * \param[in] arg input PERR_Subject
 * \param[in] Location new PERR_Location to set in #arg
 * \return A new PERR_Subject with Location at the right place, or a core dump if #arg did not have a Location
 */
PERR_Subject PERR_setSubjectLocation(PERR_Subject arg, PERR_Location Location) {
  if (PERR_isSubjectLocalized(arg)) {
    return (PERR_Subject)ATsetArgument((ATermAppl)arg, (ATerm)ATreplace((ATermList)ATgetArgument((ATermAppl)arg, 1), (ATerm)((ATerm) Location), 8), 1);
  }

  ATabort("Subject has no Location: %t\n", arg);
  return (PERR_Subject)NULL;
}

/**
 * Set the ws-after-Location of a PERR_Subject. The precondition being that this PERR_Subject actually has a ws-after-Location
 * \param[in] arg input PERR_Subject
 * \param[in] wsAfterLocation new PERR_OptLayout to set in #arg
 * \return A new PERR_Subject with wsAfterLocation at the right place, or a core dump if #arg did not have a wsAfterLocation
 */
PERR_Subject PERR_setSubjectWsAfterLocation(PERR_Subject arg, PERR_OptLayout wsAfterLocation) {
  if (PERR_isSubjectLocalized(arg)) {
    return (PERR_Subject)ATsetArgument((ATermAppl)arg, (ATerm)ATreplace((ATermList)ATgetArgument((ATermAppl)arg, 1), (ATerm)((ATerm) wsAfterLocation), 9), 1);
  }

  ATabort("Subject has no WsAfterLocation: %t\n", arg);
  return (PERR_Subject)NULL;
}

/**
 * Assert whether a PERR_Error is any of the valid alternatives, or not. This analysis does not go any deeper than the top level
 * \param[in] arg input PERR_Error
 * \return ATtrue if #arg corresponds to the expected signature, or ATfalse otherwise
 */
ATbool PERR_isValidError(PERR_Error arg) {
  if (PERR_isErrorInfo(arg)) {
    return ATtrue;
  }
  else if (PERR_isErrorWarning(arg)) {
    return ATtrue;
  }
  else if (PERR_isErrorError(arg)) {
    return ATtrue;
  }
  else if (PERR_isErrorFatal(arg)) {
    return ATtrue;
  }
  return ATfalse;
}

/**
 * Assert whether a PERR_Error is a info by checking against the following ATerm pattern: appl(prod([lit("info"),cf(opt(layout)),lit("("),cf(opt(layout)),cf(sort("StrCon")),cf(opt(layout)),lit(","),cf(opt(layout)),lit("["),cf(opt(layout)),cf(iter-star-sep(sort("Subject"),lit(","))),cf(opt(layout)),lit("]"),cf(opt(layout)),lit(")")],cf(sort("Error")),attrs([term(cons("info"))])),[appl(prod([char-class([105]),char-class([110]),char-class([102]),char-class([111])],lit("info"),no-attrs),[105,110,102,111]),<"ws-after-info"("OptLayout")>,appl(prod([char-class([40])],lit("("),no-attrs),[40]),<"ws-after-("("OptLayout")>,<"description"("StrCon")>,<"ws-after-description"("OptLayout")>,appl(prod([char-class([44])],lit(","),no-attrs),[44]),<"ws-after-,"("OptLayout")>,appl(prod([char-class([91])],lit("["),no-attrs),[91]),<"ws-after-["("OptLayout")>,appl(list(cf(iter-star-sep(sort("Subject"),lit(",")))),<"list"("Subject-list")>),<"ws-after-list"("OptLayout")>,appl(prod([char-class([93])],lit("]"),no-attrs),[93]),<"ws-after-]"("OptLayout")>,appl(prod([char-class([41])],lit(")"),no-attrs),[41])]). May not be used to assert correctness of the PERR_Error
 * \param[in] arg input PERR_Error
 * \return ATtrue if #arg corresponds to the signature of a info, or ATfalse otherwise
 */
inline ATbool PERR_isErrorInfo(PERR_Error arg){
  /* checking for: appl */
  if (ATgetType((ATerm)arg) == AT_APPL && ATgetAFun((ATermAppl)arg) == PERR_afun0) {
    ATerm arg_arg0 = ATgetArgument(arg, 0);
    /* checking for: prod */
    if (ATgetType((ATerm)arg_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0) == PERR_afun1) {
      ATerm arg_arg0_arg0 = ATgetArgument(arg_arg0, 0);
      if (ATgetType((ATerm)arg_arg0_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0) == ATfalse) {
        ATerm arg_arg0_arg0_head;
        ATermList arg_arg0_arg0_list = (ATermList)arg_arg0_arg0;
        arg_arg0_arg0_head = ATgetFirst(arg_arg0_arg0_list);
        arg_arg0_arg0_list = ATgetNext(arg_arg0_arg0_list);
        /* checking for: lit */
        if (ATgetType((ATerm)arg_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head) == PERR_afun9) {
          ATerm arg_arg0_arg0_head_arg0 = ATgetArgument(arg_arg0_arg0_head, 0);
          /* checking for: info */
          if (ATgetType((ATerm)arg_arg0_arg0_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0) == PERR_afun48) {
            if (ATgetType((ATerm)arg_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0_list) == ATfalse) {
              arg_arg0_arg0_head = ATgetFirst(arg_arg0_arg0_list);
              arg_arg0_arg0_list = ATgetNext(arg_arg0_arg0_list);
              /* checking for: cf */
              if (ATgetType((ATerm)arg_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head) == PERR_afun2) {
                ATerm arg_arg0_arg0_head_arg0 = ATgetArgument(arg_arg0_arg0_head, 0);
                /* checking for: opt */
                if (ATgetType((ATerm)arg_arg0_arg0_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0) == PERR_afun3) {
                  ATerm arg_arg0_arg0_head_arg0_arg0 = ATgetArgument(arg_arg0_arg0_head_arg0, 0);
                  /* checking for: layout */
                  if (ATgetType((ATerm)arg_arg0_arg0_head_arg0_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0_arg0) == PERR_afun4) {
                    if (ATgetType((ATerm)arg_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0_list) == ATfalse) {
                      arg_arg0_arg0_head = ATgetFirst(arg_arg0_arg0_list);
                      arg_arg0_arg0_list = ATgetNext(arg_arg0_arg0_list);
                      /* checking for: lit */
                      if (ATgetType((ATerm)arg_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head) == PERR_afun9) {
                        ATerm arg_arg0_arg0_head_arg0 = ATgetArgument(arg_arg0_arg0_head, 0);
                        /* checking for: ( */
                        if (ATgetType((ATerm)arg_arg0_arg0_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0) == PERR_afun39) {
                          if (ATgetType((ATerm)arg_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0_list) == ATfalse) {
                            arg_arg0_arg0_head = ATgetFirst(arg_arg0_arg0_list);
                            arg_arg0_arg0_list = ATgetNext(arg_arg0_arg0_list);
                            /* checking for: cf */
                            if (ATgetType((ATerm)arg_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head) == PERR_afun2) {
                              ATerm arg_arg0_arg0_head_arg0 = ATgetArgument(arg_arg0_arg0_head, 0);
                              /* checking for: opt */
                              if (ATgetType((ATerm)arg_arg0_arg0_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0) == PERR_afun3) {
                                ATerm arg_arg0_arg0_head_arg0_arg0 = ATgetArgument(arg_arg0_arg0_head_arg0, 0);
                                /* checking for: layout */
                                if (ATgetType((ATerm)arg_arg0_arg0_head_arg0_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0_arg0) == PERR_afun4) {
                                  if (ATgetType((ATerm)arg_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0_list) == ATfalse) {
                                    arg_arg0_arg0_head = ATgetFirst(arg_arg0_arg0_list);
                                    arg_arg0_arg0_list = ATgetNext(arg_arg0_arg0_list);
                                    /* checking for: cf */
                                    if (ATgetType((ATerm)arg_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head) == PERR_afun2) {
                                      ATerm arg_arg0_arg0_head_arg0 = ATgetArgument(arg_arg0_arg0_head, 0);
                                      /* checking for: sort */
                                      if (ATgetType((ATerm)arg_arg0_arg0_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0) == PERR_afun11) {
                                        ATerm arg_arg0_arg0_head_arg0_arg0 = ATgetArgument(arg_arg0_arg0_head_arg0, 0);
                                        /* checking for: StrCon */
                                        if (ATgetType((ATerm)arg_arg0_arg0_head_arg0_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0_arg0) == PERR_afun29) {
                                          if (ATgetType((ATerm)arg_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0_list) == ATfalse) {
                                            arg_arg0_arg0_head = ATgetFirst(arg_arg0_arg0_list);
                                            arg_arg0_arg0_list = ATgetNext(arg_arg0_arg0_list);
                                            /* checking for: cf */
                                            if (ATgetType((ATerm)arg_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head) == PERR_afun2) {
                                              ATerm arg_arg0_arg0_head_arg0 = ATgetArgument(arg_arg0_arg0_head, 0);
                                              /* checking for: opt */
                                              if (ATgetType((ATerm)arg_arg0_arg0_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0) == PERR_afun3) {
                                                ATerm arg_arg0_arg0_head_arg0_arg0 = ATgetArgument(arg_arg0_arg0_head_arg0, 0);
                                                /* checking for: layout */
                                                if (ATgetType((ATerm)arg_arg0_arg0_head_arg0_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0_arg0) == PERR_afun4) {
                                                  if (ATgetType((ATerm)arg_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0_list) == ATfalse) {
                                                    arg_arg0_arg0_head = ATgetFirst(arg_arg0_arg0_list);
                                                    arg_arg0_arg0_list = ATgetNext(arg_arg0_arg0_list);
                                                    /* checking for: lit */
                                                    if (ATgetType((ATerm)arg_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head) == PERR_afun9) {
                                                      ATerm arg_arg0_arg0_head_arg0 = ATgetArgument(arg_arg0_arg0_head, 0);
                                                      /* checking for: , */
                                                      if (ATgetType((ATerm)arg_arg0_arg0_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0) == PERR_afun43) {
                                                        if (ATgetType((ATerm)arg_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0_list) == ATfalse) {
                                                          arg_arg0_arg0_head = ATgetFirst(arg_arg0_arg0_list);
                                                          arg_arg0_arg0_list = ATgetNext(arg_arg0_arg0_list);
                                                          /* checking for: cf */
                                                          if (ATgetType((ATerm)arg_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head) == PERR_afun2) {
                                                            ATerm arg_arg0_arg0_head_arg0 = ATgetArgument(arg_arg0_arg0_head, 0);
                                                            /* checking for: opt */
                                                            if (ATgetType((ATerm)arg_arg0_arg0_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0) == PERR_afun3) {
                                                              ATerm arg_arg0_arg0_head_arg0_arg0 = ATgetArgument(arg_arg0_arg0_head_arg0, 0);
                                                              /* checking for: layout */
                                                              if (ATgetType((ATerm)arg_arg0_arg0_head_arg0_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0_arg0) == PERR_afun4) {
                                                                if (ATgetType((ATerm)arg_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0_list) == ATfalse) {
                                                                  arg_arg0_arg0_head = ATgetFirst(arg_arg0_arg0_list);
                                                                  arg_arg0_arg0_list = ATgetNext(arg_arg0_arg0_list);
                                                                  /* checking for: lit */
                                                                  if (ATgetType((ATerm)arg_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head) == PERR_afun9) {
                                                                    ATerm arg_arg0_arg0_head_arg0 = ATgetArgument(arg_arg0_arg0_head, 0);
                                                                    /* checking for: [ */
                                                                    if (ATgetType((ATerm)arg_arg0_arg0_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0) == PERR_afun47) {
                                                                      if (ATgetType((ATerm)arg_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0_list) == ATfalse) {
                                                                        arg_arg0_arg0_head = ATgetFirst(arg_arg0_arg0_list);
                                                                        arg_arg0_arg0_list = ATgetNext(arg_arg0_arg0_list);
                                                                        /* checking for: cf */
                                                                        if (ATgetType((ATerm)arg_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head) == PERR_afun2) {
                                                                          ATerm arg_arg0_arg0_head_arg0 = ATgetArgument(arg_arg0_arg0_head, 0);
                                                                          /* checking for: opt */
                                                                          if (ATgetType((ATerm)arg_arg0_arg0_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0) == PERR_afun3) {
                                                                            ATerm arg_arg0_arg0_head_arg0_arg0 = ATgetArgument(arg_arg0_arg0_head_arg0, 0);
                                                                            /* checking for: layout */
                                                                            if (ATgetType((ATerm)arg_arg0_arg0_head_arg0_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0_arg0) == PERR_afun4) {
                                                                              if (ATgetType((ATerm)arg_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0_list) == ATfalse) {
                                                                                arg_arg0_arg0_head = ATgetFirst(arg_arg0_arg0_list);
                                                                                arg_arg0_arg0_list = ATgetNext(arg_arg0_arg0_list);
                                                                                /* checking for: cf */
                                                                                if (ATgetType((ATerm)arg_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head) == PERR_afun2) {
                                                                                  ATerm arg_arg0_arg0_head_arg0 = ATgetArgument(arg_arg0_arg0_head, 0);
                                                                                  /* checking for: iter-star-sep */
                                                                                  if (ATgetType((ATerm)arg_arg0_arg0_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0) == PERR_afun46) {
                                                                                    ATerm arg_arg0_arg0_head_arg0_arg0 = ATgetArgument(arg_arg0_arg0_head_arg0, 0);
                                                                                    /* checking for: sort */
                                                                                    if (ATgetType((ATerm)arg_arg0_arg0_head_arg0_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0_arg0) == PERR_afun11) {
                                                                                      ATerm arg_arg0_arg0_head_arg0_arg0_arg0 = ATgetArgument(arg_arg0_arg0_head_arg0_arg0, 0);
                                                                                      /* checking for: Subject */
                                                                                      if (ATgetType((ATerm)arg_arg0_arg0_head_arg0_arg0_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0_arg0_arg0) == PERR_afun41) {
                                                                                        ATerm arg_arg0_arg0_head_arg0_arg1 = ATgetArgument(arg_arg0_arg0_head_arg0, 1);
                                                                                        /* checking for: lit */
                                                                                        if (ATgetType((ATerm)arg_arg0_arg0_head_arg0_arg1) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0_arg1) == PERR_afun9) {
                                                                                          ATerm arg_arg0_arg0_head_arg0_arg1_arg0 = ATgetArgument(arg_arg0_arg0_head_arg0_arg1, 0);
                                                                                          /* checking for: , */
                                                                                          if (ATgetType((ATerm)arg_arg0_arg0_head_arg0_arg1_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0_arg1_arg0) == PERR_afun43) {
                                                                                            if (ATgetType((ATerm)arg_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0_list) == ATfalse) {
                                                                                              arg_arg0_arg0_head = ATgetFirst(arg_arg0_arg0_list);
                                                                                              arg_arg0_arg0_list = ATgetNext(arg_arg0_arg0_list);
                                                                                              /* checking for: cf */
                                                                                              if (ATgetType((ATerm)arg_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head) == PERR_afun2) {
                                                                                                ATerm arg_arg0_arg0_head_arg0 = ATgetArgument(arg_arg0_arg0_head, 0);
                                                                                                /* checking for: opt */
                                                                                                if (ATgetType((ATerm)arg_arg0_arg0_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0) == PERR_afun3) {
                                                                                                  ATerm arg_arg0_arg0_head_arg0_arg0 = ATgetArgument(arg_arg0_arg0_head_arg0, 0);
                                                                                                  /* checking for: layout */
                                                                                                  if (ATgetType((ATerm)arg_arg0_arg0_head_arg0_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0_arg0) == PERR_afun4) {
                                                                                                    if (ATgetType((ATerm)arg_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0_list) == ATfalse) {
                                                                                                      arg_arg0_arg0_head = ATgetFirst(arg_arg0_arg0_list);
                                                                                                      arg_arg0_arg0_list = ATgetNext(arg_arg0_arg0_list);
                                                                                                      /* checking for: lit */
                                                                                                      if (ATgetType((ATerm)arg_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head) == PERR_afun9) {
                                                                                                        ATerm arg_arg0_arg0_head_arg0 = ATgetArgument(arg_arg0_arg0_head, 0);
                                                                                                        /* checking for: ] */
                                                                                                        if (ATgetType((ATerm)arg_arg0_arg0_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0) == PERR_afun45) {
                                                                                                          if (ATgetType((ATerm)arg_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0_list) == ATfalse) {
                                                                                                            arg_arg0_arg0_head = ATgetFirst(arg_arg0_arg0_list);
                                                                                                            arg_arg0_arg0_list = ATgetNext(arg_arg0_arg0_list);
                                                                                                            /* checking for: cf */
                                                                                                            if (ATgetType((ATerm)arg_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head) == PERR_afun2) {
                                                                                                              ATerm arg_arg0_arg0_head_arg0 = ATgetArgument(arg_arg0_arg0_head, 0);
                                                                                                              /* checking for: opt */
                                                                                                              if (ATgetType((ATerm)arg_arg0_arg0_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0) == PERR_afun3) {
                                                                                                                ATerm arg_arg0_arg0_head_arg0_arg0 = ATgetArgument(arg_arg0_arg0_head_arg0, 0);
                                                                                                                /* checking for: layout */
                                                                                                                if (ATgetType((ATerm)arg_arg0_arg0_head_arg0_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0_arg0) == PERR_afun4) {
                                                                                                                  if (ATgetType((ATerm)arg_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0_list) == ATfalse) {
                                                                                                                    arg_arg0_arg0_head = ATgetFirst(arg_arg0_arg0_list);
                                                                                                                    arg_arg0_arg0_list = ATgetNext(arg_arg0_arg0_list);
                                                                                                                    /* checking for: lit */
                                                                                                                    if (ATgetType((ATerm)arg_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head) == PERR_afun9) {
                                                                                                                      ATerm arg_arg0_arg0_head_arg0 = ATgetArgument(arg_arg0_arg0_head, 0);
                                                                                                                      /* checking for: ) */
                                                                                                                      if (ATgetType((ATerm)arg_arg0_arg0_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0) == PERR_afun38) {
                                                                                                                        if (ATgetType((ATerm)arg_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0_list) == ATtrue) {
                                                                                                                          ATerm arg_arg0_arg1 = ATgetArgument(arg_arg0, 1);
                                                                                                                          /* checking for: cf */
                                                                                                                          if (ATgetType((ATerm)arg_arg0_arg1) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg1) == PERR_afun2) {
                                                                                                                            ATerm arg_arg0_arg1_arg0 = ATgetArgument(arg_arg0_arg1, 0);
                                                                                                                            /* checking for: sort */
                                                                                                                            if (ATgetType((ATerm)arg_arg0_arg1_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg1_arg0) == PERR_afun11) {
                                                                                                                              ATerm arg_arg0_arg1_arg0_arg0 = ATgetArgument(arg_arg0_arg1_arg0, 0);
                                                                                                                              /* checking for: Error */
                                                                                                                              if (ATgetType((ATerm)arg_arg0_arg1_arg0_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg1_arg0_arg0) == PERR_afun37) {
                                                                                                                                ATerm arg_arg0_arg2 = ATgetArgument(arg_arg0, 2);
                                                                                                                                /* checking for: attrs */
                                                                                                                                if (ATgetType((ATerm)arg_arg0_arg2) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg2) == PERR_afun13) {
                                                                                                                                  ATerm arg_arg0_arg2_arg0 = ATgetArgument(arg_arg0_arg2, 0);
                                                                                                                                  if (ATgetType((ATerm)arg_arg0_arg2_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg2_arg0) == ATfalse) {
                                                                                                                                    ATerm arg_arg0_arg2_arg0_head;
                                                                                                                                    ATermList arg_arg0_arg2_arg0_list = (ATermList)arg_arg0_arg2_arg0;
                                                                                                                                    arg_arg0_arg2_arg0_head = ATgetFirst(arg_arg0_arg2_arg0_list);
                                                                                                                                    arg_arg0_arg2_arg0_list = ATgetNext(arg_arg0_arg2_arg0_list);
                                                                                                                                    /* checking for: term */
                                                                                                                                    if (ATgetType((ATerm)arg_arg0_arg2_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg2_arg0_head) == PERR_afun14) {
                                                                                                                                      ATerm arg_arg0_arg2_arg0_head_arg0 = ATgetArgument(arg_arg0_arg2_arg0_head, 0);
                                                                                                                                      /* checking for: cons */
                                                                                                                                      if (ATgetType((ATerm)arg_arg0_arg2_arg0_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg2_arg0_head_arg0) == PERR_afun15) {
                                                                                                                                        ATerm arg_arg0_arg2_arg0_head_arg0_arg0 = ATgetArgument(arg_arg0_arg2_arg0_head_arg0, 0);
                                                                                                                                        /* checking for: info */
                                                                                                                                        if (ATgetType((ATerm)arg_arg0_arg2_arg0_head_arg0_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg2_arg0_head_arg0_arg0) == PERR_afun48) {
                                                                                                                                          if (ATgetType((ATerm)arg_arg0_arg2_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg2_arg0_list) == ATtrue) {
                                                                                                                                            ATerm arg_arg1 = ATgetArgument(arg, 1);
                                                                                                                                            if (ATgetType((ATerm)arg_arg1) == AT_LIST && ATisEmpty((ATermList)arg_arg1) == ATfalse) {
                                                                                                                                              ATerm arg_arg1_head;
                                                                                                                                              ATermList arg_arg1_list = (ATermList)arg_arg1;
                                                                                                                                              arg_arg1_head = ATgetFirst(arg_arg1_list);
                                                                                                                                              arg_arg1_list = ATgetNext(arg_arg1_list);
                                                                                                                                              /* checking for: appl */
                                                                                                                                              if (ATgetType((ATerm)arg_arg1_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head) == PERR_afun0) {
                                                                                                                                                ATerm arg_arg1_head_arg0 = ATgetArgument(arg_arg1_head, 0);
                                                                                                                                                /* checking for: prod */
                                                                                                                                                if (ATgetType((ATerm)arg_arg1_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0) == PERR_afun1) {
                                                                                                                                                  ATerm arg_arg1_head_arg0_arg0 = ATgetArgument(arg_arg1_head_arg0, 0);
                                                                                                                                                  if (ATgetType((ATerm)arg_arg1_head_arg0_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0) == ATfalse) {
                                                                                                                                                    ATerm arg_arg1_head_arg0_arg0_head;
                                                                                                                                                    ATermList arg_arg1_head_arg0_arg0_list = (ATermList)arg_arg1_head_arg0_arg0;
                                                                                                                                                    arg_arg1_head_arg0_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_list);
                                                                                                                                                    arg_arg1_head_arg0_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_list);
                                                                                                                                                    /* checking for: char-class */
                                                                                                                                                    if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg0_head) == PERR_afun17) {
                                                                                                                                                      ATerm arg_arg1_head_arg0_arg0_head_arg0 = ATgetArgument(arg_arg1_head_arg0_arg0_head, 0);
                                                                                                                                                      if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0) == ATfalse) {
                                                                                                                                                        ATerm arg_arg1_head_arg0_arg0_head_arg0_head;
                                                                                                                                                        ATermList arg_arg1_head_arg0_arg0_head_arg0_list = (ATermList)arg_arg1_head_arg0_arg0_head_arg0;
                                                                                                                                                        arg_arg1_head_arg0_arg0_head_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                                                                        arg_arg1_head_arg0_arg0_head_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                                                                        if (ATgetInt((ATermInt)arg_arg1_head_arg0_arg0_head_arg0_head) == 105) {

                                                                                                                                                          if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0_list) == ATtrue) {
                                                                                                                                                            if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_list) == ATfalse) {
                                                                                                                                                              arg_arg1_head_arg0_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_list);
                                                                                                                                                              arg_arg1_head_arg0_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_list);
                                                                                                                                                              /* checking for: char-class */
                                                                                                                                                              if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg0_head) == PERR_afun17) {
                                                                                                                                                                ATerm arg_arg1_head_arg0_arg0_head_arg0 = ATgetArgument(arg_arg1_head_arg0_arg0_head, 0);
                                                                                                                                                                if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0) == ATfalse) {
                                                                                                                                                                  ATerm arg_arg1_head_arg0_arg0_head_arg0_head;
                                                                                                                                                                  ATermList arg_arg1_head_arg0_arg0_head_arg0_list = (ATermList)arg_arg1_head_arg0_arg0_head_arg0;
                                                                                                                                                                  arg_arg1_head_arg0_arg0_head_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                                                                                  arg_arg1_head_arg0_arg0_head_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                                                                                  if (ATgetInt((ATermInt)arg_arg1_head_arg0_arg0_head_arg0_head) == 110) {

                                                                                                                                                                    if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0_list) == ATtrue) {
                                                                                                                                                                      if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_list) == ATfalse) {
                                                                                                                                                                        arg_arg1_head_arg0_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_list);
                                                                                                                                                                        arg_arg1_head_arg0_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_list);
                                                                                                                                                                        /* checking for: char-class */
                                                                                                                                                                        if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg0_head) == PERR_afun17) {
                                                                                                                                                                          ATerm arg_arg1_head_arg0_arg0_head_arg0 = ATgetArgument(arg_arg1_head_arg0_arg0_head, 0);
                                                                                                                                                                          if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0) == ATfalse) {
                                                                                                                                                                            ATerm arg_arg1_head_arg0_arg0_head_arg0_head;
                                                                                                                                                                            ATermList arg_arg1_head_arg0_arg0_head_arg0_list = (ATermList)arg_arg1_head_arg0_arg0_head_arg0;
                                                                                                                                                                            arg_arg1_head_arg0_arg0_head_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                                                                                            arg_arg1_head_arg0_arg0_head_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                                                                                            if (ATgetInt((ATermInt)arg_arg1_head_arg0_arg0_head_arg0_head) == 102) {

                                                                                                                                                                              if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0_list) == ATtrue) {
                                                                                                                                                                                if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_list) == ATfalse) {
                                                                                                                                                                                  arg_arg1_head_arg0_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_list);
                                                                                                                                                                                  arg_arg1_head_arg0_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_list);
                                                                                                                                                                                  /* checking for: char-class */
                                                                                                                                                                                  if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg0_head) == PERR_afun17) {
                                                                                                                                                                                    ATerm arg_arg1_head_arg0_arg0_head_arg0 = ATgetArgument(arg_arg1_head_arg0_arg0_head, 0);
                                                                                                                                                                                    if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0) == ATfalse) {
                                                                                                                                                                                      ATerm arg_arg1_head_arg0_arg0_head_arg0_head;
                                                                                                                                                                                      ATermList arg_arg1_head_arg0_arg0_head_arg0_list = (ATermList)arg_arg1_head_arg0_arg0_head_arg0;
                                                                                                                                                                                      arg_arg1_head_arg0_arg0_head_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                                                                                                      arg_arg1_head_arg0_arg0_head_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                                                                                                      if (ATgetInt((ATermInt)arg_arg1_head_arg0_arg0_head_arg0_head) == 111) {

                                                                                                                                                                                        if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0_list) == ATtrue) {
                                                                                                                                                                                          if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_list) == ATtrue) {
                                                                                                                                                                                            ATerm arg_arg1_head_arg0_arg1 = ATgetArgument(arg_arg1_head_arg0, 1);
                                                                                                                                                                                            /* checking for: lit */
                                                                                                                                                                                            if (ATgetType((ATerm)arg_arg1_head_arg0_arg1) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg1) == PERR_afun9) {
                                                                                                                                                                                              ATerm arg_arg1_head_arg0_arg1_arg0 = ATgetArgument(arg_arg1_head_arg0_arg1, 0);
                                                                                                                                                                                              /* checking for: info */
                                                                                                                                                                                              if (ATgetType((ATerm)arg_arg1_head_arg0_arg1_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg1_arg0) == PERR_afun48) {
                                                                                                                                                                                                ATerm arg_arg1_head_arg0_arg2 = ATgetArgument(arg_arg1_head_arg0, 2);
                                                                                                                                                                                                /* checking for: no-attrs */
                                                                                                                                                                                                if (ATgetType((ATerm)arg_arg1_head_arg0_arg2) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg2) == PERR_afun5) {
                                                                                                                                                                                                  ATerm arg_arg1_head_arg1 = ATgetArgument(arg_arg1_head, 1);
                                                                                                                                                                                                  if (ATgetType((ATerm)arg_arg1_head_arg1) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg1) == ATfalse) {
                                                                                                                                                                                                    ATerm arg_arg1_head_arg1_head;
                                                                                                                                                                                                    ATermList arg_arg1_head_arg1_list = (ATermList)arg_arg1_head_arg1;
                                                                                                                                                                                                    arg_arg1_head_arg1_head = ATgetFirst(arg_arg1_head_arg1_list);
                                                                                                                                                                                                    arg_arg1_head_arg1_list = ATgetNext(arg_arg1_head_arg1_list);
                                                                                                                                                                                                    if (ATgetInt((ATermInt)arg_arg1_head_arg1_head) == 105) {

                                                                                                                                                                                                      if (ATgetType((ATerm)arg_arg1_head_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg1_list) == ATfalse) {
                                                                                                                                                                                                        arg_arg1_head_arg1_head = ATgetFirst(arg_arg1_head_arg1_list);
                                                                                                                                                                                                        arg_arg1_head_arg1_list = ATgetNext(arg_arg1_head_arg1_list);
                                                                                                                                                                                                        if (ATgetInt((ATermInt)arg_arg1_head_arg1_head) == 110) {

                                                                                                                                                                                                          if (ATgetType((ATerm)arg_arg1_head_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg1_list) == ATfalse) {
                                                                                                                                                                                                            arg_arg1_head_arg1_head = ATgetFirst(arg_arg1_head_arg1_list);
                                                                                                                                                                                                            arg_arg1_head_arg1_list = ATgetNext(arg_arg1_head_arg1_list);
                                                                                                                                                                                                            if (ATgetInt((ATermInt)arg_arg1_head_arg1_head) == 102) {

                                                                                                                                                                                                              if (ATgetType((ATerm)arg_arg1_head_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg1_list) == ATfalse) {
                                                                                                                                                                                                                arg_arg1_head_arg1_head = ATgetFirst(arg_arg1_head_arg1_list);
                                                                                                                                                                                                                arg_arg1_head_arg1_list = ATgetNext(arg_arg1_head_arg1_list);
                                                                                                                                                                                                                if (ATgetInt((ATermInt)arg_arg1_head_arg1_head) == 111) {

                                                                                                                                                                                                                  if (ATgetType((ATerm)arg_arg1_head_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg1_list) == ATtrue) {
                                                                                                                                                                                                                    if (ATgetType((ATerm)arg_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_list) == ATfalse) {
                                                                                                                                                                                                                      arg_arg1_list = ATgetNext(arg_arg1_list);
                                                                                                                                                                                                                      if (ATgetType((ATerm)arg_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_list) == ATfalse) {
                                                                                                                                                                                                                        arg_arg1_head = ATgetFirst(arg_arg1_list);
                                                                                                                                                                                                                        arg_arg1_list = ATgetNext(arg_arg1_list);
                                                                                                                                                                                                                        /* checking for: appl */
                                                                                                                                                                                                                        if (ATgetType((ATerm)arg_arg1_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head) == PERR_afun0) {
                                                                                                                                                                                                                          ATerm arg_arg1_head_arg0 = ATgetArgument(arg_arg1_head, 0);
                                                                                                                                                                                                                          /* checking for: prod */
                                                                                                                                                                                                                          if (ATgetType((ATerm)arg_arg1_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0) == PERR_afun1) {
                                                                                                                                                                                                                            ATerm arg_arg1_head_arg0_arg0 = ATgetArgument(arg_arg1_head_arg0, 0);
                                                                                                                                                                                                                            if (ATgetType((ATerm)arg_arg1_head_arg0_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0) == ATfalse) {
                                                                                                                                                                                                                              ATerm arg_arg1_head_arg0_arg0_head;
                                                                                                                                                                                                                              ATermList arg_arg1_head_arg0_arg0_list = (ATermList)arg_arg1_head_arg0_arg0;
                                                                                                                                                                                                                              arg_arg1_head_arg0_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_list);
                                                                                                                                                                                                                              arg_arg1_head_arg0_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_list);
                                                                                                                                                                                                                              /* checking for: char-class */
                                                                                                                                                                                                                              if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg0_head) == PERR_afun17) {
                                                                                                                                                                                                                                ATerm arg_arg1_head_arg0_arg0_head_arg0 = ATgetArgument(arg_arg1_head_arg0_arg0_head, 0);
                                                                                                                                                                                                                                if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0) == ATfalse) {
                                                                                                                                                                                                                                  ATerm arg_arg1_head_arg0_arg0_head_arg0_head;
                                                                                                                                                                                                                                  ATermList arg_arg1_head_arg0_arg0_head_arg0_list = (ATermList)arg_arg1_head_arg0_arg0_head_arg0;
                                                                                                                                                                                                                                  arg_arg1_head_arg0_arg0_head_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                                                                                                                                                  arg_arg1_head_arg0_arg0_head_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                                                                                                                                                  if (ATgetInt((ATermInt)arg_arg1_head_arg0_arg0_head_arg0_head) == 40) {

                                                                                                                                                                                                                                    if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0_list) == ATtrue) {
                                                                                                                                                                                                                                      if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_list) == ATtrue) {
                                                                                                                                                                                                                                        ATerm arg_arg1_head_arg0_arg1 = ATgetArgument(arg_arg1_head_arg0, 1);
                                                                                                                                                                                                                                        /* checking for: lit */
                                                                                                                                                                                                                                        if (ATgetType((ATerm)arg_arg1_head_arg0_arg1) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg1) == PERR_afun9) {
                                                                                                                                                                                                                                          ATerm arg_arg1_head_arg0_arg1_arg0 = ATgetArgument(arg_arg1_head_arg0_arg1, 0);
                                                                                                                                                                                                                                          /* checking for: ( */
                                                                                                                                                                                                                                          if (ATgetType((ATerm)arg_arg1_head_arg0_arg1_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg1_arg0) == PERR_afun39) {
                                                                                                                                                                                                                                            ATerm arg_arg1_head_arg0_arg2 = ATgetArgument(arg_arg1_head_arg0, 2);
                                                                                                                                                                                                                                            /* checking for: no-attrs */
                                                                                                                                                                                                                                            if (ATgetType((ATerm)arg_arg1_head_arg0_arg2) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg2) == PERR_afun5) {
                                                                                                                                                                                                                                              ATerm arg_arg1_head_arg1 = ATgetArgument(arg_arg1_head, 1);
                                                                                                                                                                                                                                              if (ATgetType((ATerm)arg_arg1_head_arg1) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg1) == ATfalse) {
                                                                                                                                                                                                                                                ATerm arg_arg1_head_arg1_head;
                                                                                                                                                                                                                                                ATermList arg_arg1_head_arg1_list = (ATermList)arg_arg1_head_arg1;
                                                                                                                                                                                                                                                arg_arg1_head_arg1_head = ATgetFirst(arg_arg1_head_arg1_list);
                                                                                                                                                                                                                                                arg_arg1_head_arg1_list = ATgetNext(arg_arg1_head_arg1_list);
                                                                                                                                                                                                                                                if (ATgetInt((ATermInt)arg_arg1_head_arg1_head) == 40) {

                                                                                                                                                                                                                                                  if (ATgetType((ATerm)arg_arg1_head_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg1_list) == ATtrue) {
                                                                                                                                                                                                                                                    if (ATgetType((ATerm)arg_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_list) == ATfalse) {
                                                                                                                                                                                                                                                      arg_arg1_list = ATgetNext(arg_arg1_list);
                                                                                                                                                                                                                                                      if (ATgetType((ATerm)arg_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_list) == ATfalse) {
                                                                                                                                                                                                                                                        arg_arg1_list = ATgetNext(arg_arg1_list);
                                                                                                                                                                                                                                                        if (ATgetType((ATerm)arg_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_list) == ATfalse) {
                                                                                                                                                                                                                                                          arg_arg1_list = ATgetNext(arg_arg1_list);
                                                                                                                                                                                                                                                          if (ATgetType((ATerm)arg_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_list) == ATfalse) {
                                                                                                                                                                                                                                                            arg_arg1_head = ATgetFirst(arg_arg1_list);
                                                                                                                                                                                                                                                            arg_arg1_list = ATgetNext(arg_arg1_list);
                                                                                                                                                                                                                                                            /* checking for: appl */
                                                                                                                                                                                                                                                            if (ATgetType((ATerm)arg_arg1_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head) == PERR_afun0) {
                                                                                                                                                                                                                                                              ATerm arg_arg1_head_arg0 = ATgetArgument(arg_arg1_head, 0);
                                                                                                                                                                                                                                                              /* checking for: prod */
                                                                                                                                                                                                                                                              if (ATgetType((ATerm)arg_arg1_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0) == PERR_afun1) {
                                                                                                                                                                                                                                                                ATerm arg_arg1_head_arg0_arg0 = ATgetArgument(arg_arg1_head_arg0, 0);
                                                                                                                                                                                                                                                                if (ATgetType((ATerm)arg_arg1_head_arg0_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0) == ATfalse) {
                                                                                                                                                                                                                                                                  ATerm arg_arg1_head_arg0_arg0_head;
                                                                                                                                                                                                                                                                  ATermList arg_arg1_head_arg0_arg0_list = (ATermList)arg_arg1_head_arg0_arg0;
                                                                                                                                                                                                                                                                  arg_arg1_head_arg0_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_list);
                                                                                                                                                                                                                                                                  arg_arg1_head_arg0_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_list);
                                                                                                                                                                                                                                                                  /* checking for: char-class */
                                                                                                                                                                                                                                                                  if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg0_head) == PERR_afun17) {
                                                                                                                                                                                                                                                                    ATerm arg_arg1_head_arg0_arg0_head_arg0 = ATgetArgument(arg_arg1_head_arg0_arg0_head, 0);
                                                                                                                                                                                                                                                                    if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0) == ATfalse) {
                                                                                                                                                                                                                                                                      ATerm arg_arg1_head_arg0_arg0_head_arg0_head;
                                                                                                                                                                                                                                                                      ATermList arg_arg1_head_arg0_arg0_head_arg0_list = (ATermList)arg_arg1_head_arg0_arg0_head_arg0;
                                                                                                                                                                                                                                                                      arg_arg1_head_arg0_arg0_head_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                                                                                                                                                                                      arg_arg1_head_arg0_arg0_head_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                                                                                                                                                                                      if (ATgetInt((ATermInt)arg_arg1_head_arg0_arg0_head_arg0_head) == 44) {

                                                                                                                                                                                                                                                                        if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0_list) == ATtrue) {
                                                                                                                                                                                                                                                                          if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_list) == ATtrue) {
                                                                                                                                                                                                                                                                            ATerm arg_arg1_head_arg0_arg1 = ATgetArgument(arg_arg1_head_arg0, 1);
                                                                                                                                                                                                                                                                            /* checking for: lit */
                                                                                                                                                                                                                                                                            if (ATgetType((ATerm)arg_arg1_head_arg0_arg1) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg1) == PERR_afun9) {
                                                                                                                                                                                                                                                                              ATerm arg_arg1_head_arg0_arg1_arg0 = ATgetArgument(arg_arg1_head_arg0_arg1, 0);
                                                                                                                                                                                                                                                                              /* checking for: , */
                                                                                                                                                                                                                                                                              if (ATgetType((ATerm)arg_arg1_head_arg0_arg1_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg1_arg0) == PERR_afun43) {
                                                                                                                                                                                                                                                                                ATerm arg_arg1_head_arg0_arg2 = ATgetArgument(arg_arg1_head_arg0, 2);
                                                                                                                                                                                                                                                                                /* checking for: no-attrs */
                                                                                                                                                                                                                                                                                if (ATgetType((ATerm)arg_arg1_head_arg0_arg2) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg2) == PERR_afun5) {
                                                                                                                                                                                                                                                                                  ATerm arg_arg1_head_arg1 = ATgetArgument(arg_arg1_head, 1);
                                                                                                                                                                                                                                                                                  if (ATgetType((ATerm)arg_arg1_head_arg1) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg1) == ATfalse) {
                                                                                                                                                                                                                                                                                    ATerm arg_arg1_head_arg1_head;
                                                                                                                                                                                                                                                                                    ATermList arg_arg1_head_arg1_list = (ATermList)arg_arg1_head_arg1;
                                                                                                                                                                                                                                                                                    arg_arg1_head_arg1_head = ATgetFirst(arg_arg1_head_arg1_list);
                                                                                                                                                                                                                                                                                    arg_arg1_head_arg1_list = ATgetNext(arg_arg1_head_arg1_list);
                                                                                                                                                                                                                                                                                    if (ATgetInt((ATermInt)arg_arg1_head_arg1_head) == 44) {

                                                                                                                                                                                                                                                                                      if (ATgetType((ATerm)arg_arg1_head_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg1_list) == ATtrue) {
                                                                                                                                                                                                                                                                                        if (ATgetType((ATerm)arg_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_list) == ATfalse) {
                                                                                                                                                                                                                                                                                          arg_arg1_list = ATgetNext(arg_arg1_list);
                                                                                                                                                                                                                                                                                          if (ATgetType((ATerm)arg_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_list) == ATfalse) {
                                                                                                                                                                                                                                                                                            arg_arg1_head = ATgetFirst(arg_arg1_list);
                                                                                                                                                                                                                                                                                            arg_arg1_list = ATgetNext(arg_arg1_list);
                                                                                                                                                                                                                                                                                            /* checking for: appl */
                                                                                                                                                                                                                                                                                            if (ATgetType((ATerm)arg_arg1_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head) == PERR_afun0) {
                                                                                                                                                                                                                                                                                              ATerm arg_arg1_head_arg0 = ATgetArgument(arg_arg1_head, 0);
                                                                                                                                                                                                                                                                                              /* checking for: prod */
                                                                                                                                                                                                                                                                                              if (ATgetType((ATerm)arg_arg1_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0) == PERR_afun1) {
                                                                                                                                                                                                                                                                                                ATerm arg_arg1_head_arg0_arg0 = ATgetArgument(arg_arg1_head_arg0, 0);
                                                                                                                                                                                                                                                                                                if (ATgetType((ATerm)arg_arg1_head_arg0_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0) == ATfalse) {
                                                                                                                                                                                                                                                                                                  ATerm arg_arg1_head_arg0_arg0_head;
                                                                                                                                                                                                                                                                                                  ATermList arg_arg1_head_arg0_arg0_list = (ATermList)arg_arg1_head_arg0_arg0;
                                                                                                                                                                                                                                                                                                  arg_arg1_head_arg0_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_list);
                                                                                                                                                                                                                                                                                                  arg_arg1_head_arg0_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_list);
                                                                                                                                                                                                                                                                                                  /* checking for: char-class */
                                                                                                                                                                                                                                                                                                  if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg0_head) == PERR_afun17) {
                                                                                                                                                                                                                                                                                                    ATerm arg_arg1_head_arg0_arg0_head_arg0 = ATgetArgument(arg_arg1_head_arg0_arg0_head, 0);
                                                                                                                                                                                                                                                                                                    if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0) == ATfalse) {
                                                                                                                                                                                                                                                                                                      ATerm arg_arg1_head_arg0_arg0_head_arg0_head;
                                                                                                                                                                                                                                                                                                      ATermList arg_arg1_head_arg0_arg0_head_arg0_list = (ATermList)arg_arg1_head_arg0_arg0_head_arg0;
                                                                                                                                                                                                                                                                                                      arg_arg1_head_arg0_arg0_head_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                                                                                                                                                                                                                      arg_arg1_head_arg0_arg0_head_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                                                                                                                                                                                                                      if (ATgetInt((ATermInt)arg_arg1_head_arg0_arg0_head_arg0_head) == 91) {

                                                                                                                                                                                                                                                                                                        if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0_list) == ATtrue) {
                                                                                                                                                                                                                                                                                                          if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_list) == ATtrue) {
                                                                                                                                                                                                                                                                                                            ATerm arg_arg1_head_arg0_arg1 = ATgetArgument(arg_arg1_head_arg0, 1);
                                                                                                                                                                                                                                                                                                            /* checking for: lit */
                                                                                                                                                                                                                                                                                                            if (ATgetType((ATerm)arg_arg1_head_arg0_arg1) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg1) == PERR_afun9) {
                                                                                                                                                                                                                                                                                                              ATerm arg_arg1_head_arg0_arg1_arg0 = ATgetArgument(arg_arg1_head_arg0_arg1, 0);
                                                                                                                                                                                                                                                                                                              /* checking for: [ */
                                                                                                                                                                                                                                                                                                              if (ATgetType((ATerm)arg_arg1_head_arg0_arg1_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg1_arg0) == PERR_afun47) {
                                                                                                                                                                                                                                                                                                                ATerm arg_arg1_head_arg0_arg2 = ATgetArgument(arg_arg1_head_arg0, 2);
                                                                                                                                                                                                                                                                                                                /* checking for: no-attrs */
                                                                                                                                                                                                                                                                                                                if (ATgetType((ATerm)arg_arg1_head_arg0_arg2) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg2) == PERR_afun5) {
                                                                                                                                                                                                                                                                                                                  ATerm arg_arg1_head_arg1 = ATgetArgument(arg_arg1_head, 1);
                                                                                                                                                                                                                                                                                                                  if (ATgetType((ATerm)arg_arg1_head_arg1) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg1) == ATfalse) {
                                                                                                                                                                                                                                                                                                                    ATerm arg_arg1_head_arg1_head;
                                                                                                                                                                                                                                                                                                                    ATermList arg_arg1_head_arg1_list = (ATermList)arg_arg1_head_arg1;
                                                                                                                                                                                                                                                                                                                    arg_arg1_head_arg1_head = ATgetFirst(arg_arg1_head_arg1_list);
                                                                                                                                                                                                                                                                                                                    arg_arg1_head_arg1_list = ATgetNext(arg_arg1_head_arg1_list);
                                                                                                                                                                                                                                                                                                                    if (ATgetInt((ATermInt)arg_arg1_head_arg1_head) == 91) {

                                                                                                                                                                                                                                                                                                                      if (ATgetType((ATerm)arg_arg1_head_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg1_list) == ATtrue) {
                                                                                                                                                                                                                                                                                                                        if (ATgetType((ATerm)arg_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_list) == ATfalse) {
                                                                                                                                                                                                                                                                                                                          arg_arg1_list = ATgetNext(arg_arg1_list);
                                                                                                                                                                                                                                                                                                                          if (ATgetType((ATerm)arg_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_list) == ATfalse) {
                                                                                                                                                                                                                                                                                                                            arg_arg1_head = ATgetFirst(arg_arg1_list);
                                                                                                                                                                                                                                                                                                                            arg_arg1_list = ATgetNext(arg_arg1_list);
                                                                                                                                                                                                                                                                                                                            /* checking for: appl */
                                                                                                                                                                                                                                                                                                                            if (ATgetType((ATerm)arg_arg1_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head) == PERR_afun0) {
                                                                                                                                                                                                                                                                                                                              ATerm arg_arg1_head_arg0 = ATgetArgument(arg_arg1_head, 0);
                                                                                                                                                                                                                                                                                                                              /* checking for: list */
                                                                                                                                                                                                                                                                                                                              if (ATgetType((ATerm)arg_arg1_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0) == PERR_afun8) {
                                                                                                                                                                                                                                                                                                                                ATerm arg_arg1_head_arg0_arg0 = ATgetArgument(arg_arg1_head_arg0, 0);
                                                                                                                                                                                                                                                                                                                                /* checking for: cf */
                                                                                                                                                                                                                                                                                                                                if (ATgetType((ATerm)arg_arg1_head_arg0_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg0) == PERR_afun2) {
                                                                                                                                                                                                                                                                                                                                  ATerm arg_arg1_head_arg0_arg0_arg0 = ATgetArgument(arg_arg1_head_arg0_arg0, 0);
                                                                                                                                                                                                                                                                                                                                  /* checking for: iter-star-sep */
                                                                                                                                                                                                                                                                                                                                  if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg0_arg0) == PERR_afun46) {
                                                                                                                                                                                                                                                                                                                                    ATerm arg_arg1_head_arg0_arg0_arg0_arg0 = ATgetArgument(arg_arg1_head_arg0_arg0_arg0, 0);
                                                                                                                                                                                                                                                                                                                                    /* checking for: sort */
                                                                                                                                                                                                                                                                                                                                    if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_arg0_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg0_arg0_arg0) == PERR_afun11) {
                                                                                                                                                                                                                                                                                                                                      ATerm arg_arg1_head_arg0_arg0_arg0_arg0_arg0 = ATgetArgument(arg_arg1_head_arg0_arg0_arg0_arg0, 0);
                                                                                                                                                                                                                                                                                                                                      /* checking for: Subject */
                                                                                                                                                                                                                                                                                                                                      if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_arg0_arg0_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg0_arg0_arg0_arg0) == PERR_afun41) {
                                                                                                                                                                                                                                                                                                                                        ATerm arg_arg1_head_arg0_arg0_arg0_arg1 = ATgetArgument(arg_arg1_head_arg0_arg0_arg0, 1);
                                                                                                                                                                                                                                                                                                                                        /* checking for: lit */
                                                                                                                                                                                                                                                                                                                                        if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_arg0_arg1) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg0_arg0_arg1) == PERR_afun9) {
                                                                                                                                                                                                                                                                                                                                          ATerm arg_arg1_head_arg0_arg0_arg0_arg1_arg0 = ATgetArgument(arg_arg1_head_arg0_arg0_arg0_arg1, 0);
                                                                                                                                                                                                                                                                                                                                          /* checking for: , */
                                                                                                                                                                                                                                                                                                                                          if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_arg0_arg1_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg0_arg0_arg1_arg0) == PERR_afun43) {
                                                                                                                                                                                                                                                                                                                                            if (ATgetType((ATerm)arg_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_list) == ATfalse) {
                                                                                                                                                                                                                                                                                                                                              arg_arg1_list = ATgetNext(arg_arg1_list);
                                                                                                                                                                                                                                                                                                                                              if (ATgetType((ATerm)arg_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_list) == ATfalse) {
                                                                                                                                                                                                                                                                                                                                                arg_arg1_head = ATgetFirst(arg_arg1_list);
                                                                                                                                                                                                                                                                                                                                                arg_arg1_list = ATgetNext(arg_arg1_list);
                                                                                                                                                                                                                                                                                                                                                /* checking for: appl */
                                                                                                                                                                                                                                                                                                                                                if (ATgetType((ATerm)arg_arg1_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head) == PERR_afun0) {
                                                                                                                                                                                                                                                                                                                                                  ATerm arg_arg1_head_arg0 = ATgetArgument(arg_arg1_head, 0);
                                                                                                                                                                                                                                                                                                                                                  /* checking for: prod */
                                                                                                                                                                                                                                                                                                                                                  if (ATgetType((ATerm)arg_arg1_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0) == PERR_afun1) {
                                                                                                                                                                                                                                                                                                                                                    ATerm arg_arg1_head_arg0_arg0 = ATgetArgument(arg_arg1_head_arg0, 0);
                                                                                                                                                                                                                                                                                                                                                    if (ATgetType((ATerm)arg_arg1_head_arg0_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0) == ATfalse) {
                                                                                                                                                                                                                                                                                                                                                      ATerm arg_arg1_head_arg0_arg0_head;
                                                                                                                                                                                                                                                                                                                                                      ATermList arg_arg1_head_arg0_arg0_list = (ATermList)arg_arg1_head_arg0_arg0;
                                                                                                                                                                                                                                                                                                                                                      arg_arg1_head_arg0_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_list);
                                                                                                                                                                                                                                                                                                                                                      arg_arg1_head_arg0_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_list);
                                                                                                                                                                                                                                                                                                                                                      /* checking for: char-class */
                                                                                                                                                                                                                                                                                                                                                      if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg0_head) == PERR_afun17) {
                                                                                                                                                                                                                                                                                                                                                        ATerm arg_arg1_head_arg0_arg0_head_arg0 = ATgetArgument(arg_arg1_head_arg0_arg0_head, 0);
                                                                                                                                                                                                                                                                                                                                                        if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0) == ATfalse) {
                                                                                                                                                                                                                                                                                                                                                          ATerm arg_arg1_head_arg0_arg0_head_arg0_head;
                                                                                                                                                                                                                                                                                                                                                          ATermList arg_arg1_head_arg0_arg0_head_arg0_list = (ATermList)arg_arg1_head_arg0_arg0_head_arg0;
                                                                                                                                                                                                                                                                                                                                                          arg_arg1_head_arg0_arg0_head_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                                                                                                                                                                                                                                                                          arg_arg1_head_arg0_arg0_head_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                                                                                                                                                                                                                                                                          if (ATgetInt((ATermInt)arg_arg1_head_arg0_arg0_head_arg0_head) == 93) {

                                                                                                                                                                                                                                                                                                                                                            if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0_list) == ATtrue) {
                                                                                                                                                                                                                                                                                                                                                              if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_list) == ATtrue) {
                                                                                                                                                                                                                                                                                                                                                                ATerm arg_arg1_head_arg0_arg1 = ATgetArgument(arg_arg1_head_arg0, 1);
                                                                                                                                                                                                                                                                                                                                                                /* checking for: lit */
                                                                                                                                                                                                                                                                                                                                                                if (ATgetType((ATerm)arg_arg1_head_arg0_arg1) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg1) == PERR_afun9) {
                                                                                                                                                                                                                                                                                                                                                                  ATerm arg_arg1_head_arg0_arg1_arg0 = ATgetArgument(arg_arg1_head_arg0_arg1, 0);
                                                                                                                                                                                                                                                                                                                                                                  /* checking for: ] */
                                                                                                                                                                                                                                                                                                                                                                  if (ATgetType((ATerm)arg_arg1_head_arg0_arg1_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg1_arg0) == PERR_afun45) {
                                                                                                                                                                                                                                                                                                                                                                    ATerm arg_arg1_head_arg0_arg2 = ATgetArgument(arg_arg1_head_arg0, 2);
                                                                                                                                                                                                                                                                                                                                                                    /* checking for: no-attrs */
                                                                                                                                                                                                                                                                                                                                                                    if (ATgetType((ATerm)arg_arg1_head_arg0_arg2) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg2) == PERR_afun5) {
                                                                                                                                                                                                                                                                                                                                                                      ATerm arg_arg1_head_arg1 = ATgetArgument(arg_arg1_head, 1);
                                                                                                                                                                                                                                                                                                                                                                      if (ATgetType((ATerm)arg_arg1_head_arg1) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg1) == ATfalse) {
                                                                                                                                                                                                                                                                                                                                                                        ATerm arg_arg1_head_arg1_head;
                                                                                                                                                                                                                                                                                                                                                                        ATermList arg_arg1_head_arg1_list = (ATermList)arg_arg1_head_arg1;
                                                                                                                                                                                                                                                                                                                                                                        arg_arg1_head_arg1_head = ATgetFirst(arg_arg1_head_arg1_list);
                                                                                                                                                                                                                                                                                                                                                                        arg_arg1_head_arg1_list = ATgetNext(arg_arg1_head_arg1_list);
                                                                                                                                                                                                                                                                                                                                                                        if (ATgetInt((ATermInt)arg_arg1_head_arg1_head) == 93) {

                                                                                                                                                                                                                                                                                                                                                                          if (ATgetType((ATerm)arg_arg1_head_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg1_list) == ATtrue) {
                                                                                                                                                                                                                                                                                                                                                                            if (ATgetType((ATerm)arg_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_list) == ATfalse) {
                                                                                                                                                                                                                                                                                                                                                                              arg_arg1_list = ATgetNext(arg_arg1_list);
                                                                                                                                                                                                                                                                                                                                                                              if (ATgetType((ATerm)arg_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_list) == ATfalse) {
                                                                                                                                                                                                                                                                                                                                                                                arg_arg1_head = ATgetFirst(arg_arg1_list);
                                                                                                                                                                                                                                                                                                                                                                                arg_arg1_list = ATgetNext(arg_arg1_list);
                                                                                                                                                                                                                                                                                                                                                                                /* checking for: appl */
                                                                                                                                                                                                                                                                                                                                                                                if (ATgetType((ATerm)arg_arg1_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head) == PERR_afun0) {
                                                                                                                                                                                                                                                                                                                                                                                  ATerm arg_arg1_head_arg0 = ATgetArgument(arg_arg1_head, 0);
                                                                                                                                                                                                                                                                                                                                                                                  /* checking for: prod */
                                                                                                                                                                                                                                                                                                                                                                                  if (ATgetType((ATerm)arg_arg1_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0) == PERR_afun1) {
                                                                                                                                                                                                                                                                                                                                                                                    ATerm arg_arg1_head_arg0_arg0 = ATgetArgument(arg_arg1_head_arg0, 0);
                                                                                                                                                                                                                                                                                                                                                                                    if (ATgetType((ATerm)arg_arg1_head_arg0_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0) == ATfalse) {
                                                                                                                                                                                                                                                                                                                                                                                      ATerm arg_arg1_head_arg0_arg0_head;
                                                                                                                                                                                                                                                                                                                                                                                      ATermList arg_arg1_head_arg0_arg0_list = (ATermList)arg_arg1_head_arg0_arg0;
                                                                                                                                                                                                                                                                                                                                                                                      arg_arg1_head_arg0_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_list);
                                                                                                                                                                                                                                                                                                                                                                                      arg_arg1_head_arg0_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_list);
                                                                                                                                                                                                                                                                                                                                                                                      /* checking for: char-class */
                                                                                                                                                                                                                                                                                                                                                                                      if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg0_head) == PERR_afun17) {
                                                                                                                                                                                                                                                                                                                                                                                        ATerm arg_arg1_head_arg0_arg0_head_arg0 = ATgetArgument(arg_arg1_head_arg0_arg0_head, 0);
                                                                                                                                                                                                                                                                                                                                                                                        if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0) == ATfalse) {
                                                                                                                                                                                                                                                                                                                                                                                          ATerm arg_arg1_head_arg0_arg0_head_arg0_head;
                                                                                                                                                                                                                                                                                                                                                                                          ATermList arg_arg1_head_arg0_arg0_head_arg0_list = (ATermList)arg_arg1_head_arg0_arg0_head_arg0;
                                                                                                                                                                                                                                                                                                                                                                                          arg_arg1_head_arg0_arg0_head_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                                                                                                                                                                                                                                                                                                          arg_arg1_head_arg0_arg0_head_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                                                                                                                                                                                                                                                                                                          if (ATgetInt((ATermInt)arg_arg1_head_arg0_arg0_head_arg0_head) == 41) {

                                                                                                                                                                                                                                                                                                                                                                                            if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0_list) == ATtrue) {
                                                                                                                                                                                                                                                                                                                                                                                              if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_list) == ATtrue) {
                                                                                                                                                                                                                                                                                                                                                                                                ATerm arg_arg1_head_arg0_arg1 = ATgetArgument(arg_arg1_head_arg0, 1);
                                                                                                                                                                                                                                                                                                                                                                                                /* checking for: lit */
                                                                                                                                                                                                                                                                                                                                                                                                if (ATgetType((ATerm)arg_arg1_head_arg0_arg1) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg1) == PERR_afun9) {
                                                                                                                                                                                                                                                                                                                                                                                                  ATerm arg_arg1_head_arg0_arg1_arg0 = ATgetArgument(arg_arg1_head_arg0_arg1, 0);
                                                                                                                                                                                                                                                                                                                                                                                                  /* checking for: ) */
                                                                                                                                                                                                                                                                                                                                                                                                  if (ATgetType((ATerm)arg_arg1_head_arg0_arg1_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg1_arg0) == PERR_afun38) {
                                                                                                                                                                                                                                                                                                                                                                                                    ATerm arg_arg1_head_arg0_arg2 = ATgetArgument(arg_arg1_head_arg0, 2);
                                                                                                                                                                                                                                                                                                                                                                                                    /* checking for: no-attrs */
                                                                                                                                                                                                                                                                                                                                                                                                    if (ATgetType((ATerm)arg_arg1_head_arg0_arg2) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg2) == PERR_afun5) {
                                                                                                                                                                                                                                                                                                                                                                                                      ATerm arg_arg1_head_arg1 = ATgetArgument(arg_arg1_head, 1);
                                                                                                                                                                                                                                                                                                                                                                                                      if (ATgetType((ATerm)arg_arg1_head_arg1) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg1) == ATfalse) {
                                                                                                                                                                                                                                                                                                                                                                                                        ATerm arg_arg1_head_arg1_head;
                                                                                                                                                                                                                                                                                                                                                                                                        ATermList arg_arg1_head_arg1_list = (ATermList)arg_arg1_head_arg1;
                                                                                                                                                                                                                                                                                                                                                                                                        arg_arg1_head_arg1_head = ATgetFirst(arg_arg1_head_arg1_list);
                                                                                                                                                                                                                                                                                                                                                                                                        arg_arg1_head_arg1_list = ATgetNext(arg_arg1_head_arg1_list);
                                                                                                                                                                                                                                                                                                                                                                                                        if (ATgetInt((ATermInt)arg_arg1_head_arg1_head) == 41) {

                                                                                                                                                                                                                                                                                                                                                                                                          if (ATgetType((ATerm)arg_arg1_head_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg1_list) == ATtrue) {
                                                                                                                                                                                                                                                                                                                                                                                                            if (ATgetType((ATerm)arg_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_list) == ATtrue) {
                                                                                                                                                                                                                                                                                                                                                                                                              return ATtrue;
                                                                                                                                                                                                                                                                                                                                                                                                            }
                                                                                                                                                                                                                                                                                                                                                                                                          }
                                                                                                                                                                                                                                                                                                                                                                                                        }
                                                                                                                                                                                                                                                                                                                                                                                                      }
                                                                                                                                                                                                                                                                                                                                                                                                    }
                                                                                                                                                                                                                                                                                                                                                                                                  }
                                                                                                                                                                                                                                                                                                                                                                                                }
                                                                                                                                                                                                                                                                                                                                                                                              }
                                                                                                                                                                                                                                                                                                                                                                                            }
                                                                                                                                                                                                                                                                                                                                                                                          }
                                                                                                                                                                                                                                                                                                                                                                                        }
                                                                                                                                                                                                                                                                                                                                                                                      }
                                                                                                                                                                                                                                                                                                                                                                                    }
                                                                                                                                                                                                                                                                                                                                                                                  }
                                                                                                                                                                                                                                                                                                                                                                                }
                                                                                                                                                                                                                                                                                                                                                                              }
                                                                                                                                                                                                                                                                                                                                                                            }
                                                                                                                                                                                                                                                                                                                                                                          }
                                                                                                                                                                                                                                                                                                                                                                        }
                                                                                                                                                                                                                                                                                                                                                                      }
                                                                                                                                                                                                                                                                                                                                                                    }
                                                                                                                                                                                                                                                                                                                                                                  }
                                                                                                                                                                                                                                                                                                                                                                }
                                                                                                                                                                                                                                                                                                                                                              }
                                                                                                                                                                                                                                                                                                                                                            }
                                                                                                                                                                                                                                                                                                                                                          }
                                                                                                                                                                                                                                                                                                                                                        }
                                                                                                                                                                                                                                                                                                                                                      }
                                                                                                                                                                                                                                                                                                                                                    }
                                                                                                                                                                                                                                                                                                                                                  }
                                                                                                                                                                                                                                                                                                                                                }
                                                                                                                                                                                                                                                                                                                                              }
                                                                                                                                                                                                                                                                                                                                            }
                                                                                                                                                                                                                                                                                                                                          }
                                                                                                                                                                                                                                                                                                                                        }
                                                                                                                                                                                                                                                                                                                                      }
                                                                                                                                                                                                                                                                                                                                    }
                                                                                                                                                                                                                                                                                                                                  }
                                                                                                                                                                                                                                                                                                                                }
                                                                                                                                                                                                                                                                                                                              }
                                                                                                                                                                                                                                                                                                                            }
                                                                                                                                                                                                                                                                                                                          }
                                                                                                                                                                                                                                                                                                                        }
                                                                                                                                                                                                                                                                                                                      }
                                                                                                                                                                                                                                                                                                                    }
                                                                                                                                                                                                                                                                                                                  }
                                                                                                                                                                                                                                                                                                                }
                                                                                                                                                                                                                                                                                                              }
                                                                                                                                                                                                                                                                                                            }
                                                                                                                                                                                                                                                                                                          }
                                                                                                                                                                                                                                                                                                        }
                                                                                                                                                                                                                                                                                                      }
                                                                                                                                                                                                                                                                                                    }
                                                                                                                                                                                                                                                                                                  }
                                                                                                                                                                                                                                                                                                }
                                                                                                                                                                                                                                                                                              }
                                                                                                                                                                                                                                                                                            }
                                                                                                                                                                                                                                                                                          }
                                                                                                                                                                                                                                                                                        }
                                                                                                                                                                                                                                                                                      }
                                                                                                                                                                                                                                                                                    }
                                                                                                                                                                                                                                                                                  }
                                                                                                                                                                                                                                                                                }
                                                                                                                                                                                                                                                                              }
                                                                                                                                                                                                                                                                            }
                                                                                                                                                                                                                                                                          }
                                                                                                                                                                                                                                                                        }
                                                                                                                                                                                                                                                                      }
                                                                                                                                                                                                                                                                    }
                                                                                                                                                                                                                                                                  }
                                                                                                                                                                                                                                                                }
                                                                                                                                                                                                                                                              }
                                                                                                                                                                                                                                                            }
                                                                                                                                                                                                                                                          }
                                                                                                                                                                                                                                                        }
                                                                                                                                                                                                                                                      }
                                                                                                                                                                                                                                                    }
                                                                                                                                                                                                                                                  }
                                                                                                                                                                                                                                                }
                                                                                                                                                                                                                                              }
                                                                                                                                                                                                                                            }
                                                                                                                                                                                                                                          }
                                                                                                                                                                                                                                        }
                                                                                                                                                                                                                                      }
                                                                                                                                                                                                                                    }
                                                                                                                                                                                                                                  }
                                                                                                                                                                                                                                }
                                                                                                                                                                                                                              }
                                                                                                                                                                                                                            }
                                                                                                                                                                                                                          }
                                                                                                                                                                                                                        }
                                                                                                                                                                                                                      }
                                                                                                                                                                                                                    }
                                                                                                                                                                                                                  }
                                                                                                                                                                                                                }
                                                                                                                                                                                                              }
                                                                                                                                                                                                            }
                                                                                                                                                                                                          }
                                                                                                                                                                                                        }
                                                                                                                                                                                                      }
                                                                                                                                                                                                    }
                                                                                                                                                                                                  }
                                                                                                                                                                                                }
                                                                                                                                                                                              }
                                                                                                                                                                                            }
                                                                                                                                                                                          }
                                                                                                                                                                                        }
                                                                                                                                                                                      }
                                                                                                                                                                                    }
                                                                                                                                                                                  }
                                                                                                                                                                                }
                                                                                                                                                                              }
                                                                                                                                                                            }
                                                                                                                                                                          }
                                                                                                                                                                        }
                                                                                                                                                                      }
                                                                                                                                                                    }
                                                                                                                                                                  }
                                                                                                                                                                }
                                                                                                                                                              }
                                                                                                                                                            }
                                                                                                                                                          }
                                                                                                                                                        }
                                                                                                                                                      }
                                                                                                                                                    }
                                                                                                                                                  }
                                                                                                                                                }
                                                                                                                                              }
                                                                                                                                            }
                                                                                                                                          }
                                                                                                                                        }
                                                                                                                                      }
                                                                                                                                    }
                                                                                                                                  }
                                                                                                                                }
                                                                                                                              }
                                                                                                                            }
                                                                                                                          }
                                                                                                                        }
                                                                                                                      }
                                                                                                                    }
                                                                                                                  }
                                                                                                                }
                                                                                                              }
                                                                                                            }
                                                                                                          }
                                                                                                        }
                                                                                                      }
                                                                                                    }
                                                                                                  }
                                                                                                }
                                                                                              }
                                                                                            }
                                                                                          }
                                                                                        }
                                                                                      }
                                                                                    }
                                                                                  }
                                                                                }
                                                                              }
                                                                            }
                                                                          }
                                                                        }
                                                                      }
                                                                    }
                                                                  }
                                                                }
                                                              }
                                                            }
                                                          }
                                                        }
                                                      }
                                                    }
                                                  }
                                                }
                                              }
                                            }
                                          }
                                        }
                                      }
                                    }
                                  }
                                }
                              }
                            }
                          }
                        }
                      }
                    }
                  }
                }
              }
            }
          }
        }
      }
    }
  }
  return ATfalse;
}

/**
 * Assert whether a PERR_Error is a warning by checking against the following ATerm pattern: appl(prod([lit("warning"),cf(opt(layout)),lit("("),cf(opt(layout)),cf(sort("StrCon")),cf(opt(layout)),lit(","),cf(opt(layout)),lit("["),cf(opt(layout)),cf(iter-star-sep(sort("Subject"),lit(","))),cf(opt(layout)),lit("]"),cf(opt(layout)),lit(")")],cf(sort("Error")),attrs([term(cons("warning"))])),[appl(prod([char-class([119]),char-class([97]),char-class([114]),char-class([110]),char-class([105]),char-class([110]),char-class([103])],lit("warning"),no-attrs),[119,97,114,110,105,110,103]),<"ws-after-warning"("OptLayout")>,appl(prod([char-class([40])],lit("("),no-attrs),[40]),<"ws-after-("("OptLayout")>,<"description"("StrCon")>,<"ws-after-description"("OptLayout")>,appl(prod([char-class([44])],lit(","),no-attrs),[44]),<"ws-after-,"("OptLayout")>,appl(prod([char-class([91])],lit("["),no-attrs),[91]),<"ws-after-["("OptLayout")>,appl(list(cf(iter-star-sep(sort("Subject"),lit(",")))),<"list"("Subject-list")>),<"ws-after-list"("OptLayout")>,appl(prod([char-class([93])],lit("]"),no-attrs),[93]),<"ws-after-]"("OptLayout")>,appl(prod([char-class([41])],lit(")"),no-attrs),[41])]). May not be used to assert correctness of the PERR_Error
 * \param[in] arg input PERR_Error
 * \return ATtrue if #arg corresponds to the signature of a warning, or ATfalse otherwise
 */
inline ATbool PERR_isErrorWarning(PERR_Error arg){
  /* checking for: appl */
  if (ATgetType((ATerm)arg) == AT_APPL && ATgetAFun((ATermAppl)arg) == PERR_afun0) {
    ATerm arg_arg0 = ATgetArgument(arg, 0);
    /* checking for: prod */
    if (ATgetType((ATerm)arg_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0) == PERR_afun1) {
      ATerm arg_arg0_arg0 = ATgetArgument(arg_arg0, 0);
      if (ATgetType((ATerm)arg_arg0_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0) == ATfalse) {
        ATerm arg_arg0_arg0_head;
        ATermList arg_arg0_arg0_list = (ATermList)arg_arg0_arg0;
        arg_arg0_arg0_head = ATgetFirst(arg_arg0_arg0_list);
        arg_arg0_arg0_list = ATgetNext(arg_arg0_arg0_list);
        /* checking for: lit */
        if (ATgetType((ATerm)arg_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head) == PERR_afun9) {
          ATerm arg_arg0_arg0_head_arg0 = ATgetArgument(arg_arg0_arg0_head, 0);
          /* checking for: warning */
          if (ATgetType((ATerm)arg_arg0_arg0_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0) == PERR_afun49) {
            if (ATgetType((ATerm)arg_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0_list) == ATfalse) {
              arg_arg0_arg0_head = ATgetFirst(arg_arg0_arg0_list);
              arg_arg0_arg0_list = ATgetNext(arg_arg0_arg0_list);
              /* checking for: cf */
              if (ATgetType((ATerm)arg_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head) == PERR_afun2) {
                ATerm arg_arg0_arg0_head_arg0 = ATgetArgument(arg_arg0_arg0_head, 0);
                /* checking for: opt */
                if (ATgetType((ATerm)arg_arg0_arg0_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0) == PERR_afun3) {
                  ATerm arg_arg0_arg0_head_arg0_arg0 = ATgetArgument(arg_arg0_arg0_head_arg0, 0);
                  /* checking for: layout */
                  if (ATgetType((ATerm)arg_arg0_arg0_head_arg0_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0_arg0) == PERR_afun4) {
                    if (ATgetType((ATerm)arg_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0_list) == ATfalse) {
                      arg_arg0_arg0_head = ATgetFirst(arg_arg0_arg0_list);
                      arg_arg0_arg0_list = ATgetNext(arg_arg0_arg0_list);
                      /* checking for: lit */
                      if (ATgetType((ATerm)arg_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head) == PERR_afun9) {
                        ATerm arg_arg0_arg0_head_arg0 = ATgetArgument(arg_arg0_arg0_head, 0);
                        /* checking for: ( */
                        if (ATgetType((ATerm)arg_arg0_arg0_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0) == PERR_afun39) {
                          if (ATgetType((ATerm)arg_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0_list) == ATfalse) {
                            arg_arg0_arg0_head = ATgetFirst(arg_arg0_arg0_list);
                            arg_arg0_arg0_list = ATgetNext(arg_arg0_arg0_list);
                            /* checking for: cf */
                            if (ATgetType((ATerm)arg_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head) == PERR_afun2) {
                              ATerm arg_arg0_arg0_head_arg0 = ATgetArgument(arg_arg0_arg0_head, 0);
                              /* checking for: opt */
                              if (ATgetType((ATerm)arg_arg0_arg0_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0) == PERR_afun3) {
                                ATerm arg_arg0_arg0_head_arg0_arg0 = ATgetArgument(arg_arg0_arg0_head_arg0, 0);
                                /* checking for: layout */
                                if (ATgetType((ATerm)arg_arg0_arg0_head_arg0_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0_arg0) == PERR_afun4) {
                                  if (ATgetType((ATerm)arg_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0_list) == ATfalse) {
                                    arg_arg0_arg0_head = ATgetFirst(arg_arg0_arg0_list);
                                    arg_arg0_arg0_list = ATgetNext(arg_arg0_arg0_list);
                                    /* checking for: cf */
                                    if (ATgetType((ATerm)arg_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head) == PERR_afun2) {
                                      ATerm arg_arg0_arg0_head_arg0 = ATgetArgument(arg_arg0_arg0_head, 0);
                                      /* checking for: sort */
                                      if (ATgetType((ATerm)arg_arg0_arg0_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0) == PERR_afun11) {
                                        ATerm arg_arg0_arg0_head_arg0_arg0 = ATgetArgument(arg_arg0_arg0_head_arg0, 0);
                                        /* checking for: StrCon */
                                        if (ATgetType((ATerm)arg_arg0_arg0_head_arg0_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0_arg0) == PERR_afun29) {
                                          if (ATgetType((ATerm)arg_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0_list) == ATfalse) {
                                            arg_arg0_arg0_head = ATgetFirst(arg_arg0_arg0_list);
                                            arg_arg0_arg0_list = ATgetNext(arg_arg0_arg0_list);
                                            /* checking for: cf */
                                            if (ATgetType((ATerm)arg_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head) == PERR_afun2) {
                                              ATerm arg_arg0_arg0_head_arg0 = ATgetArgument(arg_arg0_arg0_head, 0);
                                              /* checking for: opt */
                                              if (ATgetType((ATerm)arg_arg0_arg0_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0) == PERR_afun3) {
                                                ATerm arg_arg0_arg0_head_arg0_arg0 = ATgetArgument(arg_arg0_arg0_head_arg0, 0);
                                                /* checking for: layout */
                                                if (ATgetType((ATerm)arg_arg0_arg0_head_arg0_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0_arg0) == PERR_afun4) {
                                                  if (ATgetType((ATerm)arg_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0_list) == ATfalse) {
                                                    arg_arg0_arg0_head = ATgetFirst(arg_arg0_arg0_list);
                                                    arg_arg0_arg0_list = ATgetNext(arg_arg0_arg0_list);
                                                    /* checking for: lit */
                                                    if (ATgetType((ATerm)arg_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head) == PERR_afun9) {
                                                      ATerm arg_arg0_arg0_head_arg0 = ATgetArgument(arg_arg0_arg0_head, 0);
                                                      /* checking for: , */
                                                      if (ATgetType((ATerm)arg_arg0_arg0_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0) == PERR_afun43) {
                                                        if (ATgetType((ATerm)arg_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0_list) == ATfalse) {
                                                          arg_arg0_arg0_head = ATgetFirst(arg_arg0_arg0_list);
                                                          arg_arg0_arg0_list = ATgetNext(arg_arg0_arg0_list);
                                                          /* checking for: cf */
                                                          if (ATgetType((ATerm)arg_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head) == PERR_afun2) {
                                                            ATerm arg_arg0_arg0_head_arg0 = ATgetArgument(arg_arg0_arg0_head, 0);
                                                            /* checking for: opt */
                                                            if (ATgetType((ATerm)arg_arg0_arg0_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0) == PERR_afun3) {
                                                              ATerm arg_arg0_arg0_head_arg0_arg0 = ATgetArgument(arg_arg0_arg0_head_arg0, 0);
                                                              /* checking for: layout */
                                                              if (ATgetType((ATerm)arg_arg0_arg0_head_arg0_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0_arg0) == PERR_afun4) {
                                                                if (ATgetType((ATerm)arg_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0_list) == ATfalse) {
                                                                  arg_arg0_arg0_head = ATgetFirst(arg_arg0_arg0_list);
                                                                  arg_arg0_arg0_list = ATgetNext(arg_arg0_arg0_list);
                                                                  /* checking for: lit */
                                                                  if (ATgetType((ATerm)arg_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head) == PERR_afun9) {
                                                                    ATerm arg_arg0_arg0_head_arg0 = ATgetArgument(arg_arg0_arg0_head, 0);
                                                                    /* checking for: [ */
                                                                    if (ATgetType((ATerm)arg_arg0_arg0_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0) == PERR_afun47) {
                                                                      if (ATgetType((ATerm)arg_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0_list) == ATfalse) {
                                                                        arg_arg0_arg0_head = ATgetFirst(arg_arg0_arg0_list);
                                                                        arg_arg0_arg0_list = ATgetNext(arg_arg0_arg0_list);
                                                                        /* checking for: cf */
                                                                        if (ATgetType((ATerm)arg_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head) == PERR_afun2) {
                                                                          ATerm arg_arg0_arg0_head_arg0 = ATgetArgument(arg_arg0_arg0_head, 0);
                                                                          /* checking for: opt */
                                                                          if (ATgetType((ATerm)arg_arg0_arg0_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0) == PERR_afun3) {
                                                                            ATerm arg_arg0_arg0_head_arg0_arg0 = ATgetArgument(arg_arg0_arg0_head_arg0, 0);
                                                                            /* checking for: layout */
                                                                            if (ATgetType((ATerm)arg_arg0_arg0_head_arg0_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0_arg0) == PERR_afun4) {
                                                                              if (ATgetType((ATerm)arg_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0_list) == ATfalse) {
                                                                                arg_arg0_arg0_head = ATgetFirst(arg_arg0_arg0_list);
                                                                                arg_arg0_arg0_list = ATgetNext(arg_arg0_arg0_list);
                                                                                /* checking for: cf */
                                                                                if (ATgetType((ATerm)arg_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head) == PERR_afun2) {
                                                                                  ATerm arg_arg0_arg0_head_arg0 = ATgetArgument(arg_arg0_arg0_head, 0);
                                                                                  /* checking for: iter-star-sep */
                                                                                  if (ATgetType((ATerm)arg_arg0_arg0_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0) == PERR_afun46) {
                                                                                    ATerm arg_arg0_arg0_head_arg0_arg0 = ATgetArgument(arg_arg0_arg0_head_arg0, 0);
                                                                                    /* checking for: sort */
                                                                                    if (ATgetType((ATerm)arg_arg0_arg0_head_arg0_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0_arg0) == PERR_afun11) {
                                                                                      ATerm arg_arg0_arg0_head_arg0_arg0_arg0 = ATgetArgument(arg_arg0_arg0_head_arg0_arg0, 0);
                                                                                      /* checking for: Subject */
                                                                                      if (ATgetType((ATerm)arg_arg0_arg0_head_arg0_arg0_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0_arg0_arg0) == PERR_afun41) {
                                                                                        ATerm arg_arg0_arg0_head_arg0_arg1 = ATgetArgument(arg_arg0_arg0_head_arg0, 1);
                                                                                        /* checking for: lit */
                                                                                        if (ATgetType((ATerm)arg_arg0_arg0_head_arg0_arg1) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0_arg1) == PERR_afun9) {
                                                                                          ATerm arg_arg0_arg0_head_arg0_arg1_arg0 = ATgetArgument(arg_arg0_arg0_head_arg0_arg1, 0);
                                                                                          /* checking for: , */
                                                                                          if (ATgetType((ATerm)arg_arg0_arg0_head_arg0_arg1_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0_arg1_arg0) == PERR_afun43) {
                                                                                            if (ATgetType((ATerm)arg_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0_list) == ATfalse) {
                                                                                              arg_arg0_arg0_head = ATgetFirst(arg_arg0_arg0_list);
                                                                                              arg_arg0_arg0_list = ATgetNext(arg_arg0_arg0_list);
                                                                                              /* checking for: cf */
                                                                                              if (ATgetType((ATerm)arg_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head) == PERR_afun2) {
                                                                                                ATerm arg_arg0_arg0_head_arg0 = ATgetArgument(arg_arg0_arg0_head, 0);
                                                                                                /* checking for: opt */
                                                                                                if (ATgetType((ATerm)arg_arg0_arg0_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0) == PERR_afun3) {
                                                                                                  ATerm arg_arg0_arg0_head_arg0_arg0 = ATgetArgument(arg_arg0_arg0_head_arg0, 0);
                                                                                                  /* checking for: layout */
                                                                                                  if (ATgetType((ATerm)arg_arg0_arg0_head_arg0_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0_arg0) == PERR_afun4) {
                                                                                                    if (ATgetType((ATerm)arg_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0_list) == ATfalse) {
                                                                                                      arg_arg0_arg0_head = ATgetFirst(arg_arg0_arg0_list);
                                                                                                      arg_arg0_arg0_list = ATgetNext(arg_arg0_arg0_list);
                                                                                                      /* checking for: lit */
                                                                                                      if (ATgetType((ATerm)arg_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head) == PERR_afun9) {
                                                                                                        ATerm arg_arg0_arg0_head_arg0 = ATgetArgument(arg_arg0_arg0_head, 0);
                                                                                                        /* checking for: ] */
                                                                                                        if (ATgetType((ATerm)arg_arg0_arg0_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0) == PERR_afun45) {
                                                                                                          if (ATgetType((ATerm)arg_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0_list) == ATfalse) {
                                                                                                            arg_arg0_arg0_head = ATgetFirst(arg_arg0_arg0_list);
                                                                                                            arg_arg0_arg0_list = ATgetNext(arg_arg0_arg0_list);
                                                                                                            /* checking for: cf */
                                                                                                            if (ATgetType((ATerm)arg_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head) == PERR_afun2) {
                                                                                                              ATerm arg_arg0_arg0_head_arg0 = ATgetArgument(arg_arg0_arg0_head, 0);
                                                                                                              /* checking for: opt */
                                                                                                              if (ATgetType((ATerm)arg_arg0_arg0_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0) == PERR_afun3) {
                                                                                                                ATerm arg_arg0_arg0_head_arg0_arg0 = ATgetArgument(arg_arg0_arg0_head_arg0, 0);
                                                                                                                /* checking for: layout */
                                                                                                                if (ATgetType((ATerm)arg_arg0_arg0_head_arg0_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0_arg0) == PERR_afun4) {
                                                                                                                  if (ATgetType((ATerm)arg_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0_list) == ATfalse) {
                                                                                                                    arg_arg0_arg0_head = ATgetFirst(arg_arg0_arg0_list);
                                                                                                                    arg_arg0_arg0_list = ATgetNext(arg_arg0_arg0_list);
                                                                                                                    /* checking for: lit */
                                                                                                                    if (ATgetType((ATerm)arg_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head) == PERR_afun9) {
                                                                                                                      ATerm arg_arg0_arg0_head_arg0 = ATgetArgument(arg_arg0_arg0_head, 0);
                                                                                                                      /* checking for: ) */
                                                                                                                      if (ATgetType((ATerm)arg_arg0_arg0_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0) == PERR_afun38) {
                                                                                                                        if (ATgetType((ATerm)arg_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0_list) == ATtrue) {
                                                                                                                          ATerm arg_arg0_arg1 = ATgetArgument(arg_arg0, 1);
                                                                                                                          /* checking for: cf */
                                                                                                                          if (ATgetType((ATerm)arg_arg0_arg1) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg1) == PERR_afun2) {
                                                                                                                            ATerm arg_arg0_arg1_arg0 = ATgetArgument(arg_arg0_arg1, 0);
                                                                                                                            /* checking for: sort */
                                                                                                                            if (ATgetType((ATerm)arg_arg0_arg1_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg1_arg0) == PERR_afun11) {
                                                                                                                              ATerm arg_arg0_arg1_arg0_arg0 = ATgetArgument(arg_arg0_arg1_arg0, 0);
                                                                                                                              /* checking for: Error */
                                                                                                                              if (ATgetType((ATerm)arg_arg0_arg1_arg0_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg1_arg0_arg0) == PERR_afun37) {
                                                                                                                                ATerm arg_arg0_arg2 = ATgetArgument(arg_arg0, 2);
                                                                                                                                /* checking for: attrs */
                                                                                                                                if (ATgetType((ATerm)arg_arg0_arg2) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg2) == PERR_afun13) {
                                                                                                                                  ATerm arg_arg0_arg2_arg0 = ATgetArgument(arg_arg0_arg2, 0);
                                                                                                                                  if (ATgetType((ATerm)arg_arg0_arg2_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg2_arg0) == ATfalse) {
                                                                                                                                    ATerm arg_arg0_arg2_arg0_head;
                                                                                                                                    ATermList arg_arg0_arg2_arg0_list = (ATermList)arg_arg0_arg2_arg0;
                                                                                                                                    arg_arg0_arg2_arg0_head = ATgetFirst(arg_arg0_arg2_arg0_list);
                                                                                                                                    arg_arg0_arg2_arg0_list = ATgetNext(arg_arg0_arg2_arg0_list);
                                                                                                                                    /* checking for: term */
                                                                                                                                    if (ATgetType((ATerm)arg_arg0_arg2_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg2_arg0_head) == PERR_afun14) {
                                                                                                                                      ATerm arg_arg0_arg2_arg0_head_arg0 = ATgetArgument(arg_arg0_arg2_arg0_head, 0);
                                                                                                                                      /* checking for: cons */
                                                                                                                                      if (ATgetType((ATerm)arg_arg0_arg2_arg0_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg2_arg0_head_arg0) == PERR_afun15) {
                                                                                                                                        ATerm arg_arg0_arg2_arg0_head_arg0_arg0 = ATgetArgument(arg_arg0_arg2_arg0_head_arg0, 0);
                                                                                                                                        /* checking for: warning */
                                                                                                                                        if (ATgetType((ATerm)arg_arg0_arg2_arg0_head_arg0_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg2_arg0_head_arg0_arg0) == PERR_afun49) {
                                                                                                                                          if (ATgetType((ATerm)arg_arg0_arg2_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg2_arg0_list) == ATtrue) {
                                                                                                                                            ATerm arg_arg1 = ATgetArgument(arg, 1);
                                                                                                                                            if (ATgetType((ATerm)arg_arg1) == AT_LIST && ATisEmpty((ATermList)arg_arg1) == ATfalse) {
                                                                                                                                              ATerm arg_arg1_head;
                                                                                                                                              ATermList arg_arg1_list = (ATermList)arg_arg1;
                                                                                                                                              arg_arg1_head = ATgetFirst(arg_arg1_list);
                                                                                                                                              arg_arg1_list = ATgetNext(arg_arg1_list);
                                                                                                                                              /* checking for: appl */
                                                                                                                                              if (ATgetType((ATerm)arg_arg1_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head) == PERR_afun0) {
                                                                                                                                                ATerm arg_arg1_head_arg0 = ATgetArgument(arg_arg1_head, 0);
                                                                                                                                                /* checking for: prod */
                                                                                                                                                if (ATgetType((ATerm)arg_arg1_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0) == PERR_afun1) {
                                                                                                                                                  ATerm arg_arg1_head_arg0_arg0 = ATgetArgument(arg_arg1_head_arg0, 0);
                                                                                                                                                  if (ATgetType((ATerm)arg_arg1_head_arg0_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0) == ATfalse) {
                                                                                                                                                    ATerm arg_arg1_head_arg0_arg0_head;
                                                                                                                                                    ATermList arg_arg1_head_arg0_arg0_list = (ATermList)arg_arg1_head_arg0_arg0;
                                                                                                                                                    arg_arg1_head_arg0_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_list);
                                                                                                                                                    arg_arg1_head_arg0_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_list);
                                                                                                                                                    /* checking for: char-class */
                                                                                                                                                    if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg0_head) == PERR_afun17) {
                                                                                                                                                      ATerm arg_arg1_head_arg0_arg0_head_arg0 = ATgetArgument(arg_arg1_head_arg0_arg0_head, 0);
                                                                                                                                                      if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0) == ATfalse) {
                                                                                                                                                        ATerm arg_arg1_head_arg0_arg0_head_arg0_head;
                                                                                                                                                        ATermList arg_arg1_head_arg0_arg0_head_arg0_list = (ATermList)arg_arg1_head_arg0_arg0_head_arg0;
                                                                                                                                                        arg_arg1_head_arg0_arg0_head_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                                                                        arg_arg1_head_arg0_arg0_head_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                                                                        if (ATgetInt((ATermInt)arg_arg1_head_arg0_arg0_head_arg0_head) == 119) {

                                                                                                                                                          if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0_list) == ATtrue) {
                                                                                                                                                            if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_list) == ATfalse) {
                                                                                                                                                              arg_arg1_head_arg0_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_list);
                                                                                                                                                              arg_arg1_head_arg0_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_list);
                                                                                                                                                              /* checking for: char-class */
                                                                                                                                                              if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg0_head) == PERR_afun17) {
                                                                                                                                                                ATerm arg_arg1_head_arg0_arg0_head_arg0 = ATgetArgument(arg_arg1_head_arg0_arg0_head, 0);
                                                                                                                                                                if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0) == ATfalse) {
                                                                                                                                                                  ATerm arg_arg1_head_arg0_arg0_head_arg0_head;
                                                                                                                                                                  ATermList arg_arg1_head_arg0_arg0_head_arg0_list = (ATermList)arg_arg1_head_arg0_arg0_head_arg0;
                                                                                                                                                                  arg_arg1_head_arg0_arg0_head_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                                                                                  arg_arg1_head_arg0_arg0_head_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                                                                                  if (ATgetInt((ATermInt)arg_arg1_head_arg0_arg0_head_arg0_head) == 97) {

                                                                                                                                                                    if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0_list) == ATtrue) {
                                                                                                                                                                      if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_list) == ATfalse) {
                                                                                                                                                                        arg_arg1_head_arg0_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_list);
                                                                                                                                                                        arg_arg1_head_arg0_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_list);
                                                                                                                                                                        /* checking for: char-class */
                                                                                                                                                                        if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg0_head) == PERR_afun17) {
                                                                                                                                                                          ATerm arg_arg1_head_arg0_arg0_head_arg0 = ATgetArgument(arg_arg1_head_arg0_arg0_head, 0);
                                                                                                                                                                          if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0) == ATfalse) {
                                                                                                                                                                            ATerm arg_arg1_head_arg0_arg0_head_arg0_head;
                                                                                                                                                                            ATermList arg_arg1_head_arg0_arg0_head_arg0_list = (ATermList)arg_arg1_head_arg0_arg0_head_arg0;
                                                                                                                                                                            arg_arg1_head_arg0_arg0_head_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                                                                                            arg_arg1_head_arg0_arg0_head_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                                                                                            if (ATgetInt((ATermInt)arg_arg1_head_arg0_arg0_head_arg0_head) == 114) {

                                                                                                                                                                              if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0_list) == ATtrue) {
                                                                                                                                                                                if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_list) == ATfalse) {
                                                                                                                                                                                  arg_arg1_head_arg0_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_list);
                                                                                                                                                                                  arg_arg1_head_arg0_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_list);
                                                                                                                                                                                  /* checking for: char-class */
                                                                                                                                                                                  if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg0_head) == PERR_afun17) {
                                                                                                                                                                                    ATerm arg_arg1_head_arg0_arg0_head_arg0 = ATgetArgument(arg_arg1_head_arg0_arg0_head, 0);
                                                                                                                                                                                    if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0) == ATfalse) {
                                                                                                                                                                                      ATerm arg_arg1_head_arg0_arg0_head_arg0_head;
                                                                                                                                                                                      ATermList arg_arg1_head_arg0_arg0_head_arg0_list = (ATermList)arg_arg1_head_arg0_arg0_head_arg0;
                                                                                                                                                                                      arg_arg1_head_arg0_arg0_head_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                                                                                                      arg_arg1_head_arg0_arg0_head_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                                                                                                      if (ATgetInt((ATermInt)arg_arg1_head_arg0_arg0_head_arg0_head) == 110) {

                                                                                                                                                                                        if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0_list) == ATtrue) {
                                                                                                                                                                                          if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_list) == ATfalse) {
                                                                                                                                                                                            arg_arg1_head_arg0_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_list);
                                                                                                                                                                                            arg_arg1_head_arg0_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_list);
                                                                                                                                                                                            /* checking for: char-class */
                                                                                                                                                                                            if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg0_head) == PERR_afun17) {
                                                                                                                                                                                              ATerm arg_arg1_head_arg0_arg0_head_arg0 = ATgetArgument(arg_arg1_head_arg0_arg0_head, 0);
                                                                                                                                                                                              if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0) == ATfalse) {
                                                                                                                                                                                                ATerm arg_arg1_head_arg0_arg0_head_arg0_head;
                                                                                                                                                                                                ATermList arg_arg1_head_arg0_arg0_head_arg0_list = (ATermList)arg_arg1_head_arg0_arg0_head_arg0;
                                                                                                                                                                                                arg_arg1_head_arg0_arg0_head_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                                                                                                                arg_arg1_head_arg0_arg0_head_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                                                                                                                if (ATgetInt((ATermInt)arg_arg1_head_arg0_arg0_head_arg0_head) == 105) {

                                                                                                                                                                                                  if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0_list) == ATtrue) {
                                                                                                                                                                                                    if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_list) == ATfalse) {
                                                                                                                                                                                                      arg_arg1_head_arg0_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_list);
                                                                                                                                                                                                      arg_arg1_head_arg0_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_list);
                                                                                                                                                                                                      /* checking for: char-class */
                                                                                                                                                                                                      if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg0_head) == PERR_afun17) {
                                                                                                                                                                                                        ATerm arg_arg1_head_arg0_arg0_head_arg0 = ATgetArgument(arg_arg1_head_arg0_arg0_head, 0);
                                                                                                                                                                                                        if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0) == ATfalse) {
                                                                                                                                                                                                          ATerm arg_arg1_head_arg0_arg0_head_arg0_head;
                                                                                                                                                                                                          ATermList arg_arg1_head_arg0_arg0_head_arg0_list = (ATermList)arg_arg1_head_arg0_arg0_head_arg0;
                                                                                                                                                                                                          arg_arg1_head_arg0_arg0_head_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                                                                                                                          arg_arg1_head_arg0_arg0_head_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                                                                                                                          if (ATgetInt((ATermInt)arg_arg1_head_arg0_arg0_head_arg0_head) == 110) {

                                                                                                                                                                                                            if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0_list) == ATtrue) {
                                                                                                                                                                                                              if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_list) == ATfalse) {
                                                                                                                                                                                                                arg_arg1_head_arg0_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_list);
                                                                                                                                                                                                                arg_arg1_head_arg0_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_list);
                                                                                                                                                                                                                /* checking for: char-class */
                                                                                                                                                                                                                if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg0_head) == PERR_afun17) {
                                                                                                                                                                                                                  ATerm arg_arg1_head_arg0_arg0_head_arg0 = ATgetArgument(arg_arg1_head_arg0_arg0_head, 0);
                                                                                                                                                                                                                  if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0) == ATfalse) {
                                                                                                                                                                                                                    ATerm arg_arg1_head_arg0_arg0_head_arg0_head;
                                                                                                                                                                                                                    ATermList arg_arg1_head_arg0_arg0_head_arg0_list = (ATermList)arg_arg1_head_arg0_arg0_head_arg0;
                                                                                                                                                                                                                    arg_arg1_head_arg0_arg0_head_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                                                                                                                                    arg_arg1_head_arg0_arg0_head_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                                                                                                                                    if (ATgetInt((ATermInt)arg_arg1_head_arg0_arg0_head_arg0_head) == 103) {

                                                                                                                                                                                                                      if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0_list) == ATtrue) {
                                                                                                                                                                                                                        if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_list) == ATtrue) {
                                                                                                                                                                                                                          ATerm arg_arg1_head_arg0_arg1 = ATgetArgument(arg_arg1_head_arg0, 1);
                                                                                                                                                                                                                          /* checking for: lit */
                                                                                                                                                                                                                          if (ATgetType((ATerm)arg_arg1_head_arg0_arg1) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg1) == PERR_afun9) {
                                                                                                                                                                                                                            ATerm arg_arg1_head_arg0_arg1_arg0 = ATgetArgument(arg_arg1_head_arg0_arg1, 0);
                                                                                                                                                                                                                            /* checking for: warning */
                                                                                                                                                                                                                            if (ATgetType((ATerm)arg_arg1_head_arg0_arg1_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg1_arg0) == PERR_afun49) {
                                                                                                                                                                                                                              ATerm arg_arg1_head_arg0_arg2 = ATgetArgument(arg_arg1_head_arg0, 2);
                                                                                                                                                                                                                              /* checking for: no-attrs */
                                                                                                                                                                                                                              if (ATgetType((ATerm)arg_arg1_head_arg0_arg2) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg2) == PERR_afun5) {
                                                                                                                                                                                                                                ATerm arg_arg1_head_arg1 = ATgetArgument(arg_arg1_head, 1);
                                                                                                                                                                                                                                if (ATgetType((ATerm)arg_arg1_head_arg1) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg1) == ATfalse) {
                                                                                                                                                                                                                                  ATerm arg_arg1_head_arg1_head;
                                                                                                                                                                                                                                  ATermList arg_arg1_head_arg1_list = (ATermList)arg_arg1_head_arg1;
                                                                                                                                                                                                                                  arg_arg1_head_arg1_head = ATgetFirst(arg_arg1_head_arg1_list);
                                                                                                                                                                                                                                  arg_arg1_head_arg1_list = ATgetNext(arg_arg1_head_arg1_list);
                                                                                                                                                                                                                                  if (ATgetInt((ATermInt)arg_arg1_head_arg1_head) == 119) {

                                                                                                                                                                                                                                    if (ATgetType((ATerm)arg_arg1_head_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg1_list) == ATfalse) {
                                                                                                                                                                                                                                      arg_arg1_head_arg1_head = ATgetFirst(arg_arg1_head_arg1_list);
                                                                                                                                                                                                                                      arg_arg1_head_arg1_list = ATgetNext(arg_arg1_head_arg1_list);
                                                                                                                                                                                                                                      if (ATgetInt((ATermInt)arg_arg1_head_arg1_head) == 97) {

                                                                                                                                                                                                                                        if (ATgetType((ATerm)arg_arg1_head_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg1_list) == ATfalse) {
                                                                                                                                                                                                                                          arg_arg1_head_arg1_head = ATgetFirst(arg_arg1_head_arg1_list);
                                                                                                                                                                                                                                          arg_arg1_head_arg1_list = ATgetNext(arg_arg1_head_arg1_list);
                                                                                                                                                                                                                                          if (ATgetInt((ATermInt)arg_arg1_head_arg1_head) == 114) {

                                                                                                                                                                                                                                            if (ATgetType((ATerm)arg_arg1_head_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg1_list) == ATfalse) {
                                                                                                                                                                                                                                              arg_arg1_head_arg1_head = ATgetFirst(arg_arg1_head_arg1_list);
                                                                                                                                                                                                                                              arg_arg1_head_arg1_list = ATgetNext(arg_arg1_head_arg1_list);
                                                                                                                                                                                                                                              if (ATgetInt((ATermInt)arg_arg1_head_arg1_head) == 110) {

                                                                                                                                                                                                                                                if (ATgetType((ATerm)arg_arg1_head_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg1_list) == ATfalse) {
                                                                                                                                                                                                                                                  arg_arg1_head_arg1_head = ATgetFirst(arg_arg1_head_arg1_list);
                                                                                                                                                                                                                                                  arg_arg1_head_arg1_list = ATgetNext(arg_arg1_head_arg1_list);
                                                                                                                                                                                                                                                  if (ATgetInt((ATermInt)arg_arg1_head_arg1_head) == 105) {

                                                                                                                                                                                                                                                    if (ATgetType((ATerm)arg_arg1_head_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg1_list) == ATfalse) {
                                                                                                                                                                                                                                                      arg_arg1_head_arg1_head = ATgetFirst(arg_arg1_head_arg1_list);
                                                                                                                                                                                                                                                      arg_arg1_head_arg1_list = ATgetNext(arg_arg1_head_arg1_list);
                                                                                                                                                                                                                                                      if (ATgetInt((ATermInt)arg_arg1_head_arg1_head) == 110) {

                                                                                                                                                                                                                                                        if (ATgetType((ATerm)arg_arg1_head_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg1_list) == ATfalse) {
                                                                                                                                                                                                                                                          arg_arg1_head_arg1_head = ATgetFirst(arg_arg1_head_arg1_list);
                                                                                                                                                                                                                                                          arg_arg1_head_arg1_list = ATgetNext(arg_arg1_head_arg1_list);
                                                                                                                                                                                                                                                          if (ATgetInt((ATermInt)arg_arg1_head_arg1_head) == 103) {

                                                                                                                                                                                                                                                            if (ATgetType((ATerm)arg_arg1_head_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg1_list) == ATtrue) {
                                                                                                                                                                                                                                                              if (ATgetType((ATerm)arg_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_list) == ATfalse) {
                                                                                                                                                                                                                                                                arg_arg1_list = ATgetNext(arg_arg1_list);
                                                                                                                                                                                                                                                                if (ATgetType((ATerm)arg_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_list) == ATfalse) {
                                                                                                                                                                                                                                                                  arg_arg1_head = ATgetFirst(arg_arg1_list);
                                                                                                                                                                                                                                                                  arg_arg1_list = ATgetNext(arg_arg1_list);
                                                                                                                                                                                                                                                                  /* checking for: appl */
                                                                                                                                                                                                                                                                  if (ATgetType((ATerm)arg_arg1_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head) == PERR_afun0) {
                                                                                                                                                                                                                                                                    ATerm arg_arg1_head_arg0 = ATgetArgument(arg_arg1_head, 0);
                                                                                                                                                                                                                                                                    /* checking for: prod */
                                                                                                                                                                                                                                                                    if (ATgetType((ATerm)arg_arg1_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0) == PERR_afun1) {
                                                                                                                                                                                                                                                                      ATerm arg_arg1_head_arg0_arg0 = ATgetArgument(arg_arg1_head_arg0, 0);
                                                                                                                                                                                                                                                                      if (ATgetType((ATerm)arg_arg1_head_arg0_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0) == ATfalse) {
                                                                                                                                                                                                                                                                        ATerm arg_arg1_head_arg0_arg0_head;
                                                                                                                                                                                                                                                                        ATermList arg_arg1_head_arg0_arg0_list = (ATermList)arg_arg1_head_arg0_arg0;
                                                                                                                                                                                                                                                                        arg_arg1_head_arg0_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_list);
                                                                                                                                                                                                                                                                        arg_arg1_head_arg0_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_list);
                                                                                                                                                                                                                                                                        /* checking for: char-class */
                                                                                                                                                                                                                                                                        if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg0_head) == PERR_afun17) {
                                                                                                                                                                                                                                                                          ATerm arg_arg1_head_arg0_arg0_head_arg0 = ATgetArgument(arg_arg1_head_arg0_arg0_head, 0);
                                                                                                                                                                                                                                                                          if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0) == ATfalse) {
                                                                                                                                                                                                                                                                            ATerm arg_arg1_head_arg0_arg0_head_arg0_head;
                                                                                                                                                                                                                                                                            ATermList arg_arg1_head_arg0_arg0_head_arg0_list = (ATermList)arg_arg1_head_arg0_arg0_head_arg0;
                                                                                                                                                                                                                                                                            arg_arg1_head_arg0_arg0_head_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                                                                                                                                                                                            arg_arg1_head_arg0_arg0_head_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                                                                                                                                                                                            if (ATgetInt((ATermInt)arg_arg1_head_arg0_arg0_head_arg0_head) == 40) {

                                                                                                                                                                                                                                                                              if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0_list) == ATtrue) {
                                                                                                                                                                                                                                                                                if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_list) == ATtrue) {
                                                                                                                                                                                                                                                                                  ATerm arg_arg1_head_arg0_arg1 = ATgetArgument(arg_arg1_head_arg0, 1);
                                                                                                                                                                                                                                                                                  /* checking for: lit */
                                                                                                                                                                                                                                                                                  if (ATgetType((ATerm)arg_arg1_head_arg0_arg1) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg1) == PERR_afun9) {
                                                                                                                                                                                                                                                                                    ATerm arg_arg1_head_arg0_arg1_arg0 = ATgetArgument(arg_arg1_head_arg0_arg1, 0);
                                                                                                                                                                                                                                                                                    /* checking for: ( */
                                                                                                                                                                                                                                                                                    if (ATgetType((ATerm)arg_arg1_head_arg0_arg1_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg1_arg0) == PERR_afun39) {
                                                                                                                                                                                                                                                                                      ATerm arg_arg1_head_arg0_arg2 = ATgetArgument(arg_arg1_head_arg0, 2);
                                                                                                                                                                                                                                                                                      /* checking for: no-attrs */
                                                                                                                                                                                                                                                                                      if (ATgetType((ATerm)arg_arg1_head_arg0_arg2) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg2) == PERR_afun5) {
                                                                                                                                                                                                                                                                                        ATerm arg_arg1_head_arg1 = ATgetArgument(arg_arg1_head, 1);
                                                                                                                                                                                                                                                                                        if (ATgetType((ATerm)arg_arg1_head_arg1) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg1) == ATfalse) {
                                                                                                                                                                                                                                                                                          ATerm arg_arg1_head_arg1_head;
                                                                                                                                                                                                                                                                                          ATermList arg_arg1_head_arg1_list = (ATermList)arg_arg1_head_arg1;
                                                                                                                                                                                                                                                                                          arg_arg1_head_arg1_head = ATgetFirst(arg_arg1_head_arg1_list);
                                                                                                                                                                                                                                                                                          arg_arg1_head_arg1_list = ATgetNext(arg_arg1_head_arg1_list);
                                                                                                                                                                                                                                                                                          if (ATgetInt((ATermInt)arg_arg1_head_arg1_head) == 40) {

                                                                                                                                                                                                                                                                                            if (ATgetType((ATerm)arg_arg1_head_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg1_list) == ATtrue) {
                                                                                                                                                                                                                                                                                              if (ATgetType((ATerm)arg_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_list) == ATfalse) {
                                                                                                                                                                                                                                                                                                arg_arg1_list = ATgetNext(arg_arg1_list);
                                                                                                                                                                                                                                                                                                if (ATgetType((ATerm)arg_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_list) == ATfalse) {
                                                                                                                                                                                                                                                                                                  arg_arg1_list = ATgetNext(arg_arg1_list);
                                                                                                                                                                                                                                                                                                  if (ATgetType((ATerm)arg_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_list) == ATfalse) {
                                                                                                                                                                                                                                                                                                    arg_arg1_list = ATgetNext(arg_arg1_list);
                                                                                                                                                                                                                                                                                                    if (ATgetType((ATerm)arg_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_list) == ATfalse) {
                                                                                                                                                                                                                                                                                                      arg_arg1_head = ATgetFirst(arg_arg1_list);
                                                                                                                                                                                                                                                                                                      arg_arg1_list = ATgetNext(arg_arg1_list);
                                                                                                                                                                                                                                                                                                      /* checking for: appl */
                                                                                                                                                                                                                                                                                                      if (ATgetType((ATerm)arg_arg1_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head) == PERR_afun0) {
                                                                                                                                                                                                                                                                                                        ATerm arg_arg1_head_arg0 = ATgetArgument(arg_arg1_head, 0);
                                                                                                                                                                                                                                                                                                        /* checking for: prod */
                                                                                                                                                                                                                                                                                                        if (ATgetType((ATerm)arg_arg1_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0) == PERR_afun1) {
                                                                                                                                                                                                                                                                                                          ATerm arg_arg1_head_arg0_arg0 = ATgetArgument(arg_arg1_head_arg0, 0);
                                                                                                                                                                                                                                                                                                          if (ATgetType((ATerm)arg_arg1_head_arg0_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0) == ATfalse) {
                                                                                                                                                                                                                                                                                                            ATerm arg_arg1_head_arg0_arg0_head;
                                                                                                                                                                                                                                                                                                            ATermList arg_arg1_head_arg0_arg0_list = (ATermList)arg_arg1_head_arg0_arg0;
                                                                                                                                                                                                                                                                                                            arg_arg1_head_arg0_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_list);
                                                                                                                                                                                                                                                                                                            arg_arg1_head_arg0_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_list);
                                                                                                                                                                                                                                                                                                            /* checking for: char-class */
                                                                                                                                                                                                                                                                                                            if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg0_head) == PERR_afun17) {
                                                                                                                                                                                                                                                                                                              ATerm arg_arg1_head_arg0_arg0_head_arg0 = ATgetArgument(arg_arg1_head_arg0_arg0_head, 0);
                                                                                                                                                                                                                                                                                                              if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0) == ATfalse) {
                                                                                                                                                                                                                                                                                                                ATerm arg_arg1_head_arg0_arg0_head_arg0_head;
                                                                                                                                                                                                                                                                                                                ATermList arg_arg1_head_arg0_arg0_head_arg0_list = (ATermList)arg_arg1_head_arg0_arg0_head_arg0;
                                                                                                                                                                                                                                                                                                                arg_arg1_head_arg0_arg0_head_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                                                                                                                                                                                                                                arg_arg1_head_arg0_arg0_head_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                                                                                                                                                                                                                                if (ATgetInt((ATermInt)arg_arg1_head_arg0_arg0_head_arg0_head) == 44) {

                                                                                                                                                                                                                                                                                                                  if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0_list) == ATtrue) {
                                                                                                                                                                                                                                                                                                                    if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_list) == ATtrue) {
                                                                                                                                                                                                                                                                                                                      ATerm arg_arg1_head_arg0_arg1 = ATgetArgument(arg_arg1_head_arg0, 1);
                                                                                                                                                                                                                                                                                                                      /* checking for: lit */
                                                                                                                                                                                                                                                                                                                      if (ATgetType((ATerm)arg_arg1_head_arg0_arg1) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg1) == PERR_afun9) {
                                                                                                                                                                                                                                                                                                                        ATerm arg_arg1_head_arg0_arg1_arg0 = ATgetArgument(arg_arg1_head_arg0_arg1, 0);
                                                                                                                                                                                                                                                                                                                        /* checking for: , */
                                                                                                                                                                                                                                                                                                                        if (ATgetType((ATerm)arg_arg1_head_arg0_arg1_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg1_arg0) == PERR_afun43) {
                                                                                                                                                                                                                                                                                                                          ATerm arg_arg1_head_arg0_arg2 = ATgetArgument(arg_arg1_head_arg0, 2);
                                                                                                                                                                                                                                                                                                                          /* checking for: no-attrs */
                                                                                                                                                                                                                                                                                                                          if (ATgetType((ATerm)arg_arg1_head_arg0_arg2) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg2) == PERR_afun5) {
                                                                                                                                                                                                                                                                                                                            ATerm arg_arg1_head_arg1 = ATgetArgument(arg_arg1_head, 1);
                                                                                                                                                                                                                                                                                                                            if (ATgetType((ATerm)arg_arg1_head_arg1) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg1) == ATfalse) {
                                                                                                                                                                                                                                                                                                                              ATerm arg_arg1_head_arg1_head;
                                                                                                                                                                                                                                                                                                                              ATermList arg_arg1_head_arg1_list = (ATermList)arg_arg1_head_arg1;
                                                                                                                                                                                                                                                                                                                              arg_arg1_head_arg1_head = ATgetFirst(arg_arg1_head_arg1_list);
                                                                                                                                                                                                                                                                                                                              arg_arg1_head_arg1_list = ATgetNext(arg_arg1_head_arg1_list);
                                                                                                                                                                                                                                                                                                                              if (ATgetInt((ATermInt)arg_arg1_head_arg1_head) == 44) {

                                                                                                                                                                                                                                                                                                                                if (ATgetType((ATerm)arg_arg1_head_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg1_list) == ATtrue) {
                                                                                                                                                                                                                                                                                                                                  if (ATgetType((ATerm)arg_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_list) == ATfalse) {
                                                                                                                                                                                                                                                                                                                                    arg_arg1_list = ATgetNext(arg_arg1_list);
                                                                                                                                                                                                                                                                                                                                    if (ATgetType((ATerm)arg_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_list) == ATfalse) {
                                                                                                                                                                                                                                                                                                                                      arg_arg1_head = ATgetFirst(arg_arg1_list);
                                                                                                                                                                                                                                                                                                                                      arg_arg1_list = ATgetNext(arg_arg1_list);
                                                                                                                                                                                                                                                                                                                                      /* checking for: appl */
                                                                                                                                                                                                                                                                                                                                      if (ATgetType((ATerm)arg_arg1_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head) == PERR_afun0) {
                                                                                                                                                                                                                                                                                                                                        ATerm arg_arg1_head_arg0 = ATgetArgument(arg_arg1_head, 0);
                                                                                                                                                                                                                                                                                                                                        /* checking for: prod */
                                                                                                                                                                                                                                                                                                                                        if (ATgetType((ATerm)arg_arg1_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0) == PERR_afun1) {
                                                                                                                                                                                                                                                                                                                                          ATerm arg_arg1_head_arg0_arg0 = ATgetArgument(arg_arg1_head_arg0, 0);
                                                                                                                                                                                                                                                                                                                                          if (ATgetType((ATerm)arg_arg1_head_arg0_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0) == ATfalse) {
                                                                                                                                                                                                                                                                                                                                            ATerm arg_arg1_head_arg0_arg0_head;
                                                                                                                                                                                                                                                                                                                                            ATermList arg_arg1_head_arg0_arg0_list = (ATermList)arg_arg1_head_arg0_arg0;
                                                                                                                                                                                                                                                                                                                                            arg_arg1_head_arg0_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_list);
                                                                                                                                                                                                                                                                                                                                            arg_arg1_head_arg0_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_list);
                                                                                                                                                                                                                                                                                                                                            /* checking for: char-class */
                                                                                                                                                                                                                                                                                                                                            if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg0_head) == PERR_afun17) {
                                                                                                                                                                                                                                                                                                                                              ATerm arg_arg1_head_arg0_arg0_head_arg0 = ATgetArgument(arg_arg1_head_arg0_arg0_head, 0);
                                                                                                                                                                                                                                                                                                                                              if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0) == ATfalse) {
                                                                                                                                                                                                                                                                                                                                                ATerm arg_arg1_head_arg0_arg0_head_arg0_head;
                                                                                                                                                                                                                                                                                                                                                ATermList arg_arg1_head_arg0_arg0_head_arg0_list = (ATermList)arg_arg1_head_arg0_arg0_head_arg0;
                                                                                                                                                                                                                                                                                                                                                arg_arg1_head_arg0_arg0_head_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                                                                                                                                                                                                                                                                arg_arg1_head_arg0_arg0_head_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                                                                                                                                                                                                                                                                if (ATgetInt((ATermInt)arg_arg1_head_arg0_arg0_head_arg0_head) == 91) {

                                                                                                                                                                                                                                                                                                                                                  if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0_list) == ATtrue) {
                                                                                                                                                                                                                                                                                                                                                    if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_list) == ATtrue) {
                                                                                                                                                                                                                                                                                                                                                      ATerm arg_arg1_head_arg0_arg1 = ATgetArgument(arg_arg1_head_arg0, 1);
                                                                                                                                                                                                                                                                                                                                                      /* checking for: lit */
                                                                                                                                                                                                                                                                                                                                                      if (ATgetType((ATerm)arg_arg1_head_arg0_arg1) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg1) == PERR_afun9) {
                                                                                                                                                                                                                                                                                                                                                        ATerm arg_arg1_head_arg0_arg1_arg0 = ATgetArgument(arg_arg1_head_arg0_arg1, 0);
                                                                                                                                                                                                                                                                                                                                                        /* checking for: [ */
                                                                                                                                                                                                                                                                                                                                                        if (ATgetType((ATerm)arg_arg1_head_arg0_arg1_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg1_arg0) == PERR_afun47) {
                                                                                                                                                                                                                                                                                                                                                          ATerm arg_arg1_head_arg0_arg2 = ATgetArgument(arg_arg1_head_arg0, 2);
                                                                                                                                                                                                                                                                                                                                                          /* checking for: no-attrs */
                                                                                                                                                                                                                                                                                                                                                          if (ATgetType((ATerm)arg_arg1_head_arg0_arg2) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg2) == PERR_afun5) {
                                                                                                                                                                                                                                                                                                                                                            ATerm arg_arg1_head_arg1 = ATgetArgument(arg_arg1_head, 1);
                                                                                                                                                                                                                                                                                                                                                            if (ATgetType((ATerm)arg_arg1_head_arg1) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg1) == ATfalse) {
                                                                                                                                                                                                                                                                                                                                                              ATerm arg_arg1_head_arg1_head;
                                                                                                                                                                                                                                                                                                                                                              ATermList arg_arg1_head_arg1_list = (ATermList)arg_arg1_head_arg1;
                                                                                                                                                                                                                                                                                                                                                              arg_arg1_head_arg1_head = ATgetFirst(arg_arg1_head_arg1_list);
                                                                                                                                                                                                                                                                                                                                                              arg_arg1_head_arg1_list = ATgetNext(arg_arg1_head_arg1_list);
                                                                                                                                                                                                                                                                                                                                                              if (ATgetInt((ATermInt)arg_arg1_head_arg1_head) == 91) {

                                                                                                                                                                                                                                                                                                                                                                if (ATgetType((ATerm)arg_arg1_head_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg1_list) == ATtrue) {
                                                                                                                                                                                                                                                                                                                                                                  if (ATgetType((ATerm)arg_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_list) == ATfalse) {
                                                                                                                                                                                                                                                                                                                                                                    arg_arg1_list = ATgetNext(arg_arg1_list);
                                                                                                                                                                                                                                                                                                                                                                    if (ATgetType((ATerm)arg_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_list) == ATfalse) {
                                                                                                                                                                                                                                                                                                                                                                      arg_arg1_head = ATgetFirst(arg_arg1_list);
                                                                                                                                                                                                                                                                                                                                                                      arg_arg1_list = ATgetNext(arg_arg1_list);
                                                                                                                                                                                                                                                                                                                                                                      /* checking for: appl */
                                                                                                                                                                                                                                                                                                                                                                      if (ATgetType((ATerm)arg_arg1_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head) == PERR_afun0) {
                                                                                                                                                                                                                                                                                                                                                                        ATerm arg_arg1_head_arg0 = ATgetArgument(arg_arg1_head, 0);
                                                                                                                                                                                                                                                                                                                                                                        /* checking for: list */
                                                                                                                                                                                                                                                                                                                                                                        if (ATgetType((ATerm)arg_arg1_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0) == PERR_afun8) {
                                                                                                                                                                                                                                                                                                                                                                          ATerm arg_arg1_head_arg0_arg0 = ATgetArgument(arg_arg1_head_arg0, 0);
                                                                                                                                                                                                                                                                                                                                                                          /* checking for: cf */
                                                                                                                                                                                                                                                                                                                                                                          if (ATgetType((ATerm)arg_arg1_head_arg0_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg0) == PERR_afun2) {
                                                                                                                                                                                                                                                                                                                                                                            ATerm arg_arg1_head_arg0_arg0_arg0 = ATgetArgument(arg_arg1_head_arg0_arg0, 0);
                                                                                                                                                                                                                                                                                                                                                                            /* checking for: iter-star-sep */
                                                                                                                                                                                                                                                                                                                                                                            if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg0_arg0) == PERR_afun46) {
                                                                                                                                                                                                                                                                                                                                                                              ATerm arg_arg1_head_arg0_arg0_arg0_arg0 = ATgetArgument(arg_arg1_head_arg0_arg0_arg0, 0);
                                                                                                                                                                                                                                                                                                                                                                              /* checking for: sort */
                                                                                                                                                                                                                                                                                                                                                                              if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_arg0_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg0_arg0_arg0) == PERR_afun11) {
                                                                                                                                                                                                                                                                                                                                                                                ATerm arg_arg1_head_arg0_arg0_arg0_arg0_arg0 = ATgetArgument(arg_arg1_head_arg0_arg0_arg0_arg0, 0);
                                                                                                                                                                                                                                                                                                                                                                                /* checking for: Subject */
                                                                                                                                                                                                                                                                                                                                                                                if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_arg0_arg0_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg0_arg0_arg0_arg0) == PERR_afun41) {
                                                                                                                                                                                                                                                                                                                                                                                  ATerm arg_arg1_head_arg0_arg0_arg0_arg1 = ATgetArgument(arg_arg1_head_arg0_arg0_arg0, 1);
                                                                                                                                                                                                                                                                                                                                                                                  /* checking for: lit */
                                                                                                                                                                                                                                                                                                                                                                                  if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_arg0_arg1) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg0_arg0_arg1) == PERR_afun9) {
                                                                                                                                                                                                                                                                                                                                                                                    ATerm arg_arg1_head_arg0_arg0_arg0_arg1_arg0 = ATgetArgument(arg_arg1_head_arg0_arg0_arg0_arg1, 0);
                                                                                                                                                                                                                                                                                                                                                                                    /* checking for: , */
                                                                                                                                                                                                                                                                                                                                                                                    if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_arg0_arg1_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg0_arg0_arg1_arg0) == PERR_afun43) {
                                                                                                                                                                                                                                                                                                                                                                                      if (ATgetType((ATerm)arg_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_list) == ATfalse) {
                                                                                                                                                                                                                                                                                                                                                                                        arg_arg1_list = ATgetNext(arg_arg1_list);
                                                                                                                                                                                                                                                                                                                                                                                        if (ATgetType((ATerm)arg_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_list) == ATfalse) {
                                                                                                                                                                                                                                                                                                                                                                                          arg_arg1_head = ATgetFirst(arg_arg1_list);
                                                                                                                                                                                                                                                                                                                                                                                          arg_arg1_list = ATgetNext(arg_arg1_list);
                                                                                                                                                                                                                                                                                                                                                                                          /* checking for: appl */
                                                                                                                                                                                                                                                                                                                                                                                          if (ATgetType((ATerm)arg_arg1_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head) == PERR_afun0) {
                                                                                                                                                                                                                                                                                                                                                                                            ATerm arg_arg1_head_arg0 = ATgetArgument(arg_arg1_head, 0);
                                                                                                                                                                                                                                                                                                                                                                                            /* checking for: prod */
                                                                                                                                                                                                                                                                                                                                                                                            if (ATgetType((ATerm)arg_arg1_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0) == PERR_afun1) {
                                                                                                                                                                                                                                                                                                                                                                                              ATerm arg_arg1_head_arg0_arg0 = ATgetArgument(arg_arg1_head_arg0, 0);
                                                                                                                                                                                                                                                                                                                                                                                              if (ATgetType((ATerm)arg_arg1_head_arg0_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0) == ATfalse) {
                                                                                                                                                                                                                                                                                                                                                                                                ATerm arg_arg1_head_arg0_arg0_head;
                                                                                                                                                                                                                                                                                                                                                                                                ATermList arg_arg1_head_arg0_arg0_list = (ATermList)arg_arg1_head_arg0_arg0;
                                                                                                                                                                                                                                                                                                                                                                                                arg_arg1_head_arg0_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_list);
                                                                                                                                                                                                                                                                                                                                                                                                arg_arg1_head_arg0_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_list);
                                                                                                                                                                                                                                                                                                                                                                                                /* checking for: char-class */
                                                                                                                                                                                                                                                                                                                                                                                                if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg0_head) == PERR_afun17) {
                                                                                                                                                                                                                                                                                                                                                                                                  ATerm arg_arg1_head_arg0_arg0_head_arg0 = ATgetArgument(arg_arg1_head_arg0_arg0_head, 0);
                                                                                                                                                                                                                                                                                                                                                                                                  if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0) == ATfalse) {
                                                                                                                                                                                                                                                                                                                                                                                                    ATerm arg_arg1_head_arg0_arg0_head_arg0_head;
                                                                                                                                                                                                                                                                                                                                                                                                    ATermList arg_arg1_head_arg0_arg0_head_arg0_list = (ATermList)arg_arg1_head_arg0_arg0_head_arg0;
                                                                                                                                                                                                                                                                                                                                                                                                    arg_arg1_head_arg0_arg0_head_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                                                                                                                                                                                                                                                                                                                    arg_arg1_head_arg0_arg0_head_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                                                                                                                                                                                                                                                                                                                    if (ATgetInt((ATermInt)arg_arg1_head_arg0_arg0_head_arg0_head) == 93) {

                                                                                                                                                                                                                                                                                                                                                                                                      if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0_list) == ATtrue) {
                                                                                                                                                                                                                                                                                                                                                                                                        if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_list) == ATtrue) {
                                                                                                                                                                                                                                                                                                                                                                                                          ATerm arg_arg1_head_arg0_arg1 = ATgetArgument(arg_arg1_head_arg0, 1);
                                                                                                                                                                                                                                                                                                                                                                                                          /* checking for: lit */
                                                                                                                                                                                                                                                                                                                                                                                                          if (ATgetType((ATerm)arg_arg1_head_arg0_arg1) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg1) == PERR_afun9) {
                                                                                                                                                                                                                                                                                                                                                                                                            ATerm arg_arg1_head_arg0_arg1_arg0 = ATgetArgument(arg_arg1_head_arg0_arg1, 0);
                                                                                                                                                                                                                                                                                                                                                                                                            /* checking for: ] */
                                                                                                                                                                                                                                                                                                                                                                                                            if (ATgetType((ATerm)arg_arg1_head_arg0_arg1_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg1_arg0) == PERR_afun45) {
                                                                                                                                                                                                                                                                                                                                                                                                              ATerm arg_arg1_head_arg0_arg2 = ATgetArgument(arg_arg1_head_arg0, 2);
                                                                                                                                                                                                                                                                                                                                                                                                              /* checking for: no-attrs */
                                                                                                                                                                                                                                                                                                                                                                                                              if (ATgetType((ATerm)arg_arg1_head_arg0_arg2) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg2) == PERR_afun5) {
                                                                                                                                                                                                                                                                                                                                                                                                                ATerm arg_arg1_head_arg1 = ATgetArgument(arg_arg1_head, 1);
                                                                                                                                                                                                                                                                                                                                                                                                                if (ATgetType((ATerm)arg_arg1_head_arg1) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg1) == ATfalse) {
                                                                                                                                                                                                                                                                                                                                                                                                                  ATerm arg_arg1_head_arg1_head;
                                                                                                                                                                                                                                                                                                                                                                                                                  ATermList arg_arg1_head_arg1_list = (ATermList)arg_arg1_head_arg1;
                                                                                                                                                                                                                                                                                                                                                                                                                  arg_arg1_head_arg1_head = ATgetFirst(arg_arg1_head_arg1_list);
                                                                                                                                                                                                                                                                                                                                                                                                                  arg_arg1_head_arg1_list = ATgetNext(arg_arg1_head_arg1_list);
                                                                                                                                                                                                                                                                                                                                                                                                                  if (ATgetInt((ATermInt)arg_arg1_head_arg1_head) == 93) {

                                                                                                                                                                                                                                                                                                                                                                                                                    if (ATgetType((ATerm)arg_arg1_head_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg1_list) == ATtrue) {
                                                                                                                                                                                                                                                                                                                                                                                                                      if (ATgetType((ATerm)arg_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_list) == ATfalse) {
                                                                                                                                                                                                                                                                                                                                                                                                                        arg_arg1_list = ATgetNext(arg_arg1_list);
                                                                                                                                                                                                                                                                                                                                                                                                                        if (ATgetType((ATerm)arg_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_list) == ATfalse) {
                                                                                                                                                                                                                                                                                                                                                                                                                          arg_arg1_head = ATgetFirst(arg_arg1_list);
                                                                                                                                                                                                                                                                                                                                                                                                                          arg_arg1_list = ATgetNext(arg_arg1_list);
                                                                                                                                                                                                                                                                                                                                                                                                                          /* checking for: appl */
                                                                                                                                                                                                                                                                                                                                                                                                                          if (ATgetType((ATerm)arg_arg1_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head) == PERR_afun0) {
                                                                                                                                                                                                                                                                                                                                                                                                                            ATerm arg_arg1_head_arg0 = ATgetArgument(arg_arg1_head, 0);
                                                                                                                                                                                                                                                                                                                                                                                                                            /* checking for: prod */
                                                                                                                                                                                                                                                                                                                                                                                                                            if (ATgetType((ATerm)arg_arg1_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0) == PERR_afun1) {
                                                                                                                                                                                                                                                                                                                                                                                                                              ATerm arg_arg1_head_arg0_arg0 = ATgetArgument(arg_arg1_head_arg0, 0);
                                                                                                                                                                                                                                                                                                                                                                                                                              if (ATgetType((ATerm)arg_arg1_head_arg0_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0) == ATfalse) {
                                                                                                                                                                                                                                                                                                                                                                                                                                ATerm arg_arg1_head_arg0_arg0_head;
                                                                                                                                                                                                                                                                                                                                                                                                                                ATermList arg_arg1_head_arg0_arg0_list = (ATermList)arg_arg1_head_arg0_arg0;
                                                                                                                                                                                                                                                                                                                                                                                                                                arg_arg1_head_arg0_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_list);
                                                                                                                                                                                                                                                                                                                                                                                                                                arg_arg1_head_arg0_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_list);
                                                                                                                                                                                                                                                                                                                                                                                                                                /* checking for: char-class */
                                                                                                                                                                                                                                                                                                                                                                                                                                if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg0_head) == PERR_afun17) {
                                                                                                                                                                                                                                                                                                                                                                                                                                  ATerm arg_arg1_head_arg0_arg0_head_arg0 = ATgetArgument(arg_arg1_head_arg0_arg0_head, 0);
                                                                                                                                                                                                                                                                                                                                                                                                                                  if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0) == ATfalse) {
                                                                                                                                                                                                                                                                                                                                                                                                                                    ATerm arg_arg1_head_arg0_arg0_head_arg0_head;
                                                                                                                                                                                                                                                                                                                                                                                                                                    ATermList arg_arg1_head_arg0_arg0_head_arg0_list = (ATermList)arg_arg1_head_arg0_arg0_head_arg0;
                                                                                                                                                                                                                                                                                                                                                                                                                                    arg_arg1_head_arg0_arg0_head_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                                                                                                                                                                                                                                                                                                                                                    arg_arg1_head_arg0_arg0_head_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                                                                                                                                                                                                                                                                                                                                                    if (ATgetInt((ATermInt)arg_arg1_head_arg0_arg0_head_arg0_head) == 41) {

                                                                                                                                                                                                                                                                                                                                                                                                                                      if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0_list) == ATtrue) {
                                                                                                                                                                                                                                                                                                                                                                                                                                        if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_list) == ATtrue) {
                                                                                                                                                                                                                                                                                                                                                                                                                                          ATerm arg_arg1_head_arg0_arg1 = ATgetArgument(arg_arg1_head_arg0, 1);
                                                                                                                                                                                                                                                                                                                                                                                                                                          /* checking for: lit */
                                                                                                                                                                                                                                                                                                                                                                                                                                          if (ATgetType((ATerm)arg_arg1_head_arg0_arg1) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg1) == PERR_afun9) {
                                                                                                                                                                                                                                                                                                                                                                                                                                            ATerm arg_arg1_head_arg0_arg1_arg0 = ATgetArgument(arg_arg1_head_arg0_arg1, 0);
                                                                                                                                                                                                                                                                                                                                                                                                                                            /* checking for: ) */
                                                                                                                                                                                                                                                                                                                                                                                                                                            if (ATgetType((ATerm)arg_arg1_head_arg0_arg1_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg1_arg0) == PERR_afun38) {
                                                                                                                                                                                                                                                                                                                                                                                                                                              ATerm arg_arg1_head_arg0_arg2 = ATgetArgument(arg_arg1_head_arg0, 2);
                                                                                                                                                                                                                                                                                                                                                                                                                                              /* checking for: no-attrs */
                                                                                                                                                                                                                                                                                                                                                                                                                                              if (ATgetType((ATerm)arg_arg1_head_arg0_arg2) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg2) == PERR_afun5) {
                                                                                                                                                                                                                                                                                                                                                                                                                                                ATerm arg_arg1_head_arg1 = ATgetArgument(arg_arg1_head, 1);
                                                                                                                                                                                                                                                                                                                                                                                                                                                if (ATgetType((ATerm)arg_arg1_head_arg1) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg1) == ATfalse) {
                                                                                                                                                                                                                                                                                                                                                                                                                                                  ATerm arg_arg1_head_arg1_head;
                                                                                                                                                                                                                                                                                                                                                                                                                                                  ATermList arg_arg1_head_arg1_list = (ATermList)arg_arg1_head_arg1;
                                                                                                                                                                                                                                                                                                                                                                                                                                                  arg_arg1_head_arg1_head = ATgetFirst(arg_arg1_head_arg1_list);
                                                                                                                                                                                                                                                                                                                                                                                                                                                  arg_arg1_head_arg1_list = ATgetNext(arg_arg1_head_arg1_list);
                                                                                                                                                                                                                                                                                                                                                                                                                                                  if (ATgetInt((ATermInt)arg_arg1_head_arg1_head) == 41) {

                                                                                                                                                                                                                                                                                                                                                                                                                                                    if (ATgetType((ATerm)arg_arg1_head_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg1_list) == ATtrue) {
                                                                                                                                                                                                                                                                                                                                                                                                                                                      if (ATgetType((ATerm)arg_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_list) == ATtrue) {
                                                                                                                                                                                                                                                                                                                                                                                                                                                        return ATtrue;
                                                                                                                                                                                                                                                                                                                                                                                                                                                      }
                                                                                                                                                                                                                                                                                                                                                                                                                                                    }
                                                                                                                                                                                                                                                                                                                                                                                                                                                  }
                                                                                                                                                                                                                                                                                                                                                                                                                                                }
                                                                                                                                                                                                                                                                                                                                                                                                                                              }
                                                                                                                                                                                                                                                                                                                                                                                                                                            }
                                                                                                                                                                                                                                                                                                                                                                                                                                          }
                                                                                                                                                                                                                                                                                                                                                                                                                                        }
                                                                                                                                                                                                                                                                                                                                                                                                                                      }
                                                                                                                                                                                                                                                                                                                                                                                                                                    }
                                                                                                                                                                                                                                                                                                                                                                                                                                  }
                                                                                                                                                                                                                                                                                                                                                                                                                                }
                                                                                                                                                                                                                                                                                                                                                                                                                              }
                                                                                                                                                                                                                                                                                                                                                                                                                            }
                                                                                                                                                                                                                                                                                                                                                                                                                          }
                                                                                                                                                                                                                                                                                                                                                                                                                        }
                                                                                                                                                                                                                                                                                                                                                                                                                      }
                                                                                                                                                                                                                                                                                                                                                                                                                    }
                                                                                                                                                                                                                                                                                                                                                                                                                  }
                                                                                                                                                                                                                                                                                                                                                                                                                }
                                                                                                                                                                                                                                                                                                                                                                                                              }
                                                                                                                                                                                                                                                                                                                                                                                                            }
                                                                                                                                                                                                                                                                                                                                                                                                          }
                                                                                                                                                                                                                                                                                                                                                                                                        }
                                                                                                                                                                                                                                                                                                                                                                                                      }
                                                                                                                                                                                                                                                                                                                                                                                                    }
                                                                                                                                                                                                                                                                                                                                                                                                  }
                                                                                                                                                                                                                                                                                                                                                                                                }
                                                                                                                                                                                                                                                                                                                                                                                              }
                                                                                                                                                                                                                                                                                                                                                                                            }
                                                                                                                                                                                                                                                                                                                                                                                          }
                                                                                                                                                                                                                                                                                                                                                                                        }
                                                                                                                                                                                                                                                                                                                                                                                      }
                                                                                                                                                                                                                                                                                                                                                                                    }
                                                                                                                                                                                                                                                                                                                                                                                  }
                                                                                                                                                                                                                                                                                                                                                                                }
                                                                                                                                                                                                                                                                                                                                                                              }
                                                                                                                                                                                                                                                                                                                                                                            }
                                                                                                                                                                                                                                                                                                                                                                          }
                                                                                                                                                                                                                                                                                                                                                                        }
                                                                                                                                                                                                                                                                                                                                                                      }
                                                                                                                                                                                                                                                                                                                                                                    }
                                                                                                                                                                                                                                                                                                                                                                  }
                                                                                                                                                                                                                                                                                                                                                                }
                                                                                                                                                                                                                                                                                                                                                              }
                                                                                                                                                                                                                                                                                                                                                            }
                                                                                                                                                                                                                                                                                                                                                          }
                                                                                                                                                                                                                                                                                                                                                        }
                                                                                                                                                                                                                                                                                                                                                      }
                                                                                                                                                                                                                                                                                                                                                    }
                                                                                                                                                                                                                                                                                                                                                  }
                                                                                                                                                                                                                                                                                                                                                }
                                                                                                                                                                                                                                                                                                                                              }
                                                                                                                                                                                                                                                                                                                                            }
                                                                                                                                                                                                                                                                                                                                          }
                                                                                                                                                                                                                                                                                                                                        }
                                                                                                                                                                                                                                                                                                                                      }
                                                                                                                                                                                                                                                                                                                                    }
                                                                                                                                                                                                                                                                                                                                  }
                                                                                                                                                                                                                                                                                                                                }
                                                                                                                                                                                                                                                                                                                              }
                                                                                                                                                                                                                                                                                                                            }
                                                                                                                                                                                                                                                                                                                          }
                                                                                                                                                                                                                                                                                                                        }
                                                                                                                                                                                                                                                                                                                      }
                                                                                                                                                                                                                                                                                                                    }
                                                                                                                                                                                                                                                                                                                  }
                                                                                                                                                                                                                                                                                                                }
                                                                                                                                                                                                                                                                                                              }
                                                                                                                                                                                                                                                                                                            }
                                                                                                                                                                                                                                                                                                          }
                                                                                                                                                                                                                                                                                                        }
                                                                                                                                                                                                                                                                                                      }
                                                                                                                                                                                                                                                                                                    }
                                                                                                                                                                                                                                                                                                  }
                                                                                                                                                                                                                                                                                                }
                                                                                                                                                                                                                                                                                              }
                                                                                                                                                                                                                                                                                            }
                                                                                                                                                                                                                                                                                          }
                                                                                                                                                                                                                                                                                        }
                                                                                                                                                                                                                                                                                      }
                                                                                                                                                                                                                                                                                    }
                                                                                                                                                                                                                                                                                  }
                                                                                                                                                                                                                                                                                }
                                                                                                                                                                                                                                                                              }
                                                                                                                                                                                                                                                                            }
                                                                                                                                                                                                                                                                          }
                                                                                                                                                                                                                                                                        }
                                                                                                                                                                                                                                                                      }
                                                                                                                                                                                                                                                                    }
                                                                                                                                                                                                                                                                  }
                                                                                                                                                                                                                                                                }
                                                                                                                                                                                                                                                              }
                                                                                                                                                                                                                                                            }
                                                                                                                                                                                                                                                          }
                                                                                                                                                                                                                                                        }
                                                                                                                                                                                                                                                      }
                                                                                                                                                                                                                                                    }
                                                                                                                                                                                                                                                  }
                                                                                                                                                                                                                                                }
                                                                                                                                                                                                                                              }
                                                                                                                                                                                                                                            }
                                                                                                                                                                                                                                          }
                                                                                                                                                                                                                                        }
                                                                                                                                                                                                                                      }
                                                                                                                                                                                                                                    }
                                                                                                                                                                                                                                  }
                                                                                                                                                                                                                                }
                                                                                                                                                                                                                              }
                                                                                                                                                                                                                            }
                                                                                                                                                                                                                          }
                                                                                                                                                                                                                        }
                                                                                                                                                                                                                      }
                                                                                                                                                                                                                    }
                                                                                                                                                                                                                  }
                                                                                                                                                                                                                }
                                                                                                                                                                                                              }
                                                                                                                                                                                                            }
                                                                                                                                                                                                          }
                                                                                                                                                                                                        }
                                                                                                                                                                                                      }
                                                                                                                                                                                                    }
                                                                                                                                                                                                  }
                                                                                                                                                                                                }
                                                                                                                                                                                              }
                                                                                                                                                                                            }
                                                                                                                                                                                          }
                                                                                                                                                                                        }
                                                                                                                                                                                      }
                                                                                                                                                                                    }
                                                                                                                                                                                  }
                                                                                                                                                                                }
                                                                                                                                                                              }
                                                                                                                                                                            }
                                                                                                                                                                          }
                                                                                                                                                                        }
                                                                                                                                                                      }
                                                                                                                                                                    }
                                                                                                                                                                  }
                                                                                                                                                                }
                                                                                                                                                              }
                                                                                                                                                            }
                                                                                                                                                          }
                                                                                                                                                        }
                                                                                                                                                      }
                                                                                                                                                    }
                                                                                                                                                  }
                                                                                                                                                }
                                                                                                                                              }
                                                                                                                                            }
                                                                                                                                          }
                                                                                                                                        }
                                                                                                                                      }
                                                                                                                                    }
                                                                                                                                  }
                                                                                                                                }
                                                                                                                              }
                                                                                                                            }
                                                                                                                          }
                                                                                                                        }
                                                                                                                      }
                                                                                                                    }
                                                                                                                  }
                                                                                                                }
                                                                                                              }
                                                                                                            }
                                                                                                          }
                                                                                                        }
                                                                                                      }
                                                                                                    }
                                                                                                  }
                                                                                                }
                                                                                              }
                                                                                            }
                                                                                          }
                                                                                        }
                                                                                      }
                                                                                    }
                                                                                  }
                                                                                }
                                                                              }
                                                                            }
                                                                          }
                                                                        }
                                                                      }
                                                                    }
                                                                  }
                                                                }
                                                              }
                                                            }
                                                          }
                                                        }
                                                      }
                                                    }
                                                  }
                                                }
                                              }
                                            }
                                          }
                                        }
                                      }
                                    }
                                  }
                                }
                              }
                            }
                          }
                        }
                      }
                    }
                  }
                }
              }
            }
          }
        }
      }
    }
  }
  return ATfalse;
}

/**
 * Assert whether a PERR_Error is a error by checking against the following ATerm pattern: appl(prod([lit("error"),cf(opt(layout)),lit("("),cf(opt(layout)),cf(sort("StrCon")),cf(opt(layout)),lit(","),cf(opt(layout)),lit("["),cf(opt(layout)),cf(iter-star-sep(sort("Subject"),lit(","))),cf(opt(layout)),lit("]"),cf(opt(layout)),lit(")")],cf(sort("Error")),attrs([term(cons("error"))])),[appl(prod([char-class([101]),char-class([114]),char-class([114]),char-class([111]),char-class([114])],lit("error"),no-attrs),[101,114,114,111,114]),<"ws-after-error"("OptLayout")>,appl(prod([char-class([40])],lit("("),no-attrs),[40]),<"ws-after-("("OptLayout")>,<"description"("StrCon")>,<"ws-after-description"("OptLayout")>,appl(prod([char-class([44])],lit(","),no-attrs),[44]),<"ws-after-,"("OptLayout")>,appl(prod([char-class([91])],lit("["),no-attrs),[91]),<"ws-after-["("OptLayout")>,appl(list(cf(iter-star-sep(sort("Subject"),lit(",")))),<"list"("Subject-list")>),<"ws-after-list"("OptLayout")>,appl(prod([char-class([93])],lit("]"),no-attrs),[93]),<"ws-after-]"("OptLayout")>,appl(prod([char-class([41])],lit(")"),no-attrs),[41])]). May not be used to assert correctness of the PERR_Error
 * \param[in] arg input PERR_Error
 * \return ATtrue if #arg corresponds to the signature of a error, or ATfalse otherwise
 */
inline ATbool PERR_isErrorError(PERR_Error arg){
  /* checking for: appl */
  if (ATgetType((ATerm)arg) == AT_APPL && ATgetAFun((ATermAppl)arg) == PERR_afun0) {
    ATerm arg_arg0 = ATgetArgument(arg, 0);
    /* checking for: prod */
    if (ATgetType((ATerm)arg_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0) == PERR_afun1) {
      ATerm arg_arg0_arg0 = ATgetArgument(arg_arg0, 0);
      if (ATgetType((ATerm)arg_arg0_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0) == ATfalse) {
        ATerm arg_arg0_arg0_head;
        ATermList arg_arg0_arg0_list = (ATermList)arg_arg0_arg0;
        arg_arg0_arg0_head = ATgetFirst(arg_arg0_arg0_list);
        arg_arg0_arg0_list = ATgetNext(arg_arg0_arg0_list);
        /* checking for: lit */
        if (ATgetType((ATerm)arg_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head) == PERR_afun9) {
          ATerm arg_arg0_arg0_head_arg0 = ATgetArgument(arg_arg0_arg0_head, 0);
          /* checking for: error */
          if (ATgetType((ATerm)arg_arg0_arg0_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0) == PERR_afun50) {
            if (ATgetType((ATerm)arg_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0_list) == ATfalse) {
              arg_arg0_arg0_head = ATgetFirst(arg_arg0_arg0_list);
              arg_arg0_arg0_list = ATgetNext(arg_arg0_arg0_list);
              /* checking for: cf */
              if (ATgetType((ATerm)arg_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head) == PERR_afun2) {
                ATerm arg_arg0_arg0_head_arg0 = ATgetArgument(arg_arg0_arg0_head, 0);
                /* checking for: opt */
                if (ATgetType((ATerm)arg_arg0_arg0_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0) == PERR_afun3) {
                  ATerm arg_arg0_arg0_head_arg0_arg0 = ATgetArgument(arg_arg0_arg0_head_arg0, 0);
                  /* checking for: layout */
                  if (ATgetType((ATerm)arg_arg0_arg0_head_arg0_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0_arg0) == PERR_afun4) {
                    if (ATgetType((ATerm)arg_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0_list) == ATfalse) {
                      arg_arg0_arg0_head = ATgetFirst(arg_arg0_arg0_list);
                      arg_arg0_arg0_list = ATgetNext(arg_arg0_arg0_list);
                      /* checking for: lit */
                      if (ATgetType((ATerm)arg_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head) == PERR_afun9) {
                        ATerm arg_arg0_arg0_head_arg0 = ATgetArgument(arg_arg0_arg0_head, 0);
                        /* checking for: ( */
                        if (ATgetType((ATerm)arg_arg0_arg0_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0) == PERR_afun39) {
                          if (ATgetType((ATerm)arg_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0_list) == ATfalse) {
                            arg_arg0_arg0_head = ATgetFirst(arg_arg0_arg0_list);
                            arg_arg0_arg0_list = ATgetNext(arg_arg0_arg0_list);
                            /* checking for: cf */
                            if (ATgetType((ATerm)arg_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head) == PERR_afun2) {
                              ATerm arg_arg0_arg0_head_arg0 = ATgetArgument(arg_arg0_arg0_head, 0);
                              /* checking for: opt */
                              if (ATgetType((ATerm)arg_arg0_arg0_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0) == PERR_afun3) {
                                ATerm arg_arg0_arg0_head_arg0_arg0 = ATgetArgument(arg_arg0_arg0_head_arg0, 0);
                                /* checking for: layout */
                                if (ATgetType((ATerm)arg_arg0_arg0_head_arg0_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0_arg0) == PERR_afun4) {
                                  if (ATgetType((ATerm)arg_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0_list) == ATfalse) {
                                    arg_arg0_arg0_head = ATgetFirst(arg_arg0_arg0_list);
                                    arg_arg0_arg0_list = ATgetNext(arg_arg0_arg0_list);
                                    /* checking for: cf */
                                    if (ATgetType((ATerm)arg_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head) == PERR_afun2) {
                                      ATerm arg_arg0_arg0_head_arg0 = ATgetArgument(arg_arg0_arg0_head, 0);
                                      /* checking for: sort */
                                      if (ATgetType((ATerm)arg_arg0_arg0_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0) == PERR_afun11) {
                                        ATerm arg_arg0_arg0_head_arg0_arg0 = ATgetArgument(arg_arg0_arg0_head_arg0, 0);
                                        /* checking for: StrCon */
                                        if (ATgetType((ATerm)arg_arg0_arg0_head_arg0_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0_arg0) == PERR_afun29) {
                                          if (ATgetType((ATerm)arg_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0_list) == ATfalse) {
                                            arg_arg0_arg0_head = ATgetFirst(arg_arg0_arg0_list);
                                            arg_arg0_arg0_list = ATgetNext(arg_arg0_arg0_list);
                                            /* checking for: cf */
                                            if (ATgetType((ATerm)arg_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head) == PERR_afun2) {
                                              ATerm arg_arg0_arg0_head_arg0 = ATgetArgument(arg_arg0_arg0_head, 0);
                                              /* checking for: opt */
                                              if (ATgetType((ATerm)arg_arg0_arg0_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0) == PERR_afun3) {
                                                ATerm arg_arg0_arg0_head_arg0_arg0 = ATgetArgument(arg_arg0_arg0_head_arg0, 0);
                                                /* checking for: layout */
                                                if (ATgetType((ATerm)arg_arg0_arg0_head_arg0_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0_arg0) == PERR_afun4) {
                                                  if (ATgetType((ATerm)arg_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0_list) == ATfalse) {
                                                    arg_arg0_arg0_head = ATgetFirst(arg_arg0_arg0_list);
                                                    arg_arg0_arg0_list = ATgetNext(arg_arg0_arg0_list);
                                                    /* checking for: lit */
                                                    if (ATgetType((ATerm)arg_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head) == PERR_afun9) {
                                                      ATerm arg_arg0_arg0_head_arg0 = ATgetArgument(arg_arg0_arg0_head, 0);
                                                      /* checking for: , */
                                                      if (ATgetType((ATerm)arg_arg0_arg0_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0) == PERR_afun43) {
                                                        if (ATgetType((ATerm)arg_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0_list) == ATfalse) {
                                                          arg_arg0_arg0_head = ATgetFirst(arg_arg0_arg0_list);
                                                          arg_arg0_arg0_list = ATgetNext(arg_arg0_arg0_list);
                                                          /* checking for: cf */
                                                          if (ATgetType((ATerm)arg_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head) == PERR_afun2) {
                                                            ATerm arg_arg0_arg0_head_arg0 = ATgetArgument(arg_arg0_arg0_head, 0);
                                                            /* checking for: opt */
                                                            if (ATgetType((ATerm)arg_arg0_arg0_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0) == PERR_afun3) {
                                                              ATerm arg_arg0_arg0_head_arg0_arg0 = ATgetArgument(arg_arg0_arg0_head_arg0, 0);
                                                              /* checking for: layout */
                                                              if (ATgetType((ATerm)arg_arg0_arg0_head_arg0_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0_arg0) == PERR_afun4) {
                                                                if (ATgetType((ATerm)arg_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0_list) == ATfalse) {
                                                                  arg_arg0_arg0_head = ATgetFirst(arg_arg0_arg0_list);
                                                                  arg_arg0_arg0_list = ATgetNext(arg_arg0_arg0_list);
                                                                  /* checking for: lit */
                                                                  if (ATgetType((ATerm)arg_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head) == PERR_afun9) {
                                                                    ATerm arg_arg0_arg0_head_arg0 = ATgetArgument(arg_arg0_arg0_head, 0);
                                                                    /* checking for: [ */
                                                                    if (ATgetType((ATerm)arg_arg0_arg0_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0) == PERR_afun47) {
                                                                      if (ATgetType((ATerm)arg_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0_list) == ATfalse) {
                                                                        arg_arg0_arg0_head = ATgetFirst(arg_arg0_arg0_list);
                                                                        arg_arg0_arg0_list = ATgetNext(arg_arg0_arg0_list);
                                                                        /* checking for: cf */
                                                                        if (ATgetType((ATerm)arg_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head) == PERR_afun2) {
                                                                          ATerm arg_arg0_arg0_head_arg0 = ATgetArgument(arg_arg0_arg0_head, 0);
                                                                          /* checking for: opt */
                                                                          if (ATgetType((ATerm)arg_arg0_arg0_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0) == PERR_afun3) {
                                                                            ATerm arg_arg0_arg0_head_arg0_arg0 = ATgetArgument(arg_arg0_arg0_head_arg0, 0);
                                                                            /* checking for: layout */
                                                                            if (ATgetType((ATerm)arg_arg0_arg0_head_arg0_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0_arg0) == PERR_afun4) {
                                                                              if (ATgetType((ATerm)arg_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0_list) == ATfalse) {
                                                                                arg_arg0_arg0_head = ATgetFirst(arg_arg0_arg0_list);
                                                                                arg_arg0_arg0_list = ATgetNext(arg_arg0_arg0_list);
                                                                                /* checking for: cf */
                                                                                if (ATgetType((ATerm)arg_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head) == PERR_afun2) {
                                                                                  ATerm arg_arg0_arg0_head_arg0 = ATgetArgument(arg_arg0_arg0_head, 0);
                                                                                  /* checking for: iter-star-sep */
                                                                                  if (ATgetType((ATerm)arg_arg0_arg0_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0) == PERR_afun46) {
                                                                                    ATerm arg_arg0_arg0_head_arg0_arg0 = ATgetArgument(arg_arg0_arg0_head_arg0, 0);
                                                                                    /* checking for: sort */
                                                                                    if (ATgetType((ATerm)arg_arg0_arg0_head_arg0_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0_arg0) == PERR_afun11) {
                                                                                      ATerm arg_arg0_arg0_head_arg0_arg0_arg0 = ATgetArgument(arg_arg0_arg0_head_arg0_arg0, 0);
                                                                                      /* checking for: Subject */
                                                                                      if (ATgetType((ATerm)arg_arg0_arg0_head_arg0_arg0_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0_arg0_arg0) == PERR_afun41) {
                                                                                        ATerm arg_arg0_arg0_head_arg0_arg1 = ATgetArgument(arg_arg0_arg0_head_arg0, 1);
                                                                                        /* checking for: lit */
                                                                                        if (ATgetType((ATerm)arg_arg0_arg0_head_arg0_arg1) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0_arg1) == PERR_afun9) {
                                                                                          ATerm arg_arg0_arg0_head_arg0_arg1_arg0 = ATgetArgument(arg_arg0_arg0_head_arg0_arg1, 0);
                                                                                          /* checking for: , */
                                                                                          if (ATgetType((ATerm)arg_arg0_arg0_head_arg0_arg1_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0_arg1_arg0) == PERR_afun43) {
                                                                                            if (ATgetType((ATerm)arg_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0_list) == ATfalse) {
                                                                                              arg_arg0_arg0_head = ATgetFirst(arg_arg0_arg0_list);
                                                                                              arg_arg0_arg0_list = ATgetNext(arg_arg0_arg0_list);
                                                                                              /* checking for: cf */
                                                                                              if (ATgetType((ATerm)arg_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head) == PERR_afun2) {
                                                                                                ATerm arg_arg0_arg0_head_arg0 = ATgetArgument(arg_arg0_arg0_head, 0);
                                                                                                /* checking for: opt */
                                                                                                if (ATgetType((ATerm)arg_arg0_arg0_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0) == PERR_afun3) {
                                                                                                  ATerm arg_arg0_arg0_head_arg0_arg0 = ATgetArgument(arg_arg0_arg0_head_arg0, 0);
                                                                                                  /* checking for: layout */
                                                                                                  if (ATgetType((ATerm)arg_arg0_arg0_head_arg0_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0_arg0) == PERR_afun4) {
                                                                                                    if (ATgetType((ATerm)arg_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0_list) == ATfalse) {
                                                                                                      arg_arg0_arg0_head = ATgetFirst(arg_arg0_arg0_list);
                                                                                                      arg_arg0_arg0_list = ATgetNext(arg_arg0_arg0_list);
                                                                                                      /* checking for: lit */
                                                                                                      if (ATgetType((ATerm)arg_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head) == PERR_afun9) {
                                                                                                        ATerm arg_arg0_arg0_head_arg0 = ATgetArgument(arg_arg0_arg0_head, 0);
                                                                                                        /* checking for: ] */
                                                                                                        if (ATgetType((ATerm)arg_arg0_arg0_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0) == PERR_afun45) {
                                                                                                          if (ATgetType((ATerm)arg_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0_list) == ATfalse) {
                                                                                                            arg_arg0_arg0_head = ATgetFirst(arg_arg0_arg0_list);
                                                                                                            arg_arg0_arg0_list = ATgetNext(arg_arg0_arg0_list);
                                                                                                            /* checking for: cf */
                                                                                                            if (ATgetType((ATerm)arg_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head) == PERR_afun2) {
                                                                                                              ATerm arg_arg0_arg0_head_arg0 = ATgetArgument(arg_arg0_arg0_head, 0);
                                                                                                              /* checking for: opt */
                                                                                                              if (ATgetType((ATerm)arg_arg0_arg0_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0) == PERR_afun3) {
                                                                                                                ATerm arg_arg0_arg0_head_arg0_arg0 = ATgetArgument(arg_arg0_arg0_head_arg0, 0);
                                                                                                                /* checking for: layout */
                                                                                                                if (ATgetType((ATerm)arg_arg0_arg0_head_arg0_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0_arg0) == PERR_afun4) {
                                                                                                                  if (ATgetType((ATerm)arg_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0_list) == ATfalse) {
                                                                                                                    arg_arg0_arg0_head = ATgetFirst(arg_arg0_arg0_list);
                                                                                                                    arg_arg0_arg0_list = ATgetNext(arg_arg0_arg0_list);
                                                                                                                    /* checking for: lit */
                                                                                                                    if (ATgetType((ATerm)arg_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head) == PERR_afun9) {
                                                                                                                      ATerm arg_arg0_arg0_head_arg0 = ATgetArgument(arg_arg0_arg0_head, 0);
                                                                                                                      /* checking for: ) */
                                                                                                                      if (ATgetType((ATerm)arg_arg0_arg0_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0) == PERR_afun38) {
                                                                                                                        if (ATgetType((ATerm)arg_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0_list) == ATtrue) {
                                                                                                                          ATerm arg_arg0_arg1 = ATgetArgument(arg_arg0, 1);
                                                                                                                          /* checking for: cf */
                                                                                                                          if (ATgetType((ATerm)arg_arg0_arg1) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg1) == PERR_afun2) {
                                                                                                                            ATerm arg_arg0_arg1_arg0 = ATgetArgument(arg_arg0_arg1, 0);
                                                                                                                            /* checking for: sort */
                                                                                                                            if (ATgetType((ATerm)arg_arg0_arg1_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg1_arg0) == PERR_afun11) {
                                                                                                                              ATerm arg_arg0_arg1_arg0_arg0 = ATgetArgument(arg_arg0_arg1_arg0, 0);
                                                                                                                              /* checking for: Error */
                                                                                                                              if (ATgetType((ATerm)arg_arg0_arg1_arg0_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg1_arg0_arg0) == PERR_afun37) {
                                                                                                                                ATerm arg_arg0_arg2 = ATgetArgument(arg_arg0, 2);
                                                                                                                                /* checking for: attrs */
                                                                                                                                if (ATgetType((ATerm)arg_arg0_arg2) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg2) == PERR_afun13) {
                                                                                                                                  ATerm arg_arg0_arg2_arg0 = ATgetArgument(arg_arg0_arg2, 0);
                                                                                                                                  if (ATgetType((ATerm)arg_arg0_arg2_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg2_arg0) == ATfalse) {
                                                                                                                                    ATerm arg_arg0_arg2_arg0_head;
                                                                                                                                    ATermList arg_arg0_arg2_arg0_list = (ATermList)arg_arg0_arg2_arg0;
                                                                                                                                    arg_arg0_arg2_arg0_head = ATgetFirst(arg_arg0_arg2_arg0_list);
                                                                                                                                    arg_arg0_arg2_arg0_list = ATgetNext(arg_arg0_arg2_arg0_list);
                                                                                                                                    /* checking for: term */
                                                                                                                                    if (ATgetType((ATerm)arg_arg0_arg2_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg2_arg0_head) == PERR_afun14) {
                                                                                                                                      ATerm arg_arg0_arg2_arg0_head_arg0 = ATgetArgument(arg_arg0_arg2_arg0_head, 0);
                                                                                                                                      /* checking for: cons */
                                                                                                                                      if (ATgetType((ATerm)arg_arg0_arg2_arg0_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg2_arg0_head_arg0) == PERR_afun15) {
                                                                                                                                        ATerm arg_arg0_arg2_arg0_head_arg0_arg0 = ATgetArgument(arg_arg0_arg2_arg0_head_arg0, 0);
                                                                                                                                        /* checking for: error */
                                                                                                                                        if (ATgetType((ATerm)arg_arg0_arg2_arg0_head_arg0_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg2_arg0_head_arg0_arg0) == PERR_afun50) {
                                                                                                                                          if (ATgetType((ATerm)arg_arg0_arg2_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg2_arg0_list) == ATtrue) {
                                                                                                                                            ATerm arg_arg1 = ATgetArgument(arg, 1);
                                                                                                                                            if (ATgetType((ATerm)arg_arg1) == AT_LIST && ATisEmpty((ATermList)arg_arg1) == ATfalse) {
                                                                                                                                              ATerm arg_arg1_head;
                                                                                                                                              ATermList arg_arg1_list = (ATermList)arg_arg1;
                                                                                                                                              arg_arg1_head = ATgetFirst(arg_arg1_list);
                                                                                                                                              arg_arg1_list = ATgetNext(arg_arg1_list);
                                                                                                                                              /* checking for: appl */
                                                                                                                                              if (ATgetType((ATerm)arg_arg1_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head) == PERR_afun0) {
                                                                                                                                                ATerm arg_arg1_head_arg0 = ATgetArgument(arg_arg1_head, 0);
                                                                                                                                                /* checking for: prod */
                                                                                                                                                if (ATgetType((ATerm)arg_arg1_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0) == PERR_afun1) {
                                                                                                                                                  ATerm arg_arg1_head_arg0_arg0 = ATgetArgument(arg_arg1_head_arg0, 0);
                                                                                                                                                  if (ATgetType((ATerm)arg_arg1_head_arg0_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0) == ATfalse) {
                                                                                                                                                    ATerm arg_arg1_head_arg0_arg0_head;
                                                                                                                                                    ATermList arg_arg1_head_arg0_arg0_list = (ATermList)arg_arg1_head_arg0_arg0;
                                                                                                                                                    arg_arg1_head_arg0_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_list);
                                                                                                                                                    arg_arg1_head_arg0_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_list);
                                                                                                                                                    /* checking for: char-class */
                                                                                                                                                    if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg0_head) == PERR_afun17) {
                                                                                                                                                      ATerm arg_arg1_head_arg0_arg0_head_arg0 = ATgetArgument(arg_arg1_head_arg0_arg0_head, 0);
                                                                                                                                                      if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0) == ATfalse) {
                                                                                                                                                        ATerm arg_arg1_head_arg0_arg0_head_arg0_head;
                                                                                                                                                        ATermList arg_arg1_head_arg0_arg0_head_arg0_list = (ATermList)arg_arg1_head_arg0_arg0_head_arg0;
                                                                                                                                                        arg_arg1_head_arg0_arg0_head_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                                                                        arg_arg1_head_arg0_arg0_head_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                                                                        if (ATgetInt((ATermInt)arg_arg1_head_arg0_arg0_head_arg0_head) == 101) {

                                                                                                                                                          if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0_list) == ATtrue) {
                                                                                                                                                            if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_list) == ATfalse) {
                                                                                                                                                              arg_arg1_head_arg0_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_list);
                                                                                                                                                              arg_arg1_head_arg0_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_list);
                                                                                                                                                              /* checking for: char-class */
                                                                                                                                                              if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg0_head) == PERR_afun17) {
                                                                                                                                                                ATerm arg_arg1_head_arg0_arg0_head_arg0 = ATgetArgument(arg_arg1_head_arg0_arg0_head, 0);
                                                                                                                                                                if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0) == ATfalse) {
                                                                                                                                                                  ATerm arg_arg1_head_arg0_arg0_head_arg0_head;
                                                                                                                                                                  ATermList arg_arg1_head_arg0_arg0_head_arg0_list = (ATermList)arg_arg1_head_arg0_arg0_head_arg0;
                                                                                                                                                                  arg_arg1_head_arg0_arg0_head_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                                                                                  arg_arg1_head_arg0_arg0_head_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                                                                                  if (ATgetInt((ATermInt)arg_arg1_head_arg0_arg0_head_arg0_head) == 114) {

                                                                                                                                                                    if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0_list) == ATtrue) {
                                                                                                                                                                      if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_list) == ATfalse) {
                                                                                                                                                                        arg_arg1_head_arg0_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_list);
                                                                                                                                                                        arg_arg1_head_arg0_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_list);
                                                                                                                                                                        /* checking for: char-class */
                                                                                                                                                                        if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg0_head) == PERR_afun17) {
                                                                                                                                                                          ATerm arg_arg1_head_arg0_arg0_head_arg0 = ATgetArgument(arg_arg1_head_arg0_arg0_head, 0);
                                                                                                                                                                          if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0) == ATfalse) {
                                                                                                                                                                            ATerm arg_arg1_head_arg0_arg0_head_arg0_head;
                                                                                                                                                                            ATermList arg_arg1_head_arg0_arg0_head_arg0_list = (ATermList)arg_arg1_head_arg0_arg0_head_arg0;
                                                                                                                                                                            arg_arg1_head_arg0_arg0_head_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                                                                                            arg_arg1_head_arg0_arg0_head_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                                                                                            if (ATgetInt((ATermInt)arg_arg1_head_arg0_arg0_head_arg0_head) == 114) {

                                                                                                                                                                              if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0_list) == ATtrue) {
                                                                                                                                                                                if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_list) == ATfalse) {
                                                                                                                                                                                  arg_arg1_head_arg0_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_list);
                                                                                                                                                                                  arg_arg1_head_arg0_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_list);
                                                                                                                                                                                  /* checking for: char-class */
                                                                                                                                                                                  if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg0_head) == PERR_afun17) {
                                                                                                                                                                                    ATerm arg_arg1_head_arg0_arg0_head_arg0 = ATgetArgument(arg_arg1_head_arg0_arg0_head, 0);
                                                                                                                                                                                    if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0) == ATfalse) {
                                                                                                                                                                                      ATerm arg_arg1_head_arg0_arg0_head_arg0_head;
                                                                                                                                                                                      ATermList arg_arg1_head_arg0_arg0_head_arg0_list = (ATermList)arg_arg1_head_arg0_arg0_head_arg0;
                                                                                                                                                                                      arg_arg1_head_arg0_arg0_head_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                                                                                                      arg_arg1_head_arg0_arg0_head_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                                                                                                      if (ATgetInt((ATermInt)arg_arg1_head_arg0_arg0_head_arg0_head) == 111) {

                                                                                                                                                                                        if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0_list) == ATtrue) {
                                                                                                                                                                                          if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_list) == ATfalse) {
                                                                                                                                                                                            arg_arg1_head_arg0_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_list);
                                                                                                                                                                                            arg_arg1_head_arg0_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_list);
                                                                                                                                                                                            /* checking for: char-class */
                                                                                                                                                                                            if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg0_head) == PERR_afun17) {
                                                                                                                                                                                              ATerm arg_arg1_head_arg0_arg0_head_arg0 = ATgetArgument(arg_arg1_head_arg0_arg0_head, 0);
                                                                                                                                                                                              if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0) == ATfalse) {
                                                                                                                                                                                                ATerm arg_arg1_head_arg0_arg0_head_arg0_head;
                                                                                                                                                                                                ATermList arg_arg1_head_arg0_arg0_head_arg0_list = (ATermList)arg_arg1_head_arg0_arg0_head_arg0;
                                                                                                                                                                                                arg_arg1_head_arg0_arg0_head_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                                                                                                                arg_arg1_head_arg0_arg0_head_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                                                                                                                if (ATgetInt((ATermInt)arg_arg1_head_arg0_arg0_head_arg0_head) == 114) {

                                                                                                                                                                                                  if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0_list) == ATtrue) {
                                                                                                                                                                                                    if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_list) == ATtrue) {
                                                                                                                                                                                                      ATerm arg_arg1_head_arg0_arg1 = ATgetArgument(arg_arg1_head_arg0, 1);
                                                                                                                                                                                                      /* checking for: lit */
                                                                                                                                                                                                      if (ATgetType((ATerm)arg_arg1_head_arg0_arg1) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg1) == PERR_afun9) {
                                                                                                                                                                                                        ATerm arg_arg1_head_arg0_arg1_arg0 = ATgetArgument(arg_arg1_head_arg0_arg1, 0);
                                                                                                                                                                                                        /* checking for: error */
                                                                                                                                                                                                        if (ATgetType((ATerm)arg_arg1_head_arg0_arg1_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg1_arg0) == PERR_afun50) {
                                                                                                                                                                                                          ATerm arg_arg1_head_arg0_arg2 = ATgetArgument(arg_arg1_head_arg0, 2);
                                                                                                                                                                                                          /* checking for: no-attrs */
                                                                                                                                                                                                          if (ATgetType((ATerm)arg_arg1_head_arg0_arg2) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg2) == PERR_afun5) {
                                                                                                                                                                                                            ATerm arg_arg1_head_arg1 = ATgetArgument(arg_arg1_head, 1);
                                                                                                                                                                                                            if (ATgetType((ATerm)arg_arg1_head_arg1) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg1) == ATfalse) {
                                                                                                                                                                                                              ATerm arg_arg1_head_arg1_head;
                                                                                                                                                                                                              ATermList arg_arg1_head_arg1_list = (ATermList)arg_arg1_head_arg1;
                                                                                                                                                                                                              arg_arg1_head_arg1_head = ATgetFirst(arg_arg1_head_arg1_list);
                                                                                                                                                                                                              arg_arg1_head_arg1_list = ATgetNext(arg_arg1_head_arg1_list);
                                                                                                                                                                                                              if (ATgetInt((ATermInt)arg_arg1_head_arg1_head) == 101) {

                                                                                                                                                                                                                if (ATgetType((ATerm)arg_arg1_head_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg1_list) == ATfalse) {
                                                                                                                                                                                                                  arg_arg1_head_arg1_head = ATgetFirst(arg_arg1_head_arg1_list);
                                                                                                                                                                                                                  arg_arg1_head_arg1_list = ATgetNext(arg_arg1_head_arg1_list);
                                                                                                                                                                                                                  if (ATgetInt((ATermInt)arg_arg1_head_arg1_head) == 114) {

                                                                                                                                                                                                                    if (ATgetType((ATerm)arg_arg1_head_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg1_list) == ATfalse) {
                                                                                                                                                                                                                      arg_arg1_head_arg1_head = ATgetFirst(arg_arg1_head_arg1_list);
                                                                                                                                                                                                                      arg_arg1_head_arg1_list = ATgetNext(arg_arg1_head_arg1_list);
                                                                                                                                                                                                                      if (ATgetInt((ATermInt)arg_arg1_head_arg1_head) == 114) {

                                                                                                                                                                                                                        if (ATgetType((ATerm)arg_arg1_head_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg1_list) == ATfalse) {
                                                                                                                                                                                                                          arg_arg1_head_arg1_head = ATgetFirst(arg_arg1_head_arg1_list);
                                                                                                                                                                                                                          arg_arg1_head_arg1_list = ATgetNext(arg_arg1_head_arg1_list);
                                                                                                                                                                                                                          if (ATgetInt((ATermInt)arg_arg1_head_arg1_head) == 111) {

                                                                                                                                                                                                                            if (ATgetType((ATerm)arg_arg1_head_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg1_list) == ATfalse) {
                                                                                                                                                                                                                              arg_arg1_head_arg1_head = ATgetFirst(arg_arg1_head_arg1_list);
                                                                                                                                                                                                                              arg_arg1_head_arg1_list = ATgetNext(arg_arg1_head_arg1_list);
                                                                                                                                                                                                                              if (ATgetInt((ATermInt)arg_arg1_head_arg1_head) == 114) {

                                                                                                                                                                                                                                if (ATgetType((ATerm)arg_arg1_head_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg1_list) == ATtrue) {
                                                                                                                                                                                                                                  if (ATgetType((ATerm)arg_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_list) == ATfalse) {
                                                                                                                                                                                                                                    arg_arg1_list = ATgetNext(arg_arg1_list);
                                                                                                                                                                                                                                    if (ATgetType((ATerm)arg_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_list) == ATfalse) {
                                                                                                                                                                                                                                      arg_arg1_head = ATgetFirst(arg_arg1_list);
                                                                                                                                                                                                                                      arg_arg1_list = ATgetNext(arg_arg1_list);
                                                                                                                                                                                                                                      /* checking for: appl */
                                                                                                                                                                                                                                      if (ATgetType((ATerm)arg_arg1_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head) == PERR_afun0) {
                                                                                                                                                                                                                                        ATerm arg_arg1_head_arg0 = ATgetArgument(arg_arg1_head, 0);
                                                                                                                                                                                                                                        /* checking for: prod */
                                                                                                                                                                                                                                        if (ATgetType((ATerm)arg_arg1_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0) == PERR_afun1) {
                                                                                                                                                                                                                                          ATerm arg_arg1_head_arg0_arg0 = ATgetArgument(arg_arg1_head_arg0, 0);
                                                                                                                                                                                                                                          if (ATgetType((ATerm)arg_arg1_head_arg0_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0) == ATfalse) {
                                                                                                                                                                                                                                            ATerm arg_arg1_head_arg0_arg0_head;
                                                                                                                                                                                                                                            ATermList arg_arg1_head_arg0_arg0_list = (ATermList)arg_arg1_head_arg0_arg0;
                                                                                                                                                                                                                                            arg_arg1_head_arg0_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_list);
                                                                                                                                                                                                                                            arg_arg1_head_arg0_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_list);
                                                                                                                                                                                                                                            /* checking for: char-class */
                                                                                                                                                                                                                                            if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg0_head) == PERR_afun17) {
                                                                                                                                                                                                                                              ATerm arg_arg1_head_arg0_arg0_head_arg0 = ATgetArgument(arg_arg1_head_arg0_arg0_head, 0);
                                                                                                                                                                                                                                              if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0) == ATfalse) {
                                                                                                                                                                                                                                                ATerm arg_arg1_head_arg0_arg0_head_arg0_head;
                                                                                                                                                                                                                                                ATermList arg_arg1_head_arg0_arg0_head_arg0_list = (ATermList)arg_arg1_head_arg0_arg0_head_arg0;
                                                                                                                                                                                                                                                arg_arg1_head_arg0_arg0_head_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                                                                                                                                                                arg_arg1_head_arg0_arg0_head_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                                                                                                                                                                if (ATgetInt((ATermInt)arg_arg1_head_arg0_arg0_head_arg0_head) == 40) {

                                                                                                                                                                                                                                                  if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0_list) == ATtrue) {
                                                                                                                                                                                                                                                    if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_list) == ATtrue) {
                                                                                                                                                                                                                                                      ATerm arg_arg1_head_arg0_arg1 = ATgetArgument(arg_arg1_head_arg0, 1);
                                                                                                                                                                                                                                                      /* checking for: lit */
                                                                                                                                                                                                                                                      if (ATgetType((ATerm)arg_arg1_head_arg0_arg1) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg1) == PERR_afun9) {
                                                                                                                                                                                                                                                        ATerm arg_arg1_head_arg0_arg1_arg0 = ATgetArgument(arg_arg1_head_arg0_arg1, 0);
                                                                                                                                                                                                                                                        /* checking for: ( */
                                                                                                                                                                                                                                                        if (ATgetType((ATerm)arg_arg1_head_arg0_arg1_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg1_arg0) == PERR_afun39) {
                                                                                                                                                                                                                                                          ATerm arg_arg1_head_arg0_arg2 = ATgetArgument(arg_arg1_head_arg0, 2);
                                                                                                                                                                                                                                                          /* checking for: no-attrs */
                                                                                                                                                                                                                                                          if (ATgetType((ATerm)arg_arg1_head_arg0_arg2) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg2) == PERR_afun5) {
                                                                                                                                                                                                                                                            ATerm arg_arg1_head_arg1 = ATgetArgument(arg_arg1_head, 1);
                                                                                                                                                                                                                                                            if (ATgetType((ATerm)arg_arg1_head_arg1) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg1) == ATfalse) {
                                                                                                                                                                                                                                                              ATerm arg_arg1_head_arg1_head;
                                                                                                                                                                                                                                                              ATermList arg_arg1_head_arg1_list = (ATermList)arg_arg1_head_arg1;
                                                                                                                                                                                                                                                              arg_arg1_head_arg1_head = ATgetFirst(arg_arg1_head_arg1_list);
                                                                                                                                                                                                                                                              arg_arg1_head_arg1_list = ATgetNext(arg_arg1_head_arg1_list);
                                                                                                                                                                                                                                                              if (ATgetInt((ATermInt)arg_arg1_head_arg1_head) == 40) {

                                                                                                                                                                                                                                                                if (ATgetType((ATerm)arg_arg1_head_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg1_list) == ATtrue) {
                                                                                                                                                                                                                                                                  if (ATgetType((ATerm)arg_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_list) == ATfalse) {
                                                                                                                                                                                                                                                                    arg_arg1_list = ATgetNext(arg_arg1_list);
                                                                                                                                                                                                                                                                    if (ATgetType((ATerm)arg_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_list) == ATfalse) {
                                                                                                                                                                                                                                                                      arg_arg1_list = ATgetNext(arg_arg1_list);
                                                                                                                                                                                                                                                                      if (ATgetType((ATerm)arg_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_list) == ATfalse) {
                                                                                                                                                                                                                                                                        arg_arg1_list = ATgetNext(arg_arg1_list);
                                                                                                                                                                                                                                                                        if (ATgetType((ATerm)arg_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_list) == ATfalse) {
                                                                                                                                                                                                                                                                          arg_arg1_head = ATgetFirst(arg_arg1_list);
                                                                                                                                                                                                                                                                          arg_arg1_list = ATgetNext(arg_arg1_list);
                                                                                                                                                                                                                                                                          /* checking for: appl */
                                                                                                                                                                                                                                                                          if (ATgetType((ATerm)arg_arg1_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head) == PERR_afun0) {
                                                                                                                                                                                                                                                                            ATerm arg_arg1_head_arg0 = ATgetArgument(arg_arg1_head, 0);
                                                                                                                                                                                                                                                                            /* checking for: prod */
                                                                                                                                                                                                                                                                            if (ATgetType((ATerm)arg_arg1_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0) == PERR_afun1) {
                                                                                                                                                                                                                                                                              ATerm arg_arg1_head_arg0_arg0 = ATgetArgument(arg_arg1_head_arg0, 0);
                                                                                                                                                                                                                                                                              if (ATgetType((ATerm)arg_arg1_head_arg0_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0) == ATfalse) {
                                                                                                                                                                                                                                                                                ATerm arg_arg1_head_arg0_arg0_head;
                                                                                                                                                                                                                                                                                ATermList arg_arg1_head_arg0_arg0_list = (ATermList)arg_arg1_head_arg0_arg0;
                                                                                                                                                                                                                                                                                arg_arg1_head_arg0_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_list);
                                                                                                                                                                                                                                                                                arg_arg1_head_arg0_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_list);
                                                                                                                                                                                                                                                                                /* checking for: char-class */
                                                                                                                                                                                                                                                                                if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg0_head) == PERR_afun17) {
                                                                                                                                                                                                                                                                                  ATerm arg_arg1_head_arg0_arg0_head_arg0 = ATgetArgument(arg_arg1_head_arg0_arg0_head, 0);
                                                                                                                                                                                                                                                                                  if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0) == ATfalse) {
                                                                                                                                                                                                                                                                                    ATerm arg_arg1_head_arg0_arg0_head_arg0_head;
                                                                                                                                                                                                                                                                                    ATermList arg_arg1_head_arg0_arg0_head_arg0_list = (ATermList)arg_arg1_head_arg0_arg0_head_arg0;
                                                                                                                                                                                                                                                                                    arg_arg1_head_arg0_arg0_head_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                                                                                                                                                                                                    arg_arg1_head_arg0_arg0_head_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                                                                                                                                                                                                    if (ATgetInt((ATermInt)arg_arg1_head_arg0_arg0_head_arg0_head) == 44) {

                                                                                                                                                                                                                                                                                      if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0_list) == ATtrue) {
                                                                                                                                                                                                                                                                                        if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_list) == ATtrue) {
                                                                                                                                                                                                                                                                                          ATerm arg_arg1_head_arg0_arg1 = ATgetArgument(arg_arg1_head_arg0, 1);
                                                                                                                                                                                                                                                                                          /* checking for: lit */
                                                                                                                                                                                                                                                                                          if (ATgetType((ATerm)arg_arg1_head_arg0_arg1) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg1) == PERR_afun9) {
                                                                                                                                                                                                                                                                                            ATerm arg_arg1_head_arg0_arg1_arg0 = ATgetArgument(arg_arg1_head_arg0_arg1, 0);
                                                                                                                                                                                                                                                                                            /* checking for: , */
                                                                                                                                                                                                                                                                                            if (ATgetType((ATerm)arg_arg1_head_arg0_arg1_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg1_arg0) == PERR_afun43) {
                                                                                                                                                                                                                                                                                              ATerm arg_arg1_head_arg0_arg2 = ATgetArgument(arg_arg1_head_arg0, 2);
                                                                                                                                                                                                                                                                                              /* checking for: no-attrs */
                                                                                                                                                                                                                                                                                              if (ATgetType((ATerm)arg_arg1_head_arg0_arg2) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg2) == PERR_afun5) {
                                                                                                                                                                                                                                                                                                ATerm arg_arg1_head_arg1 = ATgetArgument(arg_arg1_head, 1);
                                                                                                                                                                                                                                                                                                if (ATgetType((ATerm)arg_arg1_head_arg1) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg1) == ATfalse) {
                                                                                                                                                                                                                                                                                                  ATerm arg_arg1_head_arg1_head;
                                                                                                                                                                                                                                                                                                  ATermList arg_arg1_head_arg1_list = (ATermList)arg_arg1_head_arg1;
                                                                                                                                                                                                                                                                                                  arg_arg1_head_arg1_head = ATgetFirst(arg_arg1_head_arg1_list);
                                                                                                                                                                                                                                                                                                  arg_arg1_head_arg1_list = ATgetNext(arg_arg1_head_arg1_list);
                                                                                                                                                                                                                                                                                                  if (ATgetInt((ATermInt)arg_arg1_head_arg1_head) == 44) {

                                                                                                                                                                                                                                                                                                    if (ATgetType((ATerm)arg_arg1_head_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg1_list) == ATtrue) {
                                                                                                                                                                                                                                                                                                      if (ATgetType((ATerm)arg_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_list) == ATfalse) {
                                                                                                                                                                                                                                                                                                        arg_arg1_list = ATgetNext(arg_arg1_list);
                                                                                                                                                                                                                                                                                                        if (ATgetType((ATerm)arg_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_list) == ATfalse) {
                                                                                                                                                                                                                                                                                                          arg_arg1_head = ATgetFirst(arg_arg1_list);
                                                                                                                                                                                                                                                                                                          arg_arg1_list = ATgetNext(arg_arg1_list);
                                                                                                                                                                                                                                                                                                          /* checking for: appl */
                                                                                                                                                                                                                                                                                                          if (ATgetType((ATerm)arg_arg1_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head) == PERR_afun0) {
                                                                                                                                                                                                                                                                                                            ATerm arg_arg1_head_arg0 = ATgetArgument(arg_arg1_head, 0);
                                                                                                                                                                                                                                                                                                            /* checking for: prod */
                                                                                                                                                                                                                                                                                                            if (ATgetType((ATerm)arg_arg1_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0) == PERR_afun1) {
                                                                                                                                                                                                                                                                                                              ATerm arg_arg1_head_arg0_arg0 = ATgetArgument(arg_arg1_head_arg0, 0);
                                                                                                                                                                                                                                                                                                              if (ATgetType((ATerm)arg_arg1_head_arg0_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0) == ATfalse) {
                                                                                                                                                                                                                                                                                                                ATerm arg_arg1_head_arg0_arg0_head;
                                                                                                                                                                                                                                                                                                                ATermList arg_arg1_head_arg0_arg0_list = (ATermList)arg_arg1_head_arg0_arg0;
                                                                                                                                                                                                                                                                                                                arg_arg1_head_arg0_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_list);
                                                                                                                                                                                                                                                                                                                arg_arg1_head_arg0_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_list);
                                                                                                                                                                                                                                                                                                                /* checking for: char-class */
                                                                                                                                                                                                                                                                                                                if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg0_head) == PERR_afun17) {
                                                                                                                                                                                                                                                                                                                  ATerm arg_arg1_head_arg0_arg0_head_arg0 = ATgetArgument(arg_arg1_head_arg0_arg0_head, 0);
                                                                                                                                                                                                                                                                                                                  if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0) == ATfalse) {
                                                                                                                                                                                                                                                                                                                    ATerm arg_arg1_head_arg0_arg0_head_arg0_head;
                                                                                                                                                                                                                                                                                                                    ATermList arg_arg1_head_arg0_arg0_head_arg0_list = (ATermList)arg_arg1_head_arg0_arg0_head_arg0;
                                                                                                                                                                                                                                                                                                                    arg_arg1_head_arg0_arg0_head_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                                                                                                                                                                                                                                    arg_arg1_head_arg0_arg0_head_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                                                                                                                                                                                                                                    if (ATgetInt((ATermInt)arg_arg1_head_arg0_arg0_head_arg0_head) == 91) {

                                                                                                                                                                                                                                                                                                                      if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0_list) == ATtrue) {
                                                                                                                                                                                                                                                                                                                        if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_list) == ATtrue) {
                                                                                                                                                                                                                                                                                                                          ATerm arg_arg1_head_arg0_arg1 = ATgetArgument(arg_arg1_head_arg0, 1);
                                                                                                                                                                                                                                                                                                                          /* checking for: lit */
                                                                                                                                                                                                                                                                                                                          if (ATgetType((ATerm)arg_arg1_head_arg0_arg1) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg1) == PERR_afun9) {
                                                                                                                                                                                                                                                                                                                            ATerm arg_arg1_head_arg0_arg1_arg0 = ATgetArgument(arg_arg1_head_arg0_arg1, 0);
                                                                                                                                                                                                                                                                                                                            /* checking for: [ */
                                                                                                                                                                                                                                                                                                                            if (ATgetType((ATerm)arg_arg1_head_arg0_arg1_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg1_arg0) == PERR_afun47) {
                                                                                                                                                                                                                                                                                                                              ATerm arg_arg1_head_arg0_arg2 = ATgetArgument(arg_arg1_head_arg0, 2);
                                                                                                                                                                                                                                                                                                                              /* checking for: no-attrs */
                                                                                                                                                                                                                                                                                                                              if (ATgetType((ATerm)arg_arg1_head_arg0_arg2) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg2) == PERR_afun5) {
                                                                                                                                                                                                                                                                                                                                ATerm arg_arg1_head_arg1 = ATgetArgument(arg_arg1_head, 1);
                                                                                                                                                                                                                                                                                                                                if (ATgetType((ATerm)arg_arg1_head_arg1) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg1) == ATfalse) {
                                                                                                                                                                                                                                                                                                                                  ATerm arg_arg1_head_arg1_head;
                                                                                                                                                                                                                                                                                                                                  ATermList arg_arg1_head_arg1_list = (ATermList)arg_arg1_head_arg1;
                                                                                                                                                                                                                                                                                                                                  arg_arg1_head_arg1_head = ATgetFirst(arg_arg1_head_arg1_list);
                                                                                                                                                                                                                                                                                                                                  arg_arg1_head_arg1_list = ATgetNext(arg_arg1_head_arg1_list);
                                                                                                                                                                                                                                                                                                                                  if (ATgetInt((ATermInt)arg_arg1_head_arg1_head) == 91) {

                                                                                                                                                                                                                                                                                                                                    if (ATgetType((ATerm)arg_arg1_head_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg1_list) == ATtrue) {
                                                                                                                                                                                                                                                                                                                                      if (ATgetType((ATerm)arg_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_list) == ATfalse) {
                                                                                                                                                                                                                                                                                                                                        arg_arg1_list = ATgetNext(arg_arg1_list);
                                                                                                                                                                                                                                                                                                                                        if (ATgetType((ATerm)arg_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_list) == ATfalse) {
                                                                                                                                                                                                                                                                                                                                          arg_arg1_head = ATgetFirst(arg_arg1_list);
                                                                                                                                                                                                                                                                                                                                          arg_arg1_list = ATgetNext(arg_arg1_list);
                                                                                                                                                                                                                                                                                                                                          /* checking for: appl */
                                                                                                                                                                                                                                                                                                                                          if (ATgetType((ATerm)arg_arg1_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head) == PERR_afun0) {
                                                                                                                                                                                                                                                                                                                                            ATerm arg_arg1_head_arg0 = ATgetArgument(arg_arg1_head, 0);
                                                                                                                                                                                                                                                                                                                                            /* checking for: list */
                                                                                                                                                                                                                                                                                                                                            if (ATgetType((ATerm)arg_arg1_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0) == PERR_afun8) {
                                                                                                                                                                                                                                                                                                                                              ATerm arg_arg1_head_arg0_arg0 = ATgetArgument(arg_arg1_head_arg0, 0);
                                                                                                                                                                                                                                                                                                                                              /* checking for: cf */
                                                                                                                                                                                                                                                                                                                                              if (ATgetType((ATerm)arg_arg1_head_arg0_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg0) == PERR_afun2) {
                                                                                                                                                                                                                                                                                                                                                ATerm arg_arg1_head_arg0_arg0_arg0 = ATgetArgument(arg_arg1_head_arg0_arg0, 0);
                                                                                                                                                                                                                                                                                                                                                /* checking for: iter-star-sep */
                                                                                                                                                                                                                                                                                                                                                if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg0_arg0) == PERR_afun46) {
                                                                                                                                                                                                                                                                                                                                                  ATerm arg_arg1_head_arg0_arg0_arg0_arg0 = ATgetArgument(arg_arg1_head_arg0_arg0_arg0, 0);
                                                                                                                                                                                                                                                                                                                                                  /* checking for: sort */
                                                                                                                                                                                                                                                                                                                                                  if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_arg0_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg0_arg0_arg0) == PERR_afun11) {
                                                                                                                                                                                                                                                                                                                                                    ATerm arg_arg1_head_arg0_arg0_arg0_arg0_arg0 = ATgetArgument(arg_arg1_head_arg0_arg0_arg0_arg0, 0);
                                                                                                                                                                                                                                                                                                                                                    /* checking for: Subject */
                                                                                                                                                                                                                                                                                                                                                    if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_arg0_arg0_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg0_arg0_arg0_arg0) == PERR_afun41) {
                                                                                                                                                                                                                                                                                                                                                      ATerm arg_arg1_head_arg0_arg0_arg0_arg1 = ATgetArgument(arg_arg1_head_arg0_arg0_arg0, 1);
                                                                                                                                                                                                                                                                                                                                                      /* checking for: lit */
                                                                                                                                                                                                                                                                                                                                                      if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_arg0_arg1) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg0_arg0_arg1) == PERR_afun9) {
                                                                                                                                                                                                                                                                                                                                                        ATerm arg_arg1_head_arg0_arg0_arg0_arg1_arg0 = ATgetArgument(arg_arg1_head_arg0_arg0_arg0_arg1, 0);
                                                                                                                                                                                                                                                                                                                                                        /* checking for: , */
                                                                                                                                                                                                                                                                                                                                                        if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_arg0_arg1_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg0_arg0_arg1_arg0) == PERR_afun43) {
                                                                                                                                                                                                                                                                                                                                                          if (ATgetType((ATerm)arg_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_list) == ATfalse) {
                                                                                                                                                                                                                                                                                                                                                            arg_arg1_list = ATgetNext(arg_arg1_list);
                                                                                                                                                                                                                                                                                                                                                            if (ATgetType((ATerm)arg_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_list) == ATfalse) {
                                                                                                                                                                                                                                                                                                                                                              arg_arg1_head = ATgetFirst(arg_arg1_list);
                                                                                                                                                                                                                                                                                                                                                              arg_arg1_list = ATgetNext(arg_arg1_list);
                                                                                                                                                                                                                                                                                                                                                              /* checking for: appl */
                                                                                                                                                                                                                                                                                                                                                              if (ATgetType((ATerm)arg_arg1_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head) == PERR_afun0) {
                                                                                                                                                                                                                                                                                                                                                                ATerm arg_arg1_head_arg0 = ATgetArgument(arg_arg1_head, 0);
                                                                                                                                                                                                                                                                                                                                                                /* checking for: prod */
                                                                                                                                                                                                                                                                                                                                                                if (ATgetType((ATerm)arg_arg1_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0) == PERR_afun1) {
                                                                                                                                                                                                                                                                                                                                                                  ATerm arg_arg1_head_arg0_arg0 = ATgetArgument(arg_arg1_head_arg0, 0);
                                                                                                                                                                                                                                                                                                                                                                  if (ATgetType((ATerm)arg_arg1_head_arg0_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0) == ATfalse) {
                                                                                                                                                                                                                                                                                                                                                                    ATerm arg_arg1_head_arg0_arg0_head;
                                                                                                                                                                                                                                                                                                                                                                    ATermList arg_arg1_head_arg0_arg0_list = (ATermList)arg_arg1_head_arg0_arg0;
                                                                                                                                                                                                                                                                                                                                                                    arg_arg1_head_arg0_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_list);
                                                                                                                                                                                                                                                                                                                                                                    arg_arg1_head_arg0_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_list);
                                                                                                                                                                                                                                                                                                                                                                    /* checking for: char-class */
                                                                                                                                                                                                                                                                                                                                                                    if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg0_head) == PERR_afun17) {
                                                                                                                                                                                                                                                                                                                                                                      ATerm arg_arg1_head_arg0_arg0_head_arg0 = ATgetArgument(arg_arg1_head_arg0_arg0_head, 0);
                                                                                                                                                                                                                                                                                                                                                                      if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0) == ATfalse) {
                                                                                                                                                                                                                                                                                                                                                                        ATerm arg_arg1_head_arg0_arg0_head_arg0_head;
                                                                                                                                                                                                                                                                                                                                                                        ATermList arg_arg1_head_arg0_arg0_head_arg0_list = (ATermList)arg_arg1_head_arg0_arg0_head_arg0;
                                                                                                                                                                                                                                                                                                                                                                        arg_arg1_head_arg0_arg0_head_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                                                                                                                                                                                                                                                                                        arg_arg1_head_arg0_arg0_head_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                                                                                                                                                                                                                                                                                        if (ATgetInt((ATermInt)arg_arg1_head_arg0_arg0_head_arg0_head) == 93) {

                                                                                                                                                                                                                                                                                                                                                                          if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0_list) == ATtrue) {
                                                                                                                                                                                                                                                                                                                                                                            if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_list) == ATtrue) {
                                                                                                                                                                                                                                                                                                                                                                              ATerm arg_arg1_head_arg0_arg1 = ATgetArgument(arg_arg1_head_arg0, 1);
                                                                                                                                                                                                                                                                                                                                                                              /* checking for: lit */
                                                                                                                                                                                                                                                                                                                                                                              if (ATgetType((ATerm)arg_arg1_head_arg0_arg1) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg1) == PERR_afun9) {
                                                                                                                                                                                                                                                                                                                                                                                ATerm arg_arg1_head_arg0_arg1_arg0 = ATgetArgument(arg_arg1_head_arg0_arg1, 0);
                                                                                                                                                                                                                                                                                                                                                                                /* checking for: ] */
                                                                                                                                                                                                                                                                                                                                                                                if (ATgetType((ATerm)arg_arg1_head_arg0_arg1_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg1_arg0) == PERR_afun45) {
                                                                                                                                                                                                                                                                                                                                                                                  ATerm arg_arg1_head_arg0_arg2 = ATgetArgument(arg_arg1_head_arg0, 2);
                                                                                                                                                                                                                                                                                                                                                                                  /* checking for: no-attrs */
                                                                                                                                                                                                                                                                                                                                                                                  if (ATgetType((ATerm)arg_arg1_head_arg0_arg2) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg2) == PERR_afun5) {
                                                                                                                                                                                                                                                                                                                                                                                    ATerm arg_arg1_head_arg1 = ATgetArgument(arg_arg1_head, 1);
                                                                                                                                                                                                                                                                                                                                                                                    if (ATgetType((ATerm)arg_arg1_head_arg1) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg1) == ATfalse) {
                                                                                                                                                                                                                                                                                                                                                                                      ATerm arg_arg1_head_arg1_head;
                                                                                                                                                                                                                                                                                                                                                                                      ATermList arg_arg1_head_arg1_list = (ATermList)arg_arg1_head_arg1;
                                                                                                                                                                                                                                                                                                                                                                                      arg_arg1_head_arg1_head = ATgetFirst(arg_arg1_head_arg1_list);
                                                                                                                                                                                                                                                                                                                                                                                      arg_arg1_head_arg1_list = ATgetNext(arg_arg1_head_arg1_list);
                                                                                                                                                                                                                                                                                                                                                                                      if (ATgetInt((ATermInt)arg_arg1_head_arg1_head) == 93) {

                                                                                                                                                                                                                                                                                                                                                                                        if (ATgetType((ATerm)arg_arg1_head_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg1_list) == ATtrue) {
                                                                                                                                                                                                                                                                                                                                                                                          if (ATgetType((ATerm)arg_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_list) == ATfalse) {
                                                                                                                                                                                                                                                                                                                                                                                            arg_arg1_list = ATgetNext(arg_arg1_list);
                                                                                                                                                                                                                                                                                                                                                                                            if (ATgetType((ATerm)arg_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_list) == ATfalse) {
                                                                                                                                                                                                                                                                                                                                                                                              arg_arg1_head = ATgetFirst(arg_arg1_list);
                                                                                                                                                                                                                                                                                                                                                                                              arg_arg1_list = ATgetNext(arg_arg1_list);
                                                                                                                                                                                                                                                                                                                                                                                              /* checking for: appl */
                                                                                                                                                                                                                                                                                                                                                                                              if (ATgetType((ATerm)arg_arg1_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head) == PERR_afun0) {
                                                                                                                                                                                                                                                                                                                                                                                                ATerm arg_arg1_head_arg0 = ATgetArgument(arg_arg1_head, 0);
                                                                                                                                                                                                                                                                                                                                                                                                /* checking for: prod */
                                                                                                                                                                                                                                                                                                                                                                                                if (ATgetType((ATerm)arg_arg1_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0) == PERR_afun1) {
                                                                                                                                                                                                                                                                                                                                                                                                  ATerm arg_arg1_head_arg0_arg0 = ATgetArgument(arg_arg1_head_arg0, 0);
                                                                                                                                                                                                                                                                                                                                                                                                  if (ATgetType((ATerm)arg_arg1_head_arg0_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0) == ATfalse) {
                                                                                                                                                                                                                                                                                                                                                                                                    ATerm arg_arg1_head_arg0_arg0_head;
                                                                                                                                                                                                                                                                                                                                                                                                    ATermList arg_arg1_head_arg0_arg0_list = (ATermList)arg_arg1_head_arg0_arg0;
                                                                                                                                                                                                                                                                                                                                                                                                    arg_arg1_head_arg0_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_list);
                                                                                                                                                                                                                                                                                                                                                                                                    arg_arg1_head_arg0_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_list);
                                                                                                                                                                                                                                                                                                                                                                                                    /* checking for: char-class */
                                                                                                                                                                                                                                                                                                                                                                                                    if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg0_head) == PERR_afun17) {
                                                                                                                                                                                                                                                                                                                                                                                                      ATerm arg_arg1_head_arg0_arg0_head_arg0 = ATgetArgument(arg_arg1_head_arg0_arg0_head, 0);
                                                                                                                                                                                                                                                                                                                                                                                                      if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0) == ATfalse) {
                                                                                                                                                                                                                                                                                                                                                                                                        ATerm arg_arg1_head_arg0_arg0_head_arg0_head;
                                                                                                                                                                                                                                                                                                                                                                                                        ATermList arg_arg1_head_arg0_arg0_head_arg0_list = (ATermList)arg_arg1_head_arg0_arg0_head_arg0;
                                                                                                                                                                                                                                                                                                                                                                                                        arg_arg1_head_arg0_arg0_head_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                                                                                                                                                                                                                                                                                                                        arg_arg1_head_arg0_arg0_head_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                                                                                                                                                                                                                                                                                                                        if (ATgetInt((ATermInt)arg_arg1_head_arg0_arg0_head_arg0_head) == 41) {

                                                                                                                                                                                                                                                                                                                                                                                                          if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0_list) == ATtrue) {
                                                                                                                                                                                                                                                                                                                                                                                                            if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_list) == ATtrue) {
                                                                                                                                                                                                                                                                                                                                                                                                              ATerm arg_arg1_head_arg0_arg1 = ATgetArgument(arg_arg1_head_arg0, 1);
                                                                                                                                                                                                                                                                                                                                                                                                              /* checking for: lit */
                                                                                                                                                                                                                                                                                                                                                                                                              if (ATgetType((ATerm)arg_arg1_head_arg0_arg1) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg1) == PERR_afun9) {
                                                                                                                                                                                                                                                                                                                                                                                                                ATerm arg_arg1_head_arg0_arg1_arg0 = ATgetArgument(arg_arg1_head_arg0_arg1, 0);
                                                                                                                                                                                                                                                                                                                                                                                                                /* checking for: ) */
                                                                                                                                                                                                                                                                                                                                                                                                                if (ATgetType((ATerm)arg_arg1_head_arg0_arg1_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg1_arg0) == PERR_afun38) {
                                                                                                                                                                                                                                                                                                                                                                                                                  ATerm arg_arg1_head_arg0_arg2 = ATgetArgument(arg_arg1_head_arg0, 2);
                                                                                                                                                                                                                                                                                                                                                                                                                  /* checking for: no-attrs */
                                                                                                                                                                                                                                                                                                                                                                                                                  if (ATgetType((ATerm)arg_arg1_head_arg0_arg2) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg2) == PERR_afun5) {
                                                                                                                                                                                                                                                                                                                                                                                                                    ATerm arg_arg1_head_arg1 = ATgetArgument(arg_arg1_head, 1);
                                                                                                                                                                                                                                                                                                                                                                                                                    if (ATgetType((ATerm)arg_arg1_head_arg1) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg1) == ATfalse) {
                                                                                                                                                                                                                                                                                                                                                                                                                      ATerm arg_arg1_head_arg1_head;
                                                                                                                                                                                                                                                                                                                                                                                                                      ATermList arg_arg1_head_arg1_list = (ATermList)arg_arg1_head_arg1;
                                                                                                                                                                                                                                                                                                                                                                                                                      arg_arg1_head_arg1_head = ATgetFirst(arg_arg1_head_arg1_list);
                                                                                                                                                                                                                                                                                                                                                                                                                      arg_arg1_head_arg1_list = ATgetNext(arg_arg1_head_arg1_list);
                                                                                                                                                                                                                                                                                                                                                                                                                      if (ATgetInt((ATermInt)arg_arg1_head_arg1_head) == 41) {

                                                                                                                                                                                                                                                                                                                                                                                                                        if (ATgetType((ATerm)arg_arg1_head_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg1_list) == ATtrue) {
                                                                                                                                                                                                                                                                                                                                                                                                                          if (ATgetType((ATerm)arg_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_list) == ATtrue) {
                                                                                                                                                                                                                                                                                                                                                                                                                            return ATtrue;
                                                                                                                                                                                                                                                                                                                                                                                                                          }
                                                                                                                                                                                                                                                                                                                                                                                                                        }
                                                                                                                                                                                                                                                                                                                                                                                                                      }
                                                                                                                                                                                                                                                                                                                                                                                                                    }
                                                                                                                                                                                                                                                                                                                                                                                                                  }
                                                                                                                                                                                                                                                                                                                                                                                                                }
                                                                                                                                                                                                                                                                                                                                                                                                              }
                                                                                                                                                                                                                                                                                                                                                                                                            }
                                                                                                                                                                                                                                                                                                                                                                                                          }
                                                                                                                                                                                                                                                                                                                                                                                                        }
                                                                                                                                                                                                                                                                                                                                                                                                      }
                                                                                                                                                                                                                                                                                                                                                                                                    }
                                                                                                                                                                                                                                                                                                                                                                                                  }
                                                                                                                                                                                                                                                                                                                                                                                                }
                                                                                                                                                                                                                                                                                                                                                                                              }
                                                                                                                                                                                                                                                                                                                                                                                            }
                                                                                                                                                                                                                                                                                                                                                                                          }
                                                                                                                                                                                                                                                                                                                                                                                        }
                                                                                                                                                                                                                                                                                                                                                                                      }
                                                                                                                                                                                                                                                                                                                                                                                    }
                                                                                                                                                                                                                                                                                                                                                                                  }
                                                                                                                                                                                                                                                                                                                                                                                }
                                                                                                                                                                                                                                                                                                                                                                              }
                                                                                                                                                                                                                                                                                                                                                                            }
                                                                                                                                                                                                                                                                                                                                                                          }
                                                                                                                                                                                                                                                                                                                                                                        }
                                                                                                                                                                                                                                                                                                                                                                      }
                                                                                                                                                                                                                                                                                                                                                                    }
                                                                                                                                                                                                                                                                                                                                                                  }
                                                                                                                                                                                                                                                                                                                                                                }
                                                                                                                                                                                                                                                                                                                                                              }
                                                                                                                                                                                                                                                                                                                                                            }
                                                                                                                                                                                                                                                                                                                                                          }
                                                                                                                                                                                                                                                                                                                                                        }
                                                                                                                                                                                                                                                                                                                                                      }
                                                                                                                                                                                                                                                                                                                                                    }
                                                                                                                                                                                                                                                                                                                                                  }
                                                                                                                                                                                                                                                                                                                                                }
                                                                                                                                                                                                                                                                                                                                              }
                                                                                                                                                                                                                                                                                                                                            }
                                                                                                                                                                                                                                                                                                                                          }
                                                                                                                                                                                                                                                                                                                                        }
                                                                                                                                                                                                                                                                                                                                      }
                                                                                                                                                                                                                                                                                                                                    }
                                                                                                                                                                                                                                                                                                                                  }
                                                                                                                                                                                                                                                                                                                                }
                                                                                                                                                                                                                                                                                                                              }
                                                                                                                                                                                                                                                                                                                            }
                                                                                                                                                                                                                                                                                                                          }
                                                                                                                                                                                                                                                                                                                        }
                                                                                                                                                                                                                                                                                                                      }
                                                                                                                                                                                                                                                                                                                    }
                                                                                                                                                                                                                                                                                                                  }
                                                                                                                                                                                                                                                                                                                }
                                                                                                                                                                                                                                                                                                              }
                                                                                                                                                                                                                                                                                                            }
                                                                                                                                                                                                                                                                                                          }
                                                                                                                                                                                                                                                                                                        }
                                                                                                                                                                                                                                                                                                      }
                                                                                                                                                                                                                                                                                                    }
                                                                                                                                                                                                                                                                                                  }
                                                                                                                                                                                                                                                                                                }
                                                                                                                                                                                                                                                                                              }
                                                                                                                                                                                                                                                                                            }
                                                                                                                                                                                                                                                                                          }
                                                                                                                                                                                                                                                                                        }
                                                                                                                                                                                                                                                                                      }
                                                                                                                                                                                                                                                                                    }
                                                                                                                                                                                                                                                                                  }
                                                                                                                                                                                                                                                                                }
                                                                                                                                                                                                                                                                              }
                                                                                                                                                                                                                                                                            }
                                                                                                                                                                                                                                                                          }
                                                                                                                                                                                                                                                                        }
                                                                                                                                                                                                                                                                      }
                                                                                                                                                                                                                                                                    }
                                                                                                                                                                                                                                                                  }
                                                                                                                                                                                                                                                                }
                                                                                                                                                                                                                                                              }
                                                                                                                                                                                                                                                            }
                                                                                                                                                                                                                                                          }
                                                                                                                                                                                                                                                        }
                                                                                                                                                                                                                                                      }
                                                                                                                                                                                                                                                    }
                                                                                                                                                                                                                                                  }
                                                                                                                                                                                                                                                }
                                                                                                                                                                                                                                              }
                                                                                                                                                                                                                                            }
                                                                                                                                                                                                                                          }
                                                                                                                                                                                                                                        }
                                                                                                                                                                                                                                      }
                                                                                                                                                                                                                                    }
                                                                                                                                                                                                                                  }
                                                                                                                                                                                                                                }
                                                                                                                                                                                                                              }
                                                                                                                                                                                                                            }
                                                                                                                                                                                                                          }
                                                                                                                                                                                                                        }
                                                                                                                                                                                                                      }
                                                                                                                                                                                                                    }
                                                                                                                                                                                                                  }
                                                                                                                                                                                                                }
                                                                                                                                                                                                              }
                                                                                                                                                                                                            }
                                                                                                                                                                                                          }
                                                                                                                                                                                                        }
                                                                                                                                                                                                      }
                                                                                                                                                                                                    }
                                                                                                                                                                                                  }
                                                                                                                                                                                                }
                                                                                                                                                                                              }
                                                                                                                                                                                            }
                                                                                                                                                                                          }
                                                                                                                                                                                        }
                                                                                                                                                                                      }
                                                                                                                                                                                    }
                                                                                                                                                                                  }
                                                                                                                                                                                }
                                                                                                                                                                              }
                                                                                                                                                                            }
                                                                                                                                                                          }
                                                                                                                                                                        }
                                                                                                                                                                      }
                                                                                                                                                                    }
                                                                                                                                                                  }
                                                                                                                                                                }
                                                                                                                                                              }
                                                                                                                                                            }
                                                                                                                                                          }
                                                                                                                                                        }
                                                                                                                                                      }
                                                                                                                                                    }
                                                                                                                                                  }
                                                                                                                                                }
                                                                                                                                              }
                                                                                                                                            }
                                                                                                                                          }
                                                                                                                                        }
                                                                                                                                      }
                                                                                                                                    }
                                                                                                                                  }
                                                                                                                                }
                                                                                                                              }
                                                                                                                            }
                                                                                                                          }
                                                                                                                        }
                                                                                                                      }
                                                                                                                    }
                                                                                                                  }
                                                                                                                }
                                                                                                              }
                                                                                                            }
                                                                                                          }
                                                                                                        }
                                                                                                      }
                                                                                                    }
                                                                                                  }
                                                                                                }
                                                                                              }
                                                                                            }
                                                                                          }
                                                                                        }
                                                                                      }
                                                                                    }
                                                                                  }
                                                                                }
                                                                              }
                                                                            }
                                                                          }
                                                                        }
                                                                      }
                                                                    }
                                                                  }
                                                                }
                                                              }
                                                            }
                                                          }
                                                        }
                                                      }
                                                    }
                                                  }
                                                }
                                              }
                                            }
                                          }
                                        }
                                      }
                                    }
                                  }
                                }
                              }
                            }
                          }
                        }
                      }
                    }
                  }
                }
              }
            }
          }
        }
      }
    }
  }
  return ATfalse;
}

/**
 * Assert whether a PERR_Error is a fatal by checking against the following ATerm pattern: appl(prod([lit("fatal"),cf(opt(layout)),lit("("),cf(opt(layout)),cf(sort("StrCon")),cf(opt(layout)),lit(","),cf(opt(layout)),lit("["),cf(opt(layout)),cf(iter-star-sep(sort("Subject"),lit(","))),cf(opt(layout)),lit("]"),cf(opt(layout)),lit(")")],cf(sort("Error")),attrs([term(cons("fatal"))])),[appl(prod([char-class([102]),char-class([97]),char-class([116]),char-class([97]),char-class([108])],lit("fatal"),no-attrs),[102,97,116,97,108]),<"ws-after-fatal"("OptLayout")>,appl(prod([char-class([40])],lit("("),no-attrs),[40]),<"ws-after-("("OptLayout")>,<"description"("StrCon")>,<"ws-after-description"("OptLayout")>,appl(prod([char-class([44])],lit(","),no-attrs),[44]),<"ws-after-,"("OptLayout")>,appl(prod([char-class([91])],lit("["),no-attrs),[91]),<"ws-after-["("OptLayout")>,appl(list(cf(iter-star-sep(sort("Subject"),lit(",")))),<"list"("Subject-list")>),<"ws-after-list"("OptLayout")>,appl(prod([char-class([93])],lit("]"),no-attrs),[93]),<"ws-after-]"("OptLayout")>,appl(prod([char-class([41])],lit(")"),no-attrs),[41])]). May not be used to assert correctness of the PERR_Error
 * \param[in] arg input PERR_Error
 * \return ATtrue if #arg corresponds to the signature of a fatal, or ATfalse otherwise
 */
inline ATbool PERR_isErrorFatal(PERR_Error arg){
  /* checking for: appl */
  if (ATgetType((ATerm)arg) == AT_APPL && ATgetAFun((ATermAppl)arg) == PERR_afun0) {
    ATerm arg_arg0 = ATgetArgument(arg, 0);
    /* checking for: prod */
    if (ATgetType((ATerm)arg_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0) == PERR_afun1) {
      ATerm arg_arg0_arg0 = ATgetArgument(arg_arg0, 0);
      if (ATgetType((ATerm)arg_arg0_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0) == ATfalse) {
        ATerm arg_arg0_arg0_head;
        ATermList arg_arg0_arg0_list = (ATermList)arg_arg0_arg0;
        arg_arg0_arg0_head = ATgetFirst(arg_arg0_arg0_list);
        arg_arg0_arg0_list = ATgetNext(arg_arg0_arg0_list);
        /* checking for: lit */
        if (ATgetType((ATerm)arg_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head) == PERR_afun9) {
          ATerm arg_arg0_arg0_head_arg0 = ATgetArgument(arg_arg0_arg0_head, 0);
          /* checking for: fatal */
          if (ATgetType((ATerm)arg_arg0_arg0_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0) == PERR_afun51) {
            if (ATgetType((ATerm)arg_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0_list) == ATfalse) {
              arg_arg0_arg0_head = ATgetFirst(arg_arg0_arg0_list);
              arg_arg0_arg0_list = ATgetNext(arg_arg0_arg0_list);
              /* checking for: cf */
              if (ATgetType((ATerm)arg_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head) == PERR_afun2) {
                ATerm arg_arg0_arg0_head_arg0 = ATgetArgument(arg_arg0_arg0_head, 0);
                /* checking for: opt */
                if (ATgetType((ATerm)arg_arg0_arg0_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0) == PERR_afun3) {
                  ATerm arg_arg0_arg0_head_arg0_arg0 = ATgetArgument(arg_arg0_arg0_head_arg0, 0);
                  /* checking for: layout */
                  if (ATgetType((ATerm)arg_arg0_arg0_head_arg0_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0_arg0) == PERR_afun4) {
                    if (ATgetType((ATerm)arg_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0_list) == ATfalse) {
                      arg_arg0_arg0_head = ATgetFirst(arg_arg0_arg0_list);
                      arg_arg0_arg0_list = ATgetNext(arg_arg0_arg0_list);
                      /* checking for: lit */
                      if (ATgetType((ATerm)arg_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head) == PERR_afun9) {
                        ATerm arg_arg0_arg0_head_arg0 = ATgetArgument(arg_arg0_arg0_head, 0);
                        /* checking for: ( */
                        if (ATgetType((ATerm)arg_arg0_arg0_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0) == PERR_afun39) {
                          if (ATgetType((ATerm)arg_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0_list) == ATfalse) {
                            arg_arg0_arg0_head = ATgetFirst(arg_arg0_arg0_list);
                            arg_arg0_arg0_list = ATgetNext(arg_arg0_arg0_list);
                            /* checking for: cf */
                            if (ATgetType((ATerm)arg_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head) == PERR_afun2) {
                              ATerm arg_arg0_arg0_head_arg0 = ATgetArgument(arg_arg0_arg0_head, 0);
                              /* checking for: opt */
                              if (ATgetType((ATerm)arg_arg0_arg0_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0) == PERR_afun3) {
                                ATerm arg_arg0_arg0_head_arg0_arg0 = ATgetArgument(arg_arg0_arg0_head_arg0, 0);
                                /* checking for: layout */
                                if (ATgetType((ATerm)arg_arg0_arg0_head_arg0_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0_arg0) == PERR_afun4) {
                                  if (ATgetType((ATerm)arg_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0_list) == ATfalse) {
                                    arg_arg0_arg0_head = ATgetFirst(arg_arg0_arg0_list);
                                    arg_arg0_arg0_list = ATgetNext(arg_arg0_arg0_list);
                                    /* checking for: cf */
                                    if (ATgetType((ATerm)arg_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head) == PERR_afun2) {
                                      ATerm arg_arg0_arg0_head_arg0 = ATgetArgument(arg_arg0_arg0_head, 0);
                                      /* checking for: sort */
                                      if (ATgetType((ATerm)arg_arg0_arg0_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0) == PERR_afun11) {
                                        ATerm arg_arg0_arg0_head_arg0_arg0 = ATgetArgument(arg_arg0_arg0_head_arg0, 0);
                                        /* checking for: StrCon */
                                        if (ATgetType((ATerm)arg_arg0_arg0_head_arg0_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0_arg0) == PERR_afun29) {
                                          if (ATgetType((ATerm)arg_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0_list) == ATfalse) {
                                            arg_arg0_arg0_head = ATgetFirst(arg_arg0_arg0_list);
                                            arg_arg0_arg0_list = ATgetNext(arg_arg0_arg0_list);
                                            /* checking for: cf */
                                            if (ATgetType((ATerm)arg_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head) == PERR_afun2) {
                                              ATerm arg_arg0_arg0_head_arg0 = ATgetArgument(arg_arg0_arg0_head, 0);
                                              /* checking for: opt */
                                              if (ATgetType((ATerm)arg_arg0_arg0_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0) == PERR_afun3) {
                                                ATerm arg_arg0_arg0_head_arg0_arg0 = ATgetArgument(arg_arg0_arg0_head_arg0, 0);
                                                /* checking for: layout */
                                                if (ATgetType((ATerm)arg_arg0_arg0_head_arg0_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0_arg0) == PERR_afun4) {
                                                  if (ATgetType((ATerm)arg_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0_list) == ATfalse) {
                                                    arg_arg0_arg0_head = ATgetFirst(arg_arg0_arg0_list);
                                                    arg_arg0_arg0_list = ATgetNext(arg_arg0_arg0_list);
                                                    /* checking for: lit */
                                                    if (ATgetType((ATerm)arg_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head) == PERR_afun9) {
                                                      ATerm arg_arg0_arg0_head_arg0 = ATgetArgument(arg_arg0_arg0_head, 0);
                                                      /* checking for: , */
                                                      if (ATgetType((ATerm)arg_arg0_arg0_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0) == PERR_afun43) {
                                                        if (ATgetType((ATerm)arg_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0_list) == ATfalse) {
                                                          arg_arg0_arg0_head = ATgetFirst(arg_arg0_arg0_list);
                                                          arg_arg0_arg0_list = ATgetNext(arg_arg0_arg0_list);
                                                          /* checking for: cf */
                                                          if (ATgetType((ATerm)arg_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head) == PERR_afun2) {
                                                            ATerm arg_arg0_arg0_head_arg0 = ATgetArgument(arg_arg0_arg0_head, 0);
                                                            /* checking for: opt */
                                                            if (ATgetType((ATerm)arg_arg0_arg0_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0) == PERR_afun3) {
                                                              ATerm arg_arg0_arg0_head_arg0_arg0 = ATgetArgument(arg_arg0_arg0_head_arg0, 0);
                                                              /* checking for: layout */
                                                              if (ATgetType((ATerm)arg_arg0_arg0_head_arg0_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0_arg0) == PERR_afun4) {
                                                                if (ATgetType((ATerm)arg_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0_list) == ATfalse) {
                                                                  arg_arg0_arg0_head = ATgetFirst(arg_arg0_arg0_list);
                                                                  arg_arg0_arg0_list = ATgetNext(arg_arg0_arg0_list);
                                                                  /* checking for: lit */
                                                                  if (ATgetType((ATerm)arg_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head) == PERR_afun9) {
                                                                    ATerm arg_arg0_arg0_head_arg0 = ATgetArgument(arg_arg0_arg0_head, 0);
                                                                    /* checking for: [ */
                                                                    if (ATgetType((ATerm)arg_arg0_arg0_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0) == PERR_afun47) {
                                                                      if (ATgetType((ATerm)arg_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0_list) == ATfalse) {
                                                                        arg_arg0_arg0_head = ATgetFirst(arg_arg0_arg0_list);
                                                                        arg_arg0_arg0_list = ATgetNext(arg_arg0_arg0_list);
                                                                        /* checking for: cf */
                                                                        if (ATgetType((ATerm)arg_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head) == PERR_afun2) {
                                                                          ATerm arg_arg0_arg0_head_arg0 = ATgetArgument(arg_arg0_arg0_head, 0);
                                                                          /* checking for: opt */
                                                                          if (ATgetType((ATerm)arg_arg0_arg0_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0) == PERR_afun3) {
                                                                            ATerm arg_arg0_arg0_head_arg0_arg0 = ATgetArgument(arg_arg0_arg0_head_arg0, 0);
                                                                            /* checking for: layout */
                                                                            if (ATgetType((ATerm)arg_arg0_arg0_head_arg0_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0_arg0) == PERR_afun4) {
                                                                              if (ATgetType((ATerm)arg_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0_list) == ATfalse) {
                                                                                arg_arg0_arg0_head = ATgetFirst(arg_arg0_arg0_list);
                                                                                arg_arg0_arg0_list = ATgetNext(arg_arg0_arg0_list);
                                                                                /* checking for: cf */
                                                                                if (ATgetType((ATerm)arg_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head) == PERR_afun2) {
                                                                                  ATerm arg_arg0_arg0_head_arg0 = ATgetArgument(arg_arg0_arg0_head, 0);
                                                                                  /* checking for: iter-star-sep */
                                                                                  if (ATgetType((ATerm)arg_arg0_arg0_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0) == PERR_afun46) {
                                                                                    ATerm arg_arg0_arg0_head_arg0_arg0 = ATgetArgument(arg_arg0_arg0_head_arg0, 0);
                                                                                    /* checking for: sort */
                                                                                    if (ATgetType((ATerm)arg_arg0_arg0_head_arg0_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0_arg0) == PERR_afun11) {
                                                                                      ATerm arg_arg0_arg0_head_arg0_arg0_arg0 = ATgetArgument(arg_arg0_arg0_head_arg0_arg0, 0);
                                                                                      /* checking for: Subject */
                                                                                      if (ATgetType((ATerm)arg_arg0_arg0_head_arg0_arg0_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0_arg0_arg0) == PERR_afun41) {
                                                                                        ATerm arg_arg0_arg0_head_arg0_arg1 = ATgetArgument(arg_arg0_arg0_head_arg0, 1);
                                                                                        /* checking for: lit */
                                                                                        if (ATgetType((ATerm)arg_arg0_arg0_head_arg0_arg1) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0_arg1) == PERR_afun9) {
                                                                                          ATerm arg_arg0_arg0_head_arg0_arg1_arg0 = ATgetArgument(arg_arg0_arg0_head_arg0_arg1, 0);
                                                                                          /* checking for: , */
                                                                                          if (ATgetType((ATerm)arg_arg0_arg0_head_arg0_arg1_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0_arg1_arg0) == PERR_afun43) {
                                                                                            if (ATgetType((ATerm)arg_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0_list) == ATfalse) {
                                                                                              arg_arg0_arg0_head = ATgetFirst(arg_arg0_arg0_list);
                                                                                              arg_arg0_arg0_list = ATgetNext(arg_arg0_arg0_list);
                                                                                              /* checking for: cf */
                                                                                              if (ATgetType((ATerm)arg_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head) == PERR_afun2) {
                                                                                                ATerm arg_arg0_arg0_head_arg0 = ATgetArgument(arg_arg0_arg0_head, 0);
                                                                                                /* checking for: opt */
                                                                                                if (ATgetType((ATerm)arg_arg0_arg0_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0) == PERR_afun3) {
                                                                                                  ATerm arg_arg0_arg0_head_arg0_arg0 = ATgetArgument(arg_arg0_arg0_head_arg0, 0);
                                                                                                  /* checking for: layout */
                                                                                                  if (ATgetType((ATerm)arg_arg0_arg0_head_arg0_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0_arg0) == PERR_afun4) {
                                                                                                    if (ATgetType((ATerm)arg_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0_list) == ATfalse) {
                                                                                                      arg_arg0_arg0_head = ATgetFirst(arg_arg0_arg0_list);
                                                                                                      arg_arg0_arg0_list = ATgetNext(arg_arg0_arg0_list);
                                                                                                      /* checking for: lit */
                                                                                                      if (ATgetType((ATerm)arg_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head) == PERR_afun9) {
                                                                                                        ATerm arg_arg0_arg0_head_arg0 = ATgetArgument(arg_arg0_arg0_head, 0);
                                                                                                        /* checking for: ] */
                                                                                                        if (ATgetType((ATerm)arg_arg0_arg0_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0) == PERR_afun45) {
                                                                                                          if (ATgetType((ATerm)arg_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0_list) == ATfalse) {
                                                                                                            arg_arg0_arg0_head = ATgetFirst(arg_arg0_arg0_list);
                                                                                                            arg_arg0_arg0_list = ATgetNext(arg_arg0_arg0_list);
                                                                                                            /* checking for: cf */
                                                                                                            if (ATgetType((ATerm)arg_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head) == PERR_afun2) {
                                                                                                              ATerm arg_arg0_arg0_head_arg0 = ATgetArgument(arg_arg0_arg0_head, 0);
                                                                                                              /* checking for: opt */
                                                                                                              if (ATgetType((ATerm)arg_arg0_arg0_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0) == PERR_afun3) {
                                                                                                                ATerm arg_arg0_arg0_head_arg0_arg0 = ATgetArgument(arg_arg0_arg0_head_arg0, 0);
                                                                                                                /* checking for: layout */
                                                                                                                if (ATgetType((ATerm)arg_arg0_arg0_head_arg0_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0_arg0) == PERR_afun4) {
                                                                                                                  if (ATgetType((ATerm)arg_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0_list) == ATfalse) {
                                                                                                                    arg_arg0_arg0_head = ATgetFirst(arg_arg0_arg0_list);
                                                                                                                    arg_arg0_arg0_list = ATgetNext(arg_arg0_arg0_list);
                                                                                                                    /* checking for: lit */
                                                                                                                    if (ATgetType((ATerm)arg_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head) == PERR_afun9) {
                                                                                                                      ATerm arg_arg0_arg0_head_arg0 = ATgetArgument(arg_arg0_arg0_head, 0);
                                                                                                                      /* checking for: ) */
                                                                                                                      if (ATgetType((ATerm)arg_arg0_arg0_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0) == PERR_afun38) {
                                                                                                                        if (ATgetType((ATerm)arg_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0_list) == ATtrue) {
                                                                                                                          ATerm arg_arg0_arg1 = ATgetArgument(arg_arg0, 1);
                                                                                                                          /* checking for: cf */
                                                                                                                          if (ATgetType((ATerm)arg_arg0_arg1) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg1) == PERR_afun2) {
                                                                                                                            ATerm arg_arg0_arg1_arg0 = ATgetArgument(arg_arg0_arg1, 0);
                                                                                                                            /* checking for: sort */
                                                                                                                            if (ATgetType((ATerm)arg_arg0_arg1_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg1_arg0) == PERR_afun11) {
                                                                                                                              ATerm arg_arg0_arg1_arg0_arg0 = ATgetArgument(arg_arg0_arg1_arg0, 0);
                                                                                                                              /* checking for: Error */
                                                                                                                              if (ATgetType((ATerm)arg_arg0_arg1_arg0_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg1_arg0_arg0) == PERR_afun37) {
                                                                                                                                ATerm arg_arg0_arg2 = ATgetArgument(arg_arg0, 2);
                                                                                                                                /* checking for: attrs */
                                                                                                                                if (ATgetType((ATerm)arg_arg0_arg2) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg2) == PERR_afun13) {
                                                                                                                                  ATerm arg_arg0_arg2_arg0 = ATgetArgument(arg_arg0_arg2, 0);
                                                                                                                                  if (ATgetType((ATerm)arg_arg0_arg2_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg2_arg0) == ATfalse) {
                                                                                                                                    ATerm arg_arg0_arg2_arg0_head;
                                                                                                                                    ATermList arg_arg0_arg2_arg0_list = (ATermList)arg_arg0_arg2_arg0;
                                                                                                                                    arg_arg0_arg2_arg0_head = ATgetFirst(arg_arg0_arg2_arg0_list);
                                                                                                                                    arg_arg0_arg2_arg0_list = ATgetNext(arg_arg0_arg2_arg0_list);
                                                                                                                                    /* checking for: term */
                                                                                                                                    if (ATgetType((ATerm)arg_arg0_arg2_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg2_arg0_head) == PERR_afun14) {
                                                                                                                                      ATerm arg_arg0_arg2_arg0_head_arg0 = ATgetArgument(arg_arg0_arg2_arg0_head, 0);
                                                                                                                                      /* checking for: cons */
                                                                                                                                      if (ATgetType((ATerm)arg_arg0_arg2_arg0_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg2_arg0_head_arg0) == PERR_afun15) {
                                                                                                                                        ATerm arg_arg0_arg2_arg0_head_arg0_arg0 = ATgetArgument(arg_arg0_arg2_arg0_head_arg0, 0);
                                                                                                                                        /* checking for: fatal */
                                                                                                                                        if (ATgetType((ATerm)arg_arg0_arg2_arg0_head_arg0_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg2_arg0_head_arg0_arg0) == PERR_afun51) {
                                                                                                                                          if (ATgetType((ATerm)arg_arg0_arg2_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg2_arg0_list) == ATtrue) {
                                                                                                                                            ATerm arg_arg1 = ATgetArgument(arg, 1);
                                                                                                                                            if (ATgetType((ATerm)arg_arg1) == AT_LIST && ATisEmpty((ATermList)arg_arg1) == ATfalse) {
                                                                                                                                              ATerm arg_arg1_head;
                                                                                                                                              ATermList arg_arg1_list = (ATermList)arg_arg1;
                                                                                                                                              arg_arg1_head = ATgetFirst(arg_arg1_list);
                                                                                                                                              arg_arg1_list = ATgetNext(arg_arg1_list);
                                                                                                                                              /* checking for: appl */
                                                                                                                                              if (ATgetType((ATerm)arg_arg1_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head) == PERR_afun0) {
                                                                                                                                                ATerm arg_arg1_head_arg0 = ATgetArgument(arg_arg1_head, 0);
                                                                                                                                                /* checking for: prod */
                                                                                                                                                if (ATgetType((ATerm)arg_arg1_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0) == PERR_afun1) {
                                                                                                                                                  ATerm arg_arg1_head_arg0_arg0 = ATgetArgument(arg_arg1_head_arg0, 0);
                                                                                                                                                  if (ATgetType((ATerm)arg_arg1_head_arg0_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0) == ATfalse) {
                                                                                                                                                    ATerm arg_arg1_head_arg0_arg0_head;
                                                                                                                                                    ATermList arg_arg1_head_arg0_arg0_list = (ATermList)arg_arg1_head_arg0_arg0;
                                                                                                                                                    arg_arg1_head_arg0_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_list);
                                                                                                                                                    arg_arg1_head_arg0_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_list);
                                                                                                                                                    /* checking for: char-class */
                                                                                                                                                    if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg0_head) == PERR_afun17) {
                                                                                                                                                      ATerm arg_arg1_head_arg0_arg0_head_arg0 = ATgetArgument(arg_arg1_head_arg0_arg0_head, 0);
                                                                                                                                                      if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0) == ATfalse) {
                                                                                                                                                        ATerm arg_arg1_head_arg0_arg0_head_arg0_head;
                                                                                                                                                        ATermList arg_arg1_head_arg0_arg0_head_arg0_list = (ATermList)arg_arg1_head_arg0_arg0_head_arg0;
                                                                                                                                                        arg_arg1_head_arg0_arg0_head_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                                                                        arg_arg1_head_arg0_arg0_head_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                                                                        if (ATgetInt((ATermInt)arg_arg1_head_arg0_arg0_head_arg0_head) == 102) {

                                                                                                                                                          if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0_list) == ATtrue) {
                                                                                                                                                            if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_list) == ATfalse) {
                                                                                                                                                              arg_arg1_head_arg0_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_list);
                                                                                                                                                              arg_arg1_head_arg0_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_list);
                                                                                                                                                              /* checking for: char-class */
                                                                                                                                                              if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg0_head) == PERR_afun17) {
                                                                                                                                                                ATerm arg_arg1_head_arg0_arg0_head_arg0 = ATgetArgument(arg_arg1_head_arg0_arg0_head, 0);
                                                                                                                                                                if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0) == ATfalse) {
                                                                                                                                                                  ATerm arg_arg1_head_arg0_arg0_head_arg0_head;
                                                                                                                                                                  ATermList arg_arg1_head_arg0_arg0_head_arg0_list = (ATermList)arg_arg1_head_arg0_arg0_head_arg0;
                                                                                                                                                                  arg_arg1_head_arg0_arg0_head_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                                                                                  arg_arg1_head_arg0_arg0_head_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                                                                                  if (ATgetInt((ATermInt)arg_arg1_head_arg0_arg0_head_arg0_head) == 97) {

                                                                                                                                                                    if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0_list) == ATtrue) {
                                                                                                                                                                      if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_list) == ATfalse) {
                                                                                                                                                                        arg_arg1_head_arg0_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_list);
                                                                                                                                                                        arg_arg1_head_arg0_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_list);
                                                                                                                                                                        /* checking for: char-class */
                                                                                                                                                                        if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg0_head) == PERR_afun17) {
                                                                                                                                                                          ATerm arg_arg1_head_arg0_arg0_head_arg0 = ATgetArgument(arg_arg1_head_arg0_arg0_head, 0);
                                                                                                                                                                          if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0) == ATfalse) {
                                                                                                                                                                            ATerm arg_arg1_head_arg0_arg0_head_arg0_head;
                                                                                                                                                                            ATermList arg_arg1_head_arg0_arg0_head_arg0_list = (ATermList)arg_arg1_head_arg0_arg0_head_arg0;
                                                                                                                                                                            arg_arg1_head_arg0_arg0_head_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                                                                                            arg_arg1_head_arg0_arg0_head_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                                                                                            if (ATgetInt((ATermInt)arg_arg1_head_arg0_arg0_head_arg0_head) == 116) {

                                                                                                                                                                              if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0_list) == ATtrue) {
                                                                                                                                                                                if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_list) == ATfalse) {
                                                                                                                                                                                  arg_arg1_head_arg0_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_list);
                                                                                                                                                                                  arg_arg1_head_arg0_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_list);
                                                                                                                                                                                  /* checking for: char-class */
                                                                                                                                                                                  if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg0_head) == PERR_afun17) {
                                                                                                                                                                                    ATerm arg_arg1_head_arg0_arg0_head_arg0 = ATgetArgument(arg_arg1_head_arg0_arg0_head, 0);
                                                                                                                                                                                    if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0) == ATfalse) {
                                                                                                                                                                                      ATerm arg_arg1_head_arg0_arg0_head_arg0_head;
                                                                                                                                                                                      ATermList arg_arg1_head_arg0_arg0_head_arg0_list = (ATermList)arg_arg1_head_arg0_arg0_head_arg0;
                                                                                                                                                                                      arg_arg1_head_arg0_arg0_head_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                                                                                                      arg_arg1_head_arg0_arg0_head_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                                                                                                      if (ATgetInt((ATermInt)arg_arg1_head_arg0_arg0_head_arg0_head) == 97) {

                                                                                                                                                                                        if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0_list) == ATtrue) {
                                                                                                                                                                                          if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_list) == ATfalse) {
                                                                                                                                                                                            arg_arg1_head_arg0_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_list);
                                                                                                                                                                                            arg_arg1_head_arg0_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_list);
                                                                                                                                                                                            /* checking for: char-class */
                                                                                                                                                                                            if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg0_head) == PERR_afun17) {
                                                                                                                                                                                              ATerm arg_arg1_head_arg0_arg0_head_arg0 = ATgetArgument(arg_arg1_head_arg0_arg0_head, 0);
                                                                                                                                                                                              if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0) == ATfalse) {
                                                                                                                                                                                                ATerm arg_arg1_head_arg0_arg0_head_arg0_head;
                                                                                                                                                                                                ATermList arg_arg1_head_arg0_arg0_head_arg0_list = (ATermList)arg_arg1_head_arg0_arg0_head_arg0;
                                                                                                                                                                                                arg_arg1_head_arg0_arg0_head_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                                                                                                                arg_arg1_head_arg0_arg0_head_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                                                                                                                if (ATgetInt((ATermInt)arg_arg1_head_arg0_arg0_head_arg0_head) == 108) {

                                                                                                                                                                                                  if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0_list) == ATtrue) {
                                                                                                                                                                                                    if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_list) == ATtrue) {
                                                                                                                                                                                                      ATerm arg_arg1_head_arg0_arg1 = ATgetArgument(arg_arg1_head_arg0, 1);
                                                                                                                                                                                                      /* checking for: lit */
                                                                                                                                                                                                      if (ATgetType((ATerm)arg_arg1_head_arg0_arg1) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg1) == PERR_afun9) {
                                                                                                                                                                                                        ATerm arg_arg1_head_arg0_arg1_arg0 = ATgetArgument(arg_arg1_head_arg0_arg1, 0);
                                                                                                                                                                                                        /* checking for: fatal */
                                                                                                                                                                                                        if (ATgetType((ATerm)arg_arg1_head_arg0_arg1_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg1_arg0) == PERR_afun51) {
                                                                                                                                                                                                          ATerm arg_arg1_head_arg0_arg2 = ATgetArgument(arg_arg1_head_arg0, 2);
                                                                                                                                                                                                          /* checking for: no-attrs */
                                                                                                                                                                                                          if (ATgetType((ATerm)arg_arg1_head_arg0_arg2) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg2) == PERR_afun5) {
                                                                                                                                                                                                            ATerm arg_arg1_head_arg1 = ATgetArgument(arg_arg1_head, 1);
                                                                                                                                                                                                            if (ATgetType((ATerm)arg_arg1_head_arg1) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg1) == ATfalse) {
                                                                                                                                                                                                              ATerm arg_arg1_head_arg1_head;
                                                                                                                                                                                                              ATermList arg_arg1_head_arg1_list = (ATermList)arg_arg1_head_arg1;
                                                                                                                                                                                                              arg_arg1_head_arg1_head = ATgetFirst(arg_arg1_head_arg1_list);
                                                                                                                                                                                                              arg_arg1_head_arg1_list = ATgetNext(arg_arg1_head_arg1_list);
                                                                                                                                                                                                              if (ATgetInt((ATermInt)arg_arg1_head_arg1_head) == 102) {

                                                                                                                                                                                                                if (ATgetType((ATerm)arg_arg1_head_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg1_list) == ATfalse) {
                                                                                                                                                                                                                  arg_arg1_head_arg1_head = ATgetFirst(arg_arg1_head_arg1_list);
                                                                                                                                                                                                                  arg_arg1_head_arg1_list = ATgetNext(arg_arg1_head_arg1_list);
                                                                                                                                                                                                                  if (ATgetInt((ATermInt)arg_arg1_head_arg1_head) == 97) {

                                                                                                                                                                                                                    if (ATgetType((ATerm)arg_arg1_head_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg1_list) == ATfalse) {
                                                                                                                                                                                                                      arg_arg1_head_arg1_head = ATgetFirst(arg_arg1_head_arg1_list);
                                                                                                                                                                                                                      arg_arg1_head_arg1_list = ATgetNext(arg_arg1_head_arg1_list);
                                                                                                                                                                                                                      if (ATgetInt((ATermInt)arg_arg1_head_arg1_head) == 116) {

                                                                                                                                                                                                                        if (ATgetType((ATerm)arg_arg1_head_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg1_list) == ATfalse) {
                                                                                                                                                                                                                          arg_arg1_head_arg1_head = ATgetFirst(arg_arg1_head_arg1_list);
                                                                                                                                                                                                                          arg_arg1_head_arg1_list = ATgetNext(arg_arg1_head_arg1_list);
                                                                                                                                                                                                                          if (ATgetInt((ATermInt)arg_arg1_head_arg1_head) == 97) {

                                                                                                                                                                                                                            if (ATgetType((ATerm)arg_arg1_head_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg1_list) == ATfalse) {
                                                                                                                                                                                                                              arg_arg1_head_arg1_head = ATgetFirst(arg_arg1_head_arg1_list);
                                                                                                                                                                                                                              arg_arg1_head_arg1_list = ATgetNext(arg_arg1_head_arg1_list);
                                                                                                                                                                                                                              if (ATgetInt((ATermInt)arg_arg1_head_arg1_head) == 108) {

                                                                                                                                                                                                                                if (ATgetType((ATerm)arg_arg1_head_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg1_list) == ATtrue) {
                                                                                                                                                                                                                                  if (ATgetType((ATerm)arg_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_list) == ATfalse) {
                                                                                                                                                                                                                                    arg_arg1_list = ATgetNext(arg_arg1_list);
                                                                                                                                                                                                                                    if (ATgetType((ATerm)arg_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_list) == ATfalse) {
                                                                                                                                                                                                                                      arg_arg1_head = ATgetFirst(arg_arg1_list);
                                                                                                                                                                                                                                      arg_arg1_list = ATgetNext(arg_arg1_list);
                                                                                                                                                                                                                                      /* checking for: appl */
                                                                                                                                                                                                                                      if (ATgetType((ATerm)arg_arg1_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head) == PERR_afun0) {
                                                                                                                                                                                                                                        ATerm arg_arg1_head_arg0 = ATgetArgument(arg_arg1_head, 0);
                                                                                                                                                                                                                                        /* checking for: prod */
                                                                                                                                                                                                                                        if (ATgetType((ATerm)arg_arg1_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0) == PERR_afun1) {
                                                                                                                                                                                                                                          ATerm arg_arg1_head_arg0_arg0 = ATgetArgument(arg_arg1_head_arg0, 0);
                                                                                                                                                                                                                                          if (ATgetType((ATerm)arg_arg1_head_arg0_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0) == ATfalse) {
                                                                                                                                                                                                                                            ATerm arg_arg1_head_arg0_arg0_head;
                                                                                                                                                                                                                                            ATermList arg_arg1_head_arg0_arg0_list = (ATermList)arg_arg1_head_arg0_arg0;
                                                                                                                                                                                                                                            arg_arg1_head_arg0_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_list);
                                                                                                                                                                                                                                            arg_arg1_head_arg0_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_list);
                                                                                                                                                                                                                                            /* checking for: char-class */
                                                                                                                                                                                                                                            if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg0_head) == PERR_afun17) {
                                                                                                                                                                                                                                              ATerm arg_arg1_head_arg0_arg0_head_arg0 = ATgetArgument(arg_arg1_head_arg0_arg0_head, 0);
                                                                                                                                                                                                                                              if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0) == ATfalse) {
                                                                                                                                                                                                                                                ATerm arg_arg1_head_arg0_arg0_head_arg0_head;
                                                                                                                                                                                                                                                ATermList arg_arg1_head_arg0_arg0_head_arg0_list = (ATermList)arg_arg1_head_arg0_arg0_head_arg0;
                                                                                                                                                                                                                                                arg_arg1_head_arg0_arg0_head_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                                                                                                                                                                arg_arg1_head_arg0_arg0_head_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                                                                                                                                                                if (ATgetInt((ATermInt)arg_arg1_head_arg0_arg0_head_arg0_head) == 40) {

                                                                                                                                                                                                                                                  if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0_list) == ATtrue) {
                                                                                                                                                                                                                                                    if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_list) == ATtrue) {
                                                                                                                                                                                                                                                      ATerm arg_arg1_head_arg0_arg1 = ATgetArgument(arg_arg1_head_arg0, 1);
                                                                                                                                                                                                                                                      /* checking for: lit */
                                                                                                                                                                                                                                                      if (ATgetType((ATerm)arg_arg1_head_arg0_arg1) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg1) == PERR_afun9) {
                                                                                                                                                                                                                                                        ATerm arg_arg1_head_arg0_arg1_arg0 = ATgetArgument(arg_arg1_head_arg0_arg1, 0);
                                                                                                                                                                                                                                                        /* checking for: ( */
                                                                                                                                                                                                                                                        if (ATgetType((ATerm)arg_arg1_head_arg0_arg1_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg1_arg0) == PERR_afun39) {
                                                                                                                                                                                                                                                          ATerm arg_arg1_head_arg0_arg2 = ATgetArgument(arg_arg1_head_arg0, 2);
                                                                                                                                                                                                                                                          /* checking for: no-attrs */
                                                                                                                                                                                                                                                          if (ATgetType((ATerm)arg_arg1_head_arg0_arg2) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg2) == PERR_afun5) {
                                                                                                                                                                                                                                                            ATerm arg_arg1_head_arg1 = ATgetArgument(arg_arg1_head, 1);
                                                                                                                                                                                                                                                            if (ATgetType((ATerm)arg_arg1_head_arg1) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg1) == ATfalse) {
                                                                                                                                                                                                                                                              ATerm arg_arg1_head_arg1_head;
                                                                                                                                                                                                                                                              ATermList arg_arg1_head_arg1_list = (ATermList)arg_arg1_head_arg1;
                                                                                                                                                                                                                                                              arg_arg1_head_arg1_head = ATgetFirst(arg_arg1_head_arg1_list);
                                                                                                                                                                                                                                                              arg_arg1_head_arg1_list = ATgetNext(arg_arg1_head_arg1_list);
                                                                                                                                                                                                                                                              if (ATgetInt((ATermInt)arg_arg1_head_arg1_head) == 40) {

                                                                                                                                                                                                                                                                if (ATgetType((ATerm)arg_arg1_head_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg1_list) == ATtrue) {
                                                                                                                                                                                                                                                                  if (ATgetType((ATerm)arg_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_list) == ATfalse) {
                                                                                                                                                                                                                                                                    arg_arg1_list = ATgetNext(arg_arg1_list);
                                                                                                                                                                                                                                                                    if (ATgetType((ATerm)arg_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_list) == ATfalse) {
                                                                                                                                                                                                                                                                      arg_arg1_list = ATgetNext(arg_arg1_list);
                                                                                                                                                                                                                                                                      if (ATgetType((ATerm)arg_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_list) == ATfalse) {
                                                                                                                                                                                                                                                                        arg_arg1_list = ATgetNext(arg_arg1_list);
                                                                                                                                                                                                                                                                        if (ATgetType((ATerm)arg_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_list) == ATfalse) {
                                                                                                                                                                                                                                                                          arg_arg1_head = ATgetFirst(arg_arg1_list);
                                                                                                                                                                                                                                                                          arg_arg1_list = ATgetNext(arg_arg1_list);
                                                                                                                                                                                                                                                                          /* checking for: appl */
                                                                                                                                                                                                                                                                          if (ATgetType((ATerm)arg_arg1_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head) == PERR_afun0) {
                                                                                                                                                                                                                                                                            ATerm arg_arg1_head_arg0 = ATgetArgument(arg_arg1_head, 0);
                                                                                                                                                                                                                                                                            /* checking for: prod */
                                                                                                                                                                                                                                                                            if (ATgetType((ATerm)arg_arg1_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0) == PERR_afun1) {
                                                                                                                                                                                                                                                                              ATerm arg_arg1_head_arg0_arg0 = ATgetArgument(arg_arg1_head_arg0, 0);
                                                                                                                                                                                                                                                                              if (ATgetType((ATerm)arg_arg1_head_arg0_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0) == ATfalse) {
                                                                                                                                                                                                                                                                                ATerm arg_arg1_head_arg0_arg0_head;
                                                                                                                                                                                                                                                                                ATermList arg_arg1_head_arg0_arg0_list = (ATermList)arg_arg1_head_arg0_arg0;
                                                                                                                                                                                                                                                                                arg_arg1_head_arg0_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_list);
                                                                                                                                                                                                                                                                                arg_arg1_head_arg0_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_list);
                                                                                                                                                                                                                                                                                /* checking for: char-class */
                                                                                                                                                                                                                                                                                if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg0_head) == PERR_afun17) {
                                                                                                                                                                                                                                                                                  ATerm arg_arg1_head_arg0_arg0_head_arg0 = ATgetArgument(arg_arg1_head_arg0_arg0_head, 0);
                                                                                                                                                                                                                                                                                  if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0) == ATfalse) {
                                                                                                                                                                                                                                                                                    ATerm arg_arg1_head_arg0_arg0_head_arg0_head;
                                                                                                                                                                                                                                                                                    ATermList arg_arg1_head_arg0_arg0_head_arg0_list = (ATermList)arg_arg1_head_arg0_arg0_head_arg0;
                                                                                                                                                                                                                                                                                    arg_arg1_head_arg0_arg0_head_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                                                                                                                                                                                                    arg_arg1_head_arg0_arg0_head_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                                                                                                                                                                                                    if (ATgetInt((ATermInt)arg_arg1_head_arg0_arg0_head_arg0_head) == 44) {

                                                                                                                                                                                                                                                                                      if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0_list) == ATtrue) {
                                                                                                                                                                                                                                                                                        if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_list) == ATtrue) {
                                                                                                                                                                                                                                                                                          ATerm arg_arg1_head_arg0_arg1 = ATgetArgument(arg_arg1_head_arg0, 1);
                                                                                                                                                                                                                                                                                          /* checking for: lit */
                                                                                                                                                                                                                                                                                          if (ATgetType((ATerm)arg_arg1_head_arg0_arg1) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg1) == PERR_afun9) {
                                                                                                                                                                                                                                                                                            ATerm arg_arg1_head_arg0_arg1_arg0 = ATgetArgument(arg_arg1_head_arg0_arg1, 0);
                                                                                                                                                                                                                                                                                            /* checking for: , */
                                                                                                                                                                                                                                                                                            if (ATgetType((ATerm)arg_arg1_head_arg0_arg1_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg1_arg0) == PERR_afun43) {
                                                                                                                                                                                                                                                                                              ATerm arg_arg1_head_arg0_arg2 = ATgetArgument(arg_arg1_head_arg0, 2);
                                                                                                                                                                                                                                                                                              /* checking for: no-attrs */
                                                                                                                                                                                                                                                                                              if (ATgetType((ATerm)arg_arg1_head_arg0_arg2) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg2) == PERR_afun5) {
                                                                                                                                                                                                                                                                                                ATerm arg_arg1_head_arg1 = ATgetArgument(arg_arg1_head, 1);
                                                                                                                                                                                                                                                                                                if (ATgetType((ATerm)arg_arg1_head_arg1) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg1) == ATfalse) {
                                                                                                                                                                                                                                                                                                  ATerm arg_arg1_head_arg1_head;
                                                                                                                                                                                                                                                                                                  ATermList arg_arg1_head_arg1_list = (ATermList)arg_arg1_head_arg1;
                                                                                                                                                                                                                                                                                                  arg_arg1_head_arg1_head = ATgetFirst(arg_arg1_head_arg1_list);
                                                                                                                                                                                                                                                                                                  arg_arg1_head_arg1_list = ATgetNext(arg_arg1_head_arg1_list);
                                                                                                                                                                                                                                                                                                  if (ATgetInt((ATermInt)arg_arg1_head_arg1_head) == 44) {

                                                                                                                                                                                                                                                                                                    if (ATgetType((ATerm)arg_arg1_head_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg1_list) == ATtrue) {
                                                                                                                                                                                                                                                                                                      if (ATgetType((ATerm)arg_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_list) == ATfalse) {
                                                                                                                                                                                                                                                                                                        arg_arg1_list = ATgetNext(arg_arg1_list);
                                                                                                                                                                                                                                                                                                        if (ATgetType((ATerm)arg_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_list) == ATfalse) {
                                                                                                                                                                                                                                                                                                          arg_arg1_head = ATgetFirst(arg_arg1_list);
                                                                                                                                                                                                                                                                                                          arg_arg1_list = ATgetNext(arg_arg1_list);
                                                                                                                                                                                                                                                                                                          /* checking for: appl */
                                                                                                                                                                                                                                                                                                          if (ATgetType((ATerm)arg_arg1_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head) == PERR_afun0) {
                                                                                                                                                                                                                                                                                                            ATerm arg_arg1_head_arg0 = ATgetArgument(arg_arg1_head, 0);
                                                                                                                                                                                                                                                                                                            /* checking for: prod */
                                                                                                                                                                                                                                                                                                            if (ATgetType((ATerm)arg_arg1_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0) == PERR_afun1) {
                                                                                                                                                                                                                                                                                                              ATerm arg_arg1_head_arg0_arg0 = ATgetArgument(arg_arg1_head_arg0, 0);
                                                                                                                                                                                                                                                                                                              if (ATgetType((ATerm)arg_arg1_head_arg0_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0) == ATfalse) {
                                                                                                                                                                                                                                                                                                                ATerm arg_arg1_head_arg0_arg0_head;
                                                                                                                                                                                                                                                                                                                ATermList arg_arg1_head_arg0_arg0_list = (ATermList)arg_arg1_head_arg0_arg0;
                                                                                                                                                                                                                                                                                                                arg_arg1_head_arg0_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_list);
                                                                                                                                                                                                                                                                                                                arg_arg1_head_arg0_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_list);
                                                                                                                                                                                                                                                                                                                /* checking for: char-class */
                                                                                                                                                                                                                                                                                                                if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg0_head) == PERR_afun17) {
                                                                                                                                                                                                                                                                                                                  ATerm arg_arg1_head_arg0_arg0_head_arg0 = ATgetArgument(arg_arg1_head_arg0_arg0_head, 0);
                                                                                                                                                                                                                                                                                                                  if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0) == ATfalse) {
                                                                                                                                                                                                                                                                                                                    ATerm arg_arg1_head_arg0_arg0_head_arg0_head;
                                                                                                                                                                                                                                                                                                                    ATermList arg_arg1_head_arg0_arg0_head_arg0_list = (ATermList)arg_arg1_head_arg0_arg0_head_arg0;
                                                                                                                                                                                                                                                                                                                    arg_arg1_head_arg0_arg0_head_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                                                                                                                                                                                                                                    arg_arg1_head_arg0_arg0_head_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                                                                                                                                                                                                                                    if (ATgetInt((ATermInt)arg_arg1_head_arg0_arg0_head_arg0_head) == 91) {

                                                                                                                                                                                                                                                                                                                      if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0_list) == ATtrue) {
                                                                                                                                                                                                                                                                                                                        if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_list) == ATtrue) {
                                                                                                                                                                                                                                                                                                                          ATerm arg_arg1_head_arg0_arg1 = ATgetArgument(arg_arg1_head_arg0, 1);
                                                                                                                                                                                                                                                                                                                          /* checking for: lit */
                                                                                                                                                                                                                                                                                                                          if (ATgetType((ATerm)arg_arg1_head_arg0_arg1) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg1) == PERR_afun9) {
                                                                                                                                                                                                                                                                                                                            ATerm arg_arg1_head_arg0_arg1_arg0 = ATgetArgument(arg_arg1_head_arg0_arg1, 0);
                                                                                                                                                                                                                                                                                                                            /* checking for: [ */
                                                                                                                                                                                                                                                                                                                            if (ATgetType((ATerm)arg_arg1_head_arg0_arg1_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg1_arg0) == PERR_afun47) {
                                                                                                                                                                                                                                                                                                                              ATerm arg_arg1_head_arg0_arg2 = ATgetArgument(arg_arg1_head_arg0, 2);
                                                                                                                                                                                                                                                                                                                              /* checking for: no-attrs */
                                                                                                                                                                                                                                                                                                                              if (ATgetType((ATerm)arg_arg1_head_arg0_arg2) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg2) == PERR_afun5) {
                                                                                                                                                                                                                                                                                                                                ATerm arg_arg1_head_arg1 = ATgetArgument(arg_arg1_head, 1);
                                                                                                                                                                                                                                                                                                                                if (ATgetType((ATerm)arg_arg1_head_arg1) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg1) == ATfalse) {
                                                                                                                                                                                                                                                                                                                                  ATerm arg_arg1_head_arg1_head;
                                                                                                                                                                                                                                                                                                                                  ATermList arg_arg1_head_arg1_list = (ATermList)arg_arg1_head_arg1;
                                                                                                                                                                                                                                                                                                                                  arg_arg1_head_arg1_head = ATgetFirst(arg_arg1_head_arg1_list);
                                                                                                                                                                                                                                                                                                                                  arg_arg1_head_arg1_list = ATgetNext(arg_arg1_head_arg1_list);
                                                                                                                                                                                                                                                                                                                                  if (ATgetInt((ATermInt)arg_arg1_head_arg1_head) == 91) {

                                                                                                                                                                                                                                                                                                                                    if (ATgetType((ATerm)arg_arg1_head_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg1_list) == ATtrue) {
                                                                                                                                                                                                                                                                                                                                      if (ATgetType((ATerm)arg_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_list) == ATfalse) {
                                                                                                                                                                                                                                                                                                                                        arg_arg1_list = ATgetNext(arg_arg1_list);
                                                                                                                                                                                                                                                                                                                                        if (ATgetType((ATerm)arg_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_list) == ATfalse) {
                                                                                                                                                                                                                                                                                                                                          arg_arg1_head = ATgetFirst(arg_arg1_list);
                                                                                                                                                                                                                                                                                                                                          arg_arg1_list = ATgetNext(arg_arg1_list);
                                                                                                                                                                                                                                                                                                                                          /* checking for: appl */
                                                                                                                                                                                                                                                                                                                                          if (ATgetType((ATerm)arg_arg1_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head) == PERR_afun0) {
                                                                                                                                                                                                                                                                                                                                            ATerm arg_arg1_head_arg0 = ATgetArgument(arg_arg1_head, 0);
                                                                                                                                                                                                                                                                                                                                            /* checking for: list */
                                                                                                                                                                                                                                                                                                                                            if (ATgetType((ATerm)arg_arg1_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0) == PERR_afun8) {
                                                                                                                                                                                                                                                                                                                                              ATerm arg_arg1_head_arg0_arg0 = ATgetArgument(arg_arg1_head_arg0, 0);
                                                                                                                                                                                                                                                                                                                                              /* checking for: cf */
                                                                                                                                                                                                                                                                                                                                              if (ATgetType((ATerm)arg_arg1_head_arg0_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg0) == PERR_afun2) {
                                                                                                                                                                                                                                                                                                                                                ATerm arg_arg1_head_arg0_arg0_arg0 = ATgetArgument(arg_arg1_head_arg0_arg0, 0);
                                                                                                                                                                                                                                                                                                                                                /* checking for: iter-star-sep */
                                                                                                                                                                                                                                                                                                                                                if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg0_arg0) == PERR_afun46) {
                                                                                                                                                                                                                                                                                                                                                  ATerm arg_arg1_head_arg0_arg0_arg0_arg0 = ATgetArgument(arg_arg1_head_arg0_arg0_arg0, 0);
                                                                                                                                                                                                                                                                                                                                                  /* checking for: sort */
                                                                                                                                                                                                                                                                                                                                                  if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_arg0_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg0_arg0_arg0) == PERR_afun11) {
                                                                                                                                                                                                                                                                                                                                                    ATerm arg_arg1_head_arg0_arg0_arg0_arg0_arg0 = ATgetArgument(arg_arg1_head_arg0_arg0_arg0_arg0, 0);
                                                                                                                                                                                                                                                                                                                                                    /* checking for: Subject */
                                                                                                                                                                                                                                                                                                                                                    if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_arg0_arg0_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg0_arg0_arg0_arg0) == PERR_afun41) {
                                                                                                                                                                                                                                                                                                                                                      ATerm arg_arg1_head_arg0_arg0_arg0_arg1 = ATgetArgument(arg_arg1_head_arg0_arg0_arg0, 1);
                                                                                                                                                                                                                                                                                                                                                      /* checking for: lit */
                                                                                                                                                                                                                                                                                                                                                      if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_arg0_arg1) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg0_arg0_arg1) == PERR_afun9) {
                                                                                                                                                                                                                                                                                                                                                        ATerm arg_arg1_head_arg0_arg0_arg0_arg1_arg0 = ATgetArgument(arg_arg1_head_arg0_arg0_arg0_arg1, 0);
                                                                                                                                                                                                                                                                                                                                                        /* checking for: , */
                                                                                                                                                                                                                                                                                                                                                        if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_arg0_arg1_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg0_arg0_arg1_arg0) == PERR_afun43) {
                                                                                                                                                                                                                                                                                                                                                          if (ATgetType((ATerm)arg_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_list) == ATfalse) {
                                                                                                                                                                                                                                                                                                                                                            arg_arg1_list = ATgetNext(arg_arg1_list);
                                                                                                                                                                                                                                                                                                                                                            if (ATgetType((ATerm)arg_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_list) == ATfalse) {
                                                                                                                                                                                                                                                                                                                                                              arg_arg1_head = ATgetFirst(arg_arg1_list);
                                                                                                                                                                                                                                                                                                                                                              arg_arg1_list = ATgetNext(arg_arg1_list);
                                                                                                                                                                                                                                                                                                                                                              /* checking for: appl */
                                                                                                                                                                                                                                                                                                                                                              if (ATgetType((ATerm)arg_arg1_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head) == PERR_afun0) {
                                                                                                                                                                                                                                                                                                                                                                ATerm arg_arg1_head_arg0 = ATgetArgument(arg_arg1_head, 0);
                                                                                                                                                                                                                                                                                                                                                                /* checking for: prod */
                                                                                                                                                                                                                                                                                                                                                                if (ATgetType((ATerm)arg_arg1_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0) == PERR_afun1) {
                                                                                                                                                                                                                                                                                                                                                                  ATerm arg_arg1_head_arg0_arg0 = ATgetArgument(arg_arg1_head_arg0, 0);
                                                                                                                                                                                                                                                                                                                                                                  if (ATgetType((ATerm)arg_arg1_head_arg0_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0) == ATfalse) {
                                                                                                                                                                                                                                                                                                                                                                    ATerm arg_arg1_head_arg0_arg0_head;
                                                                                                                                                                                                                                                                                                                                                                    ATermList arg_arg1_head_arg0_arg0_list = (ATermList)arg_arg1_head_arg0_arg0;
                                                                                                                                                                                                                                                                                                                                                                    arg_arg1_head_arg0_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_list);
                                                                                                                                                                                                                                                                                                                                                                    arg_arg1_head_arg0_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_list);
                                                                                                                                                                                                                                                                                                                                                                    /* checking for: char-class */
                                                                                                                                                                                                                                                                                                                                                                    if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg0_head) == PERR_afun17) {
                                                                                                                                                                                                                                                                                                                                                                      ATerm arg_arg1_head_arg0_arg0_head_arg0 = ATgetArgument(arg_arg1_head_arg0_arg0_head, 0);
                                                                                                                                                                                                                                                                                                                                                                      if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0) == ATfalse) {
                                                                                                                                                                                                                                                                                                                                                                        ATerm arg_arg1_head_arg0_arg0_head_arg0_head;
                                                                                                                                                                                                                                                                                                                                                                        ATermList arg_arg1_head_arg0_arg0_head_arg0_list = (ATermList)arg_arg1_head_arg0_arg0_head_arg0;
                                                                                                                                                                                                                                                                                                                                                                        arg_arg1_head_arg0_arg0_head_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                                                                                                                                                                                                                                                                                        arg_arg1_head_arg0_arg0_head_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                                                                                                                                                                                                                                                                                        if (ATgetInt((ATermInt)arg_arg1_head_arg0_arg0_head_arg0_head) == 93) {

                                                                                                                                                                                                                                                                                                                                                                          if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0_list) == ATtrue) {
                                                                                                                                                                                                                                                                                                                                                                            if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_list) == ATtrue) {
                                                                                                                                                                                                                                                                                                                                                                              ATerm arg_arg1_head_arg0_arg1 = ATgetArgument(arg_arg1_head_arg0, 1);
                                                                                                                                                                                                                                                                                                                                                                              /* checking for: lit */
                                                                                                                                                                                                                                                                                                                                                                              if (ATgetType((ATerm)arg_arg1_head_arg0_arg1) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg1) == PERR_afun9) {
                                                                                                                                                                                                                                                                                                                                                                                ATerm arg_arg1_head_arg0_arg1_arg0 = ATgetArgument(arg_arg1_head_arg0_arg1, 0);
                                                                                                                                                                                                                                                                                                                                                                                /* checking for: ] */
                                                                                                                                                                                                                                                                                                                                                                                if (ATgetType((ATerm)arg_arg1_head_arg0_arg1_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg1_arg0) == PERR_afun45) {
                                                                                                                                                                                                                                                                                                                                                                                  ATerm arg_arg1_head_arg0_arg2 = ATgetArgument(arg_arg1_head_arg0, 2);
                                                                                                                                                                                                                                                                                                                                                                                  /* checking for: no-attrs */
                                                                                                                                                                                                                                                                                                                                                                                  if (ATgetType((ATerm)arg_arg1_head_arg0_arg2) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg2) == PERR_afun5) {
                                                                                                                                                                                                                                                                                                                                                                                    ATerm arg_arg1_head_arg1 = ATgetArgument(arg_arg1_head, 1);
                                                                                                                                                                                                                                                                                                                                                                                    if (ATgetType((ATerm)arg_arg1_head_arg1) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg1) == ATfalse) {
                                                                                                                                                                                                                                                                                                                                                                                      ATerm arg_arg1_head_arg1_head;
                                                                                                                                                                                                                                                                                                                                                                                      ATermList arg_arg1_head_arg1_list = (ATermList)arg_arg1_head_arg1;
                                                                                                                                                                                                                                                                                                                                                                                      arg_arg1_head_arg1_head = ATgetFirst(arg_arg1_head_arg1_list);
                                                                                                                                                                                                                                                                                                                                                                                      arg_arg1_head_arg1_list = ATgetNext(arg_arg1_head_arg1_list);
                                                                                                                                                                                                                                                                                                                                                                                      if (ATgetInt((ATermInt)arg_arg1_head_arg1_head) == 93) {

                                                                                                                                                                                                                                                                                                                                                                                        if (ATgetType((ATerm)arg_arg1_head_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg1_list) == ATtrue) {
                                                                                                                                                                                                                                                                                                                                                                                          if (ATgetType((ATerm)arg_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_list) == ATfalse) {
                                                                                                                                                                                                                                                                                                                                                                                            arg_arg1_list = ATgetNext(arg_arg1_list);
                                                                                                                                                                                                                                                                                                                                                                                            if (ATgetType((ATerm)arg_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_list) == ATfalse) {
                                                                                                                                                                                                                                                                                                                                                                                              arg_arg1_head = ATgetFirst(arg_arg1_list);
                                                                                                                                                                                                                                                                                                                                                                                              arg_arg1_list = ATgetNext(arg_arg1_list);
                                                                                                                                                                                                                                                                                                                                                                                              /* checking for: appl */
                                                                                                                                                                                                                                                                                                                                                                                              if (ATgetType((ATerm)arg_arg1_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head) == PERR_afun0) {
                                                                                                                                                                                                                                                                                                                                                                                                ATerm arg_arg1_head_arg0 = ATgetArgument(arg_arg1_head, 0);
                                                                                                                                                                                                                                                                                                                                                                                                /* checking for: prod */
                                                                                                                                                                                                                                                                                                                                                                                                if (ATgetType((ATerm)arg_arg1_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0) == PERR_afun1) {
                                                                                                                                                                                                                                                                                                                                                                                                  ATerm arg_arg1_head_arg0_arg0 = ATgetArgument(arg_arg1_head_arg0, 0);
                                                                                                                                                                                                                                                                                                                                                                                                  if (ATgetType((ATerm)arg_arg1_head_arg0_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0) == ATfalse) {
                                                                                                                                                                                                                                                                                                                                                                                                    ATerm arg_arg1_head_arg0_arg0_head;
                                                                                                                                                                                                                                                                                                                                                                                                    ATermList arg_arg1_head_arg0_arg0_list = (ATermList)arg_arg1_head_arg0_arg0;
                                                                                                                                                                                                                                                                                                                                                                                                    arg_arg1_head_arg0_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_list);
                                                                                                                                                                                                                                                                                                                                                                                                    arg_arg1_head_arg0_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_list);
                                                                                                                                                                                                                                                                                                                                                                                                    /* checking for: char-class */
                                                                                                                                                                                                                                                                                                                                                                                                    if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg0_head) == PERR_afun17) {
                                                                                                                                                                                                                                                                                                                                                                                                      ATerm arg_arg1_head_arg0_arg0_head_arg0 = ATgetArgument(arg_arg1_head_arg0_arg0_head, 0);
                                                                                                                                                                                                                                                                                                                                                                                                      if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0) == ATfalse) {
                                                                                                                                                                                                                                                                                                                                                                                                        ATerm arg_arg1_head_arg0_arg0_head_arg0_head;
                                                                                                                                                                                                                                                                                                                                                                                                        ATermList arg_arg1_head_arg0_arg0_head_arg0_list = (ATermList)arg_arg1_head_arg0_arg0_head_arg0;
                                                                                                                                                                                                                                                                                                                                                                                                        arg_arg1_head_arg0_arg0_head_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                                                                                                                                                                                                                                                                                                                        arg_arg1_head_arg0_arg0_head_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                                                                                                                                                                                                                                                                                                                        if (ATgetInt((ATermInt)arg_arg1_head_arg0_arg0_head_arg0_head) == 41) {

                                                                                                                                                                                                                                                                                                                                                                                                          if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0_list) == ATtrue) {
                                                                                                                                                                                                                                                                                                                                                                                                            if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_list) == ATtrue) {
                                                                                                                                                                                                                                                                                                                                                                                                              ATerm arg_arg1_head_arg0_arg1 = ATgetArgument(arg_arg1_head_arg0, 1);
                                                                                                                                                                                                                                                                                                                                                                                                              /* checking for: lit */
                                                                                                                                                                                                                                                                                                                                                                                                              if (ATgetType((ATerm)arg_arg1_head_arg0_arg1) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg1) == PERR_afun9) {
                                                                                                                                                                                                                                                                                                                                                                                                                ATerm arg_arg1_head_arg0_arg1_arg0 = ATgetArgument(arg_arg1_head_arg0_arg1, 0);
                                                                                                                                                                                                                                                                                                                                                                                                                /* checking for: ) */
                                                                                                                                                                                                                                                                                                                                                                                                                if (ATgetType((ATerm)arg_arg1_head_arg0_arg1_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg1_arg0) == PERR_afun38) {
                                                                                                                                                                                                                                                                                                                                                                                                                  ATerm arg_arg1_head_arg0_arg2 = ATgetArgument(arg_arg1_head_arg0, 2);
                                                                                                                                                                                                                                                                                                                                                                                                                  /* checking for: no-attrs */
                                                                                                                                                                                                                                                                                                                                                                                                                  if (ATgetType((ATerm)arg_arg1_head_arg0_arg2) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg2) == PERR_afun5) {
                                                                                                                                                                                                                                                                                                                                                                                                                    ATerm arg_arg1_head_arg1 = ATgetArgument(arg_arg1_head, 1);
                                                                                                                                                                                                                                                                                                                                                                                                                    if (ATgetType((ATerm)arg_arg1_head_arg1) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg1) == ATfalse) {
                                                                                                                                                                                                                                                                                                                                                                                                                      ATerm arg_arg1_head_arg1_head;
                                                                                                                                                                                                                                                                                                                                                                                                                      ATermList arg_arg1_head_arg1_list = (ATermList)arg_arg1_head_arg1;
                                                                                                                                                                                                                                                                                                                                                                                                                      arg_arg1_head_arg1_head = ATgetFirst(arg_arg1_head_arg1_list);
                                                                                                                                                                                                                                                                                                                                                                                                                      arg_arg1_head_arg1_list = ATgetNext(arg_arg1_head_arg1_list);
                                                                                                                                                                                                                                                                                                                                                                                                                      if (ATgetInt((ATermInt)arg_arg1_head_arg1_head) == 41) {

                                                                                                                                                                                                                                                                                                                                                                                                                        if (ATgetType((ATerm)arg_arg1_head_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg1_list) == ATtrue) {
                                                                                                                                                                                                                                                                                                                                                                                                                          if (ATgetType((ATerm)arg_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_list) == ATtrue) {
                                                                                                                                                                                                                                                                                                                                                                                                                            return ATtrue;
                                                                                                                                                                                                                                                                                                                                                                                                                          }
                                                                                                                                                                                                                                                                                                                                                                                                                        }
                                                                                                                                                                                                                                                                                                                                                                                                                      }
                                                                                                                                                                                                                                                                                                                                                                                                                    }
                                                                                                                                                                                                                                                                                                                                                                                                                  }
                                                                                                                                                                                                                                                                                                                                                                                                                }
                                                                                                                                                                                                                                                                                                                                                                                                              }
                                                                                                                                                                                                                                                                                                                                                                                                            }
                                                                                                                                                                                                                                                                                                                                                                                                          }
                                                                                                                                                                                                                                                                                                                                                                                                        }
                                                                                                                                                                                                                                                                                                                                                                                                      }
                                                                                                                                                                                                                                                                                                                                                                                                    }
                                                                                                                                                                                                                                                                                                                                                                                                  }
                                                                                                                                                                                                                                                                                                                                                                                                }
                                                                                                                                                                                                                                                                                                                                                                                              }
                                                                                                                                                                                                                                                                                                                                                                                            }
                                                                                                                                                                                                                                                                                                                                                                                          }
                                                                                                                                                                                                                                                                                                                                                                                        }
                                                                                                                                                                                                                                                                                                                                                                                      }
                                                                                                                                                                                                                                                                                                                                                                                    }
                                                                                                                                                                                                                                                                                                                                                                                  }
                                                                                                                                                                                                                                                                                                                                                                                }
                                                                                                                                                                                                                                                                                                                                                                              }
                                                                                                                                                                                                                                                                                                                                                                            }
                                                                                                                                                                                                                                                                                                                                                                          }
                                                                                                                                                                                                                                                                                                                                                                        }
                                                                                                                                                                                                                                                                                                                                                                      }
                                                                                                                                                                                                                                                                                                                                                                    }
                                                                                                                                                                                                                                                                                                                                                                  }
                                                                                                                                                                                                                                                                                                                                                                }
                                                                                                                                                                                                                                                                                                                                                              }
                                                                                                                                                                                                                                                                                                                                                            }
                                                                                                                                                                                                                                                                                                                                                          }
                                                                                                                                                                                                                                                                                                                                                        }
                                                                                                                                                                                                                                                                                                                                                      }
                                                                                                                                                                                                                                                                                                                                                    }
                                                                                                                                                                                                                                                                                                                                                  }
                                                                                                                                                                                                                                                                                                                                                }
                                                                                                                                                                                                                                                                                                                                              }
                                                                                                                                                                                                                                                                                                                                            }
                                                                                                                                                                                                                                                                                                                                          }
                                                                                                                                                                                                                                                                                                                                        }
                                                                                                                                                                                                                                                                                                                                      }
                                                                                                                                                                                                                                                                                                                                    }
                                                                                                                                                                                                                                                                                                                                  }
                                                                                                                                                                                                                                                                                                                                }
                                                                                                                                                                                                                                                                                                                              }
                                                                                                                                                                                                                                                                                                                            }
                                                                                                                                                                                                                                                                                                                          }
                                                                                                                                                                                                                                                                                                                        }
                                                                                                                                                                                                                                                                                                                      }
                                                                                                                                                                                                                                                                                                                    }
                                                                                                                                                                                                                                                                                                                  }
                                                                                                                                                                                                                                                                                                                }
                                                                                                                                                                                                                                                                                                              }
                                                                                                                                                                                                                                                                                                            }
                                                                                                                                                                                                                                                                                                          }
                                                                                                                                                                                                                                                                                                        }
                                                                                                                                                                                                                                                                                                      }
                                                                                                                                                                                                                                                                                                    }
                                                                                                                                                                                                                                                                                                  }
                                                                                                                                                                                                                                                                                                }
                                                                                                                                                                                                                                                                                              }
                                                                                                                                                                                                                                                                                            }
                                                                                                                                                                                                                                                                                          }
                                                                                                                                                                                                                                                                                        }
                                                                                                                                                                                                                                                                                      }
                                                                                                                                                                                                                                                                                    }
                                                                                                                                                                                                                                                                                  }
                                                                                                                                                                                                                                                                                }
                                                                                                                                                                                                                                                                              }
                                                                                                                                                                                                                                                                            }
                                                                                                                                                                                                                                                                          }
                                                                                                                                                                                                                                                                        }
                                                                                                                                                                                                                                                                      }
                                                                                                                                                                                                                                                                    }
                                                                                                                                                                                                                                                                  }
                                                                                                                                                                                                                                                                }
                                                                                                                                                                                                                                                              }
                                                                                                                                                                                                                                                            }
                                                                                                                                                                                                                                                          }
                                                                                                                                                                                                                                                        }
                                                                                                                                                                                                                                                      }
                                                                                                                                                                                                                                                    }
                                                                                                                                                                                                                                                  }
                                                                                                                                                                                                                                                }
                                                                                                                                                                                                                                              }
                                                                                                                                                                                                                                            }
                                                                                                                                                                                                                                          }
                                                                                                                                                                                                                                        }
                                                                                                                                                                                                                                      }
                                                                                                                                                                                                                                    }
                                                                                                                                                                                                                                  }
                                                                                                                                                                                                                                }
                                                                                                                                                                                                                              }
                                                                                                                                                                                                                            }
                                                                                                                                                                                                                          }
                                                                                                                                                                                                                        }
                                                                                                                                                                                                                      }
                                                                                                                                                                                                                    }
                                                                                                                                                                                                                  }
                                                                                                                                                                                                                }
                                                                                                                                                                                                              }
                                                                                                                                                                                                            }
                                                                                                                                                                                                          }
                                                                                                                                                                                                        }
                                                                                                                                                                                                      }
                                                                                                                                                                                                    }
                                                                                                                                                                                                  }
                                                                                                                                                                                                }
                                                                                                                                                                                              }
                                                                                                                                                                                            }
                                                                                                                                                                                          }
                                                                                                                                                                                        }
                                                                                                                                                                                      }
                                                                                                                                                                                    }
                                                                                                                                                                                  }
                                                                                                                                                                                }
                                                                                                                                                                              }
                                                                                                                                                                            }
                                                                                                                                                                          }
                                                                                                                                                                        }
                                                                                                                                                                      }
                                                                                                                                                                    }
                                                                                                                                                                  }
                                                                                                                                                                }
                                                                                                                                                              }
                                                                                                                                                            }
                                                                                                                                                          }
                                                                                                                                                        }
                                                                                                                                                      }
                                                                                                                                                    }
                                                                                                                                                  }
                                                                                                                                                }
                                                                                                                                              }
                                                                                                                                            }
                                                                                                                                          }
                                                                                                                                        }
                                                                                                                                      }
                                                                                                                                    }
                                                                                                                                  }
                                                                                                                                }
                                                                                                                              }
                                                                                                                            }
                                                                                                                          }
                                                                                                                        }
                                                                                                                      }
                                                                                                                    }
                                                                                                                  }
                                                                                                                }
                                                                                                              }
                                                                                                            }
                                                                                                          }
                                                                                                        }
                                                                                                      }
                                                                                                    }
                                                                                                  }
                                                                                                }
                                                                                              }
                                                                                            }
                                                                                          }
                                                                                        }
                                                                                      }
                                                                                    }
                                                                                  }
                                                                                }
                                                                              }
                                                                            }
                                                                          }
                                                                        }
                                                                      }
                                                                    }
                                                                  }
                                                                }
                                                              }
                                                            }
                                                          }
                                                        }
                                                      }
                                                    }
                                                  }
                                                }
                                              }
                                            }
                                          }
                                        }
                                      }
                                    }
                                  }
                                }
                              }
                            }
                          }
                        }
                      }
                    }
                  }
                }
              }
            }
          }
        }
      }
    }
  }
  return ATfalse;
}

/**
 * Assert whether a PERR_Error has a ws-after-info. 
 * \param[in] arg input PERR_Error
 * \return ATtrue if the PERR_Error had a ws-after-info, or ATfalse otherwise
 */
ATbool PERR_hasErrorWsAfterInfo(PERR_Error arg) {
  if (PERR_isErrorInfo(arg)) {
    return ATtrue;
  }
  return ATfalse;
}

/**
 * Assert whether a PERR_Error has a ws-after-(. 
 * \param[in] arg input PERR_Error
 * \return ATtrue if the PERR_Error had a ws-after-(, or ATfalse otherwise
 */
ATbool PERR_hasErrorWsAfterParenOpen(PERR_Error arg) {
  if (PERR_isErrorInfo(arg)) {
    return ATtrue;
  }
  else if (PERR_isErrorWarning(arg)) {
    return ATtrue;
  }
  else if (PERR_isErrorError(arg)) {
    return ATtrue;
  }
  else if (PERR_isErrorFatal(arg)) {
    return ATtrue;
  }
  return ATfalse;
}

/**
 * Assert whether a PERR_Error has a description. 
 * \param[in] arg input PERR_Error
 * \return ATtrue if the PERR_Error had a description, or ATfalse otherwise
 */
ATbool PERR_hasErrorDescription(PERR_Error arg) {
  if (PERR_isErrorInfo(arg)) {
    return ATtrue;
  }
  else if (PERR_isErrorWarning(arg)) {
    return ATtrue;
  }
  else if (PERR_isErrorError(arg)) {
    return ATtrue;
  }
  else if (PERR_isErrorFatal(arg)) {
    return ATtrue;
  }
  return ATfalse;
}

/**
 * Assert whether a PERR_Error has a ws-after-description. 
 * \param[in] arg input PERR_Error
 * \return ATtrue if the PERR_Error had a ws-after-description, or ATfalse otherwise
 */
ATbool PERR_hasErrorWsAfterDescription(PERR_Error arg) {
  if (PERR_isErrorInfo(arg)) {
    return ATtrue;
  }
  else if (PERR_isErrorWarning(arg)) {
    return ATtrue;
  }
  else if (PERR_isErrorError(arg)) {
    return ATtrue;
  }
  else if (PERR_isErrorFatal(arg)) {
    return ATtrue;
  }
  return ATfalse;
}

/**
 * Assert whether a PERR_Error has a ws-after-,. 
 * \param[in] arg input PERR_Error
 * \return ATtrue if the PERR_Error had a ws-after-,, or ATfalse otherwise
 */
ATbool PERR_hasErrorWsAfterComma(PERR_Error arg) {
  if (PERR_isErrorInfo(arg)) {
    return ATtrue;
  }
  else if (PERR_isErrorWarning(arg)) {
    return ATtrue;
  }
  else if (PERR_isErrorError(arg)) {
    return ATtrue;
  }
  else if (PERR_isErrorFatal(arg)) {
    return ATtrue;
  }
  return ATfalse;
}

/**
 * Assert whether a PERR_Error has a ws-after-[. 
 * \param[in] arg input PERR_Error
 * \return ATtrue if the PERR_Error had a ws-after-[, or ATfalse otherwise
 */
ATbool PERR_hasErrorWsAfterBracketOpen(PERR_Error arg) {
  if (PERR_isErrorInfo(arg)) {
    return ATtrue;
  }
  else if (PERR_isErrorWarning(arg)) {
    return ATtrue;
  }
  else if (PERR_isErrorError(arg)) {
    return ATtrue;
  }
  else if (PERR_isErrorFatal(arg)) {
    return ATtrue;
  }
  return ATfalse;
}

/**
 * Assert whether a PERR_Error has a list. 
 * \param[in] arg input PERR_Error
 * \return ATtrue if the PERR_Error had a list, or ATfalse otherwise
 */
ATbool PERR_hasErrorList(PERR_Error arg) {
  if (PERR_isErrorInfo(arg)) {
    return ATtrue;
  }
  else if (PERR_isErrorWarning(arg)) {
    return ATtrue;
  }
  else if (PERR_isErrorError(arg)) {
    return ATtrue;
  }
  else if (PERR_isErrorFatal(arg)) {
    return ATtrue;
  }
  return ATfalse;
}

/**
 * Assert whether a PERR_Error has a ws-after-list. 
 * \param[in] arg input PERR_Error
 * \return ATtrue if the PERR_Error had a ws-after-list, or ATfalse otherwise
 */
ATbool PERR_hasErrorWsAfterList(PERR_Error arg) {
  if (PERR_isErrorInfo(arg)) {
    return ATtrue;
  }
  else if (PERR_isErrorWarning(arg)) {
    return ATtrue;
  }
  else if (PERR_isErrorError(arg)) {
    return ATtrue;
  }
  else if (PERR_isErrorFatal(arg)) {
    return ATtrue;
  }
  return ATfalse;
}

/**
 * Assert whether a PERR_Error has a ws-after-]. 
 * \param[in] arg input PERR_Error
 * \return ATtrue if the PERR_Error had a ws-after-], or ATfalse otherwise
 */
ATbool PERR_hasErrorWsAfterBracketClose(PERR_Error arg) {
  if (PERR_isErrorInfo(arg)) {
    return ATtrue;
  }
  else if (PERR_isErrorWarning(arg)) {
    return ATtrue;
  }
  else if (PERR_isErrorError(arg)) {
    return ATtrue;
  }
  else if (PERR_isErrorFatal(arg)) {
    return ATtrue;
  }
  return ATfalse;
}

/**
 * Assert whether a PERR_Error has a ws-after-warning. 
 * \param[in] arg input PERR_Error
 * \return ATtrue if the PERR_Error had a ws-after-warning, or ATfalse otherwise
 */
ATbool PERR_hasErrorWsAfterWarning(PERR_Error arg) {
  if (PERR_isErrorWarning(arg)) {
    return ATtrue;
  }
  return ATfalse;
}

/**
 * Assert whether a PERR_Error has a ws-after-error. 
 * \param[in] arg input PERR_Error
 * \return ATtrue if the PERR_Error had a ws-after-error, or ATfalse otherwise
 */
ATbool PERR_hasErrorWsAfterError(PERR_Error arg) {
  if (PERR_isErrorError(arg)) {
    return ATtrue;
  }
  return ATfalse;
}

/**
 * Assert whether a PERR_Error has a ws-after-fatal. 
 * \param[in] arg input PERR_Error
 * \return ATtrue if the PERR_Error had a ws-after-fatal, or ATfalse otherwise
 */
ATbool PERR_hasErrorWsAfterFatal(PERR_Error arg) {
  if (PERR_isErrorFatal(arg)) {
    return ATtrue;
  }
  return ATfalse;
}

/**
 * Get the ws-after-info PERR_OptLayout of a PERR_Error. Note that the precondition is that this PERR_Error actually has a ws-after-info
 * \param[in] arg input PERR_Error
 * \return the ws-after-info of #arg, if it exist or an undefined value if it does not
 */
PERR_OptLayout PERR_getErrorWsAfterInfo(PERR_Error arg) {
  
    return (PERR_OptLayout)ATelementAt((ATermList)ATgetArgument((ATermAppl)arg, 1), 1);
}

/**
 * Get the ws-after-( PERR_OptLayout of a PERR_Error. Note that the precondition is that this PERR_Error actually has a ws-after-(
 * \param[in] arg input PERR_Error
 * \return the ws-after-( of #arg, if it exist or an undefined value if it does not
 */
PERR_OptLayout PERR_getErrorWsAfterParenOpen(PERR_Error arg) {
  if (PERR_isErrorInfo(arg)) {
    return (PERR_OptLayout)ATelementAt((ATermList)ATgetArgument((ATermAppl)arg, 1), 3);
  }
  else if (PERR_isErrorWarning(arg)) {
    return (PERR_OptLayout)ATelementAt((ATermList)ATgetArgument((ATermAppl)arg, 1), 3);
  }
  else if (PERR_isErrorError(arg)) {
    return (PERR_OptLayout)ATelementAt((ATermList)ATgetArgument((ATermAppl)arg, 1), 3);
  }
  else 
    return (PERR_OptLayout)ATelementAt((ATermList)ATgetArgument((ATermAppl)arg, 1), 3);
}

/**
 * Get the description PERR_StrCon of a PERR_Error. Note that the precondition is that this PERR_Error actually has a description
 * \param[in] arg input PERR_Error
 * \return the description of #arg, if it exist or an undefined value if it does not
 */
PERR_StrCon PERR_getErrorDescription(PERR_Error arg) {
  if (PERR_isErrorInfo(arg)) {
    return (PERR_StrCon)ATelementAt((ATermList)ATgetArgument((ATermAppl)arg, 1), 4);
  }
  else if (PERR_isErrorWarning(arg)) {
    return (PERR_StrCon)ATelementAt((ATermList)ATgetArgument((ATermAppl)arg, 1), 4);
  }
  else if (PERR_isErrorError(arg)) {
    return (PERR_StrCon)ATelementAt((ATermList)ATgetArgument((ATermAppl)arg, 1), 4);
  }
  else 
    return (PERR_StrCon)ATelementAt((ATermList)ATgetArgument((ATermAppl)arg, 1), 4);
}

/**
 * Get the ws-after-description PERR_OptLayout of a PERR_Error. Note that the precondition is that this PERR_Error actually has a ws-after-description
 * \param[in] arg input PERR_Error
 * \return the ws-after-description of #arg, if it exist or an undefined value if it does not
 */
PERR_OptLayout PERR_getErrorWsAfterDescription(PERR_Error arg) {
  if (PERR_isErrorInfo(arg)) {
    return (PERR_OptLayout)ATelementAt((ATermList)ATgetArgument((ATermAppl)arg, 1), 5);
  }
  else if (PERR_isErrorWarning(arg)) {
    return (PERR_OptLayout)ATelementAt((ATermList)ATgetArgument((ATermAppl)arg, 1), 5);
  }
  else if (PERR_isErrorError(arg)) {
    return (PERR_OptLayout)ATelementAt((ATermList)ATgetArgument((ATermAppl)arg, 1), 5);
  }
  else 
    return (PERR_OptLayout)ATelementAt((ATermList)ATgetArgument((ATermAppl)arg, 1), 5);
}

/**
 * Get the ws-after-, PERR_OptLayout of a PERR_Error. Note that the precondition is that this PERR_Error actually has a ws-after-,
 * \param[in] arg input PERR_Error
 * \return the ws-after-, of #arg, if it exist or an undefined value if it does not
 */
PERR_OptLayout PERR_getErrorWsAfterComma(PERR_Error arg) {
  if (PERR_isErrorInfo(arg)) {
    return (PERR_OptLayout)ATelementAt((ATermList)ATgetArgument((ATermAppl)arg, 1), 7);
  }
  else if (PERR_isErrorWarning(arg)) {
    return (PERR_OptLayout)ATelementAt((ATermList)ATgetArgument((ATermAppl)arg, 1), 7);
  }
  else if (PERR_isErrorError(arg)) {
    return (PERR_OptLayout)ATelementAt((ATermList)ATgetArgument((ATermAppl)arg, 1), 7);
  }
  else 
    return (PERR_OptLayout)ATelementAt((ATermList)ATgetArgument((ATermAppl)arg, 1), 7);
}

/**
 * Get the ws-after-[ PERR_OptLayout of a PERR_Error. Note that the precondition is that this PERR_Error actually has a ws-after-[
 * \param[in] arg input PERR_Error
 * \return the ws-after-[ of #arg, if it exist or an undefined value if it does not
 */
PERR_OptLayout PERR_getErrorWsAfterBracketOpen(PERR_Error arg) {
  if (PERR_isErrorInfo(arg)) {
    return (PERR_OptLayout)ATelementAt((ATermList)ATgetArgument((ATermAppl)arg, 1), 9);
  }
  else if (PERR_isErrorWarning(arg)) {
    return (PERR_OptLayout)ATelementAt((ATermList)ATgetArgument((ATermAppl)arg, 1), 9);
  }
  else if (PERR_isErrorError(arg)) {
    return (PERR_OptLayout)ATelementAt((ATermList)ATgetArgument((ATermAppl)arg, 1), 9);
  }
  else 
    return (PERR_OptLayout)ATelementAt((ATermList)ATgetArgument((ATermAppl)arg, 1), 9);
}

/**
 * Get the list PERR_SubjectList of a PERR_Error. Note that the precondition is that this PERR_Error actually has a list
 * \param[in] arg input PERR_Error
 * \return the list of #arg, if it exist or an undefined value if it does not
 */
PERR_SubjectList PERR_getErrorList(PERR_Error arg) {
  if (PERR_isErrorInfo(arg)) {
    return (PERR_SubjectList)ATgetArgument((ATermAppl)ATelementAt((ATermList)ATgetArgument((ATermAppl)arg, 1), 10), 1);
  }
  else if (PERR_isErrorWarning(arg)) {
    return (PERR_SubjectList)ATgetArgument((ATermAppl)ATelementAt((ATermList)ATgetArgument((ATermAppl)arg, 1), 10), 1);
  }
  else if (PERR_isErrorError(arg)) {
    return (PERR_SubjectList)ATgetArgument((ATermAppl)ATelementAt((ATermList)ATgetArgument((ATermAppl)arg, 1), 10), 1);
  }
  else 
    return (PERR_SubjectList)ATgetArgument((ATermAppl)ATelementAt((ATermList)ATgetArgument((ATermAppl)arg, 1), 10), 1);
}

/**
 * Get the ws-after-list PERR_OptLayout of a PERR_Error. Note that the precondition is that this PERR_Error actually has a ws-after-list
 * \param[in] arg input PERR_Error
 * \return the ws-after-list of #arg, if it exist or an undefined value if it does not
 */
PERR_OptLayout PERR_getErrorWsAfterList(PERR_Error arg) {
  if (PERR_isErrorInfo(arg)) {
    return (PERR_OptLayout)ATelementAt((ATermList)ATgetArgument((ATermAppl)arg, 1), 11);
  }
  else if (PERR_isErrorWarning(arg)) {
    return (PERR_OptLayout)ATelementAt((ATermList)ATgetArgument((ATermAppl)arg, 1), 11);
  }
  else if (PERR_isErrorError(arg)) {
    return (PERR_OptLayout)ATelementAt((ATermList)ATgetArgument((ATermAppl)arg, 1), 11);
  }
  else 
    return (PERR_OptLayout)ATelementAt((ATermList)ATgetArgument((ATermAppl)arg, 1), 11);
}

/**
 * Get the ws-after-] PERR_OptLayout of a PERR_Error. Note that the precondition is that this PERR_Error actually has a ws-after-]
 * \param[in] arg input PERR_Error
 * \return the ws-after-] of #arg, if it exist or an undefined value if it does not
 */
PERR_OptLayout PERR_getErrorWsAfterBracketClose(PERR_Error arg) {
  if (PERR_isErrorInfo(arg)) {
    return (PERR_OptLayout)ATelementAt((ATermList)ATgetArgument((ATermAppl)arg, 1), 13);
  }
  else if (PERR_isErrorWarning(arg)) {
    return (PERR_OptLayout)ATelementAt((ATermList)ATgetArgument((ATermAppl)arg, 1), 13);
  }
  else if (PERR_isErrorError(arg)) {
    return (PERR_OptLayout)ATelementAt((ATermList)ATgetArgument((ATermAppl)arg, 1), 13);
  }
  else 
    return (PERR_OptLayout)ATelementAt((ATermList)ATgetArgument((ATermAppl)arg, 1), 13);
}

/**
 * Get the ws-after-warning PERR_OptLayout of a PERR_Error. Note that the precondition is that this PERR_Error actually has a ws-after-warning
 * \param[in] arg input PERR_Error
 * \return the ws-after-warning of #arg, if it exist or an undefined value if it does not
 */
PERR_OptLayout PERR_getErrorWsAfterWarning(PERR_Error arg) {
  
    return (PERR_OptLayout)ATelementAt((ATermList)ATgetArgument((ATermAppl)arg, 1), 1);
}

/**
 * Get the ws-after-error PERR_OptLayout of a PERR_Error. Note that the precondition is that this PERR_Error actually has a ws-after-error
 * \param[in] arg input PERR_Error
 * \return the ws-after-error of #arg, if it exist or an undefined value if it does not
 */
PERR_OptLayout PERR_getErrorWsAfterError(PERR_Error arg) {
  
    return (PERR_OptLayout)ATelementAt((ATermList)ATgetArgument((ATermAppl)arg, 1), 1);
}

/**
 * Get the ws-after-fatal PERR_OptLayout of a PERR_Error. Note that the precondition is that this PERR_Error actually has a ws-after-fatal
 * \param[in] arg input PERR_Error
 * \return the ws-after-fatal of #arg, if it exist or an undefined value if it does not
 */
PERR_OptLayout PERR_getErrorWsAfterFatal(PERR_Error arg) {
  
    return (PERR_OptLayout)ATelementAt((ATermList)ATgetArgument((ATermAppl)arg, 1), 1);
}

/**
 * Set the ws-after-info of a PERR_Error. The precondition being that this PERR_Error actually has a ws-after-info
 * \param[in] arg input PERR_Error
 * \param[in] wsAfterInfo new PERR_OptLayout to set in #arg
 * \return A new PERR_Error with wsAfterInfo at the right place, or a core dump if #arg did not have a wsAfterInfo
 */
PERR_Error PERR_setErrorWsAfterInfo(PERR_Error arg, PERR_OptLayout wsAfterInfo) {
  if (PERR_isErrorInfo(arg)) {
    return (PERR_Error)ATsetArgument((ATermAppl)arg, (ATerm)ATreplace((ATermList)ATgetArgument((ATermAppl)arg, 1), (ATerm)((ATerm) wsAfterInfo), 1), 1);
  }

  ATabort("Error has no WsAfterInfo: %t\n", arg);
  return (PERR_Error)NULL;
}

/**
 * Set the ws-after-( of a PERR_Error. The precondition being that this PERR_Error actually has a ws-after-(
 * \param[in] arg input PERR_Error
 * \param[in] wsAfterParenOpen new PERR_OptLayout to set in #arg
 * \return A new PERR_Error with wsAfterParenOpen at the right place, or a core dump if #arg did not have a wsAfterParenOpen
 */
PERR_Error PERR_setErrorWsAfterParenOpen(PERR_Error arg, PERR_OptLayout wsAfterParenOpen) {
  if (PERR_isErrorInfo(arg)) {
    return (PERR_Error)ATsetArgument((ATermAppl)arg, (ATerm)ATreplace((ATermList)ATgetArgument((ATermAppl)arg, 1), (ATerm)((ATerm) wsAfterParenOpen), 3), 1);
  }
  else if (PERR_isErrorWarning(arg)) {
    return (PERR_Error)ATsetArgument((ATermAppl)arg, (ATerm)ATreplace((ATermList)ATgetArgument((ATermAppl)arg, 1), (ATerm)((ATerm) wsAfterParenOpen), 3), 1);
  }
  else if (PERR_isErrorError(arg)) {
    return (PERR_Error)ATsetArgument((ATermAppl)arg, (ATerm)ATreplace((ATermList)ATgetArgument((ATermAppl)arg, 1), (ATerm)((ATerm) wsAfterParenOpen), 3), 1);
  }
  else if (PERR_isErrorFatal(arg)) {
    return (PERR_Error)ATsetArgument((ATermAppl)arg, (ATerm)ATreplace((ATermList)ATgetArgument((ATermAppl)arg, 1), (ATerm)((ATerm) wsAfterParenOpen), 3), 1);
  }

  ATabort("Error has no WsAfterParenOpen: %t\n", arg);
  return (PERR_Error)NULL;
}

/**
 * Set the description of a PERR_Error. The precondition being that this PERR_Error actually has a description
 * \param[in] arg input PERR_Error
 * \param[in] description new PERR_StrCon to set in #arg
 * \return A new PERR_Error with description at the right place, or a core dump if #arg did not have a description
 */
PERR_Error PERR_setErrorDescription(PERR_Error arg, PERR_StrCon description) {
  if (PERR_isErrorInfo(arg)) {
    return (PERR_Error)ATsetArgument((ATermAppl)arg, (ATerm)ATreplace((ATermList)ATgetArgument((ATermAppl)arg, 1), (ATerm)((ATerm) description), 4), 1);
  }
  else if (PERR_isErrorWarning(arg)) {
    return (PERR_Error)ATsetArgument((ATermAppl)arg, (ATerm)ATreplace((ATermList)ATgetArgument((ATermAppl)arg, 1), (ATerm)((ATerm) description), 4), 1);
  }
  else if (PERR_isErrorError(arg)) {
    return (PERR_Error)ATsetArgument((ATermAppl)arg, (ATerm)ATreplace((ATermList)ATgetArgument((ATermAppl)arg, 1), (ATerm)((ATerm) description), 4), 1);
  }
  else if (PERR_isErrorFatal(arg)) {
    return (PERR_Error)ATsetArgument((ATermAppl)arg, (ATerm)ATreplace((ATermList)ATgetArgument((ATermAppl)arg, 1), (ATerm)((ATerm) description), 4), 1);
  }

  ATabort("Error has no Description: %t\n", arg);
  return (PERR_Error)NULL;
}

/**
 * Set the ws-after-description of a PERR_Error. The precondition being that this PERR_Error actually has a ws-after-description
 * \param[in] arg input PERR_Error
 * \param[in] wsAfterDescription new PERR_OptLayout to set in #arg
 * \return A new PERR_Error with wsAfterDescription at the right place, or a core dump if #arg did not have a wsAfterDescription
 */
PERR_Error PERR_setErrorWsAfterDescription(PERR_Error arg, PERR_OptLayout wsAfterDescription) {
  if (PERR_isErrorInfo(arg)) {
    return (PERR_Error)ATsetArgument((ATermAppl)arg, (ATerm)ATreplace((ATermList)ATgetArgument((ATermAppl)arg, 1), (ATerm)((ATerm) wsAfterDescription), 5), 1);
  }
  else if (PERR_isErrorWarning(arg)) {
    return (PERR_Error)ATsetArgument((ATermAppl)arg, (ATerm)ATreplace((ATermList)ATgetArgument((ATermAppl)arg, 1), (ATerm)((ATerm) wsAfterDescription), 5), 1);
  }
  else if (PERR_isErrorError(arg)) {
    return (PERR_Error)ATsetArgument((ATermAppl)arg, (ATerm)ATreplace((ATermList)ATgetArgument((ATermAppl)arg, 1), (ATerm)((ATerm) wsAfterDescription), 5), 1);
  }
  else if (PERR_isErrorFatal(arg)) {
    return (PERR_Error)ATsetArgument((ATermAppl)arg, (ATerm)ATreplace((ATermList)ATgetArgument((ATermAppl)arg, 1), (ATerm)((ATerm) wsAfterDescription), 5), 1);
  }

  ATabort("Error has no WsAfterDescription: %t\n", arg);
  return (PERR_Error)NULL;
}

/**
 * Set the ws-after-, of a PERR_Error. The precondition being that this PERR_Error actually has a ws-after-,
 * \param[in] arg input PERR_Error
 * \param[in] wsAfterComma new PERR_OptLayout to set in #arg
 * \return A new PERR_Error with wsAfterComma at the right place, or a core dump if #arg did not have a wsAfterComma
 */
PERR_Error PERR_setErrorWsAfterComma(PERR_Error arg, PERR_OptLayout wsAfterComma) {
  if (PERR_isErrorInfo(arg)) {
    return (PERR_Error)ATsetArgument((ATermAppl)arg, (ATerm)ATreplace((ATermList)ATgetArgument((ATermAppl)arg, 1), (ATerm)((ATerm) wsAfterComma), 7), 1);
  }
  else if (PERR_isErrorWarning(arg)) {
    return (PERR_Error)ATsetArgument((ATermAppl)arg, (ATerm)ATreplace((ATermList)ATgetArgument((ATermAppl)arg, 1), (ATerm)((ATerm) wsAfterComma), 7), 1);
  }
  else if (PERR_isErrorError(arg)) {
    return (PERR_Error)ATsetArgument((ATermAppl)arg, (ATerm)ATreplace((ATermList)ATgetArgument((ATermAppl)arg, 1), (ATerm)((ATerm) wsAfterComma), 7), 1);
  }
  else if (PERR_isErrorFatal(arg)) {
    return (PERR_Error)ATsetArgument((ATermAppl)arg, (ATerm)ATreplace((ATermList)ATgetArgument((ATermAppl)arg, 1), (ATerm)((ATerm) wsAfterComma), 7), 1);
  }

  ATabort("Error has no WsAfterComma: %t\n", arg);
  return (PERR_Error)NULL;
}

/**
 * Set the ws-after-[ of a PERR_Error. The precondition being that this PERR_Error actually has a ws-after-[
 * \param[in] arg input PERR_Error
 * \param[in] wsAfterBracketOpen new PERR_OptLayout to set in #arg
 * \return A new PERR_Error with wsAfterBracketOpen at the right place, or a core dump if #arg did not have a wsAfterBracketOpen
 */
PERR_Error PERR_setErrorWsAfterBracketOpen(PERR_Error arg, PERR_OptLayout wsAfterBracketOpen) {
  if (PERR_isErrorInfo(arg)) {
    return (PERR_Error)ATsetArgument((ATermAppl)arg, (ATerm)ATreplace((ATermList)ATgetArgument((ATermAppl)arg, 1), (ATerm)((ATerm) wsAfterBracketOpen), 9), 1);
  }
  else if (PERR_isErrorWarning(arg)) {
    return (PERR_Error)ATsetArgument((ATermAppl)arg, (ATerm)ATreplace((ATermList)ATgetArgument((ATermAppl)arg, 1), (ATerm)((ATerm) wsAfterBracketOpen), 9), 1);
  }
  else if (PERR_isErrorError(arg)) {
    return (PERR_Error)ATsetArgument((ATermAppl)arg, (ATerm)ATreplace((ATermList)ATgetArgument((ATermAppl)arg, 1), (ATerm)((ATerm) wsAfterBracketOpen), 9), 1);
  }
  else if (PERR_isErrorFatal(arg)) {
    return (PERR_Error)ATsetArgument((ATermAppl)arg, (ATerm)ATreplace((ATermList)ATgetArgument((ATermAppl)arg, 1), (ATerm)((ATerm) wsAfterBracketOpen), 9), 1);
  }

  ATabort("Error has no WsAfterBracketOpen: %t\n", arg);
  return (PERR_Error)NULL;
}

/**
 * Set the list of a PERR_Error. The precondition being that this PERR_Error actually has a list
 * \param[in] arg input PERR_Error
 * \param[in] list new PERR_SubjectList to set in #arg
 * \return A new PERR_Error with list at the right place, or a core dump if #arg did not have a list
 */
PERR_Error PERR_setErrorList(PERR_Error arg, PERR_SubjectList list) {
  if (PERR_isErrorInfo(arg)) {
    return (PERR_Error)ATsetArgument((ATermAppl)arg, (ATerm)ATreplace((ATermList)ATgetArgument((ATermAppl)arg, 1), (ATerm)ATsetArgument((ATermAppl)ATelementAt((ATermList)ATgetArgument((ATermAppl)arg, 1), 10), (ATerm)((ATerm) list), 1), 10), 1);
  }
  else if (PERR_isErrorWarning(arg)) {
    return (PERR_Error)ATsetArgument((ATermAppl)arg, (ATerm)ATreplace((ATermList)ATgetArgument((ATermAppl)arg, 1), (ATerm)ATsetArgument((ATermAppl)ATelementAt((ATermList)ATgetArgument((ATermAppl)arg, 1), 10), (ATerm)((ATerm) list), 1), 10), 1);
  }
  else if (PERR_isErrorError(arg)) {
    return (PERR_Error)ATsetArgument((ATermAppl)arg, (ATerm)ATreplace((ATermList)ATgetArgument((ATermAppl)arg, 1), (ATerm)ATsetArgument((ATermAppl)ATelementAt((ATermList)ATgetArgument((ATermAppl)arg, 1), 10), (ATerm)((ATerm) list), 1), 10), 1);
  }
  else if (PERR_isErrorFatal(arg)) {
    return (PERR_Error)ATsetArgument((ATermAppl)arg, (ATerm)ATreplace((ATermList)ATgetArgument((ATermAppl)arg, 1), (ATerm)ATsetArgument((ATermAppl)ATelementAt((ATermList)ATgetArgument((ATermAppl)arg, 1), 10), (ATerm)((ATerm) list), 1), 10), 1);
  }

  ATabort("Error has no List: %t\n", arg);
  return (PERR_Error)NULL;
}

/**
 * Set the ws-after-list of a PERR_Error. The precondition being that this PERR_Error actually has a ws-after-list
 * \param[in] arg input PERR_Error
 * \param[in] wsAfterList new PERR_OptLayout to set in #arg
 * \return A new PERR_Error with wsAfterList at the right place, or a core dump if #arg did not have a wsAfterList
 */
PERR_Error PERR_setErrorWsAfterList(PERR_Error arg, PERR_OptLayout wsAfterList) {
  if (PERR_isErrorInfo(arg)) {
    return (PERR_Error)ATsetArgument((ATermAppl)arg, (ATerm)ATreplace((ATermList)ATgetArgument((ATermAppl)arg, 1), (ATerm)((ATerm) wsAfterList), 11), 1);
  }
  else if (PERR_isErrorWarning(arg)) {
    return (PERR_Error)ATsetArgument((ATermAppl)arg, (ATerm)ATreplace((ATermList)ATgetArgument((ATermAppl)arg, 1), (ATerm)((ATerm) wsAfterList), 11), 1);
  }
  else if (PERR_isErrorError(arg)) {
    return (PERR_Error)ATsetArgument((ATermAppl)arg, (ATerm)ATreplace((ATermList)ATgetArgument((ATermAppl)arg, 1), (ATerm)((ATerm) wsAfterList), 11), 1);
  }
  else if (PERR_isErrorFatal(arg)) {
    return (PERR_Error)ATsetArgument((ATermAppl)arg, (ATerm)ATreplace((ATermList)ATgetArgument((ATermAppl)arg, 1), (ATerm)((ATerm) wsAfterList), 11), 1);
  }

  ATabort("Error has no WsAfterList: %t\n", arg);
  return (PERR_Error)NULL;
}

/**
 * Set the ws-after-] of a PERR_Error. The precondition being that this PERR_Error actually has a ws-after-]
 * \param[in] arg input PERR_Error
 * \param[in] wsAfterBracketClose new PERR_OptLayout to set in #arg
 * \return A new PERR_Error with wsAfterBracketClose at the right place, or a core dump if #arg did not have a wsAfterBracketClose
 */
PERR_Error PERR_setErrorWsAfterBracketClose(PERR_Error arg, PERR_OptLayout wsAfterBracketClose) {
  if (PERR_isErrorInfo(arg)) {
    return (PERR_Error)ATsetArgument((ATermAppl)arg, (ATerm)ATreplace((ATermList)ATgetArgument((ATermAppl)arg, 1), (ATerm)((ATerm) wsAfterBracketClose), 13), 1);
  }
  else if (PERR_isErrorWarning(arg)) {
    return (PERR_Error)ATsetArgument((ATermAppl)arg, (ATerm)ATreplace((ATermList)ATgetArgument((ATermAppl)arg, 1), (ATerm)((ATerm) wsAfterBracketClose), 13), 1);
  }
  else if (PERR_isErrorError(arg)) {
    return (PERR_Error)ATsetArgument((ATermAppl)arg, (ATerm)ATreplace((ATermList)ATgetArgument((ATermAppl)arg, 1), (ATerm)((ATerm) wsAfterBracketClose), 13), 1);
  }
  else if (PERR_isErrorFatal(arg)) {
    return (PERR_Error)ATsetArgument((ATermAppl)arg, (ATerm)ATreplace((ATermList)ATgetArgument((ATermAppl)arg, 1), (ATerm)((ATerm) wsAfterBracketClose), 13), 1);
  }

  ATabort("Error has no WsAfterBracketClose: %t\n", arg);
  return (PERR_Error)NULL;
}

/**
 * Set the ws-after-warning of a PERR_Error. The precondition being that this PERR_Error actually has a ws-after-warning
 * \param[in] arg input PERR_Error
 * \param[in] wsAfterWarning new PERR_OptLayout to set in #arg
 * \return A new PERR_Error with wsAfterWarning at the right place, or a core dump if #arg did not have a wsAfterWarning
 */
PERR_Error PERR_setErrorWsAfterWarning(PERR_Error arg, PERR_OptLayout wsAfterWarning) {
  if (PERR_isErrorWarning(arg)) {
    return (PERR_Error)ATsetArgument((ATermAppl)arg, (ATerm)ATreplace((ATermList)ATgetArgument((ATermAppl)arg, 1), (ATerm)((ATerm) wsAfterWarning), 1), 1);
  }

  ATabort("Error has no WsAfterWarning: %t\n", arg);
  return (PERR_Error)NULL;
}

/**
 * Set the ws-after-error of a PERR_Error. The precondition being that this PERR_Error actually has a ws-after-error
 * \param[in] arg input PERR_Error
 * \param[in] wsAfterError new PERR_OptLayout to set in #arg
 * \return A new PERR_Error with wsAfterError at the right place, or a core dump if #arg did not have a wsAfterError
 */
PERR_Error PERR_setErrorWsAfterError(PERR_Error arg, PERR_OptLayout wsAfterError) {
  if (PERR_isErrorError(arg)) {
    return (PERR_Error)ATsetArgument((ATermAppl)arg, (ATerm)ATreplace((ATermList)ATgetArgument((ATermAppl)arg, 1), (ATerm)((ATerm) wsAfterError), 1), 1);
  }

  ATabort("Error has no WsAfterError: %t\n", arg);
  return (PERR_Error)NULL;
}

/**
 * Set the ws-after-fatal of a PERR_Error. The precondition being that this PERR_Error actually has a ws-after-fatal
 * \param[in] arg input PERR_Error
 * \param[in] wsAfterFatal new PERR_OptLayout to set in #arg
 * \return A new PERR_Error with wsAfterFatal at the right place, or a core dump if #arg did not have a wsAfterFatal
 */
PERR_Error PERR_setErrorWsAfterFatal(PERR_Error arg, PERR_OptLayout wsAfterFatal) {
  if (PERR_isErrorFatal(arg)) {
    return (PERR_Error)ATsetArgument((ATermAppl)arg, (ATerm)ATreplace((ATermList)ATgetArgument((ATermAppl)arg, 1), (ATerm)((ATerm) wsAfterFatal), 1), 1);
  }

  ATabort("Error has no WsAfterFatal: %t\n", arg);
  return (PERR_Error)NULL;
}

/**
 * Assert whether a PERR_Summary is any of the valid alternatives, or not. This analysis does not go any deeper than the top level
 * \param[in] arg input PERR_Summary
 * \return ATtrue if #arg corresponds to the expected signature, or ATfalse otherwise
 */
ATbool PERR_isValidSummary(PERR_Summary arg) {
  if (PERR_isSummarySummary(arg)) {
    return ATtrue;
  }
  return ATfalse;
}

/**
 * Assert whether a PERR_Summary is a summary by checking against the following ATerm pattern: appl(prod([lit("summary"),cf(opt(layout)),lit("("),cf(opt(layout)),cf(sort("StrCon")),cf(opt(layout)),lit(","),cf(opt(layout)),cf(sort("StrCon")),cf(opt(layout)),lit(","),cf(opt(layout)),lit("["),cf(opt(layout)),cf(iter-star-sep(sort("Error"),lit(","))),cf(opt(layout)),lit("]"),cf(opt(layout)),lit(")")],cf(sort("Summary")),attrs([term(cons("summary"))])),[appl(prod([char-class([115]),char-class([117]),char-class([109]),char-class([109]),char-class([97]),char-class([114]),char-class([121])],lit("summary"),no-attrs),[115,117,109,109,97,114,121]),<"ws-after-summary"("OptLayout")>,appl(prod([char-class([40])],lit("("),no-attrs),[40]),<"ws-after-("("OptLayout")>,<"producer"("StrCon")>,<"ws-after-producer"("OptLayout")>,appl(prod([char-class([44])],lit(","),no-attrs),[44]),<"ws-after-,"("OptLayout")>,<"id"("StrCon")>,<"ws-after-id"("OptLayout")>,appl(prod([char-class([44])],lit(","),no-attrs),[44]),<"ws-after-,-1"("OptLayout")>,appl(prod([char-class([91])],lit("["),no-attrs),[91]),<"ws-after-["("OptLayout")>,appl(list(cf(iter-star-sep(sort("Error"),lit(",")))),<"list"("Error-list")>),<"ws-after-list"("OptLayout")>,appl(prod([char-class([93])],lit("]"),no-attrs),[93]),<"ws-after-]"("OptLayout")>,appl(prod([char-class([41])],lit(")"),no-attrs),[41])]). Always returns ATtrue
 * \param[in] arg input PERR_Summary
 * \return ATtrue if #arg corresponds to the signature of a summary, or ATfalse otherwise
 */
inline ATbool PERR_isSummarySummary(PERR_Summary arg){
  /* checking for: appl */
  if (ATgetType((ATerm)arg) == AT_APPL && ATgetAFun((ATermAppl)arg) == PERR_afun0) {
    ATerm arg_arg0 = ATgetArgument(arg, 0);
    /* checking for: prod */
    if (ATgetType((ATerm)arg_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0) == PERR_afun1) {
      ATerm arg_arg0_arg0 = ATgetArgument(arg_arg0, 0);
      if (ATgetType((ATerm)arg_arg0_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0) == ATfalse) {
        ATerm arg_arg0_arg0_head;
        ATermList arg_arg0_arg0_list = (ATermList)arg_arg0_arg0;
        arg_arg0_arg0_head = ATgetFirst(arg_arg0_arg0_list);
        arg_arg0_arg0_list = ATgetNext(arg_arg0_arg0_list);
        /* checking for: lit */
        if (ATgetType((ATerm)arg_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head) == PERR_afun9) {
          ATerm arg_arg0_arg0_head_arg0 = ATgetArgument(arg_arg0_arg0_head, 0);
          /* checking for: summary */
          if (ATgetType((ATerm)arg_arg0_arg0_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0) == PERR_afun52) {
            if (ATgetType((ATerm)arg_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0_list) == ATfalse) {
              arg_arg0_arg0_head = ATgetFirst(arg_arg0_arg0_list);
              arg_arg0_arg0_list = ATgetNext(arg_arg0_arg0_list);
              /* checking for: cf */
              if (ATgetType((ATerm)arg_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head) == PERR_afun2) {
                ATerm arg_arg0_arg0_head_arg0 = ATgetArgument(arg_arg0_arg0_head, 0);
                /* checking for: opt */
                if (ATgetType((ATerm)arg_arg0_arg0_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0) == PERR_afun3) {
                  ATerm arg_arg0_arg0_head_arg0_arg0 = ATgetArgument(arg_arg0_arg0_head_arg0, 0);
                  /* checking for: layout */
                  if (ATgetType((ATerm)arg_arg0_arg0_head_arg0_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0_arg0) == PERR_afun4) {
                    if (ATgetType((ATerm)arg_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0_list) == ATfalse) {
                      arg_arg0_arg0_head = ATgetFirst(arg_arg0_arg0_list);
                      arg_arg0_arg0_list = ATgetNext(arg_arg0_arg0_list);
                      /* checking for: lit */
                      if (ATgetType((ATerm)arg_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head) == PERR_afun9) {
                        ATerm arg_arg0_arg0_head_arg0 = ATgetArgument(arg_arg0_arg0_head, 0);
                        /* checking for: ( */
                        if (ATgetType((ATerm)arg_arg0_arg0_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0) == PERR_afun39) {
                          if (ATgetType((ATerm)arg_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0_list) == ATfalse) {
                            arg_arg0_arg0_head = ATgetFirst(arg_arg0_arg0_list);
                            arg_arg0_arg0_list = ATgetNext(arg_arg0_arg0_list);
                            /* checking for: cf */
                            if (ATgetType((ATerm)arg_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head) == PERR_afun2) {
                              ATerm arg_arg0_arg0_head_arg0 = ATgetArgument(arg_arg0_arg0_head, 0);
                              /* checking for: opt */
                              if (ATgetType((ATerm)arg_arg0_arg0_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0) == PERR_afun3) {
                                ATerm arg_arg0_arg0_head_arg0_arg0 = ATgetArgument(arg_arg0_arg0_head_arg0, 0);
                                /* checking for: layout */
                                if (ATgetType((ATerm)arg_arg0_arg0_head_arg0_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0_arg0) == PERR_afun4) {
                                  if (ATgetType((ATerm)arg_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0_list) == ATfalse) {
                                    arg_arg0_arg0_head = ATgetFirst(arg_arg0_arg0_list);
                                    arg_arg0_arg0_list = ATgetNext(arg_arg0_arg0_list);
                                    /* checking for: cf */
                                    if (ATgetType((ATerm)arg_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head) == PERR_afun2) {
                                      ATerm arg_arg0_arg0_head_arg0 = ATgetArgument(arg_arg0_arg0_head, 0);
                                      /* checking for: sort */
                                      if (ATgetType((ATerm)arg_arg0_arg0_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0) == PERR_afun11) {
                                        ATerm arg_arg0_arg0_head_arg0_arg0 = ATgetArgument(arg_arg0_arg0_head_arg0, 0);
                                        /* checking for: StrCon */
                                        if (ATgetType((ATerm)arg_arg0_arg0_head_arg0_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0_arg0) == PERR_afun29) {
                                          if (ATgetType((ATerm)arg_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0_list) == ATfalse) {
                                            arg_arg0_arg0_head = ATgetFirst(arg_arg0_arg0_list);
                                            arg_arg0_arg0_list = ATgetNext(arg_arg0_arg0_list);
                                            /* checking for: cf */
                                            if (ATgetType((ATerm)arg_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head) == PERR_afun2) {
                                              ATerm arg_arg0_arg0_head_arg0 = ATgetArgument(arg_arg0_arg0_head, 0);
                                              /* checking for: opt */
                                              if (ATgetType((ATerm)arg_arg0_arg0_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0) == PERR_afun3) {
                                                ATerm arg_arg0_arg0_head_arg0_arg0 = ATgetArgument(arg_arg0_arg0_head_arg0, 0);
                                                /* checking for: layout */
                                                if (ATgetType((ATerm)arg_arg0_arg0_head_arg0_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0_arg0) == PERR_afun4) {
                                                  if (ATgetType((ATerm)arg_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0_list) == ATfalse) {
                                                    arg_arg0_arg0_head = ATgetFirst(arg_arg0_arg0_list);
                                                    arg_arg0_arg0_list = ATgetNext(arg_arg0_arg0_list);
                                                    /* checking for: lit */
                                                    if (ATgetType((ATerm)arg_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head) == PERR_afun9) {
                                                      ATerm arg_arg0_arg0_head_arg0 = ATgetArgument(arg_arg0_arg0_head, 0);
                                                      /* checking for: , */
                                                      if (ATgetType((ATerm)arg_arg0_arg0_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0) == PERR_afun43) {
                                                        if (ATgetType((ATerm)arg_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0_list) == ATfalse) {
                                                          arg_arg0_arg0_head = ATgetFirst(arg_arg0_arg0_list);
                                                          arg_arg0_arg0_list = ATgetNext(arg_arg0_arg0_list);
                                                          /* checking for: cf */
                                                          if (ATgetType((ATerm)arg_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head) == PERR_afun2) {
                                                            ATerm arg_arg0_arg0_head_arg0 = ATgetArgument(arg_arg0_arg0_head, 0);
                                                            /* checking for: opt */
                                                            if (ATgetType((ATerm)arg_arg0_arg0_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0) == PERR_afun3) {
                                                              ATerm arg_arg0_arg0_head_arg0_arg0 = ATgetArgument(arg_arg0_arg0_head_arg0, 0);
                                                              /* checking for: layout */
                                                              if (ATgetType((ATerm)arg_arg0_arg0_head_arg0_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0_arg0) == PERR_afun4) {
                                                                if (ATgetType((ATerm)arg_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0_list) == ATfalse) {
                                                                  arg_arg0_arg0_head = ATgetFirst(arg_arg0_arg0_list);
                                                                  arg_arg0_arg0_list = ATgetNext(arg_arg0_arg0_list);
                                                                  /* checking for: cf */
                                                                  if (ATgetType((ATerm)arg_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head) == PERR_afun2) {
                                                                    ATerm arg_arg0_arg0_head_arg0 = ATgetArgument(arg_arg0_arg0_head, 0);
                                                                    /* checking for: sort */
                                                                    if (ATgetType((ATerm)arg_arg0_arg0_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0) == PERR_afun11) {
                                                                      ATerm arg_arg0_arg0_head_arg0_arg0 = ATgetArgument(arg_arg0_arg0_head_arg0, 0);
                                                                      /* checking for: StrCon */
                                                                      if (ATgetType((ATerm)arg_arg0_arg0_head_arg0_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0_arg0) == PERR_afun29) {
                                                                        if (ATgetType((ATerm)arg_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0_list) == ATfalse) {
                                                                          arg_arg0_arg0_head = ATgetFirst(arg_arg0_arg0_list);
                                                                          arg_arg0_arg0_list = ATgetNext(arg_arg0_arg0_list);
                                                                          /* checking for: cf */
                                                                          if (ATgetType((ATerm)arg_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head) == PERR_afun2) {
                                                                            ATerm arg_arg0_arg0_head_arg0 = ATgetArgument(arg_arg0_arg0_head, 0);
                                                                            /* checking for: opt */
                                                                            if (ATgetType((ATerm)arg_arg0_arg0_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0) == PERR_afun3) {
                                                                              ATerm arg_arg0_arg0_head_arg0_arg0 = ATgetArgument(arg_arg0_arg0_head_arg0, 0);
                                                                              /* checking for: layout */
                                                                              if (ATgetType((ATerm)arg_arg0_arg0_head_arg0_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0_arg0) == PERR_afun4) {
                                                                                if (ATgetType((ATerm)arg_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0_list) == ATfalse) {
                                                                                  arg_arg0_arg0_head = ATgetFirst(arg_arg0_arg0_list);
                                                                                  arg_arg0_arg0_list = ATgetNext(arg_arg0_arg0_list);
                                                                                  /* checking for: lit */
                                                                                  if (ATgetType((ATerm)arg_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head) == PERR_afun9) {
                                                                                    ATerm arg_arg0_arg0_head_arg0 = ATgetArgument(arg_arg0_arg0_head, 0);
                                                                                    /* checking for: , */
                                                                                    if (ATgetType((ATerm)arg_arg0_arg0_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0) == PERR_afun43) {
                                                                                      if (ATgetType((ATerm)arg_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0_list) == ATfalse) {
                                                                                        arg_arg0_arg0_head = ATgetFirst(arg_arg0_arg0_list);
                                                                                        arg_arg0_arg0_list = ATgetNext(arg_arg0_arg0_list);
                                                                                        /* checking for: cf */
                                                                                        if (ATgetType((ATerm)arg_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head) == PERR_afun2) {
                                                                                          ATerm arg_arg0_arg0_head_arg0 = ATgetArgument(arg_arg0_arg0_head, 0);
                                                                                          /* checking for: opt */
                                                                                          if (ATgetType((ATerm)arg_arg0_arg0_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0) == PERR_afun3) {
                                                                                            ATerm arg_arg0_arg0_head_arg0_arg0 = ATgetArgument(arg_arg0_arg0_head_arg0, 0);
                                                                                            /* checking for: layout */
                                                                                            if (ATgetType((ATerm)arg_arg0_arg0_head_arg0_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0_arg0) == PERR_afun4) {
                                                                                              if (ATgetType((ATerm)arg_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0_list) == ATfalse) {
                                                                                                arg_arg0_arg0_head = ATgetFirst(arg_arg0_arg0_list);
                                                                                                arg_arg0_arg0_list = ATgetNext(arg_arg0_arg0_list);
                                                                                                /* checking for: lit */
                                                                                                if (ATgetType((ATerm)arg_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head) == PERR_afun9) {
                                                                                                  ATerm arg_arg0_arg0_head_arg0 = ATgetArgument(arg_arg0_arg0_head, 0);
                                                                                                  /* checking for: [ */
                                                                                                  if (ATgetType((ATerm)arg_arg0_arg0_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0) == PERR_afun47) {
                                                                                                    if (ATgetType((ATerm)arg_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0_list) == ATfalse) {
                                                                                                      arg_arg0_arg0_head = ATgetFirst(arg_arg0_arg0_list);
                                                                                                      arg_arg0_arg0_list = ATgetNext(arg_arg0_arg0_list);
                                                                                                      /* checking for: cf */
                                                                                                      if (ATgetType((ATerm)arg_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head) == PERR_afun2) {
                                                                                                        ATerm arg_arg0_arg0_head_arg0 = ATgetArgument(arg_arg0_arg0_head, 0);
                                                                                                        /* checking for: opt */
                                                                                                        if (ATgetType((ATerm)arg_arg0_arg0_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0) == PERR_afun3) {
                                                                                                          ATerm arg_arg0_arg0_head_arg0_arg0 = ATgetArgument(arg_arg0_arg0_head_arg0, 0);
                                                                                                          /* checking for: layout */
                                                                                                          if (ATgetType((ATerm)arg_arg0_arg0_head_arg0_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0_arg0) == PERR_afun4) {
                                                                                                            if (ATgetType((ATerm)arg_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0_list) == ATfalse) {
                                                                                                              arg_arg0_arg0_head = ATgetFirst(arg_arg0_arg0_list);
                                                                                                              arg_arg0_arg0_list = ATgetNext(arg_arg0_arg0_list);
                                                                                                              /* checking for: cf */
                                                                                                              if (ATgetType((ATerm)arg_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head) == PERR_afun2) {
                                                                                                                ATerm arg_arg0_arg0_head_arg0 = ATgetArgument(arg_arg0_arg0_head, 0);
                                                                                                                /* checking for: iter-star-sep */
                                                                                                                if (ATgetType((ATerm)arg_arg0_arg0_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0) == PERR_afun46) {
                                                                                                                  ATerm arg_arg0_arg0_head_arg0_arg0 = ATgetArgument(arg_arg0_arg0_head_arg0, 0);
                                                                                                                  /* checking for: sort */
                                                                                                                  if (ATgetType((ATerm)arg_arg0_arg0_head_arg0_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0_arg0) == PERR_afun11) {
                                                                                                                    ATerm arg_arg0_arg0_head_arg0_arg0_arg0 = ATgetArgument(arg_arg0_arg0_head_arg0_arg0, 0);
                                                                                                                    /* checking for: Error */
                                                                                                                    if (ATgetType((ATerm)arg_arg0_arg0_head_arg0_arg0_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0_arg0_arg0) == PERR_afun37) {
                                                                                                                      ATerm arg_arg0_arg0_head_arg0_arg1 = ATgetArgument(arg_arg0_arg0_head_arg0, 1);
                                                                                                                      /* checking for: lit */
                                                                                                                      if (ATgetType((ATerm)arg_arg0_arg0_head_arg0_arg1) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0_arg1) == PERR_afun9) {
                                                                                                                        ATerm arg_arg0_arg0_head_arg0_arg1_arg0 = ATgetArgument(arg_arg0_arg0_head_arg0_arg1, 0);
                                                                                                                        /* checking for: , */
                                                                                                                        if (ATgetType((ATerm)arg_arg0_arg0_head_arg0_arg1_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0_arg1_arg0) == PERR_afun43) {
                                                                                                                          if (ATgetType((ATerm)arg_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0_list) == ATfalse) {
                                                                                                                            arg_arg0_arg0_head = ATgetFirst(arg_arg0_arg0_list);
                                                                                                                            arg_arg0_arg0_list = ATgetNext(arg_arg0_arg0_list);
                                                                                                                            /* checking for: cf */
                                                                                                                            if (ATgetType((ATerm)arg_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head) == PERR_afun2) {
                                                                                                                              ATerm arg_arg0_arg0_head_arg0 = ATgetArgument(arg_arg0_arg0_head, 0);
                                                                                                                              /* checking for: opt */
                                                                                                                              if (ATgetType((ATerm)arg_arg0_arg0_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0) == PERR_afun3) {
                                                                                                                                ATerm arg_arg0_arg0_head_arg0_arg0 = ATgetArgument(arg_arg0_arg0_head_arg0, 0);
                                                                                                                                /* checking for: layout */
                                                                                                                                if (ATgetType((ATerm)arg_arg0_arg0_head_arg0_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0_arg0) == PERR_afun4) {
                                                                                                                                  if (ATgetType((ATerm)arg_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0_list) == ATfalse) {
                                                                                                                                    arg_arg0_arg0_head = ATgetFirst(arg_arg0_arg0_list);
                                                                                                                                    arg_arg0_arg0_list = ATgetNext(arg_arg0_arg0_list);
                                                                                                                                    /* checking for: lit */
                                                                                                                                    if (ATgetType((ATerm)arg_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head) == PERR_afun9) {
                                                                                                                                      ATerm arg_arg0_arg0_head_arg0 = ATgetArgument(arg_arg0_arg0_head, 0);
                                                                                                                                      /* checking for: ] */
                                                                                                                                      if (ATgetType((ATerm)arg_arg0_arg0_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0) == PERR_afun45) {
                                                                                                                                        if (ATgetType((ATerm)arg_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0_list) == ATfalse) {
                                                                                                                                          arg_arg0_arg0_head = ATgetFirst(arg_arg0_arg0_list);
                                                                                                                                          arg_arg0_arg0_list = ATgetNext(arg_arg0_arg0_list);
                                                                                                                                          /* checking for: cf */
                                                                                                                                          if (ATgetType((ATerm)arg_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head) == PERR_afun2) {
                                                                                                                                            ATerm arg_arg0_arg0_head_arg0 = ATgetArgument(arg_arg0_arg0_head, 0);
                                                                                                                                            /* checking for: opt */
                                                                                                                                            if (ATgetType((ATerm)arg_arg0_arg0_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0) == PERR_afun3) {
                                                                                                                                              ATerm arg_arg0_arg0_head_arg0_arg0 = ATgetArgument(arg_arg0_arg0_head_arg0, 0);
                                                                                                                                              /* checking for: layout */
                                                                                                                                              if (ATgetType((ATerm)arg_arg0_arg0_head_arg0_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0_arg0) == PERR_afun4) {
                                                                                                                                                if (ATgetType((ATerm)arg_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0_list) == ATfalse) {
                                                                                                                                                  arg_arg0_arg0_head = ATgetFirst(arg_arg0_arg0_list);
                                                                                                                                                  arg_arg0_arg0_list = ATgetNext(arg_arg0_arg0_list);
                                                                                                                                                  /* checking for: lit */
                                                                                                                                                  if (ATgetType((ATerm)arg_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head) == PERR_afun9) {
                                                                                                                                                    ATerm arg_arg0_arg0_head_arg0 = ATgetArgument(arg_arg0_arg0_head, 0);
                                                                                                                                                    /* checking for: ) */
                                                                                                                                                    if (ATgetType((ATerm)arg_arg0_arg0_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0) == PERR_afun38) {
                                                                                                                                                      if (ATgetType((ATerm)arg_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0_list) == ATtrue) {
                                                                                                                                                        ATerm arg_arg0_arg1 = ATgetArgument(arg_arg0, 1);
                                                                                                                                                        /* checking for: cf */
                                                                                                                                                        if (ATgetType((ATerm)arg_arg0_arg1) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg1) == PERR_afun2) {
                                                                                                                                                          ATerm arg_arg0_arg1_arg0 = ATgetArgument(arg_arg0_arg1, 0);
                                                                                                                                                          /* checking for: sort */
                                                                                                                                                          if (ATgetType((ATerm)arg_arg0_arg1_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg1_arg0) == PERR_afun11) {
                                                                                                                                                            ATerm arg_arg0_arg1_arg0_arg0 = ATgetArgument(arg_arg0_arg1_arg0, 0);
                                                                                                                                                            /* checking for: Summary */
                                                                                                                                                            if (ATgetType((ATerm)arg_arg0_arg1_arg0_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg1_arg0_arg0) == PERR_afun35) {
                                                                                                                                                              ATerm arg_arg0_arg2 = ATgetArgument(arg_arg0, 2);
                                                                                                                                                              /* checking for: attrs */
                                                                                                                                                              if (ATgetType((ATerm)arg_arg0_arg2) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg2) == PERR_afun13) {
                                                                                                                                                                ATerm arg_arg0_arg2_arg0 = ATgetArgument(arg_arg0_arg2, 0);
                                                                                                                                                                if (ATgetType((ATerm)arg_arg0_arg2_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg2_arg0) == ATfalse) {
                                                                                                                                                                  ATerm arg_arg0_arg2_arg0_head;
                                                                                                                                                                  ATermList arg_arg0_arg2_arg0_list = (ATermList)arg_arg0_arg2_arg0;
                                                                                                                                                                  arg_arg0_arg2_arg0_head = ATgetFirst(arg_arg0_arg2_arg0_list);
                                                                                                                                                                  arg_arg0_arg2_arg0_list = ATgetNext(arg_arg0_arg2_arg0_list);
                                                                                                                                                                  /* checking for: term */
                                                                                                                                                                  if (ATgetType((ATerm)arg_arg0_arg2_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg2_arg0_head) == PERR_afun14) {
                                                                                                                                                                    ATerm arg_arg0_arg2_arg0_head_arg0 = ATgetArgument(arg_arg0_arg2_arg0_head, 0);
                                                                                                                                                                    /* checking for: cons */
                                                                                                                                                                    if (ATgetType((ATerm)arg_arg0_arg2_arg0_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg2_arg0_head_arg0) == PERR_afun15) {
                                                                                                                                                                      ATerm arg_arg0_arg2_arg0_head_arg0_arg0 = ATgetArgument(arg_arg0_arg2_arg0_head_arg0, 0);
                                                                                                                                                                      /* checking for: summary */
                                                                                                                                                                      if (ATgetType((ATerm)arg_arg0_arg2_arg0_head_arg0_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg2_arg0_head_arg0_arg0) == PERR_afun52) {
                                                                                                                                                                        if (ATgetType((ATerm)arg_arg0_arg2_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg2_arg0_list) == ATtrue) {
                                                                                                                                                                          ATerm arg_arg1 = ATgetArgument(arg, 1);
                                                                                                                                                                          if (ATgetType((ATerm)arg_arg1) == AT_LIST && ATisEmpty((ATermList)arg_arg1) == ATfalse) {
                                                                                                                                                                            ATerm arg_arg1_head;
                                                                                                                                                                            ATermList arg_arg1_list = (ATermList)arg_arg1;
                                                                                                                                                                            arg_arg1_head = ATgetFirst(arg_arg1_list);
                                                                                                                                                                            arg_arg1_list = ATgetNext(arg_arg1_list);
                                                                                                                                                                            /* checking for: appl */
                                                                                                                                                                            if (ATgetType((ATerm)arg_arg1_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head) == PERR_afun0) {
                                                                                                                                                                              ATerm arg_arg1_head_arg0 = ATgetArgument(arg_arg1_head, 0);
                                                                                                                                                                              /* checking for: prod */
                                                                                                                                                                              if (ATgetType((ATerm)arg_arg1_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0) == PERR_afun1) {
                                                                                                                                                                                ATerm arg_arg1_head_arg0_arg0 = ATgetArgument(arg_arg1_head_arg0, 0);
                                                                                                                                                                                if (ATgetType((ATerm)arg_arg1_head_arg0_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0) == ATfalse) {
                                                                                                                                                                                  ATerm arg_arg1_head_arg0_arg0_head;
                                                                                                                                                                                  ATermList arg_arg1_head_arg0_arg0_list = (ATermList)arg_arg1_head_arg0_arg0;
                                                                                                                                                                                  arg_arg1_head_arg0_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_list);
                                                                                                                                                                                  arg_arg1_head_arg0_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_list);
                                                                                                                                                                                  /* checking for: char-class */
                                                                                                                                                                                  if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg0_head) == PERR_afun17) {
                                                                                                                                                                                    ATerm arg_arg1_head_arg0_arg0_head_arg0 = ATgetArgument(arg_arg1_head_arg0_arg0_head, 0);
                                                                                                                                                                                    if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0) == ATfalse) {
                                                                                                                                                                                      ATerm arg_arg1_head_arg0_arg0_head_arg0_head;
                                                                                                                                                                                      ATermList arg_arg1_head_arg0_arg0_head_arg0_list = (ATermList)arg_arg1_head_arg0_arg0_head_arg0;
                                                                                                                                                                                      arg_arg1_head_arg0_arg0_head_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                                                                                                      arg_arg1_head_arg0_arg0_head_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                                                                                                      if (ATgetInt((ATermInt)arg_arg1_head_arg0_arg0_head_arg0_head) == 115) {

                                                                                                                                                                                        if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0_list) == ATtrue) {
                                                                                                                                                                                          if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_list) == ATfalse) {
                                                                                                                                                                                            arg_arg1_head_arg0_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_list);
                                                                                                                                                                                            arg_arg1_head_arg0_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_list);
                                                                                                                                                                                            /* checking for: char-class */
                                                                                                                                                                                            if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg0_head) == PERR_afun17) {
                                                                                                                                                                                              ATerm arg_arg1_head_arg0_arg0_head_arg0 = ATgetArgument(arg_arg1_head_arg0_arg0_head, 0);
                                                                                                                                                                                              if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0) == ATfalse) {
                                                                                                                                                                                                ATerm arg_arg1_head_arg0_arg0_head_arg0_head;
                                                                                                                                                                                                ATermList arg_arg1_head_arg0_arg0_head_arg0_list = (ATermList)arg_arg1_head_arg0_arg0_head_arg0;
                                                                                                                                                                                                arg_arg1_head_arg0_arg0_head_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                                                                                                                arg_arg1_head_arg0_arg0_head_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                                                                                                                if (ATgetInt((ATermInt)arg_arg1_head_arg0_arg0_head_arg0_head) == 117) {

                                                                                                                                                                                                  if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0_list) == ATtrue) {
                                                                                                                                                                                                    if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_list) == ATfalse) {
                                                                                                                                                                                                      arg_arg1_head_arg0_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_list);
                                                                                                                                                                                                      arg_arg1_head_arg0_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_list);
                                                                                                                                                                                                      /* checking for: char-class */
                                                                                                                                                                                                      if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg0_head) == PERR_afun17) {
                                                                                                                                                                                                        ATerm arg_arg1_head_arg0_arg0_head_arg0 = ATgetArgument(arg_arg1_head_arg0_arg0_head, 0);
                                                                                                                                                                                                        if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0) == ATfalse) {
                                                                                                                                                                                                          ATerm arg_arg1_head_arg0_arg0_head_arg0_head;
                                                                                                                                                                                                          ATermList arg_arg1_head_arg0_arg0_head_arg0_list = (ATermList)arg_arg1_head_arg0_arg0_head_arg0;
                                                                                                                                                                                                          arg_arg1_head_arg0_arg0_head_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                                                                                                                          arg_arg1_head_arg0_arg0_head_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                                                                                                                          if (ATgetInt((ATermInt)arg_arg1_head_arg0_arg0_head_arg0_head) == 109) {

                                                                                                                                                                                                            if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0_list) == ATtrue) {
                                                                                                                                                                                                              if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_list) == ATfalse) {
                                                                                                                                                                                                                arg_arg1_head_arg0_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_list);
                                                                                                                                                                                                                arg_arg1_head_arg0_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_list);
                                                                                                                                                                                                                /* checking for: char-class */
                                                                                                                                                                                                                if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg0_head) == PERR_afun17) {
                                                                                                                                                                                                                  ATerm arg_arg1_head_arg0_arg0_head_arg0 = ATgetArgument(arg_arg1_head_arg0_arg0_head, 0);
                                                                                                                                                                                                                  if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0) == ATfalse) {
                                                                                                                                                                                                                    ATerm arg_arg1_head_arg0_arg0_head_arg0_head;
                                                                                                                                                                                                                    ATermList arg_arg1_head_arg0_arg0_head_arg0_list = (ATermList)arg_arg1_head_arg0_arg0_head_arg0;
                                                                                                                                                                                                                    arg_arg1_head_arg0_arg0_head_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                                                                                                                                    arg_arg1_head_arg0_arg0_head_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                                                                                                                                    if (ATgetInt((ATermInt)arg_arg1_head_arg0_arg0_head_arg0_head) == 109) {

                                                                                                                                                                                                                      if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0_list) == ATtrue) {
                                                                                                                                                                                                                        if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_list) == ATfalse) {
                                                                                                                                                                                                                          arg_arg1_head_arg0_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_list);
                                                                                                                                                                                                                          arg_arg1_head_arg0_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_list);
                                                                                                                                                                                                                          /* checking for: char-class */
                                                                                                                                                                                                                          if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg0_head) == PERR_afun17) {
                                                                                                                                                                                                                            ATerm arg_arg1_head_arg0_arg0_head_arg0 = ATgetArgument(arg_arg1_head_arg0_arg0_head, 0);
                                                                                                                                                                                                                            if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0) == ATfalse) {
                                                                                                                                                                                                                              ATerm arg_arg1_head_arg0_arg0_head_arg0_head;
                                                                                                                                                                                                                              ATermList arg_arg1_head_arg0_arg0_head_arg0_list = (ATermList)arg_arg1_head_arg0_arg0_head_arg0;
                                                                                                                                                                                                                              arg_arg1_head_arg0_arg0_head_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                                                                                                                                              arg_arg1_head_arg0_arg0_head_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                                                                                                                                              if (ATgetInt((ATermInt)arg_arg1_head_arg0_arg0_head_arg0_head) == 97) {

                                                                                                                                                                                                                                if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0_list) == ATtrue) {
                                                                                                                                                                                                                                  if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_list) == ATfalse) {
                                                                                                                                                                                                                                    arg_arg1_head_arg0_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_list);
                                                                                                                                                                                                                                    arg_arg1_head_arg0_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_list);
                                                                                                                                                                                                                                    /* checking for: char-class */
                                                                                                                                                                                                                                    if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg0_head) == PERR_afun17) {
                                                                                                                                                                                                                                      ATerm arg_arg1_head_arg0_arg0_head_arg0 = ATgetArgument(arg_arg1_head_arg0_arg0_head, 0);
                                                                                                                                                                                                                                      if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0) == ATfalse) {
                                                                                                                                                                                                                                        ATerm arg_arg1_head_arg0_arg0_head_arg0_head;
                                                                                                                                                                                                                                        ATermList arg_arg1_head_arg0_arg0_head_arg0_list = (ATermList)arg_arg1_head_arg0_arg0_head_arg0;
                                                                                                                                                                                                                                        arg_arg1_head_arg0_arg0_head_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                                                                                                                                                        arg_arg1_head_arg0_arg0_head_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                                                                                                                                                        if (ATgetInt((ATermInt)arg_arg1_head_arg0_arg0_head_arg0_head) == 114) {

                                                                                                                                                                                                                                          if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0_list) == ATtrue) {
                                                                                                                                                                                                                                            if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_list) == ATfalse) {
                                                                                                                                                                                                                                              arg_arg1_head_arg0_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_list);
                                                                                                                                                                                                                                              arg_arg1_head_arg0_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_list);
                                                                                                                                                                                                                                              /* checking for: char-class */
                                                                                                                                                                                                                                              if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg0_head) == PERR_afun17) {
                                                                                                                                                                                                                                                ATerm arg_arg1_head_arg0_arg0_head_arg0 = ATgetArgument(arg_arg1_head_arg0_arg0_head, 0);
                                                                                                                                                                                                                                                if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0) == ATfalse) {
                                                                                                                                                                                                                                                  ATerm arg_arg1_head_arg0_arg0_head_arg0_head;
                                                                                                                                                                                                                                                  ATermList arg_arg1_head_arg0_arg0_head_arg0_list = (ATermList)arg_arg1_head_arg0_arg0_head_arg0;
                                                                                                                                                                                                                                                  arg_arg1_head_arg0_arg0_head_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                                                                                                                                                                  arg_arg1_head_arg0_arg0_head_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                                                                                                                                                                  if (ATgetInt((ATermInt)arg_arg1_head_arg0_arg0_head_arg0_head) == 121) {

                                                                                                                                                                                                                                                    if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0_list) == ATtrue) {
                                                                                                                                                                                                                                                      if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_list) == ATtrue) {
                                                                                                                                                                                                                                                        ATerm arg_arg1_head_arg0_arg1 = ATgetArgument(arg_arg1_head_arg0, 1);
                                                                                                                                                                                                                                                        /* checking for: lit */
                                                                                                                                                                                                                                                        if (ATgetType((ATerm)arg_arg1_head_arg0_arg1) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg1) == PERR_afun9) {
                                                                                                                                                                                                                                                          ATerm arg_arg1_head_arg0_arg1_arg0 = ATgetArgument(arg_arg1_head_arg0_arg1, 0);
                                                                                                                                                                                                                                                          /* checking for: summary */
                                                                                                                                                                                                                                                          if (ATgetType((ATerm)arg_arg1_head_arg0_arg1_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg1_arg0) == PERR_afun52) {
                                                                                                                                                                                                                                                            ATerm arg_arg1_head_arg0_arg2 = ATgetArgument(arg_arg1_head_arg0, 2);
                                                                                                                                                                                                                                                            /* checking for: no-attrs */
                                                                                                                                                                                                                                                            if (ATgetType((ATerm)arg_arg1_head_arg0_arg2) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg2) == PERR_afun5) {
                                                                                                                                                                                                                                                              ATerm arg_arg1_head_arg1 = ATgetArgument(arg_arg1_head, 1);
                                                                                                                                                                                                                                                              if (ATgetType((ATerm)arg_arg1_head_arg1) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg1) == ATfalse) {
                                                                                                                                                                                                                                                                ATerm arg_arg1_head_arg1_head;
                                                                                                                                                                                                                                                                ATermList arg_arg1_head_arg1_list = (ATermList)arg_arg1_head_arg1;
                                                                                                                                                                                                                                                                arg_arg1_head_arg1_head = ATgetFirst(arg_arg1_head_arg1_list);
                                                                                                                                                                                                                                                                arg_arg1_head_arg1_list = ATgetNext(arg_arg1_head_arg1_list);
                                                                                                                                                                                                                                                                if (ATgetInt((ATermInt)arg_arg1_head_arg1_head) == 115) {

                                                                                                                                                                                                                                                                  if (ATgetType((ATerm)arg_arg1_head_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg1_list) == ATfalse) {
                                                                                                                                                                                                                                                                    arg_arg1_head_arg1_head = ATgetFirst(arg_arg1_head_arg1_list);
                                                                                                                                                                                                                                                                    arg_arg1_head_arg1_list = ATgetNext(arg_arg1_head_arg1_list);
                                                                                                                                                                                                                                                                    if (ATgetInt((ATermInt)arg_arg1_head_arg1_head) == 117) {

                                                                                                                                                                                                                                                                      if (ATgetType((ATerm)arg_arg1_head_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg1_list) == ATfalse) {
                                                                                                                                                                                                                                                                        arg_arg1_head_arg1_head = ATgetFirst(arg_arg1_head_arg1_list);
                                                                                                                                                                                                                                                                        arg_arg1_head_arg1_list = ATgetNext(arg_arg1_head_arg1_list);
                                                                                                                                                                                                                                                                        if (ATgetInt((ATermInt)arg_arg1_head_arg1_head) == 109) {

                                                                                                                                                                                                                                                                          if (ATgetType((ATerm)arg_arg1_head_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg1_list) == ATfalse) {
                                                                                                                                                                                                                                                                            arg_arg1_head_arg1_head = ATgetFirst(arg_arg1_head_arg1_list);
                                                                                                                                                                                                                                                                            arg_arg1_head_arg1_list = ATgetNext(arg_arg1_head_arg1_list);
                                                                                                                                                                                                                                                                            if (ATgetInt((ATermInt)arg_arg1_head_arg1_head) == 109) {

                                                                                                                                                                                                                                                                              if (ATgetType((ATerm)arg_arg1_head_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg1_list) == ATfalse) {
                                                                                                                                                                                                                                                                                arg_arg1_head_arg1_head = ATgetFirst(arg_arg1_head_arg1_list);
                                                                                                                                                                                                                                                                                arg_arg1_head_arg1_list = ATgetNext(arg_arg1_head_arg1_list);
                                                                                                                                                                                                                                                                                if (ATgetInt((ATermInt)arg_arg1_head_arg1_head) == 97) {

                                                                                                                                                                                                                                                                                  if (ATgetType((ATerm)arg_arg1_head_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg1_list) == ATfalse) {
                                                                                                                                                                                                                                                                                    arg_arg1_head_arg1_head = ATgetFirst(arg_arg1_head_arg1_list);
                                                                                                                                                                                                                                                                                    arg_arg1_head_arg1_list = ATgetNext(arg_arg1_head_arg1_list);
                                                                                                                                                                                                                                                                                    if (ATgetInt((ATermInt)arg_arg1_head_arg1_head) == 114) {

                                                                                                                                                                                                                                                                                      if (ATgetType((ATerm)arg_arg1_head_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg1_list) == ATfalse) {
                                                                                                                                                                                                                                                                                        arg_arg1_head_arg1_head = ATgetFirst(arg_arg1_head_arg1_list);
                                                                                                                                                                                                                                                                                        arg_arg1_head_arg1_list = ATgetNext(arg_arg1_head_arg1_list);
                                                                                                                                                                                                                                                                                        if (ATgetInt((ATermInt)arg_arg1_head_arg1_head) == 121) {

                                                                                                                                                                                                                                                                                          if (ATgetType((ATerm)arg_arg1_head_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg1_list) == ATtrue) {
                                                                                                                                                                                                                                                                                            if (ATgetType((ATerm)arg_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_list) == ATfalse) {
                                                                                                                                                                                                                                                                                              arg_arg1_list = ATgetNext(arg_arg1_list);
                                                                                                                                                                                                                                                                                              if (ATgetType((ATerm)arg_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_list) == ATfalse) {
                                                                                                                                                                                                                                                                                                arg_arg1_head = ATgetFirst(arg_arg1_list);
                                                                                                                                                                                                                                                                                                arg_arg1_list = ATgetNext(arg_arg1_list);
                                                                                                                                                                                                                                                                                                /* checking for: appl */
                                                                                                                                                                                                                                                                                                if (ATgetType((ATerm)arg_arg1_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head) == PERR_afun0) {
                                                                                                                                                                                                                                                                                                  ATerm arg_arg1_head_arg0 = ATgetArgument(arg_arg1_head, 0);
                                                                                                                                                                                                                                                                                                  /* checking for: prod */
                                                                                                                                                                                                                                                                                                  if (ATgetType((ATerm)arg_arg1_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0) == PERR_afun1) {
                                                                                                                                                                                                                                                                                                    ATerm arg_arg1_head_arg0_arg0 = ATgetArgument(arg_arg1_head_arg0, 0);
                                                                                                                                                                                                                                                                                                    if (ATgetType((ATerm)arg_arg1_head_arg0_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0) == ATfalse) {
                                                                                                                                                                                                                                                                                                      ATerm arg_arg1_head_arg0_arg0_head;
                                                                                                                                                                                                                                                                                                      ATermList arg_arg1_head_arg0_arg0_list = (ATermList)arg_arg1_head_arg0_arg0;
                                                                                                                                                                                                                                                                                                      arg_arg1_head_arg0_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_list);
                                                                                                                                                                                                                                                                                                      arg_arg1_head_arg0_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_list);
                                                                                                                                                                                                                                                                                                      /* checking for: char-class */
                                                                                                                                                                                                                                                                                                      if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg0_head) == PERR_afun17) {
                                                                                                                                                                                                                                                                                                        ATerm arg_arg1_head_arg0_arg0_head_arg0 = ATgetArgument(arg_arg1_head_arg0_arg0_head, 0);
                                                                                                                                                                                                                                                                                                        if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0) == ATfalse) {
                                                                                                                                                                                                                                                                                                          ATerm arg_arg1_head_arg0_arg0_head_arg0_head;
                                                                                                                                                                                                                                                                                                          ATermList arg_arg1_head_arg0_arg0_head_arg0_list = (ATermList)arg_arg1_head_arg0_arg0_head_arg0;
                                                                                                                                                                                                                                                                                                          arg_arg1_head_arg0_arg0_head_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                                                                                                                                                                                                                          arg_arg1_head_arg0_arg0_head_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                                                                                                                                                                                                                          if (ATgetInt((ATermInt)arg_arg1_head_arg0_arg0_head_arg0_head) == 40) {

                                                                                                                                                                                                                                                                                                            if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0_list) == ATtrue) {
                                                                                                                                                                                                                                                                                                              if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_list) == ATtrue) {
                                                                                                                                                                                                                                                                                                                ATerm arg_arg1_head_arg0_arg1 = ATgetArgument(arg_arg1_head_arg0, 1);
                                                                                                                                                                                                                                                                                                                /* checking for: lit */
                                                                                                                                                                                                                                                                                                                if (ATgetType((ATerm)arg_arg1_head_arg0_arg1) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg1) == PERR_afun9) {
                                                                                                                                                                                                                                                                                                                  ATerm arg_arg1_head_arg0_arg1_arg0 = ATgetArgument(arg_arg1_head_arg0_arg1, 0);
                                                                                                                                                                                                                                                                                                                  /* checking for: ( */
                                                                                                                                                                                                                                                                                                                  if (ATgetType((ATerm)arg_arg1_head_arg0_arg1_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg1_arg0) == PERR_afun39) {
                                                                                                                                                                                                                                                                                                                    ATerm arg_arg1_head_arg0_arg2 = ATgetArgument(arg_arg1_head_arg0, 2);
                                                                                                                                                                                                                                                                                                                    /* checking for: no-attrs */
                                                                                                                                                                                                                                                                                                                    if (ATgetType((ATerm)arg_arg1_head_arg0_arg2) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg2) == PERR_afun5) {
                                                                                                                                                                                                                                                                                                                      ATerm arg_arg1_head_arg1 = ATgetArgument(arg_arg1_head, 1);
                                                                                                                                                                                                                                                                                                                      if (ATgetType((ATerm)arg_arg1_head_arg1) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg1) == ATfalse) {
                                                                                                                                                                                                                                                                                                                        ATerm arg_arg1_head_arg1_head;
                                                                                                                                                                                                                                                                                                                        ATermList arg_arg1_head_arg1_list = (ATermList)arg_arg1_head_arg1;
                                                                                                                                                                                                                                                                                                                        arg_arg1_head_arg1_head = ATgetFirst(arg_arg1_head_arg1_list);
                                                                                                                                                                                                                                                                                                                        arg_arg1_head_arg1_list = ATgetNext(arg_arg1_head_arg1_list);
                                                                                                                                                                                                                                                                                                                        if (ATgetInt((ATermInt)arg_arg1_head_arg1_head) == 40) {

                                                                                                                                                                                                                                                                                                                          if (ATgetType((ATerm)arg_arg1_head_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg1_list) == ATtrue) {
                                                                                                                                                                                                                                                                                                                            if (ATgetType((ATerm)arg_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_list) == ATfalse) {
                                                                                                                                                                                                                                                                                                                              arg_arg1_list = ATgetNext(arg_arg1_list);
                                                                                                                                                                                                                                                                                                                              if (ATgetType((ATerm)arg_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_list) == ATfalse) {
                                                                                                                                                                                                                                                                                                                                arg_arg1_list = ATgetNext(arg_arg1_list);
                                                                                                                                                                                                                                                                                                                                if (ATgetType((ATerm)arg_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_list) == ATfalse) {
                                                                                                                                                                                                                                                                                                                                  arg_arg1_list = ATgetNext(arg_arg1_list);
                                                                                                                                                                                                                                                                                                                                  if (ATgetType((ATerm)arg_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_list) == ATfalse) {
                                                                                                                                                                                                                                                                                                                                    arg_arg1_head = ATgetFirst(arg_arg1_list);
                                                                                                                                                                                                                                                                                                                                    arg_arg1_list = ATgetNext(arg_arg1_list);
                                                                                                                                                                                                                                                                                                                                    /* checking for: appl */
                                                                                                                                                                                                                                                                                                                                    if (ATgetType((ATerm)arg_arg1_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head) == PERR_afun0) {
                                                                                                                                                                                                                                                                                                                                      ATerm arg_arg1_head_arg0 = ATgetArgument(arg_arg1_head, 0);
                                                                                                                                                                                                                                                                                                                                      /* checking for: prod */
                                                                                                                                                                                                                                                                                                                                      if (ATgetType((ATerm)arg_arg1_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0) == PERR_afun1) {
                                                                                                                                                                                                                                                                                                                                        ATerm arg_arg1_head_arg0_arg0 = ATgetArgument(arg_arg1_head_arg0, 0);
                                                                                                                                                                                                                                                                                                                                        if (ATgetType((ATerm)arg_arg1_head_arg0_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0) == ATfalse) {
                                                                                                                                                                                                                                                                                                                                          ATerm arg_arg1_head_arg0_arg0_head;
                                                                                                                                                                                                                                                                                                                                          ATermList arg_arg1_head_arg0_arg0_list = (ATermList)arg_arg1_head_arg0_arg0;
                                                                                                                                                                                                                                                                                                                                          arg_arg1_head_arg0_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_list);
                                                                                                                                                                                                                                                                                                                                          arg_arg1_head_arg0_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_list);
                                                                                                                                                                                                                                                                                                                                          /* checking for: char-class */
                                                                                                                                                                                                                                                                                                                                          if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg0_head) == PERR_afun17) {
                                                                                                                                                                                                                                                                                                                                            ATerm arg_arg1_head_arg0_arg0_head_arg0 = ATgetArgument(arg_arg1_head_arg0_arg0_head, 0);
                                                                                                                                                                                                                                                                                                                                            if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0) == ATfalse) {
                                                                                                                                                                                                                                                                                                                                              ATerm arg_arg1_head_arg0_arg0_head_arg0_head;
                                                                                                                                                                                                                                                                                                                                              ATermList arg_arg1_head_arg0_arg0_head_arg0_list = (ATermList)arg_arg1_head_arg0_arg0_head_arg0;
                                                                                                                                                                                                                                                                                                                                              arg_arg1_head_arg0_arg0_head_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                                                                                                                                                                                                                                                              arg_arg1_head_arg0_arg0_head_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                                                                                                                                                                                                                                                              if (ATgetInt((ATermInt)arg_arg1_head_arg0_arg0_head_arg0_head) == 44) {

                                                                                                                                                                                                                                                                                                                                                if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0_list) == ATtrue) {
                                                                                                                                                                                                                                                                                                                                                  if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_list) == ATtrue) {
                                                                                                                                                                                                                                                                                                                                                    ATerm arg_arg1_head_arg0_arg1 = ATgetArgument(arg_arg1_head_arg0, 1);
                                                                                                                                                                                                                                                                                                                                                    /* checking for: lit */
                                                                                                                                                                                                                                                                                                                                                    if (ATgetType((ATerm)arg_arg1_head_arg0_arg1) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg1) == PERR_afun9) {
                                                                                                                                                                                                                                                                                                                                                      ATerm arg_arg1_head_arg0_arg1_arg0 = ATgetArgument(arg_arg1_head_arg0_arg1, 0);
                                                                                                                                                                                                                                                                                                                                                      /* checking for: , */
                                                                                                                                                                                                                                                                                                                                                      if (ATgetType((ATerm)arg_arg1_head_arg0_arg1_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg1_arg0) == PERR_afun43) {
                                                                                                                                                                                                                                                                                                                                                        ATerm arg_arg1_head_arg0_arg2 = ATgetArgument(arg_arg1_head_arg0, 2);
                                                                                                                                                                                                                                                                                                                                                        /* checking for: no-attrs */
                                                                                                                                                                                                                                                                                                                                                        if (ATgetType((ATerm)arg_arg1_head_arg0_arg2) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg2) == PERR_afun5) {
                                                                                                                                                                                                                                                                                                                                                          ATerm arg_arg1_head_arg1 = ATgetArgument(arg_arg1_head, 1);
                                                                                                                                                                                                                                                                                                                                                          if (ATgetType((ATerm)arg_arg1_head_arg1) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg1) == ATfalse) {
                                                                                                                                                                                                                                                                                                                                                            ATerm arg_arg1_head_arg1_head;
                                                                                                                                                                                                                                                                                                                                                            ATermList arg_arg1_head_arg1_list = (ATermList)arg_arg1_head_arg1;
                                                                                                                                                                                                                                                                                                                                                            arg_arg1_head_arg1_head = ATgetFirst(arg_arg1_head_arg1_list);
                                                                                                                                                                                                                                                                                                                                                            arg_arg1_head_arg1_list = ATgetNext(arg_arg1_head_arg1_list);
                                                                                                                                                                                                                                                                                                                                                            if (ATgetInt((ATermInt)arg_arg1_head_arg1_head) == 44) {

                                                                                                                                                                                                                                                                                                                                                              if (ATgetType((ATerm)arg_arg1_head_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg1_list) == ATtrue) {
                                                                                                                                                                                                                                                                                                                                                                if (ATgetType((ATerm)arg_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_list) == ATfalse) {
                                                                                                                                                                                                                                                                                                                                                                  arg_arg1_list = ATgetNext(arg_arg1_list);
                                                                                                                                                                                                                                                                                                                                                                  if (ATgetType((ATerm)arg_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_list) == ATfalse) {
                                                                                                                                                                                                                                                                                                                                                                    arg_arg1_list = ATgetNext(arg_arg1_list);
                                                                                                                                                                                                                                                                                                                                                                    if (ATgetType((ATerm)arg_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_list) == ATfalse) {
                                                                                                                                                                                                                                                                                                                                                                      arg_arg1_list = ATgetNext(arg_arg1_list);
                                                                                                                                                                                                                                                                                                                                                                      if (ATgetType((ATerm)arg_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_list) == ATfalse) {
                                                                                                                                                                                                                                                                                                                                                                        arg_arg1_head = ATgetFirst(arg_arg1_list);
                                                                                                                                                                                                                                                                                                                                                                        arg_arg1_list = ATgetNext(arg_arg1_list);
                                                                                                                                                                                                                                                                                                                                                                        /* checking for: appl */
                                                                                                                                                                                                                                                                                                                                                                        if (ATgetType((ATerm)arg_arg1_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head) == PERR_afun0) {
                                                                                                                                                                                                                                                                                                                                                                          ATerm arg_arg1_head_arg0 = ATgetArgument(arg_arg1_head, 0);
                                                                                                                                                                                                                                                                                                                                                                          /* checking for: prod */
                                                                                                                                                                                                                                                                                                                                                                          if (ATgetType((ATerm)arg_arg1_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0) == PERR_afun1) {
                                                                                                                                                                                                                                                                                                                                                                            ATerm arg_arg1_head_arg0_arg0 = ATgetArgument(arg_arg1_head_arg0, 0);
                                                                                                                                                                                                                                                                                                                                                                            if (ATgetType((ATerm)arg_arg1_head_arg0_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0) == ATfalse) {
                                                                                                                                                                                                                                                                                                                                                                              ATerm arg_arg1_head_arg0_arg0_head;
                                                                                                                                                                                                                                                                                                                                                                              ATermList arg_arg1_head_arg0_arg0_list = (ATermList)arg_arg1_head_arg0_arg0;
                                                                                                                                                                                                                                                                                                                                                                              arg_arg1_head_arg0_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_list);
                                                                                                                                                                                                                                                                                                                                                                              arg_arg1_head_arg0_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_list);
                                                                                                                                                                                                                                                                                                                                                                              /* checking for: char-class */
                                                                                                                                                                                                                                                                                                                                                                              if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg0_head) == PERR_afun17) {
                                                                                                                                                                                                                                                                                                                                                                                ATerm arg_arg1_head_arg0_arg0_head_arg0 = ATgetArgument(arg_arg1_head_arg0_arg0_head, 0);
                                                                                                                                                                                                                                                                                                                                                                                if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0) == ATfalse) {
                                                                                                                                                                                                                                                                                                                                                                                  ATerm arg_arg1_head_arg0_arg0_head_arg0_head;
                                                                                                                                                                                                                                                                                                                                                                                  ATermList arg_arg1_head_arg0_arg0_head_arg0_list = (ATermList)arg_arg1_head_arg0_arg0_head_arg0;
                                                                                                                                                                                                                                                                                                                                                                                  arg_arg1_head_arg0_arg0_head_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                                                                                                                                                                                                                                                                                                  arg_arg1_head_arg0_arg0_head_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                                                                                                                                                                                                                                                                                                  if (ATgetInt((ATermInt)arg_arg1_head_arg0_arg0_head_arg0_head) == 44) {

                                                                                                                                                                                                                                                                                                                                                                                    if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0_list) == ATtrue) {
                                                                                                                                                                                                                                                                                                                                                                                      if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_list) == ATtrue) {
                                                                                                                                                                                                                                                                                                                                                                                        ATerm arg_arg1_head_arg0_arg1 = ATgetArgument(arg_arg1_head_arg0, 1);
                                                                                                                                                                                                                                                                                                                                                                                        /* checking for: lit */
                                                                                                                                                                                                                                                                                                                                                                                        if (ATgetType((ATerm)arg_arg1_head_arg0_arg1) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg1) == PERR_afun9) {
                                                                                                                                                                                                                                                                                                                                                                                          ATerm arg_arg1_head_arg0_arg1_arg0 = ATgetArgument(arg_arg1_head_arg0_arg1, 0);
                                                                                                                                                                                                                                                                                                                                                                                          /* checking for: , */
                                                                                                                                                                                                                                                                                                                                                                                          if (ATgetType((ATerm)arg_arg1_head_arg0_arg1_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg1_arg0) == PERR_afun43) {
                                                                                                                                                                                                                                                                                                                                                                                            ATerm arg_arg1_head_arg0_arg2 = ATgetArgument(arg_arg1_head_arg0, 2);
                                                                                                                                                                                                                                                                                                                                                                                            /* checking for: no-attrs */
                                                                                                                                                                                                                                                                                                                                                                                            if (ATgetType((ATerm)arg_arg1_head_arg0_arg2) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg2) == PERR_afun5) {
                                                                                                                                                                                                                                                                                                                                                                                              ATerm arg_arg1_head_arg1 = ATgetArgument(arg_arg1_head, 1);
                                                                                                                                                                                                                                                                                                                                                                                              if (ATgetType((ATerm)arg_arg1_head_arg1) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg1) == ATfalse) {
                                                                                                                                                                                                                                                                                                                                                                                                ATerm arg_arg1_head_arg1_head;
                                                                                                                                                                                                                                                                                                                                                                                                ATermList arg_arg1_head_arg1_list = (ATermList)arg_arg1_head_arg1;
                                                                                                                                                                                                                                                                                                                                                                                                arg_arg1_head_arg1_head = ATgetFirst(arg_arg1_head_arg1_list);
                                                                                                                                                                                                                                                                                                                                                                                                arg_arg1_head_arg1_list = ATgetNext(arg_arg1_head_arg1_list);
                                                                                                                                                                                                                                                                                                                                                                                                if (ATgetInt((ATermInt)arg_arg1_head_arg1_head) == 44) {

                                                                                                                                                                                                                                                                                                                                                                                                  if (ATgetType((ATerm)arg_arg1_head_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg1_list) == ATtrue) {
                                                                                                                                                                                                                                                                                                                                                                                                    if (ATgetType((ATerm)arg_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_list) == ATfalse) {
                                                                                                                                                                                                                                                                                                                                                                                                      arg_arg1_list = ATgetNext(arg_arg1_list);
                                                                                                                                                                                                                                                                                                                                                                                                      if (ATgetType((ATerm)arg_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_list) == ATfalse) {
                                                                                                                                                                                                                                                                                                                                                                                                        arg_arg1_head = ATgetFirst(arg_arg1_list);
                                                                                                                                                                                                                                                                                                                                                                                                        arg_arg1_list = ATgetNext(arg_arg1_list);
                                                                                                                                                                                                                                                                                                                                                                                                        /* checking for: appl */
                                                                                                                                                                                                                                                                                                                                                                                                        if (ATgetType((ATerm)arg_arg1_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head) == PERR_afun0) {
                                                                                                                                                                                                                                                                                                                                                                                                          ATerm arg_arg1_head_arg0 = ATgetArgument(arg_arg1_head, 0);
                                                                                                                                                                                                                                                                                                                                                                                                          /* checking for: prod */
                                                                                                                                                                                                                                                                                                                                                                                                          if (ATgetType((ATerm)arg_arg1_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0) == PERR_afun1) {
                                                                                                                                                                                                                                                                                                                                                                                                            ATerm arg_arg1_head_arg0_arg0 = ATgetArgument(arg_arg1_head_arg0, 0);
                                                                                                                                                                                                                                                                                                                                                                                                            if (ATgetType((ATerm)arg_arg1_head_arg0_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0) == ATfalse) {
                                                                                                                                                                                                                                                                                                                                                                                                              ATerm arg_arg1_head_arg0_arg0_head;
                                                                                                                                                                                                                                                                                                                                                                                                              ATermList arg_arg1_head_arg0_arg0_list = (ATermList)arg_arg1_head_arg0_arg0;
                                                                                                                                                                                                                                                                                                                                                                                                              arg_arg1_head_arg0_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_list);
                                                                                                                                                                                                                                                                                                                                                                                                              arg_arg1_head_arg0_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_list);
                                                                                                                                                                                                                                                                                                                                                                                                              /* checking for: char-class */
                                                                                                                                                                                                                                                                                                                                                                                                              if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg0_head) == PERR_afun17) {
                                                                                                                                                                                                                                                                                                                                                                                                                ATerm arg_arg1_head_arg0_arg0_head_arg0 = ATgetArgument(arg_arg1_head_arg0_arg0_head, 0);
                                                                                                                                                                                                                                                                                                                                                                                                                if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0) == ATfalse) {
                                                                                                                                                                                                                                                                                                                                                                                                                  ATerm arg_arg1_head_arg0_arg0_head_arg0_head;
                                                                                                                                                                                                                                                                                                                                                                                                                  ATermList arg_arg1_head_arg0_arg0_head_arg0_list = (ATermList)arg_arg1_head_arg0_arg0_head_arg0;
                                                                                                                                                                                                                                                                                                                                                                                                                  arg_arg1_head_arg0_arg0_head_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                                                                                                                                                                                                                                                                                                                                  arg_arg1_head_arg0_arg0_head_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                                                                                                                                                                                                                                                                                                                                  if (ATgetInt((ATermInt)arg_arg1_head_arg0_arg0_head_arg0_head) == 91) {

                                                                                                                                                                                                                                                                                                                                                                                                                    if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0_list) == ATtrue) {
                                                                                                                                                                                                                                                                                                                                                                                                                      if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_list) == ATtrue) {
                                                                                                                                                                                                                                                                                                                                                                                                                        ATerm arg_arg1_head_arg0_arg1 = ATgetArgument(arg_arg1_head_arg0, 1);
                                                                                                                                                                                                                                                                                                                                                                                                                        /* checking for: lit */
                                                                                                                                                                                                                                                                                                                                                                                                                        if (ATgetType((ATerm)arg_arg1_head_arg0_arg1) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg1) == PERR_afun9) {
                                                                                                                                                                                                                                                                                                                                                                                                                          ATerm arg_arg1_head_arg0_arg1_arg0 = ATgetArgument(arg_arg1_head_arg0_arg1, 0);
                                                                                                                                                                                                                                                                                                                                                                                                                          /* checking for: [ */
                                                                                                                                                                                                                                                                                                                                                                                                                          if (ATgetType((ATerm)arg_arg1_head_arg0_arg1_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg1_arg0) == PERR_afun47) {
                                                                                                                                                                                                                                                                                                                                                                                                                            ATerm arg_arg1_head_arg0_arg2 = ATgetArgument(arg_arg1_head_arg0, 2);
                                                                                                                                                                                                                                                                                                                                                                                                                            /* checking for: no-attrs */
                                                                                                                                                                                                                                                                                                                                                                                                                            if (ATgetType((ATerm)arg_arg1_head_arg0_arg2) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg2) == PERR_afun5) {
                                                                                                                                                                                                                                                                                                                                                                                                                              ATerm arg_arg1_head_arg1 = ATgetArgument(arg_arg1_head, 1);
                                                                                                                                                                                                                                                                                                                                                                                                                              if (ATgetType((ATerm)arg_arg1_head_arg1) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg1) == ATfalse) {
                                                                                                                                                                                                                                                                                                                                                                                                                                ATerm arg_arg1_head_arg1_head;
                                                                                                                                                                                                                                                                                                                                                                                                                                ATermList arg_arg1_head_arg1_list = (ATermList)arg_arg1_head_arg1;
                                                                                                                                                                                                                                                                                                                                                                                                                                arg_arg1_head_arg1_head = ATgetFirst(arg_arg1_head_arg1_list);
                                                                                                                                                                                                                                                                                                                                                                                                                                arg_arg1_head_arg1_list = ATgetNext(arg_arg1_head_arg1_list);
                                                                                                                                                                                                                                                                                                                                                                                                                                if (ATgetInt((ATermInt)arg_arg1_head_arg1_head) == 91) {

                                                                                                                                                                                                                                                                                                                                                                                                                                  if (ATgetType((ATerm)arg_arg1_head_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg1_list) == ATtrue) {
                                                                                                                                                                                                                                                                                                                                                                                                                                    if (ATgetType((ATerm)arg_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_list) == ATfalse) {
                                                                                                                                                                                                                                                                                                                                                                                                                                      arg_arg1_list = ATgetNext(arg_arg1_list);
                                                                                                                                                                                                                                                                                                                                                                                                                                      if (ATgetType((ATerm)arg_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_list) == ATfalse) {
                                                                                                                                                                                                                                                                                                                                                                                                                                        arg_arg1_head = ATgetFirst(arg_arg1_list);
                                                                                                                                                                                                                                                                                                                                                                                                                                        arg_arg1_list = ATgetNext(arg_arg1_list);
                                                                                                                                                                                                                                                                                                                                                                                                                                        /* checking for: appl */
                                                                                                                                                                                                                                                                                                                                                                                                                                        if (ATgetType((ATerm)arg_arg1_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head) == PERR_afun0) {
                                                                                                                                                                                                                                                                                                                                                                                                                                          ATerm arg_arg1_head_arg0 = ATgetArgument(arg_arg1_head, 0);
                                                                                                                                                                                                                                                                                                                                                                                                                                          /* checking for: list */
                                                                                                                                                                                                                                                                                                                                                                                                                                          if (ATgetType((ATerm)arg_arg1_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0) == PERR_afun8) {
                                                                                                                                                                                                                                                                                                                                                                                                                                            ATerm arg_arg1_head_arg0_arg0 = ATgetArgument(arg_arg1_head_arg0, 0);
                                                                                                                                                                                                                                                                                                                                                                                                                                            /* checking for: cf */
                                                                                                                                                                                                                                                                                                                                                                                                                                            if (ATgetType((ATerm)arg_arg1_head_arg0_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg0) == PERR_afun2) {
                                                                                                                                                                                                                                                                                                                                                                                                                                              ATerm arg_arg1_head_arg0_arg0_arg0 = ATgetArgument(arg_arg1_head_arg0_arg0, 0);
                                                                                                                                                                                                                                                                                                                                                                                                                                              /* checking for: iter-star-sep */
                                                                                                                                                                                                                                                                                                                                                                                                                                              if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg0_arg0) == PERR_afun46) {
                                                                                                                                                                                                                                                                                                                                                                                                                                                ATerm arg_arg1_head_arg0_arg0_arg0_arg0 = ATgetArgument(arg_arg1_head_arg0_arg0_arg0, 0);
                                                                                                                                                                                                                                                                                                                                                                                                                                                /* checking for: sort */
                                                                                                                                                                                                                                                                                                                                                                                                                                                if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_arg0_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg0_arg0_arg0) == PERR_afun11) {
                                                                                                                                                                                                                                                                                                                                                                                                                                                  ATerm arg_arg1_head_arg0_arg0_arg0_arg0_arg0 = ATgetArgument(arg_arg1_head_arg0_arg0_arg0_arg0, 0);
                                                                                                                                                                                                                                                                                                                                                                                                                                                  /* checking for: Error */
                                                                                                                                                                                                                                                                                                                                                                                                                                                  if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_arg0_arg0_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg0_arg0_arg0_arg0) == PERR_afun37) {
                                                                                                                                                                                                                                                                                                                                                                                                                                                    ATerm arg_arg1_head_arg0_arg0_arg0_arg1 = ATgetArgument(arg_arg1_head_arg0_arg0_arg0, 1);
                                                                                                                                                                                                                                                                                                                                                                                                                                                    /* checking for: lit */
                                                                                                                                                                                                                                                                                                                                                                                                                                                    if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_arg0_arg1) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg0_arg0_arg1) == PERR_afun9) {
                                                                                                                                                                                                                                                                                                                                                                                                                                                      ATerm arg_arg1_head_arg0_arg0_arg0_arg1_arg0 = ATgetArgument(arg_arg1_head_arg0_arg0_arg0_arg1, 0);
                                                                                                                                                                                                                                                                                                                                                                                                                                                      /* checking for: , */
                                                                                                                                                                                                                                                                                                                                                                                                                                                      if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_arg0_arg1_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg0_arg0_arg1_arg0) == PERR_afun43) {
                                                                                                                                                                                                                                                                                                                                                                                                                                                        if (ATgetType((ATerm)arg_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_list) == ATfalse) {
                                                                                                                                                                                                                                                                                                                                                                                                                                                          arg_arg1_list = ATgetNext(arg_arg1_list);
                                                                                                                                                                                                                                                                                                                                                                                                                                                          if (ATgetType((ATerm)arg_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_list) == ATfalse) {
                                                                                                                                                                                                                                                                                                                                                                                                                                                            arg_arg1_head = ATgetFirst(arg_arg1_list);
                                                                                                                                                                                                                                                                                                                                                                                                                                                            arg_arg1_list = ATgetNext(arg_arg1_list);
                                                                                                                                                                                                                                                                                                                                                                                                                                                            /* checking for: appl */
                                                                                                                                                                                                                                                                                                                                                                                                                                                            if (ATgetType((ATerm)arg_arg1_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head) == PERR_afun0) {
                                                                                                                                                                                                                                                                                                                                                                                                                                                              ATerm arg_arg1_head_arg0 = ATgetArgument(arg_arg1_head, 0);
                                                                                                                                                                                                                                                                                                                                                                                                                                                              /* checking for: prod */
                                                                                                                                                                                                                                                                                                                                                                                                                                                              if (ATgetType((ATerm)arg_arg1_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0) == PERR_afun1) {
                                                                                                                                                                                                                                                                                                                                                                                                                                                                ATerm arg_arg1_head_arg0_arg0 = ATgetArgument(arg_arg1_head_arg0, 0);
                                                                                                                                                                                                                                                                                                                                                                                                                                                                if (ATgetType((ATerm)arg_arg1_head_arg0_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0) == ATfalse) {
                                                                                                                                                                                                                                                                                                                                                                                                                                                                  ATerm arg_arg1_head_arg0_arg0_head;
                                                                                                                                                                                                                                                                                                                                                                                                                                                                  ATermList arg_arg1_head_arg0_arg0_list = (ATermList)arg_arg1_head_arg0_arg0;
                                                                                                                                                                                                                                                                                                                                                                                                                                                                  arg_arg1_head_arg0_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_list);
                                                                                                                                                                                                                                                                                                                                                                                                                                                                  arg_arg1_head_arg0_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_list);
                                                                                                                                                                                                                                                                                                                                                                                                                                                                  /* checking for: char-class */
                                                                                                                                                                                                                                                                                                                                                                                                                                                                  if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg0_head) == PERR_afun17) {
                                                                                                                                                                                                                                                                                                                                                                                                                                                                    ATerm arg_arg1_head_arg0_arg0_head_arg0 = ATgetArgument(arg_arg1_head_arg0_arg0_head, 0);
                                                                                                                                                                                                                                                                                                                                                                                                                                                                    if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0) == ATfalse) {
                                                                                                                                                                                                                                                                                                                                                                                                                                                                      ATerm arg_arg1_head_arg0_arg0_head_arg0_head;
                                                                                                                                                                                                                                                                                                                                                                                                                                                                      ATermList arg_arg1_head_arg0_arg0_head_arg0_list = (ATermList)arg_arg1_head_arg0_arg0_head_arg0;
                                                                                                                                                                                                                                                                                                                                                                                                                                                                      arg_arg1_head_arg0_arg0_head_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                                                                                                                                                                                                                                                                                                                                                                                      arg_arg1_head_arg0_arg0_head_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                                                                                                                                                                                                                                                                                                                                                                                      if (ATgetInt((ATermInt)arg_arg1_head_arg0_arg0_head_arg0_head) == 93) {

                                                                                                                                                                                                                                                                                                                                                                                                                                                                        if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0_list) == ATtrue) {
                                                                                                                                                                                                                                                                                                                                                                                                                                                                          if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_list) == ATtrue) {
                                                                                                                                                                                                                                                                                                                                                                                                                                                                            ATerm arg_arg1_head_arg0_arg1 = ATgetArgument(arg_arg1_head_arg0, 1);
                                                                                                                                                                                                                                                                                                                                                                                                                                                                            /* checking for: lit */
                                                                                                                                                                                                                                                                                                                                                                                                                                                                            if (ATgetType((ATerm)arg_arg1_head_arg0_arg1) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg1) == PERR_afun9) {
                                                                                                                                                                                                                                                                                                                                                                                                                                                                              ATerm arg_arg1_head_arg0_arg1_arg0 = ATgetArgument(arg_arg1_head_arg0_arg1, 0);
                                                                                                                                                                                                                                                                                                                                                                                                                                                                              /* checking for: ] */
                                                                                                                                                                                                                                                                                                                                                                                                                                                                              if (ATgetType((ATerm)arg_arg1_head_arg0_arg1_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg1_arg0) == PERR_afun45) {
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                ATerm arg_arg1_head_arg0_arg2 = ATgetArgument(arg_arg1_head_arg0, 2);
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                /* checking for: no-attrs */
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                if (ATgetType((ATerm)arg_arg1_head_arg0_arg2) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg2) == PERR_afun5) {
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                  ATerm arg_arg1_head_arg1 = ATgetArgument(arg_arg1_head, 1);
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                  if (ATgetType((ATerm)arg_arg1_head_arg1) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg1) == ATfalse) {
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                    ATerm arg_arg1_head_arg1_head;
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                    ATermList arg_arg1_head_arg1_list = (ATermList)arg_arg1_head_arg1;
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                    arg_arg1_head_arg1_head = ATgetFirst(arg_arg1_head_arg1_list);
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                    arg_arg1_head_arg1_list = ATgetNext(arg_arg1_head_arg1_list);
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                    if (ATgetInt((ATermInt)arg_arg1_head_arg1_head) == 93) {

                                                                                                                                                                                                                                                                                                                                                                                                                                                                                      if (ATgetType((ATerm)arg_arg1_head_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg1_list) == ATtrue) {
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                        if (ATgetType((ATerm)arg_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_list) == ATfalse) {
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                          arg_arg1_list = ATgetNext(arg_arg1_list);
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                          if (ATgetType((ATerm)arg_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_list) == ATfalse) {
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                            arg_arg1_head = ATgetFirst(arg_arg1_list);
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                            arg_arg1_list = ATgetNext(arg_arg1_list);
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                            /* checking for: appl */
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                            if (ATgetType((ATerm)arg_arg1_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head) == PERR_afun0) {
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                              ATerm arg_arg1_head_arg0 = ATgetArgument(arg_arg1_head, 0);
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                              /* checking for: prod */
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                              if (ATgetType((ATerm)arg_arg1_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0) == PERR_afun1) {
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                ATerm arg_arg1_head_arg0_arg0 = ATgetArgument(arg_arg1_head_arg0, 0);
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                if (ATgetType((ATerm)arg_arg1_head_arg0_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0) == ATfalse) {
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                  ATerm arg_arg1_head_arg0_arg0_head;
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                  ATermList arg_arg1_head_arg0_arg0_list = (ATermList)arg_arg1_head_arg0_arg0;
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                  arg_arg1_head_arg0_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_list);
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                  arg_arg1_head_arg0_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_list);
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                  /* checking for: char-class */
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                  if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg0_head) == PERR_afun17) {
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                    ATerm arg_arg1_head_arg0_arg0_head_arg0 = ATgetArgument(arg_arg1_head_arg0_arg0_head, 0);
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                    if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0) == ATfalse) {
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                      ATerm arg_arg1_head_arg0_arg0_head_arg0_head;
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                      ATermList arg_arg1_head_arg0_arg0_head_arg0_list = (ATermList)arg_arg1_head_arg0_arg0_head_arg0;
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                      arg_arg1_head_arg0_arg0_head_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                      arg_arg1_head_arg0_arg0_head_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                      if (ATgetInt((ATermInt)arg_arg1_head_arg0_arg0_head_arg0_head) == 41) {

                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                        if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0_list) == ATtrue) {
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                          if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_list) == ATtrue) {
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                            ATerm arg_arg1_head_arg0_arg1 = ATgetArgument(arg_arg1_head_arg0, 1);
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                            /* checking for: lit */
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                            if (ATgetType((ATerm)arg_arg1_head_arg0_arg1) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg1) == PERR_afun9) {
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                              ATerm arg_arg1_head_arg0_arg1_arg0 = ATgetArgument(arg_arg1_head_arg0_arg1, 0);
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                              /* checking for: ) */
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                              if (ATgetType((ATerm)arg_arg1_head_arg0_arg1_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg1_arg0) == PERR_afun38) {
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                ATerm arg_arg1_head_arg0_arg2 = ATgetArgument(arg_arg1_head_arg0, 2);
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                /* checking for: no-attrs */
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                if (ATgetType((ATerm)arg_arg1_head_arg0_arg2) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg2) == PERR_afun5) {
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                  ATerm arg_arg1_head_arg1 = ATgetArgument(arg_arg1_head, 1);
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                  if (ATgetType((ATerm)arg_arg1_head_arg1) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg1) == ATfalse) {
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                    ATerm arg_arg1_head_arg1_head;
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                    ATermList arg_arg1_head_arg1_list = (ATermList)arg_arg1_head_arg1;
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                    arg_arg1_head_arg1_head = ATgetFirst(arg_arg1_head_arg1_list);
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                    arg_arg1_head_arg1_list = ATgetNext(arg_arg1_head_arg1_list);
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                    if (ATgetInt((ATermInt)arg_arg1_head_arg1_head) == 41) {

                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                      if (ATgetType((ATerm)arg_arg1_head_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg1_list) == ATtrue) {
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                        if (ATgetType((ATerm)arg_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_list) == ATtrue) {
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                          return ATtrue;
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                        }
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                      }
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                    }
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                  }
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                }
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                              }
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                            }
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                          }
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                        }
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                      }
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                    }
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                  }
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                }
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                              }
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                            }
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                          }
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                        }
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                      }
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                    }
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                  }
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                }
                                                                                                                                                                                                                                                                                                                                                                                                                                                                              }
                                                                                                                                                                                                                                                                                                                                                                                                                                                                            }
                                                                                                                                                                                                                                                                                                                                                                                                                                                                          }
                                                                                                                                                                                                                                                                                                                                                                                                                                                                        }
                                                                                                                                                                                                                                                                                                                                                                                                                                                                      }
                                                                                                                                                                                                                                                                                                                                                                                                                                                                    }
                                                                                                                                                                                                                                                                                                                                                                                                                                                                  }
                                                                                                                                                                                                                                                                                                                                                                                                                                                                }
                                                                                                                                                                                                                                                                                                                                                                                                                                                              }
                                                                                                                                                                                                                                                                                                                                                                                                                                                            }
                                                                                                                                                                                                                                                                                                                                                                                                                                                          }
                                                                                                                                                                                                                                                                                                                                                                                                                                                        }
                                                                                                                                                                                                                                                                                                                                                                                                                                                      }
                                                                                                                                                                                                                                                                                                                                                                                                                                                    }
                                                                                                                                                                                                                                                                                                                                                                                                                                                  }
                                                                                                                                                                                                                                                                                                                                                                                                                                                }
                                                                                                                                                                                                                                                                                                                                                                                                                                              }
                                                                                                                                                                                                                                                                                                                                                                                                                                            }
                                                                                                                                                                                                                                                                                                                                                                                                                                          }
                                                                                                                                                                                                                                                                                                                                                                                                                                        }
                                                                                                                                                                                                                                                                                                                                                                                                                                      }
                                                                                                                                                                                                                                                                                                                                                                                                                                    }
                                                                                                                                                                                                                                                                                                                                                                                                                                  }
                                                                                                                                                                                                                                                                                                                                                                                                                                }
                                                                                                                                                                                                                                                                                                                                                                                                                              }
                                                                                                                                                                                                                                                                                                                                                                                                                            }
                                                                                                                                                                                                                                                                                                                                                                                                                          }
                                                                                                                                                                                                                                                                                                                                                                                                                        }
                                                                                                                                                                                                                                                                                                                                                                                                                      }
                                                                                                                                                                                                                                                                                                                                                                                                                    }
                                                                                                                                                                                                                                                                                                                                                                                                                  }
                                                                                                                                                                                                                                                                                                                                                                                                                }
                                                                                                                                                                                                                                                                                                                                                                                                              }
                                                                                                                                                                                                                                                                                                                                                                                                            }
                                                                                                                                                                                                                                                                                                                                                                                                          }
                                                                                                                                                                                                                                                                                                                                                                                                        }
                                                                                                                                                                                                                                                                                                                                                                                                      }
                                                                                                                                                                                                                                                                                                                                                                                                    }
                                                                                                                                                                                                                                                                                                                                                                                                  }
                                                                                                                                                                                                                                                                                                                                                                                                }
                                                                                                                                                                                                                                                                                                                                                                                              }
                                                                                                                                                                                                                                                                                                                                                                                            }
                                                                                                                                                                                                                                                                                                                                                                                          }
                                                                                                                                                                                                                                                                                                                                                                                        }
                                                                                                                                                                                                                                                                                                                                                                                      }
                                                                                                                                                                                                                                                                                                                                                                                    }
                                                                                                                                                                                                                                                                                                                                                                                  }
                                                                                                                                                                                                                                                                                                                                                                                }
                                                                                                                                                                                                                                                                                                                                                                              }
                                                                                                                                                                                                                                                                                                                                                                            }
                                                                                                                                                                                                                                                                                                                                                                          }
                                                                                                                                                                                                                                                                                                                                                                        }
                                                                                                                                                                                                                                                                                                                                                                      }
                                                                                                                                                                                                                                                                                                                                                                    }
                                                                                                                                                                                                                                                                                                                                                                  }
                                                                                                                                                                                                                                                                                                                                                                }
                                                                                                                                                                                                                                                                                                                                                              }
                                                                                                                                                                                                                                                                                                                                                            }
                                                                                                                                                                                                                                                                                                                                                          }
                                                                                                                                                                                                                                                                                                                                                        }
                                                                                                                                                                                                                                                                                                                                                      }
                                                                                                                                                                                                                                                                                                                                                    }
                                                                                                                                                                                                                                                                                                                                                  }
                                                                                                                                                                                                                                                                                                                                                }
                                                                                                                                                                                                                                                                                                                                              }
                                                                                                                                                                                                                                                                                                                                            }
                                                                                                                                                                                                                                                                                                                                          }
                                                                                                                                                                                                                                                                                                                                        }
                                                                                                                                                                                                                                                                                                                                      }
                                                                                                                                                                                                                                                                                                                                    }
                                                                                                                                                                                                                                                                                                                                  }
                                                                                                                                                                                                                                                                                                                                }
                                                                                                                                                                                                                                                                                                                              }
                                                                                                                                                                                                                                                                                                                            }
                                                                                                                                                                                                                                                                                                                          }
                                                                                                                                                                                                                                                                                                                        }
                                                                                                                                                                                                                                                                                                                      }
                                                                                                                                                                                                                                                                                                                    }
                                                                                                                                                                                                                                                                                                                  }
                                                                                                                                                                                                                                                                                                                }
                                                                                                                                                                                                                                                                                                              }
                                                                                                                                                                                                                                                                                                            }
                                                                                                                                                                                                                                                                                                          }
                                                                                                                                                                                                                                                                                                        }
                                                                                                                                                                                                                                                                                                      }
                                                                                                                                                                                                                                                                                                    }
                                                                                                                                                                                                                                                                                                  }
                                                                                                                                                                                                                                                                                                }
                                                                                                                                                                                                                                                                                              }
                                                                                                                                                                                                                                                                                            }
                                                                                                                                                                                                                                                                                          }
                                                                                                                                                                                                                                                                                        }
                                                                                                                                                                                                                                                                                      }
                                                                                                                                                                                                                                                                                    }
                                                                                                                                                                                                                                                                                  }
                                                                                                                                                                                                                                                                                }
                                                                                                                                                                                                                                                                              }
                                                                                                                                                                                                                                                                            }
                                                                                                                                                                                                                                                                          }
                                                                                                                                                                                                                                                                        }
                                                                                                                                                                                                                                                                      }
                                                                                                                                                                                                                                                                    }
                                                                                                                                                                                                                                                                  }
                                                                                                                                                                                                                                                                }
                                                                                                                                                                                                                                                              }
                                                                                                                                                                                                                                                            }
                                                                                                                                                                                                                                                          }
                                                                                                                                                                                                                                                        }
                                                                                                                                                                                                                                                      }
                                                                                                                                                                                                                                                    }
                                                                                                                                                                                                                                                  }
                                                                                                                                                                                                                                                }
                                                                                                                                                                                                                                              }
                                                                                                                                                                                                                                            }
                                                                                                                                                                                                                                          }
                                                                                                                                                                                                                                        }
                                                                                                                                                                                                                                      }
                                                                                                                                                                                                                                    }
                                                                                                                                                                                                                                  }
                                                                                                                                                                                                                                }
                                                                                                                                                                                                                              }
                                                                                                                                                                                                                            }
                                                                                                                                                                                                                          }
                                                                                                                                                                                                                        }
                                                                                                                                                                                                                      }
                                                                                                                                                                                                                    }
                                                                                                                                                                                                                  }
                                                                                                                                                                                                                }
                                                                                                                                                                                                              }
                                                                                                                                                                                                            }
                                                                                                                                                                                                          }
                                                                                                                                                                                                        }
                                                                                                                                                                                                      }
                                                                                                                                                                                                    }
                                                                                                                                                                                                  }
                                                                                                                                                                                                }
                                                                                                                                                                                              }
                                                                                                                                                                                            }
                                                                                                                                                                                          }
                                                                                                                                                                                        }
                                                                                                                                                                                      }
                                                                                                                                                                                    }
                                                                                                                                                                                  }
                                                                                                                                                                                }
                                                                                                                                                                              }
                                                                                                                                                                            }
                                                                                                                                                                          }
                                                                                                                                                                        }
                                                                                                                                                                      }
                                                                                                                                                                    }
                                                                                                                                                                  }
                                                                                                                                                                }
                                                                                                                                                              }
                                                                                                                                                            }
                                                                                                                                                          }
                                                                                                                                                        }
                                                                                                                                                      }
                                                                                                                                                    }
                                                                                                                                                  }
                                                                                                                                                }
                                                                                                                                              }
                                                                                                                                            }
                                                                                                                                          }
                                                                                                                                        }
                                                                                                                                      }
                                                                                                                                    }
                                                                                                                                  }
                                                                                                                                }
                                                                                                                              }
                                                                                                                            }
                                                                                                                          }
                                                                                                                        }
                                                                                                                      }
                                                                                                                    }
                                                                                                                  }
                                                                                                                }
                                                                                                              }
                                                                                                            }
                                                                                                          }
                                                                                                        }
                                                                                                      }
                                                                                                    }
                                                                                                  }
                                                                                                }
                                                                                              }
                                                                                            }
                                                                                          }
                                                                                        }
                                                                                      }
                                                                                    }
                                                                                  }
                                                                                }
                                                                              }
                                                                            }
                                                                          }
                                                                        }
                                                                      }
                                                                    }
                                                                  }
                                                                }
                                                              }
                                                            }
                                                          }
                                                        }
                                                      }
                                                    }
                                                  }
                                                }
                                              }
                                            }
                                          }
                                        }
                                      }
                                    }
                                  }
                                }
                              }
                            }
                          }
                        }
                      }
                    }
                  }
                }
              }
            }
          }
        }
      }
    }
  }
  return ATfalse;
}

/**
 * Assert whether a PERR_Summary has a ws-after-summary. 
 * \param[in] arg input PERR_Summary
 * \return ATtrue if the PERR_Summary had a ws-after-summary, or ATfalse otherwise
 */
ATbool PERR_hasSummaryWsAfterSummary(PERR_Summary arg) {
  if (PERR_isSummarySummary(arg)) {
    return ATtrue;
  }
  return ATfalse;
}

/**
 * Assert whether a PERR_Summary has a ws-after-(. 
 * \param[in] arg input PERR_Summary
 * \return ATtrue if the PERR_Summary had a ws-after-(, or ATfalse otherwise
 */
ATbool PERR_hasSummaryWsAfterParenOpen(PERR_Summary arg) {
  if (PERR_isSummarySummary(arg)) {
    return ATtrue;
  }
  return ATfalse;
}

/**
 * Assert whether a PERR_Summary has a producer. 
 * \param[in] arg input PERR_Summary
 * \return ATtrue if the PERR_Summary had a producer, or ATfalse otherwise
 */
ATbool PERR_hasSummaryProducer(PERR_Summary arg) {
  if (PERR_isSummarySummary(arg)) {
    return ATtrue;
  }
  return ATfalse;
}

/**
 * Assert whether a PERR_Summary has a ws-after-producer. 
 * \param[in] arg input PERR_Summary
 * \return ATtrue if the PERR_Summary had a ws-after-producer, or ATfalse otherwise
 */
ATbool PERR_hasSummaryWsAfterProducer(PERR_Summary arg) {
  if (PERR_isSummarySummary(arg)) {
    return ATtrue;
  }
  return ATfalse;
}

/**
 * Assert whether a PERR_Summary has a ws-after-,. 
 * \param[in] arg input PERR_Summary
 * \return ATtrue if the PERR_Summary had a ws-after-,, or ATfalse otherwise
 */
ATbool PERR_hasSummaryWsAfterComma(PERR_Summary arg) {
  if (PERR_isSummarySummary(arg)) {
    return ATtrue;
  }
  return ATfalse;
}

/**
 * Assert whether a PERR_Summary has a id. 
 * \param[in] arg input PERR_Summary
 * \return ATtrue if the PERR_Summary had a id, or ATfalse otherwise
 */
ATbool PERR_hasSummaryId(PERR_Summary arg) {
  if (PERR_isSummarySummary(arg)) {
    return ATtrue;
  }
  return ATfalse;
}

/**
 * Assert whether a PERR_Summary has a ws-after-id. 
 * \param[in] arg input PERR_Summary
 * \return ATtrue if the PERR_Summary had a ws-after-id, or ATfalse otherwise
 */
ATbool PERR_hasSummaryWsAfterId(PERR_Summary arg) {
  if (PERR_isSummarySummary(arg)) {
    return ATtrue;
  }
  return ATfalse;
}

/**
 * Assert whether a PERR_Summary has a ws-after-,-1. 
 * \param[in] arg input PERR_Summary
 * \return ATtrue if the PERR_Summary had a ws-after-,-1, or ATfalse otherwise
 */
ATbool PERR_hasSummaryWsAfterComma1(PERR_Summary arg) {
  if (PERR_isSummarySummary(arg)) {
    return ATtrue;
  }
  return ATfalse;
}

/**
 * Assert whether a PERR_Summary has a ws-after-[. 
 * \param[in] arg input PERR_Summary
 * \return ATtrue if the PERR_Summary had a ws-after-[, or ATfalse otherwise
 */
ATbool PERR_hasSummaryWsAfterBracketOpen(PERR_Summary arg) {
  if (PERR_isSummarySummary(arg)) {
    return ATtrue;
  }
  return ATfalse;
}

/**
 * Assert whether a PERR_Summary has a list. 
 * \param[in] arg input PERR_Summary
 * \return ATtrue if the PERR_Summary had a list, or ATfalse otherwise
 */
ATbool PERR_hasSummaryList(PERR_Summary arg) {
  if (PERR_isSummarySummary(arg)) {
    return ATtrue;
  }
  return ATfalse;
}

/**
 * Assert whether a PERR_Summary has a ws-after-list. 
 * \param[in] arg input PERR_Summary
 * \return ATtrue if the PERR_Summary had a ws-after-list, or ATfalse otherwise
 */
ATbool PERR_hasSummaryWsAfterList(PERR_Summary arg) {
  if (PERR_isSummarySummary(arg)) {
    return ATtrue;
  }
  return ATfalse;
}

/**
 * Assert whether a PERR_Summary has a ws-after-]. 
 * \param[in] arg input PERR_Summary
 * \return ATtrue if the PERR_Summary had a ws-after-], or ATfalse otherwise
 */
ATbool PERR_hasSummaryWsAfterBracketClose(PERR_Summary arg) {
  if (PERR_isSummarySummary(arg)) {
    return ATtrue;
  }
  return ATfalse;
}

/**
 * Get the ws-after-summary PERR_OptLayout of a PERR_Summary. Note that the precondition is that this PERR_Summary actually has a ws-after-summary
 * \param[in] arg input PERR_Summary
 * \return the ws-after-summary of #arg, if it exist or an undefined value if it does not
 */
PERR_OptLayout PERR_getSummaryWsAfterSummary(PERR_Summary arg) {
  
    return (PERR_OptLayout)ATelementAt((ATermList)ATgetArgument((ATermAppl)arg, 1), 1);
}

/**
 * Get the ws-after-( PERR_OptLayout of a PERR_Summary. Note that the precondition is that this PERR_Summary actually has a ws-after-(
 * \param[in] arg input PERR_Summary
 * \return the ws-after-( of #arg, if it exist or an undefined value if it does not
 */
PERR_OptLayout PERR_getSummaryWsAfterParenOpen(PERR_Summary arg) {
  
    return (PERR_OptLayout)ATelementAt((ATermList)ATgetArgument((ATermAppl)arg, 1), 3);
}

/**
 * Get the producer PERR_StrCon of a PERR_Summary. Note that the precondition is that this PERR_Summary actually has a producer
 * \param[in] arg input PERR_Summary
 * \return the producer of #arg, if it exist or an undefined value if it does not
 */
PERR_StrCon PERR_getSummaryProducer(PERR_Summary arg) {
  
    return (PERR_StrCon)ATelementAt((ATermList)ATgetArgument((ATermAppl)arg, 1), 4);
}

/**
 * Get the ws-after-producer PERR_OptLayout of a PERR_Summary. Note that the precondition is that this PERR_Summary actually has a ws-after-producer
 * \param[in] arg input PERR_Summary
 * \return the ws-after-producer of #arg, if it exist or an undefined value if it does not
 */
PERR_OptLayout PERR_getSummaryWsAfterProducer(PERR_Summary arg) {
  
    return (PERR_OptLayout)ATelementAt((ATermList)ATgetArgument((ATermAppl)arg, 1), 5);
}

/**
 * Get the ws-after-, PERR_OptLayout of a PERR_Summary. Note that the precondition is that this PERR_Summary actually has a ws-after-,
 * \param[in] arg input PERR_Summary
 * \return the ws-after-, of #arg, if it exist or an undefined value if it does not
 */
PERR_OptLayout PERR_getSummaryWsAfterComma(PERR_Summary arg) {
  
    return (PERR_OptLayout)ATelementAt((ATermList)ATgetArgument((ATermAppl)arg, 1), 7);
}

/**
 * Get the id PERR_StrCon of a PERR_Summary. Note that the precondition is that this PERR_Summary actually has a id
 * \param[in] arg input PERR_Summary
 * \return the id of #arg, if it exist or an undefined value if it does not
 */
PERR_StrCon PERR_getSummaryId(PERR_Summary arg) {
  
    return (PERR_StrCon)ATelementAt((ATermList)ATgetArgument((ATermAppl)arg, 1), 8);
}

/**
 * Get the ws-after-id PERR_OptLayout of a PERR_Summary. Note that the precondition is that this PERR_Summary actually has a ws-after-id
 * \param[in] arg input PERR_Summary
 * \return the ws-after-id of #arg, if it exist or an undefined value if it does not
 */
PERR_OptLayout PERR_getSummaryWsAfterId(PERR_Summary arg) {
  
    return (PERR_OptLayout)ATelementAt((ATermList)ATgetArgument((ATermAppl)arg, 1), 9);
}

/**
 * Get the ws-after-,-1 PERR_OptLayout of a PERR_Summary. Note that the precondition is that this PERR_Summary actually has a ws-after-,-1
 * \param[in] arg input PERR_Summary
 * \return the ws-after-,-1 of #arg, if it exist or an undefined value if it does not
 */
PERR_OptLayout PERR_getSummaryWsAfterComma1(PERR_Summary arg) {
  
    return (PERR_OptLayout)ATelementAt((ATermList)ATgetArgument((ATermAppl)arg, 1), 11);
}

/**
 * Get the ws-after-[ PERR_OptLayout of a PERR_Summary. Note that the precondition is that this PERR_Summary actually has a ws-after-[
 * \param[in] arg input PERR_Summary
 * \return the ws-after-[ of #arg, if it exist or an undefined value if it does not
 */
PERR_OptLayout PERR_getSummaryWsAfterBracketOpen(PERR_Summary arg) {
  
    return (PERR_OptLayout)ATelementAt((ATermList)ATgetArgument((ATermAppl)arg, 1), 13);
}

/**
 * Get the list PERR_ErrorList of a PERR_Summary. Note that the precondition is that this PERR_Summary actually has a list
 * \param[in] arg input PERR_Summary
 * \return the list of #arg, if it exist or an undefined value if it does not
 */
PERR_ErrorList PERR_getSummaryList(PERR_Summary arg) {
  
    return (PERR_ErrorList)ATgetArgument((ATermAppl)ATelementAt((ATermList)ATgetArgument((ATermAppl)arg, 1), 14), 1);
}

/**
 * Get the ws-after-list PERR_OptLayout of a PERR_Summary. Note that the precondition is that this PERR_Summary actually has a ws-after-list
 * \param[in] arg input PERR_Summary
 * \return the ws-after-list of #arg, if it exist or an undefined value if it does not
 */
PERR_OptLayout PERR_getSummaryWsAfterList(PERR_Summary arg) {
  
    return (PERR_OptLayout)ATelementAt((ATermList)ATgetArgument((ATermAppl)arg, 1), 15);
}

/**
 * Get the ws-after-] PERR_OptLayout of a PERR_Summary. Note that the precondition is that this PERR_Summary actually has a ws-after-]
 * \param[in] arg input PERR_Summary
 * \return the ws-after-] of #arg, if it exist or an undefined value if it does not
 */
PERR_OptLayout PERR_getSummaryWsAfterBracketClose(PERR_Summary arg) {
  
    return (PERR_OptLayout)ATelementAt((ATermList)ATgetArgument((ATermAppl)arg, 1), 17);
}

/**
 * Set the ws-after-summary of a PERR_Summary. The precondition being that this PERR_Summary actually has a ws-after-summary
 * \param[in] arg input PERR_Summary
 * \param[in] wsAfterSummary new PERR_OptLayout to set in #arg
 * \return A new PERR_Summary with wsAfterSummary at the right place, or a core dump if #arg did not have a wsAfterSummary
 */
PERR_Summary PERR_setSummaryWsAfterSummary(PERR_Summary arg, PERR_OptLayout wsAfterSummary) {
  if (PERR_isSummarySummary(arg)) {
    return (PERR_Summary)ATsetArgument((ATermAppl)arg, (ATerm)ATreplace((ATermList)ATgetArgument((ATermAppl)arg, 1), (ATerm)((ATerm) wsAfterSummary), 1), 1);
  }

  ATabort("Summary has no WsAfterSummary: %t\n", arg);
  return (PERR_Summary)NULL;
}

/**
 * Set the ws-after-( of a PERR_Summary. The precondition being that this PERR_Summary actually has a ws-after-(
 * \param[in] arg input PERR_Summary
 * \param[in] wsAfterParenOpen new PERR_OptLayout to set in #arg
 * \return A new PERR_Summary with wsAfterParenOpen at the right place, or a core dump if #arg did not have a wsAfterParenOpen
 */
PERR_Summary PERR_setSummaryWsAfterParenOpen(PERR_Summary arg, PERR_OptLayout wsAfterParenOpen) {
  if (PERR_isSummarySummary(arg)) {
    return (PERR_Summary)ATsetArgument((ATermAppl)arg, (ATerm)ATreplace((ATermList)ATgetArgument((ATermAppl)arg, 1), (ATerm)((ATerm) wsAfterParenOpen), 3), 1);
  }

  ATabort("Summary has no WsAfterParenOpen: %t\n", arg);
  return (PERR_Summary)NULL;
}

/**
 * Set the producer of a PERR_Summary. The precondition being that this PERR_Summary actually has a producer
 * \param[in] arg input PERR_Summary
 * \param[in] producer new PERR_StrCon to set in #arg
 * \return A new PERR_Summary with producer at the right place, or a core dump if #arg did not have a producer
 */
PERR_Summary PERR_setSummaryProducer(PERR_Summary arg, PERR_StrCon producer) {
  if (PERR_isSummarySummary(arg)) {
    return (PERR_Summary)ATsetArgument((ATermAppl)arg, (ATerm)ATreplace((ATermList)ATgetArgument((ATermAppl)arg, 1), (ATerm)((ATerm) producer), 4), 1);
  }

  ATabort("Summary has no Producer: %t\n", arg);
  return (PERR_Summary)NULL;
}

/**
 * Set the ws-after-producer of a PERR_Summary. The precondition being that this PERR_Summary actually has a ws-after-producer
 * \param[in] arg input PERR_Summary
 * \param[in] wsAfterProducer new PERR_OptLayout to set in #arg
 * \return A new PERR_Summary with wsAfterProducer at the right place, or a core dump if #arg did not have a wsAfterProducer
 */
PERR_Summary PERR_setSummaryWsAfterProducer(PERR_Summary arg, PERR_OptLayout wsAfterProducer) {
  if (PERR_isSummarySummary(arg)) {
    return (PERR_Summary)ATsetArgument((ATermAppl)arg, (ATerm)ATreplace((ATermList)ATgetArgument((ATermAppl)arg, 1), (ATerm)((ATerm) wsAfterProducer), 5), 1);
  }

  ATabort("Summary has no WsAfterProducer: %t\n", arg);
  return (PERR_Summary)NULL;
}

/**
 * Set the ws-after-, of a PERR_Summary. The precondition being that this PERR_Summary actually has a ws-after-,
 * \param[in] arg input PERR_Summary
 * \param[in] wsAfterComma new PERR_OptLayout to set in #arg
 * \return A new PERR_Summary with wsAfterComma at the right place, or a core dump if #arg did not have a wsAfterComma
 */
PERR_Summary PERR_setSummaryWsAfterComma(PERR_Summary arg, PERR_OptLayout wsAfterComma) {
  if (PERR_isSummarySummary(arg)) {
    return (PERR_Summary)ATsetArgument((ATermAppl)arg, (ATerm)ATreplace((ATermList)ATgetArgument((ATermAppl)arg, 1), (ATerm)((ATerm) wsAfterComma), 7), 1);
  }

  ATabort("Summary has no WsAfterComma: %t\n", arg);
  return (PERR_Summary)NULL;
}

/**
 * Set the id of a PERR_Summary. The precondition being that this PERR_Summary actually has a id
 * \param[in] arg input PERR_Summary
 * \param[in] id new PERR_StrCon to set in #arg
 * \return A new PERR_Summary with id at the right place, or a core dump if #arg did not have a id
 */
PERR_Summary PERR_setSummaryId(PERR_Summary arg, PERR_StrCon id) {
  if (PERR_isSummarySummary(arg)) {
    return (PERR_Summary)ATsetArgument((ATermAppl)arg, (ATerm)ATreplace((ATermList)ATgetArgument((ATermAppl)arg, 1), (ATerm)((ATerm) id), 8), 1);
  }

  ATabort("Summary has no Id: %t\n", arg);
  return (PERR_Summary)NULL;
}

/**
 * Set the ws-after-id of a PERR_Summary. The precondition being that this PERR_Summary actually has a ws-after-id
 * \param[in] arg input PERR_Summary
 * \param[in] wsAfterId new PERR_OptLayout to set in #arg
 * \return A new PERR_Summary with wsAfterId at the right place, or a core dump if #arg did not have a wsAfterId
 */
PERR_Summary PERR_setSummaryWsAfterId(PERR_Summary arg, PERR_OptLayout wsAfterId) {
  if (PERR_isSummarySummary(arg)) {
    return (PERR_Summary)ATsetArgument((ATermAppl)arg, (ATerm)ATreplace((ATermList)ATgetArgument((ATermAppl)arg, 1), (ATerm)((ATerm) wsAfterId), 9), 1);
  }

  ATabort("Summary has no WsAfterId: %t\n", arg);
  return (PERR_Summary)NULL;
}

/**
 * Set the ws-after-,-1 of a PERR_Summary. The precondition being that this PERR_Summary actually has a ws-after-,-1
 * \param[in] arg input PERR_Summary
 * \param[in] wsAfterComma1 new PERR_OptLayout to set in #arg
 * \return A new PERR_Summary with wsAfterComma1 at the right place, or a core dump if #arg did not have a wsAfterComma1
 */
PERR_Summary PERR_setSummaryWsAfterComma1(PERR_Summary arg, PERR_OptLayout wsAfterComma1) {
  if (PERR_isSummarySummary(arg)) {
    return (PERR_Summary)ATsetArgument((ATermAppl)arg, (ATerm)ATreplace((ATermList)ATgetArgument((ATermAppl)arg, 1), (ATerm)((ATerm) wsAfterComma1), 11), 1);
  }

  ATabort("Summary has no WsAfterComma1: %t\n", arg);
  return (PERR_Summary)NULL;
}

/**
 * Set the ws-after-[ of a PERR_Summary. The precondition being that this PERR_Summary actually has a ws-after-[
 * \param[in] arg input PERR_Summary
 * \param[in] wsAfterBracketOpen new PERR_OptLayout to set in #arg
 * \return A new PERR_Summary with wsAfterBracketOpen at the right place, or a core dump if #arg did not have a wsAfterBracketOpen
 */
PERR_Summary PERR_setSummaryWsAfterBracketOpen(PERR_Summary arg, PERR_OptLayout wsAfterBracketOpen) {
  if (PERR_isSummarySummary(arg)) {
    return (PERR_Summary)ATsetArgument((ATermAppl)arg, (ATerm)ATreplace((ATermList)ATgetArgument((ATermAppl)arg, 1), (ATerm)((ATerm) wsAfterBracketOpen), 13), 1);
  }

  ATabort("Summary has no WsAfterBracketOpen: %t\n", arg);
  return (PERR_Summary)NULL;
}

/**
 * Set the list of a PERR_Summary. The precondition being that this PERR_Summary actually has a list
 * \param[in] arg input PERR_Summary
 * \param[in] list new PERR_ErrorList to set in #arg
 * \return A new PERR_Summary with list at the right place, or a core dump if #arg did not have a list
 */
PERR_Summary PERR_setSummaryList(PERR_Summary arg, PERR_ErrorList list) {
  if (PERR_isSummarySummary(arg)) {
    return (PERR_Summary)ATsetArgument((ATermAppl)arg, (ATerm)ATreplace((ATermList)ATgetArgument((ATermAppl)arg, 1), (ATerm)ATsetArgument((ATermAppl)ATelementAt((ATermList)ATgetArgument((ATermAppl)arg, 1), 14), (ATerm)((ATerm) list), 1), 14), 1);
  }

  ATabort("Summary has no List: %t\n", arg);
  return (PERR_Summary)NULL;
}

/**
 * Set the ws-after-list of a PERR_Summary. The precondition being that this PERR_Summary actually has a ws-after-list
 * \param[in] arg input PERR_Summary
 * \param[in] wsAfterList new PERR_OptLayout to set in #arg
 * \return A new PERR_Summary with wsAfterList at the right place, or a core dump if #arg did not have a wsAfterList
 */
PERR_Summary PERR_setSummaryWsAfterList(PERR_Summary arg, PERR_OptLayout wsAfterList) {
  if (PERR_isSummarySummary(arg)) {
    return (PERR_Summary)ATsetArgument((ATermAppl)arg, (ATerm)ATreplace((ATermList)ATgetArgument((ATermAppl)arg, 1), (ATerm)((ATerm) wsAfterList), 15), 1);
  }

  ATabort("Summary has no WsAfterList: %t\n", arg);
  return (PERR_Summary)NULL;
}

/**
 * Set the ws-after-] of a PERR_Summary. The precondition being that this PERR_Summary actually has a ws-after-]
 * \param[in] arg input PERR_Summary
 * \param[in] wsAfterBracketClose new PERR_OptLayout to set in #arg
 * \return A new PERR_Summary with wsAfterBracketClose at the right place, or a core dump if #arg did not have a wsAfterBracketClose
 */
PERR_Summary PERR_setSummaryWsAfterBracketClose(PERR_Summary arg, PERR_OptLayout wsAfterBracketClose) {
  if (PERR_isSummarySummary(arg)) {
    return (PERR_Summary)ATsetArgument((ATermAppl)arg, (ATerm)ATreplace((ATermList)ATgetArgument((ATermAppl)arg, 1), (ATerm)((ATerm) wsAfterBracketClose), 17), 1);
  }

  ATabort("Summary has no WsAfterBracketClose: %t\n", arg);
  return (PERR_Summary)NULL;
}

/**
 * Assert whether a PERR_SubjectList is any of the valid alternatives, or not. This analysis does not go any deeper than the top level
 * \param[in] arg input PERR_SubjectList
 * \return ATtrue if #arg corresponds to the expected signature, or ATfalse otherwise
 */
ATbool PERR_isValidSubjectList(PERR_SubjectList arg) {
  if (PERR_isSubjectListEmpty(arg)) {
    return ATtrue;
  }
  else if (PERR_isSubjectListSingle(arg)) {
    return ATtrue;
  }
  else if (PERR_isSubjectListMany(arg)) {
    return ATtrue;
  }
  return ATfalse;
}

/**
 * Assert whether a PERR_SubjectList is a empty by checking against the following ATerm pattern: []. May not be used to assert correctness of the PERR_SubjectList
 * \param[in] arg input PERR_SubjectList
 * \return ATtrue if #arg corresponds to the signature of a empty, or ATfalse otherwise
 */
inline ATbool PERR_isSubjectListEmpty(PERR_SubjectList arg){
  if (ATisEmpty((ATermList)arg)) {
    return ATtrue;
  }
  return ATfalse;
}

/**
 * Assert whether a PERR_SubjectList is a single by checking against the following ATerm pattern: [<head(Subject)>]. May not be used to assert correctness of the PERR_SubjectList
 * \param[in] arg input PERR_SubjectList
 * \return ATtrue if #arg corresponds to the signature of a single, or ATfalse otherwise
 */
inline ATbool PERR_isSubjectListSingle(PERR_SubjectList arg){
  if (ATgetType((ATerm)arg) == AT_LIST && ATisEmpty((ATermList)arg) == ATfalse) {
    ATermList arg_list = (ATermList)arg;
    arg_list = ATgetNext(arg_list);
    if (ATgetType((ATerm)arg_list) == AT_LIST && ATisEmpty((ATermList)arg_list) == ATtrue) {
      return ATtrue;
    }
  }
  return ATfalse;
}

/**
 * Assert whether a PERR_SubjectList is a many by checking against the following ATerm pattern: [<head(Subject)>,<ws-after-head("OptLayout")>,appl(prod([char-class([44])],lit(","),no-attrs),[44]),<ws-after-sep("OptLayout")>,<[tail(Subject-list)]>]. May not be used to assert correctness of the PERR_SubjectList
 * \param[in] arg input PERR_SubjectList
 * \return ATtrue if #arg corresponds to the signature of a many, or ATfalse otherwise
 */
inline ATbool PERR_isSubjectListMany(PERR_SubjectList arg){
  if (ATgetType((ATerm)arg) == AT_LIST && ATisEmpty((ATermList)arg) == ATfalse) {
    ATerm arg_head;
    ATermList arg_list = (ATermList)arg;
    arg_list = ATgetNext(arg_list);
    if (ATgetType((ATerm)arg_list) == AT_LIST && ATisEmpty((ATermList)arg_list) == ATfalse) {
      arg_list = ATgetNext(arg_list);
      if (ATgetType((ATerm)arg_list) == AT_LIST && ATisEmpty((ATermList)arg_list) == ATfalse) {
        arg_head = ATgetFirst(arg_list);
        arg_list = ATgetNext(arg_list);
        /* checking for: appl */
        if (ATgetType((ATerm)arg_head) == AT_APPL && ATgetAFun((ATermAppl)arg_head) == PERR_afun0) {
          ATerm arg_head_arg0 = ATgetArgument(arg_head, 0);
          /* checking for: prod */
          if (ATgetType((ATerm)arg_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_head_arg0) == PERR_afun1) {
            ATerm arg_head_arg0_arg0 = ATgetArgument(arg_head_arg0, 0);
            if (ATgetType((ATerm)arg_head_arg0_arg0) == AT_LIST && ATisEmpty((ATermList)arg_head_arg0_arg0) == ATfalse) {
              ATerm arg_head_arg0_arg0_head;
              ATermList arg_head_arg0_arg0_list = (ATermList)arg_head_arg0_arg0;
              arg_head_arg0_arg0_head = ATgetFirst(arg_head_arg0_arg0_list);
              arg_head_arg0_arg0_list = ATgetNext(arg_head_arg0_arg0_list);
              /* checking for: char-class */
              if (ATgetType((ATerm)arg_head_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_head_arg0_arg0_head) == PERR_afun17) {
                ATerm arg_head_arg0_arg0_head_arg0 = ATgetArgument(arg_head_arg0_arg0_head, 0);
                if (ATgetType((ATerm)arg_head_arg0_arg0_head_arg0) == AT_LIST && ATisEmpty((ATermList)arg_head_arg0_arg0_head_arg0) == ATfalse) {
                  ATerm arg_head_arg0_arg0_head_arg0_head;
                  ATermList arg_head_arg0_arg0_head_arg0_list = (ATermList)arg_head_arg0_arg0_head_arg0;
                  arg_head_arg0_arg0_head_arg0_head = ATgetFirst(arg_head_arg0_arg0_head_arg0_list);
                  arg_head_arg0_arg0_head_arg0_list = ATgetNext(arg_head_arg0_arg0_head_arg0_list);
                  if (ATgetInt((ATermInt)arg_head_arg0_arg0_head_arg0_head) == 44) {

                    if (ATgetType((ATerm)arg_head_arg0_arg0_head_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_head_arg0_arg0_head_arg0_list) == ATtrue) {
                      if (ATgetType((ATerm)arg_head_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_head_arg0_arg0_list) == ATtrue) {
                        ATerm arg_head_arg0_arg1 = ATgetArgument(arg_head_arg0, 1);
                        /* checking for: lit */
                        if (ATgetType((ATerm)arg_head_arg0_arg1) == AT_APPL && ATgetAFun((ATermAppl)arg_head_arg0_arg1) == PERR_afun9) {
                          ATerm arg_head_arg0_arg1_arg0 = ATgetArgument(arg_head_arg0_arg1, 0);
                          /* checking for: , */
                          if (ATgetType((ATerm)arg_head_arg0_arg1_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_head_arg0_arg1_arg0) == PERR_afun43) {
                            ATerm arg_head_arg0_arg2 = ATgetArgument(arg_head_arg0, 2);
                            /* checking for: no-attrs */
                            if (ATgetType((ATerm)arg_head_arg0_arg2) == AT_APPL && ATgetAFun((ATermAppl)arg_head_arg0_arg2) == PERR_afun5) {
                              ATerm arg_head_arg1 = ATgetArgument(arg_head, 1);
                              if (ATgetType((ATerm)arg_head_arg1) == AT_LIST && ATisEmpty((ATermList)arg_head_arg1) == ATfalse) {
                                ATerm arg_head_arg1_head;
                                ATermList arg_head_arg1_list = (ATermList)arg_head_arg1;
                                arg_head_arg1_head = ATgetFirst(arg_head_arg1_list);
                                arg_head_arg1_list = ATgetNext(arg_head_arg1_list);
                                if (ATgetInt((ATermInt)arg_head_arg1_head) == 44) {

                                  if (ATgetType((ATerm)arg_head_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_head_arg1_list) == ATtrue) {
                                    if (ATgetType((ATerm)arg_list) == AT_LIST && ATisEmpty((ATermList)arg_list) == ATfalse) {
                                      return ATtrue;
                                    }
                                  }
                                }
                              }
                            }
                          }
                        }
                      }
                    }
                  }
                }
              }
            }
          }
        }
      }
    }
  }
  return ATfalse;
}

/**
 * Assert whether a PERR_SubjectList has a head. 
 * \param[in] arg input PERR_SubjectList
 * \return ATtrue if the PERR_SubjectList had a head, or ATfalse otherwise
 */
ATbool PERR_hasSubjectListHead(PERR_SubjectList arg) {
  if (PERR_isSubjectListSingle(arg)) {
    return ATtrue;
  }
  else if (PERR_isSubjectListMany(arg)) {
    return ATtrue;
  }
  return ATfalse;
}

/**
 * Assert whether a PERR_SubjectList has a ws-after-head. 
 * \param[in] arg input PERR_SubjectList
 * \return ATtrue if the PERR_SubjectList had a ws-after-head, or ATfalse otherwise
 */
ATbool PERR_hasSubjectListWsAfterHead(PERR_SubjectList arg) {
  if (PERR_isSubjectListMany(arg)) {
    return ATtrue;
  }
  return ATfalse;
}

/**
 * Assert whether a PERR_SubjectList has a ws-after-sep. 
 * \param[in] arg input PERR_SubjectList
 * \return ATtrue if the PERR_SubjectList had a ws-after-sep, or ATfalse otherwise
 */
ATbool PERR_hasSubjectListWsAfterSep(PERR_SubjectList arg) {
  if (PERR_isSubjectListMany(arg)) {
    return ATtrue;
  }
  return ATfalse;
}

/**
 * Assert whether a PERR_SubjectList has a tail. 
 * \param[in] arg input PERR_SubjectList
 * \return ATtrue if the PERR_SubjectList had a tail, or ATfalse otherwise
 */
ATbool PERR_hasSubjectListTail(PERR_SubjectList arg) {
  if (PERR_isSubjectListMany(arg)) {
    return ATtrue;
  }
  return ATfalse;
}

/**
 * Returns a list of all but the first element of a PERR_SubjectList. 
 * \param[in] arg input PERR_SubjectList
 * \return A new PERR_SubjectList, without the first element and the separator(s) just after it.)
 */
PERR_SubjectList PERR_getSubjectListTail(PERR_SubjectList arg) {
  assert(!PERR_isSubjectListEmpty(arg) && "getTail on an empty list");
  if (PERR_isSubjectListSingle(arg)) {
    return (PERR_SubjectList) PERR_makeSubjectListEmpty();
  }
  else {
  
    return (PERR_SubjectList)ATgetTail((ATermList)arg, 4);
  }
}

/**
 * Get the head PERR_Subject of a PERR_SubjectList. Note that the precondition is that this PERR_SubjectList actually has a head
 * \param[in] arg input PERR_SubjectList
 * \return the head of #arg, if it exist or an undefined value if it does not
 */
PERR_Subject PERR_getSubjectListHead(PERR_SubjectList arg) {
  if (PERR_isSubjectListSingle(arg)) {
    return (PERR_Subject)ATgetFirst((ATermList)arg);
  }
  else 
    return (PERR_Subject)ATgetFirst((ATermList)arg);
}

/**
 * Get the ws-after-head PERR_OptLayout of a PERR_SubjectList. Note that the precondition is that this PERR_SubjectList actually has a ws-after-head
 * \param[in] arg input PERR_SubjectList
 * \return the ws-after-head of #arg, if it exist or an undefined value if it does not
 */
PERR_OptLayout PERR_getSubjectListWsAfterHead(PERR_SubjectList arg) {
  
    return (PERR_OptLayout)ATelementAt((ATermList)arg, 1);
}

/**
 * Get the ws-after-sep PERR_OptLayout of a PERR_SubjectList. Note that the precondition is that this PERR_SubjectList actually has a ws-after-sep
 * \param[in] arg input PERR_SubjectList
 * \return the ws-after-sep of #arg, if it exist or an undefined value if it does not
 */
PERR_OptLayout PERR_getSubjectListWsAfterSep(PERR_SubjectList arg) {
  
    return (PERR_OptLayout)ATelementAt((ATermList)arg, 3);
}

/**
 * Set the head of a PERR_SubjectList. The precondition being that this PERR_SubjectList actually has a head
 * \param[in] arg input PERR_SubjectList
 * \param[in] head new PERR_Subject to set in #arg
 * \return A new PERR_SubjectList with head at the right place, or a core dump if #arg did not have a head
 */
PERR_SubjectList PERR_setSubjectListHead(PERR_SubjectList arg, PERR_Subject head) {
  if (PERR_isSubjectListSingle(arg)) {
    return (PERR_SubjectList)ATreplace((ATermList)arg, (ATerm)((ATerm) head), 0);
  }
  else if (PERR_isSubjectListMany(arg)) {
    return (PERR_SubjectList)ATreplace((ATermList)arg, (ATerm)((ATerm) head), 0);
  }

  ATabort("SubjectList has no Head: %t\n", arg);
  return (PERR_SubjectList)NULL;
}

/**
 * Set the ws-after-head of a PERR_SubjectList. The precondition being that this PERR_SubjectList actually has a ws-after-head
 * \param[in] arg input PERR_SubjectList
 * \param[in] wsAfterHead new PERR_OptLayout to set in #arg
 * \return A new PERR_SubjectList with wsAfterHead at the right place, or a core dump if #arg did not have a wsAfterHead
 */
PERR_SubjectList PERR_setSubjectListWsAfterHead(PERR_SubjectList arg, PERR_OptLayout wsAfterHead) {
  if (PERR_isSubjectListMany(arg)) {
    return (PERR_SubjectList)ATreplace((ATermList)arg, (ATerm)((ATerm) wsAfterHead), 1);
  }

  ATabort("SubjectList has no WsAfterHead: %t\n", arg);
  return (PERR_SubjectList)NULL;
}

/**
 * Set the ws-after-sep of a PERR_SubjectList. The precondition being that this PERR_SubjectList actually has a ws-after-sep
 * \param[in] arg input PERR_SubjectList
 * \param[in] wsAfterSep new PERR_OptLayout to set in #arg
 * \return A new PERR_SubjectList with wsAfterSep at the right place, or a core dump if #arg did not have a wsAfterSep
 */
PERR_SubjectList PERR_setSubjectListWsAfterSep(PERR_SubjectList arg, PERR_OptLayout wsAfterSep) {
  if (PERR_isSubjectListMany(arg)) {
    return (PERR_SubjectList)ATreplace((ATermList)arg, (ATerm)((ATerm) wsAfterSep), 3);
  }

  ATabort("SubjectList has no WsAfterSep: %t\n", arg);
  return (PERR_SubjectList)NULL;
}

/**
 * Set the tail of a PERR_SubjectList. The precondition being that this PERR_SubjectList actually has a tail
 * \param[in] arg input PERR_SubjectList
 * \param[in] tail new PERR_SubjectList to set in #arg
 * \return A new PERR_SubjectList with tail at the right place, or a core dump if #arg did not have a tail
 */
PERR_SubjectList PERR_setSubjectListTail(PERR_SubjectList arg, PERR_SubjectList tail) {
  if (PERR_isSubjectListMany(arg)) {
    return (PERR_SubjectList)ATreplaceTail((ATermList)arg, (ATermList)((ATerm) tail), 4);
  }

  ATabort("SubjectList has no Tail: %t\n", arg);
  return (PERR_SubjectList)NULL;
}

/**
 * Assert whether a PERR_ErrorList is any of the valid alternatives, or not. This analysis does not go any deeper than the top level
 * \param[in] arg input PERR_ErrorList
 * \return ATtrue if #arg corresponds to the expected signature, or ATfalse otherwise
 */
ATbool PERR_isValidErrorList(PERR_ErrorList arg) {
  if (PERR_isErrorListEmpty(arg)) {
    return ATtrue;
  }
  else if (PERR_isErrorListSingle(arg)) {
    return ATtrue;
  }
  else if (PERR_isErrorListMany(arg)) {
    return ATtrue;
  }
  return ATfalse;
}

/**
 * Assert whether a PERR_ErrorList is a empty by checking against the following ATerm pattern: []. May not be used to assert correctness of the PERR_ErrorList
 * \param[in] arg input PERR_ErrorList
 * \return ATtrue if #arg corresponds to the signature of a empty, or ATfalse otherwise
 */
inline ATbool PERR_isErrorListEmpty(PERR_ErrorList arg){
  if (ATisEmpty((ATermList)arg)) {
    return ATtrue;
  }
  return ATfalse;
}

/**
 * Assert whether a PERR_ErrorList is a single by checking against the following ATerm pattern: [<head(Error)>]. May not be used to assert correctness of the PERR_ErrorList
 * \param[in] arg input PERR_ErrorList
 * \return ATtrue if #arg corresponds to the signature of a single, or ATfalse otherwise
 */
inline ATbool PERR_isErrorListSingle(PERR_ErrorList arg){
  if (ATgetType((ATerm)arg) == AT_LIST && ATisEmpty((ATermList)arg) == ATfalse) {
    ATermList arg_list = (ATermList)arg;
    arg_list = ATgetNext(arg_list);
    if (ATgetType((ATerm)arg_list) == AT_LIST && ATisEmpty((ATermList)arg_list) == ATtrue) {
      return ATtrue;
    }
  }
  return ATfalse;
}

/**
 * Assert whether a PERR_ErrorList is a many by checking against the following ATerm pattern: [<head(Error)>,<ws-after-head("OptLayout")>,appl(prod([char-class([44])],lit(","),no-attrs),[44]),<ws-after-sep("OptLayout")>,<[tail(Error-list)]>]. May not be used to assert correctness of the PERR_ErrorList
 * \param[in] arg input PERR_ErrorList
 * \return ATtrue if #arg corresponds to the signature of a many, or ATfalse otherwise
 */
inline ATbool PERR_isErrorListMany(PERR_ErrorList arg){
  if (ATgetType((ATerm)arg) == AT_LIST && ATisEmpty((ATermList)arg) == ATfalse) {
    ATerm arg_head;
    ATermList arg_list = (ATermList)arg;
    arg_list = ATgetNext(arg_list);
    if (ATgetType((ATerm)arg_list) == AT_LIST && ATisEmpty((ATermList)arg_list) == ATfalse) {
      arg_list = ATgetNext(arg_list);
      if (ATgetType((ATerm)arg_list) == AT_LIST && ATisEmpty((ATermList)arg_list) == ATfalse) {
        arg_head = ATgetFirst(arg_list);
        arg_list = ATgetNext(arg_list);
        /* checking for: appl */
        if (ATgetType((ATerm)arg_head) == AT_APPL && ATgetAFun((ATermAppl)arg_head) == PERR_afun0) {
          ATerm arg_head_arg0 = ATgetArgument(arg_head, 0);
          /* checking for: prod */
          if (ATgetType((ATerm)arg_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_head_arg0) == PERR_afun1) {
            ATerm arg_head_arg0_arg0 = ATgetArgument(arg_head_arg0, 0);
            if (ATgetType((ATerm)arg_head_arg0_arg0) == AT_LIST && ATisEmpty((ATermList)arg_head_arg0_arg0) == ATfalse) {
              ATerm arg_head_arg0_arg0_head;
              ATermList arg_head_arg0_arg0_list = (ATermList)arg_head_arg0_arg0;
              arg_head_arg0_arg0_head = ATgetFirst(arg_head_arg0_arg0_list);
              arg_head_arg0_arg0_list = ATgetNext(arg_head_arg0_arg0_list);
              /* checking for: char-class */
              if (ATgetType((ATerm)arg_head_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_head_arg0_arg0_head) == PERR_afun17) {
                ATerm arg_head_arg0_arg0_head_arg0 = ATgetArgument(arg_head_arg0_arg0_head, 0);
                if (ATgetType((ATerm)arg_head_arg0_arg0_head_arg0) == AT_LIST && ATisEmpty((ATermList)arg_head_arg0_arg0_head_arg0) == ATfalse) {
                  ATerm arg_head_arg0_arg0_head_arg0_head;
                  ATermList arg_head_arg0_arg0_head_arg0_list = (ATermList)arg_head_arg0_arg0_head_arg0;
                  arg_head_arg0_arg0_head_arg0_head = ATgetFirst(arg_head_arg0_arg0_head_arg0_list);
                  arg_head_arg0_arg0_head_arg0_list = ATgetNext(arg_head_arg0_arg0_head_arg0_list);
                  if (ATgetInt((ATermInt)arg_head_arg0_arg0_head_arg0_head) == 44) {

                    if (ATgetType((ATerm)arg_head_arg0_arg0_head_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_head_arg0_arg0_head_arg0_list) == ATtrue) {
                      if (ATgetType((ATerm)arg_head_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_head_arg0_arg0_list) == ATtrue) {
                        ATerm arg_head_arg0_arg1 = ATgetArgument(arg_head_arg0, 1);
                        /* checking for: lit */
                        if (ATgetType((ATerm)arg_head_arg0_arg1) == AT_APPL && ATgetAFun((ATermAppl)arg_head_arg0_arg1) == PERR_afun9) {
                          ATerm arg_head_arg0_arg1_arg0 = ATgetArgument(arg_head_arg0_arg1, 0);
                          /* checking for: , */
                          if (ATgetType((ATerm)arg_head_arg0_arg1_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_head_arg0_arg1_arg0) == PERR_afun43) {
                            ATerm arg_head_arg0_arg2 = ATgetArgument(arg_head_arg0, 2);
                            /* checking for: no-attrs */
                            if (ATgetType((ATerm)arg_head_arg0_arg2) == AT_APPL && ATgetAFun((ATermAppl)arg_head_arg0_arg2) == PERR_afun5) {
                              ATerm arg_head_arg1 = ATgetArgument(arg_head, 1);
                              if (ATgetType((ATerm)arg_head_arg1) == AT_LIST && ATisEmpty((ATermList)arg_head_arg1) == ATfalse) {
                                ATerm arg_head_arg1_head;
                                ATermList arg_head_arg1_list = (ATermList)arg_head_arg1;
                                arg_head_arg1_head = ATgetFirst(arg_head_arg1_list);
                                arg_head_arg1_list = ATgetNext(arg_head_arg1_list);
                                if (ATgetInt((ATermInt)arg_head_arg1_head) == 44) {

                                  if (ATgetType((ATerm)arg_head_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_head_arg1_list) == ATtrue) {
                                    if (ATgetType((ATerm)arg_list) == AT_LIST && ATisEmpty((ATermList)arg_list) == ATfalse) {
                                      return ATtrue;
                                    }
                                  }
                                }
                              }
                            }
                          }
                        }
                      }
                    }
                  }
                }
              }
            }
          }
        }
      }
    }
  }
  return ATfalse;
}

/**
 * Assert whether a PERR_ErrorList has a head. 
 * \param[in] arg input PERR_ErrorList
 * \return ATtrue if the PERR_ErrorList had a head, or ATfalse otherwise
 */
ATbool PERR_hasErrorListHead(PERR_ErrorList arg) {
  if (PERR_isErrorListSingle(arg)) {
    return ATtrue;
  }
  else if (PERR_isErrorListMany(arg)) {
    return ATtrue;
  }
  return ATfalse;
}

/**
 * Assert whether a PERR_ErrorList has a ws-after-head. 
 * \param[in] arg input PERR_ErrorList
 * \return ATtrue if the PERR_ErrorList had a ws-after-head, or ATfalse otherwise
 */
ATbool PERR_hasErrorListWsAfterHead(PERR_ErrorList arg) {
  if (PERR_isErrorListMany(arg)) {
    return ATtrue;
  }
  return ATfalse;
}

/**
 * Assert whether a PERR_ErrorList has a ws-after-sep. 
 * \param[in] arg input PERR_ErrorList
 * \return ATtrue if the PERR_ErrorList had a ws-after-sep, or ATfalse otherwise
 */
ATbool PERR_hasErrorListWsAfterSep(PERR_ErrorList arg) {
  if (PERR_isErrorListMany(arg)) {
    return ATtrue;
  }
  return ATfalse;
}

/**
 * Assert whether a PERR_ErrorList has a tail. 
 * \param[in] arg input PERR_ErrorList
 * \return ATtrue if the PERR_ErrorList had a tail, or ATfalse otherwise
 */
ATbool PERR_hasErrorListTail(PERR_ErrorList arg) {
  if (PERR_isErrorListMany(arg)) {
    return ATtrue;
  }
  return ATfalse;
}

/**
 * Returns a list of all but the first element of a PERR_ErrorList. 
 * \param[in] arg input PERR_ErrorList
 * \return A new PERR_ErrorList, without the first element and the separator(s) just after it.)
 */
PERR_ErrorList PERR_getErrorListTail(PERR_ErrorList arg) {
  assert(!PERR_isErrorListEmpty(arg) && "getTail on an empty list");
  if (PERR_isErrorListSingle(arg)) {
    return (PERR_ErrorList) PERR_makeErrorListEmpty();
  }
  else {
  
    return (PERR_ErrorList)ATgetTail((ATermList)arg, 4);
  }
}

/**
 * Get the head PERR_Error of a PERR_ErrorList. Note that the precondition is that this PERR_ErrorList actually has a head
 * \param[in] arg input PERR_ErrorList
 * \return the head of #arg, if it exist or an undefined value if it does not
 */
PERR_Error PERR_getErrorListHead(PERR_ErrorList arg) {
  if (PERR_isErrorListSingle(arg)) {
    return (PERR_Error)ATgetFirst((ATermList)arg);
  }
  else 
    return (PERR_Error)ATgetFirst((ATermList)arg);
}

/**
 * Get the ws-after-head PERR_OptLayout of a PERR_ErrorList. Note that the precondition is that this PERR_ErrorList actually has a ws-after-head
 * \param[in] arg input PERR_ErrorList
 * \return the ws-after-head of #arg, if it exist or an undefined value if it does not
 */
PERR_OptLayout PERR_getErrorListWsAfterHead(PERR_ErrorList arg) {
  
    return (PERR_OptLayout)ATelementAt((ATermList)arg, 1);
}

/**
 * Get the ws-after-sep PERR_OptLayout of a PERR_ErrorList. Note that the precondition is that this PERR_ErrorList actually has a ws-after-sep
 * \param[in] arg input PERR_ErrorList
 * \return the ws-after-sep of #arg, if it exist or an undefined value if it does not
 */
PERR_OptLayout PERR_getErrorListWsAfterSep(PERR_ErrorList arg) {
  
    return (PERR_OptLayout)ATelementAt((ATermList)arg, 3);
}

/**
 * Set the head of a PERR_ErrorList. The precondition being that this PERR_ErrorList actually has a head
 * \param[in] arg input PERR_ErrorList
 * \param[in] head new PERR_Error to set in #arg
 * \return A new PERR_ErrorList with head at the right place, or a core dump if #arg did not have a head
 */
PERR_ErrorList PERR_setErrorListHead(PERR_ErrorList arg, PERR_Error head) {
  if (PERR_isErrorListSingle(arg)) {
    return (PERR_ErrorList)ATreplace((ATermList)arg, (ATerm)((ATerm) head), 0);
  }
  else if (PERR_isErrorListMany(arg)) {
    return (PERR_ErrorList)ATreplace((ATermList)arg, (ATerm)((ATerm) head), 0);
  }

  ATabort("ErrorList has no Head: %t\n", arg);
  return (PERR_ErrorList)NULL;
}

/**
 * Set the ws-after-head of a PERR_ErrorList. The precondition being that this PERR_ErrorList actually has a ws-after-head
 * \param[in] arg input PERR_ErrorList
 * \param[in] wsAfterHead new PERR_OptLayout to set in #arg
 * \return A new PERR_ErrorList with wsAfterHead at the right place, or a core dump if #arg did not have a wsAfterHead
 */
PERR_ErrorList PERR_setErrorListWsAfterHead(PERR_ErrorList arg, PERR_OptLayout wsAfterHead) {
  if (PERR_isErrorListMany(arg)) {
    return (PERR_ErrorList)ATreplace((ATermList)arg, (ATerm)((ATerm) wsAfterHead), 1);
  }

  ATabort("ErrorList has no WsAfterHead: %t\n", arg);
  return (PERR_ErrorList)NULL;
}

/**
 * Set the ws-after-sep of a PERR_ErrorList. The precondition being that this PERR_ErrorList actually has a ws-after-sep
 * \param[in] arg input PERR_ErrorList
 * \param[in] wsAfterSep new PERR_OptLayout to set in #arg
 * \return A new PERR_ErrorList with wsAfterSep at the right place, or a core dump if #arg did not have a wsAfterSep
 */
PERR_ErrorList PERR_setErrorListWsAfterSep(PERR_ErrorList arg, PERR_OptLayout wsAfterSep) {
  if (PERR_isErrorListMany(arg)) {
    return (PERR_ErrorList)ATreplace((ATermList)arg, (ATerm)((ATerm) wsAfterSep), 3);
  }

  ATabort("ErrorList has no WsAfterSep: %t\n", arg);
  return (PERR_ErrorList)NULL;
}

/**
 * Set the tail of a PERR_ErrorList. The precondition being that this PERR_ErrorList actually has a tail
 * \param[in] arg input PERR_ErrorList
 * \param[in] tail new PERR_ErrorList to set in #arg
 * \return A new PERR_ErrorList with tail at the right place, or a core dump if #arg did not have a tail
 */
PERR_ErrorList PERR_setErrorListTail(PERR_ErrorList arg, PERR_ErrorList tail) {
  if (PERR_isErrorListMany(arg)) {
    return (PERR_ErrorList)ATreplaceTail((ATermList)arg, (ATermList)((ATerm) tail), 4);
  }

  ATabort("ErrorList has no Tail: %t\n", arg);
  return (PERR_ErrorList)NULL;
}

/**
 * Assert whether a PERR_Location is any of the valid alternatives, or not. This analysis does not go any deeper than the top level
 * \param[in] arg input PERR_Location
 * \return ATtrue if #arg corresponds to the expected signature, or ATfalse otherwise
 */
ATbool PERR_isValidLocation(PERR_Location arg) {
  if (PERR_isLocationFile(arg)) {
    return ATtrue;
  }
  else if (PERR_isLocationArea(arg)) {
    return ATtrue;
  }
  else if (PERR_isLocationAreaInFile(arg)) {
    return ATtrue;
  }
  return ATfalse;
}

/**
 * Assert whether a PERR_Location is a file by checking against the following ATerm pattern: appl(prod([lit("file"),cf(opt(layout)),lit("("),cf(opt(layout)),cf(sort("StrCon")),cf(opt(layout)),lit(")")],cf(sort("Location")),attrs([term(cons("file"))])),[appl(prod([char-class([102]),char-class([105]),char-class([108]),char-class([101])],lit("file"),no-attrs),[102,105,108,101]),<"ws-after-file"("OptLayout")>,appl(prod([char-class([40])],lit("("),no-attrs),[40]),<"ws-after-("("OptLayout")>,<"filename"("StrCon")>,<"ws-after-filename"("OptLayout")>,appl(prod([char-class([41])],lit(")"),no-attrs),[41])]). May not be used to assert correctness of the PERR_Location
 * \param[in] arg input PERR_Location
 * \return ATtrue if #arg corresponds to the signature of a file, or ATfalse otherwise
 */
inline ATbool PERR_isLocationFile(PERR_Location arg){
  /* checking for: appl */
  if (ATgetType((ATerm)arg) == AT_APPL && ATgetAFun((ATermAppl)arg) == PERR_afun0) {
    ATerm arg_arg0 = ATgetArgument(arg, 0);
    /* checking for: prod */
    if (ATgetType((ATerm)arg_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0) == PERR_afun1) {
      ATerm arg_arg0_arg0 = ATgetArgument(arg_arg0, 0);
      if (ATgetType((ATerm)arg_arg0_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0) == ATfalse) {
        ATerm arg_arg0_arg0_head;
        ATermList arg_arg0_arg0_list = (ATermList)arg_arg0_arg0;
        arg_arg0_arg0_head = ATgetFirst(arg_arg0_arg0_list);
        arg_arg0_arg0_list = ATgetNext(arg_arg0_arg0_list);
        /* checking for: lit */
        if (ATgetType((ATerm)arg_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head) == PERR_afun9) {
          ATerm arg_arg0_arg0_head_arg0 = ATgetArgument(arg_arg0_arg0_head, 0);
          /* checking for: file */
          if (ATgetType((ATerm)arg_arg0_arg0_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0) == PERR_afun53) {
            if (ATgetType((ATerm)arg_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0_list) == ATfalse) {
              arg_arg0_arg0_head = ATgetFirst(arg_arg0_arg0_list);
              arg_arg0_arg0_list = ATgetNext(arg_arg0_arg0_list);
              /* checking for: cf */
              if (ATgetType((ATerm)arg_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head) == PERR_afun2) {
                ATerm arg_arg0_arg0_head_arg0 = ATgetArgument(arg_arg0_arg0_head, 0);
                /* checking for: opt */
                if (ATgetType((ATerm)arg_arg0_arg0_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0) == PERR_afun3) {
                  ATerm arg_arg0_arg0_head_arg0_arg0 = ATgetArgument(arg_arg0_arg0_head_arg0, 0);
                  /* checking for: layout */
                  if (ATgetType((ATerm)arg_arg0_arg0_head_arg0_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0_arg0) == PERR_afun4) {
                    if (ATgetType((ATerm)arg_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0_list) == ATfalse) {
                      arg_arg0_arg0_head = ATgetFirst(arg_arg0_arg0_list);
                      arg_arg0_arg0_list = ATgetNext(arg_arg0_arg0_list);
                      /* checking for: lit */
                      if (ATgetType((ATerm)arg_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head) == PERR_afun9) {
                        ATerm arg_arg0_arg0_head_arg0 = ATgetArgument(arg_arg0_arg0_head, 0);
                        /* checking for: ( */
                        if (ATgetType((ATerm)arg_arg0_arg0_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0) == PERR_afun39) {
                          if (ATgetType((ATerm)arg_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0_list) == ATfalse) {
                            arg_arg0_arg0_head = ATgetFirst(arg_arg0_arg0_list);
                            arg_arg0_arg0_list = ATgetNext(arg_arg0_arg0_list);
                            /* checking for: cf */
                            if (ATgetType((ATerm)arg_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head) == PERR_afun2) {
                              ATerm arg_arg0_arg0_head_arg0 = ATgetArgument(arg_arg0_arg0_head, 0);
                              /* checking for: opt */
                              if (ATgetType((ATerm)arg_arg0_arg0_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0) == PERR_afun3) {
                                ATerm arg_arg0_arg0_head_arg0_arg0 = ATgetArgument(arg_arg0_arg0_head_arg0, 0);
                                /* checking for: layout */
                                if (ATgetType((ATerm)arg_arg0_arg0_head_arg0_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0_arg0) == PERR_afun4) {
                                  if (ATgetType((ATerm)arg_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0_list) == ATfalse) {
                                    arg_arg0_arg0_head = ATgetFirst(arg_arg0_arg0_list);
                                    arg_arg0_arg0_list = ATgetNext(arg_arg0_arg0_list);
                                    /* checking for: cf */
                                    if (ATgetType((ATerm)arg_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head) == PERR_afun2) {
                                      ATerm arg_arg0_arg0_head_arg0 = ATgetArgument(arg_arg0_arg0_head, 0);
                                      /* checking for: sort */
                                      if (ATgetType((ATerm)arg_arg0_arg0_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0) == PERR_afun11) {
                                        ATerm arg_arg0_arg0_head_arg0_arg0 = ATgetArgument(arg_arg0_arg0_head_arg0, 0);
                                        /* checking for: StrCon */
                                        if (ATgetType((ATerm)arg_arg0_arg0_head_arg0_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0_arg0) == PERR_afun29) {
                                          if (ATgetType((ATerm)arg_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0_list) == ATfalse) {
                                            arg_arg0_arg0_head = ATgetFirst(arg_arg0_arg0_list);
                                            arg_arg0_arg0_list = ATgetNext(arg_arg0_arg0_list);
                                            /* checking for: cf */
                                            if (ATgetType((ATerm)arg_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head) == PERR_afun2) {
                                              ATerm arg_arg0_arg0_head_arg0 = ATgetArgument(arg_arg0_arg0_head, 0);
                                              /* checking for: opt */
                                              if (ATgetType((ATerm)arg_arg0_arg0_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0) == PERR_afun3) {
                                                ATerm arg_arg0_arg0_head_arg0_arg0 = ATgetArgument(arg_arg0_arg0_head_arg0, 0);
                                                /* checking for: layout */
                                                if (ATgetType((ATerm)arg_arg0_arg0_head_arg0_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0_arg0) == PERR_afun4) {
                                                  if (ATgetType((ATerm)arg_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0_list) == ATfalse) {
                                                    arg_arg0_arg0_head = ATgetFirst(arg_arg0_arg0_list);
                                                    arg_arg0_arg0_list = ATgetNext(arg_arg0_arg0_list);
                                                    /* checking for: lit */
                                                    if (ATgetType((ATerm)arg_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head) == PERR_afun9) {
                                                      ATerm arg_arg0_arg0_head_arg0 = ATgetArgument(arg_arg0_arg0_head, 0);
                                                      /* checking for: ) */
                                                      if (ATgetType((ATerm)arg_arg0_arg0_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0) == PERR_afun38) {
                                                        if (ATgetType((ATerm)arg_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0_list) == ATtrue) {
                                                          ATerm arg_arg0_arg1 = ATgetArgument(arg_arg0, 1);
                                                          /* checking for: cf */
                                                          if (ATgetType((ATerm)arg_arg0_arg1) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg1) == PERR_afun2) {
                                                            ATerm arg_arg0_arg1_arg0 = ATgetArgument(arg_arg0_arg1, 0);
                                                            /* checking for: sort */
                                                            if (ATgetType((ATerm)arg_arg0_arg1_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg1_arg0) == PERR_afun11) {
                                                              ATerm arg_arg0_arg1_arg0_arg0 = ATgetArgument(arg_arg0_arg1_arg0, 0);
                                                              /* checking for: Location */
                                                              if (ATgetType((ATerm)arg_arg0_arg1_arg0_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg1_arg0_arg0) == PERR_afun42) {
                                                                ATerm arg_arg0_arg2 = ATgetArgument(arg_arg0, 2);
                                                                /* checking for: attrs */
                                                                if (ATgetType((ATerm)arg_arg0_arg2) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg2) == PERR_afun13) {
                                                                  ATerm arg_arg0_arg2_arg0 = ATgetArgument(arg_arg0_arg2, 0);
                                                                  if (ATgetType((ATerm)arg_arg0_arg2_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg2_arg0) == ATfalse) {
                                                                    ATerm arg_arg0_arg2_arg0_head;
                                                                    ATermList arg_arg0_arg2_arg0_list = (ATermList)arg_arg0_arg2_arg0;
                                                                    arg_arg0_arg2_arg0_head = ATgetFirst(arg_arg0_arg2_arg0_list);
                                                                    arg_arg0_arg2_arg0_list = ATgetNext(arg_arg0_arg2_arg0_list);
                                                                    /* checking for: term */
                                                                    if (ATgetType((ATerm)arg_arg0_arg2_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg2_arg0_head) == PERR_afun14) {
                                                                      ATerm arg_arg0_arg2_arg0_head_arg0 = ATgetArgument(arg_arg0_arg2_arg0_head, 0);
                                                                      /* checking for: cons */
                                                                      if (ATgetType((ATerm)arg_arg0_arg2_arg0_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg2_arg0_head_arg0) == PERR_afun15) {
                                                                        ATerm arg_arg0_arg2_arg0_head_arg0_arg0 = ATgetArgument(arg_arg0_arg2_arg0_head_arg0, 0);
                                                                        /* checking for: file */
                                                                        if (ATgetType((ATerm)arg_arg0_arg2_arg0_head_arg0_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg2_arg0_head_arg0_arg0) == PERR_afun53) {
                                                                          if (ATgetType((ATerm)arg_arg0_arg2_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg2_arg0_list) == ATtrue) {
                                                                            ATerm arg_arg1 = ATgetArgument(arg, 1);
                                                                            if (ATgetType((ATerm)arg_arg1) == AT_LIST && ATisEmpty((ATermList)arg_arg1) == ATfalse) {
                                                                              ATerm arg_arg1_head;
                                                                              ATermList arg_arg1_list = (ATermList)arg_arg1;
                                                                              arg_arg1_head = ATgetFirst(arg_arg1_list);
                                                                              arg_arg1_list = ATgetNext(arg_arg1_list);
                                                                              /* checking for: appl */
                                                                              if (ATgetType((ATerm)arg_arg1_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head) == PERR_afun0) {
                                                                                ATerm arg_arg1_head_arg0 = ATgetArgument(arg_arg1_head, 0);
                                                                                /* checking for: prod */
                                                                                if (ATgetType((ATerm)arg_arg1_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0) == PERR_afun1) {
                                                                                  ATerm arg_arg1_head_arg0_arg0 = ATgetArgument(arg_arg1_head_arg0, 0);
                                                                                  if (ATgetType((ATerm)arg_arg1_head_arg0_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0) == ATfalse) {
                                                                                    ATerm arg_arg1_head_arg0_arg0_head;
                                                                                    ATermList arg_arg1_head_arg0_arg0_list = (ATermList)arg_arg1_head_arg0_arg0;
                                                                                    arg_arg1_head_arg0_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_list);
                                                                                    arg_arg1_head_arg0_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_list);
                                                                                    /* checking for: char-class */
                                                                                    if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg0_head) == PERR_afun17) {
                                                                                      ATerm arg_arg1_head_arg0_arg0_head_arg0 = ATgetArgument(arg_arg1_head_arg0_arg0_head, 0);
                                                                                      if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0) == ATfalse) {
                                                                                        ATerm arg_arg1_head_arg0_arg0_head_arg0_head;
                                                                                        ATermList arg_arg1_head_arg0_arg0_head_arg0_list = (ATermList)arg_arg1_head_arg0_arg0_head_arg0;
                                                                                        arg_arg1_head_arg0_arg0_head_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                        arg_arg1_head_arg0_arg0_head_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                        if (ATgetInt((ATermInt)arg_arg1_head_arg0_arg0_head_arg0_head) == 102) {

                                                                                          if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0_list) == ATtrue) {
                                                                                            if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_list) == ATfalse) {
                                                                                              arg_arg1_head_arg0_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_list);
                                                                                              arg_arg1_head_arg0_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_list);
                                                                                              /* checking for: char-class */
                                                                                              if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg0_head) == PERR_afun17) {
                                                                                                ATerm arg_arg1_head_arg0_arg0_head_arg0 = ATgetArgument(arg_arg1_head_arg0_arg0_head, 0);
                                                                                                if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0) == ATfalse) {
                                                                                                  ATerm arg_arg1_head_arg0_arg0_head_arg0_head;
                                                                                                  ATermList arg_arg1_head_arg0_arg0_head_arg0_list = (ATermList)arg_arg1_head_arg0_arg0_head_arg0;
                                                                                                  arg_arg1_head_arg0_arg0_head_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                  arg_arg1_head_arg0_arg0_head_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                  if (ATgetInt((ATermInt)arg_arg1_head_arg0_arg0_head_arg0_head) == 105) {

                                                                                                    if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0_list) == ATtrue) {
                                                                                                      if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_list) == ATfalse) {
                                                                                                        arg_arg1_head_arg0_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_list);
                                                                                                        arg_arg1_head_arg0_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_list);
                                                                                                        /* checking for: char-class */
                                                                                                        if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg0_head) == PERR_afun17) {
                                                                                                          ATerm arg_arg1_head_arg0_arg0_head_arg0 = ATgetArgument(arg_arg1_head_arg0_arg0_head, 0);
                                                                                                          if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0) == ATfalse) {
                                                                                                            ATerm arg_arg1_head_arg0_arg0_head_arg0_head;
                                                                                                            ATermList arg_arg1_head_arg0_arg0_head_arg0_list = (ATermList)arg_arg1_head_arg0_arg0_head_arg0;
                                                                                                            arg_arg1_head_arg0_arg0_head_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                            arg_arg1_head_arg0_arg0_head_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                            if (ATgetInt((ATermInt)arg_arg1_head_arg0_arg0_head_arg0_head) == 108) {

                                                                                                              if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0_list) == ATtrue) {
                                                                                                                if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_list) == ATfalse) {
                                                                                                                  arg_arg1_head_arg0_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_list);
                                                                                                                  arg_arg1_head_arg0_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_list);
                                                                                                                  /* checking for: char-class */
                                                                                                                  if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg0_head) == PERR_afun17) {
                                                                                                                    ATerm arg_arg1_head_arg0_arg0_head_arg0 = ATgetArgument(arg_arg1_head_arg0_arg0_head, 0);
                                                                                                                    if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0) == ATfalse) {
                                                                                                                      ATerm arg_arg1_head_arg0_arg0_head_arg0_head;
                                                                                                                      ATermList arg_arg1_head_arg0_arg0_head_arg0_list = (ATermList)arg_arg1_head_arg0_arg0_head_arg0;
                                                                                                                      arg_arg1_head_arg0_arg0_head_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                                      arg_arg1_head_arg0_arg0_head_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                                      if (ATgetInt((ATermInt)arg_arg1_head_arg0_arg0_head_arg0_head) == 101) {

                                                                                                                        if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0_list) == ATtrue) {
                                                                                                                          if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_list) == ATtrue) {
                                                                                                                            ATerm arg_arg1_head_arg0_arg1 = ATgetArgument(arg_arg1_head_arg0, 1);
                                                                                                                            /* checking for: lit */
                                                                                                                            if (ATgetType((ATerm)arg_arg1_head_arg0_arg1) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg1) == PERR_afun9) {
                                                                                                                              ATerm arg_arg1_head_arg0_arg1_arg0 = ATgetArgument(arg_arg1_head_arg0_arg1, 0);
                                                                                                                              /* checking for: file */
                                                                                                                              if (ATgetType((ATerm)arg_arg1_head_arg0_arg1_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg1_arg0) == PERR_afun53) {
                                                                                                                                ATerm arg_arg1_head_arg0_arg2 = ATgetArgument(arg_arg1_head_arg0, 2);
                                                                                                                                /* checking for: no-attrs */
                                                                                                                                if (ATgetType((ATerm)arg_arg1_head_arg0_arg2) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg2) == PERR_afun5) {
                                                                                                                                  ATerm arg_arg1_head_arg1 = ATgetArgument(arg_arg1_head, 1);
                                                                                                                                  if (ATgetType((ATerm)arg_arg1_head_arg1) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg1) == ATfalse) {
                                                                                                                                    ATerm arg_arg1_head_arg1_head;
                                                                                                                                    ATermList arg_arg1_head_arg1_list = (ATermList)arg_arg1_head_arg1;
                                                                                                                                    arg_arg1_head_arg1_head = ATgetFirst(arg_arg1_head_arg1_list);
                                                                                                                                    arg_arg1_head_arg1_list = ATgetNext(arg_arg1_head_arg1_list);
                                                                                                                                    if (ATgetInt((ATermInt)arg_arg1_head_arg1_head) == 102) {

                                                                                                                                      if (ATgetType((ATerm)arg_arg1_head_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg1_list) == ATfalse) {
                                                                                                                                        arg_arg1_head_arg1_head = ATgetFirst(arg_arg1_head_arg1_list);
                                                                                                                                        arg_arg1_head_arg1_list = ATgetNext(arg_arg1_head_arg1_list);
                                                                                                                                        if (ATgetInt((ATermInt)arg_arg1_head_arg1_head) == 105) {

                                                                                                                                          if (ATgetType((ATerm)arg_arg1_head_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg1_list) == ATfalse) {
                                                                                                                                            arg_arg1_head_arg1_head = ATgetFirst(arg_arg1_head_arg1_list);
                                                                                                                                            arg_arg1_head_arg1_list = ATgetNext(arg_arg1_head_arg1_list);
                                                                                                                                            if (ATgetInt((ATermInt)arg_arg1_head_arg1_head) == 108) {

                                                                                                                                              if (ATgetType((ATerm)arg_arg1_head_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg1_list) == ATfalse) {
                                                                                                                                                arg_arg1_head_arg1_head = ATgetFirst(arg_arg1_head_arg1_list);
                                                                                                                                                arg_arg1_head_arg1_list = ATgetNext(arg_arg1_head_arg1_list);
                                                                                                                                                if (ATgetInt((ATermInt)arg_arg1_head_arg1_head) == 101) {

                                                                                                                                                  if (ATgetType((ATerm)arg_arg1_head_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg1_list) == ATtrue) {
                                                                                                                                                    if (ATgetType((ATerm)arg_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_list) == ATfalse) {
                                                                                                                                                      arg_arg1_list = ATgetNext(arg_arg1_list);
                                                                                                                                                      if (ATgetType((ATerm)arg_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_list) == ATfalse) {
                                                                                                                                                        arg_arg1_head = ATgetFirst(arg_arg1_list);
                                                                                                                                                        arg_arg1_list = ATgetNext(arg_arg1_list);
                                                                                                                                                        /* checking for: appl */
                                                                                                                                                        if (ATgetType((ATerm)arg_arg1_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head) == PERR_afun0) {
                                                                                                                                                          ATerm arg_arg1_head_arg0 = ATgetArgument(arg_arg1_head, 0);
                                                                                                                                                          /* checking for: prod */
                                                                                                                                                          if (ATgetType((ATerm)arg_arg1_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0) == PERR_afun1) {
                                                                                                                                                            ATerm arg_arg1_head_arg0_arg0 = ATgetArgument(arg_arg1_head_arg0, 0);
                                                                                                                                                            if (ATgetType((ATerm)arg_arg1_head_arg0_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0) == ATfalse) {
                                                                                                                                                              ATerm arg_arg1_head_arg0_arg0_head;
                                                                                                                                                              ATermList arg_arg1_head_arg0_arg0_list = (ATermList)arg_arg1_head_arg0_arg0;
                                                                                                                                                              arg_arg1_head_arg0_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_list);
                                                                                                                                                              arg_arg1_head_arg0_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_list);
                                                                                                                                                              /* checking for: char-class */
                                                                                                                                                              if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg0_head) == PERR_afun17) {
                                                                                                                                                                ATerm arg_arg1_head_arg0_arg0_head_arg0 = ATgetArgument(arg_arg1_head_arg0_arg0_head, 0);
                                                                                                                                                                if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0) == ATfalse) {
                                                                                                                                                                  ATerm arg_arg1_head_arg0_arg0_head_arg0_head;
                                                                                                                                                                  ATermList arg_arg1_head_arg0_arg0_head_arg0_list = (ATermList)arg_arg1_head_arg0_arg0_head_arg0;
                                                                                                                                                                  arg_arg1_head_arg0_arg0_head_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                                                                                  arg_arg1_head_arg0_arg0_head_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                                                                                  if (ATgetInt((ATermInt)arg_arg1_head_arg0_arg0_head_arg0_head) == 40) {

                                                                                                                                                                    if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0_list) == ATtrue) {
                                                                                                                                                                      if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_list) == ATtrue) {
                                                                                                                                                                        ATerm arg_arg1_head_arg0_arg1 = ATgetArgument(arg_arg1_head_arg0, 1);
                                                                                                                                                                        /* checking for: lit */
                                                                                                                                                                        if (ATgetType((ATerm)arg_arg1_head_arg0_arg1) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg1) == PERR_afun9) {
                                                                                                                                                                          ATerm arg_arg1_head_arg0_arg1_arg0 = ATgetArgument(arg_arg1_head_arg0_arg1, 0);
                                                                                                                                                                          /* checking for: ( */
                                                                                                                                                                          if (ATgetType((ATerm)arg_arg1_head_arg0_arg1_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg1_arg0) == PERR_afun39) {
                                                                                                                                                                            ATerm arg_arg1_head_arg0_arg2 = ATgetArgument(arg_arg1_head_arg0, 2);
                                                                                                                                                                            /* checking for: no-attrs */
                                                                                                                                                                            if (ATgetType((ATerm)arg_arg1_head_arg0_arg2) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg2) == PERR_afun5) {
                                                                                                                                                                              ATerm arg_arg1_head_arg1 = ATgetArgument(arg_arg1_head, 1);
                                                                                                                                                                              if (ATgetType((ATerm)arg_arg1_head_arg1) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg1) == ATfalse) {
                                                                                                                                                                                ATerm arg_arg1_head_arg1_head;
                                                                                                                                                                                ATermList arg_arg1_head_arg1_list = (ATermList)arg_arg1_head_arg1;
                                                                                                                                                                                arg_arg1_head_arg1_head = ATgetFirst(arg_arg1_head_arg1_list);
                                                                                                                                                                                arg_arg1_head_arg1_list = ATgetNext(arg_arg1_head_arg1_list);
                                                                                                                                                                                if (ATgetInt((ATermInt)arg_arg1_head_arg1_head) == 40) {

                                                                                                                                                                                  if (ATgetType((ATerm)arg_arg1_head_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg1_list) == ATtrue) {
                                                                                                                                                                                    if (ATgetType((ATerm)arg_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_list) == ATfalse) {
                                                                                                                                                                                      arg_arg1_list = ATgetNext(arg_arg1_list);
                                                                                                                                                                                      if (ATgetType((ATerm)arg_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_list) == ATfalse) {
                                                                                                                                                                                        arg_arg1_list = ATgetNext(arg_arg1_list);
                                                                                                                                                                                        if (ATgetType((ATerm)arg_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_list) == ATfalse) {
                                                                                                                                                                                          arg_arg1_list = ATgetNext(arg_arg1_list);
                                                                                                                                                                                          if (ATgetType((ATerm)arg_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_list) == ATfalse) {
                                                                                                                                                                                            arg_arg1_head = ATgetFirst(arg_arg1_list);
                                                                                                                                                                                            arg_arg1_list = ATgetNext(arg_arg1_list);
                                                                                                                                                                                            /* checking for: appl */
                                                                                                                                                                                            if (ATgetType((ATerm)arg_arg1_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head) == PERR_afun0) {
                                                                                                                                                                                              ATerm arg_arg1_head_arg0 = ATgetArgument(arg_arg1_head, 0);
                                                                                                                                                                                              /* checking for: prod */
                                                                                                                                                                                              if (ATgetType((ATerm)arg_arg1_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0) == PERR_afun1) {
                                                                                                                                                                                                ATerm arg_arg1_head_arg0_arg0 = ATgetArgument(arg_arg1_head_arg0, 0);
                                                                                                                                                                                                if (ATgetType((ATerm)arg_arg1_head_arg0_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0) == ATfalse) {
                                                                                                                                                                                                  ATerm arg_arg1_head_arg0_arg0_head;
                                                                                                                                                                                                  ATermList arg_arg1_head_arg0_arg0_list = (ATermList)arg_arg1_head_arg0_arg0;
                                                                                                                                                                                                  arg_arg1_head_arg0_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_list);
                                                                                                                                                                                                  arg_arg1_head_arg0_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_list);
                                                                                                                                                                                                  /* checking for: char-class */
                                                                                                                                                                                                  if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg0_head) == PERR_afun17) {
                                                                                                                                                                                                    ATerm arg_arg1_head_arg0_arg0_head_arg0 = ATgetArgument(arg_arg1_head_arg0_arg0_head, 0);
                                                                                                                                                                                                    if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0) == ATfalse) {
                                                                                                                                                                                                      ATerm arg_arg1_head_arg0_arg0_head_arg0_head;
                                                                                                                                                                                                      ATermList arg_arg1_head_arg0_arg0_head_arg0_list = (ATermList)arg_arg1_head_arg0_arg0_head_arg0;
                                                                                                                                                                                                      arg_arg1_head_arg0_arg0_head_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                                                                                                                      arg_arg1_head_arg0_arg0_head_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                                                                                                                      if (ATgetInt((ATermInt)arg_arg1_head_arg0_arg0_head_arg0_head) == 41) {

                                                                                                                                                                                                        if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0_list) == ATtrue) {
                                                                                                                                                                                                          if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_list) == ATtrue) {
                                                                                                                                                                                                            ATerm arg_arg1_head_arg0_arg1 = ATgetArgument(arg_arg1_head_arg0, 1);
                                                                                                                                                                                                            /* checking for: lit */
                                                                                                                                                                                                            if (ATgetType((ATerm)arg_arg1_head_arg0_arg1) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg1) == PERR_afun9) {
                                                                                                                                                                                                              ATerm arg_arg1_head_arg0_arg1_arg0 = ATgetArgument(arg_arg1_head_arg0_arg1, 0);
                                                                                                                                                                                                              /* checking for: ) */
                                                                                                                                                                                                              if (ATgetType((ATerm)arg_arg1_head_arg0_arg1_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg1_arg0) == PERR_afun38) {
                                                                                                                                                                                                                ATerm arg_arg1_head_arg0_arg2 = ATgetArgument(arg_arg1_head_arg0, 2);
                                                                                                                                                                                                                /* checking for: no-attrs */
                                                                                                                                                                                                                if (ATgetType((ATerm)arg_arg1_head_arg0_arg2) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg2) == PERR_afun5) {
                                                                                                                                                                                                                  ATerm arg_arg1_head_arg1 = ATgetArgument(arg_arg1_head, 1);
                                                                                                                                                                                                                  if (ATgetType((ATerm)arg_arg1_head_arg1) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg1) == ATfalse) {
                                                                                                                                                                                                                    ATerm arg_arg1_head_arg1_head;
                                                                                                                                                                                                                    ATermList arg_arg1_head_arg1_list = (ATermList)arg_arg1_head_arg1;
                                                                                                                                                                                                                    arg_arg1_head_arg1_head = ATgetFirst(arg_arg1_head_arg1_list);
                                                                                                                                                                                                                    arg_arg1_head_arg1_list = ATgetNext(arg_arg1_head_arg1_list);
                                                                                                                                                                                                                    if (ATgetInt((ATermInt)arg_arg1_head_arg1_head) == 41) {

                                                                                                                                                                                                                      if (ATgetType((ATerm)arg_arg1_head_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg1_list) == ATtrue) {
                                                                                                                                                                                                                        if (ATgetType((ATerm)arg_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_list) == ATtrue) {
                                                                                                                                                                                                                          return ATtrue;
                                                                                                                                                                                                                        }
                                                                                                                                                                                                                      }
                                                                                                                                                                                                                    }
                                                                                                                                                                                                                  }
                                                                                                                                                                                                                }
                                                                                                                                                                                                              }
                                                                                                                                                                                                            }
                                                                                                                                                                                                          }
                                                                                                                                                                                                        }
                                                                                                                                                                                                      }
                                                                                                                                                                                                    }
                                                                                                                                                                                                  }
                                                                                                                                                                                                }
                                                                                                                                                                                              }
                                                                                                                                                                                            }
                                                                                                                                                                                          }
                                                                                                                                                                                        }
                                                                                                                                                                                      }
                                                                                                                                                                                    }
                                                                                                                                                                                  }
                                                                                                                                                                                }
                                                                                                                                                                              }
                                                                                                                                                                            }
                                                                                                                                                                          }
                                                                                                                                                                        }
                                                                                                                                                                      }
                                                                                                                                                                    }
                                                                                                                                                                  }
                                                                                                                                                                }
                                                                                                                                                              }
                                                                                                                                                            }
                                                                                                                                                          }
                                                                                                                                                        }
                                                                                                                                                      }
                                                                                                                                                    }
                                                                                                                                                  }
                                                                                                                                                }
                                                                                                                                              }
                                                                                                                                            }
                                                                                                                                          }
                                                                                                                                        }
                                                                                                                                      }
                                                                                                                                    }
                                                                                                                                  }
                                                                                                                                }
                                                                                                                              }
                                                                                                                            }
                                                                                                                          }
                                                                                                                        }
                                                                                                                      }
                                                                                                                    }
                                                                                                                  }
                                                                                                                }
                                                                                                              }
                                                                                                            }
                                                                                                          }
                                                                                                        }
                                                                                                      }
                                                                                                    }
                                                                                                  }
                                                                                                }
                                                                                              }
                                                                                            }
                                                                                          }
                                                                                        }
                                                                                      }
                                                                                    }
                                                                                  }
                                                                                }
                                                                              }
                                                                            }
                                                                          }
                                                                        }
                                                                      }
                                                                    }
                                                                  }
                                                                }
                                                              }
                                                            }
                                                          }
                                                        }
                                                      }
                                                    }
                                                  }
                                                }
                                              }
                                            }
                                          }
                                        }
                                      }
                                    }
                                  }
                                }
                              }
                            }
                          }
                        }
                      }
                    }
                  }
                }
              }
            }
          }
        }
      }
    }
  }
  return ATfalse;
}

/**
 * Assert whether a PERR_Location is a area by checking against the following ATerm pattern: appl(prod([lit("area"),cf(opt(layout)),lit("("),cf(opt(layout)),cf(sort("Area")),cf(opt(layout)),lit(")")],cf(sort("Location")),attrs([term(cons("area"))])),[appl(prod([char-class([97]),char-class([114]),char-class([101]),char-class([97])],lit("area"),no-attrs),[97,114,101,97]),<"ws-after-a"("OptLayout")>,appl(prod([char-class([40])],lit("("),no-attrs),[40]),<"ws-after-("("OptLayout")>,<"Area"("Area")>,<"ws-after-Area"("OptLayout")>,appl(prod([char-class([41])],lit(")"),no-attrs),[41])]). May not be used to assert correctness of the PERR_Location
 * \param[in] arg input PERR_Location
 * \return ATtrue if #arg corresponds to the signature of a area, or ATfalse otherwise
 */
inline ATbool PERR_isLocationArea(PERR_Location arg){
  /* checking for: appl */
  if (ATgetType((ATerm)arg) == AT_APPL && ATgetAFun((ATermAppl)arg) == PERR_afun0) {
    ATerm arg_arg0 = ATgetArgument(arg, 0);
    /* checking for: prod */
    if (ATgetType((ATerm)arg_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0) == PERR_afun1) {
      ATerm arg_arg0_arg0 = ATgetArgument(arg_arg0, 0);
      if (ATgetType((ATerm)arg_arg0_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0) == ATfalse) {
        ATerm arg_arg0_arg0_head;
        ATermList arg_arg0_arg0_list = (ATermList)arg_arg0_arg0;
        arg_arg0_arg0_head = ATgetFirst(arg_arg0_arg0_list);
        arg_arg0_arg0_list = ATgetNext(arg_arg0_arg0_list);
        /* checking for: lit */
        if (ATgetType((ATerm)arg_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head) == PERR_afun9) {
          ATerm arg_arg0_arg0_head_arg0 = ATgetArgument(arg_arg0_arg0_head, 0);
          /* checking for: area */
          if (ATgetType((ATerm)arg_arg0_arg0_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0) == PERR_afun55) {
            if (ATgetType((ATerm)arg_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0_list) == ATfalse) {
              arg_arg0_arg0_head = ATgetFirst(arg_arg0_arg0_list);
              arg_arg0_arg0_list = ATgetNext(arg_arg0_arg0_list);
              /* checking for: cf */
              if (ATgetType((ATerm)arg_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head) == PERR_afun2) {
                ATerm arg_arg0_arg0_head_arg0 = ATgetArgument(arg_arg0_arg0_head, 0);
                /* checking for: opt */
                if (ATgetType((ATerm)arg_arg0_arg0_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0) == PERR_afun3) {
                  ATerm arg_arg0_arg0_head_arg0_arg0 = ATgetArgument(arg_arg0_arg0_head_arg0, 0);
                  /* checking for: layout */
                  if (ATgetType((ATerm)arg_arg0_arg0_head_arg0_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0_arg0) == PERR_afun4) {
                    if (ATgetType((ATerm)arg_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0_list) == ATfalse) {
                      arg_arg0_arg0_head = ATgetFirst(arg_arg0_arg0_list);
                      arg_arg0_arg0_list = ATgetNext(arg_arg0_arg0_list);
                      /* checking for: lit */
                      if (ATgetType((ATerm)arg_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head) == PERR_afun9) {
                        ATerm arg_arg0_arg0_head_arg0 = ATgetArgument(arg_arg0_arg0_head, 0);
                        /* checking for: ( */
                        if (ATgetType((ATerm)arg_arg0_arg0_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0) == PERR_afun39) {
                          if (ATgetType((ATerm)arg_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0_list) == ATfalse) {
                            arg_arg0_arg0_head = ATgetFirst(arg_arg0_arg0_list);
                            arg_arg0_arg0_list = ATgetNext(arg_arg0_arg0_list);
                            /* checking for: cf */
                            if (ATgetType((ATerm)arg_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head) == PERR_afun2) {
                              ATerm arg_arg0_arg0_head_arg0 = ATgetArgument(arg_arg0_arg0_head, 0);
                              /* checking for: opt */
                              if (ATgetType((ATerm)arg_arg0_arg0_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0) == PERR_afun3) {
                                ATerm arg_arg0_arg0_head_arg0_arg0 = ATgetArgument(arg_arg0_arg0_head_arg0, 0);
                                /* checking for: layout */
                                if (ATgetType((ATerm)arg_arg0_arg0_head_arg0_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0_arg0) == PERR_afun4) {
                                  if (ATgetType((ATerm)arg_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0_list) == ATfalse) {
                                    arg_arg0_arg0_head = ATgetFirst(arg_arg0_arg0_list);
                                    arg_arg0_arg0_list = ATgetNext(arg_arg0_arg0_list);
                                    /* checking for: cf */
                                    if (ATgetType((ATerm)arg_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head) == PERR_afun2) {
                                      ATerm arg_arg0_arg0_head_arg0 = ATgetArgument(arg_arg0_arg0_head, 0);
                                      /* checking for: sort */
                                      if (ATgetType((ATerm)arg_arg0_arg0_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0) == PERR_afun11) {
                                        ATerm arg_arg0_arg0_head_arg0_arg0 = ATgetArgument(arg_arg0_arg0_head_arg0, 0);
                                        /* checking for: Area */
                                        if (ATgetType((ATerm)arg_arg0_arg0_head_arg0_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0_arg0) == PERR_afun54) {
                                          if (ATgetType((ATerm)arg_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0_list) == ATfalse) {
                                            arg_arg0_arg0_head = ATgetFirst(arg_arg0_arg0_list);
                                            arg_arg0_arg0_list = ATgetNext(arg_arg0_arg0_list);
                                            /* checking for: cf */
                                            if (ATgetType((ATerm)arg_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head) == PERR_afun2) {
                                              ATerm arg_arg0_arg0_head_arg0 = ATgetArgument(arg_arg0_arg0_head, 0);
                                              /* checking for: opt */
                                              if (ATgetType((ATerm)arg_arg0_arg0_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0) == PERR_afun3) {
                                                ATerm arg_arg0_arg0_head_arg0_arg0 = ATgetArgument(arg_arg0_arg0_head_arg0, 0);
                                                /* checking for: layout */
                                                if (ATgetType((ATerm)arg_arg0_arg0_head_arg0_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0_arg0) == PERR_afun4) {
                                                  if (ATgetType((ATerm)arg_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0_list) == ATfalse) {
                                                    arg_arg0_arg0_head = ATgetFirst(arg_arg0_arg0_list);
                                                    arg_arg0_arg0_list = ATgetNext(arg_arg0_arg0_list);
                                                    /* checking for: lit */
                                                    if (ATgetType((ATerm)arg_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head) == PERR_afun9) {
                                                      ATerm arg_arg0_arg0_head_arg0 = ATgetArgument(arg_arg0_arg0_head, 0);
                                                      /* checking for: ) */
                                                      if (ATgetType((ATerm)arg_arg0_arg0_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0) == PERR_afun38) {
                                                        if (ATgetType((ATerm)arg_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0_list) == ATtrue) {
                                                          ATerm arg_arg0_arg1 = ATgetArgument(arg_arg0, 1);
                                                          /* checking for: cf */
                                                          if (ATgetType((ATerm)arg_arg0_arg1) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg1) == PERR_afun2) {
                                                            ATerm arg_arg0_arg1_arg0 = ATgetArgument(arg_arg0_arg1, 0);
                                                            /* checking for: sort */
                                                            if (ATgetType((ATerm)arg_arg0_arg1_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg1_arg0) == PERR_afun11) {
                                                              ATerm arg_arg0_arg1_arg0_arg0 = ATgetArgument(arg_arg0_arg1_arg0, 0);
                                                              /* checking for: Location */
                                                              if (ATgetType((ATerm)arg_arg0_arg1_arg0_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg1_arg0_arg0) == PERR_afun42) {
                                                                ATerm arg_arg0_arg2 = ATgetArgument(arg_arg0, 2);
                                                                /* checking for: attrs */
                                                                if (ATgetType((ATerm)arg_arg0_arg2) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg2) == PERR_afun13) {
                                                                  ATerm arg_arg0_arg2_arg0 = ATgetArgument(arg_arg0_arg2, 0);
                                                                  if (ATgetType((ATerm)arg_arg0_arg2_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg2_arg0) == ATfalse) {
                                                                    ATerm arg_arg0_arg2_arg0_head;
                                                                    ATermList arg_arg0_arg2_arg0_list = (ATermList)arg_arg0_arg2_arg0;
                                                                    arg_arg0_arg2_arg0_head = ATgetFirst(arg_arg0_arg2_arg0_list);
                                                                    arg_arg0_arg2_arg0_list = ATgetNext(arg_arg0_arg2_arg0_list);
                                                                    /* checking for: term */
                                                                    if (ATgetType((ATerm)arg_arg0_arg2_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg2_arg0_head) == PERR_afun14) {
                                                                      ATerm arg_arg0_arg2_arg0_head_arg0 = ATgetArgument(arg_arg0_arg2_arg0_head, 0);
                                                                      /* checking for: cons */
                                                                      if (ATgetType((ATerm)arg_arg0_arg2_arg0_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg2_arg0_head_arg0) == PERR_afun15) {
                                                                        ATerm arg_arg0_arg2_arg0_head_arg0_arg0 = ATgetArgument(arg_arg0_arg2_arg0_head_arg0, 0);
                                                                        /* checking for: area */
                                                                        if (ATgetType((ATerm)arg_arg0_arg2_arg0_head_arg0_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg2_arg0_head_arg0_arg0) == PERR_afun55) {
                                                                          if (ATgetType((ATerm)arg_arg0_arg2_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg2_arg0_list) == ATtrue) {
                                                                            ATerm arg_arg1 = ATgetArgument(arg, 1);
                                                                            if (ATgetType((ATerm)arg_arg1) == AT_LIST && ATisEmpty((ATermList)arg_arg1) == ATfalse) {
                                                                              ATerm arg_arg1_head;
                                                                              ATermList arg_arg1_list = (ATermList)arg_arg1;
                                                                              arg_arg1_head = ATgetFirst(arg_arg1_list);
                                                                              arg_arg1_list = ATgetNext(arg_arg1_list);
                                                                              /* checking for: appl */
                                                                              if (ATgetType((ATerm)arg_arg1_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head) == PERR_afun0) {
                                                                                ATerm arg_arg1_head_arg0 = ATgetArgument(arg_arg1_head, 0);
                                                                                /* checking for: prod */
                                                                                if (ATgetType((ATerm)arg_arg1_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0) == PERR_afun1) {
                                                                                  ATerm arg_arg1_head_arg0_arg0 = ATgetArgument(arg_arg1_head_arg0, 0);
                                                                                  if (ATgetType((ATerm)arg_arg1_head_arg0_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0) == ATfalse) {
                                                                                    ATerm arg_arg1_head_arg0_arg0_head;
                                                                                    ATermList arg_arg1_head_arg0_arg0_list = (ATermList)arg_arg1_head_arg0_arg0;
                                                                                    arg_arg1_head_arg0_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_list);
                                                                                    arg_arg1_head_arg0_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_list);
                                                                                    /* checking for: char-class */
                                                                                    if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg0_head) == PERR_afun17) {
                                                                                      ATerm arg_arg1_head_arg0_arg0_head_arg0 = ATgetArgument(arg_arg1_head_arg0_arg0_head, 0);
                                                                                      if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0) == ATfalse) {
                                                                                        ATerm arg_arg1_head_arg0_arg0_head_arg0_head;
                                                                                        ATermList arg_arg1_head_arg0_arg0_head_arg0_list = (ATermList)arg_arg1_head_arg0_arg0_head_arg0;
                                                                                        arg_arg1_head_arg0_arg0_head_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                        arg_arg1_head_arg0_arg0_head_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                        if (ATgetInt((ATermInt)arg_arg1_head_arg0_arg0_head_arg0_head) == 97) {

                                                                                          if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0_list) == ATtrue) {
                                                                                            if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_list) == ATfalse) {
                                                                                              arg_arg1_head_arg0_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_list);
                                                                                              arg_arg1_head_arg0_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_list);
                                                                                              /* checking for: char-class */
                                                                                              if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg0_head) == PERR_afun17) {
                                                                                                ATerm arg_arg1_head_arg0_arg0_head_arg0 = ATgetArgument(arg_arg1_head_arg0_arg0_head, 0);
                                                                                                if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0) == ATfalse) {
                                                                                                  ATerm arg_arg1_head_arg0_arg0_head_arg0_head;
                                                                                                  ATermList arg_arg1_head_arg0_arg0_head_arg0_list = (ATermList)arg_arg1_head_arg0_arg0_head_arg0;
                                                                                                  arg_arg1_head_arg0_arg0_head_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                  arg_arg1_head_arg0_arg0_head_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                  if (ATgetInt((ATermInt)arg_arg1_head_arg0_arg0_head_arg0_head) == 114) {

                                                                                                    if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0_list) == ATtrue) {
                                                                                                      if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_list) == ATfalse) {
                                                                                                        arg_arg1_head_arg0_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_list);
                                                                                                        arg_arg1_head_arg0_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_list);
                                                                                                        /* checking for: char-class */
                                                                                                        if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg0_head) == PERR_afun17) {
                                                                                                          ATerm arg_arg1_head_arg0_arg0_head_arg0 = ATgetArgument(arg_arg1_head_arg0_arg0_head, 0);
                                                                                                          if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0) == ATfalse) {
                                                                                                            ATerm arg_arg1_head_arg0_arg0_head_arg0_head;
                                                                                                            ATermList arg_arg1_head_arg0_arg0_head_arg0_list = (ATermList)arg_arg1_head_arg0_arg0_head_arg0;
                                                                                                            arg_arg1_head_arg0_arg0_head_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                            arg_arg1_head_arg0_arg0_head_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                            if (ATgetInt((ATermInt)arg_arg1_head_arg0_arg0_head_arg0_head) == 101) {

                                                                                                              if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0_list) == ATtrue) {
                                                                                                                if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_list) == ATfalse) {
                                                                                                                  arg_arg1_head_arg0_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_list);
                                                                                                                  arg_arg1_head_arg0_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_list);
                                                                                                                  /* checking for: char-class */
                                                                                                                  if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg0_head) == PERR_afun17) {
                                                                                                                    ATerm arg_arg1_head_arg0_arg0_head_arg0 = ATgetArgument(arg_arg1_head_arg0_arg0_head, 0);
                                                                                                                    if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0) == ATfalse) {
                                                                                                                      ATerm arg_arg1_head_arg0_arg0_head_arg0_head;
                                                                                                                      ATermList arg_arg1_head_arg0_arg0_head_arg0_list = (ATermList)arg_arg1_head_arg0_arg0_head_arg0;
                                                                                                                      arg_arg1_head_arg0_arg0_head_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                                      arg_arg1_head_arg0_arg0_head_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                                      if (ATgetInt((ATermInt)arg_arg1_head_arg0_arg0_head_arg0_head) == 97) {

                                                                                                                        if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0_list) == ATtrue) {
                                                                                                                          if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_list) == ATtrue) {
                                                                                                                            ATerm arg_arg1_head_arg0_arg1 = ATgetArgument(arg_arg1_head_arg0, 1);
                                                                                                                            /* checking for: lit */
                                                                                                                            if (ATgetType((ATerm)arg_arg1_head_arg0_arg1) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg1) == PERR_afun9) {
                                                                                                                              ATerm arg_arg1_head_arg0_arg1_arg0 = ATgetArgument(arg_arg1_head_arg0_arg1, 0);
                                                                                                                              /* checking for: area */
                                                                                                                              if (ATgetType((ATerm)arg_arg1_head_arg0_arg1_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg1_arg0) == PERR_afun55) {
                                                                                                                                ATerm arg_arg1_head_arg0_arg2 = ATgetArgument(arg_arg1_head_arg0, 2);
                                                                                                                                /* checking for: no-attrs */
                                                                                                                                if (ATgetType((ATerm)arg_arg1_head_arg0_arg2) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg2) == PERR_afun5) {
                                                                                                                                  ATerm arg_arg1_head_arg1 = ATgetArgument(arg_arg1_head, 1);
                                                                                                                                  if (ATgetType((ATerm)arg_arg1_head_arg1) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg1) == ATfalse) {
                                                                                                                                    ATerm arg_arg1_head_arg1_head;
                                                                                                                                    ATermList arg_arg1_head_arg1_list = (ATermList)arg_arg1_head_arg1;
                                                                                                                                    arg_arg1_head_arg1_head = ATgetFirst(arg_arg1_head_arg1_list);
                                                                                                                                    arg_arg1_head_arg1_list = ATgetNext(arg_arg1_head_arg1_list);
                                                                                                                                    if (ATgetInt((ATermInt)arg_arg1_head_arg1_head) == 97) {

                                                                                                                                      if (ATgetType((ATerm)arg_arg1_head_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg1_list) == ATfalse) {
                                                                                                                                        arg_arg1_head_arg1_head = ATgetFirst(arg_arg1_head_arg1_list);
                                                                                                                                        arg_arg1_head_arg1_list = ATgetNext(arg_arg1_head_arg1_list);
                                                                                                                                        if (ATgetInt((ATermInt)arg_arg1_head_arg1_head) == 114) {

                                                                                                                                          if (ATgetType((ATerm)arg_arg1_head_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg1_list) == ATfalse) {
                                                                                                                                            arg_arg1_head_arg1_head = ATgetFirst(arg_arg1_head_arg1_list);
                                                                                                                                            arg_arg1_head_arg1_list = ATgetNext(arg_arg1_head_arg1_list);
                                                                                                                                            if (ATgetInt((ATermInt)arg_arg1_head_arg1_head) == 101) {

                                                                                                                                              if (ATgetType((ATerm)arg_arg1_head_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg1_list) == ATfalse) {
                                                                                                                                                arg_arg1_head_arg1_head = ATgetFirst(arg_arg1_head_arg1_list);
                                                                                                                                                arg_arg1_head_arg1_list = ATgetNext(arg_arg1_head_arg1_list);
                                                                                                                                                if (ATgetInt((ATermInt)arg_arg1_head_arg1_head) == 97) {

                                                                                                                                                  if (ATgetType((ATerm)arg_arg1_head_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg1_list) == ATtrue) {
                                                                                                                                                    if (ATgetType((ATerm)arg_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_list) == ATfalse) {
                                                                                                                                                      arg_arg1_list = ATgetNext(arg_arg1_list);
                                                                                                                                                      if (ATgetType((ATerm)arg_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_list) == ATfalse) {
                                                                                                                                                        arg_arg1_head = ATgetFirst(arg_arg1_list);
                                                                                                                                                        arg_arg1_list = ATgetNext(arg_arg1_list);
                                                                                                                                                        /* checking for: appl */
                                                                                                                                                        if (ATgetType((ATerm)arg_arg1_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head) == PERR_afun0) {
                                                                                                                                                          ATerm arg_arg1_head_arg0 = ATgetArgument(arg_arg1_head, 0);
                                                                                                                                                          /* checking for: prod */
                                                                                                                                                          if (ATgetType((ATerm)arg_arg1_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0) == PERR_afun1) {
                                                                                                                                                            ATerm arg_arg1_head_arg0_arg0 = ATgetArgument(arg_arg1_head_arg0, 0);
                                                                                                                                                            if (ATgetType((ATerm)arg_arg1_head_arg0_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0) == ATfalse) {
                                                                                                                                                              ATerm arg_arg1_head_arg0_arg0_head;
                                                                                                                                                              ATermList arg_arg1_head_arg0_arg0_list = (ATermList)arg_arg1_head_arg0_arg0;
                                                                                                                                                              arg_arg1_head_arg0_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_list);
                                                                                                                                                              arg_arg1_head_arg0_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_list);
                                                                                                                                                              /* checking for: char-class */
                                                                                                                                                              if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg0_head) == PERR_afun17) {
                                                                                                                                                                ATerm arg_arg1_head_arg0_arg0_head_arg0 = ATgetArgument(arg_arg1_head_arg0_arg0_head, 0);
                                                                                                                                                                if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0) == ATfalse) {
                                                                                                                                                                  ATerm arg_arg1_head_arg0_arg0_head_arg0_head;
                                                                                                                                                                  ATermList arg_arg1_head_arg0_arg0_head_arg0_list = (ATermList)arg_arg1_head_arg0_arg0_head_arg0;
                                                                                                                                                                  arg_arg1_head_arg0_arg0_head_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                                                                                  arg_arg1_head_arg0_arg0_head_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                                                                                  if (ATgetInt((ATermInt)arg_arg1_head_arg0_arg0_head_arg0_head) == 40) {

                                                                                                                                                                    if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0_list) == ATtrue) {
                                                                                                                                                                      if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_list) == ATtrue) {
                                                                                                                                                                        ATerm arg_arg1_head_arg0_arg1 = ATgetArgument(arg_arg1_head_arg0, 1);
                                                                                                                                                                        /* checking for: lit */
                                                                                                                                                                        if (ATgetType((ATerm)arg_arg1_head_arg0_arg1) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg1) == PERR_afun9) {
                                                                                                                                                                          ATerm arg_arg1_head_arg0_arg1_arg0 = ATgetArgument(arg_arg1_head_arg0_arg1, 0);
                                                                                                                                                                          /* checking for: ( */
                                                                                                                                                                          if (ATgetType((ATerm)arg_arg1_head_arg0_arg1_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg1_arg0) == PERR_afun39) {
                                                                                                                                                                            ATerm arg_arg1_head_arg0_arg2 = ATgetArgument(arg_arg1_head_arg0, 2);
                                                                                                                                                                            /* checking for: no-attrs */
                                                                                                                                                                            if (ATgetType((ATerm)arg_arg1_head_arg0_arg2) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg2) == PERR_afun5) {
                                                                                                                                                                              ATerm arg_arg1_head_arg1 = ATgetArgument(arg_arg1_head, 1);
                                                                                                                                                                              if (ATgetType((ATerm)arg_arg1_head_arg1) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg1) == ATfalse) {
                                                                                                                                                                                ATerm arg_arg1_head_arg1_head;
                                                                                                                                                                                ATermList arg_arg1_head_arg1_list = (ATermList)arg_arg1_head_arg1;
                                                                                                                                                                                arg_arg1_head_arg1_head = ATgetFirst(arg_arg1_head_arg1_list);
                                                                                                                                                                                arg_arg1_head_arg1_list = ATgetNext(arg_arg1_head_arg1_list);
                                                                                                                                                                                if (ATgetInt((ATermInt)arg_arg1_head_arg1_head) == 40) {

                                                                                                                                                                                  if (ATgetType((ATerm)arg_arg1_head_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg1_list) == ATtrue) {
                                                                                                                                                                                    if (ATgetType((ATerm)arg_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_list) == ATfalse) {
                                                                                                                                                                                      arg_arg1_list = ATgetNext(arg_arg1_list);
                                                                                                                                                                                      if (ATgetType((ATerm)arg_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_list) == ATfalse) {
                                                                                                                                                                                        arg_arg1_list = ATgetNext(arg_arg1_list);
                                                                                                                                                                                        if (ATgetType((ATerm)arg_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_list) == ATfalse) {
                                                                                                                                                                                          arg_arg1_list = ATgetNext(arg_arg1_list);
                                                                                                                                                                                          if (ATgetType((ATerm)arg_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_list) == ATfalse) {
                                                                                                                                                                                            arg_arg1_head = ATgetFirst(arg_arg1_list);
                                                                                                                                                                                            arg_arg1_list = ATgetNext(arg_arg1_list);
                                                                                                                                                                                            /* checking for: appl */
                                                                                                                                                                                            if (ATgetType((ATerm)arg_arg1_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head) == PERR_afun0) {
                                                                                                                                                                                              ATerm arg_arg1_head_arg0 = ATgetArgument(arg_arg1_head, 0);
                                                                                                                                                                                              /* checking for: prod */
                                                                                                                                                                                              if (ATgetType((ATerm)arg_arg1_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0) == PERR_afun1) {
                                                                                                                                                                                                ATerm arg_arg1_head_arg0_arg0 = ATgetArgument(arg_arg1_head_arg0, 0);
                                                                                                                                                                                                if (ATgetType((ATerm)arg_arg1_head_arg0_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0) == ATfalse) {
                                                                                                                                                                                                  ATerm arg_arg1_head_arg0_arg0_head;
                                                                                                                                                                                                  ATermList arg_arg1_head_arg0_arg0_list = (ATermList)arg_arg1_head_arg0_arg0;
                                                                                                                                                                                                  arg_arg1_head_arg0_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_list);
                                                                                                                                                                                                  arg_arg1_head_arg0_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_list);
                                                                                                                                                                                                  /* checking for: char-class */
                                                                                                                                                                                                  if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg0_head) == PERR_afun17) {
                                                                                                                                                                                                    ATerm arg_arg1_head_arg0_arg0_head_arg0 = ATgetArgument(arg_arg1_head_arg0_arg0_head, 0);
                                                                                                                                                                                                    if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0) == ATfalse) {
                                                                                                                                                                                                      ATerm arg_arg1_head_arg0_arg0_head_arg0_head;
                                                                                                                                                                                                      ATermList arg_arg1_head_arg0_arg0_head_arg0_list = (ATermList)arg_arg1_head_arg0_arg0_head_arg0;
                                                                                                                                                                                                      arg_arg1_head_arg0_arg0_head_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                                                                                                                      arg_arg1_head_arg0_arg0_head_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                                                                                                                      if (ATgetInt((ATermInt)arg_arg1_head_arg0_arg0_head_arg0_head) == 41) {

                                                                                                                                                                                                        if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0_list) == ATtrue) {
                                                                                                                                                                                                          if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_list) == ATtrue) {
                                                                                                                                                                                                            ATerm arg_arg1_head_arg0_arg1 = ATgetArgument(arg_arg1_head_arg0, 1);
                                                                                                                                                                                                            /* checking for: lit */
                                                                                                                                                                                                            if (ATgetType((ATerm)arg_arg1_head_arg0_arg1) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg1) == PERR_afun9) {
                                                                                                                                                                                                              ATerm arg_arg1_head_arg0_arg1_arg0 = ATgetArgument(arg_arg1_head_arg0_arg1, 0);
                                                                                                                                                                                                              /* checking for: ) */
                                                                                                                                                                                                              if (ATgetType((ATerm)arg_arg1_head_arg0_arg1_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg1_arg0) == PERR_afun38) {
                                                                                                                                                                                                                ATerm arg_arg1_head_arg0_arg2 = ATgetArgument(arg_arg1_head_arg0, 2);
                                                                                                                                                                                                                /* checking for: no-attrs */
                                                                                                                                                                                                                if (ATgetType((ATerm)arg_arg1_head_arg0_arg2) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg2) == PERR_afun5) {
                                                                                                                                                                                                                  ATerm arg_arg1_head_arg1 = ATgetArgument(arg_arg1_head, 1);
                                                                                                                                                                                                                  if (ATgetType((ATerm)arg_arg1_head_arg1) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg1) == ATfalse) {
                                                                                                                                                                                                                    ATerm arg_arg1_head_arg1_head;
                                                                                                                                                                                                                    ATermList arg_arg1_head_arg1_list = (ATermList)arg_arg1_head_arg1;
                                                                                                                                                                                                                    arg_arg1_head_arg1_head = ATgetFirst(arg_arg1_head_arg1_list);
                                                                                                                                                                                                                    arg_arg1_head_arg1_list = ATgetNext(arg_arg1_head_arg1_list);
                                                                                                                                                                                                                    if (ATgetInt((ATermInt)arg_arg1_head_arg1_head) == 41) {

                                                                                                                                                                                                                      if (ATgetType((ATerm)arg_arg1_head_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg1_list) == ATtrue) {
                                                                                                                                                                                                                        if (ATgetType((ATerm)arg_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_list) == ATtrue) {
                                                                                                                                                                                                                          return ATtrue;
                                                                                                                                                                                                                        }
                                                                                                                                                                                                                      }
                                                                                                                                                                                                                    }
                                                                                                                                                                                                                  }
                                                                                                                                                                                                                }
                                                                                                                                                                                                              }
                                                                                                                                                                                                            }
                                                                                                                                                                                                          }
                                                                                                                                                                                                        }
                                                                                                                                                                                                      }
                                                                                                                                                                                                    }
                                                                                                                                                                                                  }
                                                                                                                                                                                                }
                                                                                                                                                                                              }
                                                                                                                                                                                            }
                                                                                                                                                                                          }
                                                                                                                                                                                        }
                                                                                                                                                                                      }
                                                                                                                                                                                    }
                                                                                                                                                                                  }
                                                                                                                                                                                }
                                                                                                                                                                              }
                                                                                                                                                                            }
                                                                                                                                                                          }
                                                                                                                                                                        }
                                                                                                                                                                      }
                                                                                                                                                                    }
                                                                                                                                                                  }
                                                                                                                                                                }
                                                                                                                                                              }
                                                                                                                                                            }
                                                                                                                                                          }
                                                                                                                                                        }
                                                                                                                                                      }
                                                                                                                                                    }
                                                                                                                                                  }
                                                                                                                                                }
                                                                                                                                              }
                                                                                                                                            }
                                                                                                                                          }
                                                                                                                                        }
                                                                                                                                      }
                                                                                                                                    }
                                                                                                                                  }
                                                                                                                                }
                                                                                                                              }
                                                                                                                            }
                                                                                                                          }
                                                                                                                        }
                                                                                                                      }
                                                                                                                    }
                                                                                                                  }
                                                                                                                }
                                                                                                              }
                                                                                                            }
                                                                                                          }
                                                                                                        }
                                                                                                      }
                                                                                                    }
                                                                                                  }
                                                                                                }
                                                                                              }
                                                                                            }
                                                                                          }
                                                                                        }
                                                                                      }
                                                                                    }
                                                                                  }
                                                                                }
                                                                              }
                                                                            }
                                                                          }
                                                                        }
                                                                      }
                                                                    }
                                                                  }
                                                                }
                                                              }
                                                            }
                                                          }
                                                        }
                                                      }
                                                    }
                                                  }
                                                }
                                              }
                                            }
                                          }
                                        }
                                      }
                                    }
                                  }
                                }
                              }
                            }
                          }
                        }
                      }
                    }
                  }
                }
              }
            }
          }
        }
      }
    }
  }
  return ATfalse;
}

/**
 * Assert whether a PERR_Location is a area-in-file by checking against the following ATerm pattern: appl(prod([lit("area-in-file"),cf(opt(layout)),lit("("),cf(opt(layout)),cf(sort("StrCon")),cf(opt(layout)),lit(","),cf(opt(layout)),cf(sort("Area")),cf(opt(layout)),lit(")")],cf(sort("Location")),attrs([term(cons("area-in-file"))])),[appl(prod([char-class([97]),char-class([114]),char-class([101]),char-class([97]),char-class([45]),char-class([105]),char-class([110]),char-class([45]),char-class([102]),char-class([105]),char-class([108]),char-class([101])],lit("area-in-file"),no-attrs),[97,114,101,97,45,105,110,45,102,105,108,101]),<"ws-after-area-in-file"("OptLayout")>,appl(prod([char-class([40])],lit("("),no-attrs),[40]),<"ws-after-("("OptLayout")>,<"filename"("StrCon")>,<"ws-after-filename"("OptLayout")>,appl(prod([char-class([44])],lit(","),no-attrs),[44]),<"ws-after-,"("OptLayout")>,<"Area"("Area")>,<"ws-after-Area"("OptLayout")>,appl(prod([char-class([41])],lit(")"),no-attrs),[41])]). May not be used to assert correctness of the PERR_Location
 * \param[in] arg input PERR_Location
 * \return ATtrue if #arg corresponds to the signature of a area-in-file, or ATfalse otherwise
 */
inline ATbool PERR_isLocationAreaInFile(PERR_Location arg){
  /* checking for: appl */
  if (ATgetType((ATerm)arg) == AT_APPL && ATgetAFun((ATermAppl)arg) == PERR_afun0) {
    ATerm arg_arg0 = ATgetArgument(arg, 0);
    /* checking for: prod */
    if (ATgetType((ATerm)arg_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0) == PERR_afun1) {
      ATerm arg_arg0_arg0 = ATgetArgument(arg_arg0, 0);
      if (ATgetType((ATerm)arg_arg0_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0) == ATfalse) {
        ATerm arg_arg0_arg0_head;
        ATermList arg_arg0_arg0_list = (ATermList)arg_arg0_arg0;
        arg_arg0_arg0_head = ATgetFirst(arg_arg0_arg0_list);
        arg_arg0_arg0_list = ATgetNext(arg_arg0_arg0_list);
        /* checking for: lit */
        if (ATgetType((ATerm)arg_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head) == PERR_afun9) {
          ATerm arg_arg0_arg0_head_arg0 = ATgetArgument(arg_arg0_arg0_head, 0);
          /* checking for: area-in-file */
          if (ATgetType((ATerm)arg_arg0_arg0_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0) == PERR_afun56) {
            if (ATgetType((ATerm)arg_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0_list) == ATfalse) {
              arg_arg0_arg0_head = ATgetFirst(arg_arg0_arg0_list);
              arg_arg0_arg0_list = ATgetNext(arg_arg0_arg0_list);
              /* checking for: cf */
              if (ATgetType((ATerm)arg_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head) == PERR_afun2) {
                ATerm arg_arg0_arg0_head_arg0 = ATgetArgument(arg_arg0_arg0_head, 0);
                /* checking for: opt */
                if (ATgetType((ATerm)arg_arg0_arg0_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0) == PERR_afun3) {
                  ATerm arg_arg0_arg0_head_arg0_arg0 = ATgetArgument(arg_arg0_arg0_head_arg0, 0);
                  /* checking for: layout */
                  if (ATgetType((ATerm)arg_arg0_arg0_head_arg0_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0_arg0) == PERR_afun4) {
                    if (ATgetType((ATerm)arg_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0_list) == ATfalse) {
                      arg_arg0_arg0_head = ATgetFirst(arg_arg0_arg0_list);
                      arg_arg0_arg0_list = ATgetNext(arg_arg0_arg0_list);
                      /* checking for: lit */
                      if (ATgetType((ATerm)arg_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head) == PERR_afun9) {
                        ATerm arg_arg0_arg0_head_arg0 = ATgetArgument(arg_arg0_arg0_head, 0);
                        /* checking for: ( */
                        if (ATgetType((ATerm)arg_arg0_arg0_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0) == PERR_afun39) {
                          if (ATgetType((ATerm)arg_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0_list) == ATfalse) {
                            arg_arg0_arg0_head = ATgetFirst(arg_arg0_arg0_list);
                            arg_arg0_arg0_list = ATgetNext(arg_arg0_arg0_list);
                            /* checking for: cf */
                            if (ATgetType((ATerm)arg_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head) == PERR_afun2) {
                              ATerm arg_arg0_arg0_head_arg0 = ATgetArgument(arg_arg0_arg0_head, 0);
                              /* checking for: opt */
                              if (ATgetType((ATerm)arg_arg0_arg0_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0) == PERR_afun3) {
                                ATerm arg_arg0_arg0_head_arg0_arg0 = ATgetArgument(arg_arg0_arg0_head_arg0, 0);
                                /* checking for: layout */
                                if (ATgetType((ATerm)arg_arg0_arg0_head_arg0_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0_arg0) == PERR_afun4) {
                                  if (ATgetType((ATerm)arg_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0_list) == ATfalse) {
                                    arg_arg0_arg0_head = ATgetFirst(arg_arg0_arg0_list);
                                    arg_arg0_arg0_list = ATgetNext(arg_arg0_arg0_list);
                                    /* checking for: cf */
                                    if (ATgetType((ATerm)arg_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head) == PERR_afun2) {
                                      ATerm arg_arg0_arg0_head_arg0 = ATgetArgument(arg_arg0_arg0_head, 0);
                                      /* checking for: sort */
                                      if (ATgetType((ATerm)arg_arg0_arg0_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0) == PERR_afun11) {
                                        ATerm arg_arg0_arg0_head_arg0_arg0 = ATgetArgument(arg_arg0_arg0_head_arg0, 0);
                                        /* checking for: StrCon */
                                        if (ATgetType((ATerm)arg_arg0_arg0_head_arg0_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0_arg0) == PERR_afun29) {
                                          if (ATgetType((ATerm)arg_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0_list) == ATfalse) {
                                            arg_arg0_arg0_head = ATgetFirst(arg_arg0_arg0_list);
                                            arg_arg0_arg0_list = ATgetNext(arg_arg0_arg0_list);
                                            /* checking for: cf */
                                            if (ATgetType((ATerm)arg_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head) == PERR_afun2) {
                                              ATerm arg_arg0_arg0_head_arg0 = ATgetArgument(arg_arg0_arg0_head, 0);
                                              /* checking for: opt */
                                              if (ATgetType((ATerm)arg_arg0_arg0_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0) == PERR_afun3) {
                                                ATerm arg_arg0_arg0_head_arg0_arg0 = ATgetArgument(arg_arg0_arg0_head_arg0, 0);
                                                /* checking for: layout */
                                                if (ATgetType((ATerm)arg_arg0_arg0_head_arg0_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0_arg0) == PERR_afun4) {
                                                  if (ATgetType((ATerm)arg_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0_list) == ATfalse) {
                                                    arg_arg0_arg0_head = ATgetFirst(arg_arg0_arg0_list);
                                                    arg_arg0_arg0_list = ATgetNext(arg_arg0_arg0_list);
                                                    /* checking for: lit */
                                                    if (ATgetType((ATerm)arg_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head) == PERR_afun9) {
                                                      ATerm arg_arg0_arg0_head_arg0 = ATgetArgument(arg_arg0_arg0_head, 0);
                                                      /* checking for: , */
                                                      if (ATgetType((ATerm)arg_arg0_arg0_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0) == PERR_afun43) {
                                                        if (ATgetType((ATerm)arg_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0_list) == ATfalse) {
                                                          arg_arg0_arg0_head = ATgetFirst(arg_arg0_arg0_list);
                                                          arg_arg0_arg0_list = ATgetNext(arg_arg0_arg0_list);
                                                          /* checking for: cf */
                                                          if (ATgetType((ATerm)arg_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head) == PERR_afun2) {
                                                            ATerm arg_arg0_arg0_head_arg0 = ATgetArgument(arg_arg0_arg0_head, 0);
                                                            /* checking for: opt */
                                                            if (ATgetType((ATerm)arg_arg0_arg0_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0) == PERR_afun3) {
                                                              ATerm arg_arg0_arg0_head_arg0_arg0 = ATgetArgument(arg_arg0_arg0_head_arg0, 0);
                                                              /* checking for: layout */
                                                              if (ATgetType((ATerm)arg_arg0_arg0_head_arg0_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0_arg0) == PERR_afun4) {
                                                                if (ATgetType((ATerm)arg_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0_list) == ATfalse) {
                                                                  arg_arg0_arg0_head = ATgetFirst(arg_arg0_arg0_list);
                                                                  arg_arg0_arg0_list = ATgetNext(arg_arg0_arg0_list);
                                                                  /* checking for: cf */
                                                                  if (ATgetType((ATerm)arg_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head) == PERR_afun2) {
                                                                    ATerm arg_arg0_arg0_head_arg0 = ATgetArgument(arg_arg0_arg0_head, 0);
                                                                    /* checking for: sort */
                                                                    if (ATgetType((ATerm)arg_arg0_arg0_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0) == PERR_afun11) {
                                                                      ATerm arg_arg0_arg0_head_arg0_arg0 = ATgetArgument(arg_arg0_arg0_head_arg0, 0);
                                                                      /* checking for: Area */
                                                                      if (ATgetType((ATerm)arg_arg0_arg0_head_arg0_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0_arg0) == PERR_afun54) {
                                                                        if (ATgetType((ATerm)arg_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0_list) == ATfalse) {
                                                                          arg_arg0_arg0_head = ATgetFirst(arg_arg0_arg0_list);
                                                                          arg_arg0_arg0_list = ATgetNext(arg_arg0_arg0_list);
                                                                          /* checking for: cf */
                                                                          if (ATgetType((ATerm)arg_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head) == PERR_afun2) {
                                                                            ATerm arg_arg0_arg0_head_arg0 = ATgetArgument(arg_arg0_arg0_head, 0);
                                                                            /* checking for: opt */
                                                                            if (ATgetType((ATerm)arg_arg0_arg0_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0) == PERR_afun3) {
                                                                              ATerm arg_arg0_arg0_head_arg0_arg0 = ATgetArgument(arg_arg0_arg0_head_arg0, 0);
                                                                              /* checking for: layout */
                                                                              if (ATgetType((ATerm)arg_arg0_arg0_head_arg0_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0_arg0) == PERR_afun4) {
                                                                                if (ATgetType((ATerm)arg_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0_list) == ATfalse) {
                                                                                  arg_arg0_arg0_head = ATgetFirst(arg_arg0_arg0_list);
                                                                                  arg_arg0_arg0_list = ATgetNext(arg_arg0_arg0_list);
                                                                                  /* checking for: lit */
                                                                                  if (ATgetType((ATerm)arg_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head) == PERR_afun9) {
                                                                                    ATerm arg_arg0_arg0_head_arg0 = ATgetArgument(arg_arg0_arg0_head, 0);
                                                                                    /* checking for: ) */
                                                                                    if (ATgetType((ATerm)arg_arg0_arg0_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0) == PERR_afun38) {
                                                                                      if (ATgetType((ATerm)arg_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0_list) == ATtrue) {
                                                                                        ATerm arg_arg0_arg1 = ATgetArgument(arg_arg0, 1);
                                                                                        /* checking for: cf */
                                                                                        if (ATgetType((ATerm)arg_arg0_arg1) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg1) == PERR_afun2) {
                                                                                          ATerm arg_arg0_arg1_arg0 = ATgetArgument(arg_arg0_arg1, 0);
                                                                                          /* checking for: sort */
                                                                                          if (ATgetType((ATerm)arg_arg0_arg1_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg1_arg0) == PERR_afun11) {
                                                                                            ATerm arg_arg0_arg1_arg0_arg0 = ATgetArgument(arg_arg0_arg1_arg0, 0);
                                                                                            /* checking for: Location */
                                                                                            if (ATgetType((ATerm)arg_arg0_arg1_arg0_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg1_arg0_arg0) == PERR_afun42) {
                                                                                              ATerm arg_arg0_arg2 = ATgetArgument(arg_arg0, 2);
                                                                                              /* checking for: attrs */
                                                                                              if (ATgetType((ATerm)arg_arg0_arg2) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg2) == PERR_afun13) {
                                                                                                ATerm arg_arg0_arg2_arg0 = ATgetArgument(arg_arg0_arg2, 0);
                                                                                                if (ATgetType((ATerm)arg_arg0_arg2_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg2_arg0) == ATfalse) {
                                                                                                  ATerm arg_arg0_arg2_arg0_head;
                                                                                                  ATermList arg_arg0_arg2_arg0_list = (ATermList)arg_arg0_arg2_arg0;
                                                                                                  arg_arg0_arg2_arg0_head = ATgetFirst(arg_arg0_arg2_arg0_list);
                                                                                                  arg_arg0_arg2_arg0_list = ATgetNext(arg_arg0_arg2_arg0_list);
                                                                                                  /* checking for: term */
                                                                                                  if (ATgetType((ATerm)arg_arg0_arg2_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg2_arg0_head) == PERR_afun14) {
                                                                                                    ATerm arg_arg0_arg2_arg0_head_arg0 = ATgetArgument(arg_arg0_arg2_arg0_head, 0);
                                                                                                    /* checking for: cons */
                                                                                                    if (ATgetType((ATerm)arg_arg0_arg2_arg0_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg2_arg0_head_arg0) == PERR_afun15) {
                                                                                                      ATerm arg_arg0_arg2_arg0_head_arg0_arg0 = ATgetArgument(arg_arg0_arg2_arg0_head_arg0, 0);
                                                                                                      /* checking for: area-in-file */
                                                                                                      if (ATgetType((ATerm)arg_arg0_arg2_arg0_head_arg0_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg2_arg0_head_arg0_arg0) == PERR_afun56) {
                                                                                                        if (ATgetType((ATerm)arg_arg0_arg2_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg2_arg0_list) == ATtrue) {
                                                                                                          ATerm arg_arg1 = ATgetArgument(arg, 1);
                                                                                                          if (ATgetType((ATerm)arg_arg1) == AT_LIST && ATisEmpty((ATermList)arg_arg1) == ATfalse) {
                                                                                                            ATerm arg_arg1_head;
                                                                                                            ATermList arg_arg1_list = (ATermList)arg_arg1;
                                                                                                            arg_arg1_head = ATgetFirst(arg_arg1_list);
                                                                                                            arg_arg1_list = ATgetNext(arg_arg1_list);
                                                                                                            /* checking for: appl */
                                                                                                            if (ATgetType((ATerm)arg_arg1_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head) == PERR_afun0) {
                                                                                                              ATerm arg_arg1_head_arg0 = ATgetArgument(arg_arg1_head, 0);
                                                                                                              /* checking for: prod */
                                                                                                              if (ATgetType((ATerm)arg_arg1_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0) == PERR_afun1) {
                                                                                                                ATerm arg_arg1_head_arg0_arg0 = ATgetArgument(arg_arg1_head_arg0, 0);
                                                                                                                if (ATgetType((ATerm)arg_arg1_head_arg0_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0) == ATfalse) {
                                                                                                                  ATerm arg_arg1_head_arg0_arg0_head;
                                                                                                                  ATermList arg_arg1_head_arg0_arg0_list = (ATermList)arg_arg1_head_arg0_arg0;
                                                                                                                  arg_arg1_head_arg0_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_list);
                                                                                                                  arg_arg1_head_arg0_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_list);
                                                                                                                  /* checking for: char-class */
                                                                                                                  if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg0_head) == PERR_afun17) {
                                                                                                                    ATerm arg_arg1_head_arg0_arg0_head_arg0 = ATgetArgument(arg_arg1_head_arg0_arg0_head, 0);
                                                                                                                    if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0) == ATfalse) {
                                                                                                                      ATerm arg_arg1_head_arg0_arg0_head_arg0_head;
                                                                                                                      ATermList arg_arg1_head_arg0_arg0_head_arg0_list = (ATermList)arg_arg1_head_arg0_arg0_head_arg0;
                                                                                                                      arg_arg1_head_arg0_arg0_head_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                                      arg_arg1_head_arg0_arg0_head_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                                      if (ATgetInt((ATermInt)arg_arg1_head_arg0_arg0_head_arg0_head) == 97) {

                                                                                                                        if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0_list) == ATtrue) {
                                                                                                                          if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_list) == ATfalse) {
                                                                                                                            arg_arg1_head_arg0_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_list);
                                                                                                                            arg_arg1_head_arg0_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_list);
                                                                                                                            /* checking for: char-class */
                                                                                                                            if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg0_head) == PERR_afun17) {
                                                                                                                              ATerm arg_arg1_head_arg0_arg0_head_arg0 = ATgetArgument(arg_arg1_head_arg0_arg0_head, 0);
                                                                                                                              if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0) == ATfalse) {
                                                                                                                                ATerm arg_arg1_head_arg0_arg0_head_arg0_head;
                                                                                                                                ATermList arg_arg1_head_arg0_arg0_head_arg0_list = (ATermList)arg_arg1_head_arg0_arg0_head_arg0;
                                                                                                                                arg_arg1_head_arg0_arg0_head_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                                                arg_arg1_head_arg0_arg0_head_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                                                if (ATgetInt((ATermInt)arg_arg1_head_arg0_arg0_head_arg0_head) == 114) {

                                                                                                                                  if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0_list) == ATtrue) {
                                                                                                                                    if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_list) == ATfalse) {
                                                                                                                                      arg_arg1_head_arg0_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_list);
                                                                                                                                      arg_arg1_head_arg0_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_list);
                                                                                                                                      /* checking for: char-class */
                                                                                                                                      if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg0_head) == PERR_afun17) {
                                                                                                                                        ATerm arg_arg1_head_arg0_arg0_head_arg0 = ATgetArgument(arg_arg1_head_arg0_arg0_head, 0);
                                                                                                                                        if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0) == ATfalse) {
                                                                                                                                          ATerm arg_arg1_head_arg0_arg0_head_arg0_head;
                                                                                                                                          ATermList arg_arg1_head_arg0_arg0_head_arg0_list = (ATermList)arg_arg1_head_arg0_arg0_head_arg0;
                                                                                                                                          arg_arg1_head_arg0_arg0_head_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                                                          arg_arg1_head_arg0_arg0_head_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                                                          if (ATgetInt((ATermInt)arg_arg1_head_arg0_arg0_head_arg0_head) == 101) {

                                                                                                                                            if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0_list) == ATtrue) {
                                                                                                                                              if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_list) == ATfalse) {
                                                                                                                                                arg_arg1_head_arg0_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_list);
                                                                                                                                                arg_arg1_head_arg0_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_list);
                                                                                                                                                /* checking for: char-class */
                                                                                                                                                if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg0_head) == PERR_afun17) {
                                                                                                                                                  ATerm arg_arg1_head_arg0_arg0_head_arg0 = ATgetArgument(arg_arg1_head_arg0_arg0_head, 0);
                                                                                                                                                  if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0) == ATfalse) {
                                                                                                                                                    ATerm arg_arg1_head_arg0_arg0_head_arg0_head;
                                                                                                                                                    ATermList arg_arg1_head_arg0_arg0_head_arg0_list = (ATermList)arg_arg1_head_arg0_arg0_head_arg0;
                                                                                                                                                    arg_arg1_head_arg0_arg0_head_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                                                                    arg_arg1_head_arg0_arg0_head_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                                                                    if (ATgetInt((ATermInt)arg_arg1_head_arg0_arg0_head_arg0_head) == 97) {

                                                                                                                                                      if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0_list) == ATtrue) {
                                                                                                                                                        if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_list) == ATfalse) {
                                                                                                                                                          arg_arg1_head_arg0_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_list);
                                                                                                                                                          arg_arg1_head_arg0_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_list);
                                                                                                                                                          /* checking for: char-class */
                                                                                                                                                          if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg0_head) == PERR_afun17) {
                                                                                                                                                            ATerm arg_arg1_head_arg0_arg0_head_arg0 = ATgetArgument(arg_arg1_head_arg0_arg0_head, 0);
                                                                                                                                                            if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0) == ATfalse) {
                                                                                                                                                              ATerm arg_arg1_head_arg0_arg0_head_arg0_head;
                                                                                                                                                              ATermList arg_arg1_head_arg0_arg0_head_arg0_list = (ATermList)arg_arg1_head_arg0_arg0_head_arg0;
                                                                                                                                                              arg_arg1_head_arg0_arg0_head_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                                                                              arg_arg1_head_arg0_arg0_head_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                                                                              if (ATgetInt((ATermInt)arg_arg1_head_arg0_arg0_head_arg0_head) == 45) {

                                                                                                                                                                if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0_list) == ATtrue) {
                                                                                                                                                                  if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_list) == ATfalse) {
                                                                                                                                                                    arg_arg1_head_arg0_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_list);
                                                                                                                                                                    arg_arg1_head_arg0_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_list);
                                                                                                                                                                    /* checking for: char-class */
                                                                                                                                                                    if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg0_head) == PERR_afun17) {
                                                                                                                                                                      ATerm arg_arg1_head_arg0_arg0_head_arg0 = ATgetArgument(arg_arg1_head_arg0_arg0_head, 0);
                                                                                                                                                                      if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0) == ATfalse) {
                                                                                                                                                                        ATerm arg_arg1_head_arg0_arg0_head_arg0_head;
                                                                                                                                                                        ATermList arg_arg1_head_arg0_arg0_head_arg0_list = (ATermList)arg_arg1_head_arg0_arg0_head_arg0;
                                                                                                                                                                        arg_arg1_head_arg0_arg0_head_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                                                                                        arg_arg1_head_arg0_arg0_head_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                                                                                        if (ATgetInt((ATermInt)arg_arg1_head_arg0_arg0_head_arg0_head) == 105) {

                                                                                                                                                                          if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0_list) == ATtrue) {
                                                                                                                                                                            if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_list) == ATfalse) {
                                                                                                                                                                              arg_arg1_head_arg0_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_list);
                                                                                                                                                                              arg_arg1_head_arg0_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_list);
                                                                                                                                                                              /* checking for: char-class */
                                                                                                                                                                              if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg0_head) == PERR_afun17) {
                                                                                                                                                                                ATerm arg_arg1_head_arg0_arg0_head_arg0 = ATgetArgument(arg_arg1_head_arg0_arg0_head, 0);
                                                                                                                                                                                if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0) == ATfalse) {
                                                                                                                                                                                  ATerm arg_arg1_head_arg0_arg0_head_arg0_head;
                                                                                                                                                                                  ATermList arg_arg1_head_arg0_arg0_head_arg0_list = (ATermList)arg_arg1_head_arg0_arg0_head_arg0;
                                                                                                                                                                                  arg_arg1_head_arg0_arg0_head_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                                                                                                  arg_arg1_head_arg0_arg0_head_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                                                                                                  if (ATgetInt((ATermInt)arg_arg1_head_arg0_arg0_head_arg0_head) == 110) {

                                                                                                                                                                                    if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0_list) == ATtrue) {
                                                                                                                                                                                      if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_list) == ATfalse) {
                                                                                                                                                                                        arg_arg1_head_arg0_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_list);
                                                                                                                                                                                        arg_arg1_head_arg0_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_list);
                                                                                                                                                                                        /* checking for: char-class */
                                                                                                                                                                                        if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg0_head) == PERR_afun17) {
                                                                                                                                                                                          ATerm arg_arg1_head_arg0_arg0_head_arg0 = ATgetArgument(arg_arg1_head_arg0_arg0_head, 0);
                                                                                                                                                                                          if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0) == ATfalse) {
                                                                                                                                                                                            ATerm arg_arg1_head_arg0_arg0_head_arg0_head;
                                                                                                                                                                                            ATermList arg_arg1_head_arg0_arg0_head_arg0_list = (ATermList)arg_arg1_head_arg0_arg0_head_arg0;
                                                                                                                                                                                            arg_arg1_head_arg0_arg0_head_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                                                                                                            arg_arg1_head_arg0_arg0_head_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                                                                                                            if (ATgetInt((ATermInt)arg_arg1_head_arg0_arg0_head_arg0_head) == 45) {

                                                                                                                                                                                              if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0_list) == ATtrue) {
                                                                                                                                                                                                if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_list) == ATfalse) {
                                                                                                                                                                                                  arg_arg1_head_arg0_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_list);
                                                                                                                                                                                                  arg_arg1_head_arg0_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_list);
                                                                                                                                                                                                  /* checking for: char-class */
                                                                                                                                                                                                  if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg0_head) == PERR_afun17) {
                                                                                                                                                                                                    ATerm arg_arg1_head_arg0_arg0_head_arg0 = ATgetArgument(arg_arg1_head_arg0_arg0_head, 0);
                                                                                                                                                                                                    if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0) == ATfalse) {
                                                                                                                                                                                                      ATerm arg_arg1_head_arg0_arg0_head_arg0_head;
                                                                                                                                                                                                      ATermList arg_arg1_head_arg0_arg0_head_arg0_list = (ATermList)arg_arg1_head_arg0_arg0_head_arg0;
                                                                                                                                                                                                      arg_arg1_head_arg0_arg0_head_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                                                                                                                      arg_arg1_head_arg0_arg0_head_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                                                                                                                      if (ATgetInt((ATermInt)arg_arg1_head_arg0_arg0_head_arg0_head) == 102) {

                                                                                                                                                                                                        if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0_list) == ATtrue) {
                                                                                                                                                                                                          if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_list) == ATfalse) {
                                                                                                                                                                                                            arg_arg1_head_arg0_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_list);
                                                                                                                                                                                                            arg_arg1_head_arg0_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_list);
                                                                                                                                                                                                            /* checking for: char-class */
                                                                                                                                                                                                            if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg0_head) == PERR_afun17) {
                                                                                                                                                                                                              ATerm arg_arg1_head_arg0_arg0_head_arg0 = ATgetArgument(arg_arg1_head_arg0_arg0_head, 0);
                                                                                                                                                                                                              if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0) == ATfalse) {
                                                                                                                                                                                                                ATerm arg_arg1_head_arg0_arg0_head_arg0_head;
                                                                                                                                                                                                                ATermList arg_arg1_head_arg0_arg0_head_arg0_list = (ATermList)arg_arg1_head_arg0_arg0_head_arg0;
                                                                                                                                                                                                                arg_arg1_head_arg0_arg0_head_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                                                                                                                                arg_arg1_head_arg0_arg0_head_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                                                                                                                                if (ATgetInt((ATermInt)arg_arg1_head_arg0_arg0_head_arg0_head) == 105) {

                                                                                                                                                                                                                  if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0_list) == ATtrue) {
                                                                                                                                                                                                                    if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_list) == ATfalse) {
                                                                                                                                                                                                                      arg_arg1_head_arg0_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_list);
                                                                                                                                                                                                                      arg_arg1_head_arg0_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_list);
                                                                                                                                                                                                                      /* checking for: char-class */
                                                                                                                                                                                                                      if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg0_head) == PERR_afun17) {
                                                                                                                                                                                                                        ATerm arg_arg1_head_arg0_arg0_head_arg0 = ATgetArgument(arg_arg1_head_arg0_arg0_head, 0);
                                                                                                                                                                                                                        if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0) == ATfalse) {
                                                                                                                                                                                                                          ATerm arg_arg1_head_arg0_arg0_head_arg0_head;
                                                                                                                                                                                                                          ATermList arg_arg1_head_arg0_arg0_head_arg0_list = (ATermList)arg_arg1_head_arg0_arg0_head_arg0;
                                                                                                                                                                                                                          arg_arg1_head_arg0_arg0_head_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                                                                                                                                          arg_arg1_head_arg0_arg0_head_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                                                                                                                                          if (ATgetInt((ATermInt)arg_arg1_head_arg0_arg0_head_arg0_head) == 108) {

                                                                                                                                                                                                                            if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0_list) == ATtrue) {
                                                                                                                                                                                                                              if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_list) == ATfalse) {
                                                                                                                                                                                                                                arg_arg1_head_arg0_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_list);
                                                                                                                                                                                                                                arg_arg1_head_arg0_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_list);
                                                                                                                                                                                                                                /* checking for: char-class */
                                                                                                                                                                                                                                if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg0_head) == PERR_afun17) {
                                                                                                                                                                                                                                  ATerm arg_arg1_head_arg0_arg0_head_arg0 = ATgetArgument(arg_arg1_head_arg0_arg0_head, 0);
                                                                                                                                                                                                                                  if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0) == ATfalse) {
                                                                                                                                                                                                                                    ATerm arg_arg1_head_arg0_arg0_head_arg0_head;
                                                                                                                                                                                                                                    ATermList arg_arg1_head_arg0_arg0_head_arg0_list = (ATermList)arg_arg1_head_arg0_arg0_head_arg0;
                                                                                                                                                                                                                                    arg_arg1_head_arg0_arg0_head_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                                                                                                                                                    arg_arg1_head_arg0_arg0_head_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                                                                                                                                                    if (ATgetInt((ATermInt)arg_arg1_head_arg0_arg0_head_arg0_head) == 101) {

                                                                                                                                                                                                                                      if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0_list) == ATtrue) {
                                                                                                                                                                                                                                        if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_list) == ATtrue) {
                                                                                                                                                                                                                                          ATerm arg_arg1_head_arg0_arg1 = ATgetArgument(arg_arg1_head_arg0, 1);
                                                                                                                                                                                                                                          /* checking for: lit */
                                                                                                                                                                                                                                          if (ATgetType((ATerm)arg_arg1_head_arg0_arg1) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg1) == PERR_afun9) {
                                                                                                                                                                                                                                            ATerm arg_arg1_head_arg0_arg1_arg0 = ATgetArgument(arg_arg1_head_arg0_arg1, 0);
                                                                                                                                                                                                                                            /* checking for: area-in-file */
                                                                                                                                                                                                                                            if (ATgetType((ATerm)arg_arg1_head_arg0_arg1_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg1_arg0) == PERR_afun56) {
                                                                                                                                                                                                                                              ATerm arg_arg1_head_arg0_arg2 = ATgetArgument(arg_arg1_head_arg0, 2);
                                                                                                                                                                                                                                              /* checking for: no-attrs */
                                                                                                                                                                                                                                              if (ATgetType((ATerm)arg_arg1_head_arg0_arg2) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg2) == PERR_afun5) {
                                                                                                                                                                                                                                                ATerm arg_arg1_head_arg1 = ATgetArgument(arg_arg1_head, 1);
                                                                                                                                                                                                                                                if (ATgetType((ATerm)arg_arg1_head_arg1) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg1) == ATfalse) {
                                                                                                                                                                                                                                                  ATerm arg_arg1_head_arg1_head;
                                                                                                                                                                                                                                                  ATermList arg_arg1_head_arg1_list = (ATermList)arg_arg1_head_arg1;
                                                                                                                                                                                                                                                  arg_arg1_head_arg1_head = ATgetFirst(arg_arg1_head_arg1_list);
                                                                                                                                                                                                                                                  arg_arg1_head_arg1_list = ATgetNext(arg_arg1_head_arg1_list);
                                                                                                                                                                                                                                                  if (ATgetInt((ATermInt)arg_arg1_head_arg1_head) == 97) {

                                                                                                                                                                                                                                                    if (ATgetType((ATerm)arg_arg1_head_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg1_list) == ATfalse) {
                                                                                                                                                                                                                                                      arg_arg1_head_arg1_head = ATgetFirst(arg_arg1_head_arg1_list);
                                                                                                                                                                                                                                                      arg_arg1_head_arg1_list = ATgetNext(arg_arg1_head_arg1_list);
                                                                                                                                                                                                                                                      if (ATgetInt((ATermInt)arg_arg1_head_arg1_head) == 114) {

                                                                                                                                                                                                                                                        if (ATgetType((ATerm)arg_arg1_head_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg1_list) == ATfalse) {
                                                                                                                                                                                                                                                          arg_arg1_head_arg1_head = ATgetFirst(arg_arg1_head_arg1_list);
                                                                                                                                                                                                                                                          arg_arg1_head_arg1_list = ATgetNext(arg_arg1_head_arg1_list);
                                                                                                                                                                                                                                                          if (ATgetInt((ATermInt)arg_arg1_head_arg1_head) == 101) {

                                                                                                                                                                                                                                                            if (ATgetType((ATerm)arg_arg1_head_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg1_list) == ATfalse) {
                                                                                                                                                                                                                                                              arg_arg1_head_arg1_head = ATgetFirst(arg_arg1_head_arg1_list);
                                                                                                                                                                                                                                                              arg_arg1_head_arg1_list = ATgetNext(arg_arg1_head_arg1_list);
                                                                                                                                                                                                                                                              if (ATgetInt((ATermInt)arg_arg1_head_arg1_head) == 97) {

                                                                                                                                                                                                                                                                if (ATgetType((ATerm)arg_arg1_head_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg1_list) == ATfalse) {
                                                                                                                                                                                                                                                                  arg_arg1_head_arg1_head = ATgetFirst(arg_arg1_head_arg1_list);
                                                                                                                                                                                                                                                                  arg_arg1_head_arg1_list = ATgetNext(arg_arg1_head_arg1_list);
                                                                                                                                                                                                                                                                  if (ATgetInt((ATermInt)arg_arg1_head_arg1_head) == 45) {

                                                                                                                                                                                                                                                                    if (ATgetType((ATerm)arg_arg1_head_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg1_list) == ATfalse) {
                                                                                                                                                                                                                                                                      arg_arg1_head_arg1_head = ATgetFirst(arg_arg1_head_arg1_list);
                                                                                                                                                                                                                                                                      arg_arg1_head_arg1_list = ATgetNext(arg_arg1_head_arg1_list);
                                                                                                                                                                                                                                                                      if (ATgetInt((ATermInt)arg_arg1_head_arg1_head) == 105) {

                                                                                                                                                                                                                                                                        if (ATgetType((ATerm)arg_arg1_head_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg1_list) == ATfalse) {
                                                                                                                                                                                                                                                                          arg_arg1_head_arg1_head = ATgetFirst(arg_arg1_head_arg1_list);
                                                                                                                                                                                                                                                                          arg_arg1_head_arg1_list = ATgetNext(arg_arg1_head_arg1_list);
                                                                                                                                                                                                                                                                          if (ATgetInt((ATermInt)arg_arg1_head_arg1_head) == 110) {

                                                                                                                                                                                                                                                                            if (ATgetType((ATerm)arg_arg1_head_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg1_list) == ATfalse) {
                                                                                                                                                                                                                                                                              arg_arg1_head_arg1_head = ATgetFirst(arg_arg1_head_arg1_list);
                                                                                                                                                                                                                                                                              arg_arg1_head_arg1_list = ATgetNext(arg_arg1_head_arg1_list);
                                                                                                                                                                                                                                                                              if (ATgetInt((ATermInt)arg_arg1_head_arg1_head) == 45) {

                                                                                                                                                                                                                                                                                if (ATgetType((ATerm)arg_arg1_head_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg1_list) == ATfalse) {
                                                                                                                                                                                                                                                                                  arg_arg1_head_arg1_head = ATgetFirst(arg_arg1_head_arg1_list);
                                                                                                                                                                                                                                                                                  arg_arg1_head_arg1_list = ATgetNext(arg_arg1_head_arg1_list);
                                                                                                                                                                                                                                                                                  if (ATgetInt((ATermInt)arg_arg1_head_arg1_head) == 102) {

                                                                                                                                                                                                                                                                                    if (ATgetType((ATerm)arg_arg1_head_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg1_list) == ATfalse) {
                                                                                                                                                                                                                                                                                      arg_arg1_head_arg1_head = ATgetFirst(arg_arg1_head_arg1_list);
                                                                                                                                                                                                                                                                                      arg_arg1_head_arg1_list = ATgetNext(arg_arg1_head_arg1_list);
                                                                                                                                                                                                                                                                                      if (ATgetInt((ATermInt)arg_arg1_head_arg1_head) == 105) {

                                                                                                                                                                                                                                                                                        if (ATgetType((ATerm)arg_arg1_head_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg1_list) == ATfalse) {
                                                                                                                                                                                                                                                                                          arg_arg1_head_arg1_head = ATgetFirst(arg_arg1_head_arg1_list);
                                                                                                                                                                                                                                                                                          arg_arg1_head_arg1_list = ATgetNext(arg_arg1_head_arg1_list);
                                                                                                                                                                                                                                                                                          if (ATgetInt((ATermInt)arg_arg1_head_arg1_head) == 108) {

                                                                                                                                                                                                                                                                                            if (ATgetType((ATerm)arg_arg1_head_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg1_list) == ATfalse) {
                                                                                                                                                                                                                                                                                              arg_arg1_head_arg1_head = ATgetFirst(arg_arg1_head_arg1_list);
                                                                                                                                                                                                                                                                                              arg_arg1_head_arg1_list = ATgetNext(arg_arg1_head_arg1_list);
                                                                                                                                                                                                                                                                                              if (ATgetInt((ATermInt)arg_arg1_head_arg1_head) == 101) {

                                                                                                                                                                                                                                                                                                if (ATgetType((ATerm)arg_arg1_head_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg1_list) == ATtrue) {
                                                                                                                                                                                                                                                                                                  if (ATgetType((ATerm)arg_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_list) == ATfalse) {
                                                                                                                                                                                                                                                                                                    arg_arg1_list = ATgetNext(arg_arg1_list);
                                                                                                                                                                                                                                                                                                    if (ATgetType((ATerm)arg_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_list) == ATfalse) {
                                                                                                                                                                                                                                                                                                      arg_arg1_head = ATgetFirst(arg_arg1_list);
                                                                                                                                                                                                                                                                                                      arg_arg1_list = ATgetNext(arg_arg1_list);
                                                                                                                                                                                                                                                                                                      /* checking for: appl */
                                                                                                                                                                                                                                                                                                      if (ATgetType((ATerm)arg_arg1_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head) == PERR_afun0) {
                                                                                                                                                                                                                                                                                                        ATerm arg_arg1_head_arg0 = ATgetArgument(arg_arg1_head, 0);
                                                                                                                                                                                                                                                                                                        /* checking for: prod */
                                                                                                                                                                                                                                                                                                        if (ATgetType((ATerm)arg_arg1_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0) == PERR_afun1) {
                                                                                                                                                                                                                                                                                                          ATerm arg_arg1_head_arg0_arg0 = ATgetArgument(arg_arg1_head_arg0, 0);
                                                                                                                                                                                                                                                                                                          if (ATgetType((ATerm)arg_arg1_head_arg0_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0) == ATfalse) {
                                                                                                                                                                                                                                                                                                            ATerm arg_arg1_head_arg0_arg0_head;
                                                                                                                                                                                                                                                                                                            ATermList arg_arg1_head_arg0_arg0_list = (ATermList)arg_arg1_head_arg0_arg0;
                                                                                                                                                                                                                                                                                                            arg_arg1_head_arg0_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_list);
                                                                                                                                                                                                                                                                                                            arg_arg1_head_arg0_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_list);
                                                                                                                                                                                                                                                                                                            /* checking for: char-class */
                                                                                                                                                                                                                                                                                                            if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg0_head) == PERR_afun17) {
                                                                                                                                                                                                                                                                                                              ATerm arg_arg1_head_arg0_arg0_head_arg0 = ATgetArgument(arg_arg1_head_arg0_arg0_head, 0);
                                                                                                                                                                                                                                                                                                              if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0) == ATfalse) {
                                                                                                                                                                                                                                                                                                                ATerm arg_arg1_head_arg0_arg0_head_arg0_head;
                                                                                                                                                                                                                                                                                                                ATermList arg_arg1_head_arg0_arg0_head_arg0_list = (ATermList)arg_arg1_head_arg0_arg0_head_arg0;
                                                                                                                                                                                                                                                                                                                arg_arg1_head_arg0_arg0_head_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                                                                                                                                                                                                                                arg_arg1_head_arg0_arg0_head_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                                                                                                                                                                                                                                if (ATgetInt((ATermInt)arg_arg1_head_arg0_arg0_head_arg0_head) == 40) {

                                                                                                                                                                                                                                                                                                                  if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0_list) == ATtrue) {
                                                                                                                                                                                                                                                                                                                    if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_list) == ATtrue) {
                                                                                                                                                                                                                                                                                                                      ATerm arg_arg1_head_arg0_arg1 = ATgetArgument(arg_arg1_head_arg0, 1);
                                                                                                                                                                                                                                                                                                                      /* checking for: lit */
                                                                                                                                                                                                                                                                                                                      if (ATgetType((ATerm)arg_arg1_head_arg0_arg1) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg1) == PERR_afun9) {
                                                                                                                                                                                                                                                                                                                        ATerm arg_arg1_head_arg0_arg1_arg0 = ATgetArgument(arg_arg1_head_arg0_arg1, 0);
                                                                                                                                                                                                                                                                                                                        /* checking for: ( */
                                                                                                                                                                                                                                                                                                                        if (ATgetType((ATerm)arg_arg1_head_arg0_arg1_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg1_arg0) == PERR_afun39) {
                                                                                                                                                                                                                                                                                                                          ATerm arg_arg1_head_arg0_arg2 = ATgetArgument(arg_arg1_head_arg0, 2);
                                                                                                                                                                                                                                                                                                                          /* checking for: no-attrs */
                                                                                                                                                                                                                                                                                                                          if (ATgetType((ATerm)arg_arg1_head_arg0_arg2) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg2) == PERR_afun5) {
                                                                                                                                                                                                                                                                                                                            ATerm arg_arg1_head_arg1 = ATgetArgument(arg_arg1_head, 1);
                                                                                                                                                                                                                                                                                                                            if (ATgetType((ATerm)arg_arg1_head_arg1) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg1) == ATfalse) {
                                                                                                                                                                                                                                                                                                                              ATerm arg_arg1_head_arg1_head;
                                                                                                                                                                                                                                                                                                                              ATermList arg_arg1_head_arg1_list = (ATermList)arg_arg1_head_arg1;
                                                                                                                                                                                                                                                                                                                              arg_arg1_head_arg1_head = ATgetFirst(arg_arg1_head_arg1_list);
                                                                                                                                                                                                                                                                                                                              arg_arg1_head_arg1_list = ATgetNext(arg_arg1_head_arg1_list);
                                                                                                                                                                                                                                                                                                                              if (ATgetInt((ATermInt)arg_arg1_head_arg1_head) == 40) {

                                                                                                                                                                                                                                                                                                                                if (ATgetType((ATerm)arg_arg1_head_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg1_list) == ATtrue) {
                                                                                                                                                                                                                                                                                                                                  if (ATgetType((ATerm)arg_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_list) == ATfalse) {
                                                                                                                                                                                                                                                                                                                                    arg_arg1_list = ATgetNext(arg_arg1_list);
                                                                                                                                                                                                                                                                                                                                    if (ATgetType((ATerm)arg_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_list) == ATfalse) {
                                                                                                                                                                                                                                                                                                                                      arg_arg1_list = ATgetNext(arg_arg1_list);
                                                                                                                                                                                                                                                                                                                                      if (ATgetType((ATerm)arg_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_list) == ATfalse) {
                                                                                                                                                                                                                                                                                                                                        arg_arg1_list = ATgetNext(arg_arg1_list);
                                                                                                                                                                                                                                                                                                                                        if (ATgetType((ATerm)arg_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_list) == ATfalse) {
                                                                                                                                                                                                                                                                                                                                          arg_arg1_head = ATgetFirst(arg_arg1_list);
                                                                                                                                                                                                                                                                                                                                          arg_arg1_list = ATgetNext(arg_arg1_list);
                                                                                                                                                                                                                                                                                                                                          /* checking for: appl */
                                                                                                                                                                                                                                                                                                                                          if (ATgetType((ATerm)arg_arg1_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head) == PERR_afun0) {
                                                                                                                                                                                                                                                                                                                                            ATerm arg_arg1_head_arg0 = ATgetArgument(arg_arg1_head, 0);
                                                                                                                                                                                                                                                                                                                                            /* checking for: prod */
                                                                                                                                                                                                                                                                                                                                            if (ATgetType((ATerm)arg_arg1_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0) == PERR_afun1) {
                                                                                                                                                                                                                                                                                                                                              ATerm arg_arg1_head_arg0_arg0 = ATgetArgument(arg_arg1_head_arg0, 0);
                                                                                                                                                                                                                                                                                                                                              if (ATgetType((ATerm)arg_arg1_head_arg0_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0) == ATfalse) {
                                                                                                                                                                                                                                                                                                                                                ATerm arg_arg1_head_arg0_arg0_head;
                                                                                                                                                                                                                                                                                                                                                ATermList arg_arg1_head_arg0_arg0_list = (ATermList)arg_arg1_head_arg0_arg0;
                                                                                                                                                                                                                                                                                                                                                arg_arg1_head_arg0_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_list);
                                                                                                                                                                                                                                                                                                                                                arg_arg1_head_arg0_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_list);
                                                                                                                                                                                                                                                                                                                                                /* checking for: char-class */
                                                                                                                                                                                                                                                                                                                                                if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg0_head) == PERR_afun17) {
                                                                                                                                                                                                                                                                                                                                                  ATerm arg_arg1_head_arg0_arg0_head_arg0 = ATgetArgument(arg_arg1_head_arg0_arg0_head, 0);
                                                                                                                                                                                                                                                                                                                                                  if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0) == ATfalse) {
                                                                                                                                                                                                                                                                                                                                                    ATerm arg_arg1_head_arg0_arg0_head_arg0_head;
                                                                                                                                                                                                                                                                                                                                                    ATermList arg_arg1_head_arg0_arg0_head_arg0_list = (ATermList)arg_arg1_head_arg0_arg0_head_arg0;
                                                                                                                                                                                                                                                                                                                                                    arg_arg1_head_arg0_arg0_head_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                                                                                                                                                                                                                                                                    arg_arg1_head_arg0_arg0_head_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                                                                                                                                                                                                                                                                    if (ATgetInt((ATermInt)arg_arg1_head_arg0_arg0_head_arg0_head) == 44) {

                                                                                                                                                                                                                                                                                                                                                      if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0_list) == ATtrue) {
                                                                                                                                                                                                                                                                                                                                                        if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_list) == ATtrue) {
                                                                                                                                                                                                                                                                                                                                                          ATerm arg_arg1_head_arg0_arg1 = ATgetArgument(arg_arg1_head_arg0, 1);
                                                                                                                                                                                                                                                                                                                                                          /* checking for: lit */
                                                                                                                                                                                                                                                                                                                                                          if (ATgetType((ATerm)arg_arg1_head_arg0_arg1) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg1) == PERR_afun9) {
                                                                                                                                                                                                                                                                                                                                                            ATerm arg_arg1_head_arg0_arg1_arg0 = ATgetArgument(arg_arg1_head_arg0_arg1, 0);
                                                                                                                                                                                                                                                                                                                                                            /* checking for: , */
                                                                                                                                                                                                                                                                                                                                                            if (ATgetType((ATerm)arg_arg1_head_arg0_arg1_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg1_arg0) == PERR_afun43) {
                                                                                                                                                                                                                                                                                                                                                              ATerm arg_arg1_head_arg0_arg2 = ATgetArgument(arg_arg1_head_arg0, 2);
                                                                                                                                                                                                                                                                                                                                                              /* checking for: no-attrs */
                                                                                                                                                                                                                                                                                                                                                              if (ATgetType((ATerm)arg_arg1_head_arg0_arg2) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg2) == PERR_afun5) {
                                                                                                                                                                                                                                                                                                                                                                ATerm arg_arg1_head_arg1 = ATgetArgument(arg_arg1_head, 1);
                                                                                                                                                                                                                                                                                                                                                                if (ATgetType((ATerm)arg_arg1_head_arg1) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg1) == ATfalse) {
                                                                                                                                                                                                                                                                                                                                                                  ATerm arg_arg1_head_arg1_head;
                                                                                                                                                                                                                                                                                                                                                                  ATermList arg_arg1_head_arg1_list = (ATermList)arg_arg1_head_arg1;
                                                                                                                                                                                                                                                                                                                                                                  arg_arg1_head_arg1_head = ATgetFirst(arg_arg1_head_arg1_list);
                                                                                                                                                                                                                                                                                                                                                                  arg_arg1_head_arg1_list = ATgetNext(arg_arg1_head_arg1_list);
                                                                                                                                                                                                                                                                                                                                                                  if (ATgetInt((ATermInt)arg_arg1_head_arg1_head) == 44) {

                                                                                                                                                                                                                                                                                                                                                                    if (ATgetType((ATerm)arg_arg1_head_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg1_list) == ATtrue) {
                                                                                                                                                                                                                                                                                                                                                                      if (ATgetType((ATerm)arg_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_list) == ATfalse) {
                                                                                                                                                                                                                                                                                                                                                                        arg_arg1_list = ATgetNext(arg_arg1_list);
                                                                                                                                                                                                                                                                                                                                                                        if (ATgetType((ATerm)arg_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_list) == ATfalse) {
                                                                                                                                                                                                                                                                                                                                                                          arg_arg1_list = ATgetNext(arg_arg1_list);
                                                                                                                                                                                                                                                                                                                                                                          if (ATgetType((ATerm)arg_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_list) == ATfalse) {
                                                                                                                                                                                                                                                                                                                                                                            arg_arg1_list = ATgetNext(arg_arg1_list);
                                                                                                                                                                                                                                                                                                                                                                            if (ATgetType((ATerm)arg_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_list) == ATfalse) {
                                                                                                                                                                                                                                                                                                                                                                              arg_arg1_head = ATgetFirst(arg_arg1_list);
                                                                                                                                                                                                                                                                                                                                                                              arg_arg1_list = ATgetNext(arg_arg1_list);
                                                                                                                                                                                                                                                                                                                                                                              /* checking for: appl */
                                                                                                                                                                                                                                                                                                                                                                              if (ATgetType((ATerm)arg_arg1_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head) == PERR_afun0) {
                                                                                                                                                                                                                                                                                                                                                                                ATerm arg_arg1_head_arg0 = ATgetArgument(arg_arg1_head, 0);
                                                                                                                                                                                                                                                                                                                                                                                /* checking for: prod */
                                                                                                                                                                                                                                                                                                                                                                                if (ATgetType((ATerm)arg_arg1_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0) == PERR_afun1) {
                                                                                                                                                                                                                                                                                                                                                                                  ATerm arg_arg1_head_arg0_arg0 = ATgetArgument(arg_arg1_head_arg0, 0);
                                                                                                                                                                                                                                                                                                                                                                                  if (ATgetType((ATerm)arg_arg1_head_arg0_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0) == ATfalse) {
                                                                                                                                                                                                                                                                                                                                                                                    ATerm arg_arg1_head_arg0_arg0_head;
                                                                                                                                                                                                                                                                                                                                                                                    ATermList arg_arg1_head_arg0_arg0_list = (ATermList)arg_arg1_head_arg0_arg0;
                                                                                                                                                                                                                                                                                                                                                                                    arg_arg1_head_arg0_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_list);
                                                                                                                                                                                                                                                                                                                                                                                    arg_arg1_head_arg0_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_list);
                                                                                                                                                                                                                                                                                                                                                                                    /* checking for: char-class */
                                                                                                                                                                                                                                                                                                                                                                                    if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg0_head) == PERR_afun17) {
                                                                                                                                                                                                                                                                                                                                                                                      ATerm arg_arg1_head_arg0_arg0_head_arg0 = ATgetArgument(arg_arg1_head_arg0_arg0_head, 0);
                                                                                                                                                                                                                                                                                                                                                                                      if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0) == ATfalse) {
                                                                                                                                                                                                                                                                                                                                                                                        ATerm arg_arg1_head_arg0_arg0_head_arg0_head;
                                                                                                                                                                                                                                                                                                                                                                                        ATermList arg_arg1_head_arg0_arg0_head_arg0_list = (ATermList)arg_arg1_head_arg0_arg0_head_arg0;
                                                                                                                                                                                                                                                                                                                                                                                        arg_arg1_head_arg0_arg0_head_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                                                                                                                                                                                                                                                                                                        arg_arg1_head_arg0_arg0_head_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                                                                                                                                                                                                                                                                                                        if (ATgetInt((ATermInt)arg_arg1_head_arg0_arg0_head_arg0_head) == 41) {

                                                                                                                                                                                                                                                                                                                                                                                          if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0_list) == ATtrue) {
                                                                                                                                                                                                                                                                                                                                                                                            if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_list) == ATtrue) {
                                                                                                                                                                                                                                                                                                                                                                                              ATerm arg_arg1_head_arg0_arg1 = ATgetArgument(arg_arg1_head_arg0, 1);
                                                                                                                                                                                                                                                                                                                                                                                              /* checking for: lit */
                                                                                                                                                                                                                                                                                                                                                                                              if (ATgetType((ATerm)arg_arg1_head_arg0_arg1) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg1) == PERR_afun9) {
                                                                                                                                                                                                                                                                                                                                                                                                ATerm arg_arg1_head_arg0_arg1_arg0 = ATgetArgument(arg_arg1_head_arg0_arg1, 0);
                                                                                                                                                                                                                                                                                                                                                                                                /* checking for: ) */
                                                                                                                                                                                                                                                                                                                                                                                                if (ATgetType((ATerm)arg_arg1_head_arg0_arg1_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg1_arg0) == PERR_afun38) {
                                                                                                                                                                                                                                                                                                                                                                                                  ATerm arg_arg1_head_arg0_arg2 = ATgetArgument(arg_arg1_head_arg0, 2);
                                                                                                                                                                                                                                                                                                                                                                                                  /* checking for: no-attrs */
                                                                                                                                                                                                                                                                                                                                                                                                  if (ATgetType((ATerm)arg_arg1_head_arg0_arg2) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg2) == PERR_afun5) {
                                                                                                                                                                                                                                                                                                                                                                                                    ATerm arg_arg1_head_arg1 = ATgetArgument(arg_arg1_head, 1);
                                                                                                                                                                                                                                                                                                                                                                                                    if (ATgetType((ATerm)arg_arg1_head_arg1) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg1) == ATfalse) {
                                                                                                                                                                                                                                                                                                                                                                                                      ATerm arg_arg1_head_arg1_head;
                                                                                                                                                                                                                                                                                                                                                                                                      ATermList arg_arg1_head_arg1_list = (ATermList)arg_arg1_head_arg1;
                                                                                                                                                                                                                                                                                                                                                                                                      arg_arg1_head_arg1_head = ATgetFirst(arg_arg1_head_arg1_list);
                                                                                                                                                                                                                                                                                                                                                                                                      arg_arg1_head_arg1_list = ATgetNext(arg_arg1_head_arg1_list);
                                                                                                                                                                                                                                                                                                                                                                                                      if (ATgetInt((ATermInt)arg_arg1_head_arg1_head) == 41) {

                                                                                                                                                                                                                                                                                                                                                                                                        if (ATgetType((ATerm)arg_arg1_head_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg1_list) == ATtrue) {
                                                                                                                                                                                                                                                                                                                                                                                                          if (ATgetType((ATerm)arg_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_list) == ATtrue) {
                                                                                                                                                                                                                                                                                                                                                                                                            return ATtrue;
                                                                                                                                                                                                                                                                                                                                                                                                          }
                                                                                                                                                                                                                                                                                                                                                                                                        }
                                                                                                                                                                                                                                                                                                                                                                                                      }
                                                                                                                                                                                                                                                                                                                                                                                                    }
                                                                                                                                                                                                                                                                                                                                                                                                  }
                                                                                                                                                                                                                                                                                                                                                                                                }
                                                                                                                                                                                                                                                                                                                                                                                              }
                                                                                                                                                                                                                                                                                                                                                                                            }
                                                                                                                                                                                                                                                                                                                                                                                          }
                                                                                                                                                                                                                                                                                                                                                                                        }
                                                                                                                                                                                                                                                                                                                                                                                      }
                                                                                                                                                                                                                                                                                                                                                                                    }
                                                                                                                                                                                                                                                                                                                                                                                  }
                                                                                                                                                                                                                                                                                                                                                                                }
                                                                                                                                                                                                                                                                                                                                                                              }
                                                                                                                                                                                                                                                                                                                                                                            }
                                                                                                                                                                                                                                                                                                                                                                          }
                                                                                                                                                                                                                                                                                                                                                                        }
                                                                                                                                                                                                                                                                                                                                                                      }
                                                                                                                                                                                                                                                                                                                                                                    }
                                                                                                                                                                                                                                                                                                                                                                  }
                                                                                                                                                                                                                                                                                                                                                                }
                                                                                                                                                                                                                                                                                                                                                              }
                                                                                                                                                                                                                                                                                                                                                            }
                                                                                                                                                                                                                                                                                                                                                          }
                                                                                                                                                                                                                                                                                                                                                        }
                                                                                                                                                                                                                                                                                                                                                      }
                                                                                                                                                                                                                                                                                                                                                    }
                                                                                                                                                                                                                                                                                                                                                  }
                                                                                                                                                                                                                                                                                                                                                }
                                                                                                                                                                                                                                                                                                                                              }
                                                                                                                                                                                                                                                                                                                                            }
                                                                                                                                                                                                                                                                                                                                          }
                                                                                                                                                                                                                                                                                                                                        }
                                                                                                                                                                                                                                                                                                                                      }
                                                                                                                                                                                                                                                                                                                                    }
                                                                                                                                                                                                                                                                                                                                  }
                                                                                                                                                                                                                                                                                                                                }
                                                                                                                                                                                                                                                                                                                              }
                                                                                                                                                                                                                                                                                                                            }
                                                                                                                                                                                                                                                                                                                          }
                                                                                                                                                                                                                                                                                                                        }
                                                                                                                                                                                                                                                                                                                      }
                                                                                                                                                                                                                                                                                                                    }
                                                                                                                                                                                                                                                                                                                  }
                                                                                                                                                                                                                                                                                                                }
                                                                                                                                                                                                                                                                                                              }
                                                                                                                                                                                                                                                                                                            }
                                                                                                                                                                                                                                                                                                          }
                                                                                                                                                                                                                                                                                                        }
                                                                                                                                                                                                                                                                                                      }
                                                                                                                                                                                                                                                                                                    }
                                                                                                                                                                                                                                                                                                  }
                                                                                                                                                                                                                                                                                                }
                                                                                                                                                                                                                                                                                              }
                                                                                                                                                                                                                                                                                            }
                                                                                                                                                                                                                                                                                          }
                                                                                                                                                                                                                                                                                        }
                                                                                                                                                                                                                                                                                      }
                                                                                                                                                                                                                                                                                    }
                                                                                                                                                                                                                                                                                  }
                                                                                                                                                                                                                                                                                }
                                                                                                                                                                                                                                                                              }
                                                                                                                                                                                                                                                                            }
                                                                                                                                                                                                                                                                          }
                                                                                                                                                                                                                                                                        }
                                                                                                                                                                                                                                                                      }
                                                                                                                                                                                                                                                                    }
                                                                                                                                                                                                                                                                  }
                                                                                                                                                                                                                                                                }
                                                                                                                                                                                                                                                              }
                                                                                                                                                                                                                                                            }
                                                                                                                                                                                                                                                          }
                                                                                                                                                                                                                                                        }
                                                                                                                                                                                                                                                      }
                                                                                                                                                                                                                                                    }
                                                                                                                                                                                                                                                  }
                                                                                                                                                                                                                                                }
                                                                                                                                                                                                                                              }
                                                                                                                                                                                                                                            }
                                                                                                                                                                                                                                          }
                                                                                                                                                                                                                                        }
                                                                                                                                                                                                                                      }
                                                                                                                                                                                                                                    }
                                                                                                                                                                                                                                  }
                                                                                                                                                                                                                                }
                                                                                                                                                                                                                              }
                                                                                                                                                                                                                            }
                                                                                                                                                                                                                          }
                                                                                                                                                                                                                        }
                                                                                                                                                                                                                      }
                                                                                                                                                                                                                    }
                                                                                                                                                                                                                  }
                                                                                                                                                                                                                }
                                                                                                                                                                                                              }
                                                                                                                                                                                                            }
                                                                                                                                                                                                          }
                                                                                                                                                                                                        }
                                                                                                                                                                                                      }
                                                                                                                                                                                                    }
                                                                                                                                                                                                  }
                                                                                                                                                                                                }
                                                                                                                                                                                              }
                                                                                                                                                                                            }
                                                                                                                                                                                          }
                                                                                                                                                                                        }
                                                                                                                                                                                      }
                                                                                                                                                                                    }
                                                                                                                                                                                  }
                                                                                                                                                                                }
                                                                                                                                                                              }
                                                                                                                                                                            }
                                                                                                                                                                          }
                                                                                                                                                                        }
                                                                                                                                                                      }
                                                                                                                                                                    }
                                                                                                                                                                  }
                                                                                                                                                                }
                                                                                                                                                              }
                                                                                                                                                            }
                                                                                                                                                          }
                                                                                                                                                        }
                                                                                                                                                      }
                                                                                                                                                    }
                                                                                                                                                  }
                                                                                                                                                }
                                                                                                                                              }
                                                                                                                                            }
                                                                                                                                          }
                                                                                                                                        }
                                                                                                                                      }
                                                                                                                                    }
                                                                                                                                  }
                                                                                                                                }
                                                                                                                              }
                                                                                                                            }
                                                                                                                          }
                                                                                                                        }
                                                                                                                      }
                                                                                                                    }
                                                                                                                  }
                                                                                                                }
                                                                                                              }
                                                                                                            }
                                                                                                          }
                                                                                                        }
                                                                                                      }
                                                                                                    }
                                                                                                  }
                                                                                                }
                                                                                              }
                                                                                            }
                                                                                          }
                                                                                        }
                                                                                      }
                                                                                    }
                                                                                  }
                                                                                }
                                                                              }
                                                                            }
                                                                          }
                                                                        }
                                                                      }
                                                                    }
                                                                  }
                                                                }
                                                              }
                                                            }
                                                          }
                                                        }
                                                      }
                                                    }
                                                  }
                                                }
                                              }
                                            }
                                          }
                                        }
                                      }
                                    }
                                  }
                                }
                              }
                            }
                          }
                        }
                      }
                    }
                  }
                }
              }
            }
          }
        }
      }
    }
  }
  return ATfalse;
}

/**
 * Assert whether a PERR_Location has a ws-after-file. 
 * \param[in] arg input PERR_Location
 * \return ATtrue if the PERR_Location had a ws-after-file, or ATfalse otherwise
 */
ATbool PERR_hasLocationWsAfterFile(PERR_Location arg) {
  if (PERR_isLocationFile(arg)) {
    return ATtrue;
  }
  return ATfalse;
}

/**
 * Assert whether a PERR_Location has a ws-after-(. 
 * \param[in] arg input PERR_Location
 * \return ATtrue if the PERR_Location had a ws-after-(, or ATfalse otherwise
 */
ATbool PERR_hasLocationWsAfterParenOpen(PERR_Location arg) {
  if (PERR_isLocationFile(arg)) {
    return ATtrue;
  }
  else if (PERR_isLocationArea(arg)) {
    return ATtrue;
  }
  else if (PERR_isLocationAreaInFile(arg)) {
    return ATtrue;
  }
  return ATfalse;
}

/**
 * Assert whether a PERR_Location has a filename. 
 * \param[in] arg input PERR_Location
 * \return ATtrue if the PERR_Location had a filename, or ATfalse otherwise
 */
ATbool PERR_hasLocationFilename(PERR_Location arg) {
  if (PERR_isLocationFile(arg)) {
    return ATtrue;
  }
  else if (PERR_isLocationAreaInFile(arg)) {
    return ATtrue;
  }
  return ATfalse;
}

/**
 * Assert whether a PERR_Location has a ws-after-filename. 
 * \param[in] arg input PERR_Location
 * \return ATtrue if the PERR_Location had a ws-after-filename, or ATfalse otherwise
 */
ATbool PERR_hasLocationWsAfterFilename(PERR_Location arg) {
  if (PERR_isLocationFile(arg)) {
    return ATtrue;
  }
  else if (PERR_isLocationAreaInFile(arg)) {
    return ATtrue;
  }
  return ATfalse;
}

/**
 * Assert whether a PERR_Location has a ws-after-a. 
 * \param[in] arg input PERR_Location
 * \return ATtrue if the PERR_Location had a ws-after-a, or ATfalse otherwise
 */
ATbool PERR_hasLocationWsAfterA(PERR_Location arg) {
  if (PERR_isLocationArea(arg)) {
    return ATtrue;
  }
  return ATfalse;
}

/**
 * Assert whether a PERR_Location has a Area. 
 * \param[in] arg input PERR_Location
 * \return ATtrue if the PERR_Location had a Area, or ATfalse otherwise
 */
ATbool PERR_hasLocationArea(PERR_Location arg) {
  if (PERR_isLocationArea(arg)) {
    return ATtrue;
  }
  else if (PERR_isLocationAreaInFile(arg)) {
    return ATtrue;
  }
  return ATfalse;
}

/**
 * Assert whether a PERR_Location has a ws-after-Area. 
 * \param[in] arg input PERR_Location
 * \return ATtrue if the PERR_Location had a ws-after-Area, or ATfalse otherwise
 */
ATbool PERR_hasLocationWsAfterArea(PERR_Location arg) {
  if (PERR_isLocationArea(arg)) {
    return ATtrue;
  }
  else if (PERR_isLocationAreaInFile(arg)) {
    return ATtrue;
  }
  return ATfalse;
}

/**
 * Assert whether a PERR_Location has a ws-after-area-in-file. 
 * \param[in] arg input PERR_Location
 * \return ATtrue if the PERR_Location had a ws-after-area-in-file, or ATfalse otherwise
 */
ATbool PERR_hasLocationWsAfterAreaInFile(PERR_Location arg) {
  if (PERR_isLocationAreaInFile(arg)) {
    return ATtrue;
  }
  return ATfalse;
}

/**
 * Assert whether a PERR_Location has a ws-after-,. 
 * \param[in] arg input PERR_Location
 * \return ATtrue if the PERR_Location had a ws-after-,, or ATfalse otherwise
 */
ATbool PERR_hasLocationWsAfterComma(PERR_Location arg) {
  if (PERR_isLocationAreaInFile(arg)) {
    return ATtrue;
  }
  return ATfalse;
}

/**
 * Get the ws-after-file PERR_OptLayout of a PERR_Location. Note that the precondition is that this PERR_Location actually has a ws-after-file
 * \param[in] arg input PERR_Location
 * \return the ws-after-file of #arg, if it exist or an undefined value if it does not
 */
PERR_OptLayout PERR_getLocationWsAfterFile(PERR_Location arg) {
  
    return (PERR_OptLayout)ATelementAt((ATermList)ATgetArgument((ATermAppl)arg, 1), 1);
}

/**
 * Get the ws-after-( PERR_OptLayout of a PERR_Location. Note that the precondition is that this PERR_Location actually has a ws-after-(
 * \param[in] arg input PERR_Location
 * \return the ws-after-( of #arg, if it exist or an undefined value if it does not
 */
PERR_OptLayout PERR_getLocationWsAfterParenOpen(PERR_Location arg) {
  if (PERR_isLocationFile(arg)) {
    return (PERR_OptLayout)ATelementAt((ATermList)ATgetArgument((ATermAppl)arg, 1), 3);
  }
  else if (PERR_isLocationArea(arg)) {
    return (PERR_OptLayout)ATelementAt((ATermList)ATgetArgument((ATermAppl)arg, 1), 3);
  }
  else 
    return (PERR_OptLayout)ATelementAt((ATermList)ATgetArgument((ATermAppl)arg, 1), 3);
}

/**
 * Get the filename PERR_StrCon of a PERR_Location. Note that the precondition is that this PERR_Location actually has a filename
 * \param[in] arg input PERR_Location
 * \return the filename of #arg, if it exist or an undefined value if it does not
 */
PERR_StrCon PERR_getLocationFilename(PERR_Location arg) {
  if (PERR_isLocationFile(arg)) {
    return (PERR_StrCon)ATelementAt((ATermList)ATgetArgument((ATermAppl)arg, 1), 4);
  }
  else 
    return (PERR_StrCon)ATelementAt((ATermList)ATgetArgument((ATermAppl)arg, 1), 4);
}

/**
 * Get the ws-after-filename PERR_OptLayout of a PERR_Location. Note that the precondition is that this PERR_Location actually has a ws-after-filename
 * \param[in] arg input PERR_Location
 * \return the ws-after-filename of #arg, if it exist or an undefined value if it does not
 */
PERR_OptLayout PERR_getLocationWsAfterFilename(PERR_Location arg) {
  if (PERR_isLocationFile(arg)) {
    return (PERR_OptLayout)ATelementAt((ATermList)ATgetArgument((ATermAppl)arg, 1), 5);
  }
  else 
    return (PERR_OptLayout)ATelementAt((ATermList)ATgetArgument((ATermAppl)arg, 1), 5);
}

/**
 * Get the ws-after-a PERR_OptLayout of a PERR_Location. Note that the precondition is that this PERR_Location actually has a ws-after-a
 * \param[in] arg input PERR_Location
 * \return the ws-after-a of #arg, if it exist or an undefined value if it does not
 */
PERR_OptLayout PERR_getLocationWsAfterA(PERR_Location arg) {
  
    return (PERR_OptLayout)ATelementAt((ATermList)ATgetArgument((ATermAppl)arg, 1), 1);
}

/**
 * Get the Area PERR_Area of a PERR_Location. Note that the precondition is that this PERR_Location actually has a Area
 * \param[in] arg input PERR_Location
 * \return the Area of #arg, if it exist or an undefined value if it does not
 */
PERR_Area PERR_getLocationArea(PERR_Location arg) {
  if (PERR_isLocationArea(arg)) {
    return (PERR_Area)ATelementAt((ATermList)ATgetArgument((ATermAppl)arg, 1), 4);
  }
  else 
    return (PERR_Area)ATelementAt((ATermList)ATgetArgument((ATermAppl)arg, 1), 8);
}

/**
 * Get the ws-after-Area PERR_OptLayout of a PERR_Location. Note that the precondition is that this PERR_Location actually has a ws-after-Area
 * \param[in] arg input PERR_Location
 * \return the ws-after-Area of #arg, if it exist or an undefined value if it does not
 */
PERR_OptLayout PERR_getLocationWsAfterArea(PERR_Location arg) {
  if (PERR_isLocationArea(arg)) {
    return (PERR_OptLayout)ATelementAt((ATermList)ATgetArgument((ATermAppl)arg, 1), 5);
  }
  else 
    return (PERR_OptLayout)ATelementAt((ATermList)ATgetArgument((ATermAppl)arg, 1), 9);
}

/**
 * Get the ws-after-area-in-file PERR_OptLayout of a PERR_Location. Note that the precondition is that this PERR_Location actually has a ws-after-area-in-file
 * \param[in] arg input PERR_Location
 * \return the ws-after-area-in-file of #arg, if it exist or an undefined value if it does not
 */
PERR_OptLayout PERR_getLocationWsAfterAreaInFile(PERR_Location arg) {
  
    return (PERR_OptLayout)ATelementAt((ATermList)ATgetArgument((ATermAppl)arg, 1), 1);
}

/**
 * Get the ws-after-, PERR_OptLayout of a PERR_Location. Note that the precondition is that this PERR_Location actually has a ws-after-,
 * \param[in] arg input PERR_Location
 * \return the ws-after-, of #arg, if it exist or an undefined value if it does not
 */
PERR_OptLayout PERR_getLocationWsAfterComma(PERR_Location arg) {
  
    return (PERR_OptLayout)ATelementAt((ATermList)ATgetArgument((ATermAppl)arg, 1), 7);
}

/**
 * Set the ws-after-file of a PERR_Location. The precondition being that this PERR_Location actually has a ws-after-file
 * \param[in] arg input PERR_Location
 * \param[in] wsAfterFile new PERR_OptLayout to set in #arg
 * \return A new PERR_Location with wsAfterFile at the right place, or a core dump if #arg did not have a wsAfterFile
 */
PERR_Location PERR_setLocationWsAfterFile(PERR_Location arg, PERR_OptLayout wsAfterFile) {
  if (PERR_isLocationFile(arg)) {
    return (PERR_Location)ATsetArgument((ATermAppl)arg, (ATerm)ATreplace((ATermList)ATgetArgument((ATermAppl)arg, 1), (ATerm)((ATerm) wsAfterFile), 1), 1);
  }

  ATabort("Location has no WsAfterFile: %t\n", arg);
  return (PERR_Location)NULL;
}

/**
 * Set the ws-after-( of a PERR_Location. The precondition being that this PERR_Location actually has a ws-after-(
 * \param[in] arg input PERR_Location
 * \param[in] wsAfterParenOpen new PERR_OptLayout to set in #arg
 * \return A new PERR_Location with wsAfterParenOpen at the right place, or a core dump if #arg did not have a wsAfterParenOpen
 */
PERR_Location PERR_setLocationWsAfterParenOpen(PERR_Location arg, PERR_OptLayout wsAfterParenOpen) {
  if (PERR_isLocationFile(arg)) {
    return (PERR_Location)ATsetArgument((ATermAppl)arg, (ATerm)ATreplace((ATermList)ATgetArgument((ATermAppl)arg, 1), (ATerm)((ATerm) wsAfterParenOpen), 3), 1);
  }
  else if (PERR_isLocationArea(arg)) {
    return (PERR_Location)ATsetArgument((ATermAppl)arg, (ATerm)ATreplace((ATermList)ATgetArgument((ATermAppl)arg, 1), (ATerm)((ATerm) wsAfterParenOpen), 3), 1);
  }
  else if (PERR_isLocationAreaInFile(arg)) {
    return (PERR_Location)ATsetArgument((ATermAppl)arg, (ATerm)ATreplace((ATermList)ATgetArgument((ATermAppl)arg, 1), (ATerm)((ATerm) wsAfterParenOpen), 3), 1);
  }

  ATabort("Location has no WsAfterParenOpen: %t\n", arg);
  return (PERR_Location)NULL;
}

/**
 * Set the filename of a PERR_Location. The precondition being that this PERR_Location actually has a filename
 * \param[in] arg input PERR_Location
 * \param[in] filename new PERR_StrCon to set in #arg
 * \return A new PERR_Location with filename at the right place, or a core dump if #arg did not have a filename
 */
PERR_Location PERR_setLocationFilename(PERR_Location arg, PERR_StrCon filename) {
  if (PERR_isLocationFile(arg)) {
    return (PERR_Location)ATsetArgument((ATermAppl)arg, (ATerm)ATreplace((ATermList)ATgetArgument((ATermAppl)arg, 1), (ATerm)((ATerm) filename), 4), 1);
  }
  else if (PERR_isLocationAreaInFile(arg)) {
    return (PERR_Location)ATsetArgument((ATermAppl)arg, (ATerm)ATreplace((ATermList)ATgetArgument((ATermAppl)arg, 1), (ATerm)((ATerm) filename), 4), 1);
  }

  ATabort("Location has no Filename: %t\n", arg);
  return (PERR_Location)NULL;
}

/**
 * Set the ws-after-filename of a PERR_Location. The precondition being that this PERR_Location actually has a ws-after-filename
 * \param[in] arg input PERR_Location
 * \param[in] wsAfterFilename new PERR_OptLayout to set in #arg
 * \return A new PERR_Location with wsAfterFilename at the right place, or a core dump if #arg did not have a wsAfterFilename
 */
PERR_Location PERR_setLocationWsAfterFilename(PERR_Location arg, PERR_OptLayout wsAfterFilename) {
  if (PERR_isLocationFile(arg)) {
    return (PERR_Location)ATsetArgument((ATermAppl)arg, (ATerm)ATreplace((ATermList)ATgetArgument((ATermAppl)arg, 1), (ATerm)((ATerm) wsAfterFilename), 5), 1);
  }
  else if (PERR_isLocationAreaInFile(arg)) {
    return (PERR_Location)ATsetArgument((ATermAppl)arg, (ATerm)ATreplace((ATermList)ATgetArgument((ATermAppl)arg, 1), (ATerm)((ATerm) wsAfterFilename), 5), 1);
  }

  ATabort("Location has no WsAfterFilename: %t\n", arg);
  return (PERR_Location)NULL;
}

/**
 * Set the ws-after-a of a PERR_Location. The precondition being that this PERR_Location actually has a ws-after-a
 * \param[in] arg input PERR_Location
 * \param[in] wsAfterA new PERR_OptLayout to set in #arg
 * \return A new PERR_Location with wsAfterA at the right place, or a core dump if #arg did not have a wsAfterA
 */
PERR_Location PERR_setLocationWsAfterA(PERR_Location arg, PERR_OptLayout wsAfterA) {
  if (PERR_isLocationArea(arg)) {
    return (PERR_Location)ATsetArgument((ATermAppl)arg, (ATerm)ATreplace((ATermList)ATgetArgument((ATermAppl)arg, 1), (ATerm)((ATerm) wsAfterA), 1), 1);
  }

  ATabort("Location has no WsAfterA: %t\n", arg);
  return (PERR_Location)NULL;
}

/**
 * Set the Area of a PERR_Location. The precondition being that this PERR_Location actually has a Area
 * \param[in] arg input PERR_Location
 * \param[in] Area new PERR_Area to set in #arg
 * \return A new PERR_Location with Area at the right place, or a core dump if #arg did not have a Area
 */
PERR_Location PERR_setLocationArea(PERR_Location arg, PERR_Area Area) {
  if (PERR_isLocationArea(arg)) {
    return (PERR_Location)ATsetArgument((ATermAppl)arg, (ATerm)ATreplace((ATermList)ATgetArgument((ATermAppl)arg, 1), (ATerm)((ATerm) Area), 4), 1);
  }
  else if (PERR_isLocationAreaInFile(arg)) {
    return (PERR_Location)ATsetArgument((ATermAppl)arg, (ATerm)ATreplace((ATermList)ATgetArgument((ATermAppl)arg, 1), (ATerm)((ATerm) Area), 8), 1);
  }

  ATabort("Location has no Area: %t\n", arg);
  return (PERR_Location)NULL;
}

/**
 * Set the ws-after-Area of a PERR_Location. The precondition being that this PERR_Location actually has a ws-after-Area
 * \param[in] arg input PERR_Location
 * \param[in] wsAfterArea new PERR_OptLayout to set in #arg
 * \return A new PERR_Location with wsAfterArea at the right place, or a core dump if #arg did not have a wsAfterArea
 */
PERR_Location PERR_setLocationWsAfterArea(PERR_Location arg, PERR_OptLayout wsAfterArea) {
  if (PERR_isLocationArea(arg)) {
    return (PERR_Location)ATsetArgument((ATermAppl)arg, (ATerm)ATreplace((ATermList)ATgetArgument((ATermAppl)arg, 1), (ATerm)((ATerm) wsAfterArea), 5), 1);
  }
  else if (PERR_isLocationAreaInFile(arg)) {
    return (PERR_Location)ATsetArgument((ATermAppl)arg, (ATerm)ATreplace((ATermList)ATgetArgument((ATermAppl)arg, 1), (ATerm)((ATerm) wsAfterArea), 9), 1);
  }

  ATabort("Location has no WsAfterArea: %t\n", arg);
  return (PERR_Location)NULL;
}

/**
 * Set the ws-after-area-in-file of a PERR_Location. The precondition being that this PERR_Location actually has a ws-after-area-in-file
 * \param[in] arg input PERR_Location
 * \param[in] wsAfterAreaInFile new PERR_OptLayout to set in #arg
 * \return A new PERR_Location with wsAfterAreaInFile at the right place, or a core dump if #arg did not have a wsAfterAreaInFile
 */
PERR_Location PERR_setLocationWsAfterAreaInFile(PERR_Location arg, PERR_OptLayout wsAfterAreaInFile) {
  if (PERR_isLocationAreaInFile(arg)) {
    return (PERR_Location)ATsetArgument((ATermAppl)arg, (ATerm)ATreplace((ATermList)ATgetArgument((ATermAppl)arg, 1), (ATerm)((ATerm) wsAfterAreaInFile), 1), 1);
  }

  ATabort("Location has no WsAfterAreaInFile: %t\n", arg);
  return (PERR_Location)NULL;
}

/**
 * Set the ws-after-, of a PERR_Location. The precondition being that this PERR_Location actually has a ws-after-,
 * \param[in] arg input PERR_Location
 * \param[in] wsAfterComma new PERR_OptLayout to set in #arg
 * \return A new PERR_Location with wsAfterComma at the right place, or a core dump if #arg did not have a wsAfterComma
 */
PERR_Location PERR_setLocationWsAfterComma(PERR_Location arg, PERR_OptLayout wsAfterComma) {
  if (PERR_isLocationAreaInFile(arg)) {
    return (PERR_Location)ATsetArgument((ATermAppl)arg, (ATerm)ATreplace((ATermList)ATgetArgument((ATermAppl)arg, 1), (ATerm)((ATerm) wsAfterComma), 7), 1);
  }

  ATabort("Location has no WsAfterComma: %t\n", arg);
  return (PERR_Location)NULL;
}

/**
 * Assert whether a PERR_Area is any of the valid alternatives, or not. This analysis does not go any deeper than the top level
 * \param[in] arg input PERR_Area
 * \return ATtrue if #arg corresponds to the expected signature, or ATfalse otherwise
 */
ATbool PERR_isValidArea(PERR_Area arg) {
  if (PERR_isAreaArea(arg)) {
    return ATtrue;
  }
  return ATfalse;
}

/**
 * Assert whether a PERR_Area is a area by checking against the following ATerm pattern: appl(prod([lit("area"),cf(opt(layout)),lit("("),cf(opt(layout)),cf(sort("NatCon")),cf(opt(layout)),lit(","),cf(opt(layout)),cf(sort("NatCon")),cf(opt(layout)),lit(","),cf(opt(layout)),cf(sort("NatCon")),cf(opt(layout)),lit(","),cf(opt(layout)),cf(sort("NatCon")),cf(opt(layout)),lit(","),cf(opt(layout)),cf(sort("NatCon")),cf(opt(layout)),lit(","),cf(opt(layout)),cf(sort("NatCon")),cf(opt(layout)),lit(")")],cf(sort("Area")),attrs([term(cons("area"))])),[appl(prod([char-class([97]),char-class([114]),char-class([101]),char-class([97])],lit("area"),no-attrs),[97,114,101,97]),<"ws-after-area"("OptLayout")>,appl(prod([char-class([40])],lit("("),no-attrs),[40]),<"ws-after-("("OptLayout")>,<"begin-line"("NatCon")>,<"ws-after-begin-line"("OptLayout")>,appl(prod([char-class([44])],lit(","),no-attrs),[44]),<"ws-after-,"("OptLayout")>,<"begin-column"("NatCon")>,<"ws-after-begin-column"("OptLayout")>,appl(prod([char-class([44])],lit(","),no-attrs),[44]),<"ws-after-,-1"("OptLayout")>,<"end-line"("NatCon")>,<"ws-after-end-line"("OptLayout")>,appl(prod([char-class([44])],lit(","),no-attrs),[44]),<"ws-after-,-2"("OptLayout")>,<"end-column"("NatCon")>,<"ws-after-end-column"("OptLayout")>,appl(prod([char-class([44])],lit(","),no-attrs),[44]),<"ws-after-,-3"("OptLayout")>,<"offset"("NatCon")>,<"ws-after-offset"("OptLayout")>,appl(prod([char-class([44])],lit(","),no-attrs),[44]),<"ws-after-,-4"("OptLayout")>,<"length"("NatCon")>,<"ws-after-length"("OptLayout")>,appl(prod([char-class([41])],lit(")"),no-attrs),[41])]). Always returns ATtrue
 * \param[in] arg input PERR_Area
 * \return ATtrue if #arg corresponds to the signature of a area, or ATfalse otherwise
 */
inline ATbool PERR_isAreaArea(PERR_Area arg){
  /* checking for: appl */
  if (ATgetType((ATerm)arg) == AT_APPL && ATgetAFun((ATermAppl)arg) == PERR_afun0) {
    ATerm arg_arg0 = ATgetArgument(arg, 0);
    /* checking for: prod */
    if (ATgetType((ATerm)arg_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0) == PERR_afun1) {
      ATerm arg_arg0_arg0 = ATgetArgument(arg_arg0, 0);
      if (ATgetType((ATerm)arg_arg0_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0) == ATfalse) {
        ATerm arg_arg0_arg0_head;
        ATermList arg_arg0_arg0_list = (ATermList)arg_arg0_arg0;
        arg_arg0_arg0_head = ATgetFirst(arg_arg0_arg0_list);
        arg_arg0_arg0_list = ATgetNext(arg_arg0_arg0_list);
        /* checking for: lit */
        if (ATgetType((ATerm)arg_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head) == PERR_afun9) {
          ATerm arg_arg0_arg0_head_arg0 = ATgetArgument(arg_arg0_arg0_head, 0);
          /* checking for: area */
          if (ATgetType((ATerm)arg_arg0_arg0_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0) == PERR_afun55) {
            if (ATgetType((ATerm)arg_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0_list) == ATfalse) {
              arg_arg0_arg0_head = ATgetFirst(arg_arg0_arg0_list);
              arg_arg0_arg0_list = ATgetNext(arg_arg0_arg0_list);
              /* checking for: cf */
              if (ATgetType((ATerm)arg_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head) == PERR_afun2) {
                ATerm arg_arg0_arg0_head_arg0 = ATgetArgument(arg_arg0_arg0_head, 0);
                /* checking for: opt */
                if (ATgetType((ATerm)arg_arg0_arg0_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0) == PERR_afun3) {
                  ATerm arg_arg0_arg0_head_arg0_arg0 = ATgetArgument(arg_arg0_arg0_head_arg0, 0);
                  /* checking for: layout */
                  if (ATgetType((ATerm)arg_arg0_arg0_head_arg0_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0_arg0) == PERR_afun4) {
                    if (ATgetType((ATerm)arg_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0_list) == ATfalse) {
                      arg_arg0_arg0_head = ATgetFirst(arg_arg0_arg0_list);
                      arg_arg0_arg0_list = ATgetNext(arg_arg0_arg0_list);
                      /* checking for: lit */
                      if (ATgetType((ATerm)arg_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head) == PERR_afun9) {
                        ATerm arg_arg0_arg0_head_arg0 = ATgetArgument(arg_arg0_arg0_head, 0);
                        /* checking for: ( */
                        if (ATgetType((ATerm)arg_arg0_arg0_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0) == PERR_afun39) {
                          if (ATgetType((ATerm)arg_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0_list) == ATfalse) {
                            arg_arg0_arg0_head = ATgetFirst(arg_arg0_arg0_list);
                            arg_arg0_arg0_list = ATgetNext(arg_arg0_arg0_list);
                            /* checking for: cf */
                            if (ATgetType((ATerm)arg_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head) == PERR_afun2) {
                              ATerm arg_arg0_arg0_head_arg0 = ATgetArgument(arg_arg0_arg0_head, 0);
                              /* checking for: opt */
                              if (ATgetType((ATerm)arg_arg0_arg0_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0) == PERR_afun3) {
                                ATerm arg_arg0_arg0_head_arg0_arg0 = ATgetArgument(arg_arg0_arg0_head_arg0, 0);
                                /* checking for: layout */
                                if (ATgetType((ATerm)arg_arg0_arg0_head_arg0_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0_arg0) == PERR_afun4) {
                                  if (ATgetType((ATerm)arg_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0_list) == ATfalse) {
                                    arg_arg0_arg0_head = ATgetFirst(arg_arg0_arg0_list);
                                    arg_arg0_arg0_list = ATgetNext(arg_arg0_arg0_list);
                                    /* checking for: cf */
                                    if (ATgetType((ATerm)arg_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head) == PERR_afun2) {
                                      ATerm arg_arg0_arg0_head_arg0 = ATgetArgument(arg_arg0_arg0_head, 0);
                                      /* checking for: sort */
                                      if (ATgetType((ATerm)arg_arg0_arg0_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0) == PERR_afun11) {
                                        ATerm arg_arg0_arg0_head_arg0_arg0 = ATgetArgument(arg_arg0_arg0_head_arg0, 0);
                                        /* checking for: NatCon */
                                        if (ATgetType((ATerm)arg_arg0_arg0_head_arg0_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0_arg0) == PERR_afun31) {
                                          if (ATgetType((ATerm)arg_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0_list) == ATfalse) {
                                            arg_arg0_arg0_head = ATgetFirst(arg_arg0_arg0_list);
                                            arg_arg0_arg0_list = ATgetNext(arg_arg0_arg0_list);
                                            /* checking for: cf */
                                            if (ATgetType((ATerm)arg_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head) == PERR_afun2) {
                                              ATerm arg_arg0_arg0_head_arg0 = ATgetArgument(arg_arg0_arg0_head, 0);
                                              /* checking for: opt */
                                              if (ATgetType((ATerm)arg_arg0_arg0_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0) == PERR_afun3) {
                                                ATerm arg_arg0_arg0_head_arg0_arg0 = ATgetArgument(arg_arg0_arg0_head_arg0, 0);
                                                /* checking for: layout */
                                                if (ATgetType((ATerm)arg_arg0_arg0_head_arg0_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0_arg0) == PERR_afun4) {
                                                  if (ATgetType((ATerm)arg_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0_list) == ATfalse) {
                                                    arg_arg0_arg0_head = ATgetFirst(arg_arg0_arg0_list);
                                                    arg_arg0_arg0_list = ATgetNext(arg_arg0_arg0_list);
                                                    /* checking for: lit */
                                                    if (ATgetType((ATerm)arg_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head) == PERR_afun9) {
                                                      ATerm arg_arg0_arg0_head_arg0 = ATgetArgument(arg_arg0_arg0_head, 0);
                                                      /* checking for: , */
                                                      if (ATgetType((ATerm)arg_arg0_arg0_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0) == PERR_afun43) {
                                                        if (ATgetType((ATerm)arg_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0_list) == ATfalse) {
                                                          arg_arg0_arg0_head = ATgetFirst(arg_arg0_arg0_list);
                                                          arg_arg0_arg0_list = ATgetNext(arg_arg0_arg0_list);
                                                          /* checking for: cf */
                                                          if (ATgetType((ATerm)arg_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head) == PERR_afun2) {
                                                            ATerm arg_arg0_arg0_head_arg0 = ATgetArgument(arg_arg0_arg0_head, 0);
                                                            /* checking for: opt */
                                                            if (ATgetType((ATerm)arg_arg0_arg0_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0) == PERR_afun3) {
                                                              ATerm arg_arg0_arg0_head_arg0_arg0 = ATgetArgument(arg_arg0_arg0_head_arg0, 0);
                                                              /* checking for: layout */
                                                              if (ATgetType((ATerm)arg_arg0_arg0_head_arg0_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0_arg0) == PERR_afun4) {
                                                                if (ATgetType((ATerm)arg_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0_list) == ATfalse) {
                                                                  arg_arg0_arg0_head = ATgetFirst(arg_arg0_arg0_list);
                                                                  arg_arg0_arg0_list = ATgetNext(arg_arg0_arg0_list);
                                                                  /* checking for: cf */
                                                                  if (ATgetType((ATerm)arg_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head) == PERR_afun2) {
                                                                    ATerm arg_arg0_arg0_head_arg0 = ATgetArgument(arg_arg0_arg0_head, 0);
                                                                    /* checking for: sort */
                                                                    if (ATgetType((ATerm)arg_arg0_arg0_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0) == PERR_afun11) {
                                                                      ATerm arg_arg0_arg0_head_arg0_arg0 = ATgetArgument(arg_arg0_arg0_head_arg0, 0);
                                                                      /* checking for: NatCon */
                                                                      if (ATgetType((ATerm)arg_arg0_arg0_head_arg0_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0_arg0) == PERR_afun31) {
                                                                        if (ATgetType((ATerm)arg_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0_list) == ATfalse) {
                                                                          arg_arg0_arg0_head = ATgetFirst(arg_arg0_arg0_list);
                                                                          arg_arg0_arg0_list = ATgetNext(arg_arg0_arg0_list);
                                                                          /* checking for: cf */
                                                                          if (ATgetType((ATerm)arg_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head) == PERR_afun2) {
                                                                            ATerm arg_arg0_arg0_head_arg0 = ATgetArgument(arg_arg0_arg0_head, 0);
                                                                            /* checking for: opt */
                                                                            if (ATgetType((ATerm)arg_arg0_arg0_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0) == PERR_afun3) {
                                                                              ATerm arg_arg0_arg0_head_arg0_arg0 = ATgetArgument(arg_arg0_arg0_head_arg0, 0);
                                                                              /* checking for: layout */
                                                                              if (ATgetType((ATerm)arg_arg0_arg0_head_arg0_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0_arg0) == PERR_afun4) {
                                                                                if (ATgetType((ATerm)arg_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0_list) == ATfalse) {
                                                                                  arg_arg0_arg0_head = ATgetFirst(arg_arg0_arg0_list);
                                                                                  arg_arg0_arg0_list = ATgetNext(arg_arg0_arg0_list);
                                                                                  /* checking for: lit */
                                                                                  if (ATgetType((ATerm)arg_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head) == PERR_afun9) {
                                                                                    ATerm arg_arg0_arg0_head_arg0 = ATgetArgument(arg_arg0_arg0_head, 0);
                                                                                    /* checking for: , */
                                                                                    if (ATgetType((ATerm)arg_arg0_arg0_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0) == PERR_afun43) {
                                                                                      if (ATgetType((ATerm)arg_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0_list) == ATfalse) {
                                                                                        arg_arg0_arg0_head = ATgetFirst(arg_arg0_arg0_list);
                                                                                        arg_arg0_arg0_list = ATgetNext(arg_arg0_arg0_list);
                                                                                        /* checking for: cf */
                                                                                        if (ATgetType((ATerm)arg_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head) == PERR_afun2) {
                                                                                          ATerm arg_arg0_arg0_head_arg0 = ATgetArgument(arg_arg0_arg0_head, 0);
                                                                                          /* checking for: opt */
                                                                                          if (ATgetType((ATerm)arg_arg0_arg0_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0) == PERR_afun3) {
                                                                                            ATerm arg_arg0_arg0_head_arg0_arg0 = ATgetArgument(arg_arg0_arg0_head_arg0, 0);
                                                                                            /* checking for: layout */
                                                                                            if (ATgetType((ATerm)arg_arg0_arg0_head_arg0_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0_arg0) == PERR_afun4) {
                                                                                              if (ATgetType((ATerm)arg_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0_list) == ATfalse) {
                                                                                                arg_arg0_arg0_head = ATgetFirst(arg_arg0_arg0_list);
                                                                                                arg_arg0_arg0_list = ATgetNext(arg_arg0_arg0_list);
                                                                                                /* checking for: cf */
                                                                                                if (ATgetType((ATerm)arg_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head) == PERR_afun2) {
                                                                                                  ATerm arg_arg0_arg0_head_arg0 = ATgetArgument(arg_arg0_arg0_head, 0);
                                                                                                  /* checking for: sort */
                                                                                                  if (ATgetType((ATerm)arg_arg0_arg0_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0) == PERR_afun11) {
                                                                                                    ATerm arg_arg0_arg0_head_arg0_arg0 = ATgetArgument(arg_arg0_arg0_head_arg0, 0);
                                                                                                    /* checking for: NatCon */
                                                                                                    if (ATgetType((ATerm)arg_arg0_arg0_head_arg0_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0_arg0) == PERR_afun31) {
                                                                                                      if (ATgetType((ATerm)arg_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0_list) == ATfalse) {
                                                                                                        arg_arg0_arg0_head = ATgetFirst(arg_arg0_arg0_list);
                                                                                                        arg_arg0_arg0_list = ATgetNext(arg_arg0_arg0_list);
                                                                                                        /* checking for: cf */
                                                                                                        if (ATgetType((ATerm)arg_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head) == PERR_afun2) {
                                                                                                          ATerm arg_arg0_arg0_head_arg0 = ATgetArgument(arg_arg0_arg0_head, 0);
                                                                                                          /* checking for: opt */
                                                                                                          if (ATgetType((ATerm)arg_arg0_arg0_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0) == PERR_afun3) {
                                                                                                            ATerm arg_arg0_arg0_head_arg0_arg0 = ATgetArgument(arg_arg0_arg0_head_arg0, 0);
                                                                                                            /* checking for: layout */
                                                                                                            if (ATgetType((ATerm)arg_arg0_arg0_head_arg0_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0_arg0) == PERR_afun4) {
                                                                                                              if (ATgetType((ATerm)arg_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0_list) == ATfalse) {
                                                                                                                arg_arg0_arg0_head = ATgetFirst(arg_arg0_arg0_list);
                                                                                                                arg_arg0_arg0_list = ATgetNext(arg_arg0_arg0_list);
                                                                                                                /* checking for: lit */
                                                                                                                if (ATgetType((ATerm)arg_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head) == PERR_afun9) {
                                                                                                                  ATerm arg_arg0_arg0_head_arg0 = ATgetArgument(arg_arg0_arg0_head, 0);
                                                                                                                  /* checking for: , */
                                                                                                                  if (ATgetType((ATerm)arg_arg0_arg0_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0) == PERR_afun43) {
                                                                                                                    if (ATgetType((ATerm)arg_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0_list) == ATfalse) {
                                                                                                                      arg_arg0_arg0_head = ATgetFirst(arg_arg0_arg0_list);
                                                                                                                      arg_arg0_arg0_list = ATgetNext(arg_arg0_arg0_list);
                                                                                                                      /* checking for: cf */
                                                                                                                      if (ATgetType((ATerm)arg_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head) == PERR_afun2) {
                                                                                                                        ATerm arg_arg0_arg0_head_arg0 = ATgetArgument(arg_arg0_arg0_head, 0);
                                                                                                                        /* checking for: opt */
                                                                                                                        if (ATgetType((ATerm)arg_arg0_arg0_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0) == PERR_afun3) {
                                                                                                                          ATerm arg_arg0_arg0_head_arg0_arg0 = ATgetArgument(arg_arg0_arg0_head_arg0, 0);
                                                                                                                          /* checking for: layout */
                                                                                                                          if (ATgetType((ATerm)arg_arg0_arg0_head_arg0_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0_arg0) == PERR_afun4) {
                                                                                                                            if (ATgetType((ATerm)arg_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0_list) == ATfalse) {
                                                                                                                              arg_arg0_arg0_head = ATgetFirst(arg_arg0_arg0_list);
                                                                                                                              arg_arg0_arg0_list = ATgetNext(arg_arg0_arg0_list);
                                                                                                                              /* checking for: cf */
                                                                                                                              if (ATgetType((ATerm)arg_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head) == PERR_afun2) {
                                                                                                                                ATerm arg_arg0_arg0_head_arg0 = ATgetArgument(arg_arg0_arg0_head, 0);
                                                                                                                                /* checking for: sort */
                                                                                                                                if (ATgetType((ATerm)arg_arg0_arg0_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0) == PERR_afun11) {
                                                                                                                                  ATerm arg_arg0_arg0_head_arg0_arg0 = ATgetArgument(arg_arg0_arg0_head_arg0, 0);
                                                                                                                                  /* checking for: NatCon */
                                                                                                                                  if (ATgetType((ATerm)arg_arg0_arg0_head_arg0_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0_arg0) == PERR_afun31) {
                                                                                                                                    if (ATgetType((ATerm)arg_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0_list) == ATfalse) {
                                                                                                                                      arg_arg0_arg0_head = ATgetFirst(arg_arg0_arg0_list);
                                                                                                                                      arg_arg0_arg0_list = ATgetNext(arg_arg0_arg0_list);
                                                                                                                                      /* checking for: cf */
                                                                                                                                      if (ATgetType((ATerm)arg_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head) == PERR_afun2) {
                                                                                                                                        ATerm arg_arg0_arg0_head_arg0 = ATgetArgument(arg_arg0_arg0_head, 0);
                                                                                                                                        /* checking for: opt */
                                                                                                                                        if (ATgetType((ATerm)arg_arg0_arg0_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0) == PERR_afun3) {
                                                                                                                                          ATerm arg_arg0_arg0_head_arg0_arg0 = ATgetArgument(arg_arg0_arg0_head_arg0, 0);
                                                                                                                                          /* checking for: layout */
                                                                                                                                          if (ATgetType((ATerm)arg_arg0_arg0_head_arg0_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0_arg0) == PERR_afun4) {
                                                                                                                                            if (ATgetType((ATerm)arg_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0_list) == ATfalse) {
                                                                                                                                              arg_arg0_arg0_head = ATgetFirst(arg_arg0_arg0_list);
                                                                                                                                              arg_arg0_arg0_list = ATgetNext(arg_arg0_arg0_list);
                                                                                                                                              /* checking for: lit */
                                                                                                                                              if (ATgetType((ATerm)arg_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head) == PERR_afun9) {
                                                                                                                                                ATerm arg_arg0_arg0_head_arg0 = ATgetArgument(arg_arg0_arg0_head, 0);
                                                                                                                                                /* checking for: , */
                                                                                                                                                if (ATgetType((ATerm)arg_arg0_arg0_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0) == PERR_afun43) {
                                                                                                                                                  if (ATgetType((ATerm)arg_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0_list) == ATfalse) {
                                                                                                                                                    arg_arg0_arg0_head = ATgetFirst(arg_arg0_arg0_list);
                                                                                                                                                    arg_arg0_arg0_list = ATgetNext(arg_arg0_arg0_list);
                                                                                                                                                    /* checking for: cf */
                                                                                                                                                    if (ATgetType((ATerm)arg_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head) == PERR_afun2) {
                                                                                                                                                      ATerm arg_arg0_arg0_head_arg0 = ATgetArgument(arg_arg0_arg0_head, 0);
                                                                                                                                                      /* checking for: opt */
                                                                                                                                                      if (ATgetType((ATerm)arg_arg0_arg0_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0) == PERR_afun3) {
                                                                                                                                                        ATerm arg_arg0_arg0_head_arg0_arg0 = ATgetArgument(arg_arg0_arg0_head_arg0, 0);
                                                                                                                                                        /* checking for: layout */
                                                                                                                                                        if (ATgetType((ATerm)arg_arg0_arg0_head_arg0_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0_arg0) == PERR_afun4) {
                                                                                                                                                          if (ATgetType((ATerm)arg_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0_list) == ATfalse) {
                                                                                                                                                            arg_arg0_arg0_head = ATgetFirst(arg_arg0_arg0_list);
                                                                                                                                                            arg_arg0_arg0_list = ATgetNext(arg_arg0_arg0_list);
                                                                                                                                                            /* checking for: cf */
                                                                                                                                                            if (ATgetType((ATerm)arg_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head) == PERR_afun2) {
                                                                                                                                                              ATerm arg_arg0_arg0_head_arg0 = ATgetArgument(arg_arg0_arg0_head, 0);
                                                                                                                                                              /* checking for: sort */
                                                                                                                                                              if (ATgetType((ATerm)arg_arg0_arg0_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0) == PERR_afun11) {
                                                                                                                                                                ATerm arg_arg0_arg0_head_arg0_arg0 = ATgetArgument(arg_arg0_arg0_head_arg0, 0);
                                                                                                                                                                /* checking for: NatCon */
                                                                                                                                                                if (ATgetType((ATerm)arg_arg0_arg0_head_arg0_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0_arg0) == PERR_afun31) {
                                                                                                                                                                  if (ATgetType((ATerm)arg_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0_list) == ATfalse) {
                                                                                                                                                                    arg_arg0_arg0_head = ATgetFirst(arg_arg0_arg0_list);
                                                                                                                                                                    arg_arg0_arg0_list = ATgetNext(arg_arg0_arg0_list);
                                                                                                                                                                    /* checking for: cf */
                                                                                                                                                                    if (ATgetType((ATerm)arg_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head) == PERR_afun2) {
                                                                                                                                                                      ATerm arg_arg0_arg0_head_arg0 = ATgetArgument(arg_arg0_arg0_head, 0);
                                                                                                                                                                      /* checking for: opt */
                                                                                                                                                                      if (ATgetType((ATerm)arg_arg0_arg0_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0) == PERR_afun3) {
                                                                                                                                                                        ATerm arg_arg0_arg0_head_arg0_arg0 = ATgetArgument(arg_arg0_arg0_head_arg0, 0);
                                                                                                                                                                        /* checking for: layout */
                                                                                                                                                                        if (ATgetType((ATerm)arg_arg0_arg0_head_arg0_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0_arg0) == PERR_afun4) {
                                                                                                                                                                          if (ATgetType((ATerm)arg_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0_list) == ATfalse) {
                                                                                                                                                                            arg_arg0_arg0_head = ATgetFirst(arg_arg0_arg0_list);
                                                                                                                                                                            arg_arg0_arg0_list = ATgetNext(arg_arg0_arg0_list);
                                                                                                                                                                            /* checking for: lit */
                                                                                                                                                                            if (ATgetType((ATerm)arg_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head) == PERR_afun9) {
                                                                                                                                                                              ATerm arg_arg0_arg0_head_arg0 = ATgetArgument(arg_arg0_arg0_head, 0);
                                                                                                                                                                              /* checking for: , */
                                                                                                                                                                              if (ATgetType((ATerm)arg_arg0_arg0_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0) == PERR_afun43) {
                                                                                                                                                                                if (ATgetType((ATerm)arg_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0_list) == ATfalse) {
                                                                                                                                                                                  arg_arg0_arg0_head = ATgetFirst(arg_arg0_arg0_list);
                                                                                                                                                                                  arg_arg0_arg0_list = ATgetNext(arg_arg0_arg0_list);
                                                                                                                                                                                  /* checking for: cf */
                                                                                                                                                                                  if (ATgetType((ATerm)arg_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head) == PERR_afun2) {
                                                                                                                                                                                    ATerm arg_arg0_arg0_head_arg0 = ATgetArgument(arg_arg0_arg0_head, 0);
                                                                                                                                                                                    /* checking for: opt */
                                                                                                                                                                                    if (ATgetType((ATerm)arg_arg0_arg0_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0) == PERR_afun3) {
                                                                                                                                                                                      ATerm arg_arg0_arg0_head_arg0_arg0 = ATgetArgument(arg_arg0_arg0_head_arg0, 0);
                                                                                                                                                                                      /* checking for: layout */
                                                                                                                                                                                      if (ATgetType((ATerm)arg_arg0_arg0_head_arg0_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0_arg0) == PERR_afun4) {
                                                                                                                                                                                        if (ATgetType((ATerm)arg_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0_list) == ATfalse) {
                                                                                                                                                                                          arg_arg0_arg0_head = ATgetFirst(arg_arg0_arg0_list);
                                                                                                                                                                                          arg_arg0_arg0_list = ATgetNext(arg_arg0_arg0_list);
                                                                                                                                                                                          /* checking for: cf */
                                                                                                                                                                                          if (ATgetType((ATerm)arg_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head) == PERR_afun2) {
                                                                                                                                                                                            ATerm arg_arg0_arg0_head_arg0 = ATgetArgument(arg_arg0_arg0_head, 0);
                                                                                                                                                                                            /* checking for: sort */
                                                                                                                                                                                            if (ATgetType((ATerm)arg_arg0_arg0_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0) == PERR_afun11) {
                                                                                                                                                                                              ATerm arg_arg0_arg0_head_arg0_arg0 = ATgetArgument(arg_arg0_arg0_head_arg0, 0);
                                                                                                                                                                                              /* checking for: NatCon */
                                                                                                                                                                                              if (ATgetType((ATerm)arg_arg0_arg0_head_arg0_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0_arg0) == PERR_afun31) {
                                                                                                                                                                                                if (ATgetType((ATerm)arg_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0_list) == ATfalse) {
                                                                                                                                                                                                  arg_arg0_arg0_head = ATgetFirst(arg_arg0_arg0_list);
                                                                                                                                                                                                  arg_arg0_arg0_list = ATgetNext(arg_arg0_arg0_list);
                                                                                                                                                                                                  /* checking for: cf */
                                                                                                                                                                                                  if (ATgetType((ATerm)arg_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head) == PERR_afun2) {
                                                                                                                                                                                                    ATerm arg_arg0_arg0_head_arg0 = ATgetArgument(arg_arg0_arg0_head, 0);
                                                                                                                                                                                                    /* checking for: opt */
                                                                                                                                                                                                    if (ATgetType((ATerm)arg_arg0_arg0_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0) == PERR_afun3) {
                                                                                                                                                                                                      ATerm arg_arg0_arg0_head_arg0_arg0 = ATgetArgument(arg_arg0_arg0_head_arg0, 0);
                                                                                                                                                                                                      /* checking for: layout */
                                                                                                                                                                                                      if (ATgetType((ATerm)arg_arg0_arg0_head_arg0_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0_arg0) == PERR_afun4) {
                                                                                                                                                                                                        if (ATgetType((ATerm)arg_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0_list) == ATfalse) {
                                                                                                                                                                                                          arg_arg0_arg0_head = ATgetFirst(arg_arg0_arg0_list);
                                                                                                                                                                                                          arg_arg0_arg0_list = ATgetNext(arg_arg0_arg0_list);
                                                                                                                                                                                                          /* checking for: lit */
                                                                                                                                                                                                          if (ATgetType((ATerm)arg_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head) == PERR_afun9) {
                                                                                                                                                                                                            ATerm arg_arg0_arg0_head_arg0 = ATgetArgument(arg_arg0_arg0_head, 0);
                                                                                                                                                                                                            /* checking for: ) */
                                                                                                                                                                                                            if (ATgetType((ATerm)arg_arg0_arg0_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0) == PERR_afun38) {
                                                                                                                                                                                                              if (ATgetType((ATerm)arg_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0_list) == ATtrue) {
                                                                                                                                                                                                                ATerm arg_arg0_arg1 = ATgetArgument(arg_arg0, 1);
                                                                                                                                                                                                                /* checking for: cf */
                                                                                                                                                                                                                if (ATgetType((ATerm)arg_arg0_arg1) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg1) == PERR_afun2) {
                                                                                                                                                                                                                  ATerm arg_arg0_arg1_arg0 = ATgetArgument(arg_arg0_arg1, 0);
                                                                                                                                                                                                                  /* checking for: sort */
                                                                                                                                                                                                                  if (ATgetType((ATerm)arg_arg0_arg1_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg1_arg0) == PERR_afun11) {
                                                                                                                                                                                                                    ATerm arg_arg0_arg1_arg0_arg0 = ATgetArgument(arg_arg0_arg1_arg0, 0);
                                                                                                                                                                                                                    /* checking for: Area */
                                                                                                                                                                                                                    if (ATgetType((ATerm)arg_arg0_arg1_arg0_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg1_arg0_arg0) == PERR_afun54) {
                                                                                                                                                                                                                      ATerm arg_arg0_arg2 = ATgetArgument(arg_arg0, 2);
                                                                                                                                                                                                                      /* checking for: attrs */
                                                                                                                                                                                                                      if (ATgetType((ATerm)arg_arg0_arg2) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg2) == PERR_afun13) {
                                                                                                                                                                                                                        ATerm arg_arg0_arg2_arg0 = ATgetArgument(arg_arg0_arg2, 0);
                                                                                                                                                                                                                        if (ATgetType((ATerm)arg_arg0_arg2_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg2_arg0) == ATfalse) {
                                                                                                                                                                                                                          ATerm arg_arg0_arg2_arg0_head;
                                                                                                                                                                                                                          ATermList arg_arg0_arg2_arg0_list = (ATermList)arg_arg0_arg2_arg0;
                                                                                                                                                                                                                          arg_arg0_arg2_arg0_head = ATgetFirst(arg_arg0_arg2_arg0_list);
                                                                                                                                                                                                                          arg_arg0_arg2_arg0_list = ATgetNext(arg_arg0_arg2_arg0_list);
                                                                                                                                                                                                                          /* checking for: term */
                                                                                                                                                                                                                          if (ATgetType((ATerm)arg_arg0_arg2_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg2_arg0_head) == PERR_afun14) {
                                                                                                                                                                                                                            ATerm arg_arg0_arg2_arg0_head_arg0 = ATgetArgument(arg_arg0_arg2_arg0_head, 0);
                                                                                                                                                                                                                            /* checking for: cons */
                                                                                                                                                                                                                            if (ATgetType((ATerm)arg_arg0_arg2_arg0_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg2_arg0_head_arg0) == PERR_afun15) {
                                                                                                                                                                                                                              ATerm arg_arg0_arg2_arg0_head_arg0_arg0 = ATgetArgument(arg_arg0_arg2_arg0_head_arg0, 0);
                                                                                                                                                                                                                              /* checking for: area */
                                                                                                                                                                                                                              if (ATgetType((ATerm)arg_arg0_arg2_arg0_head_arg0_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg2_arg0_head_arg0_arg0) == PERR_afun55) {
                                                                                                                                                                                                                                if (ATgetType((ATerm)arg_arg0_arg2_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg2_arg0_list) == ATtrue) {
                                                                                                                                                                                                                                  ATerm arg_arg1 = ATgetArgument(arg, 1);
                                                                                                                                                                                                                                  if (ATgetType((ATerm)arg_arg1) == AT_LIST && ATisEmpty((ATermList)arg_arg1) == ATfalse) {
                                                                                                                                                                                                                                    ATerm arg_arg1_head;
                                                                                                                                                                                                                                    ATermList arg_arg1_list = (ATermList)arg_arg1;
                                                                                                                                                                                                                                    arg_arg1_head = ATgetFirst(arg_arg1_list);
                                                                                                                                                                                                                                    arg_arg1_list = ATgetNext(arg_arg1_list);
                                                                                                                                                                                                                                    /* checking for: appl */
                                                                                                                                                                                                                                    if (ATgetType((ATerm)arg_arg1_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head) == PERR_afun0) {
                                                                                                                                                                                                                                      ATerm arg_arg1_head_arg0 = ATgetArgument(arg_arg1_head, 0);
                                                                                                                                                                                                                                      /* checking for: prod */
                                                                                                                                                                                                                                      if (ATgetType((ATerm)arg_arg1_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0) == PERR_afun1) {
                                                                                                                                                                                                                                        ATerm arg_arg1_head_arg0_arg0 = ATgetArgument(arg_arg1_head_arg0, 0);
                                                                                                                                                                                                                                        if (ATgetType((ATerm)arg_arg1_head_arg0_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0) == ATfalse) {
                                                                                                                                                                                                                                          ATerm arg_arg1_head_arg0_arg0_head;
                                                                                                                                                                                                                                          ATermList arg_arg1_head_arg0_arg0_list = (ATermList)arg_arg1_head_arg0_arg0;
                                                                                                                                                                                                                                          arg_arg1_head_arg0_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_list);
                                                                                                                                                                                                                                          arg_arg1_head_arg0_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_list);
                                                                                                                                                                                                                                          /* checking for: char-class */
                                                                                                                                                                                                                                          if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg0_head) == PERR_afun17) {
                                                                                                                                                                                                                                            ATerm arg_arg1_head_arg0_arg0_head_arg0 = ATgetArgument(arg_arg1_head_arg0_arg0_head, 0);
                                                                                                                                                                                                                                            if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0) == ATfalse) {
                                                                                                                                                                                                                                              ATerm arg_arg1_head_arg0_arg0_head_arg0_head;
                                                                                                                                                                                                                                              ATermList arg_arg1_head_arg0_arg0_head_arg0_list = (ATermList)arg_arg1_head_arg0_arg0_head_arg0;
                                                                                                                                                                                                                                              arg_arg1_head_arg0_arg0_head_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                                                                                                                                                              arg_arg1_head_arg0_arg0_head_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                                                                                                                                                              if (ATgetInt((ATermInt)arg_arg1_head_arg0_arg0_head_arg0_head) == 97) {

                                                                                                                                                                                                                                                if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0_list) == ATtrue) {
                                                                                                                                                                                                                                                  if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_list) == ATfalse) {
                                                                                                                                                                                                                                                    arg_arg1_head_arg0_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_list);
                                                                                                                                                                                                                                                    arg_arg1_head_arg0_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_list);
                                                                                                                                                                                                                                                    /* checking for: char-class */
                                                                                                                                                                                                                                                    if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg0_head) == PERR_afun17) {
                                                                                                                                                                                                                                                      ATerm arg_arg1_head_arg0_arg0_head_arg0 = ATgetArgument(arg_arg1_head_arg0_arg0_head, 0);
                                                                                                                                                                                                                                                      if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0) == ATfalse) {
                                                                                                                                                                                                                                                        ATerm arg_arg1_head_arg0_arg0_head_arg0_head;
                                                                                                                                                                                                                                                        ATermList arg_arg1_head_arg0_arg0_head_arg0_list = (ATermList)arg_arg1_head_arg0_arg0_head_arg0;
                                                                                                                                                                                                                                                        arg_arg1_head_arg0_arg0_head_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                                                                                                                                                                        arg_arg1_head_arg0_arg0_head_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                                                                                                                                                                        if (ATgetInt((ATermInt)arg_arg1_head_arg0_arg0_head_arg0_head) == 114) {

                                                                                                                                                                                                                                                          if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0_list) == ATtrue) {
                                                                                                                                                                                                                                                            if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_list) == ATfalse) {
                                                                                                                                                                                                                                                              arg_arg1_head_arg0_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_list);
                                                                                                                                                                                                                                                              arg_arg1_head_arg0_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_list);
                                                                                                                                                                                                                                                              /* checking for: char-class */
                                                                                                                                                                                                                                                              if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg0_head) == PERR_afun17) {
                                                                                                                                                                                                                                                                ATerm arg_arg1_head_arg0_arg0_head_arg0 = ATgetArgument(arg_arg1_head_arg0_arg0_head, 0);
                                                                                                                                                                                                                                                                if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0) == ATfalse) {
                                                                                                                                                                                                                                                                  ATerm arg_arg1_head_arg0_arg0_head_arg0_head;
                                                                                                                                                                                                                                                                  ATermList arg_arg1_head_arg0_arg0_head_arg0_list = (ATermList)arg_arg1_head_arg0_arg0_head_arg0;
                                                                                                                                                                                                                                                                  arg_arg1_head_arg0_arg0_head_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                                                                                                                                                                                  arg_arg1_head_arg0_arg0_head_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                                                                                                                                                                                  if (ATgetInt((ATermInt)arg_arg1_head_arg0_arg0_head_arg0_head) == 101) {

                                                                                                                                                                                                                                                                    if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0_list) == ATtrue) {
                                                                                                                                                                                                                                                                      if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_list) == ATfalse) {
                                                                                                                                                                                                                                                                        arg_arg1_head_arg0_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_list);
                                                                                                                                                                                                                                                                        arg_arg1_head_arg0_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_list);
                                                                                                                                                                                                                                                                        /* checking for: char-class */
                                                                                                                                                                                                                                                                        if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg0_head) == PERR_afun17) {
                                                                                                                                                                                                                                                                          ATerm arg_arg1_head_arg0_arg0_head_arg0 = ATgetArgument(arg_arg1_head_arg0_arg0_head, 0);
                                                                                                                                                                                                                                                                          if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0) == ATfalse) {
                                                                                                                                                                                                                                                                            ATerm arg_arg1_head_arg0_arg0_head_arg0_head;
                                                                                                                                                                                                                                                                            ATermList arg_arg1_head_arg0_arg0_head_arg0_list = (ATermList)arg_arg1_head_arg0_arg0_head_arg0;
                                                                                                                                                                                                                                                                            arg_arg1_head_arg0_arg0_head_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                                                                                                                                                                                            arg_arg1_head_arg0_arg0_head_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                                                                                                                                                                                            if (ATgetInt((ATermInt)arg_arg1_head_arg0_arg0_head_arg0_head) == 97) {

                                                                                                                                                                                                                                                                              if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0_list) == ATtrue) {
                                                                                                                                                                                                                                                                                if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_list) == ATtrue) {
                                                                                                                                                                                                                                                                                  ATerm arg_arg1_head_arg0_arg1 = ATgetArgument(arg_arg1_head_arg0, 1);
                                                                                                                                                                                                                                                                                  /* checking for: lit */
                                                                                                                                                                                                                                                                                  if (ATgetType((ATerm)arg_arg1_head_arg0_arg1) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg1) == PERR_afun9) {
                                                                                                                                                                                                                                                                                    ATerm arg_arg1_head_arg0_arg1_arg0 = ATgetArgument(arg_arg1_head_arg0_arg1, 0);
                                                                                                                                                                                                                                                                                    /* checking for: area */
                                                                                                                                                                                                                                                                                    if (ATgetType((ATerm)arg_arg1_head_arg0_arg1_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg1_arg0) == PERR_afun55) {
                                                                                                                                                                                                                                                                                      ATerm arg_arg1_head_arg0_arg2 = ATgetArgument(arg_arg1_head_arg0, 2);
                                                                                                                                                                                                                                                                                      /* checking for: no-attrs */
                                                                                                                                                                                                                                                                                      if (ATgetType((ATerm)arg_arg1_head_arg0_arg2) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg2) == PERR_afun5) {
                                                                                                                                                                                                                                                                                        ATerm arg_arg1_head_arg1 = ATgetArgument(arg_arg1_head, 1);
                                                                                                                                                                                                                                                                                        if (ATgetType((ATerm)arg_arg1_head_arg1) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg1) == ATfalse) {
                                                                                                                                                                                                                                                                                          ATerm arg_arg1_head_arg1_head;
                                                                                                                                                                                                                                                                                          ATermList arg_arg1_head_arg1_list = (ATermList)arg_arg1_head_arg1;
                                                                                                                                                                                                                                                                                          arg_arg1_head_arg1_head = ATgetFirst(arg_arg1_head_arg1_list);
                                                                                                                                                                                                                                                                                          arg_arg1_head_arg1_list = ATgetNext(arg_arg1_head_arg1_list);
                                                                                                                                                                                                                                                                                          if (ATgetInt((ATermInt)arg_arg1_head_arg1_head) == 97) {

                                                                                                                                                                                                                                                                                            if (ATgetType((ATerm)arg_arg1_head_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg1_list) == ATfalse) {
                                                                                                                                                                                                                                                                                              arg_arg1_head_arg1_head = ATgetFirst(arg_arg1_head_arg1_list);
                                                                                                                                                                                                                                                                                              arg_arg1_head_arg1_list = ATgetNext(arg_arg1_head_arg1_list);
                                                                                                                                                                                                                                                                                              if (ATgetInt((ATermInt)arg_arg1_head_arg1_head) == 114) {

                                                                                                                                                                                                                                                                                                if (ATgetType((ATerm)arg_arg1_head_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg1_list) == ATfalse) {
                                                                                                                                                                                                                                                                                                  arg_arg1_head_arg1_head = ATgetFirst(arg_arg1_head_arg1_list);
                                                                                                                                                                                                                                                                                                  arg_arg1_head_arg1_list = ATgetNext(arg_arg1_head_arg1_list);
                                                                                                                                                                                                                                                                                                  if (ATgetInt((ATermInt)arg_arg1_head_arg1_head) == 101) {

                                                                                                                                                                                                                                                                                                    if (ATgetType((ATerm)arg_arg1_head_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg1_list) == ATfalse) {
                                                                                                                                                                                                                                                                                                      arg_arg1_head_arg1_head = ATgetFirst(arg_arg1_head_arg1_list);
                                                                                                                                                                                                                                                                                                      arg_arg1_head_arg1_list = ATgetNext(arg_arg1_head_arg1_list);
                                                                                                                                                                                                                                                                                                      if (ATgetInt((ATermInt)arg_arg1_head_arg1_head) == 97) {

                                                                                                                                                                                                                                                                                                        if (ATgetType((ATerm)arg_arg1_head_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg1_list) == ATtrue) {
                                                                                                                                                                                                                                                                                                          if (ATgetType((ATerm)arg_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_list) == ATfalse) {
                                                                                                                                                                                                                                                                                                            arg_arg1_list = ATgetNext(arg_arg1_list);
                                                                                                                                                                                                                                                                                                            if (ATgetType((ATerm)arg_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_list) == ATfalse) {
                                                                                                                                                                                                                                                                                                              arg_arg1_head = ATgetFirst(arg_arg1_list);
                                                                                                                                                                                                                                                                                                              arg_arg1_list = ATgetNext(arg_arg1_list);
                                                                                                                                                                                                                                                                                                              /* checking for: appl */
                                                                                                                                                                                                                                                                                                              if (ATgetType((ATerm)arg_arg1_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head) == PERR_afun0) {
                                                                                                                                                                                                                                                                                                                ATerm arg_arg1_head_arg0 = ATgetArgument(arg_arg1_head, 0);
                                                                                                                                                                                                                                                                                                                /* checking for: prod */
                                                                                                                                                                                                                                                                                                                if (ATgetType((ATerm)arg_arg1_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0) == PERR_afun1) {
                                                                                                                                                                                                                                                                                                                  ATerm arg_arg1_head_arg0_arg0 = ATgetArgument(arg_arg1_head_arg0, 0);
                                                                                                                                                                                                                                                                                                                  if (ATgetType((ATerm)arg_arg1_head_arg0_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0) == ATfalse) {
                                                                                                                                                                                                                                                                                                                    ATerm arg_arg1_head_arg0_arg0_head;
                                                                                                                                                                                                                                                                                                                    ATermList arg_arg1_head_arg0_arg0_list = (ATermList)arg_arg1_head_arg0_arg0;
                                                                                                                                                                                                                                                                                                                    arg_arg1_head_arg0_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_list);
                                                                                                                                                                                                                                                                                                                    arg_arg1_head_arg0_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_list);
                                                                                                                                                                                                                                                                                                                    /* checking for: char-class */
                                                                                                                                                                                                                                                                                                                    if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg0_head) == PERR_afun17) {
                                                                                                                                                                                                                                                                                                                      ATerm arg_arg1_head_arg0_arg0_head_arg0 = ATgetArgument(arg_arg1_head_arg0_arg0_head, 0);
                                                                                                                                                                                                                                                                                                                      if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0) == ATfalse) {
                                                                                                                                                                                                                                                                                                                        ATerm arg_arg1_head_arg0_arg0_head_arg0_head;
                                                                                                                                                                                                                                                                                                                        ATermList arg_arg1_head_arg0_arg0_head_arg0_list = (ATermList)arg_arg1_head_arg0_arg0_head_arg0;
                                                                                                                                                                                                                                                                                                                        arg_arg1_head_arg0_arg0_head_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                                                                                                                                                                                                                                        arg_arg1_head_arg0_arg0_head_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                                                                                                                                                                                                                                        if (ATgetInt((ATermInt)arg_arg1_head_arg0_arg0_head_arg0_head) == 40) {

                                                                                                                                                                                                                                                                                                                          if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0_list) == ATtrue) {
                                                                                                                                                                                                                                                                                                                            if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_list) == ATtrue) {
                                                                                                                                                                                                                                                                                                                              ATerm arg_arg1_head_arg0_arg1 = ATgetArgument(arg_arg1_head_arg0, 1);
                                                                                                                                                                                                                                                                                                                              /* checking for: lit */
                                                                                                                                                                                                                                                                                                                              if (ATgetType((ATerm)arg_arg1_head_arg0_arg1) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg1) == PERR_afun9) {
                                                                                                                                                                                                                                                                                                                                ATerm arg_arg1_head_arg0_arg1_arg0 = ATgetArgument(arg_arg1_head_arg0_arg1, 0);
                                                                                                                                                                                                                                                                                                                                /* checking for: ( */
                                                                                                                                                                                                                                                                                                                                if (ATgetType((ATerm)arg_arg1_head_arg0_arg1_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg1_arg0) == PERR_afun39) {
                                                                                                                                                                                                                                                                                                                                  ATerm arg_arg1_head_arg0_arg2 = ATgetArgument(arg_arg1_head_arg0, 2);
                                                                                                                                                                                                                                                                                                                                  /* checking for: no-attrs */
                                                                                                                                                                                                                                                                                                                                  if (ATgetType((ATerm)arg_arg1_head_arg0_arg2) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg2) == PERR_afun5) {
                                                                                                                                                                                                                                                                                                                                    ATerm arg_arg1_head_arg1 = ATgetArgument(arg_arg1_head, 1);
                                                                                                                                                                                                                                                                                                                                    if (ATgetType((ATerm)arg_arg1_head_arg1) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg1) == ATfalse) {
                                                                                                                                                                                                                                                                                                                                      ATerm arg_arg1_head_arg1_head;
                                                                                                                                                                                                                                                                                                                                      ATermList arg_arg1_head_arg1_list = (ATermList)arg_arg1_head_arg1;
                                                                                                                                                                                                                                                                                                                                      arg_arg1_head_arg1_head = ATgetFirst(arg_arg1_head_arg1_list);
                                                                                                                                                                                                                                                                                                                                      arg_arg1_head_arg1_list = ATgetNext(arg_arg1_head_arg1_list);
                                                                                                                                                                                                                                                                                                                                      if (ATgetInt((ATermInt)arg_arg1_head_arg1_head) == 40) {

                                                                                                                                                                                                                                                                                                                                        if (ATgetType((ATerm)arg_arg1_head_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg1_list) == ATtrue) {
                                                                                                                                                                                                                                                                                                                                          if (ATgetType((ATerm)arg_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_list) == ATfalse) {
                                                                                                                                                                                                                                                                                                                                            arg_arg1_list = ATgetNext(arg_arg1_list);
                                                                                                                                                                                                                                                                                                                                            if (ATgetType((ATerm)arg_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_list) == ATfalse) {
                                                                                                                                                                                                                                                                                                                                              arg_arg1_list = ATgetNext(arg_arg1_list);
                                                                                                                                                                                                                                                                                                                                              if (ATgetType((ATerm)arg_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_list) == ATfalse) {
                                                                                                                                                                                                                                                                                                                                                arg_arg1_list = ATgetNext(arg_arg1_list);
                                                                                                                                                                                                                                                                                                                                                if (ATgetType((ATerm)arg_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_list) == ATfalse) {
                                                                                                                                                                                                                                                                                                                                                  arg_arg1_head = ATgetFirst(arg_arg1_list);
                                                                                                                                                                                                                                                                                                                                                  arg_arg1_list = ATgetNext(arg_arg1_list);
                                                                                                                                                                                                                                                                                                                                                  /* checking for: appl */
                                                                                                                                                                                                                                                                                                                                                  if (ATgetType((ATerm)arg_arg1_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head) == PERR_afun0) {
                                                                                                                                                                                                                                                                                                                                                    ATerm arg_arg1_head_arg0 = ATgetArgument(arg_arg1_head, 0);
                                                                                                                                                                                                                                                                                                                                                    /* checking for: prod */
                                                                                                                                                                                                                                                                                                                                                    if (ATgetType((ATerm)arg_arg1_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0) == PERR_afun1) {
                                                                                                                                                                                                                                                                                                                                                      ATerm arg_arg1_head_arg0_arg0 = ATgetArgument(arg_arg1_head_arg0, 0);
                                                                                                                                                                                                                                                                                                                                                      if (ATgetType((ATerm)arg_arg1_head_arg0_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0) == ATfalse) {
                                                                                                                                                                                                                                                                                                                                                        ATerm arg_arg1_head_arg0_arg0_head;
                                                                                                                                                                                                                                                                                                                                                        ATermList arg_arg1_head_arg0_arg0_list = (ATermList)arg_arg1_head_arg0_arg0;
                                                                                                                                                                                                                                                                                                                                                        arg_arg1_head_arg0_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_list);
                                                                                                                                                                                                                                                                                                                                                        arg_arg1_head_arg0_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_list);
                                                                                                                                                                                                                                                                                                                                                        /* checking for: char-class */
                                                                                                                                                                                                                                                                                                                                                        if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg0_head) == PERR_afun17) {
                                                                                                                                                                                                                                                                                                                                                          ATerm arg_arg1_head_arg0_arg0_head_arg0 = ATgetArgument(arg_arg1_head_arg0_arg0_head, 0);
                                                                                                                                                                                                                                                                                                                                                          if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0) == ATfalse) {
                                                                                                                                                                                                                                                                                                                                                            ATerm arg_arg1_head_arg0_arg0_head_arg0_head;
                                                                                                                                                                                                                                                                                                                                                            ATermList arg_arg1_head_arg0_arg0_head_arg0_list = (ATermList)arg_arg1_head_arg0_arg0_head_arg0;
                                                                                                                                                                                                                                                                                                                                                            arg_arg1_head_arg0_arg0_head_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                                                                                                                                                                                                                                                                            arg_arg1_head_arg0_arg0_head_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                                                                                                                                                                                                                                                                            if (ATgetInt((ATermInt)arg_arg1_head_arg0_arg0_head_arg0_head) == 44) {

                                                                                                                                                                                                                                                                                                                                                              if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0_list) == ATtrue) {
                                                                                                                                                                                                                                                                                                                                                                if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_list) == ATtrue) {
                                                                                                                                                                                                                                                                                                                                                                  ATerm arg_arg1_head_arg0_arg1 = ATgetArgument(arg_arg1_head_arg0, 1);
                                                                                                                                                                                                                                                                                                                                                                  /* checking for: lit */
                                                                                                                                                                                                                                                                                                                                                                  if (ATgetType((ATerm)arg_arg1_head_arg0_arg1) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg1) == PERR_afun9) {
                                                                                                                                                                                                                                                                                                                                                                    ATerm arg_arg1_head_arg0_arg1_arg0 = ATgetArgument(arg_arg1_head_arg0_arg1, 0);
                                                                                                                                                                                                                                                                                                                                                                    /* checking for: , */
                                                                                                                                                                                                                                                                                                                                                                    if (ATgetType((ATerm)arg_arg1_head_arg0_arg1_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg1_arg0) == PERR_afun43) {
                                                                                                                                                                                                                                                                                                                                                                      ATerm arg_arg1_head_arg0_arg2 = ATgetArgument(arg_arg1_head_arg0, 2);
                                                                                                                                                                                                                                                                                                                                                                      /* checking for: no-attrs */
                                                                                                                                                                                                                                                                                                                                                                      if (ATgetType((ATerm)arg_arg1_head_arg0_arg2) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg2) == PERR_afun5) {
                                                                                                                                                                                                                                                                                                                                                                        ATerm arg_arg1_head_arg1 = ATgetArgument(arg_arg1_head, 1);
                                                                                                                                                                                                                                                                                                                                                                        if (ATgetType((ATerm)arg_arg1_head_arg1) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg1) == ATfalse) {
                                                                                                                                                                                                                                                                                                                                                                          ATerm arg_arg1_head_arg1_head;
                                                                                                                                                                                                                                                                                                                                                                          ATermList arg_arg1_head_arg1_list = (ATermList)arg_arg1_head_arg1;
                                                                                                                                                                                                                                                                                                                                                                          arg_arg1_head_arg1_head = ATgetFirst(arg_arg1_head_arg1_list);
                                                                                                                                                                                                                                                                                                                                                                          arg_arg1_head_arg1_list = ATgetNext(arg_arg1_head_arg1_list);
                                                                                                                                                                                                                                                                                                                                                                          if (ATgetInt((ATermInt)arg_arg1_head_arg1_head) == 44) {

                                                                                                                                                                                                                                                                                                                                                                            if (ATgetType((ATerm)arg_arg1_head_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg1_list) == ATtrue) {
                                                                                                                                                                                                                                                                                                                                                                              if (ATgetType((ATerm)arg_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_list) == ATfalse) {
                                                                                                                                                                                                                                                                                                                                                                                arg_arg1_list = ATgetNext(arg_arg1_list);
                                                                                                                                                                                                                                                                                                                                                                                if (ATgetType((ATerm)arg_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_list) == ATfalse) {
                                                                                                                                                                                                                                                                                                                                                                                  arg_arg1_list = ATgetNext(arg_arg1_list);
                                                                                                                                                                                                                                                                                                                                                                                  if (ATgetType((ATerm)arg_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_list) == ATfalse) {
                                                                                                                                                                                                                                                                                                                                                                                    arg_arg1_list = ATgetNext(arg_arg1_list);
                                                                                                                                                                                                                                                                                                                                                                                    if (ATgetType((ATerm)arg_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_list) == ATfalse) {
                                                                                                                                                                                                                                                                                                                                                                                      arg_arg1_head = ATgetFirst(arg_arg1_list);
                                                                                                                                                                                                                                                                                                                                                                                      arg_arg1_list = ATgetNext(arg_arg1_list);
                                                                                                                                                                                                                                                                                                                                                                                      /* checking for: appl */
                                                                                                                                                                                                                                                                                                                                                                                      if (ATgetType((ATerm)arg_arg1_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head) == PERR_afun0) {
                                                                                                                                                                                                                                                                                                                                                                                        ATerm arg_arg1_head_arg0 = ATgetArgument(arg_arg1_head, 0);
                                                                                                                                                                                                                                                                                                                                                                                        /* checking for: prod */
                                                                                                                                                                                                                                                                                                                                                                                        if (ATgetType((ATerm)arg_arg1_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0) == PERR_afun1) {
                                                                                                                                                                                                                                                                                                                                                                                          ATerm arg_arg1_head_arg0_arg0 = ATgetArgument(arg_arg1_head_arg0, 0);
                                                                                                                                                                                                                                                                                                                                                                                          if (ATgetType((ATerm)arg_arg1_head_arg0_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0) == ATfalse) {
                                                                                                                                                                                                                                                                                                                                                                                            ATerm arg_arg1_head_arg0_arg0_head;
                                                                                                                                                                                                                                                                                                                                                                                            ATermList arg_arg1_head_arg0_arg0_list = (ATermList)arg_arg1_head_arg0_arg0;
                                                                                                                                                                                                                                                                                                                                                                                            arg_arg1_head_arg0_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_list);
                                                                                                                                                                                                                                                                                                                                                                                            arg_arg1_head_arg0_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_list);
                                                                                                                                                                                                                                                                                                                                                                                            /* checking for: char-class */
                                                                                                                                                                                                                                                                                                                                                                                            if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg0_head) == PERR_afun17) {
                                                                                                                                                                                                                                                                                                                                                                                              ATerm arg_arg1_head_arg0_arg0_head_arg0 = ATgetArgument(arg_arg1_head_arg0_arg0_head, 0);
                                                                                                                                                                                                                                                                                                                                                                                              if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0) == ATfalse) {
                                                                                                                                                                                                                                                                                                                                                                                                ATerm arg_arg1_head_arg0_arg0_head_arg0_head;
                                                                                                                                                                                                                                                                                                                                                                                                ATermList arg_arg1_head_arg0_arg0_head_arg0_list = (ATermList)arg_arg1_head_arg0_arg0_head_arg0;
                                                                                                                                                                                                                                                                                                                                                                                                arg_arg1_head_arg0_arg0_head_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                                                                                                                                                                                                                                                                                                                arg_arg1_head_arg0_arg0_head_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                                                                                                                                                                                                                                                                                                                if (ATgetInt((ATermInt)arg_arg1_head_arg0_arg0_head_arg0_head) == 44) {

                                                                                                                                                                                                                                                                                                                                                                                                  if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0_list) == ATtrue) {
                                                                                                                                                                                                                                                                                                                                                                                                    if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_list) == ATtrue) {
                                                                                                                                                                                                                                                                                                                                                                                                      ATerm arg_arg1_head_arg0_arg1 = ATgetArgument(arg_arg1_head_arg0, 1);
                                                                                                                                                                                                                                                                                                                                                                                                      /* checking for: lit */
                                                                                                                                                                                                                                                                                                                                                                                                      if (ATgetType((ATerm)arg_arg1_head_arg0_arg1) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg1) == PERR_afun9) {
                                                                                                                                                                                                                                                                                                                                                                                                        ATerm arg_arg1_head_arg0_arg1_arg0 = ATgetArgument(arg_arg1_head_arg0_arg1, 0);
                                                                                                                                                                                                                                                                                                                                                                                                        /* checking for: , */
                                                                                                                                                                                                                                                                                                                                                                                                        if (ATgetType((ATerm)arg_arg1_head_arg0_arg1_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg1_arg0) == PERR_afun43) {
                                                                                                                                                                                                                                                                                                                                                                                                          ATerm arg_arg1_head_arg0_arg2 = ATgetArgument(arg_arg1_head_arg0, 2);
                                                                                                                                                                                                                                                                                                                                                                                                          /* checking for: no-attrs */
                                                                                                                                                                                                                                                                                                                                                                                                          if (ATgetType((ATerm)arg_arg1_head_arg0_arg2) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg2) == PERR_afun5) {
                                                                                                                                                                                                                                                                                                                                                                                                            ATerm arg_arg1_head_arg1 = ATgetArgument(arg_arg1_head, 1);
                                                                                                                                                                                                                                                                                                                                                                                                            if (ATgetType((ATerm)arg_arg1_head_arg1) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg1) == ATfalse) {
                                                                                                                                                                                                                                                                                                                                                                                                              ATerm arg_arg1_head_arg1_head;
                                                                                                                                                                                                                                                                                                                                                                                                              ATermList arg_arg1_head_arg1_list = (ATermList)arg_arg1_head_arg1;
                                                                                                                                                                                                                                                                                                                                                                                                              arg_arg1_head_arg1_head = ATgetFirst(arg_arg1_head_arg1_list);
                                                                                                                                                                                                                                                                                                                                                                                                              arg_arg1_head_arg1_list = ATgetNext(arg_arg1_head_arg1_list);
                                                                                                                                                                                                                                                                                                                                                                                                              if (ATgetInt((ATermInt)arg_arg1_head_arg1_head) == 44) {

                                                                                                                                                                                                                                                                                                                                                                                                                if (ATgetType((ATerm)arg_arg1_head_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg1_list) == ATtrue) {
                                                                                                                                                                                                                                                                                                                                                                                                                  if (ATgetType((ATerm)arg_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_list) == ATfalse) {
                                                                                                                                                                                                                                                                                                                                                                                                                    arg_arg1_list = ATgetNext(arg_arg1_list);
                                                                                                                                                                                                                                                                                                                                                                                                                    if (ATgetType((ATerm)arg_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_list) == ATfalse) {
                                                                                                                                                                                                                                                                                                                                                                                                                      arg_arg1_list = ATgetNext(arg_arg1_list);
                                                                                                                                                                                                                                                                                                                                                                                                                      if (ATgetType((ATerm)arg_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_list) == ATfalse) {
                                                                                                                                                                                                                                                                                                                                                                                                                        arg_arg1_list = ATgetNext(arg_arg1_list);
                                                                                                                                                                                                                                                                                                                                                                                                                        if (ATgetType((ATerm)arg_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_list) == ATfalse) {
                                                                                                                                                                                                                                                                                                                                                                                                                          arg_arg1_head = ATgetFirst(arg_arg1_list);
                                                                                                                                                                                                                                                                                                                                                                                                                          arg_arg1_list = ATgetNext(arg_arg1_list);
                                                                                                                                                                                                                                                                                                                                                                                                                          /* checking for: appl */
                                                                                                                                                                                                                                                                                                                                                                                                                          if (ATgetType((ATerm)arg_arg1_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head) == PERR_afun0) {
                                                                                                                                                                                                                                                                                                                                                                                                                            ATerm arg_arg1_head_arg0 = ATgetArgument(arg_arg1_head, 0);
                                                                                                                                                                                                                                                                                                                                                                                                                            /* checking for: prod */
                                                                                                                                                                                                                                                                                                                                                                                                                            if (ATgetType((ATerm)arg_arg1_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0) == PERR_afun1) {
                                                                                                                                                                                                                                                                                                                                                                                                                              ATerm arg_arg1_head_arg0_arg0 = ATgetArgument(arg_arg1_head_arg0, 0);
                                                                                                                                                                                                                                                                                                                                                                                                                              if (ATgetType((ATerm)arg_arg1_head_arg0_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0) == ATfalse) {
                                                                                                                                                                                                                                                                                                                                                                                                                                ATerm arg_arg1_head_arg0_arg0_head;
                                                                                                                                                                                                                                                                                                                                                                                                                                ATermList arg_arg1_head_arg0_arg0_list = (ATermList)arg_arg1_head_arg0_arg0;
                                                                                                                                                                                                                                                                                                                                                                                                                                arg_arg1_head_arg0_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_list);
                                                                                                                                                                                                                                                                                                                                                                                                                                arg_arg1_head_arg0_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_list);
                                                                                                                                                                                                                                                                                                                                                                                                                                /* checking for: char-class */
                                                                                                                                                                                                                                                                                                                                                                                                                                if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg0_head) == PERR_afun17) {
                                                                                                                                                                                                                                                                                                                                                                                                                                  ATerm arg_arg1_head_arg0_arg0_head_arg0 = ATgetArgument(arg_arg1_head_arg0_arg0_head, 0);
                                                                                                                                                                                                                                                                                                                                                                                                                                  if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0) == ATfalse) {
                                                                                                                                                                                                                                                                                                                                                                                                                                    ATerm arg_arg1_head_arg0_arg0_head_arg0_head;
                                                                                                                                                                                                                                                                                                                                                                                                                                    ATermList arg_arg1_head_arg0_arg0_head_arg0_list = (ATermList)arg_arg1_head_arg0_arg0_head_arg0;
                                                                                                                                                                                                                                                                                                                                                                                                                                    arg_arg1_head_arg0_arg0_head_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                                                                                                                                                                                                                                                                                                                                                    arg_arg1_head_arg0_arg0_head_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                                                                                                                                                                                                                                                                                                                                                    if (ATgetInt((ATermInt)arg_arg1_head_arg0_arg0_head_arg0_head) == 44) {

                                                                                                                                                                                                                                                                                                                                                                                                                                      if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0_list) == ATtrue) {
                                                                                                                                                                                                                                                                                                                                                                                                                                        if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_list) == ATtrue) {
                                                                                                                                                                                                                                                                                                                                                                                                                                          ATerm arg_arg1_head_arg0_arg1 = ATgetArgument(arg_arg1_head_arg0, 1);
                                                                                                                                                                                                                                                                                                                                                                                                                                          /* checking for: lit */
                                                                                                                                                                                                                                                                                                                                                                                                                                          if (ATgetType((ATerm)arg_arg1_head_arg0_arg1) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg1) == PERR_afun9) {
                                                                                                                                                                                                                                                                                                                                                                                                                                            ATerm arg_arg1_head_arg0_arg1_arg0 = ATgetArgument(arg_arg1_head_arg0_arg1, 0);
                                                                                                                                                                                                                                                                                                                                                                                                                                            /* checking for: , */
                                                                                                                                                                                                                                                                                                                                                                                                                                            if (ATgetType((ATerm)arg_arg1_head_arg0_arg1_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg1_arg0) == PERR_afun43) {
                                                                                                                                                                                                                                                                                                                                                                                                                                              ATerm arg_arg1_head_arg0_arg2 = ATgetArgument(arg_arg1_head_arg0, 2);
                                                                                                                                                                                                                                                                                                                                                                                                                                              /* checking for: no-attrs */
                                                                                                                                                                                                                                                                                                                                                                                                                                              if (ATgetType((ATerm)arg_arg1_head_arg0_arg2) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg2) == PERR_afun5) {
                                                                                                                                                                                                                                                                                                                                                                                                                                                ATerm arg_arg1_head_arg1 = ATgetArgument(arg_arg1_head, 1);
                                                                                                                                                                                                                                                                                                                                                                                                                                                if (ATgetType((ATerm)arg_arg1_head_arg1) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg1) == ATfalse) {
                                                                                                                                                                                                                                                                                                                                                                                                                                                  ATerm arg_arg1_head_arg1_head;
                                                                                                                                                                                                                                                                                                                                                                                                                                                  ATermList arg_arg1_head_arg1_list = (ATermList)arg_arg1_head_arg1;
                                                                                                                                                                                                                                                                                                                                                                                                                                                  arg_arg1_head_arg1_head = ATgetFirst(arg_arg1_head_arg1_list);
                                                                                                                                                                                                                                                                                                                                                                                                                                                  arg_arg1_head_arg1_list = ATgetNext(arg_arg1_head_arg1_list);
                                                                                                                                                                                                                                                                                                                                                                                                                                                  if (ATgetInt((ATermInt)arg_arg1_head_arg1_head) == 44) {

                                                                                                                                                                                                                                                                                                                                                                                                                                                    if (ATgetType((ATerm)arg_arg1_head_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg1_list) == ATtrue) {
                                                                                                                                                                                                                                                                                                                                                                                                                                                      if (ATgetType((ATerm)arg_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_list) == ATfalse) {
                                                                                                                                                                                                                                                                                                                                                                                                                                                        arg_arg1_list = ATgetNext(arg_arg1_list);
                                                                                                                                                                                                                                                                                                                                                                                                                                                        if (ATgetType((ATerm)arg_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_list) == ATfalse) {
                                                                                                                                                                                                                                                                                                                                                                                                                                                          arg_arg1_list = ATgetNext(arg_arg1_list);
                                                                                                                                                                                                                                                                                                                                                                                                                                                          if (ATgetType((ATerm)arg_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_list) == ATfalse) {
                                                                                                                                                                                                                                                                                                                                                                                                                                                            arg_arg1_list = ATgetNext(arg_arg1_list);
                                                                                                                                                                                                                                                                                                                                                                                                                                                            if (ATgetType((ATerm)arg_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_list) == ATfalse) {
                                                                                                                                                                                                                                                                                                                                                                                                                                                              arg_arg1_head = ATgetFirst(arg_arg1_list);
                                                                                                                                                                                                                                                                                                                                                                                                                                                              arg_arg1_list = ATgetNext(arg_arg1_list);
                                                                                                                                                                                                                                                                                                                                                                                                                                                              /* checking for: appl */
                                                                                                                                                                                                                                                                                                                                                                                                                                                              if (ATgetType((ATerm)arg_arg1_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head) == PERR_afun0) {
                                                                                                                                                                                                                                                                                                                                                                                                                                                                ATerm arg_arg1_head_arg0 = ATgetArgument(arg_arg1_head, 0);
                                                                                                                                                                                                                                                                                                                                                                                                                                                                /* checking for: prod */
                                                                                                                                                                                                                                                                                                                                                                                                                                                                if (ATgetType((ATerm)arg_arg1_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0) == PERR_afun1) {
                                                                                                                                                                                                                                                                                                                                                                                                                                                                  ATerm arg_arg1_head_arg0_arg0 = ATgetArgument(arg_arg1_head_arg0, 0);
                                                                                                                                                                                                                                                                                                                                                                                                                                                                  if (ATgetType((ATerm)arg_arg1_head_arg0_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0) == ATfalse) {
                                                                                                                                                                                                                                                                                                                                                                                                                                                                    ATerm arg_arg1_head_arg0_arg0_head;
                                                                                                                                                                                                                                                                                                                                                                                                                                                                    ATermList arg_arg1_head_arg0_arg0_list = (ATermList)arg_arg1_head_arg0_arg0;
                                                                                                                                                                                                                                                                                                                                                                                                                                                                    arg_arg1_head_arg0_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_list);
                                                                                                                                                                                                                                                                                                                                                                                                                                                                    arg_arg1_head_arg0_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_list);
                                                                                                                                                                                                                                                                                                                                                                                                                                                                    /* checking for: char-class */
                                                                                                                                                                                                                                                                                                                                                                                                                                                                    if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg0_head) == PERR_afun17) {
                                                                                                                                                                                                                                                                                                                                                                                                                                                                      ATerm arg_arg1_head_arg0_arg0_head_arg0 = ATgetArgument(arg_arg1_head_arg0_arg0_head, 0);
                                                                                                                                                                                                                                                                                                                                                                                                                                                                      if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0) == ATfalse) {
                                                                                                                                                                                                                                                                                                                                                                                                                                                                        ATerm arg_arg1_head_arg0_arg0_head_arg0_head;
                                                                                                                                                                                                                                                                                                                                                                                                                                                                        ATermList arg_arg1_head_arg0_arg0_head_arg0_list = (ATermList)arg_arg1_head_arg0_arg0_head_arg0;
                                                                                                                                                                                                                                                                                                                                                                                                                                                                        arg_arg1_head_arg0_arg0_head_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                                                                                                                                                                                                                                                                                                                                                                                        arg_arg1_head_arg0_arg0_head_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                                                                                                                                                                                                                                                                                                                                                                                        if (ATgetInt((ATermInt)arg_arg1_head_arg0_arg0_head_arg0_head) == 44) {

                                                                                                                                                                                                                                                                                                                                                                                                                                                                          if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0_list) == ATtrue) {
                                                                                                                                                                                                                                                                                                                                                                                                                                                                            if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_list) == ATtrue) {
                                                                                                                                                                                                                                                                                                                                                                                                                                                                              ATerm arg_arg1_head_arg0_arg1 = ATgetArgument(arg_arg1_head_arg0, 1);
                                                                                                                                                                                                                                                                                                                                                                                                                                                                              /* checking for: lit */
                                                                                                                                                                                                                                                                                                                                                                                                                                                                              if (ATgetType((ATerm)arg_arg1_head_arg0_arg1) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg1) == PERR_afun9) {
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                ATerm arg_arg1_head_arg0_arg1_arg0 = ATgetArgument(arg_arg1_head_arg0_arg1, 0);
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                /* checking for: , */
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                if (ATgetType((ATerm)arg_arg1_head_arg0_arg1_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg1_arg0) == PERR_afun43) {
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                  ATerm arg_arg1_head_arg0_arg2 = ATgetArgument(arg_arg1_head_arg0, 2);
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                  /* checking for: no-attrs */
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                  if (ATgetType((ATerm)arg_arg1_head_arg0_arg2) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg2) == PERR_afun5) {
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                    ATerm arg_arg1_head_arg1 = ATgetArgument(arg_arg1_head, 1);
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                    if (ATgetType((ATerm)arg_arg1_head_arg1) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg1) == ATfalse) {
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                      ATerm arg_arg1_head_arg1_head;
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                      ATermList arg_arg1_head_arg1_list = (ATermList)arg_arg1_head_arg1;
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                      arg_arg1_head_arg1_head = ATgetFirst(arg_arg1_head_arg1_list);
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                      arg_arg1_head_arg1_list = ATgetNext(arg_arg1_head_arg1_list);
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                      if (ATgetInt((ATermInt)arg_arg1_head_arg1_head) == 44) {

                                                                                                                                                                                                                                                                                                                                                                                                                                                                                        if (ATgetType((ATerm)arg_arg1_head_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg1_list) == ATtrue) {
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                          if (ATgetType((ATerm)arg_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_list) == ATfalse) {
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                            arg_arg1_list = ATgetNext(arg_arg1_list);
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                            if (ATgetType((ATerm)arg_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_list) == ATfalse) {
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                              arg_arg1_list = ATgetNext(arg_arg1_list);
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                              if (ATgetType((ATerm)arg_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_list) == ATfalse) {
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                arg_arg1_list = ATgetNext(arg_arg1_list);
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                if (ATgetType((ATerm)arg_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_list) == ATfalse) {
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                  arg_arg1_head = ATgetFirst(arg_arg1_list);
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                  arg_arg1_list = ATgetNext(arg_arg1_list);
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                  /* checking for: appl */
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                  if (ATgetType((ATerm)arg_arg1_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head) == PERR_afun0) {
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                    ATerm arg_arg1_head_arg0 = ATgetArgument(arg_arg1_head, 0);
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                    /* checking for: prod */
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                    if (ATgetType((ATerm)arg_arg1_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0) == PERR_afun1) {
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                      ATerm arg_arg1_head_arg0_arg0 = ATgetArgument(arg_arg1_head_arg0, 0);
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                      if (ATgetType((ATerm)arg_arg1_head_arg0_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0) == ATfalse) {
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                        ATerm arg_arg1_head_arg0_arg0_head;
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                        ATermList arg_arg1_head_arg0_arg0_list = (ATermList)arg_arg1_head_arg0_arg0;
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                        arg_arg1_head_arg0_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_list);
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                        arg_arg1_head_arg0_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_list);
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                        /* checking for: char-class */
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                        if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg0_head) == PERR_afun17) {
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                          ATerm arg_arg1_head_arg0_arg0_head_arg0 = ATgetArgument(arg_arg1_head_arg0_arg0_head, 0);
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                          if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0) == ATfalse) {
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                            ATerm arg_arg1_head_arg0_arg0_head_arg0_head;
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                            ATermList arg_arg1_head_arg0_arg0_head_arg0_list = (ATermList)arg_arg1_head_arg0_arg0_head_arg0;
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                            arg_arg1_head_arg0_arg0_head_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                            arg_arg1_head_arg0_arg0_head_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                            if (ATgetInt((ATermInt)arg_arg1_head_arg0_arg0_head_arg0_head) == 44) {

                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                              if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0_list) == ATtrue) {
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_list) == ATtrue) {
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                  ATerm arg_arg1_head_arg0_arg1 = ATgetArgument(arg_arg1_head_arg0, 1);
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                  /* checking for: lit */
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                  if (ATgetType((ATerm)arg_arg1_head_arg0_arg1) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg1) == PERR_afun9) {
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                    ATerm arg_arg1_head_arg0_arg1_arg0 = ATgetArgument(arg_arg1_head_arg0_arg1, 0);
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                    /* checking for: , */
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                    if (ATgetType((ATerm)arg_arg1_head_arg0_arg1_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg1_arg0) == PERR_afun43) {
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                      ATerm arg_arg1_head_arg0_arg2 = ATgetArgument(arg_arg1_head_arg0, 2);
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                      /* checking for: no-attrs */
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                      if (ATgetType((ATerm)arg_arg1_head_arg0_arg2) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg2) == PERR_afun5) {
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                        ATerm arg_arg1_head_arg1 = ATgetArgument(arg_arg1_head, 1);
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                        if (ATgetType((ATerm)arg_arg1_head_arg1) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg1) == ATfalse) {
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                          ATerm arg_arg1_head_arg1_head;
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                          ATermList arg_arg1_head_arg1_list = (ATermList)arg_arg1_head_arg1;
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                          arg_arg1_head_arg1_head = ATgetFirst(arg_arg1_head_arg1_list);
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                          arg_arg1_head_arg1_list = ATgetNext(arg_arg1_head_arg1_list);
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                          if (ATgetInt((ATermInt)arg_arg1_head_arg1_head) == 44) {

                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                            if (ATgetType((ATerm)arg_arg1_head_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg1_list) == ATtrue) {
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                              if (ATgetType((ATerm)arg_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_list) == ATfalse) {
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                arg_arg1_list = ATgetNext(arg_arg1_list);
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                if (ATgetType((ATerm)arg_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_list) == ATfalse) {
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                  arg_arg1_list = ATgetNext(arg_arg1_list);
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                  if (ATgetType((ATerm)arg_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_list) == ATfalse) {
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                    arg_arg1_list = ATgetNext(arg_arg1_list);
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                    if (ATgetType((ATerm)arg_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_list) == ATfalse) {
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                      arg_arg1_head = ATgetFirst(arg_arg1_list);
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                      arg_arg1_list = ATgetNext(arg_arg1_list);
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                      /* checking for: appl */
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                      if (ATgetType((ATerm)arg_arg1_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head) == PERR_afun0) {
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                        ATerm arg_arg1_head_arg0 = ATgetArgument(arg_arg1_head, 0);
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                        /* checking for: prod */
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                        if (ATgetType((ATerm)arg_arg1_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0) == PERR_afun1) {
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                          ATerm arg_arg1_head_arg0_arg0 = ATgetArgument(arg_arg1_head_arg0, 0);
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                          if (ATgetType((ATerm)arg_arg1_head_arg0_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0) == ATfalse) {
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                            ATerm arg_arg1_head_arg0_arg0_head;
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                            ATermList arg_arg1_head_arg0_arg0_list = (ATermList)arg_arg1_head_arg0_arg0;
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                            arg_arg1_head_arg0_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_list);
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                            arg_arg1_head_arg0_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_list);
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                            /* checking for: char-class */
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                            if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg0_head) == PERR_afun17) {
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                              ATerm arg_arg1_head_arg0_arg0_head_arg0 = ATgetArgument(arg_arg1_head_arg0_arg0_head, 0);
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                              if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0) == ATfalse) {
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                ATerm arg_arg1_head_arg0_arg0_head_arg0_head;
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                ATermList arg_arg1_head_arg0_arg0_head_arg0_list = (ATermList)arg_arg1_head_arg0_arg0_head_arg0;
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                arg_arg1_head_arg0_arg0_head_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                arg_arg1_head_arg0_arg0_head_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                if (ATgetInt((ATermInt)arg_arg1_head_arg0_arg0_head_arg0_head) == 41) {

                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                  if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0_list) == ATtrue) {
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                    if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_list) == ATtrue) {
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                      ATerm arg_arg1_head_arg0_arg1 = ATgetArgument(arg_arg1_head_arg0, 1);
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                      /* checking for: lit */
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                      if (ATgetType((ATerm)arg_arg1_head_arg0_arg1) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg1) == PERR_afun9) {
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                        ATerm arg_arg1_head_arg0_arg1_arg0 = ATgetArgument(arg_arg1_head_arg0_arg1, 0);
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                        /* checking for: ) */
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                        if (ATgetType((ATerm)arg_arg1_head_arg0_arg1_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg1_arg0) == PERR_afun38) {
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                          ATerm arg_arg1_head_arg0_arg2 = ATgetArgument(arg_arg1_head_arg0, 2);
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                          /* checking for: no-attrs */
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                          if (ATgetType((ATerm)arg_arg1_head_arg0_arg2) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg2) == PERR_afun5) {
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                            ATerm arg_arg1_head_arg1 = ATgetArgument(arg_arg1_head, 1);
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                            if (ATgetType((ATerm)arg_arg1_head_arg1) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg1) == ATfalse) {
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                              ATerm arg_arg1_head_arg1_head;
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                              ATermList arg_arg1_head_arg1_list = (ATermList)arg_arg1_head_arg1;
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                              arg_arg1_head_arg1_head = ATgetFirst(arg_arg1_head_arg1_list);
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                              arg_arg1_head_arg1_list = ATgetNext(arg_arg1_head_arg1_list);
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                              if (ATgetInt((ATermInt)arg_arg1_head_arg1_head) == 41) {

                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                if (ATgetType((ATerm)arg_arg1_head_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg1_list) == ATtrue) {
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                  if (ATgetType((ATerm)arg_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_list) == ATtrue) {
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                    return ATtrue;
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                  }
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                }
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                              }
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                            }
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                          }
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                        }
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                      }
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                    }
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                  }
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                }
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                              }
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                            }
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                          }
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                        }
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                      }
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                    }
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                  }
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                }
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                              }
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                            }
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                          }
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                        }
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                      }
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                    }
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                  }
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                }
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                              }
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                            }
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                          }
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                        }
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                      }
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                    }
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                  }
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                }
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                              }
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                            }
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                          }
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                        }
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                      }
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                    }
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                  }
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                }
                                                                                                                                                                                                                                                                                                                                                                                                                                                                              }
                                                                                                                                                                                                                                                                                                                                                                                                                                                                            }
                                                                                                                                                                                                                                                                                                                                                                                                                                                                          }
                                                                                                                                                                                                                                                                                                                                                                                                                                                                        }
                                                                                                                                                                                                                                                                                                                                                                                                                                                                      }
                                                                                                                                                                                                                                                                                                                                                                                                                                                                    }
                                                                                                                                                                                                                                                                                                                                                                                                                                                                  }
                                                                                                                                                                                                                                                                                                                                                                                                                                                                }
                                                                                                                                                                                                                                                                                                                                                                                                                                                              }
                                                                                                                                                                                                                                                                                                                                                                                                                                                            }
                                                                                                                                                                                                                                                                                                                                                                                                                                                          }
                                                                                                                                                                                                                                                                                                                                                                                                                                                        }
                                                                                                                                                                                                                                                                                                                                                                                                                                                      }
                                                                                                                                                                                                                                                                                                                                                                                                                                                    }
                                                                                                                                                                                                                                                                                                                                                                                                                                                  }
                                                                                                                                                                                                                                                                                                                                                                                                                                                }
                                                                                                                                                                                                                                                                                                                                                                                                                                              }
                                                                                                                                                                                                                                                                                                                                                                                                                                            }
                                                                                                                                                                                                                                                                                                                                                                                                                                          }
                                                                                                                                                                                                                                                                                                                                                                                                                                        }
                                                                                                                                                                                                                                                                                                                                                                                                                                      }
                                                                                                                                                                                                                                                                                                                                                                                                                                    }
                                                                                                                                                                                                                                                                                                                                                                                                                                  }
                                                                                                                                                                                                                                                                                                                                                                                                                                }
                                                                                                                                                                                                                                                                                                                                                                                                                              }
                                                                                                                                                                                                                                                                                                                                                                                                                            }
                                                                                                                                                                                                                                                                                                                                                                                                                          }
                                                                                                                                                                                                                                                                                                                                                                                                                        }
                                                                                                                                                                                                                                                                                                                                                                                                                      }
                                                                                                                                                                                                                                                                                                                                                                                                                    }
                                                                                                                                                                                                                                                                                                                                                                                                                  }
                                                                                                                                                                                                                                                                                                                                                                                                                }
                                                                                                                                                                                                                                                                                                                                                                                                              }
                                                                                                                                                                                                                                                                                                                                                                                                            }
                                                                                                                                                                                                                                                                                                                                                                                                          }
                                                                                                                                                                                                                                                                                                                                                                                                        }
                                                                                                                                                                                                                                                                                                                                                                                                      }
                                                                                                                                                                                                                                                                                                                                                                                                    }
                                                                                                                                                                                                                                                                                                                                                                                                  }
                                                                                                                                                                                                                                                                                                                                                                                                }
                                                                                                                                                                                                                                                                                                                                                                                              }
                                                                                                                                                                                                                                                                                                                                                                                            }
                                                                                                                                                                                                                                                                                                                                                                                          }
                                                                                                                                                                                                                                                                                                                                                                                        }
                                                                                                                                                                                                                                                                                                                                                                                      }
                                                                                                                                                                                                                                                                                                                                                                                    }
                                                                                                                                                                                                                                                                                                                                                                                  }
                                                                                                                                                                                                                                                                                                                                                                                }
                                                                                                                                                                                                                                                                                                                                                                              }
                                                                                                                                                                                                                                                                                                                                                                            }
                                                                                                                                                                                                                                                                                                                                                                          }
                                                                                                                                                                                                                                                                                                                                                                        }
                                                                                                                                                                                                                                                                                                                                                                      }
                                                                                                                                                                                                                                                                                                                                                                    }
                                                                                                                                                                                                                                                                                                                                                                  }
                                                                                                                                                                                                                                                                                                                                                                }
                                                                                                                                                                                                                                                                                                                                                              }
                                                                                                                                                                                                                                                                                                                                                            }
                                                                                                                                                                                                                                                                                                                                                          }
                                                                                                                                                                                                                                                                                                                                                        }
                                                                                                                                                                                                                                                                                                                                                      }
                                                                                                                                                                                                                                                                                                                                                    }
                                                                                                                                                                                                                                                                                                                                                  }
                                                                                                                                                                                                                                                                                                                                                }
                                                                                                                                                                                                                                                                                                                                              }
                                                                                                                                                                                                                                                                                                                                            }
                                                                                                                                                                                                                                                                                                                                          }
                                                                                                                                                                                                                                                                                                                                        }
                                                                                                                                                                                                                                                                                                                                      }
                                                                                                                                                                                                                                                                                                                                    }
                                                                                                                                                                                                                                                                                                                                  }
                                                                                                                                                                                                                                                                                                                                }
                                                                                                                                                                                                                                                                                                                              }
                                                                                                                                                                                                                                                                                                                            }
                                                                                                                                                                                                                                                                                                                          }
                                                                                                                                                                                                                                                                                                                        }
                                                                                                                                                                                                                                                                                                                      }
                                                                                                                                                                                                                                                                                                                    }
                                                                                                                                                                                                                                                                                                                  }
                                                                                                                                                                                                                                                                                                                }
                                                                                                                                                                                                                                                                                                              }
                                                                                                                                                                                                                                                                                                            }
                                                                                                                                                                                                                                                                                                          }
                                                                                                                                                                                                                                                                                                        }
                                                                                                                                                                                                                                                                                                      }
                                                                                                                                                                                                                                                                                                    }
                                                                                                                                                                                                                                                                                                  }
                                                                                                                                                                                                                                                                                                }
                                                                                                                                                                                                                                                                                              }
                                                                                                                                                                                                                                                                                            }
                                                                                                                                                                                                                                                                                          }
                                                                                                                                                                                                                                                                                        }
                                                                                                                                                                                                                                                                                      }
                                                                                                                                                                                                                                                                                    }
                                                                                                                                                                                                                                                                                  }
                                                                                                                                                                                                                                                                                }
                                                                                                                                                                                                                                                                              }
                                                                                                                                                                                                                                                                            }
                                                                                                                                                                                                                                                                          }
                                                                                                                                                                                                                                                                        }
                                                                                                                                                                                                                                                                      }
                                                                                                                                                                                                                                                                    }
                                                                                                                                                                                                                                                                  }
                                                                                                                                                                                                                                                                }
                                                                                                                                                                                                                                                              }
                                                                                                                                                                                                                                                            }
                                                                                                                                                                                                                                                          }
                                                                                                                                                                                                                                                        }
                                                                                                                                                                                                                                                      }
                                                                                                                                                                                                                                                    }
                                                                                                                                                                                                                                                  }
                                                                                                                                                                                                                                                }
                                                                                                                                                                                                                                              }
                                                                                                                                                                                                                                            }
                                                                                                                                                                                                                                          }
                                                                                                                                                                                                                                        }
                                                                                                                                                                                                                                      }
                                                                                                                                                                                                                                    }
                                                                                                                                                                                                                                  }
                                                                                                                                                                                                                                }
                                                                                                                                                                                                                              }
                                                                                                                                                                                                                            }
                                                                                                                                                                                                                          }
                                                                                                                                                                                                                        }
                                                                                                                                                                                                                      }
                                                                                                                                                                                                                    }
                                                                                                                                                                                                                  }
                                                                                                                                                                                                                }
                                                                                                                                                                                                              }
                                                                                                                                                                                                            }
                                                                                                                                                                                                          }
                                                                                                                                                                                                        }
                                                                                                                                                                                                      }
                                                                                                                                                                                                    }
                                                                                                                                                                                                  }
                                                                                                                                                                                                }
                                                                                                                                                                                              }
                                                                                                                                                                                            }
                                                                                                                                                                                          }
                                                                                                                                                                                        }
                                                                                                                                                                                      }
                                                                                                                                                                                    }
                                                                                                                                                                                  }
                                                                                                                                                                                }
                                                                                                                                                                              }
                                                                                                                                                                            }
                                                                                                                                                                          }
                                                                                                                                                                        }
                                                                                                                                                                      }
                                                                                                                                                                    }
                                                                                                                                                                  }
                                                                                                                                                                }
                                                                                                                                                              }
                                                                                                                                                            }
                                                                                                                                                          }
                                                                                                                                                        }
                                                                                                                                                      }
                                                                                                                                                    }
                                                                                                                                                  }
                                                                                                                                                }
                                                                                                                                              }
                                                                                                                                            }
                                                                                                                                          }
                                                                                                                                        }
                                                                                                                                      }
                                                                                                                                    }
                                                                                                                                  }
                                                                                                                                }
                                                                                                                              }
                                                                                                                            }
                                                                                                                          }
                                                                                                                        }
                                                                                                                      }
                                                                                                                    }
                                                                                                                  }
                                                                                                                }
                                                                                                              }
                                                                                                            }
                                                                                                          }
                                                                                                        }
                                                                                                      }
                                                                                                    }
                                                                                                  }
                                                                                                }
                                                                                              }
                                                                                            }
                                                                                          }
                                                                                        }
                                                                                      }
                                                                                    }
                                                                                  }
                                                                                }
                                                                              }
                                                                            }
                                                                          }
                                                                        }
                                                                      }
                                                                    }
                                                                  }
                                                                }
                                                              }
                                                            }
                                                          }
                                                        }
                                                      }
                                                    }
                                                  }
                                                }
                                              }
                                            }
                                          }
                                        }
                                      }
                                    }
                                  }
                                }
                              }
                            }
                          }
                        }
                      }
                    }
                  }
                }
              }
            }
          }
        }
      }
    }
  }
  return ATfalse;
}

/**
 * Assert whether a PERR_Area has a ws-after-area. 
 * \param[in] arg input PERR_Area
 * \return ATtrue if the PERR_Area had a ws-after-area, or ATfalse otherwise
 */
ATbool PERR_hasAreaWsAfterArea(PERR_Area arg) {
  if (PERR_isAreaArea(arg)) {
    return ATtrue;
  }
  return ATfalse;
}

/**
 * Assert whether a PERR_Area has a ws-after-(. 
 * \param[in] arg input PERR_Area
 * \return ATtrue if the PERR_Area had a ws-after-(, or ATfalse otherwise
 */
ATbool PERR_hasAreaWsAfterParenOpen(PERR_Area arg) {
  if (PERR_isAreaArea(arg)) {
    return ATtrue;
  }
  return ATfalse;
}

/**
 * Assert whether a PERR_Area has a begin-line. 
 * \param[in] arg input PERR_Area
 * \return ATtrue if the PERR_Area had a begin-line, or ATfalse otherwise
 */
ATbool PERR_hasAreaBeginLine(PERR_Area arg) {
  if (PERR_isAreaArea(arg)) {
    return ATtrue;
  }
  return ATfalse;
}

/**
 * Assert whether a PERR_Area has a ws-after-begin-line. 
 * \param[in] arg input PERR_Area
 * \return ATtrue if the PERR_Area had a ws-after-begin-line, or ATfalse otherwise
 */
ATbool PERR_hasAreaWsAfterBeginLine(PERR_Area arg) {
  if (PERR_isAreaArea(arg)) {
    return ATtrue;
  }
  return ATfalse;
}

/**
 * Assert whether a PERR_Area has a ws-after-,. 
 * \param[in] arg input PERR_Area
 * \return ATtrue if the PERR_Area had a ws-after-,, or ATfalse otherwise
 */
ATbool PERR_hasAreaWsAfterComma(PERR_Area arg) {
  if (PERR_isAreaArea(arg)) {
    return ATtrue;
  }
  return ATfalse;
}

/**
 * Assert whether a PERR_Area has a begin-column. 
 * \param[in] arg input PERR_Area
 * \return ATtrue if the PERR_Area had a begin-column, or ATfalse otherwise
 */
ATbool PERR_hasAreaBeginColumn(PERR_Area arg) {
  if (PERR_isAreaArea(arg)) {
    return ATtrue;
  }
  return ATfalse;
}

/**
 * Assert whether a PERR_Area has a ws-after-begin-column. 
 * \param[in] arg input PERR_Area
 * \return ATtrue if the PERR_Area had a ws-after-begin-column, or ATfalse otherwise
 */
ATbool PERR_hasAreaWsAfterBeginColumn(PERR_Area arg) {
  if (PERR_isAreaArea(arg)) {
    return ATtrue;
  }
  return ATfalse;
}

/**
 * Assert whether a PERR_Area has a ws-after-,-1. 
 * \param[in] arg input PERR_Area
 * \return ATtrue if the PERR_Area had a ws-after-,-1, or ATfalse otherwise
 */
ATbool PERR_hasAreaWsAfterComma1(PERR_Area arg) {
  if (PERR_isAreaArea(arg)) {
    return ATtrue;
  }
  return ATfalse;
}

/**
 * Assert whether a PERR_Area has a end-line. 
 * \param[in] arg input PERR_Area
 * \return ATtrue if the PERR_Area had a end-line, or ATfalse otherwise
 */
ATbool PERR_hasAreaEndLine(PERR_Area arg) {
  if (PERR_isAreaArea(arg)) {
    return ATtrue;
  }
  return ATfalse;
}

/**
 * Assert whether a PERR_Area has a ws-after-end-line. 
 * \param[in] arg input PERR_Area
 * \return ATtrue if the PERR_Area had a ws-after-end-line, or ATfalse otherwise
 */
ATbool PERR_hasAreaWsAfterEndLine(PERR_Area arg) {
  if (PERR_isAreaArea(arg)) {
    return ATtrue;
  }
  return ATfalse;
}

/**
 * Assert whether a PERR_Area has a ws-after-,-2. 
 * \param[in] arg input PERR_Area
 * \return ATtrue if the PERR_Area had a ws-after-,-2, or ATfalse otherwise
 */
ATbool PERR_hasAreaWsAfterComma2(PERR_Area arg) {
  if (PERR_isAreaArea(arg)) {
    return ATtrue;
  }
  return ATfalse;
}

/**
 * Assert whether a PERR_Area has a end-column. 
 * \param[in] arg input PERR_Area
 * \return ATtrue if the PERR_Area had a end-column, or ATfalse otherwise
 */
ATbool PERR_hasAreaEndColumn(PERR_Area arg) {
  if (PERR_isAreaArea(arg)) {
    return ATtrue;
  }
  return ATfalse;
}

/**
 * Assert whether a PERR_Area has a ws-after-end-column. 
 * \param[in] arg input PERR_Area
 * \return ATtrue if the PERR_Area had a ws-after-end-column, or ATfalse otherwise
 */
ATbool PERR_hasAreaWsAfterEndColumn(PERR_Area arg) {
  if (PERR_isAreaArea(arg)) {
    return ATtrue;
  }
  return ATfalse;
}

/**
 * Assert whether a PERR_Area has a ws-after-,-3. 
 * \param[in] arg input PERR_Area
 * \return ATtrue if the PERR_Area had a ws-after-,-3, or ATfalse otherwise
 */
ATbool PERR_hasAreaWsAfterComma3(PERR_Area arg) {
  if (PERR_isAreaArea(arg)) {
    return ATtrue;
  }
  return ATfalse;
}

/**
 * Assert whether a PERR_Area has a offset. 
 * \param[in] arg input PERR_Area
 * \return ATtrue if the PERR_Area had a offset, or ATfalse otherwise
 */
ATbool PERR_hasAreaOffset(PERR_Area arg) {
  if (PERR_isAreaArea(arg)) {
    return ATtrue;
  }
  return ATfalse;
}

/**
 * Assert whether a PERR_Area has a ws-after-offset. 
 * \param[in] arg input PERR_Area
 * \return ATtrue if the PERR_Area had a ws-after-offset, or ATfalse otherwise
 */
ATbool PERR_hasAreaWsAfterOffset(PERR_Area arg) {
  if (PERR_isAreaArea(arg)) {
    return ATtrue;
  }
  return ATfalse;
}

/**
 * Assert whether a PERR_Area has a ws-after-,-4. 
 * \param[in] arg input PERR_Area
 * \return ATtrue if the PERR_Area had a ws-after-,-4, or ATfalse otherwise
 */
ATbool PERR_hasAreaWsAfterComma4(PERR_Area arg) {
  if (PERR_isAreaArea(arg)) {
    return ATtrue;
  }
  return ATfalse;
}

/**
 * Assert whether a PERR_Area has a length. 
 * \param[in] arg input PERR_Area
 * \return ATtrue if the PERR_Area had a length, or ATfalse otherwise
 */
ATbool PERR_hasAreaLength(PERR_Area arg) {
  if (PERR_isAreaArea(arg)) {
    return ATtrue;
  }
  return ATfalse;
}

/**
 * Assert whether a PERR_Area has a ws-after-length. 
 * \param[in] arg input PERR_Area
 * \return ATtrue if the PERR_Area had a ws-after-length, or ATfalse otherwise
 */
ATbool PERR_hasAreaWsAfterLength(PERR_Area arg) {
  if (PERR_isAreaArea(arg)) {
    return ATtrue;
  }
  return ATfalse;
}

/**
 * Get the ws-after-area PERR_OptLayout of a PERR_Area. Note that the precondition is that this PERR_Area actually has a ws-after-area
 * \param[in] arg input PERR_Area
 * \return the ws-after-area of #arg, if it exist or an undefined value if it does not
 */
PERR_OptLayout PERR_getAreaWsAfterArea(PERR_Area arg) {
  
    return (PERR_OptLayout)ATelementAt((ATermList)ATgetArgument((ATermAppl)arg, 1), 1);
}

/**
 * Get the ws-after-( PERR_OptLayout of a PERR_Area. Note that the precondition is that this PERR_Area actually has a ws-after-(
 * \param[in] arg input PERR_Area
 * \return the ws-after-( of #arg, if it exist or an undefined value if it does not
 */
PERR_OptLayout PERR_getAreaWsAfterParenOpen(PERR_Area arg) {
  
    return (PERR_OptLayout)ATelementAt((ATermList)ATgetArgument((ATermAppl)arg, 1), 3);
}

/**
 * Get the begin-line PERR_NatCon of a PERR_Area. Note that the precondition is that this PERR_Area actually has a begin-line
 * \param[in] arg input PERR_Area
 * \return the begin-line of #arg, if it exist or an undefined value if it does not
 */
PERR_NatCon PERR_getAreaBeginLine(PERR_Area arg) {
  
    return (PERR_NatCon)ATelementAt((ATermList)ATgetArgument((ATermAppl)arg, 1), 4);
}

/**
 * Get the ws-after-begin-line PERR_OptLayout of a PERR_Area. Note that the precondition is that this PERR_Area actually has a ws-after-begin-line
 * \param[in] arg input PERR_Area
 * \return the ws-after-begin-line of #arg, if it exist or an undefined value if it does not
 */
PERR_OptLayout PERR_getAreaWsAfterBeginLine(PERR_Area arg) {
  
    return (PERR_OptLayout)ATelementAt((ATermList)ATgetArgument((ATermAppl)arg, 1), 5);
}

/**
 * Get the ws-after-, PERR_OptLayout of a PERR_Area. Note that the precondition is that this PERR_Area actually has a ws-after-,
 * \param[in] arg input PERR_Area
 * \return the ws-after-, of #arg, if it exist or an undefined value if it does not
 */
PERR_OptLayout PERR_getAreaWsAfterComma(PERR_Area arg) {
  
    return (PERR_OptLayout)ATelementAt((ATermList)ATgetArgument((ATermAppl)arg, 1), 7);
}

/**
 * Get the begin-column PERR_NatCon of a PERR_Area. Note that the precondition is that this PERR_Area actually has a begin-column
 * \param[in] arg input PERR_Area
 * \return the begin-column of #arg, if it exist or an undefined value if it does not
 */
PERR_NatCon PERR_getAreaBeginColumn(PERR_Area arg) {
  
    return (PERR_NatCon)ATelementAt((ATermList)ATgetArgument((ATermAppl)arg, 1), 8);
}

/**
 * Get the ws-after-begin-column PERR_OptLayout of a PERR_Area. Note that the precondition is that this PERR_Area actually has a ws-after-begin-column
 * \param[in] arg input PERR_Area
 * \return the ws-after-begin-column of #arg, if it exist or an undefined value if it does not
 */
PERR_OptLayout PERR_getAreaWsAfterBeginColumn(PERR_Area arg) {
  
    return (PERR_OptLayout)ATelementAt((ATermList)ATgetArgument((ATermAppl)arg, 1), 9);
}

/**
 * Get the ws-after-,-1 PERR_OptLayout of a PERR_Area. Note that the precondition is that this PERR_Area actually has a ws-after-,-1
 * \param[in] arg input PERR_Area
 * \return the ws-after-,-1 of #arg, if it exist or an undefined value if it does not
 */
PERR_OptLayout PERR_getAreaWsAfterComma1(PERR_Area arg) {
  
    return (PERR_OptLayout)ATelementAt((ATermList)ATgetArgument((ATermAppl)arg, 1), 11);
}

/**
 * Get the end-line PERR_NatCon of a PERR_Area. Note that the precondition is that this PERR_Area actually has a end-line
 * \param[in] arg input PERR_Area
 * \return the end-line of #arg, if it exist or an undefined value if it does not
 */
PERR_NatCon PERR_getAreaEndLine(PERR_Area arg) {
  
    return (PERR_NatCon)ATelementAt((ATermList)ATgetArgument((ATermAppl)arg, 1), 12);
}

/**
 * Get the ws-after-end-line PERR_OptLayout of a PERR_Area. Note that the precondition is that this PERR_Area actually has a ws-after-end-line
 * \param[in] arg input PERR_Area
 * \return the ws-after-end-line of #arg, if it exist or an undefined value if it does not
 */
PERR_OptLayout PERR_getAreaWsAfterEndLine(PERR_Area arg) {
  
    return (PERR_OptLayout)ATelementAt((ATermList)ATgetArgument((ATermAppl)arg, 1), 13);
}

/**
 * Get the ws-after-,-2 PERR_OptLayout of a PERR_Area. Note that the precondition is that this PERR_Area actually has a ws-after-,-2
 * \param[in] arg input PERR_Area
 * \return the ws-after-,-2 of #arg, if it exist or an undefined value if it does not
 */
PERR_OptLayout PERR_getAreaWsAfterComma2(PERR_Area arg) {
  
    return (PERR_OptLayout)ATelementAt((ATermList)ATgetArgument((ATermAppl)arg, 1), 15);
}

/**
 * Get the end-column PERR_NatCon of a PERR_Area. Note that the precondition is that this PERR_Area actually has a end-column
 * \param[in] arg input PERR_Area
 * \return the end-column of #arg, if it exist or an undefined value if it does not
 */
PERR_NatCon PERR_getAreaEndColumn(PERR_Area arg) {
  
    return (PERR_NatCon)ATelementAt((ATermList)ATgetArgument((ATermAppl)arg, 1), 16);
}

/**
 * Get the ws-after-end-column PERR_OptLayout of a PERR_Area. Note that the precondition is that this PERR_Area actually has a ws-after-end-column
 * \param[in] arg input PERR_Area
 * \return the ws-after-end-column of #arg, if it exist or an undefined value if it does not
 */
PERR_OptLayout PERR_getAreaWsAfterEndColumn(PERR_Area arg) {
  
    return (PERR_OptLayout)ATelementAt((ATermList)ATgetArgument((ATermAppl)arg, 1), 17);
}

/**
 * Get the ws-after-,-3 PERR_OptLayout of a PERR_Area. Note that the precondition is that this PERR_Area actually has a ws-after-,-3
 * \param[in] arg input PERR_Area
 * \return the ws-after-,-3 of #arg, if it exist or an undefined value if it does not
 */
PERR_OptLayout PERR_getAreaWsAfterComma3(PERR_Area arg) {
  
    return (PERR_OptLayout)ATelementAt((ATermList)ATgetArgument((ATermAppl)arg, 1), 19);
}

/**
 * Get the offset PERR_NatCon of a PERR_Area. Note that the precondition is that this PERR_Area actually has a offset
 * \param[in] arg input PERR_Area
 * \return the offset of #arg, if it exist or an undefined value if it does not
 */
PERR_NatCon PERR_getAreaOffset(PERR_Area arg) {
  
    return (PERR_NatCon)ATelementAt((ATermList)ATgetArgument((ATermAppl)arg, 1), 20);
}

/**
 * Get the ws-after-offset PERR_OptLayout of a PERR_Area. Note that the precondition is that this PERR_Area actually has a ws-after-offset
 * \param[in] arg input PERR_Area
 * \return the ws-after-offset of #arg, if it exist or an undefined value if it does not
 */
PERR_OptLayout PERR_getAreaWsAfterOffset(PERR_Area arg) {
  
    return (PERR_OptLayout)ATelementAt((ATermList)ATgetArgument((ATermAppl)arg, 1), 21);
}

/**
 * Get the ws-after-,-4 PERR_OptLayout of a PERR_Area. Note that the precondition is that this PERR_Area actually has a ws-after-,-4
 * \param[in] arg input PERR_Area
 * \return the ws-after-,-4 of #arg, if it exist or an undefined value if it does not
 */
PERR_OptLayout PERR_getAreaWsAfterComma4(PERR_Area arg) {
  
    return (PERR_OptLayout)ATelementAt((ATermList)ATgetArgument((ATermAppl)arg, 1), 23);
}

/**
 * Get the length PERR_NatCon of a PERR_Area. Note that the precondition is that this PERR_Area actually has a length
 * \param[in] arg input PERR_Area
 * \return the length of #arg, if it exist or an undefined value if it does not
 */
PERR_NatCon PERR_getAreaLength(PERR_Area arg) {
  
    return (PERR_NatCon)ATelementAt((ATermList)ATgetArgument((ATermAppl)arg, 1), 24);
}

/**
 * Get the ws-after-length PERR_OptLayout of a PERR_Area. Note that the precondition is that this PERR_Area actually has a ws-after-length
 * \param[in] arg input PERR_Area
 * \return the ws-after-length of #arg, if it exist or an undefined value if it does not
 */
PERR_OptLayout PERR_getAreaWsAfterLength(PERR_Area arg) {
  
    return (PERR_OptLayout)ATelementAt((ATermList)ATgetArgument((ATermAppl)arg, 1), 25);
}

/**
 * Set the ws-after-area of a PERR_Area. The precondition being that this PERR_Area actually has a ws-after-area
 * \param[in] arg input PERR_Area
 * \param[in] wsAfterArea new PERR_OptLayout to set in #arg
 * \return A new PERR_Area with wsAfterArea at the right place, or a core dump if #arg did not have a wsAfterArea
 */
PERR_Area PERR_setAreaWsAfterArea(PERR_Area arg, PERR_OptLayout wsAfterArea) {
  if (PERR_isAreaArea(arg)) {
    return (PERR_Area)ATsetArgument((ATermAppl)arg, (ATerm)ATreplace((ATermList)ATgetArgument((ATermAppl)arg, 1), (ATerm)((ATerm) wsAfterArea), 1), 1);
  }

  ATabort("Area has no WsAfterArea: %t\n", arg);
  return (PERR_Area)NULL;
}

/**
 * Set the ws-after-( of a PERR_Area. The precondition being that this PERR_Area actually has a ws-after-(
 * \param[in] arg input PERR_Area
 * \param[in] wsAfterParenOpen new PERR_OptLayout to set in #arg
 * \return A new PERR_Area with wsAfterParenOpen at the right place, or a core dump if #arg did not have a wsAfterParenOpen
 */
PERR_Area PERR_setAreaWsAfterParenOpen(PERR_Area arg, PERR_OptLayout wsAfterParenOpen) {
  if (PERR_isAreaArea(arg)) {
    return (PERR_Area)ATsetArgument((ATermAppl)arg, (ATerm)ATreplace((ATermList)ATgetArgument((ATermAppl)arg, 1), (ATerm)((ATerm) wsAfterParenOpen), 3), 1);
  }

  ATabort("Area has no WsAfterParenOpen: %t\n", arg);
  return (PERR_Area)NULL;
}

/**
 * Set the begin-line of a PERR_Area. The precondition being that this PERR_Area actually has a begin-line
 * \param[in] arg input PERR_Area
 * \param[in] beginLine new PERR_NatCon to set in #arg
 * \return A new PERR_Area with beginLine at the right place, or a core dump if #arg did not have a beginLine
 */
PERR_Area PERR_setAreaBeginLine(PERR_Area arg, PERR_NatCon beginLine) {
  if (PERR_isAreaArea(arg)) {
    return (PERR_Area)ATsetArgument((ATermAppl)arg, (ATerm)ATreplace((ATermList)ATgetArgument((ATermAppl)arg, 1), (ATerm)((ATerm) beginLine), 4), 1);
  }

  ATabort("Area has no BeginLine: %t\n", arg);
  return (PERR_Area)NULL;
}

/**
 * Set the ws-after-begin-line of a PERR_Area. The precondition being that this PERR_Area actually has a ws-after-begin-line
 * \param[in] arg input PERR_Area
 * \param[in] wsAfterBeginLine new PERR_OptLayout to set in #arg
 * \return A new PERR_Area with wsAfterBeginLine at the right place, or a core dump if #arg did not have a wsAfterBeginLine
 */
PERR_Area PERR_setAreaWsAfterBeginLine(PERR_Area arg, PERR_OptLayout wsAfterBeginLine) {
  if (PERR_isAreaArea(arg)) {
    return (PERR_Area)ATsetArgument((ATermAppl)arg, (ATerm)ATreplace((ATermList)ATgetArgument((ATermAppl)arg, 1), (ATerm)((ATerm) wsAfterBeginLine), 5), 1);
  }

  ATabort("Area has no WsAfterBeginLine: %t\n", arg);
  return (PERR_Area)NULL;
}

/**
 * Set the ws-after-, of a PERR_Area. The precondition being that this PERR_Area actually has a ws-after-,
 * \param[in] arg input PERR_Area
 * \param[in] wsAfterComma new PERR_OptLayout to set in #arg
 * \return A new PERR_Area with wsAfterComma at the right place, or a core dump if #arg did not have a wsAfterComma
 */
PERR_Area PERR_setAreaWsAfterComma(PERR_Area arg, PERR_OptLayout wsAfterComma) {
  if (PERR_isAreaArea(arg)) {
    return (PERR_Area)ATsetArgument((ATermAppl)arg, (ATerm)ATreplace((ATermList)ATgetArgument((ATermAppl)arg, 1), (ATerm)((ATerm) wsAfterComma), 7), 1);
  }

  ATabort("Area has no WsAfterComma: %t\n", arg);
  return (PERR_Area)NULL;
}

/**
 * Set the begin-column of a PERR_Area. The precondition being that this PERR_Area actually has a begin-column
 * \param[in] arg input PERR_Area
 * \param[in] beginColumn new PERR_NatCon to set in #arg
 * \return A new PERR_Area with beginColumn at the right place, or a core dump if #arg did not have a beginColumn
 */
PERR_Area PERR_setAreaBeginColumn(PERR_Area arg, PERR_NatCon beginColumn) {
  if (PERR_isAreaArea(arg)) {
    return (PERR_Area)ATsetArgument((ATermAppl)arg, (ATerm)ATreplace((ATermList)ATgetArgument((ATermAppl)arg, 1), (ATerm)((ATerm) beginColumn), 8), 1);
  }

  ATabort("Area has no BeginColumn: %t\n", arg);
  return (PERR_Area)NULL;
}

/**
 * Set the ws-after-begin-column of a PERR_Area. The precondition being that this PERR_Area actually has a ws-after-begin-column
 * \param[in] arg input PERR_Area
 * \param[in] wsAfterBeginColumn new PERR_OptLayout to set in #arg
 * \return A new PERR_Area with wsAfterBeginColumn at the right place, or a core dump if #arg did not have a wsAfterBeginColumn
 */
PERR_Area PERR_setAreaWsAfterBeginColumn(PERR_Area arg, PERR_OptLayout wsAfterBeginColumn) {
  if (PERR_isAreaArea(arg)) {
    return (PERR_Area)ATsetArgument((ATermAppl)arg, (ATerm)ATreplace((ATermList)ATgetArgument((ATermAppl)arg, 1), (ATerm)((ATerm) wsAfterBeginColumn), 9), 1);
  }

  ATabort("Area has no WsAfterBeginColumn: %t\n", arg);
  return (PERR_Area)NULL;
}

/**
 * Set the ws-after-,-1 of a PERR_Area. The precondition being that this PERR_Area actually has a ws-after-,-1
 * \param[in] arg input PERR_Area
 * \param[in] wsAfterComma1 new PERR_OptLayout to set in #arg
 * \return A new PERR_Area with wsAfterComma1 at the right place, or a core dump if #arg did not have a wsAfterComma1
 */
PERR_Area PERR_setAreaWsAfterComma1(PERR_Area arg, PERR_OptLayout wsAfterComma1) {
  if (PERR_isAreaArea(arg)) {
    return (PERR_Area)ATsetArgument((ATermAppl)arg, (ATerm)ATreplace((ATermList)ATgetArgument((ATermAppl)arg, 1), (ATerm)((ATerm) wsAfterComma1), 11), 1);
  }

  ATabort("Area has no WsAfterComma1: %t\n", arg);
  return (PERR_Area)NULL;
}

/**
 * Set the end-line of a PERR_Area. The precondition being that this PERR_Area actually has a end-line
 * \param[in] arg input PERR_Area
 * \param[in] endLine new PERR_NatCon to set in #arg
 * \return A new PERR_Area with endLine at the right place, or a core dump if #arg did not have a endLine
 */
PERR_Area PERR_setAreaEndLine(PERR_Area arg, PERR_NatCon endLine) {
  if (PERR_isAreaArea(arg)) {
    return (PERR_Area)ATsetArgument((ATermAppl)arg, (ATerm)ATreplace((ATermList)ATgetArgument((ATermAppl)arg, 1), (ATerm)((ATerm) endLine), 12), 1);
  }

  ATabort("Area has no EndLine: %t\n", arg);
  return (PERR_Area)NULL;
}

/**
 * Set the ws-after-end-line of a PERR_Area. The precondition being that this PERR_Area actually has a ws-after-end-line
 * \param[in] arg input PERR_Area
 * \param[in] wsAfterEndLine new PERR_OptLayout to set in #arg
 * \return A new PERR_Area with wsAfterEndLine at the right place, or a core dump if #arg did not have a wsAfterEndLine
 */
PERR_Area PERR_setAreaWsAfterEndLine(PERR_Area arg, PERR_OptLayout wsAfterEndLine) {
  if (PERR_isAreaArea(arg)) {
    return (PERR_Area)ATsetArgument((ATermAppl)arg, (ATerm)ATreplace((ATermList)ATgetArgument((ATermAppl)arg, 1), (ATerm)((ATerm) wsAfterEndLine), 13), 1);
  }

  ATabort("Area has no WsAfterEndLine: %t\n", arg);
  return (PERR_Area)NULL;
}

/**
 * Set the ws-after-,-2 of a PERR_Area. The precondition being that this PERR_Area actually has a ws-after-,-2
 * \param[in] arg input PERR_Area
 * \param[in] wsAfterComma2 new PERR_OptLayout to set in #arg
 * \return A new PERR_Area with wsAfterComma2 at the right place, or a core dump if #arg did not have a wsAfterComma2
 */
PERR_Area PERR_setAreaWsAfterComma2(PERR_Area arg, PERR_OptLayout wsAfterComma2) {
  if (PERR_isAreaArea(arg)) {
    return (PERR_Area)ATsetArgument((ATermAppl)arg, (ATerm)ATreplace((ATermList)ATgetArgument((ATermAppl)arg, 1), (ATerm)((ATerm) wsAfterComma2), 15), 1);
  }

  ATabort("Area has no WsAfterComma2: %t\n", arg);
  return (PERR_Area)NULL;
}

/**
 * Set the end-column of a PERR_Area. The precondition being that this PERR_Area actually has a end-column
 * \param[in] arg input PERR_Area
 * \param[in] endColumn new PERR_NatCon to set in #arg
 * \return A new PERR_Area with endColumn at the right place, or a core dump if #arg did not have a endColumn
 */
PERR_Area PERR_setAreaEndColumn(PERR_Area arg, PERR_NatCon endColumn) {
  if (PERR_isAreaArea(arg)) {
    return (PERR_Area)ATsetArgument((ATermAppl)arg, (ATerm)ATreplace((ATermList)ATgetArgument((ATermAppl)arg, 1), (ATerm)((ATerm) endColumn), 16), 1);
  }

  ATabort("Area has no EndColumn: %t\n", arg);
  return (PERR_Area)NULL;
}

/**
 * Set the ws-after-end-column of a PERR_Area. The precondition being that this PERR_Area actually has a ws-after-end-column
 * \param[in] arg input PERR_Area
 * \param[in] wsAfterEndColumn new PERR_OptLayout to set in #arg
 * \return A new PERR_Area with wsAfterEndColumn at the right place, or a core dump if #arg did not have a wsAfterEndColumn
 */
PERR_Area PERR_setAreaWsAfterEndColumn(PERR_Area arg, PERR_OptLayout wsAfterEndColumn) {
  if (PERR_isAreaArea(arg)) {
    return (PERR_Area)ATsetArgument((ATermAppl)arg, (ATerm)ATreplace((ATermList)ATgetArgument((ATermAppl)arg, 1), (ATerm)((ATerm) wsAfterEndColumn), 17), 1);
  }

  ATabort("Area has no WsAfterEndColumn: %t\n", arg);
  return (PERR_Area)NULL;
}

/**
 * Set the ws-after-,-3 of a PERR_Area. The precondition being that this PERR_Area actually has a ws-after-,-3
 * \param[in] arg input PERR_Area
 * \param[in] wsAfterComma3 new PERR_OptLayout to set in #arg
 * \return A new PERR_Area with wsAfterComma3 at the right place, or a core dump if #arg did not have a wsAfterComma3
 */
PERR_Area PERR_setAreaWsAfterComma3(PERR_Area arg, PERR_OptLayout wsAfterComma3) {
  if (PERR_isAreaArea(arg)) {
    return (PERR_Area)ATsetArgument((ATermAppl)arg, (ATerm)ATreplace((ATermList)ATgetArgument((ATermAppl)arg, 1), (ATerm)((ATerm) wsAfterComma3), 19), 1);
  }

  ATabort("Area has no WsAfterComma3: %t\n", arg);
  return (PERR_Area)NULL;
}

/**
 * Set the offset of a PERR_Area. The precondition being that this PERR_Area actually has a offset
 * \param[in] arg input PERR_Area
 * \param[in] offset new PERR_NatCon to set in #arg
 * \return A new PERR_Area with offset at the right place, or a core dump if #arg did not have a offset
 */
PERR_Area PERR_setAreaOffset(PERR_Area arg, PERR_NatCon offset) {
  if (PERR_isAreaArea(arg)) {
    return (PERR_Area)ATsetArgument((ATermAppl)arg, (ATerm)ATreplace((ATermList)ATgetArgument((ATermAppl)arg, 1), (ATerm)((ATerm) offset), 20), 1);
  }

  ATabort("Area has no Offset: %t\n", arg);
  return (PERR_Area)NULL;
}

/**
 * Set the ws-after-offset of a PERR_Area. The precondition being that this PERR_Area actually has a ws-after-offset
 * \param[in] arg input PERR_Area
 * \param[in] wsAfterOffset new PERR_OptLayout to set in #arg
 * \return A new PERR_Area with wsAfterOffset at the right place, or a core dump if #arg did not have a wsAfterOffset
 */
PERR_Area PERR_setAreaWsAfterOffset(PERR_Area arg, PERR_OptLayout wsAfterOffset) {
  if (PERR_isAreaArea(arg)) {
    return (PERR_Area)ATsetArgument((ATermAppl)arg, (ATerm)ATreplace((ATermList)ATgetArgument((ATermAppl)arg, 1), (ATerm)((ATerm) wsAfterOffset), 21), 1);
  }

  ATabort("Area has no WsAfterOffset: %t\n", arg);
  return (PERR_Area)NULL;
}

/**
 * Set the ws-after-,-4 of a PERR_Area. The precondition being that this PERR_Area actually has a ws-after-,-4
 * \param[in] arg input PERR_Area
 * \param[in] wsAfterComma4 new PERR_OptLayout to set in #arg
 * \return A new PERR_Area with wsAfterComma4 at the right place, or a core dump if #arg did not have a wsAfterComma4
 */
PERR_Area PERR_setAreaWsAfterComma4(PERR_Area arg, PERR_OptLayout wsAfterComma4) {
  if (PERR_isAreaArea(arg)) {
    return (PERR_Area)ATsetArgument((ATermAppl)arg, (ATerm)ATreplace((ATermList)ATgetArgument((ATermAppl)arg, 1), (ATerm)((ATerm) wsAfterComma4), 23), 1);
  }

  ATabort("Area has no WsAfterComma4: %t\n", arg);
  return (PERR_Area)NULL;
}

/**
 * Set the length of a PERR_Area. The precondition being that this PERR_Area actually has a length
 * \param[in] arg input PERR_Area
 * \param[in] length new PERR_NatCon to set in #arg
 * \return A new PERR_Area with length at the right place, or a core dump if #arg did not have a length
 */
PERR_Area PERR_setAreaLength(PERR_Area arg, PERR_NatCon length) {
  if (PERR_isAreaArea(arg)) {
    return (PERR_Area)ATsetArgument((ATermAppl)arg, (ATerm)ATreplace((ATermList)ATgetArgument((ATermAppl)arg, 1), (ATerm)((ATerm) length), 24), 1);
  }

  ATabort("Area has no Length: %t\n", arg);
  return (PERR_Area)NULL;
}

/**
 * Set the ws-after-length of a PERR_Area. The precondition being that this PERR_Area actually has a ws-after-length
 * \param[in] arg input PERR_Area
 * \param[in] wsAfterLength new PERR_OptLayout to set in #arg
 * \return A new PERR_Area with wsAfterLength at the right place, or a core dump if #arg did not have a wsAfterLength
 */
PERR_Area PERR_setAreaWsAfterLength(PERR_Area arg, PERR_OptLayout wsAfterLength) {
  if (PERR_isAreaArea(arg)) {
    return (PERR_Area)ATsetArgument((ATermAppl)arg, (ATerm)ATreplace((ATermList)ATgetArgument((ATermAppl)arg, 1), (ATerm)((ATerm) wsAfterLength), 25), 1);
  }

  ATabort("Area has no WsAfterLength: %t\n", arg);
  return (PERR_Area)NULL;
}

/**
 * Assert whether a PERR_Slice is any of the valid alternatives, or not. This analysis does not go any deeper than the top level
 * \param[in] arg input PERR_Slice
 * \return ATtrue if #arg corresponds to the expected signature, or ATfalse otherwise
 */
ATbool PERR_isValidSlice(PERR_Slice arg) {
  if (PERR_isSliceSlice(arg)) {
    return ATtrue;
  }
  return ATfalse;
}

/**
 * Assert whether a PERR_Slice is a slice by checking against the following ATerm pattern: appl(prod([lit("slice"),cf(opt(layout)),lit("("),cf(opt(layout)),cf(sort("StrCon")),cf(opt(layout)),lit(","),cf(opt(layout)),lit("["),cf(opt(layout)),cf(iter-star-sep(sort("Area"),lit(","))),cf(opt(layout)),lit("]"),cf(opt(layout)),lit(")")],cf(sort("Slice")),attrs([term(cons("slice"))])),[appl(prod([char-class([115]),char-class([108]),char-class([105]),char-class([99]),char-class([101])],lit("slice"),no-attrs),[115,108,105,99,101]),<"ws-after-slice"("OptLayout")>,appl(prod([char-class([40])],lit("("),no-attrs),[40]),<"ws-after-("("OptLayout")>,<"id"("StrCon")>,<"ws-after-id"("OptLayout")>,appl(prod([char-class([44])],lit(","),no-attrs),[44]),<"ws-after-,"("OptLayout")>,appl(prod([char-class([91])],lit("["),no-attrs),[91]),<"ws-after-["("OptLayout")>,appl(list(cf(iter-star-sep(sort("Area"),lit(",")))),<"areas"("Area-areas")>),<"ws-after-areas"("OptLayout")>,appl(prod([char-class([93])],lit("]"),no-attrs),[93]),<"ws-after-]"("OptLayout")>,appl(prod([char-class([41])],lit(")"),no-attrs),[41])]). Always returns ATtrue
 * \param[in] arg input PERR_Slice
 * \return ATtrue if #arg corresponds to the signature of a slice, or ATfalse otherwise
 */
inline ATbool PERR_isSliceSlice(PERR_Slice arg){
  /* checking for: appl */
  if (ATgetType((ATerm)arg) == AT_APPL && ATgetAFun((ATermAppl)arg) == PERR_afun0) {
    ATerm arg_arg0 = ATgetArgument(arg, 0);
    /* checking for: prod */
    if (ATgetType((ATerm)arg_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0) == PERR_afun1) {
      ATerm arg_arg0_arg0 = ATgetArgument(arg_arg0, 0);
      if (ATgetType((ATerm)arg_arg0_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0) == ATfalse) {
        ATerm arg_arg0_arg0_head;
        ATermList arg_arg0_arg0_list = (ATermList)arg_arg0_arg0;
        arg_arg0_arg0_head = ATgetFirst(arg_arg0_arg0_list);
        arg_arg0_arg0_list = ATgetNext(arg_arg0_arg0_list);
        /* checking for: lit */
        if (ATgetType((ATerm)arg_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head) == PERR_afun9) {
          ATerm arg_arg0_arg0_head_arg0 = ATgetArgument(arg_arg0_arg0_head, 0);
          /* checking for: slice */
          if (ATgetType((ATerm)arg_arg0_arg0_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0) == PERR_afun57) {
            if (ATgetType((ATerm)arg_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0_list) == ATfalse) {
              arg_arg0_arg0_head = ATgetFirst(arg_arg0_arg0_list);
              arg_arg0_arg0_list = ATgetNext(arg_arg0_arg0_list);
              /* checking for: cf */
              if (ATgetType((ATerm)arg_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head) == PERR_afun2) {
                ATerm arg_arg0_arg0_head_arg0 = ATgetArgument(arg_arg0_arg0_head, 0);
                /* checking for: opt */
                if (ATgetType((ATerm)arg_arg0_arg0_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0) == PERR_afun3) {
                  ATerm arg_arg0_arg0_head_arg0_arg0 = ATgetArgument(arg_arg0_arg0_head_arg0, 0);
                  /* checking for: layout */
                  if (ATgetType((ATerm)arg_arg0_arg0_head_arg0_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0_arg0) == PERR_afun4) {
                    if (ATgetType((ATerm)arg_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0_list) == ATfalse) {
                      arg_arg0_arg0_head = ATgetFirst(arg_arg0_arg0_list);
                      arg_arg0_arg0_list = ATgetNext(arg_arg0_arg0_list);
                      /* checking for: lit */
                      if (ATgetType((ATerm)arg_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head) == PERR_afun9) {
                        ATerm arg_arg0_arg0_head_arg0 = ATgetArgument(arg_arg0_arg0_head, 0);
                        /* checking for: ( */
                        if (ATgetType((ATerm)arg_arg0_arg0_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0) == PERR_afun39) {
                          if (ATgetType((ATerm)arg_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0_list) == ATfalse) {
                            arg_arg0_arg0_head = ATgetFirst(arg_arg0_arg0_list);
                            arg_arg0_arg0_list = ATgetNext(arg_arg0_arg0_list);
                            /* checking for: cf */
                            if (ATgetType((ATerm)arg_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head) == PERR_afun2) {
                              ATerm arg_arg0_arg0_head_arg0 = ATgetArgument(arg_arg0_arg0_head, 0);
                              /* checking for: opt */
                              if (ATgetType((ATerm)arg_arg0_arg0_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0) == PERR_afun3) {
                                ATerm arg_arg0_arg0_head_arg0_arg0 = ATgetArgument(arg_arg0_arg0_head_arg0, 0);
                                /* checking for: layout */
                                if (ATgetType((ATerm)arg_arg0_arg0_head_arg0_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0_arg0) == PERR_afun4) {
                                  if (ATgetType((ATerm)arg_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0_list) == ATfalse) {
                                    arg_arg0_arg0_head = ATgetFirst(arg_arg0_arg0_list);
                                    arg_arg0_arg0_list = ATgetNext(arg_arg0_arg0_list);
                                    /* checking for: cf */
                                    if (ATgetType((ATerm)arg_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head) == PERR_afun2) {
                                      ATerm arg_arg0_arg0_head_arg0 = ATgetArgument(arg_arg0_arg0_head, 0);
                                      /* checking for: sort */
                                      if (ATgetType((ATerm)arg_arg0_arg0_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0) == PERR_afun11) {
                                        ATerm arg_arg0_arg0_head_arg0_arg0 = ATgetArgument(arg_arg0_arg0_head_arg0, 0);
                                        /* checking for: StrCon */
                                        if (ATgetType((ATerm)arg_arg0_arg0_head_arg0_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0_arg0) == PERR_afun29) {
                                          if (ATgetType((ATerm)arg_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0_list) == ATfalse) {
                                            arg_arg0_arg0_head = ATgetFirst(arg_arg0_arg0_list);
                                            arg_arg0_arg0_list = ATgetNext(arg_arg0_arg0_list);
                                            /* checking for: cf */
                                            if (ATgetType((ATerm)arg_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head) == PERR_afun2) {
                                              ATerm arg_arg0_arg0_head_arg0 = ATgetArgument(arg_arg0_arg0_head, 0);
                                              /* checking for: opt */
                                              if (ATgetType((ATerm)arg_arg0_arg0_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0) == PERR_afun3) {
                                                ATerm arg_arg0_arg0_head_arg0_arg0 = ATgetArgument(arg_arg0_arg0_head_arg0, 0);
                                                /* checking for: layout */
                                                if (ATgetType((ATerm)arg_arg0_arg0_head_arg0_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0_arg0) == PERR_afun4) {
                                                  if (ATgetType((ATerm)arg_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0_list) == ATfalse) {
                                                    arg_arg0_arg0_head = ATgetFirst(arg_arg0_arg0_list);
                                                    arg_arg0_arg0_list = ATgetNext(arg_arg0_arg0_list);
                                                    /* checking for: lit */
                                                    if (ATgetType((ATerm)arg_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head) == PERR_afun9) {
                                                      ATerm arg_arg0_arg0_head_arg0 = ATgetArgument(arg_arg0_arg0_head, 0);
                                                      /* checking for: , */
                                                      if (ATgetType((ATerm)arg_arg0_arg0_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0) == PERR_afun43) {
                                                        if (ATgetType((ATerm)arg_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0_list) == ATfalse) {
                                                          arg_arg0_arg0_head = ATgetFirst(arg_arg0_arg0_list);
                                                          arg_arg0_arg0_list = ATgetNext(arg_arg0_arg0_list);
                                                          /* checking for: cf */
                                                          if (ATgetType((ATerm)arg_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head) == PERR_afun2) {
                                                            ATerm arg_arg0_arg0_head_arg0 = ATgetArgument(arg_arg0_arg0_head, 0);
                                                            /* checking for: opt */
                                                            if (ATgetType((ATerm)arg_arg0_arg0_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0) == PERR_afun3) {
                                                              ATerm arg_arg0_arg0_head_arg0_arg0 = ATgetArgument(arg_arg0_arg0_head_arg0, 0);
                                                              /* checking for: layout */
                                                              if (ATgetType((ATerm)arg_arg0_arg0_head_arg0_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0_arg0) == PERR_afun4) {
                                                                if (ATgetType((ATerm)arg_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0_list) == ATfalse) {
                                                                  arg_arg0_arg0_head = ATgetFirst(arg_arg0_arg0_list);
                                                                  arg_arg0_arg0_list = ATgetNext(arg_arg0_arg0_list);
                                                                  /* checking for: lit */
                                                                  if (ATgetType((ATerm)arg_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head) == PERR_afun9) {
                                                                    ATerm arg_arg0_arg0_head_arg0 = ATgetArgument(arg_arg0_arg0_head, 0);
                                                                    /* checking for: [ */
                                                                    if (ATgetType((ATerm)arg_arg0_arg0_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0) == PERR_afun47) {
                                                                      if (ATgetType((ATerm)arg_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0_list) == ATfalse) {
                                                                        arg_arg0_arg0_head = ATgetFirst(arg_arg0_arg0_list);
                                                                        arg_arg0_arg0_list = ATgetNext(arg_arg0_arg0_list);
                                                                        /* checking for: cf */
                                                                        if (ATgetType((ATerm)arg_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head) == PERR_afun2) {
                                                                          ATerm arg_arg0_arg0_head_arg0 = ATgetArgument(arg_arg0_arg0_head, 0);
                                                                          /* checking for: opt */
                                                                          if (ATgetType((ATerm)arg_arg0_arg0_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0) == PERR_afun3) {
                                                                            ATerm arg_arg0_arg0_head_arg0_arg0 = ATgetArgument(arg_arg0_arg0_head_arg0, 0);
                                                                            /* checking for: layout */
                                                                            if (ATgetType((ATerm)arg_arg0_arg0_head_arg0_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0_arg0) == PERR_afun4) {
                                                                              if (ATgetType((ATerm)arg_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0_list) == ATfalse) {
                                                                                arg_arg0_arg0_head = ATgetFirst(arg_arg0_arg0_list);
                                                                                arg_arg0_arg0_list = ATgetNext(arg_arg0_arg0_list);
                                                                                /* checking for: cf */
                                                                                if (ATgetType((ATerm)arg_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head) == PERR_afun2) {
                                                                                  ATerm arg_arg0_arg0_head_arg0 = ATgetArgument(arg_arg0_arg0_head, 0);
                                                                                  /* checking for: iter-star-sep */
                                                                                  if (ATgetType((ATerm)arg_arg0_arg0_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0) == PERR_afun46) {
                                                                                    ATerm arg_arg0_arg0_head_arg0_arg0 = ATgetArgument(arg_arg0_arg0_head_arg0, 0);
                                                                                    /* checking for: sort */
                                                                                    if (ATgetType((ATerm)arg_arg0_arg0_head_arg0_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0_arg0) == PERR_afun11) {
                                                                                      ATerm arg_arg0_arg0_head_arg0_arg0_arg0 = ATgetArgument(arg_arg0_arg0_head_arg0_arg0, 0);
                                                                                      /* checking for: Area */
                                                                                      if (ATgetType((ATerm)arg_arg0_arg0_head_arg0_arg0_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0_arg0_arg0) == PERR_afun54) {
                                                                                        ATerm arg_arg0_arg0_head_arg0_arg1 = ATgetArgument(arg_arg0_arg0_head_arg0, 1);
                                                                                        /* checking for: lit */
                                                                                        if (ATgetType((ATerm)arg_arg0_arg0_head_arg0_arg1) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0_arg1) == PERR_afun9) {
                                                                                          ATerm arg_arg0_arg0_head_arg0_arg1_arg0 = ATgetArgument(arg_arg0_arg0_head_arg0_arg1, 0);
                                                                                          /* checking for: , */
                                                                                          if (ATgetType((ATerm)arg_arg0_arg0_head_arg0_arg1_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0_arg1_arg0) == PERR_afun43) {
                                                                                            if (ATgetType((ATerm)arg_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0_list) == ATfalse) {
                                                                                              arg_arg0_arg0_head = ATgetFirst(arg_arg0_arg0_list);
                                                                                              arg_arg0_arg0_list = ATgetNext(arg_arg0_arg0_list);
                                                                                              /* checking for: cf */
                                                                                              if (ATgetType((ATerm)arg_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head) == PERR_afun2) {
                                                                                                ATerm arg_arg0_arg0_head_arg0 = ATgetArgument(arg_arg0_arg0_head, 0);
                                                                                                /* checking for: opt */
                                                                                                if (ATgetType((ATerm)arg_arg0_arg0_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0) == PERR_afun3) {
                                                                                                  ATerm arg_arg0_arg0_head_arg0_arg0 = ATgetArgument(arg_arg0_arg0_head_arg0, 0);
                                                                                                  /* checking for: layout */
                                                                                                  if (ATgetType((ATerm)arg_arg0_arg0_head_arg0_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0_arg0) == PERR_afun4) {
                                                                                                    if (ATgetType((ATerm)arg_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0_list) == ATfalse) {
                                                                                                      arg_arg0_arg0_head = ATgetFirst(arg_arg0_arg0_list);
                                                                                                      arg_arg0_arg0_list = ATgetNext(arg_arg0_arg0_list);
                                                                                                      /* checking for: lit */
                                                                                                      if (ATgetType((ATerm)arg_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head) == PERR_afun9) {
                                                                                                        ATerm arg_arg0_arg0_head_arg0 = ATgetArgument(arg_arg0_arg0_head, 0);
                                                                                                        /* checking for: ] */
                                                                                                        if (ATgetType((ATerm)arg_arg0_arg0_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0) == PERR_afun45) {
                                                                                                          if (ATgetType((ATerm)arg_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0_list) == ATfalse) {
                                                                                                            arg_arg0_arg0_head = ATgetFirst(arg_arg0_arg0_list);
                                                                                                            arg_arg0_arg0_list = ATgetNext(arg_arg0_arg0_list);
                                                                                                            /* checking for: cf */
                                                                                                            if (ATgetType((ATerm)arg_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head) == PERR_afun2) {
                                                                                                              ATerm arg_arg0_arg0_head_arg0 = ATgetArgument(arg_arg0_arg0_head, 0);
                                                                                                              /* checking for: opt */
                                                                                                              if (ATgetType((ATerm)arg_arg0_arg0_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0) == PERR_afun3) {
                                                                                                                ATerm arg_arg0_arg0_head_arg0_arg0 = ATgetArgument(arg_arg0_arg0_head_arg0, 0);
                                                                                                                /* checking for: layout */
                                                                                                                if (ATgetType((ATerm)arg_arg0_arg0_head_arg0_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0_arg0) == PERR_afun4) {
                                                                                                                  if (ATgetType((ATerm)arg_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0_list) == ATfalse) {
                                                                                                                    arg_arg0_arg0_head = ATgetFirst(arg_arg0_arg0_list);
                                                                                                                    arg_arg0_arg0_list = ATgetNext(arg_arg0_arg0_list);
                                                                                                                    /* checking for: lit */
                                                                                                                    if (ATgetType((ATerm)arg_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head) == PERR_afun9) {
                                                                                                                      ATerm arg_arg0_arg0_head_arg0 = ATgetArgument(arg_arg0_arg0_head, 0);
                                                                                                                      /* checking for: ) */
                                                                                                                      if (ATgetType((ATerm)arg_arg0_arg0_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg0_head_arg0) == PERR_afun38) {
                                                                                                                        if (ATgetType((ATerm)arg_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg0_list) == ATtrue) {
                                                                                                                          ATerm arg_arg0_arg1 = ATgetArgument(arg_arg0, 1);
                                                                                                                          /* checking for: cf */
                                                                                                                          if (ATgetType((ATerm)arg_arg0_arg1) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg1) == PERR_afun2) {
                                                                                                                            ATerm arg_arg0_arg1_arg0 = ATgetArgument(arg_arg0_arg1, 0);
                                                                                                                            /* checking for: sort */
                                                                                                                            if (ATgetType((ATerm)arg_arg0_arg1_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg1_arg0) == PERR_afun11) {
                                                                                                                              ATerm arg_arg0_arg1_arg0_arg0 = ATgetArgument(arg_arg0_arg1_arg0, 0);
                                                                                                                              /* checking for: Slice */
                                                                                                                              if (ATgetType((ATerm)arg_arg0_arg1_arg0_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg1_arg0_arg0) == PERR_afun58) {
                                                                                                                                ATerm arg_arg0_arg2 = ATgetArgument(arg_arg0, 2);
                                                                                                                                /* checking for: attrs */
                                                                                                                                if (ATgetType((ATerm)arg_arg0_arg2) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg2) == PERR_afun13) {
                                                                                                                                  ATerm arg_arg0_arg2_arg0 = ATgetArgument(arg_arg0_arg2, 0);
                                                                                                                                  if (ATgetType((ATerm)arg_arg0_arg2_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg2_arg0) == ATfalse) {
                                                                                                                                    ATerm arg_arg0_arg2_arg0_head;
                                                                                                                                    ATermList arg_arg0_arg2_arg0_list = (ATermList)arg_arg0_arg2_arg0;
                                                                                                                                    arg_arg0_arg2_arg0_head = ATgetFirst(arg_arg0_arg2_arg0_list);
                                                                                                                                    arg_arg0_arg2_arg0_list = ATgetNext(arg_arg0_arg2_arg0_list);
                                                                                                                                    /* checking for: term */
                                                                                                                                    if (ATgetType((ATerm)arg_arg0_arg2_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg2_arg0_head) == PERR_afun14) {
                                                                                                                                      ATerm arg_arg0_arg2_arg0_head_arg0 = ATgetArgument(arg_arg0_arg2_arg0_head, 0);
                                                                                                                                      /* checking for: cons */
                                                                                                                                      if (ATgetType((ATerm)arg_arg0_arg2_arg0_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg2_arg0_head_arg0) == PERR_afun15) {
                                                                                                                                        ATerm arg_arg0_arg2_arg0_head_arg0_arg0 = ATgetArgument(arg_arg0_arg2_arg0_head_arg0, 0);
                                                                                                                                        /* checking for: slice */
                                                                                                                                        if (ATgetType((ATerm)arg_arg0_arg2_arg0_head_arg0_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg0_arg2_arg0_head_arg0_arg0) == PERR_afun57) {
                                                                                                                                          if (ATgetType((ATerm)arg_arg0_arg2_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg0_arg2_arg0_list) == ATtrue) {
                                                                                                                                            ATerm arg_arg1 = ATgetArgument(arg, 1);
                                                                                                                                            if (ATgetType((ATerm)arg_arg1) == AT_LIST && ATisEmpty((ATermList)arg_arg1) == ATfalse) {
                                                                                                                                              ATerm arg_arg1_head;
                                                                                                                                              ATermList arg_arg1_list = (ATermList)arg_arg1;
                                                                                                                                              arg_arg1_head = ATgetFirst(arg_arg1_list);
                                                                                                                                              arg_arg1_list = ATgetNext(arg_arg1_list);
                                                                                                                                              /* checking for: appl */
                                                                                                                                              if (ATgetType((ATerm)arg_arg1_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head) == PERR_afun0) {
                                                                                                                                                ATerm arg_arg1_head_arg0 = ATgetArgument(arg_arg1_head, 0);
                                                                                                                                                /* checking for: prod */
                                                                                                                                                if (ATgetType((ATerm)arg_arg1_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0) == PERR_afun1) {
                                                                                                                                                  ATerm arg_arg1_head_arg0_arg0 = ATgetArgument(arg_arg1_head_arg0, 0);
                                                                                                                                                  if (ATgetType((ATerm)arg_arg1_head_arg0_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0) == ATfalse) {
                                                                                                                                                    ATerm arg_arg1_head_arg0_arg0_head;
                                                                                                                                                    ATermList arg_arg1_head_arg0_arg0_list = (ATermList)arg_arg1_head_arg0_arg0;
                                                                                                                                                    arg_arg1_head_arg0_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_list);
                                                                                                                                                    arg_arg1_head_arg0_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_list);
                                                                                                                                                    /* checking for: char-class */
                                                                                                                                                    if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg0_head) == PERR_afun17) {
                                                                                                                                                      ATerm arg_arg1_head_arg0_arg0_head_arg0 = ATgetArgument(arg_arg1_head_arg0_arg0_head, 0);
                                                                                                                                                      if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0) == ATfalse) {
                                                                                                                                                        ATerm arg_arg1_head_arg0_arg0_head_arg0_head;
                                                                                                                                                        ATermList arg_arg1_head_arg0_arg0_head_arg0_list = (ATermList)arg_arg1_head_arg0_arg0_head_arg0;
                                                                                                                                                        arg_arg1_head_arg0_arg0_head_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                                                                        arg_arg1_head_arg0_arg0_head_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                                                                        if (ATgetInt((ATermInt)arg_arg1_head_arg0_arg0_head_arg0_head) == 115) {

                                                                                                                                                          if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0_list) == ATtrue) {
                                                                                                                                                            if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_list) == ATfalse) {
                                                                                                                                                              arg_arg1_head_arg0_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_list);
                                                                                                                                                              arg_arg1_head_arg0_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_list);
                                                                                                                                                              /* checking for: char-class */
                                                                                                                                                              if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg0_head) == PERR_afun17) {
                                                                                                                                                                ATerm arg_arg1_head_arg0_arg0_head_arg0 = ATgetArgument(arg_arg1_head_arg0_arg0_head, 0);
                                                                                                                                                                if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0) == ATfalse) {
                                                                                                                                                                  ATerm arg_arg1_head_arg0_arg0_head_arg0_head;
                                                                                                                                                                  ATermList arg_arg1_head_arg0_arg0_head_arg0_list = (ATermList)arg_arg1_head_arg0_arg0_head_arg0;
                                                                                                                                                                  arg_arg1_head_arg0_arg0_head_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                                                                                  arg_arg1_head_arg0_arg0_head_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                                                                                  if (ATgetInt((ATermInt)arg_arg1_head_arg0_arg0_head_arg0_head) == 108) {

                                                                                                                                                                    if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0_list) == ATtrue) {
                                                                                                                                                                      if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_list) == ATfalse) {
                                                                                                                                                                        arg_arg1_head_arg0_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_list);
                                                                                                                                                                        arg_arg1_head_arg0_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_list);
                                                                                                                                                                        /* checking for: char-class */
                                                                                                                                                                        if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg0_head) == PERR_afun17) {
                                                                                                                                                                          ATerm arg_arg1_head_arg0_arg0_head_arg0 = ATgetArgument(arg_arg1_head_arg0_arg0_head, 0);
                                                                                                                                                                          if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0) == ATfalse) {
                                                                                                                                                                            ATerm arg_arg1_head_arg0_arg0_head_arg0_head;
                                                                                                                                                                            ATermList arg_arg1_head_arg0_arg0_head_arg0_list = (ATermList)arg_arg1_head_arg0_arg0_head_arg0;
                                                                                                                                                                            arg_arg1_head_arg0_arg0_head_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                                                                                            arg_arg1_head_arg0_arg0_head_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                                                                                            if (ATgetInt((ATermInt)arg_arg1_head_arg0_arg0_head_arg0_head) == 105) {

                                                                                                                                                                              if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0_list) == ATtrue) {
                                                                                                                                                                                if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_list) == ATfalse) {
                                                                                                                                                                                  arg_arg1_head_arg0_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_list);
                                                                                                                                                                                  arg_arg1_head_arg0_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_list);
                                                                                                                                                                                  /* checking for: char-class */
                                                                                                                                                                                  if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg0_head) == PERR_afun17) {
                                                                                                                                                                                    ATerm arg_arg1_head_arg0_arg0_head_arg0 = ATgetArgument(arg_arg1_head_arg0_arg0_head, 0);
                                                                                                                                                                                    if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0) == ATfalse) {
                                                                                                                                                                                      ATerm arg_arg1_head_arg0_arg0_head_arg0_head;
                                                                                                                                                                                      ATermList arg_arg1_head_arg0_arg0_head_arg0_list = (ATermList)arg_arg1_head_arg0_arg0_head_arg0;
                                                                                                                                                                                      arg_arg1_head_arg0_arg0_head_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                                                                                                      arg_arg1_head_arg0_arg0_head_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                                                                                                      if (ATgetInt((ATermInt)arg_arg1_head_arg0_arg0_head_arg0_head) == 99) {

                                                                                                                                                                                        if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0_list) == ATtrue) {
                                                                                                                                                                                          if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_list) == ATfalse) {
                                                                                                                                                                                            arg_arg1_head_arg0_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_list);
                                                                                                                                                                                            arg_arg1_head_arg0_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_list);
                                                                                                                                                                                            /* checking for: char-class */
                                                                                                                                                                                            if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg0_head) == PERR_afun17) {
                                                                                                                                                                                              ATerm arg_arg1_head_arg0_arg0_head_arg0 = ATgetArgument(arg_arg1_head_arg0_arg0_head, 0);
                                                                                                                                                                                              if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0) == ATfalse) {
                                                                                                                                                                                                ATerm arg_arg1_head_arg0_arg0_head_arg0_head;
                                                                                                                                                                                                ATermList arg_arg1_head_arg0_arg0_head_arg0_list = (ATermList)arg_arg1_head_arg0_arg0_head_arg0;
                                                                                                                                                                                                arg_arg1_head_arg0_arg0_head_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                                                                                                                arg_arg1_head_arg0_arg0_head_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                                                                                                                if (ATgetInt((ATermInt)arg_arg1_head_arg0_arg0_head_arg0_head) == 101) {

                                                                                                                                                                                                  if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0_list) == ATtrue) {
                                                                                                                                                                                                    if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_list) == ATtrue) {
                                                                                                                                                                                                      ATerm arg_arg1_head_arg0_arg1 = ATgetArgument(arg_arg1_head_arg0, 1);
                                                                                                                                                                                                      /* checking for: lit */
                                                                                                                                                                                                      if (ATgetType((ATerm)arg_arg1_head_arg0_arg1) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg1) == PERR_afun9) {
                                                                                                                                                                                                        ATerm arg_arg1_head_arg0_arg1_arg0 = ATgetArgument(arg_arg1_head_arg0_arg1, 0);
                                                                                                                                                                                                        /* checking for: slice */
                                                                                                                                                                                                        if (ATgetType((ATerm)arg_arg1_head_arg0_arg1_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg1_arg0) == PERR_afun57) {
                                                                                                                                                                                                          ATerm arg_arg1_head_arg0_arg2 = ATgetArgument(arg_arg1_head_arg0, 2);
                                                                                                                                                                                                          /* checking for: no-attrs */
                                                                                                                                                                                                          if (ATgetType((ATerm)arg_arg1_head_arg0_arg2) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg2) == PERR_afun5) {
                                                                                                                                                                                                            ATerm arg_arg1_head_arg1 = ATgetArgument(arg_arg1_head, 1);
                                                                                                                                                                                                            if (ATgetType((ATerm)arg_arg1_head_arg1) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg1) == ATfalse) {
                                                                                                                                                                                                              ATerm arg_arg1_head_arg1_head;
                                                                                                                                                                                                              ATermList arg_arg1_head_arg1_list = (ATermList)arg_arg1_head_arg1;
                                                                                                                                                                                                              arg_arg1_head_arg1_head = ATgetFirst(arg_arg1_head_arg1_list);
                                                                                                                                                                                                              arg_arg1_head_arg1_list = ATgetNext(arg_arg1_head_arg1_list);
                                                                                                                                                                                                              if (ATgetInt((ATermInt)arg_arg1_head_arg1_head) == 115) {

                                                                                                                                                                                                                if (ATgetType((ATerm)arg_arg1_head_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg1_list) == ATfalse) {
                                                                                                                                                                                                                  arg_arg1_head_arg1_head = ATgetFirst(arg_arg1_head_arg1_list);
                                                                                                                                                                                                                  arg_arg1_head_arg1_list = ATgetNext(arg_arg1_head_arg1_list);
                                                                                                                                                                                                                  if (ATgetInt((ATermInt)arg_arg1_head_arg1_head) == 108) {

                                                                                                                                                                                                                    if (ATgetType((ATerm)arg_arg1_head_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg1_list) == ATfalse) {
                                                                                                                                                                                                                      arg_arg1_head_arg1_head = ATgetFirst(arg_arg1_head_arg1_list);
                                                                                                                                                                                                                      arg_arg1_head_arg1_list = ATgetNext(arg_arg1_head_arg1_list);
                                                                                                                                                                                                                      if (ATgetInt((ATermInt)arg_arg1_head_arg1_head) == 105) {

                                                                                                                                                                                                                        if (ATgetType((ATerm)arg_arg1_head_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg1_list) == ATfalse) {
                                                                                                                                                                                                                          arg_arg1_head_arg1_head = ATgetFirst(arg_arg1_head_arg1_list);
                                                                                                                                                                                                                          arg_arg1_head_arg1_list = ATgetNext(arg_arg1_head_arg1_list);
                                                                                                                                                                                                                          if (ATgetInt((ATermInt)arg_arg1_head_arg1_head) == 99) {

                                                                                                                                                                                                                            if (ATgetType((ATerm)arg_arg1_head_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg1_list) == ATfalse) {
                                                                                                                                                                                                                              arg_arg1_head_arg1_head = ATgetFirst(arg_arg1_head_arg1_list);
                                                                                                                                                                                                                              arg_arg1_head_arg1_list = ATgetNext(arg_arg1_head_arg1_list);
                                                                                                                                                                                                                              if (ATgetInt((ATermInt)arg_arg1_head_arg1_head) == 101) {

                                                                                                                                                                                                                                if (ATgetType((ATerm)arg_arg1_head_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg1_list) == ATtrue) {
                                                                                                                                                                                                                                  if (ATgetType((ATerm)arg_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_list) == ATfalse) {
                                                                                                                                                                                                                                    arg_arg1_list = ATgetNext(arg_arg1_list);
                                                                                                                                                                                                                                    if (ATgetType((ATerm)arg_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_list) == ATfalse) {
                                                                                                                                                                                                                                      arg_arg1_head = ATgetFirst(arg_arg1_list);
                                                                                                                                                                                                                                      arg_arg1_list = ATgetNext(arg_arg1_list);
                                                                                                                                                                                                                                      /* checking for: appl */
                                                                                                                                                                                                                                      if (ATgetType((ATerm)arg_arg1_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head) == PERR_afun0) {
                                                                                                                                                                                                                                        ATerm arg_arg1_head_arg0 = ATgetArgument(arg_arg1_head, 0);
                                                                                                                                                                                                                                        /* checking for: prod */
                                                                                                                                                                                                                                        if (ATgetType((ATerm)arg_arg1_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0) == PERR_afun1) {
                                                                                                                                                                                                                                          ATerm arg_arg1_head_arg0_arg0 = ATgetArgument(arg_arg1_head_arg0, 0);
                                                                                                                                                                                                                                          if (ATgetType((ATerm)arg_arg1_head_arg0_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0) == ATfalse) {
                                                                                                                                                                                                                                            ATerm arg_arg1_head_arg0_arg0_head;
                                                                                                                                                                                                                                            ATermList arg_arg1_head_arg0_arg0_list = (ATermList)arg_arg1_head_arg0_arg0;
                                                                                                                                                                                                                                            arg_arg1_head_arg0_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_list);
                                                                                                                                                                                                                                            arg_arg1_head_arg0_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_list);
                                                                                                                                                                                                                                            /* checking for: char-class */
                                                                                                                                                                                                                                            if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg0_head) == PERR_afun17) {
                                                                                                                                                                                                                                              ATerm arg_arg1_head_arg0_arg0_head_arg0 = ATgetArgument(arg_arg1_head_arg0_arg0_head, 0);
                                                                                                                                                                                                                                              if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0) == ATfalse) {
                                                                                                                                                                                                                                                ATerm arg_arg1_head_arg0_arg0_head_arg0_head;
                                                                                                                                                                                                                                                ATermList arg_arg1_head_arg0_arg0_head_arg0_list = (ATermList)arg_arg1_head_arg0_arg0_head_arg0;
                                                                                                                                                                                                                                                arg_arg1_head_arg0_arg0_head_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                                                                                                                                                                arg_arg1_head_arg0_arg0_head_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                                                                                                                                                                if (ATgetInt((ATermInt)arg_arg1_head_arg0_arg0_head_arg0_head) == 40) {

                                                                                                                                                                                                                                                  if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0_list) == ATtrue) {
                                                                                                                                                                                                                                                    if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_list) == ATtrue) {
                                                                                                                                                                                                                                                      ATerm arg_arg1_head_arg0_arg1 = ATgetArgument(arg_arg1_head_arg0, 1);
                                                                                                                                                                                                                                                      /* checking for: lit */
                                                                                                                                                                                                                                                      if (ATgetType((ATerm)arg_arg1_head_arg0_arg1) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg1) == PERR_afun9) {
                                                                                                                                                                                                                                                        ATerm arg_arg1_head_arg0_arg1_arg0 = ATgetArgument(arg_arg1_head_arg0_arg1, 0);
                                                                                                                                                                                                                                                        /* checking for: ( */
                                                                                                                                                                                                                                                        if (ATgetType((ATerm)arg_arg1_head_arg0_arg1_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg1_arg0) == PERR_afun39) {
                                                                                                                                                                                                                                                          ATerm arg_arg1_head_arg0_arg2 = ATgetArgument(arg_arg1_head_arg0, 2);
                                                                                                                                                                                                                                                          /* checking for: no-attrs */
                                                                                                                                                                                                                                                          if (ATgetType((ATerm)arg_arg1_head_arg0_arg2) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg2) == PERR_afun5) {
                                                                                                                                                                                                                                                            ATerm arg_arg1_head_arg1 = ATgetArgument(arg_arg1_head, 1);
                                                                                                                                                                                                                                                            if (ATgetType((ATerm)arg_arg1_head_arg1) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg1) == ATfalse) {
                                                                                                                                                                                                                                                              ATerm arg_arg1_head_arg1_head;
                                                                                                                                                                                                                                                              ATermList arg_arg1_head_arg1_list = (ATermList)arg_arg1_head_arg1;
                                                                                                                                                                                                                                                              arg_arg1_head_arg1_head = ATgetFirst(arg_arg1_head_arg1_list);
                                                                                                                                                                                                                                                              arg_arg1_head_arg1_list = ATgetNext(arg_arg1_head_arg1_list);
                                                                                                                                                                                                                                                              if (ATgetInt((ATermInt)arg_arg1_head_arg1_head) == 40) {

                                                                                                                                                                                                                                                                if (ATgetType((ATerm)arg_arg1_head_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg1_list) == ATtrue) {
                                                                                                                                                                                                                                                                  if (ATgetType((ATerm)arg_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_list) == ATfalse) {
                                                                                                                                                                                                                                                                    arg_arg1_list = ATgetNext(arg_arg1_list);
                                                                                                                                                                                                                                                                    if (ATgetType((ATerm)arg_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_list) == ATfalse) {
                                                                                                                                                                                                                                                                      arg_arg1_list = ATgetNext(arg_arg1_list);
                                                                                                                                                                                                                                                                      if (ATgetType((ATerm)arg_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_list) == ATfalse) {
                                                                                                                                                                                                                                                                        arg_arg1_list = ATgetNext(arg_arg1_list);
                                                                                                                                                                                                                                                                        if (ATgetType((ATerm)arg_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_list) == ATfalse) {
                                                                                                                                                                                                                                                                          arg_arg1_head = ATgetFirst(arg_arg1_list);
                                                                                                                                                                                                                                                                          arg_arg1_list = ATgetNext(arg_arg1_list);
                                                                                                                                                                                                                                                                          /* checking for: appl */
                                                                                                                                                                                                                                                                          if (ATgetType((ATerm)arg_arg1_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head) == PERR_afun0) {
                                                                                                                                                                                                                                                                            ATerm arg_arg1_head_arg0 = ATgetArgument(arg_arg1_head, 0);
                                                                                                                                                                                                                                                                            /* checking for: prod */
                                                                                                                                                                                                                                                                            if (ATgetType((ATerm)arg_arg1_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0) == PERR_afun1) {
                                                                                                                                                                                                                                                                              ATerm arg_arg1_head_arg0_arg0 = ATgetArgument(arg_arg1_head_arg0, 0);
                                                                                                                                                                                                                                                                              if (ATgetType((ATerm)arg_arg1_head_arg0_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0) == ATfalse) {
                                                                                                                                                                                                                                                                                ATerm arg_arg1_head_arg0_arg0_head;
                                                                                                                                                                                                                                                                                ATermList arg_arg1_head_arg0_arg0_list = (ATermList)arg_arg1_head_arg0_arg0;
                                                                                                                                                                                                                                                                                arg_arg1_head_arg0_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_list);
                                                                                                                                                                                                                                                                                arg_arg1_head_arg0_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_list);
                                                                                                                                                                                                                                                                                /* checking for: char-class */
                                                                                                                                                                                                                                                                                if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg0_head) == PERR_afun17) {
                                                                                                                                                                                                                                                                                  ATerm arg_arg1_head_arg0_arg0_head_arg0 = ATgetArgument(arg_arg1_head_arg0_arg0_head, 0);
                                                                                                                                                                                                                                                                                  if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0) == ATfalse) {
                                                                                                                                                                                                                                                                                    ATerm arg_arg1_head_arg0_arg0_head_arg0_head;
                                                                                                                                                                                                                                                                                    ATermList arg_arg1_head_arg0_arg0_head_arg0_list = (ATermList)arg_arg1_head_arg0_arg0_head_arg0;
                                                                                                                                                                                                                                                                                    arg_arg1_head_arg0_arg0_head_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                                                                                                                                                                                                    arg_arg1_head_arg0_arg0_head_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                                                                                                                                                                                                    if (ATgetInt((ATermInt)arg_arg1_head_arg0_arg0_head_arg0_head) == 44) {

                                                                                                                                                                                                                                                                                      if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0_list) == ATtrue) {
                                                                                                                                                                                                                                                                                        if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_list) == ATtrue) {
                                                                                                                                                                                                                                                                                          ATerm arg_arg1_head_arg0_arg1 = ATgetArgument(arg_arg1_head_arg0, 1);
                                                                                                                                                                                                                                                                                          /* checking for: lit */
                                                                                                                                                                                                                                                                                          if (ATgetType((ATerm)arg_arg1_head_arg0_arg1) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg1) == PERR_afun9) {
                                                                                                                                                                                                                                                                                            ATerm arg_arg1_head_arg0_arg1_arg0 = ATgetArgument(arg_arg1_head_arg0_arg1, 0);
                                                                                                                                                                                                                                                                                            /* checking for: , */
                                                                                                                                                                                                                                                                                            if (ATgetType((ATerm)arg_arg1_head_arg0_arg1_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg1_arg0) == PERR_afun43) {
                                                                                                                                                                                                                                                                                              ATerm arg_arg1_head_arg0_arg2 = ATgetArgument(arg_arg1_head_arg0, 2);
                                                                                                                                                                                                                                                                                              /* checking for: no-attrs */
                                                                                                                                                                                                                                                                                              if (ATgetType((ATerm)arg_arg1_head_arg0_arg2) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg2) == PERR_afun5) {
                                                                                                                                                                                                                                                                                                ATerm arg_arg1_head_arg1 = ATgetArgument(arg_arg1_head, 1);
                                                                                                                                                                                                                                                                                                if (ATgetType((ATerm)arg_arg1_head_arg1) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg1) == ATfalse) {
                                                                                                                                                                                                                                                                                                  ATerm arg_arg1_head_arg1_head;
                                                                                                                                                                                                                                                                                                  ATermList arg_arg1_head_arg1_list = (ATermList)arg_arg1_head_arg1;
                                                                                                                                                                                                                                                                                                  arg_arg1_head_arg1_head = ATgetFirst(arg_arg1_head_arg1_list);
                                                                                                                                                                                                                                                                                                  arg_arg1_head_arg1_list = ATgetNext(arg_arg1_head_arg1_list);
                                                                                                                                                                                                                                                                                                  if (ATgetInt((ATermInt)arg_arg1_head_arg1_head) == 44) {

                                                                                                                                                                                                                                                                                                    if (ATgetType((ATerm)arg_arg1_head_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg1_list) == ATtrue) {
                                                                                                                                                                                                                                                                                                      if (ATgetType((ATerm)arg_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_list) == ATfalse) {
                                                                                                                                                                                                                                                                                                        arg_arg1_list = ATgetNext(arg_arg1_list);
                                                                                                                                                                                                                                                                                                        if (ATgetType((ATerm)arg_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_list) == ATfalse) {
                                                                                                                                                                                                                                                                                                          arg_arg1_head = ATgetFirst(arg_arg1_list);
                                                                                                                                                                                                                                                                                                          arg_arg1_list = ATgetNext(arg_arg1_list);
                                                                                                                                                                                                                                                                                                          /* checking for: appl */
                                                                                                                                                                                                                                                                                                          if (ATgetType((ATerm)arg_arg1_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head) == PERR_afun0) {
                                                                                                                                                                                                                                                                                                            ATerm arg_arg1_head_arg0 = ATgetArgument(arg_arg1_head, 0);
                                                                                                                                                                                                                                                                                                            /* checking for: prod */
                                                                                                                                                                                                                                                                                                            if (ATgetType((ATerm)arg_arg1_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0) == PERR_afun1) {
                                                                                                                                                                                                                                                                                                              ATerm arg_arg1_head_arg0_arg0 = ATgetArgument(arg_arg1_head_arg0, 0);
                                                                                                                                                                                                                                                                                                              if (ATgetType((ATerm)arg_arg1_head_arg0_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0) == ATfalse) {
                                                                                                                                                                                                                                                                                                                ATerm arg_arg1_head_arg0_arg0_head;
                                                                                                                                                                                                                                                                                                                ATermList arg_arg1_head_arg0_arg0_list = (ATermList)arg_arg1_head_arg0_arg0;
                                                                                                                                                                                                                                                                                                                arg_arg1_head_arg0_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_list);
                                                                                                                                                                                                                                                                                                                arg_arg1_head_arg0_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_list);
                                                                                                                                                                                                                                                                                                                /* checking for: char-class */
                                                                                                                                                                                                                                                                                                                if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg0_head) == PERR_afun17) {
                                                                                                                                                                                                                                                                                                                  ATerm arg_arg1_head_arg0_arg0_head_arg0 = ATgetArgument(arg_arg1_head_arg0_arg0_head, 0);
                                                                                                                                                                                                                                                                                                                  if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0) == ATfalse) {
                                                                                                                                                                                                                                                                                                                    ATerm arg_arg1_head_arg0_arg0_head_arg0_head;
                                                                                                                                                                                                                                                                                                                    ATermList arg_arg1_head_arg0_arg0_head_arg0_list = (ATermList)arg_arg1_head_arg0_arg0_head_arg0;
                                                                                                                                                                                                                                                                                                                    arg_arg1_head_arg0_arg0_head_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                                                                                                                                                                                                                                    arg_arg1_head_arg0_arg0_head_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                                                                                                                                                                                                                                    if (ATgetInt((ATermInt)arg_arg1_head_arg0_arg0_head_arg0_head) == 91) {

                                                                                                                                                                                                                                                                                                                      if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0_list) == ATtrue) {
                                                                                                                                                                                                                                                                                                                        if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_list) == ATtrue) {
                                                                                                                                                                                                                                                                                                                          ATerm arg_arg1_head_arg0_arg1 = ATgetArgument(arg_arg1_head_arg0, 1);
                                                                                                                                                                                                                                                                                                                          /* checking for: lit */
                                                                                                                                                                                                                                                                                                                          if (ATgetType((ATerm)arg_arg1_head_arg0_arg1) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg1) == PERR_afun9) {
                                                                                                                                                                                                                                                                                                                            ATerm arg_arg1_head_arg0_arg1_arg0 = ATgetArgument(arg_arg1_head_arg0_arg1, 0);
                                                                                                                                                                                                                                                                                                                            /* checking for: [ */
                                                                                                                                                                                                                                                                                                                            if (ATgetType((ATerm)arg_arg1_head_arg0_arg1_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg1_arg0) == PERR_afun47) {
                                                                                                                                                                                                                                                                                                                              ATerm arg_arg1_head_arg0_arg2 = ATgetArgument(arg_arg1_head_arg0, 2);
                                                                                                                                                                                                                                                                                                                              /* checking for: no-attrs */
                                                                                                                                                                                                                                                                                                                              if (ATgetType((ATerm)arg_arg1_head_arg0_arg2) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg2) == PERR_afun5) {
                                                                                                                                                                                                                                                                                                                                ATerm arg_arg1_head_arg1 = ATgetArgument(arg_arg1_head, 1);
                                                                                                                                                                                                                                                                                                                                if (ATgetType((ATerm)arg_arg1_head_arg1) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg1) == ATfalse) {
                                                                                                                                                                                                                                                                                                                                  ATerm arg_arg1_head_arg1_head;
                                                                                                                                                                                                                                                                                                                                  ATermList arg_arg1_head_arg1_list = (ATermList)arg_arg1_head_arg1;
                                                                                                                                                                                                                                                                                                                                  arg_arg1_head_arg1_head = ATgetFirst(arg_arg1_head_arg1_list);
                                                                                                                                                                                                                                                                                                                                  arg_arg1_head_arg1_list = ATgetNext(arg_arg1_head_arg1_list);
                                                                                                                                                                                                                                                                                                                                  if (ATgetInt((ATermInt)arg_arg1_head_arg1_head) == 91) {

                                                                                                                                                                                                                                                                                                                                    if (ATgetType((ATerm)arg_arg1_head_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg1_list) == ATtrue) {
                                                                                                                                                                                                                                                                                                                                      if (ATgetType((ATerm)arg_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_list) == ATfalse) {
                                                                                                                                                                                                                                                                                                                                        arg_arg1_list = ATgetNext(arg_arg1_list);
                                                                                                                                                                                                                                                                                                                                        if (ATgetType((ATerm)arg_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_list) == ATfalse) {
                                                                                                                                                                                                                                                                                                                                          arg_arg1_head = ATgetFirst(arg_arg1_list);
                                                                                                                                                                                                                                                                                                                                          arg_arg1_list = ATgetNext(arg_arg1_list);
                                                                                                                                                                                                                                                                                                                                          /* checking for: appl */
                                                                                                                                                                                                                                                                                                                                          if (ATgetType((ATerm)arg_arg1_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head) == PERR_afun0) {
                                                                                                                                                                                                                                                                                                                                            ATerm arg_arg1_head_arg0 = ATgetArgument(arg_arg1_head, 0);
                                                                                                                                                                                                                                                                                                                                            /* checking for: list */
                                                                                                                                                                                                                                                                                                                                            if (ATgetType((ATerm)arg_arg1_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0) == PERR_afun8) {
                                                                                                                                                                                                                                                                                                                                              ATerm arg_arg1_head_arg0_arg0 = ATgetArgument(arg_arg1_head_arg0, 0);
                                                                                                                                                                                                                                                                                                                                              /* checking for: cf */
                                                                                                                                                                                                                                                                                                                                              if (ATgetType((ATerm)arg_arg1_head_arg0_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg0) == PERR_afun2) {
                                                                                                                                                                                                                                                                                                                                                ATerm arg_arg1_head_arg0_arg0_arg0 = ATgetArgument(arg_arg1_head_arg0_arg0, 0);
                                                                                                                                                                                                                                                                                                                                                /* checking for: iter-star-sep */
                                                                                                                                                                                                                                                                                                                                                if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg0_arg0) == PERR_afun46) {
                                                                                                                                                                                                                                                                                                                                                  ATerm arg_arg1_head_arg0_arg0_arg0_arg0 = ATgetArgument(arg_arg1_head_arg0_arg0_arg0, 0);
                                                                                                                                                                                                                                                                                                                                                  /* checking for: sort */
                                                                                                                                                                                                                                                                                                                                                  if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_arg0_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg0_arg0_arg0) == PERR_afun11) {
                                                                                                                                                                                                                                                                                                                                                    ATerm arg_arg1_head_arg0_arg0_arg0_arg0_arg0 = ATgetArgument(arg_arg1_head_arg0_arg0_arg0_arg0, 0);
                                                                                                                                                                                                                                                                                                                                                    /* checking for: Area */
                                                                                                                                                                                                                                                                                                                                                    if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_arg0_arg0_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg0_arg0_arg0_arg0) == PERR_afun54) {
                                                                                                                                                                                                                                                                                                                                                      ATerm arg_arg1_head_arg0_arg0_arg0_arg1 = ATgetArgument(arg_arg1_head_arg0_arg0_arg0, 1);
                                                                                                                                                                                                                                                                                                                                                      /* checking for: lit */
                                                                                                                                                                                                                                                                                                                                                      if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_arg0_arg1) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg0_arg0_arg1) == PERR_afun9) {
                                                                                                                                                                                                                                                                                                                                                        ATerm arg_arg1_head_arg0_arg0_arg0_arg1_arg0 = ATgetArgument(arg_arg1_head_arg0_arg0_arg0_arg1, 0);
                                                                                                                                                                                                                                                                                                                                                        /* checking for: , */
                                                                                                                                                                                                                                                                                                                                                        if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_arg0_arg1_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg0_arg0_arg1_arg0) == PERR_afun43) {
                                                                                                                                                                                                                                                                                                                                                          if (ATgetType((ATerm)arg_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_list) == ATfalse) {
                                                                                                                                                                                                                                                                                                                                                            arg_arg1_list = ATgetNext(arg_arg1_list);
                                                                                                                                                                                                                                                                                                                                                            if (ATgetType((ATerm)arg_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_list) == ATfalse) {
                                                                                                                                                                                                                                                                                                                                                              arg_arg1_head = ATgetFirst(arg_arg1_list);
                                                                                                                                                                                                                                                                                                                                                              arg_arg1_list = ATgetNext(arg_arg1_list);
                                                                                                                                                                                                                                                                                                                                                              /* checking for: appl */
                                                                                                                                                                                                                                                                                                                                                              if (ATgetType((ATerm)arg_arg1_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head) == PERR_afun0) {
                                                                                                                                                                                                                                                                                                                                                                ATerm arg_arg1_head_arg0 = ATgetArgument(arg_arg1_head, 0);
                                                                                                                                                                                                                                                                                                                                                                /* checking for: prod */
                                                                                                                                                                                                                                                                                                                                                                if (ATgetType((ATerm)arg_arg1_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0) == PERR_afun1) {
                                                                                                                                                                                                                                                                                                                                                                  ATerm arg_arg1_head_arg0_arg0 = ATgetArgument(arg_arg1_head_arg0, 0);
                                                                                                                                                                                                                                                                                                                                                                  if (ATgetType((ATerm)arg_arg1_head_arg0_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0) == ATfalse) {
                                                                                                                                                                                                                                                                                                                                                                    ATerm arg_arg1_head_arg0_arg0_head;
                                                                                                                                                                                                                                                                                                                                                                    ATermList arg_arg1_head_arg0_arg0_list = (ATermList)arg_arg1_head_arg0_arg0;
                                                                                                                                                                                                                                                                                                                                                                    arg_arg1_head_arg0_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_list);
                                                                                                                                                                                                                                                                                                                                                                    arg_arg1_head_arg0_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_list);
                                                                                                                                                                                                                                                                                                                                                                    /* checking for: char-class */
                                                                                                                                                                                                                                                                                                                                                                    if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg0_head) == PERR_afun17) {
                                                                                                                                                                                                                                                                                                                                                                      ATerm arg_arg1_head_arg0_arg0_head_arg0 = ATgetArgument(arg_arg1_head_arg0_arg0_head, 0);
                                                                                                                                                                                                                                                                                                                                                                      if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0) == ATfalse) {
                                                                                                                                                                                                                                                                                                                                                                        ATerm arg_arg1_head_arg0_arg0_head_arg0_head;
                                                                                                                                                                                                                                                                                                                                                                        ATermList arg_arg1_head_arg0_arg0_head_arg0_list = (ATermList)arg_arg1_head_arg0_arg0_head_arg0;
                                                                                                                                                                                                                                                                                                                                                                        arg_arg1_head_arg0_arg0_head_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                                                                                                                                                                                                                                                                                        arg_arg1_head_arg0_arg0_head_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                                                                                                                                                                                                                                                                                        if (ATgetInt((ATermInt)arg_arg1_head_arg0_arg0_head_arg0_head) == 93) {

                                                                                                                                                                                                                                                                                                                                                                          if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0_list) == ATtrue) {
                                                                                                                                                                                                                                                                                                                                                                            if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_list) == ATtrue) {
                                                                                                                                                                                                                                                                                                                                                                              ATerm arg_arg1_head_arg0_arg1 = ATgetArgument(arg_arg1_head_arg0, 1);
                                                                                                                                                                                                                                                                                                                                                                              /* checking for: lit */
                                                                                                                                                                                                                                                                                                                                                                              if (ATgetType((ATerm)arg_arg1_head_arg0_arg1) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg1) == PERR_afun9) {
                                                                                                                                                                                                                                                                                                                                                                                ATerm arg_arg1_head_arg0_arg1_arg0 = ATgetArgument(arg_arg1_head_arg0_arg1, 0);
                                                                                                                                                                                                                                                                                                                                                                                /* checking for: ] */
                                                                                                                                                                                                                                                                                                                                                                                if (ATgetType((ATerm)arg_arg1_head_arg0_arg1_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg1_arg0) == PERR_afun45) {
                                                                                                                                                                                                                                                                                                                                                                                  ATerm arg_arg1_head_arg0_arg2 = ATgetArgument(arg_arg1_head_arg0, 2);
                                                                                                                                                                                                                                                                                                                                                                                  /* checking for: no-attrs */
                                                                                                                                                                                                                                                                                                                                                                                  if (ATgetType((ATerm)arg_arg1_head_arg0_arg2) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg2) == PERR_afun5) {
                                                                                                                                                                                                                                                                                                                                                                                    ATerm arg_arg1_head_arg1 = ATgetArgument(arg_arg1_head, 1);
                                                                                                                                                                                                                                                                                                                                                                                    if (ATgetType((ATerm)arg_arg1_head_arg1) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg1) == ATfalse) {
                                                                                                                                                                                                                                                                                                                                                                                      ATerm arg_arg1_head_arg1_head;
                                                                                                                                                                                                                                                                                                                                                                                      ATermList arg_arg1_head_arg1_list = (ATermList)arg_arg1_head_arg1;
                                                                                                                                                                                                                                                                                                                                                                                      arg_arg1_head_arg1_head = ATgetFirst(arg_arg1_head_arg1_list);
                                                                                                                                                                                                                                                                                                                                                                                      arg_arg1_head_arg1_list = ATgetNext(arg_arg1_head_arg1_list);
                                                                                                                                                                                                                                                                                                                                                                                      if (ATgetInt((ATermInt)arg_arg1_head_arg1_head) == 93) {

                                                                                                                                                                                                                                                                                                                                                                                        if (ATgetType((ATerm)arg_arg1_head_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg1_list) == ATtrue) {
                                                                                                                                                                                                                                                                                                                                                                                          if (ATgetType((ATerm)arg_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_list) == ATfalse) {
                                                                                                                                                                                                                                                                                                                                                                                            arg_arg1_list = ATgetNext(arg_arg1_list);
                                                                                                                                                                                                                                                                                                                                                                                            if (ATgetType((ATerm)arg_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_list) == ATfalse) {
                                                                                                                                                                                                                                                                                                                                                                                              arg_arg1_head = ATgetFirst(arg_arg1_list);
                                                                                                                                                                                                                                                                                                                                                                                              arg_arg1_list = ATgetNext(arg_arg1_list);
                                                                                                                                                                                                                                                                                                                                                                                              /* checking for: appl */
                                                                                                                                                                                                                                                                                                                                                                                              if (ATgetType((ATerm)arg_arg1_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head) == PERR_afun0) {
                                                                                                                                                                                                                                                                                                                                                                                                ATerm arg_arg1_head_arg0 = ATgetArgument(arg_arg1_head, 0);
                                                                                                                                                                                                                                                                                                                                                                                                /* checking for: prod */
                                                                                                                                                                                                                                                                                                                                                                                                if (ATgetType((ATerm)arg_arg1_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0) == PERR_afun1) {
                                                                                                                                                                                                                                                                                                                                                                                                  ATerm arg_arg1_head_arg0_arg0 = ATgetArgument(arg_arg1_head_arg0, 0);
                                                                                                                                                                                                                                                                                                                                                                                                  if (ATgetType((ATerm)arg_arg1_head_arg0_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0) == ATfalse) {
                                                                                                                                                                                                                                                                                                                                                                                                    ATerm arg_arg1_head_arg0_arg0_head;
                                                                                                                                                                                                                                                                                                                                                                                                    ATermList arg_arg1_head_arg0_arg0_list = (ATermList)arg_arg1_head_arg0_arg0;
                                                                                                                                                                                                                                                                                                                                                                                                    arg_arg1_head_arg0_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_list);
                                                                                                                                                                                                                                                                                                                                                                                                    arg_arg1_head_arg0_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_list);
                                                                                                                                                                                                                                                                                                                                                                                                    /* checking for: char-class */
                                                                                                                                                                                                                                                                                                                                                                                                    if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg0_head) == PERR_afun17) {
                                                                                                                                                                                                                                                                                                                                                                                                      ATerm arg_arg1_head_arg0_arg0_head_arg0 = ATgetArgument(arg_arg1_head_arg0_arg0_head, 0);
                                                                                                                                                                                                                                                                                                                                                                                                      if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0) == ATfalse) {
                                                                                                                                                                                                                                                                                                                                                                                                        ATerm arg_arg1_head_arg0_arg0_head_arg0_head;
                                                                                                                                                                                                                                                                                                                                                                                                        ATermList arg_arg1_head_arg0_arg0_head_arg0_list = (ATermList)arg_arg1_head_arg0_arg0_head_arg0;
                                                                                                                                                                                                                                                                                                                                                                                                        arg_arg1_head_arg0_arg0_head_arg0_head = ATgetFirst(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                                                                                                                                                                                                                                                                                                                        arg_arg1_head_arg0_arg0_head_arg0_list = ATgetNext(arg_arg1_head_arg0_arg0_head_arg0_list);
                                                                                                                                                                                                                                                                                                                                                                                                        if (ATgetInt((ATermInt)arg_arg1_head_arg0_arg0_head_arg0_head) == 41) {

                                                                                                                                                                                                                                                                                                                                                                                                          if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_head_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_head_arg0_list) == ATtrue) {
                                                                                                                                                                                                                                                                                                                                                                                                            if (ATgetType((ATerm)arg_arg1_head_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg0_arg0_list) == ATtrue) {
                                                                                                                                                                                                                                                                                                                                                                                                              ATerm arg_arg1_head_arg0_arg1 = ATgetArgument(arg_arg1_head_arg0, 1);
                                                                                                                                                                                                                                                                                                                                                                                                              /* checking for: lit */
                                                                                                                                                                                                                                                                                                                                                                                                              if (ATgetType((ATerm)arg_arg1_head_arg0_arg1) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg1) == PERR_afun9) {
                                                                                                                                                                                                                                                                                                                                                                                                                ATerm arg_arg1_head_arg0_arg1_arg0 = ATgetArgument(arg_arg1_head_arg0_arg1, 0);
                                                                                                                                                                                                                                                                                                                                                                                                                /* checking for: ) */
                                                                                                                                                                                                                                                                                                                                                                                                                if (ATgetType((ATerm)arg_arg1_head_arg0_arg1_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg1_arg0) == PERR_afun38) {
                                                                                                                                                                                                                                                                                                                                                                                                                  ATerm arg_arg1_head_arg0_arg2 = ATgetArgument(arg_arg1_head_arg0, 2);
                                                                                                                                                                                                                                                                                                                                                                                                                  /* checking for: no-attrs */
                                                                                                                                                                                                                                                                                                                                                                                                                  if (ATgetType((ATerm)arg_arg1_head_arg0_arg2) == AT_APPL && ATgetAFun((ATermAppl)arg_arg1_head_arg0_arg2) == PERR_afun5) {
                                                                                                                                                                                                                                                                                                                                                                                                                    ATerm arg_arg1_head_arg1 = ATgetArgument(arg_arg1_head, 1);
                                                                                                                                                                                                                                                                                                                                                                                                                    if (ATgetType((ATerm)arg_arg1_head_arg1) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg1) == ATfalse) {
                                                                                                                                                                                                                                                                                                                                                                                                                      ATerm arg_arg1_head_arg1_head;
                                                                                                                                                                                                                                                                                                                                                                                                                      ATermList arg_arg1_head_arg1_list = (ATermList)arg_arg1_head_arg1;
                                                                                                                                                                                                                                                                                                                                                                                                                      arg_arg1_head_arg1_head = ATgetFirst(arg_arg1_head_arg1_list);
                                                                                                                                                                                                                                                                                                                                                                                                                      arg_arg1_head_arg1_list = ATgetNext(arg_arg1_head_arg1_list);
                                                                                                                                                                                                                                                                                                                                                                                                                      if (ATgetInt((ATermInt)arg_arg1_head_arg1_head) == 41) {

                                                                                                                                                                                                                                                                                                                                                                                                                        if (ATgetType((ATerm)arg_arg1_head_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_head_arg1_list) == ATtrue) {
                                                                                                                                                                                                                                                                                                                                                                                                                          if (ATgetType((ATerm)arg_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_arg1_list) == ATtrue) {
                                                                                                                                                                                                                                                                                                                                                                                                                            return ATtrue;
                                                                                                                                                                                                                                                                                                                                                                                                                          }
                                                                                                                                                                                                                                                                                                                                                                                                                        }
                                                                                                                                                                                                                                                                                                                                                                                                                      }
                                                                                                                                                                                                                                                                                                                                                                                                                    }
                                                                                                                                                                                                                                                                                                                                                                                                                  }
                                                                                                                                                                                                                                                                                                                                                                                                                }
                                                                                                                                                                                                                                                                                                                                                                                                              }
                                                                                                                                                                                                                                                                                                                                                                                                            }
                                                                                                                                                                                                                                                                                                                                                                                                          }
                                                                                                                                                                                                                                                                                                                                                                                                        }
                                                                                                                                                                                                                                                                                                                                                                                                      }
                                                                                                                                                                                                                                                                                                                                                                                                    }
                                                                                                                                                                                                                                                                                                                                                                                                  }
                                                                                                                                                                                                                                                                                                                                                                                                }
                                                                                                                                                                                                                                                                                                                                                                                              }
                                                                                                                                                                                                                                                                                                                                                                                            }
                                                                                                                                                                                                                                                                                                                                                                                          }
                                                                                                                                                                                                                                                                                                                                                                                        }
                                                                                                                                                                                                                                                                                                                                                                                      }
                                                                                                                                                                                                                                                                                                                                                                                    }
                                                                                                                                                                                                                                                                                                                                                                                  }
                                                                                                                                                                                                                                                                                                                                                                                }
                                                                                                                                                                                                                                                                                                                                                                              }
                                                                                                                                                                                                                                                                                                                                                                            }
                                                                                                                                                                                                                                                                                                                                                                          }
                                                                                                                                                                                                                                                                                                                                                                        }
                                                                                                                                                                                                                                                                                                                                                                      }
                                                                                                                                                                                                                                                                                                                                                                    }
                                                                                                                                                                                                                                                                                                                                                                  }
                                                                                                                                                                                                                                                                                                                                                                }
                                                                                                                                                                                                                                                                                                                                                              }
                                                                                                                                                                                                                                                                                                                                                            }
                                                                                                                                                                                                                                                                                                                                                          }
                                                                                                                                                                                                                                                                                                                                                        }
                                                                                                                                                                                                                                                                                                                                                      }
                                                                                                                                                                                                                                                                                                                                                    }
                                                                                                                                                                                                                                                                                                                                                  }
                                                                                                                                                                                                                                                                                                                                                }
                                                                                                                                                                                                                                                                                                                                              }
                                                                                                                                                                                                                                                                                                                                            }
                                                                                                                                                                                                                                                                                                                                          }
                                                                                                                                                                                                                                                                                                                                        }
                                                                                                                                                                                                                                                                                                                                      }
                                                                                                                                                                                                                                                                                                                                    }
                                                                                                                                                                                                                                                                                                                                  }
                                                                                                                                                                                                                                                                                                                                }
                                                                                                                                                                                                                                                                                                                              }
                                                                                                                                                                                                                                                                                                                            }
                                                                                                                                                                                                                                                                                                                          }
                                                                                                                                                                                                                                                                                                                        }
                                                                                                                                                                                                                                                                                                                      }
                                                                                                                                                                                                                                                                                                                    }
                                                                                                                                                                                                                                                                                                                  }
                                                                                                                                                                                                                                                                                                                }
                                                                                                                                                                                                                                                                                                              }
                                                                                                                                                                                                                                                                                                            }
                                                                                                                                                                                                                                                                                                          }
                                                                                                                                                                                                                                                                                                        }
                                                                                                                                                                                                                                                                                                      }
                                                                                                                                                                                                                                                                                                    }
                                                                                                                                                                                                                                                                                                  }
                                                                                                                                                                                                                                                                                                }
                                                                                                                                                                                                                                                                                              }
                                                                                                                                                                                                                                                                                            }
                                                                                                                                                                                                                                                                                          }
                                                                                                                                                                                                                                                                                        }
                                                                                                                                                                                                                                                                                      }
                                                                                                                                                                                                                                                                                    }
                                                                                                                                                                                                                                                                                  }
                                                                                                                                                                                                                                                                                }
                                                                                                                                                                                                                                                                              }
                                                                                                                                                                                                                                                                            }
                                                                                                                                                                                                                                                                          }
                                                                                                                                                                                                                                                                        }
                                                                                                                                                                                                                                                                      }
                                                                                                                                                                                                                                                                    }
                                                                                                                                                                                                                                                                  }
                                                                                                                                                                                                                                                                }
                                                                                                                                                                                                                                                              }
                                                                                                                                                                                                                                                            }
                                                                                                                                                                                                                                                          }
                                                                                                                                                                                                                                                        }
                                                                                                                                                                                                                                                      }
                                                                                                                                                                                                                                                    }
                                                                                                                                                                                                                                                  }
                                                                                                                                                                                                                                                }
                                                                                                                                                                                                                                              }
                                                                                                                                                                                                                                            }
                                                                                                                                                                                                                                          }
                                                                                                                                                                                                                                        }
                                                                                                                                                                                                                                      }
                                                                                                                                                                                                                                    }
                                                                                                                                                                                                                                  }
                                                                                                                                                                                                                                }
                                                                                                                                                                                                                              }
                                                                                                                                                                                                                            }
                                                                                                                                                                                                                          }
                                                                                                                                                                                                                        }
                                                                                                                                                                                                                      }
                                                                                                                                                                                                                    }
                                                                                                                                                                                                                  }
                                                                                                                                                                                                                }
                                                                                                                                                                                                              }
                                                                                                                                                                                                            }
                                                                                                                                                                                                          }
                                                                                                                                                                                                        }
                                                                                                                                                                                                      }
                                                                                                                                                                                                    }
                                                                                                                                                                                                  }
                                                                                                                                                                                                }
                                                                                                                                                                                              }
                                                                                                                                                                                            }
                                                                                                                                                                                          }
                                                                                                                                                                                        }
                                                                                                                                                                                      }
                                                                                                                                                                                    }
                                                                                                                                                                                  }
                                                                                                                                                                                }
                                                                                                                                                                              }
                                                                                                                                                                            }
                                                                                                                                                                          }
                                                                                                                                                                        }
                                                                                                                                                                      }
                                                                                                                                                                    }
                                                                                                                                                                  }
                                                                                                                                                                }
                                                                                                                                                              }
                                                                                                                                                            }
                                                                                                                                                          }
                                                                                                                                                        }
                                                                                                                                                      }
                                                                                                                                                    }
                                                                                                                                                  }
                                                                                                                                                }
                                                                                                                                              }
                                                                                                                                            }
                                                                                                                                          }
                                                                                                                                        }
                                                                                                                                      }
                                                                                                                                    }
                                                                                                                                  }
                                                                                                                                }
                                                                                                                              }
                                                                                                                            }
                                                                                                                          }
                                                                                                                        }
                                                                                                                      }
                                                                                                                    }
                                                                                                                  }
                                                                                                                }
                                                                                                              }
                                                                                                            }
                                                                                                          }
                                                                                                        }
                                                                                                      }
                                                                                                    }
                                                                                                  }
                                                                                                }
                                                                                              }
                                                                                            }
                                                                                          }
                                                                                        }
                                                                                      }
                                                                                    }
                                                                                  }
                                                                                }
                                                                              }
                                                                            }
                                                                          }
                                                                        }
                                                                      }
                                                                    }
                                                                  }
                                                                }
                                                              }
                                                            }
                                                          }
                                                        }
                                                      }
                                                    }
                                                  }
                                                }
                                              }
                                            }
                                          }
                                        }
                                      }
                                    }
                                  }
                                }
                              }
                            }
                          }
                        }
                      }
                    }
                  }
                }
              }
            }
          }
        }
      }
    }
  }
  return ATfalse;
}

/**
 * Assert whether a PERR_Slice has a ws-after-slice. 
 * \param[in] arg input PERR_Slice
 * \return ATtrue if the PERR_Slice had a ws-after-slice, or ATfalse otherwise
 */
ATbool PERR_hasSliceWsAfterSlice(PERR_Slice arg) {
  if (PERR_isSliceSlice(arg)) {
    return ATtrue;
  }
  return ATfalse;
}

/**
 * Assert whether a PERR_Slice has a ws-after-(. 
 * \param[in] arg input PERR_Slice
 * \return ATtrue if the PERR_Slice had a ws-after-(, or ATfalse otherwise
 */
ATbool PERR_hasSliceWsAfterParenOpen(PERR_Slice arg) {
  if (PERR_isSliceSlice(arg)) {
    return ATtrue;
  }
  return ATfalse;
}

/**
 * Assert whether a PERR_Slice has a id. 
 * \param[in] arg input PERR_Slice
 * \return ATtrue if the PERR_Slice had a id, or ATfalse otherwise
 */
ATbool PERR_hasSliceId(PERR_Slice arg) {
  if (PERR_isSliceSlice(arg)) {
    return ATtrue;
  }
  return ATfalse;
}

/**
 * Assert whether a PERR_Slice has a ws-after-id. 
 * \param[in] arg input PERR_Slice
 * \return ATtrue if the PERR_Slice had a ws-after-id, or ATfalse otherwise
 */
ATbool PERR_hasSliceWsAfterId(PERR_Slice arg) {
  if (PERR_isSliceSlice(arg)) {
    return ATtrue;
  }
  return ATfalse;
}

/**
 * Assert whether a PERR_Slice has a ws-after-,. 
 * \param[in] arg input PERR_Slice
 * \return ATtrue if the PERR_Slice had a ws-after-,, or ATfalse otherwise
 */
ATbool PERR_hasSliceWsAfterComma(PERR_Slice arg) {
  if (PERR_isSliceSlice(arg)) {
    return ATtrue;
  }
  return ATfalse;
}

/**
 * Assert whether a PERR_Slice has a ws-after-[. 
 * \param[in] arg input PERR_Slice
 * \return ATtrue if the PERR_Slice had a ws-after-[, or ATfalse otherwise
 */
ATbool PERR_hasSliceWsAfterBracketOpen(PERR_Slice arg) {
  if (PERR_isSliceSlice(arg)) {
    return ATtrue;
  }
  return ATfalse;
}

/**
 * Assert whether a PERR_Slice has a areas. 
 * \param[in] arg input PERR_Slice
 * \return ATtrue if the PERR_Slice had a areas, or ATfalse otherwise
 */
ATbool PERR_hasSliceAreas(PERR_Slice arg) {
  if (PERR_isSliceSlice(arg)) {
    return ATtrue;
  }
  return ATfalse;
}

/**
 * Assert whether a PERR_Slice has a ws-after-areas. 
 * \param[in] arg input PERR_Slice
 * \return ATtrue if the PERR_Slice had a ws-after-areas, or ATfalse otherwise
 */
ATbool PERR_hasSliceWsAfterAreas(PERR_Slice arg) {
  if (PERR_isSliceSlice(arg)) {
    return ATtrue;
  }
  return ATfalse;
}

/**
 * Assert whether a PERR_Slice has a ws-after-]. 
 * \param[in] arg input PERR_Slice
 * \return ATtrue if the PERR_Slice had a ws-after-], or ATfalse otherwise
 */
ATbool PERR_hasSliceWsAfterBracketClose(PERR_Slice arg) {
  if (PERR_isSliceSlice(arg)) {
    return ATtrue;
  }
  return ATfalse;
}

/**
 * Get the ws-after-slice PERR_OptLayout of a PERR_Slice. Note that the precondition is that this PERR_Slice actually has a ws-after-slice
 * \param[in] arg input PERR_Slice
 * \return the ws-after-slice of #arg, if it exist or an undefined value if it does not
 */
PERR_OptLayout PERR_getSliceWsAfterSlice(PERR_Slice arg) {
  
    return (PERR_OptLayout)ATelementAt((ATermList)ATgetArgument((ATermAppl)arg, 1), 1);
}

/**
 * Get the ws-after-( PERR_OptLayout of a PERR_Slice. Note that the precondition is that this PERR_Slice actually has a ws-after-(
 * \param[in] arg input PERR_Slice
 * \return the ws-after-( of #arg, if it exist or an undefined value if it does not
 */
PERR_OptLayout PERR_getSliceWsAfterParenOpen(PERR_Slice arg) {
  
    return (PERR_OptLayout)ATelementAt((ATermList)ATgetArgument((ATermAppl)arg, 1), 3);
}

/**
 * Get the id PERR_StrCon of a PERR_Slice. Note that the precondition is that this PERR_Slice actually has a id
 * \param[in] arg input PERR_Slice
 * \return the id of #arg, if it exist or an undefined value if it does not
 */
PERR_StrCon PERR_getSliceId(PERR_Slice arg) {
  
    return (PERR_StrCon)ATelementAt((ATermList)ATgetArgument((ATermAppl)arg, 1), 4);
}

/**
 * Get the ws-after-id PERR_OptLayout of a PERR_Slice. Note that the precondition is that this PERR_Slice actually has a ws-after-id
 * \param[in] arg input PERR_Slice
 * \return the ws-after-id of #arg, if it exist or an undefined value if it does not
 */
PERR_OptLayout PERR_getSliceWsAfterId(PERR_Slice arg) {
  
    return (PERR_OptLayout)ATelementAt((ATermList)ATgetArgument((ATermAppl)arg, 1), 5);
}

/**
 * Get the ws-after-, PERR_OptLayout of a PERR_Slice. Note that the precondition is that this PERR_Slice actually has a ws-after-,
 * \param[in] arg input PERR_Slice
 * \return the ws-after-, of #arg, if it exist or an undefined value if it does not
 */
PERR_OptLayout PERR_getSliceWsAfterComma(PERR_Slice arg) {
  
    return (PERR_OptLayout)ATelementAt((ATermList)ATgetArgument((ATermAppl)arg, 1), 7);
}

/**
 * Get the ws-after-[ PERR_OptLayout of a PERR_Slice. Note that the precondition is that this PERR_Slice actually has a ws-after-[
 * \param[in] arg input PERR_Slice
 * \return the ws-after-[ of #arg, if it exist or an undefined value if it does not
 */
PERR_OptLayout PERR_getSliceWsAfterBracketOpen(PERR_Slice arg) {
  
    return (PERR_OptLayout)ATelementAt((ATermList)ATgetArgument((ATermAppl)arg, 1), 9);
}

/**
 * Get the areas PERR_AreaAreas of a PERR_Slice. Note that the precondition is that this PERR_Slice actually has a areas
 * \param[in] arg input PERR_Slice
 * \return the areas of #arg, if it exist or an undefined value if it does not
 */
PERR_AreaAreas PERR_getSliceAreas(PERR_Slice arg) {
  
    return (PERR_AreaAreas)ATgetArgument((ATermAppl)ATelementAt((ATermList)ATgetArgument((ATermAppl)arg, 1), 10), 1);
}

/**
 * Get the ws-after-areas PERR_OptLayout of a PERR_Slice. Note that the precondition is that this PERR_Slice actually has a ws-after-areas
 * \param[in] arg input PERR_Slice
 * \return the ws-after-areas of #arg, if it exist or an undefined value if it does not
 */
PERR_OptLayout PERR_getSliceWsAfterAreas(PERR_Slice arg) {
  
    return (PERR_OptLayout)ATelementAt((ATermList)ATgetArgument((ATermAppl)arg, 1), 11);
}

/**
 * Get the ws-after-] PERR_OptLayout of a PERR_Slice. Note that the precondition is that this PERR_Slice actually has a ws-after-]
 * \param[in] arg input PERR_Slice
 * \return the ws-after-] of #arg, if it exist or an undefined value if it does not
 */
PERR_OptLayout PERR_getSliceWsAfterBracketClose(PERR_Slice arg) {
  
    return (PERR_OptLayout)ATelementAt((ATermList)ATgetArgument((ATermAppl)arg, 1), 13);
}

/**
 * Set the ws-after-slice of a PERR_Slice. The precondition being that this PERR_Slice actually has a ws-after-slice
 * \param[in] arg input PERR_Slice
 * \param[in] wsAfterSlice new PERR_OptLayout to set in #arg
 * \return A new PERR_Slice with wsAfterSlice at the right place, or a core dump if #arg did not have a wsAfterSlice
 */
PERR_Slice PERR_setSliceWsAfterSlice(PERR_Slice arg, PERR_OptLayout wsAfterSlice) {
  if (PERR_isSliceSlice(arg)) {
    return (PERR_Slice)ATsetArgument((ATermAppl)arg, (ATerm)ATreplace((ATermList)ATgetArgument((ATermAppl)arg, 1), (ATerm)((ATerm) wsAfterSlice), 1), 1);
  }

  ATabort("Slice has no WsAfterSlice: %t\n", arg);
  return (PERR_Slice)NULL;
}

/**
 * Set the ws-after-( of a PERR_Slice. The precondition being that this PERR_Slice actually has a ws-after-(
 * \param[in] arg input PERR_Slice
 * \param[in] wsAfterParenOpen new PERR_OptLayout to set in #arg
 * \return A new PERR_Slice with wsAfterParenOpen at the right place, or a core dump if #arg did not have a wsAfterParenOpen
 */
PERR_Slice PERR_setSliceWsAfterParenOpen(PERR_Slice arg, PERR_OptLayout wsAfterParenOpen) {
  if (PERR_isSliceSlice(arg)) {
    return (PERR_Slice)ATsetArgument((ATermAppl)arg, (ATerm)ATreplace((ATermList)ATgetArgument((ATermAppl)arg, 1), (ATerm)((ATerm) wsAfterParenOpen), 3), 1);
  }

  ATabort("Slice has no WsAfterParenOpen: %t\n", arg);
  return (PERR_Slice)NULL;
}

/**
 * Set the id of a PERR_Slice. The precondition being that this PERR_Slice actually has a id
 * \param[in] arg input PERR_Slice
 * \param[in] id new PERR_StrCon to set in #arg
 * \return A new PERR_Slice with id at the right place, or a core dump if #arg did not have a id
 */
PERR_Slice PERR_setSliceId(PERR_Slice arg, PERR_StrCon id) {
  if (PERR_isSliceSlice(arg)) {
    return (PERR_Slice)ATsetArgument((ATermAppl)arg, (ATerm)ATreplace((ATermList)ATgetArgument((ATermAppl)arg, 1), (ATerm)((ATerm) id), 4), 1);
  }

  ATabort("Slice has no Id: %t\n", arg);
  return (PERR_Slice)NULL;
}

/**
 * Set the ws-after-id of a PERR_Slice. The precondition being that this PERR_Slice actually has a ws-after-id
 * \param[in] arg input PERR_Slice
 * \param[in] wsAfterId new PERR_OptLayout to set in #arg
 * \return A new PERR_Slice with wsAfterId at the right place, or a core dump if #arg did not have a wsAfterId
 */
PERR_Slice PERR_setSliceWsAfterId(PERR_Slice arg, PERR_OptLayout wsAfterId) {
  if (PERR_isSliceSlice(arg)) {
    return (PERR_Slice)ATsetArgument((ATermAppl)arg, (ATerm)ATreplace((ATermList)ATgetArgument((ATermAppl)arg, 1), (ATerm)((ATerm) wsAfterId), 5), 1);
  }

  ATabort("Slice has no WsAfterId: %t\n", arg);
  return (PERR_Slice)NULL;
}

/**
 * Set the ws-after-, of a PERR_Slice. The precondition being that this PERR_Slice actually has a ws-after-,
 * \param[in] arg input PERR_Slice
 * \param[in] wsAfterComma new PERR_OptLayout to set in #arg
 * \return A new PERR_Slice with wsAfterComma at the right place, or a core dump if #arg did not have a wsAfterComma
 */
PERR_Slice PERR_setSliceWsAfterComma(PERR_Slice arg, PERR_OptLayout wsAfterComma) {
  if (PERR_isSliceSlice(arg)) {
    return (PERR_Slice)ATsetArgument((ATermAppl)arg, (ATerm)ATreplace((ATermList)ATgetArgument((ATermAppl)arg, 1), (ATerm)((ATerm) wsAfterComma), 7), 1);
  }

  ATabort("Slice has no WsAfterComma: %t\n", arg);
  return (PERR_Slice)NULL;
}

/**
 * Set the ws-after-[ of a PERR_Slice. The precondition being that this PERR_Slice actually has a ws-after-[
 * \param[in] arg input PERR_Slice
 * \param[in] wsAfterBracketOpen new PERR_OptLayout to set in #arg
 * \return A new PERR_Slice with wsAfterBracketOpen at the right place, or a core dump if #arg did not have a wsAfterBracketOpen
 */
PERR_Slice PERR_setSliceWsAfterBracketOpen(PERR_Slice arg, PERR_OptLayout wsAfterBracketOpen) {
  if (PERR_isSliceSlice(arg)) {
    return (PERR_Slice)ATsetArgument((ATermAppl)arg, (ATerm)ATreplace((ATermList)ATgetArgument((ATermAppl)arg, 1), (ATerm)((ATerm) wsAfterBracketOpen), 9), 1);
  }

  ATabort("Slice has no WsAfterBracketOpen: %t\n", arg);
  return (PERR_Slice)NULL;
}

/**
 * Set the areas of a PERR_Slice. The precondition being that this PERR_Slice actually has a areas
 * \param[in] arg input PERR_Slice
 * \param[in] areas new PERR_AreaAreas to set in #arg
 * \return A new PERR_Slice with areas at the right place, or a core dump if #arg did not have a areas
 */
PERR_Slice PERR_setSliceAreas(PERR_Slice arg, PERR_AreaAreas areas) {
  if (PERR_isSliceSlice(arg)) {
    return (PERR_Slice)ATsetArgument((ATermAppl)arg, (ATerm)ATreplace((ATermList)ATgetArgument((ATermAppl)arg, 1), (ATerm)ATsetArgument((ATermAppl)ATelementAt((ATermList)ATgetArgument((ATermAppl)arg, 1), 10), (ATerm)((ATerm) areas), 1), 10), 1);
  }

  ATabort("Slice has no Areas: %t\n", arg);
  return (PERR_Slice)NULL;
}

/**
 * Set the ws-after-areas of a PERR_Slice. The precondition being that this PERR_Slice actually has a ws-after-areas
 * \param[in] arg input PERR_Slice
 * \param[in] wsAfterAreas new PERR_OptLayout to set in #arg
 * \return A new PERR_Slice with wsAfterAreas at the right place, or a core dump if #arg did not have a wsAfterAreas
 */
PERR_Slice PERR_setSliceWsAfterAreas(PERR_Slice arg, PERR_OptLayout wsAfterAreas) {
  if (PERR_isSliceSlice(arg)) {
    return (PERR_Slice)ATsetArgument((ATermAppl)arg, (ATerm)ATreplace((ATermList)ATgetArgument((ATermAppl)arg, 1), (ATerm)((ATerm) wsAfterAreas), 11), 1);
  }

  ATabort("Slice has no WsAfterAreas: %t\n", arg);
  return (PERR_Slice)NULL;
}

/**
 * Set the ws-after-] of a PERR_Slice. The precondition being that this PERR_Slice actually has a ws-after-]
 * \param[in] arg input PERR_Slice
 * \param[in] wsAfterBracketClose new PERR_OptLayout to set in #arg
 * \return A new PERR_Slice with wsAfterBracketClose at the right place, or a core dump if #arg did not have a wsAfterBracketClose
 */
PERR_Slice PERR_setSliceWsAfterBracketClose(PERR_Slice arg, PERR_OptLayout wsAfterBracketClose) {
  if (PERR_isSliceSlice(arg)) {
    return (PERR_Slice)ATsetArgument((ATermAppl)arg, (ATerm)ATreplace((ATermList)ATgetArgument((ATermAppl)arg, 1), (ATerm)((ATerm) wsAfterBracketClose), 13), 1);
  }

  ATabort("Slice has no WsAfterBracketClose: %t\n", arg);
  return (PERR_Slice)NULL;
}

/**
 * Assert whether a PERR_AreaAreas is any of the valid alternatives, or not. This analysis does not go any deeper than the top level
 * \param[in] arg input PERR_AreaAreas
 * \return ATtrue if #arg corresponds to the expected signature, or ATfalse otherwise
 */
ATbool PERR_isValidAreaAreas(PERR_AreaAreas arg) {
  if (PERR_isAreaAreasEmpty(arg)) {
    return ATtrue;
  }
  else if (PERR_isAreaAreasSingle(arg)) {
    return ATtrue;
  }
  else if (PERR_isAreaAreasMany(arg)) {
    return ATtrue;
  }
  return ATfalse;
}

/**
 * Assert whether a PERR_AreaAreas is a empty by checking against the following ATerm pattern: []. May not be used to assert correctness of the PERR_AreaAreas
 * \param[in] arg input PERR_AreaAreas
 * \return ATtrue if #arg corresponds to the signature of a empty, or ATfalse otherwise
 */
inline ATbool PERR_isAreaAreasEmpty(PERR_AreaAreas arg){
  if (ATisEmpty((ATermList)arg)) {
    return ATtrue;
  }
  return ATfalse;
}

/**
 * Assert whether a PERR_AreaAreas is a single by checking against the following ATerm pattern: [<head(Area)>]. May not be used to assert correctness of the PERR_AreaAreas
 * \param[in] arg input PERR_AreaAreas
 * \return ATtrue if #arg corresponds to the signature of a single, or ATfalse otherwise
 */
inline ATbool PERR_isAreaAreasSingle(PERR_AreaAreas arg){
  if (ATgetType((ATerm)arg) == AT_LIST && ATisEmpty((ATermList)arg) == ATfalse) {
    ATermList arg_list = (ATermList)arg;
    arg_list = ATgetNext(arg_list);
    if (ATgetType((ATerm)arg_list) == AT_LIST && ATisEmpty((ATermList)arg_list) == ATtrue) {
      return ATtrue;
    }
  }
  return ATfalse;
}

/**
 * Assert whether a PERR_AreaAreas is a many by checking against the following ATerm pattern: [<head(Area)>,<ws-after-head("OptLayout")>,appl(prod([char-class([44])],lit(","),no-attrs),[44]),<ws-after-sep("OptLayout")>,<[tail(Area-areas)]>]. May not be used to assert correctness of the PERR_AreaAreas
 * \param[in] arg input PERR_AreaAreas
 * \return ATtrue if #arg corresponds to the signature of a many, or ATfalse otherwise
 */
inline ATbool PERR_isAreaAreasMany(PERR_AreaAreas arg){
  if (ATgetType((ATerm)arg) == AT_LIST && ATisEmpty((ATermList)arg) == ATfalse) {
    ATerm arg_head;
    ATermList arg_list = (ATermList)arg;
    arg_list = ATgetNext(arg_list);
    if (ATgetType((ATerm)arg_list) == AT_LIST && ATisEmpty((ATermList)arg_list) == ATfalse) {
      arg_list = ATgetNext(arg_list);
      if (ATgetType((ATerm)arg_list) == AT_LIST && ATisEmpty((ATermList)arg_list) == ATfalse) {
        arg_head = ATgetFirst(arg_list);
        arg_list = ATgetNext(arg_list);
        /* checking for: appl */
        if (ATgetType((ATerm)arg_head) == AT_APPL && ATgetAFun((ATermAppl)arg_head) == PERR_afun0) {
          ATerm arg_head_arg0 = ATgetArgument(arg_head, 0);
          /* checking for: prod */
          if (ATgetType((ATerm)arg_head_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_head_arg0) == PERR_afun1) {
            ATerm arg_head_arg0_arg0 = ATgetArgument(arg_head_arg0, 0);
            if (ATgetType((ATerm)arg_head_arg0_arg0) == AT_LIST && ATisEmpty((ATermList)arg_head_arg0_arg0) == ATfalse) {
              ATerm arg_head_arg0_arg0_head;
              ATermList arg_head_arg0_arg0_list = (ATermList)arg_head_arg0_arg0;
              arg_head_arg0_arg0_head = ATgetFirst(arg_head_arg0_arg0_list);
              arg_head_arg0_arg0_list = ATgetNext(arg_head_arg0_arg0_list);
              /* checking for: char-class */
              if (ATgetType((ATerm)arg_head_arg0_arg0_head) == AT_APPL && ATgetAFun((ATermAppl)arg_head_arg0_arg0_head) == PERR_afun17) {
                ATerm arg_head_arg0_arg0_head_arg0 = ATgetArgument(arg_head_arg0_arg0_head, 0);
                if (ATgetType((ATerm)arg_head_arg0_arg0_head_arg0) == AT_LIST && ATisEmpty((ATermList)arg_head_arg0_arg0_head_arg0) == ATfalse) {
                  ATerm arg_head_arg0_arg0_head_arg0_head;
                  ATermList arg_head_arg0_arg0_head_arg0_list = (ATermList)arg_head_arg0_arg0_head_arg0;
                  arg_head_arg0_arg0_head_arg0_head = ATgetFirst(arg_head_arg0_arg0_head_arg0_list);
                  arg_head_arg0_arg0_head_arg0_list = ATgetNext(arg_head_arg0_arg0_head_arg0_list);
                  if (ATgetInt((ATermInt)arg_head_arg0_arg0_head_arg0_head) == 44) {

                    if (ATgetType((ATerm)arg_head_arg0_arg0_head_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_head_arg0_arg0_head_arg0_list) == ATtrue) {
                      if (ATgetType((ATerm)arg_head_arg0_arg0_list) == AT_LIST && ATisEmpty((ATermList)arg_head_arg0_arg0_list) == ATtrue) {
                        ATerm arg_head_arg0_arg1 = ATgetArgument(arg_head_arg0, 1);
                        /* checking for: lit */
                        if (ATgetType((ATerm)arg_head_arg0_arg1) == AT_APPL && ATgetAFun((ATermAppl)arg_head_arg0_arg1) == PERR_afun9) {
                          ATerm arg_head_arg0_arg1_arg0 = ATgetArgument(arg_head_arg0_arg1, 0);
                          /* checking for: , */
                          if (ATgetType((ATerm)arg_head_arg0_arg1_arg0) == AT_APPL && ATgetAFun((ATermAppl)arg_head_arg0_arg1_arg0) == PERR_afun43) {
                            ATerm arg_head_arg0_arg2 = ATgetArgument(arg_head_arg0, 2);
                            /* checking for: no-attrs */
                            if (ATgetType((ATerm)arg_head_arg0_arg2) == AT_APPL && ATgetAFun((ATermAppl)arg_head_arg0_arg2) == PERR_afun5) {
                              ATerm arg_head_arg1 = ATgetArgument(arg_head, 1);
                              if (ATgetType((ATerm)arg_head_arg1) == AT_LIST && ATisEmpty((ATermList)arg_head_arg1) == ATfalse) {
                                ATerm arg_head_arg1_head;
                                ATermList arg_head_arg1_list = (ATermList)arg_head_arg1;
                                arg_head_arg1_head = ATgetFirst(arg_head_arg1_list);
                                arg_head_arg1_list = ATgetNext(arg_head_arg1_list);
                                if (ATgetInt((ATermInt)arg_head_arg1_head) == 44) {

                                  if (ATgetType((ATerm)arg_head_arg1_list) == AT_LIST && ATisEmpty((ATermList)arg_head_arg1_list) == ATtrue) {
                                    if (ATgetType((ATerm)arg_list) == AT_LIST && ATisEmpty((ATermList)arg_list) == ATfalse) {
                                      return ATtrue;
                                    }
                                  }
                                }
                              }
                            }
                          }
                        }
                      }
                    }
                  }
                }
              }
            }
          }
        }
      }
    }
  }
  return ATfalse;
}

/**
 * Assert whether a PERR_AreaAreas has a head. 
 * \param[in] arg input PERR_AreaAreas
 * \return ATtrue if the PERR_AreaAreas had a head, or ATfalse otherwise
 */
ATbool PERR_hasAreaAreasHead(PERR_AreaAreas arg) {
  if (PERR_isAreaAreasSingle(arg)) {
    return ATtrue;
  }
  else if (PERR_isAreaAreasMany(arg)) {
    return ATtrue;
  }
  return ATfalse;
}

/**
 * Assert whether a PERR_AreaAreas has a ws-after-head. 
 * \param[in] arg input PERR_AreaAreas
 * \return ATtrue if the PERR_AreaAreas had a ws-after-head, or ATfalse otherwise
 */
ATbool PERR_hasAreaAreasWsAfterHead(PERR_AreaAreas arg) {
  if (PERR_isAreaAreasMany(arg)) {
    return ATtrue;
  }
  return ATfalse;
}

/**
 * Assert whether a PERR_AreaAreas has a ws-after-sep. 
 * \param[in] arg input PERR_AreaAreas
 * \return ATtrue if the PERR_AreaAreas had a ws-after-sep, or ATfalse otherwise
 */
ATbool PERR_hasAreaAreasWsAfterSep(PERR_AreaAreas arg) {
  if (PERR_isAreaAreasMany(arg)) {
    return ATtrue;
  }
  return ATfalse;
}

/**
 * Assert whether a PERR_AreaAreas has a tail. 
 * \param[in] arg input PERR_AreaAreas
 * \return ATtrue if the PERR_AreaAreas had a tail, or ATfalse otherwise
 */
ATbool PERR_hasAreaAreasTail(PERR_AreaAreas arg) {
  if (PERR_isAreaAreasMany(arg)) {
    return ATtrue;
  }
  return ATfalse;
}

/**
 * Returns a list of all but the first element of a PERR_AreaAreas. 
 * \param[in] arg input PERR_AreaAreas
 * \return A new PERR_AreaAreas, without the first element and the separator(s) just after it.)
 */
PERR_AreaAreas PERR_getAreaAreasTail(PERR_AreaAreas arg) {
  assert(!PERR_isAreaAreasEmpty(arg) && "getTail on an empty list");
  if (PERR_isAreaAreasSingle(arg)) {
    return (PERR_AreaAreas) PERR_makeAreaAreasEmpty();
  }
  else {
  
    return (PERR_AreaAreas)ATgetTail((ATermList)arg, 4);
  }
}

/**
 * Get the head PERR_Area of a PERR_AreaAreas. Note that the precondition is that this PERR_AreaAreas actually has a head
 * \param[in] arg input PERR_AreaAreas
 * \return the head of #arg, if it exist or an undefined value if it does not
 */
PERR_Area PERR_getAreaAreasHead(PERR_AreaAreas arg) {
  if (PERR_isAreaAreasSingle(arg)) {
    return (PERR_Area)ATgetFirst((ATermList)arg);
  }
  else 
    return (PERR_Area)ATgetFirst((ATermList)arg);
}

/**
 * Get the ws-after-head PERR_OptLayout of a PERR_AreaAreas. Note that the precondition is that this PERR_AreaAreas actually has a ws-after-head
 * \param[in] arg input PERR_AreaAreas
 * \return the ws-after-head of #arg, if it exist or an undefined value if it does not
 */
PERR_OptLayout PERR_getAreaAreasWsAfterHead(PERR_AreaAreas arg) {
  
    return (PERR_OptLayout)ATelementAt((ATermList)arg, 1);
}

/**
 * Get the ws-after-sep PERR_OptLayout of a PERR_AreaAreas. Note that the precondition is that this PERR_AreaAreas actually has a ws-after-sep
 * \param[in] arg input PERR_AreaAreas
 * \return the ws-after-sep of #arg, if it exist or an undefined value if it does not
 */
PERR_OptLayout PERR_getAreaAreasWsAfterSep(PERR_AreaAreas arg) {
  
    return (PERR_OptLayout)ATelementAt((ATermList)arg, 3);
}

/**
 * Set the head of a PERR_AreaAreas. The precondition being that this PERR_AreaAreas actually has a head
 * \param[in] arg input PERR_AreaAreas
 * \param[in] head new PERR_Area to set in #arg
 * \return A new PERR_AreaAreas with head at the right place, or a core dump if #arg did not have a head
 */
PERR_AreaAreas PERR_setAreaAreasHead(PERR_AreaAreas arg, PERR_Area head) {
  if (PERR_isAreaAreasSingle(arg)) {
    return (PERR_AreaAreas)ATreplace((ATermList)arg, (ATerm)((ATerm) head), 0);
  }
  else if (PERR_isAreaAreasMany(arg)) {
    return (PERR_AreaAreas)ATreplace((ATermList)arg, (ATerm)((ATerm) head), 0);
  }

  ATabort("AreaAreas has no Head: %t\n", arg);
  return (PERR_AreaAreas)NULL;
}

/**
 * Set the ws-after-head of a PERR_AreaAreas. The precondition being that this PERR_AreaAreas actually has a ws-after-head
 * \param[in] arg input PERR_AreaAreas
 * \param[in] wsAfterHead new PERR_OptLayout to set in #arg
 * \return A new PERR_AreaAreas with wsAfterHead at the right place, or a core dump if #arg did not have a wsAfterHead
 */
PERR_AreaAreas PERR_setAreaAreasWsAfterHead(PERR_AreaAreas arg, PERR_OptLayout wsAfterHead) {
  if (PERR_isAreaAreasMany(arg)) {
    return (PERR_AreaAreas)ATreplace((ATermList)arg, (ATerm)((ATerm) wsAfterHead), 1);
  }

  ATabort("AreaAreas has no WsAfterHead: %t\n", arg);
  return (PERR_AreaAreas)NULL;
}

/**
 * Set the ws-after-sep of a PERR_AreaAreas. The precondition being that this PERR_AreaAreas actually has a ws-after-sep
 * \param[in] arg input PERR_AreaAreas
 * \param[in] wsAfterSep new PERR_OptLayout to set in #arg
 * \return A new PERR_AreaAreas with wsAfterSep at the right place, or a core dump if #arg did not have a wsAfterSep
 */
PERR_AreaAreas PERR_setAreaAreasWsAfterSep(PERR_AreaAreas arg, PERR_OptLayout wsAfterSep) {
  if (PERR_isAreaAreasMany(arg)) {
    return (PERR_AreaAreas)ATreplace((ATermList)arg, (ATerm)((ATerm) wsAfterSep), 3);
  }

  ATabort("AreaAreas has no WsAfterSep: %t\n", arg);
  return (PERR_AreaAreas)NULL;
}

/**
 * Set the tail of a PERR_AreaAreas. The precondition being that this PERR_AreaAreas actually has a tail
 * \param[in] arg input PERR_AreaAreas
 * \param[in] tail new PERR_AreaAreas to set in #arg
 * \return A new PERR_AreaAreas with tail at the right place, or a core dump if #arg did not have a tail
 */
PERR_AreaAreas PERR_setAreaAreasTail(PERR_AreaAreas arg, PERR_AreaAreas tail) {
  if (PERR_isAreaAreasMany(arg)) {
    return (PERR_AreaAreas)ATreplaceTail((ATermList)arg, (ATermList)((ATerm) tail), 4);
  }

  ATabort("AreaAreas has no Tail: %t\n", arg);
  return (PERR_AreaAreas)NULL;
}

/**
 * Apply functions to the children of a PERR_OptLayout. 
 * \return A new PERR_OptLayout with new children where the argument functions might have applied
 */
PERR_OptLayout PERR_visitOptLayout(PERR_OptLayout arg, PERR_Layout (*acceptLayout)(PERR_Layout)) {
  if (PERR_isOptLayoutAbsent(arg)) {
    return PERR_makeOptLayoutAbsent();
  }
  if (PERR_isOptLayoutPresent(arg)) {
    return PERR_makeOptLayoutPresent(
        acceptLayout ? acceptLayout(PERR_getOptLayoutLayout(arg)) : PERR_getOptLayoutLayout(arg));
  }
  ATabort("not a OptLayout: %t\n", arg);
  return (PERR_OptLayout)NULL;
}
/**
 * Apply functions to the children of a PERR_Layout. 
 * \return A new PERR_Layout with new children where the argument functions might have applied
 */
PERR_Layout PERR_visitLayout(PERR_Layout arg, PERR_LexLayoutList (*acceptList)(PERR_LexLayoutList)) {
  if (PERR_isLayoutLexToCf(arg)) {
    return PERR_makeLayoutLexToCf(
        acceptList ? acceptList(PERR_getLayoutList(arg)) : PERR_getLayoutList(arg));
  }
  ATabort("not a Layout: %t\n", arg);
  return (PERR_Layout)NULL;
}
/**
 * Apply functions to the children of a PERR_LexLayoutList. 
 * \return A new PERR_LexLayoutList with new children where the argument functions might have applied
 */
PERR_LexLayoutList PERR_visitLexLayoutList(PERR_LexLayoutList arg, PERR_LexLayout (*acceptHead)(PERR_LexLayout)) {
  if (PERR_isLexLayoutListEmpty(arg)) {
    return PERR_makeLexLayoutListEmpty();
  }
  if (PERR_isLexLayoutListSingle(arg)) {
    return PERR_makeLexLayoutListSingle(
        acceptHead ? acceptHead(PERR_getLexLayoutListHead(arg)) : PERR_getLexLayoutListHead(arg));
  }
  if (PERR_isLexLayoutListMany(arg)) {
    return PERR_makeLexLayoutListMany(
        acceptHead ? acceptHead(PERR_getLexLayoutListHead(arg)) : PERR_getLexLayoutListHead(arg),
        PERR_visitLexLayoutList(PERR_getLexLayoutListTail(arg), acceptHead));
  }
  ATabort("not a LexLayoutList: %t\n", arg);
  return (PERR_LexLayoutList)NULL;
}
/**
 * Apply functions to the children of a PERR_LexStrChar. 
 * \return A new PERR_LexStrChar with new children where the argument functions might have applied
 */
PERR_LexStrChar PERR_visitLexStrChar(PERR_LexStrChar arg, char (*acceptA)(char), char (*acceptB)(char), char (*acceptC)(char), char (*acceptCh)(char)) {
  if (PERR_isLexStrCharNewline(arg)) {
    return PERR_makeLexStrCharNewline();
  }
  if (PERR_isLexStrCharTab(arg)) {
    return PERR_makeLexStrCharTab();
  }
  if (PERR_isLexStrCharQuote(arg)) {
    return PERR_makeLexStrCharQuote();
  }
  if (PERR_isLexStrCharBackslash(arg)) {
    return PERR_makeLexStrCharBackslash();
  }
  if (PERR_isLexStrCharDecimal(arg)) {
    return PERR_makeLexStrCharDecimal(
        acceptA ? acceptA(PERR_getLexStrCharA(arg)) : PERR_getLexStrCharA(arg),
        acceptB ? acceptB(PERR_getLexStrCharB(arg)) : PERR_getLexStrCharB(arg),
        acceptC ? acceptC(PERR_getLexStrCharC(arg)) : PERR_getLexStrCharC(arg));
  }
  if (PERR_isLexStrCharNormal(arg)) {
    return PERR_makeLexStrCharNormal(
        acceptCh ? acceptCh(PERR_getLexStrCharCh(arg)) : PERR_getLexStrCharCh(arg));
  }
  ATabort("not a LexStrChar: %t\n", arg);
  return (PERR_LexStrChar)NULL;
}
/**
 * Apply functions to the children of a PERR_StrChar. 
 * \return A new PERR_StrChar with new children where the argument functions might have applied
 */
PERR_StrChar PERR_visitStrChar(PERR_StrChar arg, PERR_LexStrChar (*acceptStrChar)(PERR_LexStrChar)) {
  if (PERR_isStrCharLexToCf(arg)) {
    return PERR_makeStrCharLexToCf(
        acceptStrChar ? acceptStrChar(PERR_getStrCharStrChar(arg)) : PERR_getStrCharStrChar(arg));
  }
  ATabort("not a StrChar: %t\n", arg);
  return (PERR_StrChar)NULL;
}
/**
 * Apply functions to the children of a PERR_LexStrCon. 
 * \return A new PERR_LexStrCon with new children where the argument functions might have applied
 */
PERR_LexStrCon PERR_visitLexStrCon(PERR_LexStrCon arg, PERR_LexStrCharChars (*acceptChars)(PERR_LexStrCharChars)) {
  if (PERR_isLexStrConDefault(arg)) {
    return PERR_makeLexStrConDefault(
        acceptChars ? acceptChars(PERR_getLexStrConChars(arg)) : PERR_getLexStrConChars(arg));
  }
  ATabort("not a LexStrCon: %t\n", arg);
  return (PERR_LexStrCon)NULL;
}
/**
 * Apply functions to the children of a PERR_StrCon. 
 * \return A new PERR_StrCon with new children where the argument functions might have applied
 */
PERR_StrCon PERR_visitStrCon(PERR_StrCon arg, PERR_LexStrCon (*acceptStrCon)(PERR_LexStrCon)) {
  if (PERR_isStrConLexToCf(arg)) {
    return PERR_makeStrConLexToCf(
        acceptStrCon ? acceptStrCon(PERR_getStrConStrCon(arg)) : PERR_getStrConStrCon(arg));
  }
  ATabort("not a StrCon: %t\n", arg);
  return (PERR_StrCon)NULL;
}
/**
 * Apply functions to the children of a PERR_LexStrCharChars. 
 * \return A new PERR_LexStrCharChars with new children where the argument functions might have applied
 */
PERR_LexStrCharChars PERR_visitLexStrCharChars(PERR_LexStrCharChars arg, PERR_LexStrChar (*acceptHead)(PERR_LexStrChar)) {
  if (PERR_isLexStrCharCharsEmpty(arg)) {
    return PERR_makeLexStrCharCharsEmpty();
  }
  if (PERR_isLexStrCharCharsSingle(arg)) {
    return PERR_makeLexStrCharCharsSingle(
        acceptHead ? acceptHead(PERR_getLexStrCharCharsHead(arg)) : PERR_getLexStrCharCharsHead(arg));
  }
  if (PERR_isLexStrCharCharsMany(arg)) {
    return PERR_makeLexStrCharCharsMany(
        acceptHead ? acceptHead(PERR_getLexStrCharCharsHead(arg)) : PERR_getLexStrCharCharsHead(arg),
        PERR_visitLexStrCharChars(PERR_getLexStrCharCharsTail(arg), acceptHead));
  }
  ATabort("not a LexStrCharChars: %t\n", arg);
  return (PERR_LexStrCharChars)NULL;
}
/**
 * Apply functions to the children of a PERR_LexNatCon. 
 * \return A new PERR_LexNatCon with new children where the argument functions might have applied
 */
PERR_LexNatCon PERR_visitLexNatCon(PERR_LexNatCon arg, char* (*acceptList)(char*)) {
  if (PERR_isLexNatConDigits(arg)) {
    return PERR_makeLexNatConDigits(
        acceptList ? acceptList(PERR_getLexNatConList(arg)) : PERR_getLexNatConList(arg));
  }
  ATabort("not a LexNatCon: %t\n", arg);
  return (PERR_LexNatCon)NULL;
}
/**
 * Apply functions to the children of a PERR_NatCon. 
 * \return A new PERR_NatCon with new children where the argument functions might have applied
 */
PERR_NatCon PERR_visitNatCon(PERR_NatCon arg, PERR_LexNatCon (*acceptNatCon)(PERR_LexNatCon)) {
  if (PERR_isNatConLexToCf(arg)) {
    return PERR_makeNatConLexToCf(
        acceptNatCon ? acceptNatCon(PERR_getNatConNatCon(arg)) : PERR_getNatConNatCon(arg));
  }
  ATabort("not a NatCon: %t\n", arg);
  return (PERR_NatCon)NULL;
}
/**
 * Apply functions to the children of a PERR_LexLayout. 
 * \return A new PERR_LexLayout with new children where the argument functions might have applied
 */
PERR_LexLayout PERR_visitLexLayout(PERR_LexLayout arg, char (*acceptCh)(char)) {
  if (PERR_isLexLayoutWhitespace(arg)) {
    return PERR_makeLexLayoutWhitespace(
        acceptCh ? acceptCh(PERR_getLexLayoutCh(arg)) : PERR_getLexLayoutCh(arg));
  }
  ATabort("not a LexLayout: %t\n", arg);
  return (PERR_LexLayout)NULL;
}
/**
 * Apply functions to the children of a PERR_Start. 
 * \return A new PERR_Start with new children where the argument functions might have applied
 */
PERR_Start PERR_visitStart(PERR_Start arg, PERR_OptLayout (*acceptWsBefore)(PERR_OptLayout), PERR_Summary (*acceptTopSummary)(PERR_Summary), PERR_OptLayout (*acceptWsAfter)(PERR_OptLayout), int (*acceptAmbCnt)(int), PERR_Error (*acceptTopError)(PERR_Error)) {
  if (PERR_isStartSummary(arg)) {
    return PERR_makeStartSummary(
        acceptWsBefore ? acceptWsBefore(PERR_getStartWsBefore(arg)) : PERR_getStartWsBefore(arg),
        acceptTopSummary ? acceptTopSummary(PERR_getStartTopSummary(arg)) : PERR_getStartTopSummary(arg),
        acceptWsAfter ? acceptWsAfter(PERR_getStartWsAfter(arg)) : PERR_getStartWsAfter(arg),
        acceptAmbCnt ? acceptAmbCnt(PERR_getStartAmbCnt(arg)) : PERR_getStartAmbCnt(arg));
  }
  if (PERR_isStartError(arg)) {
    return PERR_makeStartError(
        acceptWsBefore ? acceptWsBefore(PERR_getStartWsBefore(arg)) : PERR_getStartWsBefore(arg),
        acceptTopError ? acceptTopError(PERR_getStartTopError(arg)) : PERR_getStartTopError(arg),
        acceptWsAfter ? acceptWsAfter(PERR_getStartWsAfter(arg)) : PERR_getStartWsAfter(arg),
        acceptAmbCnt ? acceptAmbCnt(PERR_getStartAmbCnt(arg)) : PERR_getStartAmbCnt(arg));
  }
  ATabort("not a Start: %t\n", arg);
  return (PERR_Start)NULL;
}
/**
 * Apply functions to the children of a PERR_Subject. 
 * \return A new PERR_Subject with new children where the argument functions might have applied
 */
PERR_Subject PERR_visitSubject(PERR_Subject arg, PERR_OptLayout (*acceptWsAfterSubject)(PERR_OptLayout), PERR_OptLayout (*acceptWsAfterParenOpen)(PERR_OptLayout), PERR_StrCon (*acceptDescription)(PERR_StrCon), PERR_OptLayout (*acceptWsAfterDescription)(PERR_OptLayout), PERR_OptLayout (*acceptWsAfterLocalized)(PERR_OptLayout), PERR_OptLayout (*acceptWsAfterComma)(PERR_OptLayout), PERR_Location (*acceptLocation)(PERR_Location), PERR_OptLayout (*acceptWsAfterLocation)(PERR_OptLayout)) {
  if (PERR_isSubjectSubject(arg)) {
    return PERR_makeSubjectSubject(
        acceptWsAfterSubject ? acceptWsAfterSubject(PERR_getSubjectWsAfterSubject(arg)) : PERR_getSubjectWsAfterSubject(arg),
        acceptWsAfterParenOpen ? acceptWsAfterParenOpen(PERR_getSubjectWsAfterParenOpen(arg)) : PERR_getSubjectWsAfterParenOpen(arg),
        acceptDescription ? acceptDescription(PERR_getSubjectDescription(arg)) : PERR_getSubjectDescription(arg),
        acceptWsAfterDescription ? acceptWsAfterDescription(PERR_getSubjectWsAfterDescription(arg)) : PERR_getSubjectWsAfterDescription(arg));
  }
  if (PERR_isSubjectLocalized(arg)) {
    return PERR_makeSubjectLocalized(
        acceptWsAfterLocalized ? acceptWsAfterLocalized(PERR_getSubjectWsAfterLocalized(arg)) : PERR_getSubjectWsAfterLocalized(arg),
        acceptWsAfterParenOpen ? acceptWsAfterParenOpen(PERR_getSubjectWsAfterParenOpen(arg)) : PERR_getSubjectWsAfterParenOpen(arg),
        acceptDescription ? acceptDescription(PERR_getSubjectDescription(arg)) : PERR_getSubjectDescription(arg),
        acceptWsAfterDescription ? acceptWsAfterDescription(PERR_getSubjectWsAfterDescription(arg)) : PERR_getSubjectWsAfterDescription(arg),
        acceptWsAfterComma ? acceptWsAfterComma(PERR_getSubjectWsAfterComma(arg)) : PERR_getSubjectWsAfterComma(arg),
        acceptLocation ? acceptLocation(PERR_getSubjectLocation(arg)) : PERR_getSubjectLocation(arg),
        acceptWsAfterLocation ? acceptWsAfterLocation(PERR_getSubjectWsAfterLocation(arg)) : PERR_getSubjectWsAfterLocation(arg));
  }
  ATabort("not a Subject: %t\n", arg);
  return (PERR_Subject)NULL;
}
/**
 * Apply functions to the children of a PERR_Error. 
 * \return A new PERR_Error with new children where the argument functions might have applied
 */
PERR_Error PERR_visitError(PERR_Error arg, PERR_OptLayout (*acceptWsAfterInfo)(PERR_OptLayout), PERR_OptLayout (*acceptWsAfterParenOpen)(PERR_OptLayout), PERR_StrCon (*acceptDescription)(PERR_StrCon), PERR_OptLayout (*acceptWsAfterDescription)(PERR_OptLayout), PERR_OptLayout (*acceptWsAfterComma)(PERR_OptLayout), PERR_OptLayout (*acceptWsAfterBracketOpen)(PERR_OptLayout), PERR_SubjectList (*acceptList)(PERR_SubjectList), PERR_OptLayout (*acceptWsAfterList)(PERR_OptLayout), PERR_OptLayout (*acceptWsAfterBracketClose)(PERR_OptLayout), PERR_OptLayout (*acceptWsAfterWarning)(PERR_OptLayout), PERR_OptLayout (*acceptWsAfterError)(PERR_OptLayout), PERR_OptLayout (*acceptWsAfterFatal)(PERR_OptLayout)) {
  if (PERR_isErrorInfo(arg)) {
    return PERR_makeErrorInfo(
        acceptWsAfterInfo ? acceptWsAfterInfo(PERR_getErrorWsAfterInfo(arg)) : PERR_getErrorWsAfterInfo(arg),
        acceptWsAfterParenOpen ? acceptWsAfterParenOpen(PERR_getErrorWsAfterParenOpen(arg)) : PERR_getErrorWsAfterParenOpen(arg),
        acceptDescription ? acceptDescription(PERR_getErrorDescription(arg)) : PERR_getErrorDescription(arg),
        acceptWsAfterDescription ? acceptWsAfterDescription(PERR_getErrorWsAfterDescription(arg)) : PERR_getErrorWsAfterDescription(arg),
        acceptWsAfterComma ? acceptWsAfterComma(PERR_getErrorWsAfterComma(arg)) : PERR_getErrorWsAfterComma(arg),
        acceptWsAfterBracketOpen ? acceptWsAfterBracketOpen(PERR_getErrorWsAfterBracketOpen(arg)) : PERR_getErrorWsAfterBracketOpen(arg),
        acceptList ? acceptList(PERR_getErrorList(arg)) : PERR_getErrorList(arg),
        acceptWsAfterList ? acceptWsAfterList(PERR_getErrorWsAfterList(arg)) : PERR_getErrorWsAfterList(arg),
        acceptWsAfterBracketClose ? acceptWsAfterBracketClose(PERR_getErrorWsAfterBracketClose(arg)) : PERR_getErrorWsAfterBracketClose(arg));
  }
  if (PERR_isErrorWarning(arg)) {
    return PERR_makeErrorWarning(
        acceptWsAfterWarning ? acceptWsAfterWarning(PERR_getErrorWsAfterWarning(arg)) : PERR_getErrorWsAfterWarning(arg),
        acceptWsAfterParenOpen ? acceptWsAfterParenOpen(PERR_getErrorWsAfterParenOpen(arg)) : PERR_getErrorWsAfterParenOpen(arg),
        acceptDescription ? acceptDescription(PERR_getErrorDescription(arg)) : PERR_getErrorDescription(arg),
        acceptWsAfterDescription ? acceptWsAfterDescription(PERR_getErrorWsAfterDescription(arg)) : PERR_getErrorWsAfterDescription(arg),
        acceptWsAfterComma ? acceptWsAfterComma(PERR_getErrorWsAfterComma(arg)) : PERR_getErrorWsAfterComma(arg),
        acceptWsAfterBracketOpen ? acceptWsAfterBracketOpen(PERR_getErrorWsAfterBracketOpen(arg)) : PERR_getErrorWsAfterBracketOpen(arg),
        acceptList ? acceptList(PERR_getErrorList(arg)) : PERR_getErrorList(arg),
        acceptWsAfterList ? acceptWsAfterList(PERR_getErrorWsAfterList(arg)) : PERR_getErrorWsAfterList(arg),
        acceptWsAfterBracketClose ? acceptWsAfterBracketClose(PERR_getErrorWsAfterBracketClose(arg)) : PERR_getErrorWsAfterBracketClose(arg));
  }
  if (PERR_isErrorError(arg)) {
    return PERR_makeErrorError(
        acceptWsAfterError ? acceptWsAfterError(PERR_getErrorWsAfterError(arg)) : PERR_getErrorWsAfterError(arg),
        acceptWsAfterParenOpen ? acceptWsAfterParenOpen(PERR_getErrorWsAfterParenOpen(arg)) : PERR_getErrorWsAfterParenOpen(arg),
        acceptDescription ? acceptDescription(PERR_getErrorDescription(arg)) : PERR_getErrorDescription(arg),
        acceptWsAfterDescription ? acceptWsAfterDescription(PERR_getErrorWsAfterDescription(arg)) : PERR_getErrorWsAfterDescription(arg),
        acceptWsAfterComma ? acceptWsAfterComma(PERR_getErrorWsAfterComma(arg)) : PERR_getErrorWsAfterComma(arg),
        acceptWsAfterBracketOpen ? acceptWsAfterBracketOpen(PERR_getErrorWsAfterBracketOpen(arg)) : PERR_getErrorWsAfterBracketOpen(arg),
        acceptList ? acceptList(PERR_getErrorList(arg)) : PERR_getErrorList(arg),
        acceptWsAfterList ? acceptWsAfterList(PERR_getErrorWsAfterList(arg)) : PERR_getErrorWsAfterList(arg),
        acceptWsAfterBracketClose ? acceptWsAfterBracketClose(PERR_getErrorWsAfterBracketClose(arg)) : PERR_getErrorWsAfterBracketClose(arg));
  }
  if (PERR_isErrorFatal(arg)) {
    return PERR_makeErrorFatal(
        acceptWsAfterFatal ? acceptWsAfterFatal(PERR_getErrorWsAfterFatal(arg)) : PERR_getErrorWsAfterFatal(arg),
        acceptWsAfterParenOpen ? acceptWsAfterParenOpen(PERR_getErrorWsAfterParenOpen(arg)) : PERR_getErrorWsAfterParenOpen(arg),
        acceptDescription ? acceptDescription(PERR_getErrorDescription(arg)) : PERR_getErrorDescription(arg),
        acceptWsAfterDescription ? acceptWsAfterDescription(PERR_getErrorWsAfterDescription(arg)) : PERR_getErrorWsAfterDescription(arg),
        acceptWsAfterComma ? acceptWsAfterComma(PERR_getErrorWsAfterComma(arg)) : PERR_getErrorWsAfterComma(arg),
        acceptWsAfterBracketOpen ? acceptWsAfterBracketOpen(PERR_getErrorWsAfterBracketOpen(arg)) : PERR_getErrorWsAfterBracketOpen(arg),
        acceptList ? acceptList(PERR_getErrorList(arg)) : PERR_getErrorList(arg),
        acceptWsAfterList ? acceptWsAfterList(PERR_getErrorWsAfterList(arg)) : PERR_getErrorWsAfterList(arg),
        acceptWsAfterBracketClose ? acceptWsAfterBracketClose(PERR_getErrorWsAfterBracketClose(arg)) : PERR_getErrorWsAfterBracketClose(arg));
  }
  ATabort("not a Error: %t\n", arg);
  return (PERR_Error)NULL;
}
/**
 * Apply functions to the children of a PERR_Summary. 
 * \return A new PERR_Summary with new children where the argument functions might have applied
 */
PERR_Summary PERR_visitSummary(PERR_Summary arg, PERR_OptLayout (*acceptWsAfterSummary)(PERR_OptLayout), PERR_OptLayout (*acceptWsAfterParenOpen)(PERR_OptLayout), PERR_StrCon (*acceptProducer)(PERR_StrCon), PERR_OptLayout (*acceptWsAfterProducer)(PERR_OptLayout), PERR_OptLayout (*acceptWsAfterComma)(PERR_OptLayout), PERR_StrCon (*acceptId)(PERR_StrCon), PERR_OptLayout (*acceptWsAfterId)(PERR_OptLayout), PERR_OptLayout (*acceptWsAfterComma1)(PERR_OptLayout), PERR_OptLayout (*acceptWsAfterBracketOpen)(PERR_OptLayout), PERR_ErrorList (*acceptList)(PERR_ErrorList), PERR_OptLayout (*acceptWsAfterList)(PERR_OptLayout), PERR_OptLayout (*acceptWsAfterBracketClose)(PERR_OptLayout)) {
  if (PERR_isSummarySummary(arg)) {
    return PERR_makeSummarySummary(
        acceptWsAfterSummary ? acceptWsAfterSummary(PERR_getSummaryWsAfterSummary(arg)) : PERR_getSummaryWsAfterSummary(arg),
        acceptWsAfterParenOpen ? acceptWsAfterParenOpen(PERR_getSummaryWsAfterParenOpen(arg)) : PERR_getSummaryWsAfterParenOpen(arg),
        acceptProducer ? acceptProducer(PERR_getSummaryProducer(arg)) : PERR_getSummaryProducer(arg),
        acceptWsAfterProducer ? acceptWsAfterProducer(PERR_getSummaryWsAfterProducer(arg)) : PERR_getSummaryWsAfterProducer(arg),
        acceptWsAfterComma ? acceptWsAfterComma(PERR_getSummaryWsAfterComma(arg)) : PERR_getSummaryWsAfterComma(arg),
        acceptId ? acceptId(PERR_getSummaryId(arg)) : PERR_getSummaryId(arg),
        acceptWsAfterId ? acceptWsAfterId(PERR_getSummaryWsAfterId(arg)) : PERR_getSummaryWsAfterId(arg),
        acceptWsAfterComma1 ? acceptWsAfterComma1(PERR_getSummaryWsAfterComma1(arg)) : PERR_getSummaryWsAfterComma1(arg),
        acceptWsAfterBracketOpen ? acceptWsAfterBracketOpen(PERR_getSummaryWsAfterBracketOpen(arg)) : PERR_getSummaryWsAfterBracketOpen(arg),
        acceptList ? acceptList(PERR_getSummaryList(arg)) : PERR_getSummaryList(arg),
        acceptWsAfterList ? acceptWsAfterList(PERR_getSummaryWsAfterList(arg)) : PERR_getSummaryWsAfterList(arg),
        acceptWsAfterBracketClose ? acceptWsAfterBracketClose(PERR_getSummaryWsAfterBracketClose(arg)) : PERR_getSummaryWsAfterBracketClose(arg));
  }
  ATabort("not a Summary: %t\n", arg);
  return (PERR_Summary)NULL;
}
/**
 * Apply functions to the children of a PERR_SubjectList. 
 * \return A new PERR_SubjectList with new children where the argument functions might have applied
 */
PERR_SubjectList PERR_visitSubjectList(PERR_SubjectList arg, PERR_Subject (*acceptHead)(PERR_Subject), PERR_OptLayout (*acceptWsAfterHead)(PERR_OptLayout), PERR_OptLayout (*acceptWsAfterSep)(PERR_OptLayout)) {
  if (PERR_isSubjectListEmpty(arg)) {
    return PERR_makeSubjectListEmpty();
  }
  if (PERR_isSubjectListSingle(arg)) {
    return PERR_makeSubjectListSingle(
        acceptHead ? acceptHead(PERR_getSubjectListHead(arg)) : PERR_getSubjectListHead(arg));
  }
  if (PERR_isSubjectListMany(arg)) {
    return PERR_makeSubjectListMany(
        acceptHead ? acceptHead(PERR_getSubjectListHead(arg)) : PERR_getSubjectListHead(arg),
        acceptWsAfterHead ? acceptWsAfterHead(PERR_getSubjectListWsAfterHead(arg)) : PERR_getSubjectListWsAfterHead(arg),
        acceptWsAfterSep ? acceptWsAfterSep(PERR_getSubjectListWsAfterSep(arg)) : PERR_getSubjectListWsAfterSep(arg),
        PERR_visitSubjectList(PERR_getSubjectListTail(arg), acceptHead, acceptWsAfterHead, acceptWsAfterSep));
  }
  ATabort("not a SubjectList: %t\n", arg);
  return (PERR_SubjectList)NULL;
}
/**
 * Apply functions to the children of a PERR_ErrorList. 
 * \return A new PERR_ErrorList with new children where the argument functions might have applied
 */
PERR_ErrorList PERR_visitErrorList(PERR_ErrorList arg, PERR_Error (*acceptHead)(PERR_Error), PERR_OptLayout (*acceptWsAfterHead)(PERR_OptLayout), PERR_OptLayout (*acceptWsAfterSep)(PERR_OptLayout)) {
  if (PERR_isErrorListEmpty(arg)) {
    return PERR_makeErrorListEmpty();
  }
  if (PERR_isErrorListSingle(arg)) {
    return PERR_makeErrorListSingle(
        acceptHead ? acceptHead(PERR_getErrorListHead(arg)) : PERR_getErrorListHead(arg));
  }
  if (PERR_isErrorListMany(arg)) {
    return PERR_makeErrorListMany(
        acceptHead ? acceptHead(PERR_getErrorListHead(arg)) : PERR_getErrorListHead(arg),
        acceptWsAfterHead ? acceptWsAfterHead(PERR_getErrorListWsAfterHead(arg)) : PERR_getErrorListWsAfterHead(arg),
        acceptWsAfterSep ? acceptWsAfterSep(PERR_getErrorListWsAfterSep(arg)) : PERR_getErrorListWsAfterSep(arg),
        PERR_visitErrorList(PERR_getErrorListTail(arg), acceptHead, acceptWsAfterHead, acceptWsAfterSep));
  }
  ATabort("not a ErrorList: %t\n", arg);
  return (PERR_ErrorList)NULL;
}
/**
 * Apply functions to the children of a PERR_Location. 
 * \return A new PERR_Location with new children where the argument functions might have applied
 */
PERR_Location PERR_visitLocation(PERR_Location arg, PERR_OptLayout (*acceptWsAfterFile)(PERR_OptLayout), PERR_OptLayout (*acceptWsAfterParenOpen)(PERR_OptLayout), PERR_StrCon (*acceptFilename)(PERR_StrCon), PERR_OptLayout (*acceptWsAfterFilename)(PERR_OptLayout), PERR_OptLayout (*acceptWsAfterA)(PERR_OptLayout), PERR_Area (*acceptArea)(PERR_Area), PERR_OptLayout (*acceptWsAfterArea)(PERR_OptLayout), PERR_OptLayout (*acceptWsAfterAreaInFile)(PERR_OptLayout), PERR_OptLayout (*acceptWsAfterComma)(PERR_OptLayout)) {
  if (PERR_isLocationFile(arg)) {
    return PERR_makeLocationFile(
        acceptWsAfterFile ? acceptWsAfterFile(PERR_getLocationWsAfterFile(arg)) : PERR_getLocationWsAfterFile(arg),
        acceptWsAfterParenOpen ? acceptWsAfterParenOpen(PERR_getLocationWsAfterParenOpen(arg)) : PERR_getLocationWsAfterParenOpen(arg),
        acceptFilename ? acceptFilename(PERR_getLocationFilename(arg)) : PERR_getLocationFilename(arg),
        acceptWsAfterFilename ? acceptWsAfterFilename(PERR_getLocationWsAfterFilename(arg)) : PERR_getLocationWsAfterFilename(arg));
  }
  if (PERR_isLocationArea(arg)) {
    return PERR_makeLocationArea(
        acceptWsAfterA ? acceptWsAfterA(PERR_getLocationWsAfterA(arg)) : PERR_getLocationWsAfterA(arg),
        acceptWsAfterParenOpen ? acceptWsAfterParenOpen(PERR_getLocationWsAfterParenOpen(arg)) : PERR_getLocationWsAfterParenOpen(arg),
        acceptArea ? acceptArea(PERR_getLocationArea(arg)) : PERR_getLocationArea(arg),
        acceptWsAfterArea ? acceptWsAfterArea(PERR_getLocationWsAfterArea(arg)) : PERR_getLocationWsAfterArea(arg));
  }
  if (PERR_isLocationAreaInFile(arg)) {
    return PERR_makeLocationAreaInFile(
        acceptWsAfterAreaInFile ? acceptWsAfterAreaInFile(PERR_getLocationWsAfterAreaInFile(arg)) : PERR_getLocationWsAfterAreaInFile(arg),
        acceptWsAfterParenOpen ? acceptWsAfterParenOpen(PERR_getLocationWsAfterParenOpen(arg)) : PERR_getLocationWsAfterParenOpen(arg),
        acceptFilename ? acceptFilename(PERR_getLocationFilename(arg)) : PERR_getLocationFilename(arg),
        acceptWsAfterFilename ? acceptWsAfterFilename(PERR_getLocationWsAfterFilename(arg)) : PERR_getLocationWsAfterFilename(arg),
        acceptWsAfterComma ? acceptWsAfterComma(PERR_getLocationWsAfterComma(arg)) : PERR_getLocationWsAfterComma(arg),
        acceptArea ? acceptArea(PERR_getLocationArea(arg)) : PERR_getLocationArea(arg),
        acceptWsAfterArea ? acceptWsAfterArea(PERR_getLocationWsAfterArea(arg)) : PERR_getLocationWsAfterArea(arg));
  }
  ATabort("not a Location: %t\n", arg);
  return (PERR_Location)NULL;
}
/**
 * Apply functions to the children of a PERR_Area. 
 * \return A new PERR_Area with new children where the argument functions might have applied
 */
PERR_Area PERR_visitArea(PERR_Area arg, PERR_OptLayout (*acceptWsAfterArea)(PERR_OptLayout), PERR_OptLayout (*acceptWsAfterParenOpen)(PERR_OptLayout), PERR_NatCon (*acceptBeginLine)(PERR_NatCon), PERR_OptLayout (*acceptWsAfterBeginLine)(PERR_OptLayout), PERR_OptLayout (*acceptWsAfterComma)(PERR_OptLayout), PERR_NatCon (*acceptBeginColumn)(PERR_NatCon), PERR_OptLayout (*acceptWsAfterBeginColumn)(PERR_OptLayout), PERR_OptLayout (*acceptWsAfterComma1)(PERR_OptLayout), PERR_NatCon (*acceptEndLine)(PERR_NatCon), PERR_OptLayout (*acceptWsAfterEndLine)(PERR_OptLayout), PERR_OptLayout (*acceptWsAfterComma2)(PERR_OptLayout), PERR_NatCon (*acceptEndColumn)(PERR_NatCon), PERR_OptLayout (*acceptWsAfterEndColumn)(PERR_OptLayout), PERR_OptLayout (*acceptWsAfterComma3)(PERR_OptLayout), PERR_NatCon (*acceptOffset)(PERR_NatCon), PERR_OptLayout (*acceptWsAfterOffset)(PERR_OptLayout), PERR_OptLayout (*acceptWsAfterComma4)(PERR_OptLayout), PERR_NatCon (*acceptLength)(PERR_NatCon), PERR_OptLayout (*acceptWsAfterLength)(PERR_OptLayout)) {
  if (PERR_isAreaArea(arg)) {
    return PERR_makeAreaArea(
        acceptWsAfterArea ? acceptWsAfterArea(PERR_getAreaWsAfterArea(arg)) : PERR_getAreaWsAfterArea(arg),
        acceptWsAfterParenOpen ? acceptWsAfterParenOpen(PERR_getAreaWsAfterParenOpen(arg)) : PERR_getAreaWsAfterParenOpen(arg),
        acceptBeginLine ? acceptBeginLine(PERR_getAreaBeginLine(arg)) : PERR_getAreaBeginLine(arg),
        acceptWsAfterBeginLine ? acceptWsAfterBeginLine(PERR_getAreaWsAfterBeginLine(arg)) : PERR_getAreaWsAfterBeginLine(arg),
        acceptWsAfterComma ? acceptWsAfterComma(PERR_getAreaWsAfterComma(arg)) : PERR_getAreaWsAfterComma(arg),
        acceptBeginColumn ? acceptBeginColumn(PERR_getAreaBeginColumn(arg)) : PERR_getAreaBeginColumn(arg),
        acceptWsAfterBeginColumn ? acceptWsAfterBeginColumn(PERR_getAreaWsAfterBeginColumn(arg)) : PERR_getAreaWsAfterBeginColumn(arg),
        acceptWsAfterComma1 ? acceptWsAfterComma1(PERR_getAreaWsAfterComma1(arg)) : PERR_getAreaWsAfterComma1(arg),
        acceptEndLine ? acceptEndLine(PERR_getAreaEndLine(arg)) : PERR_getAreaEndLine(arg),
        acceptWsAfterEndLine ? acceptWsAfterEndLine(PERR_getAreaWsAfterEndLine(arg)) : PERR_getAreaWsAfterEndLine(arg),
        acceptWsAfterComma2 ? acceptWsAfterComma2(PERR_getAreaWsAfterComma2(arg)) : PERR_getAreaWsAfterComma2(arg),
        acceptEndColumn ? acceptEndColumn(PERR_getAreaEndColumn(arg)) : PERR_getAreaEndColumn(arg),
        acceptWsAfterEndColumn ? acceptWsAfterEndColumn(PERR_getAreaWsAfterEndColumn(arg)) : PERR_getAreaWsAfterEndColumn(arg),
        acceptWsAfterComma3 ? acceptWsAfterComma3(PERR_getAreaWsAfterComma3(arg)) : PERR_getAreaWsAfterComma3(arg),
        acceptOffset ? acceptOffset(PERR_getAreaOffset(arg)) : PERR_getAreaOffset(arg),
        acceptWsAfterOffset ? acceptWsAfterOffset(PERR_getAreaWsAfterOffset(arg)) : PERR_getAreaWsAfterOffset(arg),
        acceptWsAfterComma4 ? acceptWsAfterComma4(PERR_getAreaWsAfterComma4(arg)) : PERR_getAreaWsAfterComma4(arg),
        acceptLength ? acceptLength(PERR_getAreaLength(arg)) : PERR_getAreaLength(arg),
        acceptWsAfterLength ? acceptWsAfterLength(PERR_getAreaWsAfterLength(arg)) : PERR_getAreaWsAfterLength(arg));
  }
  ATabort("not a Area: %t\n", arg);
  return (PERR_Area)NULL;
}
/**
 * Apply functions to the children of a PERR_Slice. 
 * \return A new PERR_Slice with new children where the argument functions might have applied
 */
PERR_Slice PERR_visitSlice(PERR_Slice arg, PERR_OptLayout (*acceptWsAfterSlice)(PERR_OptLayout), PERR_OptLayout (*acceptWsAfterParenOpen)(PERR_OptLayout), PERR_StrCon (*acceptId)(PERR_StrCon), PERR_OptLayout (*acceptWsAfterId)(PERR_OptLayout), PERR_OptLayout (*acceptWsAfterComma)(PERR_OptLayout), PERR_OptLayout (*acceptWsAfterBracketOpen)(PERR_OptLayout), PERR_AreaAreas (*acceptAreas)(PERR_AreaAreas), PERR_OptLayout (*acceptWsAfterAreas)(PERR_OptLayout), PERR_OptLayout (*acceptWsAfterBracketClose)(PERR_OptLayout)) {
  if (PERR_isSliceSlice(arg)) {
    return PERR_makeSliceSlice(
        acceptWsAfterSlice ? acceptWsAfterSlice(PERR_getSliceWsAfterSlice(arg)) : PERR_getSliceWsAfterSlice(arg),
        acceptWsAfterParenOpen ? acceptWsAfterParenOpen(PERR_getSliceWsAfterParenOpen(arg)) : PERR_getSliceWsAfterParenOpen(arg),
        acceptId ? acceptId(PERR_getSliceId(arg)) : PERR_getSliceId(arg),
        acceptWsAfterId ? acceptWsAfterId(PERR_getSliceWsAfterId(arg)) : PERR_getSliceWsAfterId(arg),
        acceptWsAfterComma ? acceptWsAfterComma(PERR_getSliceWsAfterComma(arg)) : PERR_getSliceWsAfterComma(arg),
        acceptWsAfterBracketOpen ? acceptWsAfterBracketOpen(PERR_getSliceWsAfterBracketOpen(arg)) : PERR_getSliceWsAfterBracketOpen(arg),
        acceptAreas ? acceptAreas(PERR_getSliceAreas(arg)) : PERR_getSliceAreas(arg),
        acceptWsAfterAreas ? acceptWsAfterAreas(PERR_getSliceWsAfterAreas(arg)) : PERR_getSliceWsAfterAreas(arg),
        acceptWsAfterBracketClose ? acceptWsAfterBracketClose(PERR_getSliceWsAfterBracketClose(arg)) : PERR_getSliceWsAfterBracketClose(arg));
  }
  ATabort("not a Slice: %t\n", arg);
  return (PERR_Slice)NULL;
}
/**
 * Apply functions to the children of a PERR_AreaAreas. 
 * \return A new PERR_AreaAreas with new children where the argument functions might have applied
 */
PERR_AreaAreas PERR_visitAreaAreas(PERR_AreaAreas arg, PERR_Area (*acceptHead)(PERR_Area), PERR_OptLayout (*acceptWsAfterHead)(PERR_OptLayout), PERR_OptLayout (*acceptWsAfterSep)(PERR_OptLayout)) {
  if (PERR_isAreaAreasEmpty(arg)) {
    return PERR_makeAreaAreasEmpty();
  }
  if (PERR_isAreaAreasSingle(arg)) {
    return PERR_makeAreaAreasSingle(
        acceptHead ? acceptHead(PERR_getAreaAreasHead(arg)) : PERR_getAreaAreasHead(arg));
  }
  if (PERR_isAreaAreasMany(arg)) {
    return PERR_makeAreaAreasMany(
        acceptHead ? acceptHead(PERR_getAreaAreasHead(arg)) : PERR_getAreaAreasHead(arg),
        acceptWsAfterHead ? acceptWsAfterHead(PERR_getAreaAreasWsAfterHead(arg)) : PERR_getAreaAreasWsAfterHead(arg),
        acceptWsAfterSep ? acceptWsAfterSep(PERR_getAreaAreasWsAfterSep(arg)) : PERR_getAreaAreasWsAfterSep(arg),
        PERR_visitAreaAreas(PERR_getAreaAreasTail(arg), acceptHead, acceptWsAfterHead, acceptWsAfterSep));
  }
  ATabort("not a AreaAreas: %t\n", arg);
  return (PERR_AreaAreas)NULL;
}

